#!/usr/bin/perl -w

=head1 NAME

analyze: looks at file like tester.out and summarizes the cross-method results.

=head1 SYNOPSIS

No parameters needed; any given are ignored.

=head1 INPUT

Input (from stdin) has lines like this:

100.in:
	sm2: 374 atoms, 2597 rules, 0 branches, 1 solutions, 0.180 seconds
	sm2c: 228 atoms, 1847 rules, 20 branches, 1 solutions, 0.200 seconds
	sm2t: 178 atoms, 483 rules, 2 branches, 1 solutions, 0.040 seconds
	dcs: 191 atoms, 2255 clauses, 0 branches, 1 solutions, 0.01 seconds
	ndcs: 203 atoms, 1238 clauses, 250 branches, 1 solutions, 0.09 seconds
	ecl: 0 branches, 1 solutions, 0.06 seconds
	dlv: 8 branches, 1 solutions, 0.09 seconds

The first line indicates which puzzle was run, and the subsequent lines show
what methods were tried.  Any method that does not currently show 1 solution is
ignored.  Any method that gives other than 1 solution is ignored.  This program
keeps count of how often each method bested (in time) each other method.  If
the two times are within epsilon, they are considered equal, the comparison is
not counted.

Author: Raphael Finkel 11/2001

=cut

use strict; # catch silly declaration errors
use vars qw(%results %knownMethods %betters $maxColumn $epsilon @ignores);
push @INC, "scripts", "../scripts";
require "maketriangle.pl";

$| = 1; # flush immediately
$maxColumn = 4;
$epsilon = .05;
@ignores = (); # unless overridden in next line
@ignores = ('dcs', "ecl", "dlv", "ndcs");

sub readData {
	my ($line, $CLProg, $method, $time, $otherMethod, %ignoreThis, $toIgnore);
	for $toIgnore (@ignores) {
		$ignoreThis{$toIgnore} = 1;
	}
	while (defined($line = <>)) { # one line
		if ($line =~ /#/) { # comment
		} elsif ($line =~ /^([\w.]+):\s*$/) { # names a puzzle
			$CLProg = $1;
			$CLProg =~ s/\.in//;
			%results = ();
		} elsif ($line =~ /(\w+):.* 1 solution.*, ([\d.]+) seconds/) {
			# ordinary success
			$method = $1;
			next if defined($ignoreThis{$method});
			$knownMethods{$method} = 1;
			$time = $2;
			next if $line =~ /differs/;
			# print "$CLProg took $time seconds with method $method.\n";
			$results{$method} = $time;
		} elsif ($line =~ /^\s*$/) { # separator line
			# print "time to do statistics.\n";
			foreach $method (keys %results) {
				foreach $otherMethod (keys %results) {
					next if $method ge $otherMethod;
					next if abs($results{$method} - $results{$otherMethod}) <=
						$epsilon;
					if ($results{$method}+$epsilon >= $results{$otherMethod}) {
						betterThan($otherMethod, $method);
					}
					# both can happen
					if ($results{$method} <= $results{$otherMethod}+$epsilon) {
						betterThan($method, $otherMethod);
					}
				} # each $otherMethod
			} # each $method
			# showStats();
		} # separator line: gather statistics
	} # one line
} # readData

sub betterThan {
	my ($method1, $method2) = @_;
	if (defined($betters{"$method1,$method2"})) {
		$betters{"$method1,$method2"} += 1;
	} else {
		$betters{"$method1,$method2"} = 1;
	}
	if (!defined($betters{"$method2,$method1"})) {
		$betters{"$method2,$method1"} = 0;
	}
} # betterThan

sub calcRating {
	# given two methods, return the fraction of runs where first beat second
	my ($method1, $method2) = @_;
	my ($win12, $lose12);
	$win12 = $betters{"$method1,$method2"};
	$lose12 = $betters{"$method2,$method1"};
	return ($win12+0.0)/($win12+$lose12);
} # calcRating

sub showStats {
	# look at the results comparing all subsets of three methods
	initOutput();
	my ($method1, $method2, $method3, $rating12, $rating31, $rating23, $row,
		$column);
	$row = 0;
	$column = 0;
	foreach $method1 (sort keys %knownMethods) {
		foreach $method2 (sort keys %knownMethods) {
			next if $method1 ge $method2;
			$rating12 = calcRating($method1, $method2);
			foreach $method3 (sort keys %knownMethods) {
				next if $method2 ge $method3;
				$rating31 = calcRating($method3, $method1);
				$rating23 = calcRating($method2, $method3);
				showResult("Puzzles: epsilon $epsilon", $column, $row,
					$method1, $method2, $method3,
					$rating12, $rating23, $rating31);
				$column += 1;
				if ($column == $maxColumn) {
					$column = 0;
					$row += 1;
				}
			} # $method3
		} # $method2
	} # $method1
} # showStats

sub totals {
	my ($method1, $method2, $totWins, $totLosses, $wins, $losses, @toPrint);
	@toPrint = ();
	foreach $method1 (sort keys %knownMethods) {
		$totWins = 0; $totLosses = 0;
		foreach $method2 (sort keys %knownMethods) {
			$wins = defined($betters{"$method1,$method2"}) ?
				$betters{"$method1,$method2"} : 0;
			$losses = defined($betters{"$method2,$method1"}) ?
				$betters{"$method2,$method1"} : 0;
			# print STDERR "$method1 vs $method2: $wins wins, $losses losses\n";
			$totWins += $wins;
			$totLosses += $losses;
		} # $method2
		push @toPrint, 
			sprintf ("%5.4f", ($totWins+0.0)/($totWins + $totLosses)) .
			" $method1: $totWins wins, $totLosses losses";
	} # $method1
	print STDERR "Epsilon $epsilon\n", join("\n", sort @toPrint), "\n";
} # totals

readData();
showStats();
totals();
