#!/usr/local/bin/perl -w


=head1 SYNOPSIS

A Perl module exporting these routines:

  Error($msg): Reports an error (typically a parse error) and dies.
  Warn($msg): Reports an error (typically a semantic error) and continues.
  resolve($dottedMember, $line): resolves into ($class, $member).
  Append($orig, $addendum, $separator)
  Prepend($prefix,$string)
  mainLoop(): calls relevant routines, parsing Constraint Lingo input
    invokes (not in this file)
      handleDeclaration($line, CLASS|PARTITION, $name, $members);
      handleVar($line, $name);
      handleDiffer($line, SAME|DIFFER, $class, $member1, $member2);
      handleOffset($line, $arithmetic, $amount, $class, $member1,
        $member2);
      handleBefore($line, $class, $member1, $member2);
      handleMatch($line, $list1, $list2);
      handleUsedOp($line, $lower, $upper, $element);
      handleUsed($line, $element);
      handleRequired($line, $member, $list);
      handleAgree($line, $member, $list);
      handleConflict($line, $list);
      handleRequiredOp($line, $member1, $member2, $op, $member3,
        $member4);
      handlePass($line, $backend, $contents);
      handleVar($line, $varname);

=head1 AUTHOR

  Raphael Finkel <raphael@cs.uky.edu> 10/2001

=cut

require 'lex.pl';

# use strict;
use vars qw (%Classes %Maps %KindOf %Members %circular $low $high %highValue 
				$line %lowValue %underspecified $classSize @contents $member 
				$idClass $method $level %headers $needVar $needSequence
				$Parent $Class @tokens);

$needVar = 0;
sub NextVar {
	$needVar += 1;
	return "v$needVar";
}

$needSequence = 0;
sub NextAtom {
	$needSequence += 1;
	return "need$needSequence";
}
	
sub Error {
	my ($message) = @_;
	print STDERR $message;
	print STDERR "\n" unless $message =~ /\n$/;
	exit(-1);
} # Error

sub Warn {
	my ($message) = @_;
	print STDERR "Warning: ", $message;
	print STDERR "\n" unless $message =~ /\n$/;
} # Warn

sub resolve { # return (class, member) resolving any dotted notation
	my ($dottedMember, $line) = @_;
	# print "% dottedMember $dottedMember \n";
	my ($answerClass, $answerMember);
	if ($dottedMember =~ /^(\w+)\.(\w+)/) { # dotted
		$answerClass = lc $1;
		$answerMember = lc $2;
		Error("$answerMember is not a member of $answerClass: $line\n")
			unless ($Classes{$answerMember} =~ /\b$answerClass\b/);
	} elsif ($dottedMember !~ /\./) { # undotted
		$answerMember = lc $dottedMember;
		$answerClass = $Classes{$answerMember};
		Error("Undefined element: $dottedMember in $line\n") 
			unless defined($answerClass);
		# Error("$answerMember ambiguous; could be " .
		#			join(" or ", split(/ /, $answerClass)) .
		#			": $line\n") if ($answerClass =~ / /);
	} else {
		Error("Unrecognizable element: $dottedMember in $line\n");
	}
	return($answerClass, $answerMember);
} # resolve

# Append($orig, $addendum, $separator)
# Returns the $orig with the $addendum at the end, separated by $separator
# If $orig is empty or undefined, no separator is used.
sub Append {
	my ($orig, $addendum, $separator) = @_;
	if (!defined($orig) or $orig eq "") {
		return($addendum);
	} elsif ($addendum eq "") {
		return($orig);
	} else {
		return($orig . $separator . $addendum);
	}
} # Append

# Prepend($prefix,$string)
# If $string is null or undefined, return ""; otherwise return $prefix.$string
sub Prepend {
	my ($prefix,$string) = @_;
	if (!defined($string) or $string eq "") {
		return("");
	} else {
		return($prefix . $string);
	}
} # Prepend

sub mainLoop {

	# This is not the only way to call the parser more than once
	# but this works good enough.
	my $counter = 0;
	while(defined($tokens[$counter])) {
		${"Parse::YYLex::".$tokens[$counter]} = undef 
			if(defined(${"Parse::YYLex::". $tokens[$counter]}));
		$counter += 3;
	}
	my $caller = caller;
	# *{"${caller}::yylex"} = undef if(defined(*{"${caller}::yylex"}));
	# print STDERR "******* $caller\n";

	# Parse::Lex->trace;
	my $lexer = new Parse::YYLex();
	$lexer->create(@tokens);

	yyparse();
}


# Routines for the new syntax

# func(member1) = member2.
# func(member1) != member2.
sub handleAssignment {
	my ($line, $class, $member1, $sign, $member2) = @_;
	my ($class2);

	if( $sign eq "=" ) {
		if($KindOf{$class} eq "CLASS") {
			if(!defined($Maps{$class})) { # class is not a mapping
				handleRequired($line, $member1, $member2); 
			} else {
				handleMapRequired($line, $class, $member1, $member2);
			}
		} else {
			if($Classes{$member2} eq $class) {
				handleAgree($line, $member2, $member1) ;
			} else {
				handleMapRequired($line, $class, $member1, $member2);
			}
		}
	} else {
		if($KindOf{$class} eq "CLASS") {
			if(!defined($Maps{$class})) { # class is not a mapping
				handleConflict($line, join(" ",($member1, $member2)));
			} else {
				handleMapConflict($line, $class, $member1, $member2);
			}
		} else {
			if($Classes{$member2} eq $class) {
				handleConflict($line, join(" ",($member2, $member1)));
			} else {
				handleMapConflict($line, $class, $member1, $member2);
			}
		}
	}
}

# func1(member1) = func2(member2)
# func1(member1) != func2(member2)
sub handleMap {
	my ($line, $class1, $member1, $sign, $class2, $member2) = @_;

	Error "$class1 is not a defined: $line" if(!defined($Members{$class1}));
	Error "$class2 is not a defined: $line" if(!defined($Members{$class2}));

	# Should we check for the type of the return value of function2?
	if($class1 eq $class2) {
		if($sign ne "!=") {
			handleSame($line, "SAME", $class1, $member1, $member2);
		} else {
			handleDiffer($line, "DIFFER", $class1, $member1, $member2);
		}
	} else {
		if($sign ne "!=") {
			print "\n\n\t%% $line\n";
			my $newVar = $member2;

			# handle the following case
			# class1(member1) = map(member2)
			# otherwise it's a normal assignment
			if(defined($Maps{$class2})) { # class is a mapping
				$newVar = NextVar;
				handleVar("VAR ".$newVar, $newVar);
				handleAssignment("SET $class2($member2)=$newVar", 
										$class2, $member2, "=", $newVar);
			}								
			handleAssignment("SET $class1($member1)=$newVar", 
									$class1, $member1, "=", $newVar);
			
		} else {
			my $newVar = $member2;
			
			# handle the following case
			# class1(member1) != map(member2)
			# otherwise it's a normal assignment
			if(defined($Maps{$class2})) { # class is a mapping
				$newVar = NextVar;
				handleVar("VAR ".$newVar, $newVar);
				handleAssignment("SET $class2($member2)=$newVar", 
										$class2, $member2, "=", $newVar);
			}								
			handleAssignment("SET $class1($member1)=$newVar", 
									$class1, $member1, "!=", $newVar);

		}
	}
}

sub handleMapOffset {
	my($line,$type,@args) = @_;
	
	if($type == 0) {
		my($class1,$var1,$arithmetic,$amount, $class2,$var2) = @args;

		# traditional offset
		if(!defined($Maps{$class1}) and $class1 eq $class2) { 
			handleOffset($line, $arithmetic, $amount, $class1, $var1, $var2);
		} elsif($class1 ne $class2) {
			handleOffsetPair($line,$arithmetic,$amount,$class1,$var1,$class2,$var2);
		}
	}
}

sub handleFunction {
	my ($line) = @_;
	my ($name, undef, $var, undef) = split(' ', $line);
	
	my $newVar = NextVar;
	handleVar("VAR ".$newVar, $newVar);
	handleMapRequired("SET ".$line." = ".$newVar, $name, $var, $newVar);
	return $newVar;
}
	
sub handleSame {
	my ($line, $kind, $class, $member1, $member2) = @_;
	handleDiffer($line, $kind, $class, $member1, $member2);
}
1;
