=head SYNOPSYS

Subroutines for converting cross-class predicate atoms, like
"child_dog(bill,fido)", into English like "Bill owns Fido".

Author: Raphael Finkel

1/2003

=cut

use vars ( # global variables
	'%classInfo', # string with comma-separated fields for generating English
	'%belongsTo', # a class belongs to its parent class
	'%classSerial', # every class has a serial number; lower prefers being a
		# subject
	'$serial', # running serial number.
	'%respell', # respelling rules
	'%replace', # replacement rules
	);

# initialized defaults
	$classInfo{'variable'} = ",%m,is,isn't,associated with variable %c,";
	$classInfo{'default'} = ",%m,is,isn't,associated with %m,";

sub assimilatePuzzle { # see if the puzzle includes precedence information
	my ($file, $line, $precedence);
	$precedence = ""; # until we learn otherwise
	if (defined($ARGV[0])) {
		$file = $ARGV[0];
		open(PUZZLE, "<$file") or
			die "Cannot open Constraint Lingo program $file for reading";
		while (defined($line = <PUZZLE>)) { # one line
			chomp $line;
			if ($line =~ /^#%\s+(.*)$/) { # found precedence.
				$precedence .= $1 . " ";
			}
		} # one line
		close PUZZLE;
	} # we have a puzzle file to look at.
	# print "precedence information: $precedence\n";
	$serial = 1;
	parsePrecedence("global", $precedence);
} # assimilatePuzzle

sub parsePrecedence {
	my ($parent, $precedence) = @_;
	my ($children, $someChildren, $child, $grandchild);
	if ($precedence =~ s/\s*([^()]+)(\(.*)?//) { # parse precedence
		$children = $1;
		$grandchild = $2;
		$children =~ s/\s*$//;
		# print "I see children $children\n";
		$someChildren = $children;
		while ($someChildren =~ s/\s*(\w+)\s*//) {
			$child = $1;
			$belongsTo{$child} = $parent;
			$classSerial{$child} = $serial;
			$serial += 1;
			# print "$child belongs to $parent\n";
			if ($someChildren =~ s/^"([^"]+)"\s*//) { # class info
				$classInfo{$child} = $1;
				# print "New class info for $child: $1\n";
			} else {
				print STDERR "Need class information for $child\n";
				exit(0);
			}
			while ($someChildren =~ s/^"([^"]+)"\s*//) { # respellings
				my $specification = $1;
				if ($specification =~ /=/) { # respelling
					my ($source, $respelling) =
						split(/\s*=\s*/, $specification);
					$respell{$source} = $respelling;
					# print "respelling $source as $respelling\n";
				} elsif ($specification =~ /^!(.*)/) { # omission; deprecated
					$replace{$1} = "";
					# print "need to avoid saying $1\n";
				} elsif ($specification =~ />/) { # replacement
					my ($source, $dest) = split(/\s*>\s*/, $specification);
					$replace{$source} = $dest;
				}
			}
		}
	}
	if (defined($grandchild)) {
		$grandchild =~ s/^\(//;
		$grandchild =~ s/\s*$//;
		$grandchild =~ s/\)$//;
		parsePrecedence($child, $grandchild);
	}
} # parsePrecedence

sub tellPrecedence {
	my $child;
	foreach $child (keys %belongsTo) {
		print "$child belongs to $belongsTo{$child}.\n";
	}
} # tellPrecedence

sub plugIn { # given a printf-style pattern, replace %m and %c (or caps)
	my ($pattern, $member, $class) = @_;
	# print "pattern $pattern member $member class $class\n";
	$member = $respell{$member} if (defined($respell{$member}));
	$pattern =~ s/%m/$member/g;
	$pattern =~ s/%M/ucfirst $member/eg;
	$pattern =~ s/%c/$class/g;
	$pattern =~ s/%C/ucfirst $class/eg;
	return($pattern);
} # plugIn

# subscripts into a classInfo array
use vars qw
	/$Equivalance $Subject $VerbPositive $VerbNegative $Predicate $Relative/;
$Equivalance = 0;
$Subject = 1;
$VerbPositive = 2;
$VerbNegative = 3;
$Predicate = 4;
$Relative = 5;

sub findPath { # determine the path of classes between $class1 and $class2
	# return (ancestor, path to class1, path to class2).
	# paths are : separated strings.
	my ($class1, $class2) = @_;
	# print "path between $class1 and $class2: ";
	my (@path1, @path2, $ancestor, @fullPath);
	$belongsTo{$class1} = "global" unless defined($belongsTo{$class1});
	$belongsTo{$class2} = "global" unless defined($belongsTo{$class2});
	while ($class1 ne "global") {
		push @path1, "$class1";
		$class1 = $belongsTo{$class1};
	}
	while ($class2 ne "global") {
		push @path2, $class2;
		$class2 = $belongsTo{$class2};
	}
	@path1 = reverse(@path1);
	@path2 = reverse(@path2);
	while ($#path1 >= 0 and $#path2 >= 0 and $path1[0] eq $path2[0]) {
		# remove common parts
		$ancestor = $path1[0];
		shift @path1;
		shift @path2;
	}
	if (!defined($ancestor)) {
		if (defined($path1[0])) {
			$ancestor = shift @path1;
		} else {
			$ancestor = shift @path2;
		}
	}
	# handle equivalences.
	# print "\nstarting equiv check: $ancestor|", join(":", @path1), "|",
	#	join(":", @path2);
	my @infoAncestor = split(',', $classInfo{$ancestor});
	if ($#path1 > 0 && $infoAncestor[$Equivalance] eq $path1[0]) {
		# print "\n$ancestor derives $path1[0]; deleting";
		shift @path1;
	} elsif ($#path2 > 0 && $infoAncestor[$Equivalance] eq $path2[0]) {
		# print "\n$ancestor derives $path2[0]; deleting";
		shift @path2;
	} elsif ($#path1 != -1 && $#path2 != -1) {
		my @infoPath1 = split(',', $classInfo{$path1[0]});
		my @infoPath2 = split(',', $classInfo{$path2[0]});
		if ($infoPath1[$Equivalance] eq $ancestor) {
			# print "\n$ancestor is derived from $path1[0]; rearranging";
			$ancestor = shift @path1;
		} elsif ($infoPath2[$Equivalance] eq $ancestor) {
			# print "\n$ancestor is derived from $path2[0]; rearranging";
			$ancestor = shift @path2;
		}
	}
	return($ancestor, join(":", @path1), join(":", @path2));
} # findPath

sub reword {
	# takes an atom (a cross-class predicate) and a truth value and returns a
	# more readable representation of that information.
	my ($atom, $truth) = @_;
	my ($firstPart, $secondPart, $answer);
	# print "\nrewording $atom $truth";
	$atom =~ /([a-z0-9]+)_([a-z0-9]+)\((\w+)(,(\w+))?\)/;
	my ($class1, $class2, $member1, $member2) = ($1, $2, $3, $5);
	if (!defined($member2)) { # result of VAR
		my @infoVar;
		if (!defined($classInfo{$class2})) {
			print "\nBad specification: The puzzle gives no information on " .
				"variable $class2.\n\n";
			$classInfo{$class2} = $classInfo{'variable'};
		}
		@infoVar = split(',', $classInfo{$class2});
		$answer = plugIn($infoVar[$Subject], $member1, $class2) . " " .
			$infoVar[$truth eq "true" ? $VerbPositive : $VerbNegative] .  " " .
			plugIn($infoVar[$Predicate], $member1, $class2);
		return($answer);
	}
	return "unintelligible information: $atom" unless defined($4);
	my ($ancestor, $path1, $path2) = findPath($class1, $class2);
	# print "\nancestor=$ancestor, path1=$path1, path2=$path2";
	# interchange 1 and 2 if necessary
	my ($path1Start, $path2Start);
	$path1Start = $1 if $path1 =~ /^([^:]+)/;
	$path2Start = $1 if $path2 =~ /^([^:]+)/;
	if (($path1 eq "" and $classSerial{$ancestor} > $classSerial{$path2Start})
			or
		($path2 eq "" and $classSerial{$ancestor} < $classSerial{$path1Start})
			or
		($path1 ne "" and $path2 ne "" and $classSerial{$path1Start} >
			$classSerial{$path2Start})
			){
		# print "\nInterchanging.";
		($path1, $path2) = ($path2, $path1);
		($class1, $member1, $class2, $member2) =
			($class2, $member2, $class1, $member1);
	};
	my @infoAncestor = split(',', $classInfo{$ancestor});
	$answer = "";
	if ($path1 eq "") {
		# print "\n plugging in $infoAncestor[$Subject], $member1, $class1";
		$answer .= plugIn($infoAncestor[$Subject], $member1, $class1) . " ";
	} else { # recite the first path
		$answer .= plugIn($infoAncestor[$Relative],"",$ancestor) . " ";
		my $notFirstTime;
		while ($path1 =~ s/^([^:]+)(:)?//) { # a single component
			my ($component, $more) = ($1, $2);
			my (@infoComponent) = split(',', $classInfo{$component});
			$answer .= $infoComponent[$VerbPositive] . " ";
			if (defined($more)) {
				$answer .= plugIn($infoComponent[$Relative], "", $component) .
					" ";
			} else {
				$answer .=
					plugIn($infoComponent[$Predicate], $member1, $component) .
					" ";
			}
			$notFirstTime = 1;
		} # a single component
	} # recite the first path.
	if ($path2 eq "") {
		$answer .= $infoAncestor[$truth eq 'true' ? 2 : 3] . " ";
		$answer .= plugIn($infoAncestor[$Predicate], $member2, $class2);
	} else { # recite the second path
		my ($notFirstTime);
		while ($path2 =~ s/^([^:]+)(:)?//) { # a single component
			my ($component, $more) = ($1, $2);
			my @infoComponent = split(',', $classInfo{$component});
			if (!defined($notFirstTime)) { # first time; put in main verb
				$answer .= $infoComponent[$truth eq 'true' ? 2 : 3] . " ";
			}
			if (defined($more)) {
				$answer .= plugIn($infoComponent[$Relative], "", $component) .
					" ";
			} else {
				$answer .= $infoComponent[$VerbPositive] . " "
					if defined($notFirstTime);
				$answer .=
					plugIn($infoComponent[$Predicate], $member2, $component) .
					" ";
			}
			$notFirstTime = 1;
		} # a single component
	} # recite the second path
	$answer =~ s/that is (?!a\b)//g;
	foreach my $pattern (keys %replace) {
		$answer =~ s/$pattern\s*/$replace{$pattern}/g;
	}
	$answer =~ s/\s*$//;
	return $answer;
	my (@info1, @info2);
	if ($belongsTo{$class1} ne "global" and $belongsTo{$class2} ne "global") {
		my $suggestion = (split /\s+/, $belongsTo{$class1})[0];
		my @info = split(',',
			defined($classInfo{$suggestion}) ? $classInfo{$suggestion} :
			$classInfo{'default'});
		my $genericName =
			(defined($info[$Relative])) ? $info[$Relative] : $suggestion;
		return("* The $genericName that $info1[$VerbPositive] " .
			plugIn($info1[$Predicate], $member1, $class1) .
			" $info2[$VerbPositive] " .
			plugIn($info2[$Predicate], $member2, $class2));
	}	
	if ($info1[0] < $info2[0]) { # reverse the order
		($class1, $member1, $class2, $member2) =
			($class2, $member2, $class1, $member1);
		my @tmp = @info1;
		@info1 = @info2;
		@info2 = @tmp;
	}
	$firstPart = plugIn($info1[$Subject], $member1, $class1);
	$secondPart = plugIn($info2[$Predicate], $member2, $class2);
	return "$firstPart " . ($truth eq 'true' ? $info2[$VerbPositive] : $info2[$VerbNegative]) .
		" $secondPart";
} # reword

1
