#!/usr/bin/perl -w

=head1 DOC

This program takes a logic puzzle and converts it to Eclipse (Prolog with
constraints).  The idea is that every line of the solution is "equal" to an
integer in the range 1..n.  If "john" appears on line 3, then the variable John
should have the value 3.

Status:  Working on it.

Author:  Raphael Finkel <raphael@cs.uky.edu> 5/2001

=cut

use IO::Handle; # intersperse errors directly in output
push @INC, "scripts", "../scripts";
require 'puzzle.common.pl';

sub myResolve { # return (class, member) resolving any dotted notation
	my ($dottedMember, $line) = @_;
	my ($answerClass, $answerMember);
	if ($dottedMember =~ /^(\w+)\.(\w+)/) { # dotted
		$answerClass = lc $1;
		$answerMember = lc $2;
		Error("$answerMember is not a member of $answerClass: $line\n")
			unless ($Classes{$answerMember} =~ /\b$answerClass\b/);
		$answerMember = (ucfirst $answerClass) . $answerMember
			if ($answerMember =~ /^\d+$/) and
				($numericId == 0 or $answerClass ne $idClass);
	} elsif ($dottedMember !~ /\./) { # undotted
		$answerMember = ucfirst (lc $dottedMember);
		$answerClass = $Classes{$answerMember};
		$answerMember = (ucfirst $answerClass) . $answerMember
		if ($answerMember =~ /^\d+$/) and defined($idClass) and
				($numericId == 0 or $answerClass ne $idClass);
			Error("Undefined element: $dottedMember in $line\n") 
				unless defined($answerClass);
			Error("$answerMember ambiguous; could be " .
				join(" or ", split(/ /, $answerClass)) .
				": $line\n") if ($answerClass =~ / /);
	} else {
		Error("Unrecognizable element: $dottedMember in $line\n");
	}
	return($answerClass, $answerMember);
} # myResolve

sub finishDecls {
	return if defined($declsFinished);
	print "% finishing declarations\n";
	$declsFinished = 1;
	if (defined($idClass)) {
		$numericId = 1;
		@rowList = @{$Members{$idClass}};
	} else { # no numeric idclass; pick idClass
		$numericId = 0;
		foreach $class (keys %Members) {
			next if(defined($underspecified{$class}));
			if ($KindOf{$class} eq "CLASS") { # got a good idClass
				$idClass = $class;
				@rowList = 1 .. $#{$Members{$idClass}}+1;
				print "% Break symmetry\n";
				$counter = 1;
				foreach $member (sort @{$Members{$class}}) {
					print "\t$member = $counter,\n";
					$idRow{$member} = $counter;
					$counter += 1;
				}
				last;
			} # got a good idClass
		} # foreach $class
	} # no numeric idclass
	$standardCount = $#{$Members{$idClass}} + 1;
	foreach $class (keys %Members) {
		if ($KindOf{$class} eq "CLASS") {
			if ($numericId) {
				next if $class eq $idClass;
				$Class = ucfirst $class;
				$excess = $#{$Members{$class}} + 1 - $standardCount;
				if ($excess > 0) {
					print "\t% underspecified range: CLASS $class has $excess ",
						"extra elements.\n";
					$extra = ", " . join(", ",
					${$Members{$idClass}}[$standardCount-1] + 1 ..
					${$Members{$idClass}}[$standardCount-1] + $excess);
				} else {
					$extra = "";
				}
				print "\t$Class :: [", join(', ', @{$Members{$idClass}}),
					"$extra],\n", "\talldifferent($Class),\n";
				# no problem for underspecified range, but it better not be the
				# first one we see.
			} else { # no numeric class
				$Class = ucfirst $class;
				print "\t$Class :: 1 .. ", $#{$Members{$class}}+1,
					", alldifferent($Class),\n";
			} # foreach $member
		} # ordinary CLASS
	} # foreach $class
	print "% rowlist is ", join(', ', @rowList), "\n";
	if (defined(@defer)) {
		foreach $line (@defer) {
			print "\t% deferred $line\n";
			if ($line =~ /(REQUIRED|AGREE)\s+([\w.-]+)\s*:\s+(.*)$/) {
				$member1 = lc $2;
				($class, $member1) = myResolve($member1, $line);
				$Class = ucfirst $class;
				foreach $member2 (split /\s+/, $3) {
					($class2, $member2) = myResolve($member2, $line);
					$Class2 = ucfirst $class2;
					$member2 = ucfirst($member2);
					if ($class2 eq $idClass) {
						$member2 = defined($idRow{$member2}) ? 
										$idRow{$member2} : 
										$member2;
						print "\t${Class}${member2} #= $member1,\n";
					} else {
						@toSay = ();
						foreach $index (@rowList) {
							if ($KindOf{$class2} eq "PARTITION") {
								$member1 =~ s/$Class/""/;
								print STDERR $member1, "\n";
								print "\t${Class2}$index #= $member2 #=> ".
									"${Class}$index #= $member1,\n";
							} else {
								push @toSay, "($member2 #= $index #=> " .
									"${Class}$index #= $member1)";
							}
						}
						print "\t(", join(" #/\\\n\t", @toSay), "),\n"
						unless $#toSay == -1;
					}
				} # foreach $member2;
			} elsif ($line =~ /^\s*PARTITION\s+([\w-]+)/) {
				$class = $1;
				$Class = ucfirst $class;
				@toSay = ();
				for $index (@rowList) {
					push @toSay, $Class . $index;
				}
				print "\t$Class = \[", join(', ', @toSay), "\],\n",
					"\t$Class :: \[", join(', ', @{$Members{$class}}), 
					"\],\n";
			}
		} # foreach deferred line
	} # there were deferred lines
	print "% declarations finished\n\n";
} # finishDecls

print "output is redirected to /tmp/$ENV{USER}.ecl.tmp\n";
open STDOUT, ">/tmp/$ENV{USER}.ecl.tmp";
autoflush STDOUT 1;
autoflush STDERR 1;
print ":- use_module(library(fd)) .\n", # constraint-propagation library
    ":- use_module(library(fd_search)) .\n", # another library
    ":- use_module(library(fd_global)) .\n", # another library
    ":- import occurrences/3 from fd_global.\n",
    ":- import alldifferent/1 from fd_global.\n\n",
    "all(Limit) :- cputime(Start), findall([], run(Limit), _), ",
      "cputime(Finish),\n",
    "\twrite(stdout, \'\\nTime = \'), Total is Finish - Start,\n",
    "\twrite(stdout, Total), write(stdout, \'\\n\') .\n",
    "a :- ['eclipse.tmp'] , all .\n\n", # for manual reload convenience
    "run(Limit) :- setval(counter, Limit),\n"; # start of main rule

open CLFILE, "<&STDIN";
mainLoop();
close CLFILE;
emitStandard();
    
sub handleDeclaration {
	my ($line, $kind, $class, $contents) = @_;
	Error("Too late to make new declaration: $line\n")
		if defined($declsFinished);
	Error("$class is both a class and a member\n")
		if defined($Classes{ucfirst $class});
	$KindOf{$class} = $kind;
	if (defined($Members{$class})) {
		Error("$class is already a defined); you can't define it again.\n");
	}
	$Members{$class} = []; # reference to initially empty array
	$circular{$class} = 0; # unless proven otherwise
	if ($contents =~ /^\s*(\d+)\s*\.\.\s*(\d+)\s*(circular)?\s*$/) { # range
		$low = $1;
		$high = $2;
		$circular{$class} = defined($3);
		if ($circular{$class}) {
			if ($low != 1) {
				Error("Bottom of circular range must be 1: $line\n");
			}
			$lowValue{$class} = $low;
			$highValue{$class} = $high;
			$dummy = $lowValue{$class} + $highValue{$class}; # avoid warning
		}
		@contents = ();
		while ($low <= $high) {
			push @contents, $low;
			$low += 1;
		}
		$contents = join " ", @contents;
	} # range
	$count = scalar(split(/\s+/, $contents));
	if(defined($totalCount)) {
		if($totalCount != $count) {
			print STDERR "Class $class has $count elements, not $totalCount.\n";
			$underspecified{$class} = 1;
		}
	} else {
		$totalCount = $count;
	} 
	foreach $member (split /\s+/, $contents) {
		Error("$member is both a class and a member\n")
			if defined($KindOf{$member});
		$member = ucfirst $member;
		if (defined($Classes{$member})) {
			# print STDERR
			#   "$member is already a member of $Classes{$member}; " ,
			#   "it is now also a member of $class.\n";
			$Classes{$member} .= " $class";
		} else {
			$Classes{$member} = $class;
		}
		$member = (ucfirst $class) . $member
			if $member =~ /^\d+$/ and defined($idClass) and $class ne $idClass; 
		$member = (ucfirst $class) . $member
			if $member =~ /^\d+$/ and defined($underspecified{$class});
		push @{$Members{$class}}, $member;
	} # foreach $member

	$idClass = $class
		unless $kind eq "PARTITION" or defined($idClass) or
			$contents =~ /[^\d\s]/ or defined($underspecified{$class});

	$Class = ucfirst $class;
	print "% $line\n";
	if (defined($idClass) and $class eq $idClass) {
		print "\t% We translate this class as a row identifier\n";
		$fullCount = $#{$Members{$class}}+1 unless defined($fullCount);
	}
	if ($kind eq "CLASS") {
		if (!defined($idClass) or $class ne $idClass) {
			print "\t$Class = \[", join(', ', @{$Members{$class}}), "\],\n";
			$fullCount = $#{$Members{$class}}+1 unless defined($fullCount);
		}
	} else { # PARTITION
		foreach $member (@{$Members{$class}}) {
			print "\t$member = \'$member\',\n";
		}
		if (defined($declsFinished)) { # emit now
			@toSay = ();
			for $index (@rowList) {
				push @toSay, $Class . $index;
			}
			print "\t$Class = \[", join(', ', @toSay), "\],\n",
				"\t$Class :: \[", join(', ', @{$Members{$class}}), "\],\n";
		} else {
			push @defer, $line;
			print "\t% deferring some output until declarations done.\n";
		}
	} # PARTITION
} # handleDeclaration

sub handleVar {
	my ($line, $member) = @_;
	$member = ucfirst $member;
	finishDecls();
	if (defined($Classes{$member})) {
		Error("$member is already a member of $Classes{$member}: $line\n");
	}
	$Classes{$member} = $member;
	$KindOf{$member} = "INDIVIDUAL";
	print "% $line\n", "\t$member :: [", join(", ", @rowList), "],\n";
} # handleVar

sub handlePass {
	my($line)=@_;
	$line =~ /^\s*PASS\s+ecl\b(.*)/ or return;
	print RULES "% pass through\n\t$1\n";
} # handlePass

sub handleRequiredOp {
	my ($line, $member1, $member2, $combiner, $member3, $member4) = @_;  
	finishDecls();
	($class1, $member1) = myResolve($member1, $line);
	($class2, $member2) = myResolve($member2, $line);
	($class3, $member3) = myResolve($member3, $line);
	($class4, $member4) = myResolve($member4, $line);
	$joiner = ($combiner eq "OR") ? "#\\/" :
	($combiner eq "IFF") ? "#<=>" : "XOR";
	print "% $line\n";
	if ($KindOf{$class1} eq "PARTITION" and $class2 eq $idClass) {
		$Class1 = ucfirst $class1;
		$first = "${Class1}$member2 #= $member1";
	} elsif ($KindOf{$class2} eq "PARTITION" and $class1 eq $idClass) {
		$Class2 = ucfirst $class2;
		$first = "${Class2}$member1 #= $member2";
	} elsif ($KindOf{$class1} ne "PARTITION" and 
		$KindOf{$class2} ne "PARTITION") {
		$first = "$member1 #= $member2";
	} else {
		Error("Not possible with partition members: $line");
	}
	if ($KindOf{$class3} eq "PARTITION" and $class4 eq $idClass) {
		$Class3 = ucfirst $class3;
		$second = "${Class3}$member4 #= $member3";
	} elsif ($KindOf{$class4} eq "PARTITION" and $class3 eq $idClass) {
		$Class4 = ucfirst $class4;
		$second = "${Class4}$member3 #= $member4";
	} elsif ($KindOf{$class3} ne "PARTITION" and 
		$KindOf{$class4} ne "PARTITION") {
		$second = "$member3 #= $member4";
	} else {
		Error("Not possible with partition members: $line");
	}
	if ($joiner ne "XOR") { 
		print "\t$first $joiner $second,\n";
	} else {
		$first =~ s/#=/#\\=/;
		print "\t$first #<=> $second,\n";
	}
} # handleRequiredOp 

sub handleConflict {
	my ($line, $part) = @_;
	finishDecls();
	print "% $line\n";
	foreach $member1 (split /\s+/, $part) {
		($class1, $member1) = myResolve($member1, $line);
		$name1 = ucfirst $member1;
		foreach $member2 (split /\s+/, $part) {
			($class2, $member2) = myResolve($member2, $line);
			$name2 = ucfirst $member2;
			next if ($member1 gt $member2); # break symmetry
			next if $class1 eq $class2 and
					($member1 eq $member2 or (
					!defined($KindOf{$member1}) and
					!defined($KindOf{$member2})
				));
			if ($KindOf{$class1} eq "PARTITION") {
				@toSay = ();
				$Class = ucfirst $class1;
				foreach $index (@rowList) {
					push @toSay, "($member2 #\\= $index #\\/ " .
						"${Class}$index #\\= $member1)";
				}
				print "\t", join(" #/\\\n\t", @toSay), ",\n",
			} elsif ($KindOf{$class2} eq "PARTITION") { 
				@toSay = ();
				$Class = ucfirst $class2;
				foreach $index (@rowList) {
					push @toSay, "($member1 #\\= $index #\\/ " .
						"${Class}$index #\\= $member2)";
				}
				print "\t", join(" #/\\\n\t", @toSay), ",\n",
			} else { # both are ordinary CLASSES
				print "\t$name1 #\\= $name2,\n";
			}
		} # for member2
	} # for member1
} # handleConflict

sub handleAgree {
	my($line, $member1, $part) = @_;
	print "% $line\n";
	($class, $member1) = myResolve($member1, $line);
	if ($KindOf{$class} ne "PARTITION") {
		Error("$member1 is not a partition: $line\n");
	}
	$Class = ucfirst $class;
	if (defined($declsFinished)) { # emit now
		foreach $member2 (split /\s+/, $part) {
			($class2, $member2) = myResolve($member2, $line);
			$Class2 = ucfirst $class2;
			$member2 = ucfirst($member2);
			if ($class2 eq $idClass) {
				$member2 = defined($idRow{$member2}) ? 
										$idRow{$member2} :
										$member2;
				print "\t${Class}${member2} #= $member1,\n";
			} else {
				@toSay = ();
				foreach $index (@rowList) {
					if ($KindOf{$class2} eq "PARTITION") {
						print "\t${Class2}$index #= $member2 #=> ".
							"${Class}$index #= $member1,\n";
					} else {
						push @toSay, "($member2 #= $index #=> " .
							"${Class}$index #= $member1)";
					}
				}
				print "\t(", join(" #/\\\n\t", @toSay), "),\n"
					unless $#toSay == -1;
			}
		} # foreach $member2;
	} else { # must defer
		push @defer, $line;
		print "\t% translator deferring output until declarations done.\n";
	} # must defer
} # handleAgree

sub handleRequired {
	my ($line, $member1, $part) = @_;
	my @parts;
	if (defined($part)) {
		@parts = split /\s+/, $part;
	} else {
		@parts = split /\s+/, $member1;
		$member1 = shift @parts;
	}
	$part = "@parts";
	finishDecls();
	print "% $line\n";
	($class1, $member1) = myResolve($member1, $line);
	Error("$member1 must not be a partition member: $line\n")
		if $KindOf{$class1} eq "PARTITION";
	$name1 = ucfirst $member1;
	for $member2 (split /\s+/, $part) {
		$member2 = lc $member2;
		($class2, $member2) = myResolve($member2, $line);
		Error("$member2 must not be a partition member: $line\n")
			if $KindOf{$class2} eq "PARTITION";
		$name2 = ucfirst $member2;
		print "\t$name1 #= $name2,\n";
	}
} # handleRequired 

sub handleUsed {
	my ($line, $member) = @_;
	finishDecls();
	print "% $line\n";
	($class, $member) = myResolve($member, $line);
	$Class = ucfirst $class;
	if ($KindOf{$class} eq "PARTITION") {
		print "\toccurrences($member, $Class, ${member}Occ),\n",
			"\t1 #<= ${member}Occ,\n";
	} else { # underspecified range; force to an actual row
		if ($class ne $idClass) {
			@toSay = ();
			foreach $index (@{$Members{$idClass}}) {
				push @toSay, "$member #= $index";
			}
			print "\t", join(" #\\/\n\t\t", @toSay), ",\n";
		} else {
			print "\t% no code necessary; $class is the row identifier.\n";
		}
	} # underspecified range
} # handleUsed

sub handleUsedOp {
	my ($line, $lower, $upper, $element) = @_; 
	finishDecls();
	print "% $line\n";
	$lower = "1" unless defined($lower);
	$upper = "" unless defined($upper);
	($class1, $member1) = myResolve($element, $line);
	if ($KindOf{$class1} ne "PARTITION") {
		Error("$element is not a partition: $line\n");
	}
	$member1 = ucfirst $member1;
	$class1 = ucfirst $class1;
	print "\toccurrences($member1, $class1, ${member1}Occ),\n",
		"\t$lower #<= ${member1}Occ,",
		$upper ne "" ? " ${member1}Occ #<= $upper,\n" : "\n";
} # handleUsedOp

sub handleMatch {
	my ($line, $part1, $part2) = @_;
	finishDecls();
	print "% $line\n";
	foreach $member1 (split /\s+/, $part1) { # each in part1 matches part2
		next if ($member1 eq "");
		(undef, $member1) = myResolve($member1, $line);
		@toSay = ();
		foreach $member2 (split /\s+/, $part2) {
			next if ($member2 eq "");
			(undef, $member2) = myResolve($member2, $line);
			push @toSay, "$member1 #= $member2";  
		}
		print "\t", join(" #\\/ ", @toSay), ",\n";
	}
	foreach $member1 (split /\s+/, $part2) { # each in part2 matches part1
		next if ($member1 eq "");
		(undef, $member1) = myResolve($member1, $line);
		@toSay = ();
		foreach $member2 (split /\s+/, $part1) {
			next if ($member2 eq "");
			(undef, $member2) = myResolve($member2, $line);
			push @toSay, "$member1 #= $member2";  
		}
		print "\t", join(" #\\/ ", @toSay), ",\n";
	}
	foreach $member1 (split /\s+/, $part1) { # none in part1 matches part1
		next if ($member1 eq "");
		($class1, $member1) = myResolve($member1, $line);
		foreach $member2 (split /\s+/, $part1) {
			next if ($member2 eq "");
			($class2, $member2) = myResolve($member2, $line);
			next if ($member2 le $member1);
			next if $class1 eq $class2 and !defined($KindOf{$member1})
				and !defined($KindOf{$member2});
			print "\t$member1 #\\= $member2,\n";
		}
	}
	foreach $member1 (split /\s+/, $part2) { # none in part2 matches part2
		next if ($member1 eq "");
		($class1, $member1) = myResolve($member1, $line);
		foreach $member2 (split /\s+/, $part2) {
			next if ($member2 eq "");
			($class2, $member2) = myResolve($member2, $line);
			next if ($member2 le $member1);
			next if $class1 eq $class2 and !defined($KindOf{$member1})
				and !defined($KindOf{$member2});
			print "\t$member1 #\\= $member2,\n";
		}
	}
} # handleMatch

sub handleBefore {
	my ($line, $class, $member1, $member2) = @_;
	finishDecls();
	($class1, $member1) = myResolve($member1, $line);
	($class2, $member2) = myResolve($member2, $line);
	print "% $line\n";
	if ($class eq $idClass) { 
		if( $KindOf{$class1} ne "PARTITION" and 
				$KindOf{$class2} ne "PARTITION" ) {
			print "\t$member1 #< $member2,\n";
		} else {
			@toSay = ();
			foreach $val1 (@{$Members{$class}}) {
				if( $KindOf{$class1} eq "PARTITION" ) { 
					$snippet = "(".ucfirst($Class).$val1;
					$snippet .= " #= $member1 "." #=>  $val1 #< $member2)";
				} elsif( $KindOf{$class2} eq "PARTITION" ) { 
					$snippet = "(".ucfirst($Class).$val1;
					$snippet .= " #= $member2 "." #=> $member1 #< $val1)";
				}
				push @toSay, $snippet;
			}
			print "\t(", join(" #/\\\n\t", @toSay), "),\n";
		}
	} else { # discriminant not idClass
		if( $KindOf{$class1} ne "PARTITION" and 
				 $KindOf{$class2} ne "PARTITION" ) {
			@toSay = ();
			foreach $val1 (@{$Members{$class}}) {
				$num1 = $val1;
				$num1 =~ s/\D+//;
				foreach $val2 (@{$Members{$class}}) {
					$num2 = $val2;
					$num2 =~ s/\D+//;
					$snippet = "($member1 #= $val1 #/\\ $member2 #= $val2)";
					push @toSay, $snippet if $num1 < $num2;
				} # for val2
			} # for val1
			print "\t(", join(" #\\/\n\t", @toSay), "),\n";
		} else { 
		    foreach $partval1 (@{$Members{$idClass}}) {
			@toSay = ();
			my $partval = $Class.$partval1;
			my $partnum = $partval1;
			foreach $val1 (@{$Members{$class}}) {
			  $num1 = $val1;
			  $num1 =~ s/\D+//;
			  foreach $val2 (@{$Members{$class}}) {
			    $num2 = $val2;
			    $num2 =~ s/\D+//;
			    if($KindOf{$class1} eq "PARTITION") {
				$snippet = "($partval #= $member1 #=> " .
				"$partnum #= $val1 #/\\ $member2 #= $val2)";
			    } else { #class2 is PARTITION
				$snippet = "($partval #= $member2 #=> " .
				"$partnum #= $val2 #/\\ $member1 #= $val1)";
			    }
			    push @toSay, $snippet if $num1 < $num2;
			  } # for val2
			} # for val1
			print "\t(", join(" #\\/\n\t", @toSay), "),\n\n";
		    }	
		}
	} # discriminant not idClass
} # handleBefore

sub handleOffset {
	my ($line, $arithmetic, $amount, $class, $member1, $member2) = @_;
	finishDecls();
	print "% $line\n";
	if ($arithmetic eq "") {$arithmetic = "+";}
	Error("$class does not exist: $line\n")
		unless defined($Members{$class});
	$highValue = $highValue{$class};
	$lowValue = $lowValue{$class};
	($class1, $member1) = myResolve($member1, $line);
	Error("$member1 is of class $class: $line\n") if $class1 eq $class;
	($class2, $member2) = myResolve($member2, $line);
	Error("$member2 is of class $class: $line\n") if $class2 eq $class;
	Error("case not covered: $member1 in a partition: $line\n") 
		if $KindOf{$class1} eq "PARTITION";
	Error("case not covered: $member2 in a partition: $line\n") 
		if $KindOf{$class2} eq "PARTITION";
	if ($circular{$class}) {
		$span = $highValue-$lowValue+1;
	}
	if ($class eq $idClass) { # lucky: discriminant is idClass 
		if ($arithmetic =~ /^[+*]$/) { # simple 
			if ($circular{$class}) {
				print "\t$member1 $arithmetic $amount #= $member2 #\\/ ",
					"$member1 $arithmetic $amount - $span #= $member2,\n";
			} else {
				print "\t$member1 $arithmetic $amount #= $member2,\n";
			}
		} elsif ($arithmetic =~ /^([!]?)\+-$/) { # +- OFFSET case
			$operator = $1 eq "!" ? "#\\=" : "#=";
			$combiner = $1 eq "!" ? "#/\\" : "#\\/";
			if ($circular{$class}) {
				print "\t$member1 + $amount $operator $member2 $combiner ",
					"$member1 - $amount $operator $member2 $combiner\n",
					"\t$member1 + $amount - $span $operator $member2 ",
					"$combiner ",
					"$member1 - $amount + $span $operator $member2,\n";
			} else {
				print "\t$member1 + $amount $operator $member2 $combiner ",
				"$member1 - $amount $operator $member2,\n";
			}
		} elsif ($arithmetic eq ">") { # > OFFSET case
			print "\t$member2 - $member1 #> $amount,\n";
		} elsif ($arithmetic eq "!") {
			if ($circular{$class}) {
				print "\t($member1 + $amount - $lowValue) ",
					"mod ($highValue-$lowValue+1)+$lowValue #\\= $member2,\n";
			} else {
				print "\t$member1 + $amount #\\= $member2,\n";
			}
		} else {
			Error("Cannot understand $line\n");
		}
	} else { # discriminant not idClass
		@toSay = ();
		foreach $val1 (@{$Members{$class}}) {
			$num1 = $val1;
			$num1 =~ s/\D+//;
			foreach $val2 (@{$Members{$class}}) {
				$num2 = $val2;
				$num2 =~ s/\D+//;
				$snippet = "($member1 #= $val1 #/\\ $member2 #= $val2)";
				$dummy = $snippet;
				if ($arithmetic =~ /^[+*]$/) { # simple 
					eval " if (\$circular{$class}) {
								push \@toSay, \$snippet
								if ($num1 $arithmetic $amount - \$lowValue)
								%(\$highValue-\$lowValue+1)+\$lowValue == $num2;
							} else {
								push \@toSay, \$snippet
								if $num1 $arithmetic $amount == $num2;
							}
						";
				} elsif ($arithmetic =~ /^([!]?)\+-$/) { # +- OFFSET case
					$operator = $1 eq "!" ? "!=" : "==";
					$combiner = $1 eq "!" ? "and" : "or";
					if ($circular{$class}) {
						eval " push \@toSay, \$snippet
							if ($num1 + $amount - $lowValue)
							% ($highValue-$lowValue+1)+$lowValue $operator
							$num2 $combiner
							($num1 - $amount - $lowValue)
							% ($highValue-$lowValue+1)+$lowValue $operator
							$num2;
						";
					} else {
						eval " push \@toSay, \$snippet
							if $num1 + $amount $operator $num2 $combiner
							$num1 - $amount $operator $num2;
						";
					}
				} elsif ($arithmetic eq ">") { # > OFFSET case
					push @toSay, $snippet if ($num2 - $num1 > $amount);
				} elsif ($arithmetic eq "!") {
					if ($circular{$class}) {
						push @toSay, $snippet
						if ($num1 + $amount - $lowValue)
						% ($highValue-$lowValue+1)+$lowValue != $num2;
					} else {
						push @toSay, $snippet
						if $num1 + $amount != $num2;
					}
				} else {
					Error("Cannot understand $line\n");
				}
			} # for val2
		} # for val1
		print "\t(", join(" #\\/\n\t", @toSay), "),\n";
	} # discriminant not idClass
} # handleOffset

sub handleAggregate {
	my ($line, $operator, $class, $amount) = @_;
	finishDecls();
    
    print "\t% $line\n";
	my ($tempClass, $member);
	if ($class eq $idClass) { # class is the determinant class
		foreach $tempClass (keys %Members) {
			next if ($tempClass eq $class or
				defined($KindOf{$tempClass}) or
				defined($underspecified{$tempClass}));
		}
		print join($operator, @{$Members{$tempClass}}), " #= $amount,\n";
	} else {
		$Class = ucfirst $class;
		$aggregateVar = 1 if (!defined($aggregateVar));
		@vars = ();
		foreach $member1 (@rowList) {
			push @vars, "Temp".$aggregateVar;
			$aggregateVar++;
		}

		$list = join(", ",@{$Members{$class}});
		$list =~ s/$Class//g;
		foreach $var (@vars) {
			print "\t$var :: [", $list,"],\n";
		}
			
		print "\n\t", join(" $operator ", @vars), " #= $amount,\n\n";

		@toPrint = ();
		foreach $var1 (@vars) {
			foreach $var2 (@vars) {
				next if ($var2 le $var1);
				push @toPrint, "$var1 #\\= $var2";
			}
		}
		print "\t( ", join(" #/\\ ", @toPrint), " ),\n\n";

		$counter = 0;
		@toPrint = ();
		foreach $var (@vars) {
			$toPrint = "\t(";
			@toprint = ();
			foreach $member (@{$Members{$class}}) {
				$member1 = $member;
				$member1 =~ s/\D//g;
				push @toprint, "(".${$Members{$idClass}}[$counter]." #= ".$Class.${member1}." #=> $var #= ${member1}".")";
			}
			$counter++;
			$toPrint .= join(" #/\\\n\t\t", @toprint);
			$toPrint .= ")";
			push  @toPrint, $toPrint;
		}
		 print join(",\n\n", @toPrint), ",\n\n";
		# print "(", join(" #/\\\n\n", @toPrint), "),\n\n";
# 		# Perl Code for power loops as described in the text
# 		# Advanced Programming Language Design, Raphael Finkel, 1995
# 		$length = $#{$Members{$idClass}} + 1;
# 		$current = 0;
# 		@positions = (0) x $length;
# 		print "\t\t";
# 	Label:
# 		# Condition to enter the loop 
# 		# Check the current value is out of range 
# 		if($positions[$current] > $#{$Members{$class}} ) {
# 			 # Special condition if the loop is the first loop 
# 			goto endLabel if($current == 0);
# 				
# 
# 			$positions[$current] = 0;
# 			$list[$current] = $lowValue{$class};
# 
# 			$current--;
# 
# 			$positions[$current]++;
# 			$list[$current] = ${$Members{$class}}[$positions[$current]];
# 			goto Label;
# 		} elsif ($positions[$current] == 0) {
# 			$list[$current] = ${$Members{$class}}[$positions[$current]];
# 		}
# 
# 		# Check if the current value is used previously
# 		for($counter = 0; $counter < $current; $counter++) {
# 			if($list[$counter] eq $list[$current]) {
# 				$positions[$current]++;
# 				$list[$current] = ${$Members{$class}}[$positions[$current]];
# 				goto Label;
# 			}
# 		}
# 
# 		# Procedure for the inner most loop 
# 		# Print the sequence if it is the last loop
# 		if($current == $length-1) {
# 			my $sum = 0;
# 			foreach $member (@list) {
# 				$num = $member;
# 				$num =~ s/\D+//;		
# 				$sum = eval  "$sum $operator $num;";
# 			}
# 			if ($sum == $amount) {
# 				$toPrint = "(";
# 				for ($counter = 0; $counter < $length; $counter++) {
# 					$toPrint .=  $rowList[$counter]." #= ".$list[$counter];
# 					$toPrint .= " #/\\ " if($counter != $length - 1);
# 				}
# 				$toPrint .= ")";
# 				push @toPrint, $toPrint;
# 			}
# 			$positions[$current]++;
# 			$list[$current] = ${$Members{$class}}[$positions[$current]] 
# 		}
# 
# 		# go to the inner loop
# 		$current++ if($current != $length-1);
# 
# 		goto Label;
# 
# 	endLabel:
# 
# 	print join(" #\\/\n\t\t", @toPrint), ",\n";
	}
} # handleAggregate

sub handleDiffer {
	my ($line, $kind, $class, $list) = @_;
	finishDecls();
	print "% $line\n";
	# Error("case not covered: $line\n");
	$Class = ucfirst $class;
	Error("$class not defined: $line\n")
		unless (defined($KindOf{$class}));
	Error("$class not a partition: $line\n")
		unless ($KindOf{$class} eq "PARTITION");
	foreach $member1 (split(/ /, $list)) {	
		foreach $member2 (split(/ /, $list)) {	
			next if($member1 le $member2);
			($class1, $member1) = myResolve($member1, $line);
			$member1 = ucfirst $member1;
			($class2, $member2) = myResolve($member2, $line);
			$member2 = ucfirst $member2;
			$combiner = ($kind eq "DIFFER") ? "#\\=" : "#=";

			foreach $index1 (@rowList) {
				next if ($class1 eq $idClass and $numericId and 
					$index1 != $member1);
				foreach $index2 (@rowList) {
					next if $index1 == $index2;
					next if ($class2 eq $idClass and $numericId and 
						$index2 != $member2);
					print "\t($member1 #= $index1 #/\\ $member2 #= $index2) ",
						"#=> ${Class}$index1 $combiner ${Class}$index2,\n";
				}
			}
		}
	}
} # handleDiffer

sub handleUnique {
	my ($line, $part) = @_;
	finishDecls();
	my ($class1, $class2, $index1, $index2, $member1, $member2, $temp1, $temp);
	my (@toPrint, @toPrint1, @toPrint2);
	my @classes = split(/ /, $part);
	print "% $line\n";
	foreach $class1 (@classes) {
		Error("$class1 is not a partition") if($KindOf{$class1} ne "PARTITION");
		$Class1 = ucfirst $class1;
		foreach $class2 (@classes) {
			next if($class1 eq $class2);
			next if($class1 gt $class2);
			Error("$class2 is not a partition") 
				if($KindOf{$class2} ne "PARTITION");
			$Class2 = ucfirst $class2;
			@toPrint1 = ();
			foreach $index1 (@rowList) {
				foreach $member11 (@{$Members{$class1}}) {
					foreach $member12 (@{$Members{$class2}}) {
						print "\t(${Class1}${index1} #= $member11 #/\\ ",
								"${Class2}${index1} #= $member12) #=> \n\t\t";
						@toPrint2 = ();
						foreach $index2 (@rowList) {
							next if($index1 == $index2);
							@toPrint1 = ();
							foreach $member21 (@{$Members{$class1}}) {
								foreach $member22 (@{$Members{$class2}}) {
									next if($member11 eq $member21 and
												$member12 eq $member22);
									push @toPrint1, "(${Class1}${index2} #= $member21 #/\\ ${Class2}${index2} #= $member22)";
								}
							}
							push @toPrint2, "(".join(" #\\/ \n\t\t", @toPrint1).")";
						}
						print join(" #/\\ \n\t\t", @toPrint2), ",\n";; 
					}
				}
			}
		}
	}
}

sub emitStandard {
	# closing material
	print "\n% Propagate information\n";
	@toSay = ();
	foreach $class (keys %Members) {
		next if $class eq $idClass and $numericId;
		push @toSay, ucfirst $class;
	}
	foreach $member (keys %KindOf) {
		next unless $KindOf{$member} eq "INDIVIDUAL";
		push @toSay, ucfirst $member;
	}
	for ($counter = 1; $counter < $aggregateVar; $counter++) {
		push @toSay, "Temp".$counter;
	}
	print "\tflatten([", join(', ', @toSay), "], AllVars),\n",
	# "\tlabeling(AllVars),\n", # deleted in favor of search/6
	# notes on params to search/6:
	# occurrence gives far fewer backtracks than first-fail (Select arg)
	# indomain seems the only usable choice (Choice arg)
	# dbs(0) is faster than complete, but misses solutions (Method arg)
		"\tsearch(AllVars, 0, occurrence, indomain, complete,\n",
		"\t\t[backtrack(BacktrackCount)]), \n",
		"\tdecval(counter), getval(counter, Current), Current >= 0,\n",
		"% print result\n",
		"\tset_flag(print_depth, ", $#rowList+2, "),\n";
	# we need to delay partitions until after idClass
	foreach $class (keys %Members) {
		next if ($KindOf{$class} eq "PARTITION");
		print "\twrite(stdout, \'\\n$class:\'),\n";
		foreach $member (@{$Members{$class}}) {
			print "\twrite(stdout, \' $member=\'), " ,
			"write(stdout, $member),\n";
		} # foreach $member
	} # foreach $class
	foreach $class (keys %Members) {
		next unless ($KindOf{$class} eq "PARTITION");
		print "\twrite(stdout, \'\\npartition $class='), ",
			"write(stdout, ", ucfirst $class, "),\n";
	} # foreach $member
	foreach $member (keys %KindOf) {
		next unless $KindOf{$member} eq "INDIVIDUAL";
		print "\twrite(stdout, \'\\nvar ", lcfirst $member, " = '), ",
			"write(stdout, ", ucfirst $member, "),\n";
	} # foreach $member
	print "\twrite(stdout, \'\\nBacktrack = \'),\n",
		"\twrite(stdout, BacktrackCount), write(stdout, \'\\n\') .\n\n",
		"% ", "vim:filetype=prolog:\n"; # so the Vim editor does syntax nicely
} # emitStandard
