#!/usr/bin/perl -w

=head1 NAME

puzzle_2t_smodels: convert PuzzleLingo to smodels.

=head1 SYNOPSIS

No parameters needed; any given are ignored.

Standard input should be a program in PuzzleLingo.	This program will be
converted to smodels and output to standard output.	This output uses features
of smodels2; it won't work with smodels1.

Author: Raphael Finkel 10/2000
Modified: Hemantha Ponnuru 1/2002 to choose idClass in separate first pass
Modified: Hemantha Ponnuru 5/2003 Aggregate (product and addition)
Modified: Hemantha Ponnuru 7/2003 translation for PowerClasses

=head1 DESCRIPTION

One of the CLASSes is chosen as prime, that is,
the identifier class whose distinct values represent rows.	 All constraints
are represented by linking them to this handle class.	This method is
partially due to Mirek Trusczynski.

Experiments show:
(1) It is not better to use 1{}1 than 1{}2 in the generated code for OFFSET +-n
(2) It is not better to use partitionMember(idClassMember) than
	partitionName_idClass(partitionMember, idClassMember)

=bugs
	speed is very sensitive to the choice of prime class (idClass).
		For instance, reverse the first two classes in molly.in; one way takes
		2 seconds, the other 53 seconds.
	often longer execution than sm2c (which is sometimes wrong), sm2
	very long execution for flight.in (cancelled after 12 minutes)
	very long execution for molly.in (53 seconds)

=cut

# use strict; # catch silly declaration errors
use vars qw($needSequence %KindOf %Classes $idClass %Members %offsets %viaCount
	$class $pass $classSize %underspecified $output $pos $tmpClass @classInput
	@otherInput $tmpLine $minVar %highValue %lowValue @contents %circular
	%Maps %newVariables $minVar $minOutput $toPrint %Power $tmpFile);
$tmpFile = "/tmp/CL$$.tmp";

push @INC, "scripts", "../scripts";
require 'puzzle.common.pl';

# Introduced by Vijay for no. of braces explored and to keep a check on the 
# no. of calls to new introduced function 
$braceCount = 0;
$exp = "exp";
$expCount = 0;

# formClause($member1, $member2, $line)
# Example: formClause(father, blue, $line)
# Returns triple: (relative_guy(father, N), color_guy(blue, N), guy(N))
# The second, third element can be empty: relative_color(father, blue) if
#	member1 or member2 is in the idClass
# The third element of the pair is any additional type info needed for a RHS
sub formClause {
	my ($member1, $member2, $line) = @_;
	# $line is the input line, in case we want to report an error.
	my ($class1, $class2, $result1, $result2, $otherResult);
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	$result1 = "";
	$result2 = "";
	$otherResult = "";
	# 9 cases: 3 x 3, for member1, member2 in {oblique class, idclass, variable}
	if ($class1 ne $idClass) { # member1 oblique
		if ($class2 ne $idClass) { # member2 oblique
			$result1 = "${class1}_$idClass($member1, N)";
			$result2 = "${class2}_$idClass($member2, N)";
			$otherResult = "$idClass(N)";
		} elsif (!defined($KindOf{$member2})) { # member2 in idClass, not var
			$result1 = "${class1}_$idClass($member1, $member2)";
		} else { # member2 is a variable
			$result1 = "${class1}_$idClass($member1, N)";
			$result2 = "variable_$member2(N)",
			$otherResult = "$idClass(N)";
		}
	} elsif (!defined($KindOf{$member1})) { # member1 in idClass, not variable
		if ($class2 ne $idClass) { # member2 oblique
			$result1 = "${class2}_$idClass($member2, $member1)";
		} elsif (!defined($KindOf{$member2})) { # member2 in idClass, not var
			#print "case 5\n";
			#Error("$member1 and $member2 are in the same class: $line\n");
		} else { # member2 is a variable
			$result1 = "variable_$member2($member1)";
		}
	} else { # member1 is a variable
		if ($class2 ne $idClass) { # member2 oblique
			$result1 = "${class2}_$idClass($member2, N)";
			$result2 = "variable_$member1(N)";
			$otherResult = "$idClass(N)";
		} elsif (!defined($KindOf{$member2})) { # member2 in idClass, not var
			$result1 = "variable_$member1($member2)";
		} else { # member2 is a variable
			$result1 = "variable_$member1(N)";
			$result2 = "variable_$member2(N)";
			$otherResult = "$idClass(N)";
		}
	}
	# print "for $member1 $member2 I form: $result1, $result2, $otherResult\n";
	return($result1, $result2, $otherResult);
} # formClause

# formOpenClause($class, $member2, $classNumber, $memberNumber, $line)
# Example: formOpenClause(person, blue, 1, 2, $line)
# Returns triple: (color_id(blue,X2), person_id(Person1, X2),
#	"person(Person1), id(X2)")
# or a simplification if $member or $class is the id class
# The third element of the pair is type info needed for a RHS
sub formOpenClause {
	my ($class, $member2, $classNumber, $memberNumber, $line) = @_;
	my ($class1, $member1, $class2, $otherResult, $result1, $result2, $IdClass);
	$class1 = $class;
	$member1 = (ucfirst $class) . $classNumber;
	($class2, $member2) = resolve($member2, $line);
	$IdClass = ucfirst $idClass;
	$result1 = "";
	$result2 = "";
	$otherResult = "";
	if ($class1 eq $idClass) {
		if($class2 ne $idClass) {
			$result1 = "${class2}_$idClass($member2, $IdClass$classNumber)";
			$otherResult = "$idClass($IdClass$classNumber)";
		}
		if (defined($KindOf{$member2})) {
			$result1 = "";
			$result2 = "variable_$member2($IdClass$classNumber)";
			$otherResult = "$idClass($IdClass$classNumber)";
		}
	} elsif ($class2 eq $idClass) {
		$otherResult = "$class1($member1)";
		if (defined($KindOf{$member2})) {
			$result1 = "${class1}_$idClass($member1, $IdClass$memberNumber)";
			$result2 = "variable_$member2($IdClass$memberNumber)";
			$otherResult .=", $idClass($IdClass$memberNumber)";
		} else {
			$result1 = "${class1}_$idClass($member1, $member2)";
		}
	} else {
		$result1 = "${class1}_$idClass($member1,$IdClass$memberNumber)";
		$result2 = "${class2}_$idClass($member2,$IdClass$memberNumber)";
		$otherResult = "$class1($member1), $idClass($IdClass$memberNumber)";
	}
	return($result1, $result2, $otherResult);
} # formOpenClause

sub incrementVia {
	my ($member1, $member2, $how) = @_;
	# increment $viaCount{$class} for all $class oblique to both given members
	# if $how == 0, don't count if $member1 == $member2.
	# if $how > 0, count extra, especially if $member1 == $member2.
	my ($class, $class1, $class2);
	return if
		(defined($KindOf{$member1}) and $KindOf{$member1} eq "INDIVIDUAL");
	return if
		(defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL");
	($class1, undef) = resolve($member1, "incrementVia");
	($class2, undef) = resolve($member2, "incrementVia");
	return unless defined($class1) and defined($class2);
	return if ($class1 eq $class2) and ($how == 0);
	foreach $class (keys %Members) {
		if ($class ne $class1 and $class ne $class2) {
			$viaCount{$class} += 1 + $how;
			if ($class1 eq $class2) {
				$viaCount{$class} += 1;
			}
		}
	} # foreach $class
} # incrementVia

# removeConnector($line)
# Example: removeConnector(N == 3|N == 4) 
# 	   removeConnector(trail_day(lark,N)^exp0_0_day(N))		   
# Returns: "N == 3, N == 4"
#	   "trail_day(lark,N), exp0_0_day(N)" 
sub removeConnector {
	my ($line) = @_;
	my $retline = "";
	foreach my $i (0 .. length($line)-1) {
		$char = substr($line,$i,1);
		if ( ($char eq '^') || ($char eq '|') ) {
			$retline=$retline.", ";
		} else {
			$retline=$retline.$char;
		}
	}
	return $retline;	
} # removeConnector

# handleMember($member)
# Example: handleMember(blue)
# Returns: color_day(blue,Day)
# Comment: Creates an atom with respect to idClass
sub handleMember {
	if ($pass == 1) {
		my ($member) = @_;
		($class,$member) = resolve($member, $member);
		if ($class eq $idClass) {
			$returnValue = "N == $member";
		} else {
			$returnValue = "${class}_${idClass}($member,N)";
		}
		return($returnValue);
	}
} # handleMember

sub handleBandBrace {
	if ($pass == 1) {
		my ($line,$band) = @_;
		print "\t% $line\n";
		$returnValue = "$exp${expCount}_${braceCount}_${idClass}(N)";
		print "\t\t$returnValue :-\n";
		print "\t\t\t$band,\n";
		print "\t\t\t$idClass(N).\n";
		++$braceCount;
		return($returnValue);
	}
} # handleBandBrace

sub handleBorBrace {
	if ($pass == 1) {
		my ($line,$bor) = @_;
		print "\t% $line\n";
		$returnValue = "$exp${expCount}_${braceCount}_${idClass}(N)";
		print "\t\t$returnValue :-\n";
		print "\t\t\t1{$bor},\n";
		print "\t\t\t$idClass(N).\n";
		++$braceCount;
		return($returnValue);
	}
} # handleBorBrace

sub handleRequired {
	my ($line, $member1, $part) = @_;
	my @parts;
	if (defined($part)) {
		@parts = split /\s+/, $part;
	} else {
		@parts = split /\s+/, $member1;
		$member1 = shift @parts;
	}
	$part = "@parts";
	if ($pass == 1) {
		print "\t% $line\n" ;
		if(defined($Classes{$member1}) or $member1=~/\./) {		
			($class1, undef) = resolve($member1, $line);
			#Error("$member1 is a partition member; use AGREE") 
			#if($KindOf{$class1} eq "PARTITION");
			for $member2 (split /\s+/, $part) {
				if(defined($Classes{$member2}) or $member2=~/\./) {
					($class2, undef) = resolve($member2, $line);
					$member2 = lc $member2;
					if ($KindOf{$class2} eq "PARTITION") {
						Error("$member2 is a partition member; use AGREE") ;
					} elsif (defined($underspecified{$class2})) {
						Error("Both $member1 and $member2 belong to an " .
							"underspecified class: $line\n") 
						if (defined($underspecified{$class1}));
						($clause1, $clause2, $rhs) = formClause($member2, $member1, $line);
					} else {
						($clause1, $clause2, $rhs) = formClause($member1, $member2, $line);
					}
					# could generate two symmetric statements, but one suffices.
					print "\t\t", Append( $clause1, Append($clause2, $rhs, ", "),
							" :- "), " .\n";
					print "\t\t", Append( $clause2, Append($clause1, $rhs, ", "), 
							" :- "), " .\n";
				} else {
					$result2 = "$member2";
					$otherResult = "$idClass(N)";
					if ($class1 ne $idClass) { # member1 oblique
						$result1 = "${class1}_$idClass($member1, N)";
					} elsif (!defined($KindOf{$member1})) { 
						# member1 in idClass, not var
						$result1 = "N  == $member1";
					} else { 
						# member1 is a variable
						$result1 = "variable_$member1(N)";
					}
					print "\t\t :- $result1, not $result2";
					print ", $otherResult.\n";
					print "\t\t :- not $result1, $result2";
					print ", $otherResult.\n";
					++$expCount;
					$braceCount = 0;
				}
			} # for member2
		} else { # for member1 a complex set of rows
			for $member2 (split /\s+/, $part) {
				if(defined($Classes{$member2}) or $member2=~/\./) {
					($class2, undef) = resolve($member2, $line);
					$result2 = "$member1";
					$otherResult = "$idClass(N)";
					if ($class2 ne $idClass) { 
						# member2 oblique
						$result1 = "${class2}_$idClass($member2, N)";
					} elsif (!defined($KindOf{$member2})) { 
						# member2 in idClass, not var
						$result1 = "N  == $member2";
					} else { 
						# member2 is a variable
						$result1 = "variable_$member2(N)";
					}
					print "\t\t :- $result1, not $result2";
					print ", $otherResult.\n";
					print "\t\t :- not $result1, $result2";
					print ", $otherResult.\n";
				} else {	
					print "\t\t :- $member1, not $member2";
					print ", $idClass(N).\n";
					print "\t\t :- not $member1, $member2";
					print ", $idClass(N).\n";
				}
				++$expCount;
				$braceCount = 0;
			} # for member2
		}	 
	} elsif ($pass==0) {
		#for $member2 (split /\s+/, $part) {
		#incrementVia($member1, $member2, 0);
		#}
	}
} # handleRequired

my ($line, $kind, $class, $Class, $contents, $low, $high,
	$member, $member1, $member2, $member3,
	$member4, $clause1, $clause2, $rhs, $atom1, $atom2, $atom3, $part,
	$part1, $part2, $toPrint, $xor, $needAtom, $arithmetic, $amount,
	$offsetter, $connect, $category,
	$clause11, $clause12, $rhs1, $rhs2, $clause21, $clause22, $class1,
	$class2, $temp);


sub handleDeclaration {
	return if ($pass >= 0);
    my ($line, $kind, $class, $contents, $range) = @_;
	$KindOf{$class} = $kind;
	$lowValue{$class} = +100000; # effectively negative infinity
	$highValue{$class} = -100000; # effectively negative infinity
	if (defined($Members{$class})) {
		Error("$class is already a defined); " .
			"you can't define it again.\n");
	}
	if ($class =~ /^(variable)$/) {
		Error("Sorry, $class is a reserved name.\n");
	}
	$Members{$class} = []; # reference to initially empty array
	$circular{$class} = 0; # unless proven otherwise
	$min =1;  $max = 1;
	if ($kind eq "POWERCLASS") {
		($min, $max) = split(/\s+/, $range);
	} 
	# NOTE: %Power is the size of sub-set with respect to the idClass member
	$Power{$class} = join(' ', ($min, $max));
	if(defined($contents)) { 
		if ($contents =~ /^\s*(\d+)\s*\.\.\s*(\d+)\s*(circular)?\s*$/) {
			# range
			$low = $1;
			$high = $2;
			$circular{$class} = defined($3);
			if ($circular{$class}) {
				if ($low != 1) {
					# Error("Bottom of circular range must be 1: $line\n");
				}
			}
			$lowValue{$class} = $low;
			$highValue{$class} = $high;
			@contents = ();
			while ($low <= $high) {
				push @contents, $low;
				$low += 1;
			}
			$contents = join " ", @contents;
		} else { # not a range
		#	$idClass = $class
		#		unless $kind eq "PARTITION" or defined($idClass) or
		#		$contents =~ /\b\d/;
			# print "Id class is $idClass\n";
		}
		foreach $member (split /\s+/, $contents) {
			if ($member =~ /^(show)$/) {
				Error("Sorry, $member is a reserved name.\n");
			}
			if (defined($Classes{$member})) {
				# print STDERR
				# 	"$member is already a member of $Classes{$member}; " ,
				# 	"it is now also a member of $class.\n";
				$Classes{$member} .= " $class";
			}	else {
				$Classes{$member} = $class;
			}
			$lowValue{$class} = $member
				if $member =~ /^\d+$/ and $member < $lowValue{$class};
			$highValue{$class} = $member
				if $member =~ /^\d+$/ and $member > $highValue{$class};
			push @{$Members{$class}}, $member;
		}
		if (defined($classSize) and $KindOf{$class} eq "CLASS") {
			if ($#{$Members{$class}} != $classSize) {
				print STDERR "Class $class has ",
					1+$#{$Members{$class}},
					" elements, not ", 1+$classSize, ".\n";
				$underspecified{$class} = 1;
			}
		} else {
			$classSize = $#{$Members{$class}} if
				$KindOf{$class} eq "CLASS";
		}
	} else {
		$Maps{$class} = 1;
	}
} # handleDeclaration

sub handleVar {
    my ($line, $member) = @_;
	if (defined($Classes{$member})) {
		Error("$member already a member of $Classes{$member}: $line\n");
	}
	$KindOf{$member} = "INDIVIDUAL";
	if ( $pass == 1) {
		$class = $idClass;
		$Classes{$member} = $class;
		print "\t% $line\n";
		# print "\t\t#hide variable_$member(X) .\n";
		print "\t\t1 {variable_$member(X):$idClass(X)} 1.\n";
	} elsif ($pass == 0) {
		# All the classes introduce a variable
		#foreach $class (keys %Members) {
		#	$viaCount{$class} += 1;
		#}
	}
} # handleVar

sub handlePass {
	my($line)=@_;
	return unless $pass == 1;
	$line =~ /^\s*PASS\s+sm2tn\b(.*)/ or return;
	print "\n\t% pass through\n\t\t$1\n";
} # handlePass

sub handleRequiredOp {
	my ($line, $member1, $member2, $combiner, $member3, $member4) = @_;
	if ($pass == 1 ) {
		print "\t% $line\n"; 
		if ($combiner =~ /OR/) { # OR and XOR
			($clause11, $clause12, $rhs1) =
				formClause($member1, $member2, $line);
			($clause21, $clause22, $rhs2) =
				formClause($member3, $member4, $line);
			$clause21 =~ s/\bN\b/N1/;
			$clause22 =~ s/\bN\b/N1/;
			$rhs2 =~ s/\bN\b/N1/;
			print "\t\t", Append("1 {$clause11, $clause21} " .
				(($combiner eq "XOR") ? "1" : "2"), 
				Append(Append($clause12, $clause22, ", "),
				Append($rhs1, $rhs2, ", "), ",\n\t\t\t"),
				" :-\n\t\t\t"), " .\n";
		} else { # IFF
			($clause1, $clause2, $rhs) =
				formClause($member1, $member2, $line);
			$atom1 = &NextAtom;
			print "\t\t", Append($atom1, Append(Append($clause1,
				$clause2, ",\n\t\t\t"), $rhs, ", "), " :- "), " .\n";
			($clause1, $clause2, $rhs) =
				formClause($member3, $member4, $line);
			$atom2 = &NextAtom;
			print "\t\t", Append($atom2, Append(Append($clause1,
				$clause2, ",\n\t\t\t"), $rhs, ", "), " :- "), " .\n";
			print "\t\t:- $atom1, not $atom2 .\n" ,
				"\t\t:- $atom2, not $atom1 .\n" ;
		} # IFF
	} elsif ($pass==0) { # For pass 0
		foreach $class (keys %Members) {
			if (defined($Classes{$member1}) and
			defined($Classes{$member2})) {
				$viaCount{$class}++ if ($Classes{$member1} ne $class and
					$Classes{$member2} ne $class); 
			} else {
				if (defined($Classes{$member1})) { 
					$viaCount{$class}++ 
									if ($Classes{$member1} ne $class);
				} else {
					$viaCount{$class}++ 
									if ($Classes{$member2} ne $class);
				}
			}
			if (defined($Classes{$member3}) and
			defined($Classes{$member4})) {
				$viaCount{$class}++ if ($Classes{$member3} ne $class and
					$Classes{$member4} ne $class);
			} else {
				if (defined($Classes{$member1})) { 
					$viaCount{$class}++ 
									if (!defined($Classes{$member3}));
					$viaCount{$class}++ if (defined($Classes{$member3})
									and $Classes{$member3} ne $class);
				} else {
					$viaCount{$class}++ 
									if (!defined($Classes{$member4}));
					$viaCount{$class}++ 
									if ($Classes{$member4} ne $class);
				}
			}
		} # foreach $class
	} # pass 0
} # handleRequiredOp

sub handleConflict {
	my ($line, $part) = @_;
	if ($pass == 1) {
		print "\t% $line\n";
		foreach $member1 (split /\s+/, $part) {
			if(defined($Classes{$member1}) or $member1=~/\./) {
				($class1, undef) = resolve($member1, $line);
				foreach $member2 (split /\s+/, $part) {
					if(defined($Classes{$member2}) or $member2=~/\./) {
						($class2, undef) = resolve($member2, $line);
						if ($member1 ge $member2) {next;}; # break symmetry
						next if ($class1 eq $class2 and 
						$KindOf{$class1} ne "POWERCLASS") and 
						($member1 eq $member2 or 
						( !defined($KindOf{$member1}) and
						!defined($KindOf{$member2})));
						($clause1, $clause2, $rhs) =
						formClause($member1, $member2, $line);
						print "\t\t:- ",
							Append($clause1, Append($clause2, $rhs, ", "), ", "), " .\n";
					} else {
						if($class1 eq $idClass) {
							$clause = "N == $member1";
						} else {
							$clause = "${class1}_${idClass}($member1,N)";
						}
						print "\t\t:- $clause, $member2";
						print ", $idClass(N) .\n";						
					}
				} # foreach $member2
			} else {
				foreach $member2 (split /\s+/, $part) {
					next if ($member1 eq $member2) or ($member1 gt $member2); # break symmetry
					if(defined($Classes{$member2}) or $member2=~/\./) {
						($class2, undef) = resolve($member2, $line);
						if($class2 eq $idClass) {
							$clause = "N == $member2";
						} else {
							$clause = "${class2}_${idClass}($member2,N)";
						}
						print "\t\t:- $member1, $clause";
						print ", $idClass(N) .\n";
					} else {
						print "\t\t:- $member1, $member2";
						print ", $idClass(N) .\n";
					}			
				} #end of foreach
			}
		} # foreach $member1
	} elsif ($pass==0) {
		#foreach $member1 (split /\s+/, $part) {
		#foreach $member2 (split /\s+/, $part) {
		#next if ($member1 gt $member2);
		#incrementVia($member1, $member2, 0);
		#} # foreach $member2
		#} # foreach $member1
	} #pass 0
} # handleConflict


sub handleAgree {
	my ($line, $member1, $part) = @_;
	($class, undef) = resolve($member1, $line);
	Error("$member1 is not a partition: $line\n")
			if ($KindOf{$class} ne "PARTITION");
	if ( $pass == 1) {
		print "\t% $line\n";
		foreach $member2 (split /\s+/, $part) {
			($clause1,$clause2,$rhs) = formClause($member1, $member2, $line);
			print "\t\t", Append($clause1, Append($clause2, $rhs, ", "), " :- "), " .\n"; 
		}
	} elsif ($pass==0) {
		foreach $member2 (split /\s+/, $part) {
			incrementVia($member1, $member2, 0);
		}
	}
	print STDERR "REQUIRED deprecated; use AGREE in $line\n" if ($line =~ /REQUIRED/);
} # handleAgree


sub handleRequiredIn {
	my ($line, $member1, $member2) = @_;
	if ($pass == 1) {
		print "\t% $line\n";
		($class1, $member1) = resolve($member1, $line);
		Error ("$member1  is not a power class member\n")
			if ($KindOf{$class1} ne "POWERCLASS");
		($clause1, $clause2, $rhs) =
			formClause($member1, $member2, $line);
		print "\t\t", Append( $clause1, Append($clause2, $rhs, ", "), " :- "), " .\n";
	}
}

sub handleSize {
	my ($line, $class, $member, $size, $sign) = @_;
	# sign can be =, <, <=, >, >=.
	if ($pass == 1) {
		print "\t% $line\n";
		$class1 = $class;
		$Class1 = ucfirst $class1;
		$IdClass = ucfirst $idClass;
		($class2, $member2) = resolve($member, $line);
		Error("$member is not a class member\n") 
			if ($KindOf{$class2} !~ /^(CLASS|POWERCLASS)$/);
		my ($lower, $upper);
		if ($sign eq '=') {
			($lower, $upper) = ($size, $size);
		} elsif ($sign eq '>') {
			($lower, $upper) = ($size+1, '');
		} elsif ($sign eq '>=') {
			($lower, $upper) = ($size, '');
		} elsif ($sign eq '<') {
			($lower, $upper) = ('', $size-1);
		} elsif ($sign eq '<=') {
			($lower, $upper) = ('', $size);
		} else {
			Error("SIZE comparison \"$size\" is not defined\n");
		}
		if ($class2 eq $idClass and !defined($KindOf{$member})) {
			print "\t$lower " .
				"{${class1}_$idClass($Class1,$member):$class1($Class1)}" .
				" $upper.\n" ;
		} elsif ($class2 eq $idClass) {
			# Variable
			print "\t$lower " .
				" {${class1}_$idClass($Class1,$IdClass):$class1($Class1)}" .
				" $upper" .
				"\n\t\t :- variable_$member($IdClass), $idClass($IdClass).\n" ;
		} else {
			print "\t$lower {${class1}_$idClass($Class1,$IdClass):$class1($Class1)}
					$upper :- ${class2}_$idClass($member,$IdClass),
					$idClass($IdClass). \n";
		}
	} # first pass
} # handleSize

sub handleUsed {
	my ($line, $member) = @_;
	$needAtom = &NextAtom;
	($class, $member) = resolve($member, $line);
	if ($pass == 1) {
		print "\t% $line\n";
		if($class ne $idClass) {
			$needAtom = &NextAtom;
			print "\t\t$needAtom :- ${class}_${idClass}(${member},X), ",
					"${idClass}(X) .\n", "\t\t:- not $needAtom .\n";
			} else {
					print "\t\t :- not ${idClass}(${member}) .\n";
			}
	} elsif ($pass == 0) {
		$viaCount{$class}+=100; # TODO: fix this line
	}
} # handleUsed


sub handleUsedOp {
	my ($line, $lower, $upper, $element) = @_;
	$lower = "1" unless defined($lower);
	$upper = "" unless defined($upper);
	($class1, $member1) = resolve($element, $line);
	if ($KindOf{$class1} ne "PARTITION" and $KindOf{$class1} ne "POWERCLASS") {
		Error("$element is not a partition: $line\n");
	}
	if ( $pass == 1) {
		print "\t% $line\n";
		my $IdClass = ucfirst($idClass);
		($clause1, undef) =
			formOpenClause($idClass, $element, "", "", $line);
		print "\t\t$lower {$clause1 : $idClass($IdClass)} $upper.\n";
	} elsif ($pass == 0) {
		# TODO: fix this
	}
} # handleUsedOp

sub handleMatch {
	my ($line, $part1, $part2) = @_;
	my ($counter, $rightSide);
	my $pos=0;
	if ($pass == 1) {
		print "\t% $line\n";
		foreach $member1 (split /\s+/, $part1) {
			# each in part1 matches part2
			if ($member1 eq "") {next;}
			$toPrint = "";
			$counter = 0;
			$rightSide = "";
			foreach $member2 (split /\s+/, $part2) {
				if ($member2 eq "") {next;}
				($clause1, $clause2, $rhs) =
					formClause($member1, $member2, $line);
				if ($clause2 ne "") {
					$counter += 1;
					if (index($clause1, "N") > -1) {
						$pos++;
					}
					elsif (index($clause2, "N") > -1) {
						$pos++;
					}
					$clause1 =~ s/N/N$counter/;
					$clause2 =~ s/N/N$counter/;
					$rhs =~ s/N/N$counter/;
					$toPrint = Append($toPrint, $clause1, ",\n\t\t\t");
					$rightSide = Append($rightSide, $clause2, ", ");
					$rightSide = Append($rightSide, $rhs, ", ");
				} else {
					$toPrint = Append($toPrint, $clause1, ",\n\t\t\t");
					$rightSide = Append($rightSide, $rhs, ", ");
				}
			}
			print "\t\t1 {$toPrint } 1 ", ($rightSide eq "") ? "" :
					":- $rightSide ", ".\n";
		}
		foreach $member1 (split /\s+/, $part2) {
			# each in part2 matches part1
			if ($member1 eq "") {next;}
			$toPrint = "";
			$counter = 0;
			$rightSide = "";
			foreach $member2 (split /\s+/, $part1) {
				if ($member2 eq "") {next;}
				($clause1, $clause2, $rhs) =
					formClause($member1, $member2, $line);
				if ($clause2 ne "") {
					$counter += 1;
					if (index($clause1, "N") > -1) {
						$pos++;
					}
					elsif (index($clause2, "N")) {
						$pos++;
					}
					if(!defined($KindOf{$member1})) {
						$clause1 =~ s/N/N$counter/;
						$clause2 =~ s/N/N$counter/;
						$rhs =~ s/N/N$counter/;
					} 
					$toPrint = Append($toPrint, $clause1, ",\n\t\t\t");
					$rightSide = Append($rightSide, $clause2, ", ");
					$rightSide = Append($rightSide, $rhs, ", ");
				} else {
					$toPrint = Append($toPrint, $clause1, ",\n\t\t\t");
					$rightSide = Append($rightSide, $rhs, ", ");
				}
			}
			print "\t\t1 {$toPrint } 1 ", ($rightSide eq "") ? "" :
					":- $rightSide ", ".\n";
		}
		foreach $member1 (split /\s+/, $part1) {
			# none in part1 matches part1
			if ($member1 eq "") {next;}
			$toPrint = "\t\t:- ";
			foreach $member2 (split /\s+/, $part1) {
				if ($member2 eq "") {next;}
				if ($member2 le $member1) {next;}
				($class1, undef) = resolve($member1, $line);
				($class2, undef) = resolve($member2, $line);
				if ($class1 eq $class2) {next;}
				($clause1, $clause2, $rhs) =
					formClause($member1, $member2, $line);
				print "\t\t:- ",
				Append($clause1, Append($clause2, $rhs, ", "), ", "),
					" .\n";
			}
		}
		foreach $member1 (split /\s+/, $part2) {
			# none in part2 matches part2
			if ($member1 eq "") {next;}
			$toPrint = "\t\t:- ";
			foreach $member2 (split /\s+/, $part2) {
				if ($member2 eq "") {next;}
				if ($member2 le $member1) {next;}
				($class1, undef) = resolve($member1, $line);
				($class2, undef) = resolve($member2, $line);
				if ($class1 eq $class2) {next;}
				($clause1, $clause2, $rhs) =
					formClause($member1, $member2, $line);
				print "\t\t:- ",
				Append($clause1, Append($clause2, $rhs, ", "), ", "),
				" .\n";
			}
		}
	} elsif ($pass == 0) {
		foreach $member1 (split /\s+/, $part1) {
			next if ($member1 eq "");
			foreach $member2 (split /\s+/, $part2) {
				if ($member2 eq "") {next;}
				incrementVia($member1, $member2, 0);
				incrementVia($member1, $member2, 0);
				incrementVia($member1, $member2, 0);
				incrementVia($member1, $member2, 0);
			}
		}
		foreach $member1 (split /\s+/, $part1) {
			next if ($member1 eq "");
			foreach $member2 (split /\s+/, $part1) {
				next if ($member2 eq "");
				incrementVia($member1, $member2, 0);
			}
		}
		foreach $member1 (split /\s+/, $part2) {
			next if ($member1 eq "");
			foreach $member2 (split /\s+/, $part2) {
				next if ($member2 eq "");
				incrementVia($member1, $member2, 0);
			}
		}
	} # pass == 0
} # handleMatch


sub handleBefore {
	my ($line, $class, $member1, $member2) = @_;
	if ( $pass == 1) {
		handleConflict($line, "$member1 $member2");
		print "\t% $line\n";
		($clause11, $clause12, $rhs1) =
			formOpenClause($class, $member1, "", "",	$line);
		($clause21, $clause22, $rhs2) =
			formOpenClause($class, $member2, 1, 1, $line);
		$Class = ucfirst $class;
		($class1, $member1) = resolve($member1, $line);
		($class2, $member2) = resolve($member2, $line);
		print "\t\t:- ", Append(
						Append(Append($clause11, $clause12, ","),
								Append($clause21, $clause22, ","),
								",\n\t\t\t"),
						Append($rhs1, $rhs2, ","),
						",\n\t\t\t"), ",\n";
		if ($class eq $class1 and !defined($KindOf{$member1})) {
			Error("Before: Cannot handle this: $line\n")
				if($class eq $class2 and !defined($KindOf{$member2}));
			print "\t\t\t$member1>=${Class}1.\n";
		} elsif ($class eq $class2 and !defined($KindOf{$member2})) {
			print "\t\t\t$Class>=$member2.\n";
		} else {
			print "\t\t\t$Class>=${Class}1.\n";
		}
	} elsif ($pass == 0) {
		incrementVia($member1, $member2, 2);
		incrementVia($member1, $member2, 2);
		$viaCount{$class} -= 8;
	}
} # handleBefore

sub handleOffset {
	my ($line, $arithmetic, $amount, $class, $member1, $member2) = @_;
	if ($arithmetic eq "") {$arithmetic = "+";}
	if ($pass == 1) {
		($class1, undef) = resolve($member1, $line);
		($class2, undef) = resolve($member2, $line);
		Error("$class does not exist: $line\n")
			if (!defined($Members{$class}));
		$Class = ucfirst $class;
		($clause11,$clause12,$rhs1) =
			formOpenClause($class, $member1, "", "", $line);
		$clause1 = Append($clause11, $clause12, ", ");
		($clause21,$clause22,$rhs2) =
			formOpenClause($class, $member2, 1, 1, $line);
		$clause2 = Append($clause21, $clause22, ", ");
		print "\t% $line\n";
		if ($arithmetic =~ /^[+*]$/) { # simple add or mult
			my $IdClass = ucfirst $idClass;
			($class2, $member) = resolve($member2, $line);
			$lhs = $clause21;
			$rhs = $clause22;
			if ($class eq $idClass and
						defined($KindOf{$member2})) {
				($lhs, $rhs) = ($rhs, $lhs);
			}
			$rhs2 =~ s/$class\(${Class}1\)([,]?)//;
			if ($circular{$class}) {
				($low, $high) = ($lowValue{$class}, $highValue{$class});
				$lhs =~ s{${Class}1}
					{((${Class}$arithmetic$amount-$low) mod ($high-$low+1)+$low)};
			}	else {
				$lhs =~ s/${Class}1/${Class}$arithmetic$amount/;
			}
			$toPrint = "\t\t$lhs :- ". 
						Append($rhs,
							Append(
								Append($clause11, $clause12, ", "),
								Append($rhs1, $rhs2, ", "), 
								",\n\t\t\t"),
							",\n\t\t\t"). ".\n";	
			print $toPrint;				
			if(!$circular{$class}) {
				print "\t\t:- ", Append(Append($clause11, $clause12, ","),
										$rhs1,
										",\n\t\t\t"), ",\n",
						"\t\t\t${Class}1=$Class$arithmetic$amount, ", 
						"not $class(${Class}1).\n";				
			}
		} elsif ($arithmetic =~ /^([!]?)\+-$/) { # +- OFFSET case
			if (!defined($offsets{$class . $amount})) {
				$offsets{$class . $amount} = 1;
					# suppress emitting again
				if ($circular{$class}) {
					($low, $high) =
						($lowValue{$class}, $highValue{$class});
					$offsetter = "((${Class}+$amount-$low) mod" .
						" ($high-$low+1))+$low";
				} else {
					$offsetter = "${Class}+$amount";
				}
				print
					"\t\tdiffer_${class}_$amount(${Class}, $offsetter) ",
						":- $class(${Class}) .\n",
					"\t\tdiffer_${class}_$amount($offsetter, ${Class})",
						" :- $class(${Class}) .\n\t\t";
			}
			$connect = ($1 eq "!") ? "" : "not ";
			print "\t\t:- ", Append( $clause1, $clause2, ", "), ",\n",
				"\t\t\t${connect}differ_${class}_$amount" ,
					"(${Class},${Class}1),\n",
				"\t\t\t", Append($rhs1, $rhs2, ", "), " .\n";

		} elsif ($arithmetic eq ">") { # > OFFSET case
			if ($circular{$class}) {
				Error("Can't handle circular case here: $line\n");
			}
			print "\t\t:- ", Append( $clause1, $clause2, ", "), ",\n",
						"\t\t\t${Class}2 = ${Class} + $amount, ",
						"${Class}2 >= ${Class}1,\n",
						"\t\t\t$rhs1, $rhs2 .\n" ;
		} elsif ($arithmetic eq "!") {
			# ! OFFSET case (not the given offset)
			if ($circular{$class}) {
				Error("Can't handle circular case here: $line\n");
			}
			print "\t\t:- ", Append($clause1, $clause2, ", "), ",\n", 
						"\t\t\t${Class}1 = ${Class} + $amount,\n",
						"\t\t\t$rhs1, $rhs2 .\n" ;
		} else {
			Error("OFFSET: Cannot understand $line\n");
		}
		handleConflict($line, "$member1 $member2");
	} elsif ($pass == 0) {
		# print "% $line\n";
		$viaCount{$class} -= 8;
		incrementVia($member1, $member2, 5);
		# experiment shows at least 5
	}
} # handleOffset

sub handleDiffer {
	my ($line, $kind, $category, $list) = @_;
	Error("$category not defined: $line\n")
		unless (defined($KindOf{$category}));
	Error("$category is neither a partition nor a powerclass: $line\n")
		if ($KindOf{$category} ne "PARTITION" and 
			$KindOf{$category} ne "POWERCLASS");
	if ($pass == 1) {
		foreach $member1 (split(/ /, $list)) {	
			next if($member1 eq "");
			foreach $member2 (split(/ /, $list)) {	
				next if($member2 eq "");
				next if($member1 le $member2);
				($clause11,$clause12,$rhs1) =
					formOpenClause($category, $member1, "", "", $line);
				($clause21,$clause22,$rhs2) =
					formOpenClause($category, $member2, "", 1, $line);
				if ($kind eq "DIFFER") {
					print "\t% $line\n\t\t" ,
					":- ", Append($clause11, $clause12, ", "), ",\n" ,
					"\t\t\t", Append($clause21, $clause22, ", "), ",\n" ,
					"\t\t\t$rhs1, $rhs2 .\n";
				} else {
					# don't really need both sides, but might be more efficient.
					print "\t% $line\n" ,
						"\t\t$clause11 :-\n\t\t\t",
							Append($clause12, Append($clause21, $clause22, ", "), ", "), ",\n",
						"\t\t\t", "$rhs1, $rhs2 .\n", 
						"\t\t$clause21 :-\n\t\t\t",
							Append($clause22,
								Append($clause11, $clause12, ", "),
								", "),
							",\n",
						"\t\t\t", "$rhs1, $rhs2 .\n", 
				}
			}
		}
	} elsif ($pass == 0) {
		foreach $member1 (split(/ /, $list)) {	
			foreach $member2 (split(/ /, $list)) {	
				next if($member1 le $member2);
				next if($member1 eq "" or $member2 eq "");
				incrementVia($member1, $member2, 0);
			}
		}
	}
} # handleDiffer


sub handleUnique {
	my ($line, $part) = @_;
	my @classes = split(/ /, $part);
	if ($pass == 1) {
		print "\t% $line\n";
		foreach $class1 (@classes) {
			Error("$class1 is not a partition") 
				if($KindOf{$class1} ne "PARTITION" and
					$KindOf{$class1} ne "POWERCLASS");
			foreach $class2 (@classes) {
				Error("$class2 is not a partition") 
					if($KindOf{$class2} ne "PARTITION" and
						$KindOf{$class2} ne "POWERCLASS");
				next if($class1 ge $class2);

				$Class1 = ucfirst $class1;
				$Class2 = ucfirst $class2;
				$IdClass = ucfirst $idClass;
				
				$clause1 = "${class1}_${idClass}(${Class1},${IdClass})"; 
				$clause11 = $clause1;
				$clause12 = $clause1;
				$clause2 = "${class2}_${idClass}(${Class2},${IdClass})"; 
				$clause21 = $clause2;
				$clause22 = $clause2;

				$clause11 =~ s/${IdClass}/${IdClass}1/; 
				$clause12 =~ s/${IdClass}/${IdClass}2/; 
				$clause21 =~ s/${IdClass}/${IdClass}1/; 
				$clause22 =~ s/${IdClass}/${IdClass}2/; 
				print "\t\t:- ${clause11}, ${clause12},\n",
						"\t\t\t${clause21}, ${clause22},\n",
						"\t\t\t$idClass(${IdClass}1),$idClass(${IdClass}2),\n", 
						"\t\t\t${class1}(${Class1}),$class2(${Class2}),\n", 
						"\t\t\t${IdClass}1 != ${IdClass}2.\n";
			}
		}
	} elsif ($pass == 0) {
		# All the classes are equally charged
	}
} # handleUnique

sub handleAggregate {
	my ($line, $operator, $class, $amount) = @_;
	if ($pass == 1) {
		# Assumption: Aggregate is only used with partitions or underspecified
		# classes which are not idClasses
		print "\t% $line\n";

		if ($operator ne "+") {
			my $count = 1;
			$Class = ucfirst $class;
			my $atom = &NextAtom;
			print "\t\t$atom :- ";
			foreach $element (@{$Members{$idClass}}) {
				($clause,undef,$rhs) = 
							formOpenClause($class, $element, $count, "", $line);
				print join(", ",($clause, $rhs)), ",\n\t\t\t";
				$count++;
			}	

			# Each variable must be different
			if($KindOf{$class} eq "CLASS") {
				for($i=1; $i<$count; $i++) {
					for($j=$i+1; $j<$count; $j++) {
						print "${Class}$i != ${Class}$j,\n\t\t\t";
					}
				}
			}

			print "${Class}1 ";
			for($i=2; $i<$count; $i++) {
				print "$operator ${Class}$i ";
			}
			print " == $amount.\n";
			print "\t\t:- not $atom.\n";
		} else { 
			# Better translation for the addition operator
			$puzzle_vars++;
			print "#weight puzzle_var_",$puzzle_vars,"(X)= X.\n";	
			print "$amount [ ";
			$count = $#{$Members{$idClass}};
			$count++;
			$toPrint = "";
			foreach $member (@{$Members{$class}}) {
				$toPrint = Append($toPrint, 
									"puzzle_var_${puzzle_vars}($member)", 
									", ");
			}
			print "$toPrint ] $amount . \n";
			print "\t\t$count { puzzle_var_${puzzle_vars}(X):$class(X) } $count .\n";
			print "\t\t1 { ${class}_$idClass(X,Y):$idClass(Y) } 1 :-",
						" puzzle_var_${puzzle_vars}(X), $class(X). \n";
		}
	} elsif ($pass == 0) {
		# All the classes are equally charged
	}
} # handleAggregate

# Routines to support new syntax
# A map will not be considered for idClass 
sub handleMapDeclaration {
	my ($line, $name, $list) = @_;
	if(defined($list) and $list =~ /ONTO/) {
		handleDeclaration($line, "CLASS", $name, undef);
	} else {
		handleDeclaration($line, "PARTITION", $name, undef);
	}
	if($pass == 1) {
		foreach (split(' ',$list)) {
			handleMapSymmetric($name) if($_ =~ /^(SYMMETRIC)$/);
			handleMapAsymmetric($name) if($_ =~ /^(ASYMMETRIC)$/);
			handleMapNonReflexive($name) if($_ =~ /^(NONREFLEXIVE)$/);
		}
	}
} # handleMapDeclaration

sub handleMapSymmetric {
	my ($class) = @_;
	print "\t% Mapping \"$class\" is symmetric\n";
	$Class = ucfirst $class;
	$IdClass = ucfirst $idClass;
	print "\t\t ${class}_${idClass}($Class,$IdClass) :- ",
						"${class}_${idClass}($IdClass, $Class),\n",
					"\t\t\t${idClass}($IdClass), ${class}($Class).\n";
	print "\t\t ${class}_${idClass}($IdClass,$Class) :- ",
						"${class}_${idClass}($Class,$IdClass),\n",
					"\t\t\t${idClass}($IdClass), ${class}($Class).\n";
} # handleMapSymmetric

sub handleMapAsymmetric {
	my ($class) = @_;
	my $parent = $idClass;
	print "\t%  Mapping \"$class\" is asymmetric\n";
	print "\t\t:- ${class}_$idClass(X,Y), ${class}_$idClass(Y,X),\n",
			"\t\t\t $parent(Y), $parent(X).\n";
} # handleMapAsymmetric

sub handleMapNonReflexive {
	my ($class) = @_;
	print "\t%  Mapping \"$class\" is non-reflexive\n";
	my $need = &NextAtom;
	print "\t\t$need :- ${class}_${idClass}(X,X), $idClass(X).\n";
	print "\t\t:- $need.\n";
} # handleMapNonReflexive

sub handleMapRequired {
	if($pass == 1) {
		my  ($line, $class, $member1, $member2) = @_;
		$Class = ucfirst $class;
		$IdClass = ucfirst $idClass;
		($class1, undef) = resolve($member1, $line);
		($class2, undef) = resolve($member2, $line);

		if(!defined($KindOf{$member1}) and !defined($KindOf{$member2})) {
			if($Classes{$member2} eq $idClass and 
				$Classes{$member1} eq $idClass ) {
				$clause = "${class}_$idClass(${member2}, ${member1})";
				$rhs = '';
			} elsif ($Classes{$member2} eq $idClass) {
				($clause, $clause2, $rhs) = formOpenClause($class, $member1, 
																"", "", $line);
				$rhs = Append($clause2, $rhs, ",\n\t\t\t");
				$rhs = Append($rhs, "${Class} == ${member2}", ",\n\t\t\t");
			} elsif ($Classes{$member1} eq $idClass) {
				($clause, $clause2, $rhs) = formOpenClause($class, $member2, 
																"", "", $line);
				$rhs = Append($clause2, $rhs, ",\n\t\t\t");
				$rhs = Append($rhs, 
								"${class}_${idClass}(${IdClass},${member1})",
								",\n\t\t\t");
			} else {
				($clause11, $clause12, $rhs1) = formOpenClause($class, $member1,
																"1","1",$line);
				($clause21, $clause22, $rhs2) = formOpenClause($class, $member2,
																"2","2",$line);
				$clause = $clause11;
				$clause22 =~ s/${IdClass}2/${Class}1/;
				$rhs = Append($clause12,
								Append($clause22, $rhs1, ","),
								",\n\t\t\t");
			}
		} elsif(!defined($KindOf{$member2})) {
			$clause = "variable_$member1($IdClass)";
			$rhs = "${class}_${idClass}($Class, $IdClass)";
			if($Classes{$member2} eq $idClass) {
				$rhs =~ s/$Class/$member2/;
				$rhs = Append($rhs, "$idClass($IdClass)", ",\n\t\t\t");
			} else {
				$rhs = Append($rhs, 
							  "${class2}_$idClass($member2, $Class)", 
							  ", ");
				$rhs = Append($rhs, 
							  "${class}($Class), $idClass($IdClass)", 
							  ",\n\t\t\t");
			}
		} elsif(!defined($KindOf{$member1})) {
			$clause = "variable_$member2($IdClass)";
			$rhs = "${class}_$idClass($Class, ${IdClass}1)";
			if ($Classes{$member1} eq $idClass) {
				$rhs =~ s/${IdClass}1/$member1/;
			    $rhs =~ s/$Class/${IdClass}/;
        		$rhs = Append($rhs, "$idClass($IdClass)", ",\n\t\t\t");
			} else {
				$rhs =~ s/$Class/${IdClass}/;
				$rhs = Append($rhs, 
							  "${class1}_$idClass(${member1},${IdClass}1)",
							  ", ");
				$rhs = Append($rhs, 
							  "$idClass(${IdClass}1), $idClass($IdClass)", 
							  ",\n\t\t\t");
		    }
		} else { # both member1 and member2 are variables
			# Yet to do this thing
			# Finished till here
			$Member = ucfirst $member2;
			$rhs = "variable_${member2}($Member), $idClass($Member)";
			$member2 = $Member;
			$Member = ucfirst $member1;
			$rhs = Append($rhs, "variable_${member1}($Member), $idClass($Member)",
							",\n\t\t\t");
			$member1 = $Member;
			$clause = "${class}_$idClass(${member2}, ${member1})";
		}
		print "\n\t% $line\n\t\t";
		my $nextAtom = &NextAtom;
		print "$nextAtom :- ", Append($clause, $rhs,", "), ".\n";
		print "\t\t:- not $nextAtom.\n";

	} elsif($pass == 0) {

	}
}

sub handleMapConflict {
	if($pass == 1) {
		my  ($line, $class, $member1, $member2) = @_;
		$Class = ucfirst $class;
		my ($class2, undef) = resolve($member2, $line);
		# member1 and member2 are not variables
		my ($clause11, $clause12, $rhs1) = formOpenClause($class, $member1, "", 
														"", $line);
		if(!defined($KindOf{$member2})) {
			if($Classes{$member2} eq $idClass) {
				$rhs1 = Append($rhs1, "$member2 == $Class", ", ");
			} else {
				$rhs1 = Append($rhs1, 
								"${class2}_$idClass($member2,${Class})",
								", ");
			}
		} else {
			$rhs1 = Append($rhs1, 
							"\n\t\t\tvariable_$member2(${IdClass}1)", 
							", ");
			$rhs1 = Append($rhs1, 
							"$idClass(${IdClass}1), ${IdClass}1 == $Class", 
							",\n\t\t\t");
		}
		
		print "\n\t% $line\n\t\t";
		print ":- ", Append($clause11,
							Append($clause12,$rhs1,",\n\t\t\t") , 
								",\n\t\t\t "), ".\n";
	}							
} # handleMapConflict

sub handleDifferMap {
	if ($pass == 1) {
		my ($line, $list) = @_;
		my @maps = split(/ /, $list);
		print "\t% $line\n";
		foreach $map1 (@maps) {
			foreach $map2 (@maps) {
				next if($map1 ge $map2);
				$IdClass = ucfirst $idClass;
				$Map1 = ucfirst $map1;
				$Map2 = ucfirst $map2;

				$clause1 = "${map1}_${idClass}(${Map1},${IdClass})"; 
				$clause2 = "${map2}_${idClass}(${Map2},${IdClass})"; 

				print "\t\t :- $clause1, $clause2, \n\t\t\t",
						"${idClass}(${IdClass}), ${idClass}(${Map1}),",
						"${idClass}(${Map2}), \n\t\t\t",
						"${Map1} == ${Map2}.\n";
			}
		}
	}
} # handleDifferMap

sub emitStandard {
	# generate all class information
	my ($class, $element, $index, $Class, $lowCount);
	if (!defined($idClass)) {
		print STDERR "$0: I couldn't figure out an idClass; failing.\n";
		exit(1);
	}
	print "\n% Members\n";
	foreach $class (sort keys %Members) {
		print "\t$class(",
			join(";", @{$Members{(defined($Maps{$class}))?$idClass:$class}}),
			") .\n";
		if ($KindOf{$class} eq "PARTITION") {next;} # don't check count
	} # foreach $class

	# existence: one hit per row, one per column
	print "\n% existence: one hit per row, one hit per column in each table\n";
	foreach $class (keys %Members) {
		$Class = ucfirst $class;
		next if ($class eq $idClass);
		my $name = $class;
		my ($min, $max) = split(' ', $Power{$class});
		print "\t$min {${name}_$idClass($Class,X):$class($Class)} $max :- ",
			"$idClass(X) .\n" ;
		next if ($KindOf{$class} eq "POWERCLASS");
		if ($KindOf{$class} eq "CLASS") {
			$lowCount = !defined($underspecified{$class}) ? 1 : 0;
			print "\t$lowCount {${class}_$idClass($Class,X):$idClass(X)} 1 :- ",
				"$class($Class) .\n" ;
		}
	} # one hit per row, column

	# uniqueness of rows for powerclasses
	$IdClass = ucfirst $idClass;
	foreach $class (keys %Members) {
		next if ($KindOf{$class} ne "POWERCLASS") ;
		$Class = ucfirst $class;
		my $need = &NextAtom;
		my ($min, $max) = split(' ', $Power{$class});
		print "% min $min max $max\n";
		if ($min == $max) {
			# the following code is faster, but only good if all sets have the
			# same cardinality
			print "\n% uniqueness of rows wrt $class (no subsets)\n\n" .
				"\tsame$Class(${IdClass}1, ${IdClass}2, $Class) :-\n" .
				"\t\t${class}_$idClass($Class, ${IdClass}1) ,\n" .
				"\t\t${class}_$idClass($Class, ${IdClass}2) ,\n" .
				"\t\t${IdClass}1 != ${IdClass}2 ,\n" .
				"\t\t$idClass(${IdClass}1;${IdClass}2) , $class($Class) .\n" .
				"\n" .
				"\t:- $min {same$Class(${IdClass}1, ${IdClass}2, $Class) : " .
					"$class($Class) },\n" .
				"\t\t${IdClass}1 != ${IdClass}2 ,\n" .
				"\t\t$idClass(${IdClass}1;${IdClass}2) .\n" ;
		} else { # $min is different from $max, so we must deal with subsets
			print "\n% uniqueness of rows wrt $class (can be subsets)\n\n" .
				"\tdifferent$Class(${IdClass}1, ${IdClass}2) :-\n" .
				"\t\t${class}_$idClass($Class, ${IdClass}1) , " .
					"not ${class}_$idClass($Class, ${IdClass}2) ,\n" .
				"\t\t${IdClass}1 != ${IdClass}2 ,\n" .
				"\t\t$idClass(${IdClass}1;${IdClass}2) , $class($Class) .\n" .
				"\n" .
				"\t:- not different$Class(${IdClass}1, ${IdClass}2) ,\n" .
				"\t\tnot different$Class(${IdClass}2, ${IdClass}1) ,\n" .
				"\t\t${IdClass}1 != ${IdClass}2 ,\n" .
				"\t\t$idClass(${IdClass}1;${IdClass}2) , $class($Class) .\n" ;
		}
	}
} # emitStandard

sub extract {
	# return numerical portion of parameter
	my ($param) = @_;
	$param =~ /([-\d]+)/ or return 0;
	return $1;
} # extract

# First Pass
open TMP, ">$tmpFile";
while(<STDIN>) {
    print TMP $_;
}
close TMP;
close STDIN;

# Second Pass
open CLFILE, $tmpFile;
$pass=-1;
mainLoop();
close CLFILE;
foreach $class (keys %Members) {
	next if(defined($Maps{$class}));
    $viaCount{$class} = 0;
}

open CLFILE, $tmpFile;
$pass++;
mainLoop();
close CLFILE;

$minVar=10000;
my @toPrint = ();
foreach $class (keys %Members) {
	next if defined($KindOf{$class}) and $KindOf{$class} eq "PARTITION";
	next if defined($KindOf{$class}) and $KindOf{$class} eq "POWERCLASS";
	next if (defined($Maps{$class}));
	next if defined($underspecified{$class}) ;
	push @toPrint, "$viaCount{$class} $class";
	if ($minVar > $viaCount{$class}) {
		$idClass = $class;
		$minVar = $viaCount{$class};
	}
}
print "% viacounts: ",
	join(", ", sort {extract($a) <=> extract($b)} (@toPrint)), "\n";

# Third Pass
$pass++;
open CLFILE, $tmpFile;
mainLoop();
emitStandard();
close CLFILE;
unlink $tmpFile;

=head1 TESTING

puzzles to test on:

airport: 0.02 position, .35 plane, .35 origin, .42 gate
	1 position, 42 plane, 45 origin, 53 gate
	SM2: .11
cloud: 0,01 order, .08 first, .14 image, .19 last
	-46 order, 28 image, 31 last, 32 first
	SM2: .14
comedian .12 order, .7 subject, .75 first, 6.3 last
	17 order, 29 first, 32 last, 34 subject
	SM2: .63
egypt 0.0 dynasty, 0.0 king, 0.03 scepter, 0.05 archaeologist, 0.05 cartouche
	 -4 dynasty, 19 king, 28 cartouche, 29 scepter, 31 archaeologist
	SM2: .04
flight 0.0 seat, .40 last, 10 first, 10 pastime
	 2 seat, 61 last, 69 first, 72 pastime
	SM2: .28
french	0.010 position .71 person, 10 soda, 10 visits
	 -17 position, 16 person, 50 soda, 62 visits
	SM2: 0.16
holiday 0.0 number .28 wife, .30 husband, .31 topping, .34 surname, 1.53 cookie
	0 number, 64 surname, 75 husband, 78 wife, 87 cookie, 88 topping
	SM2: .13
house .14 grant, .17 first, .22 year, .79 last
	2 year 4 first, 6 grant, 11 last
	SM2: 5.27
molly .04 relative, .21 color, .35 costume, 1.39 location, 10 doll
	43 relative, 66 costume, 66 color, 71 location, 72 doll
	SM2: .20
paws 0.0 room, .03 dog, .04 groomer, .12 customer
	 -3 room, 10 dog, 24 groomer, 28 customer
	SM2: .04
piano .59 pianist, 1.290 order, 1.81 composer, 28.10 rag
	7 order, 26 pianist, 30 composer, 37 rag
	SM2: .33
post .03 customer, 0.07 time, .11 state
	 0 customer, 17 time, 22 state
	SM2: .29
price 0.0 order, .06 game, .07 contestant, .07 prize
	 -14 order, 9 game, 11 contestant, 16 prize
	SM2: .41
trek .08 villain, .11 character, .38 souvenir, 3.94 friend
	43 character, 44 villain, 52 souvenir, 57 friend
	SM2: .05

=cut
