#!/usr/bin/perl -w
=head SYNOPSYS

Reads intermediate files "rules.tmp" and "facts.tmp_combine.tmp.aspps" from an
ndcs (aspps) run using the version of the solver that outputs extra
information.  

Generates an "explanation" of the grounding that a human can follow to get the
ssame result.  This explanation is keyed to the constraint-lingo program.

Authors: Hemantha Ponnuru, Raphael Finkel
Modified: Raphael Finkel 2/1/2003 to generate proper english text.
Modified: Hemantha Ponnuru 10/1/2003 added the visual extention.

Subroutines:
	assimilateClauses -> Store the grounded rules
	assimilateGrounding -> Maps the grounded rules to the translated rules
	assimilateRules -> Maps the translated rules to the actual CL rules
	assimilateSolution -> parse the output of the aspps solver
	
	getBranch -> returns the branch id for an atom with a truth value.
	insertAtom -> inserts a true atom into the list of true atoms.
	isTrueFalse -> checks if an atom is true or false.

	reason -> returns the reason why a particular atom is true or false from
				the solver information.
	groundReason -> returns the reason why a particular atom is true or false
				from the grounder information.
	generateReasons -> Generate the reason for the actual solution.
	generateAssumptions -> Generate the reasons for the assumptions made for
							the actual solution.

	createLbl -> creates a label with the provided information.
	generateNode -> creates a node for a tree with the provided information.
	setUp -> generates the order in which the "supporting" atoms are set to true 
			or false for an atom to be true or false.
	showRow -> creates the row on the solution table

	showAssumptionTree -> generate the Assumption Tree
	showSolution -> generate the Solution Table
	showReason -> generate the tree that gives the reason for the atom
	showSubTree -> generate the sub tree for an assumption

1/2003

=cut


use strict;
push @INC, "scripts", "../scripts";
require 'puzzleEnglish.pl';
require 'reasonTable.pl';
require 'reasonTree.pl';
require 'reasonNode.pl';

use vars ( # global variables
	# assimilateRules
	'%constraintLingo', # snippet of CL program associated with a rule.
	'@actualRules', # Translated CL rules
	
	# assimilateGrounding
	'@rules', # the rules generated by the grounder.
	'@ruleIndices', # indices to the actual translated rules.
	'@atoms',	# Atoms that have to be assigned truth values.
	'$numAtoms', # number of atoms that are yet to be assigned truth values
	'%groundAtoms',
	'%groundClauses',
	'@AgroundClauses',

	# assimilateSolution
	'$idClass',	# idClass decided by the translator
	'$root',	# Root of the assumption tree
	'$assumedAtom',
	'@Branch',

	'$serial' , # serial number of steps for cross-referencing
	'@toPrint',
	'%explained',
	'%visited', # Used in finding the atoms required for reasons
	'$tab',

	'@lines', # input
	'@tableValues', # information for the visual table

	# command line options
	'$simpleSolution',
	'$outputText',
	'$prune',
	'$visual',
	'$visualTable',
	'$assumptionTree',	
	'$completeSolution',	

	
	# Files used
	'$rulesFile',
	'$grounderFile',
	'$solverFile',
	'$puzzleFile',
	'$combineFile',
	);


sub assimilateRules { # Read all the rules
	my ($file, $line, $currentRule, $currentCL);
	$file = $rulesFile;
	open(RULES, "<$file");
	@actualRules = [];
	while(<RULES>) {
		$line = $_;
		if ($line =~ /\s*%.*/) {
			$line =~ s/\t//g;
			$currentCL = $line;
		} elsif( $line !~ /(.*)\./) {
			chomp($line);
			$currentRule .= $line;
		} else {
			$currentRule .= $line;
			$currentRule =~ s/\t//g;
			chop($currentRule);
			push @actualRules, $currentRule;
			$constraintLingo{$currentRule} = $currentCL;
			$constraintLingo{$currentRule} =~ s/%/The clue tells us/;
			$constraintLingo{$currentRule} =~ s/\s*$/, so /;
			$constraintLingo{$currentRule} = "By elimination, "
				if $constraintLingo{$currentRule} =~ /unique|existence/i;
			$constraintLingo{$currentRule} = "To be consistent, "
				if $constraintLingo{$currentRule} =~ /closure/i;
			$currentRule = "";
		}
	} 
	close RULES;
} # assimilateRules

sub assimilateGrounding {
	my ($line, $file, $atom, $value, $rules, $counter);
	$file = $combineFile;
	open (RULES, "<$file");
	$line = <RULES>;
	chop($line);
	(undef, $numAtoms, $rules) = split(/ /, $line);
	$counter = 1;
	@rules = ();
	@ruleIndices = ();
	while($counter <= $rules) {
		$line = <RULES>;
		chop($line);
		(undef, $ruleIndices[$counter]) = split(/ /, $line);
		$line = <RULES>;
		chop($line);
		$rules[$counter] = $line;
		$counter++;
	}
	$counter = 1;
	@atoms = ();
	while($counter <= $numAtoms) {
		$line = <RULES>;
		chop($line);
		(undef, $value, $atom) = split(/ /, $line);
		if($value > 0) {
			my $newAtom = Atom->new	(	'Name' => $atom,
										'True' => reword($atom,"true"),
										'False' => reword($atom,"false"),
									);
			$atoms[$counter] = $newAtom;
			$counter += 1;
		}
	}
	close RULES;
} # assimilateGrounding


sub assimilateSolution {
	my $line;
	my $answer = 0;
	my $currentBranch = 1;
	my $counter = 0;
	my @parentNode = [];

	my $file = $solverFile;
	open(SOLVER, "<$file");
	push @lines, $_ while (<SOLVER>);
	close SOLVER;
	
	$root = Node->new( Atom => '-1' );
	my $currentAssumedNode = $root;

	while($line = shift @lines) {
		# print STDERR "$line";
		chop $line;
		# info : rule branch forcingAtom forcingAtomTruth
		if($line=~ /rule ([0-9]+) atom ([0-9]*) (true|false) atom ([0-9]*) ([0-1])\s*$/) {
			my $node = Node->new(	Atom => $2,
									Rule => $1,
									Truth => $3 eq "true" ? 1 : 0,
									FAtom => $4,
									FTruth => $5,
									Assumption => 0,
								);
			$currentAssumedNode->addForcedAtom($node);
		} elsif ($line =~ /(true|false) atom ([0-9]+) meta ([0-9]+).\s*/) {
			my ($metaCount, $ruleCount, $atom1, $value1);
			$atom1 = $2;
			$value1 = $1;
			$metaCount = $3 - $numAtoms;
			$ruleCount = 1;
			while ($metaCount > 0) {
				if ($rules[$ruleCount] =~ /(.*){(.*)}(.*)/) {
					$metaCount--;
				}
				$ruleCount++;
			}
			$ruleCount -= 2;
			my $node = Node->new(	Atom => $2,
									Rule => $ruleCount,
									Truth => $value1 eq "true" ? 1 : 0,
									FAtom => -1,
									Assumption => 0,
								);
			$currentAssumedNode->addForcedAtom($node);
		} elsif ($line =~ /aaaa\(([a-z0-9]+)\)/) {
			$idClass = $1;
		} elsif ($answer) {
			$line =~ /\s([a-zA-Z\(,\)_0-9]*)\s$/;
			# unshift @answerAtoms, $1;
		} elsif ($line =~ /Answer-set\s*[0-9]*:$/){
			$answer += 1;
			last if($answer > 1);
		} elsif ($line =~ /Contradict atom ([0-9]+) (meta|rule) ([0-9]+)(.*)/) {
			my ($metaCount, $ruleCount, $atom1, $value1);
			# Check if the atom is a meta atom or a normal atom
			# if atom is not a meta atom
			if ($1 <= $numAtoms) {
				$atom1 = $1;
				my $fatom = $currentAssumedNode->getNode($1);
				if ($fatom->{'Truth'} == 1) {
					$value1 = 0;
				} else {
					$value1 = 1;
				}
				if ($2 eq "meta") {
					$metaCount = $3 - $numAtoms;
					$ruleCount = 1;
					while ($metaCount > 0) {
						if ($rules[$ruleCount] =~ /(.*){(.*)}(.*)/) {
							$metaCount--;
						}
						$ruleCount++;
					}
					$ruleCount -= 2;
				} else {
					$ruleCount = $3;
				}
				my $node = Node->new(	Atom => $atom1,
										Rule => $ruleCount,
										Truth => $value1,
										Assumption => 0,
										FAtom => -1
									);
				$node->{'Contradiction'} = $fatom;
				$fatom->{'Contradiction'} = $node;
				$currentAssumedNode->addForcedAtom($node);
				%visited = ();
				$node->setContext;
				%visited = ();
				$fatom->setContext;
			} else { # atom is a meta atom
				# As of now we consider cases like a1, a2 -> m { ... } n
				$atom1 = $1;
				$ruleCount = $3;
				$currentAssumedNode->{'Rule'} = $ruleCount;
			}
		} elsif ($line =~ /(.*)Backtrack=([0-9]+) literal=([0-9]+)(.*)/) {
			my $found = 0;
			$assumedAtom = $3;
			while (!$found) {
				my $counter = 0;
				foreach my $nextAssumption (@{$Branch[$currentBranch]}) {
					if ($nextAssumption == $assumedAtom) {
						my $node = Node->new(	Atom => $assumedAtom,
												Truth => 1,
												Assumption => 1,
												Branch => $currentBranch
											);
						$parentNode[$currentBranch]->addChild($node);
						delete ${$Branch[$currentBranch]}[$counter];
						$currentAssumedNode = $node;
						$found = 1;
					}
					last if($found);
					$counter++;
				}
				last if ($found);
				$currentBranch--;
			}
		} elsif ($line =~ /branch=([0-9]+) literal\[([0-9]+)\]=([0-9]+)\s*/) {
			my $branch = $1;
			push @{$Branch[$1]}, $3;
		} elsif ($line =~ /(.*)branch=([0-9]+) literal=([0-9]+)(.*)/) {
			# Assumption is made
			$currentBranch = $2;
			$assumedAtom = $3;
			my $node = Node->new(	Atom => $assumedAtom,
									Truth => 1,
									Assumption => 1,
									Branch => $currentBranch
								);
			push @parentNode, $currentAssumedNode;
			$currentAssumedNode->addChild($node);
			$currentAssumedNode = $node;
		} elsif ($line =~ /End of Solution(.*)/) {
			# Do nothing right now!
		}
	} 
	$root->checkValidAssumption;
	# &setSubSetNodes;
} # assimilateSolution

sub setSubSetNodes {
	push my (@queue), $root;
	while ($#queue > -1) {
		my $pnode = pop @queue;
		subsetNode($pnode);
		if (defined(@{$pnode->{'Children'}})) {
			foreach (reverse @{$pnode->{'Children'}}) {
				push @queue, $_;
			}
		}
	}
}

sub completeReason {

	# Start with the root and do an postfix travel on the assumption tree and
	# explain each contradiction.
	my @queue = ();
	%explained = ();
	push @queue, $root;

	while ($#queue >= 0) { # untill the queue is empty
		my $node = pop @queue;
		
		if ($node->{'Atom'} == -1) { # Root
			foreach (reverse @{$node->{'Children'}}) {
				push @queue, $_;
			}
			next;
		}

		my $printText;
		if ($outputText) {
			$printText = $atoms[$node->{'Atom'}]->text(1);
		} else {
			$printText = $atoms[$node->{'Atom'}]->{'Name'};
			$printText .= " is ".$node->{'Truth'};
		}

		print "step $serial: Assuming $printText {\n\n";
		$explained{$node->getMyIndex} = $serial;
		$serial++;
		if (!($node->{'InValid'})) {
			# A valid assumption. So explain the assumption and the true
			# atoms in the assumption.
			my @tatoms = ();
			foreach (@{$node->{'ForcedAtoms'}}) {
				if ($_->{'Truth'}) {
					$_->getAtomsRequiredForReason(1, \@tatoms);
				}
			}
			&printReason(\@tatoms);
		} else {
			# 1. If this assumption has a contradiction, explain it.
			# 2. If this assumption produced the contradiction. 
			#	a. Explain the contradiction.
			#	b. Show the assumption is not true.
			
			if (defined($node->{'Rule'}) and $node->{'Rule'} > 0) { 
				# The contradiction is a cardinality atom
				my $Rule = $rules[$node->{'Rule'}+1];
				if ($Rule =~ /^(.*) , { (.*) }/) {
					# The contradiction is because of the cardinality
					# constraint
					my @body = split(/ /, $1);
					my (undef, undef, @head) = split(/ /, $2);
					my @finalSteps = ();
					foreach (@body) {
						my $tempNode = $node->getAtom($_);	
						if(defined($tempNode)) {
							my @tatoms = ();
							$tempNode->getAtomsRequiredForReason(1, \@tatoms);
							&printReason(\@tatoms);
							push @finalSteps, $explained{$tempNode->getMyIndex};
						}
					}

					foreach (@head) {
						my $tempNode = $node->getAtom($_);	
						if(defined($tempNode)) {
							my @tatoms = ();
							$tempNode->getAtomsRequiredForReason(1, \@tatoms);
							&printReason(\@tatoms);
							push @finalSteps, $explained{$tempNode->getMyIndex};
						}
					}
					my $prefix = $constraintLingo{$actualRules[$ruleIndices[$node->{'Rule'}+1]+1]};
					print "From steps ", join(',', @finalSteps), " and $prefix",
							" the assumption made at step ",
							$explained{$node->getMyIndex},
							" is invalid. }\n\n";
				}

			} else {
				foreach (@{$node->{'ForcedAtoms'}}) {
					if (defined($_->{'Contradiction'})) {
						my @tatoms = ();
						$_->getAtomsRequiredForReason(1, \@tatoms);
						&printReason(\@tatoms);

						my $thisIndex = $_->getMyIndex;
						my $tempAtom = $_->{'Contradiction'};
						my $tempIndex = $tempAtom->getMyIndex;
						if ($explained{$tempIndex}) {
							# Step 2b
							print "From steps ",
									$explained{$tempIndex}," and ",
									$explained{$thisIndex},
									" the assumption made at step ", 
									$explained{$node->getMyIndex},
									" is invalid. }\n\n";
							last;
						}
					} # If Contradicting atom
				} # End of forced atoms for contradictions
			}
		} # if invalid assumption
	
			
		# Note: I didn't do this kind of testing at other places in this code
		# somehow it seems to work with out the test. But for some reason
		# here the code breaks with out the test. At this point of time I
		# have no idea why it breaks, so I just added the test to avoid the
		# annoyance.
		if ($#{@{$node->{'Children'}}} > -1) {
			foreach (reverse @{$node->{'Children'}}) {
				push @queue, $_;
			}
		} else {
			# If the current assumption is the last of the sibilings,
			# perculate up the contradiction to the parent and it's parent
			# and so on.
			if ($node->{'InValid'}) {
				if ($node->{'ChildIndex'} ==
						$#{@{$node->{'Parent'}->{'Children'}}}) {
					my @perculateUpQueue = ();
					push @perculateUpQueue, $node->{'Parent'};
					while($#perculateUpQueue > -1) {
						my $pnode = pop @perculateUpQueue;
						next if ($pnode->{'Atom'} == -1);
						if ($pnode->{'InValid'}) {
							my @csteps = ();
							foreach (@{$pnode->{'Children'}}) {
								push @csteps, $explained{$_->getMyIndex};
							}
							print "The assumption made at step ",
									$explained{$pnode->getMyIndex}, 
									" is not valid because assumptions",
									" made at steps ", join(', ',@csteps),
									" are all invalid. }\n\n";
						} else {
							print "Assumption made at step ",
									$explained{$pnode->getMyIndex},
									" is valid. } \n\n";
						}
									
						if ($pnode->{'ChildIndex'} ==
								$#{@{$pnode->{'Parent'}->{'Children'}}}) {
							next if($pnode->{'Parent'}->{'Atom'} == -1);
							push @perculateUpQueue, $pnode->{'Parent'};
						}
					} # end of while
				} # if this is the last sibiling
			} else {
				my @perculateUpQueue = ();
				my $currentParent = $node;
				while ($currentParent->{'Branch'} != 1) {
					push @perculateUpQueue, 
							$explained{$currentParent->getMyIndex};
					$currentParent = $currentParent->{'Parent'};
				}
				push @perculateUpQueue,
						$explained{$currentParent->getMyIndex};
				print "At this stage we have all the assumptions made at steps ",
						join(", ", @perculateUpQueue), " are true. }\n\n";	
				if ($node->{'ChildIndex'} ==
						$#{@{$node->{'Parent'}->{'Children'}}}) {
					@perculateUpQueue = ();
					push @perculateUpQueue, $node;
					while ($#perculateUpQueue > -1) {
						my $pnode = pop @perculateUpQueue;
						next if ($pnode->{'Atom'} == -1);
						if ($pnode != $node) {
							print "Assumption made at step ",
									$explained{$pnode->getMyIndex},
									" is valid. } \n\n";
						} 
						if ($pnode->{'ChildIndex'} ==
								$#{@{$pnode->{'Parent'}->{'Children'}}}) {
							next if($pnode->{'Parent'}->{'Atom'} == -1);
							push @perculateUpQueue, $pnode->{'Parent'};
						}
					}
				}
			} # if the node is a valid assumption
		} # if there are no children
	} # end of main while
}

sub simpleReason {
	my @orderAtoms = ();

	%explained = ();
	foreach (@{$root->{'Children'}}) {
		next if ($_->{'InValid'});
		$_->simpleReason(0, \@orderAtoms);
	}
	%explained = ();
	&printReason(\@orderAtoms);
}

sub printReason {
	my $orderAtoms = shift;
	my $prevVisitedStep=0;
	my @steps = ();

	foreach (@{$orderAtoms}) {
		my $node = $_->{'Atom'};
		my $print ;
		my $prefix = "";
		my $explainedIndex;
		$explainedIndex = $_->{'Atom'}->getMyIndex;
		
		# Generate the prefix and the print statements
		if ($outputText == 1) { 
			$print = $atoms[$node->{'Atom'}]->text($node->{'Truth'});
			if ($node->{'Assumption'}) {
				$prefix = "Let's assume";
			} else {
				my $rule = $node->{'Rule'};
				$prefix = $constraintLingo{$actualRules[$ruleIndices[$rule+1]]}
													if(defined($rule));
				if ($prevVisitedStep > 0) {
					$prefix =~ s/, so//g;
					$prefix = $prefix . "and from step $prevVisitedStep, ";
				}
			}
		} else {
			$print = $atoms[$node->{'Atom'}]->{'Name'}." is ".$node->{'Truth'};
		}

		# print the result
		if (defined($explained{$explainedIndex})) {
			$prevVisitedStep = $explained{$explainedIndex};
			# print "".($tab x $_->{'Tab'}), "We already know ", "(step ", 
			# $explained{$_->{'Atom'}->{'Atom'}}, ") that $print.\n\n";
			next;
		}

		# Extra information
		if ($_->{'Tab'} == 1) {
			push @steps, $serial;
		} elsif ($_->{'Tab'} == 0 and $#steps > -1) {
			$prefix =~ s/, so//g;
			$prefix .= "(steps ". join(", ", @steps)."), so";
			@steps = ();
		}

		print "".($tab x $_->{'Tab'}), 
				"step $serial: $prefix $print.\n\n"; 
		$explained{$explainedIndex} = $serial;
		$prevVisitedStep = 0;
		$serial++;
	}
}

sub Node::setContext {
	my $self = shift;
	my @tatoms = ();
	$self->getAtomsRequiredForReason(0, \@tatoms);
	
	# Remove all the atoms except assumptions
	my @tassumptions = ();
	foreach (@tatoms) {
		my $tnode = $_->{'Atom'};
		if ($tnode->{'Assumption'}) {
			push @tassumptions, $tnode;
		}
	}
	# add the assumed atoms used to reason this atom to the context for the
	# parent and it's parent and so on
	my $node = $self->{'Parent'};
	while($node->{'Atom'} != -1) { # Go up till the root
		foreach (@tassumptions) {
			if ($_->{'Assumption'} and
					$_->{'Branch'} <= $node->{'Branch'} and
					$_ != $node) {
				$node->addToContext($_);
			}
		}
		$node = $node->{'Parent'};
	}

}

sub Node::simpleReason {
	my $self = shift;
	my $depth = shift;
	my $tatoms = shift;

	if ($self->{'Atom'} == -1) { # root
		foreach (@{$self->{'Children'}}) {
			next if ($_->{'InValid'});
			$_->simpleReason($depth, $tatoms);
		}
		return;
	}
	if ($self->{'InValid'}) {
		print $atoms[$self->{'Atom'}]->{'Name'}, 
			" is an invalid assumption, so there is no true branch in",
			" this sub assumption tree.\n";
		return;
	}

	# self is a valid assumption.
	# 1. Print all the true atoms due to this assumption.
	# 2. Print simple reason for the next true child.

	my %temp = ();
	$temp{'Atom'} = $self;
	$temp{'Tab'} = $depth;
	push @{$tatoms}, \%temp;
	$visited{$self->getMyIndex} = 1;
	
	foreach (@{$self->{'ForcedAtoms'}}) {
		if ($_->{'Truth'}) {
			$_->getAtomsRequiredForReason($depth, $tatoms);
		}
	}

	foreach (@{$self->{'Children'}}) {
		next if ($_->{'InValid'});
		$_->simpleReason($depth, $tatoms);
	}
}

sub Node::getAtomsRequiredForReason {
	my $self = shift;	
	my $depth = shift;
	my $tatoms = shift;

	my $visitedIndex = $self->getMyIndex;
	if ($visited{$visitedIndex}) {
		my %temp = ();
		$temp{'Atom'} = $self;
		$temp{'Tab'} = $depth;
		push @{$tatoms}, \%temp;
		return;
	}
	$visited{$visitedIndex} = 1;
	if ($self->{'Assumption'}) {
		my %temp = ();
		$temp{'Atom'} = $self;
		$temp{'Tab'} = $depth;
		push @{$tatoms}, \%temp;
	} else {
		# self is a forced atom and not an assumption
		my $Rule = $rules[$self->{'Rule'}+1];

		if ($Rule =~ /^, { (.*) }$/) { # cardinality constraint
			my ($min, $max, @values) = split (/ /, $1);
			foreach (@values) {
				next if ($_ eq "");
				next if ($_ eq $self->{'Atom'});
				my $node = $self->{'Parent'}->getAtom($_);
				next if (!defined($node));
				if ($self->{'Truth'} == 0) {
					# only get the true atom that force this atom to false
					next if ($node->{'Truth'} == 0);
				}
				$node->getAtomsRequiredForReason($depth+1, $tatoms);
			} 
		} elsif ($Rule =~ /^([0-9]+ )+, ([0-9]+) $/) { # head :- atom1, ...
			my ($body, $head) = ($1, $2);
			if ($self->{'Atom'} != $head) {
				my $node = $self->{'Parent'}->getAtom($head);
				$node->getAtomsRequiredForReason($depth+1, $tatoms)
							if (defined($node));
			}
			foreach (split(/ /, $body)) {
				next if ($_ == $self->{'Atom'});
				my $node = $self->{'Parent'}->getAtom($_);
				next if (!defined($node));
				$node->getAtomsRequiredForReason($depth+1, $tatoms);
			}
		} elsif ($Rule =~ /(.*) , $/) { # nohead :- body	
			foreach (split(/ /, $1)) {
				next if ($_ eq "" or $_ == $self->{'Atom'});
				my $node = $self->{'Parent'}->getAtom($_);
				next if (!defined($node));
				$node->getAtomsRequiredForReason($depth+1, $tatoms);
			}
		} elsif ($Rule =~ /^([0-9]+ )+, { ([0-9]+ )+}$/) {
			my (undef, undef, @headAtoms) = split(/ /, $2);

			my $isInBody = 0;
			foreach (split(/ /, $1)) {
				$isInBody = 1 if ($_ == $self->{'Atom'});
			}
			if (!$isInBody) {
				# Explain all the atoms in the body that force the head
				# to be true
				foreach (split(/ /, $1)) {
					next if($_ !~ /[0-9]+/);
					my $node = $self->{'Parent'}->getAtom($_);
					next if (!defined($node));
					$node->getAtomsRequiredForReason($depth+1, $tatoms);
				}
			}
			foreach (@headAtoms) {
				next if($_ !~ /[0-9]+/);
				next if($_ == $self->{'Atom'});
				my $node = $self->{'Parent'}->getAtom($_);
				next if (!defined($node));
				if (!$isInBody and $self->{'Truth'} == 0) {
					# only get the true atom that force this atom to false
					next if ($node->{'Truth'} == 0);
				}
				$node->getAtomsRequiredForReason($depth+1, $tatoms);
			}
		} elsif ($Rule =~ /(.*) , (.*)$/) { # disjunctive clause
			if ($1 ne "") {
				foreach (split(/ /, $1)) {
					next if($_ !~ /[0-9]+/ or $_ == $self->{'Atom'});
					my $node = $self->{'Parent'}->getAtom($_);
					next if (!defined($node));
					$node->getAtomsRequiredForReason($depth+1, $tatoms);
				}
			}
			foreach (split(/ /, $2)) {
				next if($_ !~ /[0-9]+/ or $_ == $self->{'Atom'});
				my $node = $self->{'Parent'}->getAtom($_);
				next if (!defined($node));
				$node->getAtomsRequiredForReason($depth+1, $tatoms);
			}
		} else {
			print "$Rule: Bummer\n";
			exit(1);
		}

		my %temp = ();
		$temp{'Atom'} = $self;
		$temp{'Tab'} = $depth;
		push @{$tatoms}, \%temp;
	}
}

# Print the assumption tree
sub Node::printAssumptionTree {
	my $self = shift;
	my $depth = shift;
	my $index = shift;

	if ($self->{'Atom'} == -1) {
		print "\nRoot\n";
		$depth = 0;
	} else {
		my $truth = $self->{'Truth'};
		my $valid = $self->{'InValid'} == 1 ? 'F':'T';
		my $num = $self->{'Branch'}.".".$index;
		print "|".("--" x $depth).">($valid) ($num) ".
				$atoms[$self->{'Atom'}]->{'Name'}."\n";
	}

	my $counter = 1;
	foreach my $child (@{$self->{"Children"}}) {
		$child->printAssumptionTree($depth+1, $counter);
		$counter++;
	}
}
# Subroutines for grounding
sub assimilateClauses { # Read all the groundClauses
	my ($file, $body, $head, $min, $max, $clause, $rule, $line);
	$serial = 1;
	$file = $grounderFile;
	open(CLAUSES, "<$file");
	my @bodies = ();
	my @heads = ();
	$body = $head = 0;
	$clause = 0;
	$min = $max = -1;
	$rule = 0;
	while (<CLAUSES>) {
		$line = $_;
		chop($line);
		next if($line eq "");
		next if($line =~ /(.*)aaa(.*)/);
		if($line =~ /New Clause ([0-9]*)/) {
			$body = 1;
			$head = 0;
			$clause = $1;
		}
		elsif ($line =~ /Head/) {
			$body = 0;
			$head = 1;
		}
		elsif($line =~ /Instantiate Rule id = ([0-9]*)\s*/) {
			$rule = $1;
		}
		elsif($line =~ /\s*End Clause\s*/) {
			my ($finalclause);
			next if(!(@bodies) and !(@heads));
			$finalclause = join(", ", @bodies) if(@bodies);
			$finalclause .= " -> " if(@bodies);
			$finalclause .= join(", ", @heads) if(@heads);
			$finalclause = "$min { ".$finalclause if($min > 0);
			$finalclause = $finalclause." } $max"  if($max > 0);
			$finalclause .= " .";
			$groundClauses{$clause} = $finalclause.":".$rule;
			push @{$AgroundClauses[$rule]}, $finalclause;
			
			$head = $body = 0;
			$min = $max = -1;
			@heads = @bodies = ();
			$finalclause = "";
		}
		elsif($line =~ /min=([0-9]+), max=([0-9]+)\s*/) {
			$min = $1;
			$max = $2;
		} elsif($line =~ /.*tmpsym = (.*) bdy.*/) {
			my $pred = $1;
			push @bodies, $pred if($body != 0);
			push @heads, $pred if($head != 0);
		} elsif($line =~ /Setting (.*) to (true|false) ([0-9]*)\./) {
			my $atom = $1;
			my $value = $2;
			my $cls = $3;
			$groundAtoms{$1} = $value.":".$cls;
			undef @toPrint;
			if ($visualTable) {
				my (undef, $cls) = split(/:/, $groundAtoms{$atom});
				my ($Clause, $Rule) = split(/:/,$groundClauses{$cls});
				# $constraintLingo{$actualRules[$Rule-1]};
				# $Rule--;
				if ($atom =~ /([a-z0-9]+)_([a-z0-9]+)\((\w+),(\w+)\)/) { 
					push @tableValues, "$3:$4:$1:$2:$value:1:$Rule:$Clause";
				} elsif ($atom =~ /([a-z0-9]+)_([a-z0-9]+)\((\w+)\)/) {
					push @tableValues, "$3:-1:$1:$2:$value:1:$Rule:$Clause";
				}					
				$idClass = $2;
			}
			&groundReason($atom, 0) if ($value eq "true");
			print join("\n", @toPrint);
		}
	}
} # assimilateClauses

sub groundReason {
	my ($atom, $level) = @_;
	my ($elements, $temp);
	if(defined($explained{$atom})) {
		my ($value, undef) = split(/:/, $groundAtoms{$atom});
		push @toPrint, "\n".($tab x $level).
				"We already know (step $explained{$atom}) that " .
				reword($atom, $value). "\n";
		return 0;
	}
	my ($value, $cls) = split(/:/, $groundAtoms{$atom});
	# There are two groundReasons why an atom is true
	# 1) It is in the set of those elements that are forced to be true.
	# 2) It is the head of the clause whose body groundAtoms are true.
	my ($Clause, $Rule) = split(/:/,$groundClauses{$cls});
	# print "\n" . $tab x $level " . $atom is $value from \n" . $tab x $level
	#		$actualRules[$Rule-1] . "\n" . $tab x $level;
	#		$constraintLingo{$actualRules[$Rule-1]}, "\n";
	if($Clause !~ /(.*) -> (.*) \./) { # No body
		if($Clause =~ /[0-9]* { (.*) } [0-9]* \./) { # Cardinality clause
			for $elements (split (/, /, $1)) {
				($temp, undef) = split(/:/, $groundAtoms{$elements});
				if ($value eq "true") {
					&groundReason ($elements, $level+1) if($temp =~ /false/);
				} else {
					&groundReason ($elements, $level+1) if($temp =~ /true/);
				}
			}
		}
	} elsif($Clause =~ /(.*) -> \./) { # Constraint
		# Check for the other predicate if it is true
	} else { # General Rule
		if ($Clause =~ /(.*) -> [0-9]* { (.*) } [0-9]* \./) { 
			# nothing to do here
		} elsif ($Clause =~ /(.*) -> (.*) \./) { # single atom in head
			if ($value eq "true") {
				# The element must be in the head
			} else {
				# The element must be in the body
				# and the rest of the elements are true 
				# and the head is false
				for $elements (split (/, /, $1)) {
					&groundReason ($elements, $level+1) if($atom ne $elements);
				}
				for $elements (split (/, /, $2)) {
					&groundReason ($elements, $level+1);
				}
			}
		} # single atom in head
	} # General Rule
	my $explain = $constraintLingo{$actualRules[$Rule]};
	$explain .= $tab x $level if $explain =~ /\n$/;
	push @toPrint, "\n" . $tab x $level . "step $serial:  " . $explain .
		reword($atom, $value) . ".\n";
	$explained{$atom} = $serial;
	$serial += 1;
} # groundReason

# main program here
$tab = "    ";
$prune = 0;
$visualTable = 0;
$visual = 0;
$simpleSolution = 0;
$completeSolution = 0;
$assumptionTree = 0;
$outputText = 1;

# Files
$rulesFile = "rules.tmp";
$grounderFile = "/tmp/ground.tmp";
$solverFile = "/tmp/solver.tmp";
$combineFile = "facts.tmp_combine.tmp.aspps";
$puzzleFile = "";

my $counter = 0;
foreach (@ARGV) {
	$counter++;
	next if($counter == 0);
	$visual = 1 if ($_ eq "-v");
	$visualTable = 1 if ($_ eq "-t");
	$simpleSolution = 1 if ($_ eq "-s");
	$completeSolution = 1 if ($_ eq "-c");
	$outputText = 0 if ($_ eq "-atoms");
	$assumptionTree = 1 if ($_ eq "-assumptions");
	$prune = 1 if($_ eq "-p");
}

assimilatePuzzle();
assimilateRules();
assimilateClauses();
assimilateGrounding();
assimilateSolution();
if ($completeSolution or $simpleSolution) { # Generate English Text
	$root->printAssumptionTree if($assumptionTree);
	if ($simpleSolution) {
		&simpleReason;	
	} else {
		&completeReason;
	}
} elsif ($visualTable) { # Show the solution Table only
	use Tk;
	use Tk::Font;
	my %visitedAtoms = ();
	my @queue = ();
	push @queue, $root;

	while ($#queue >= 0) {
		my $node = pop @queue;
		if ($node->{'Atom'} != -1 and $node->{'InValid'} == 0) {
			if ($atoms[$node->{'Atom'}]->{'Name'} =~ 
					/([a-z0-9]+)_([a-z0-9]+)\((\w+),(\w+)\)/) {
				push @tableValues, "$3:$4:$1:$2:true:2:-1";
			} elsif ($atoms[$node->{'Atom'}]->{'Name'} =~ 
					/([a-z0-9]+)_([a-z0-9]+)\((\w+)\)/) {
				push @tableValues, "$3:-1:$1:$2:true:2:-1";
			}
			$visitedAtoms{$node->{'Atom'}} = 1;
			foreach (@{$node->{'ForcedAtoms'}}) {
				next if ($visitedAtoms{$_->{'Atom'}});
				my $tvalue = ($_->{'Truth'} == 1)?"true":"false";
				my $clause;
				if (defined($_->{'FAtom'})) {
					my $catom = $atoms[$_->{'Atom'}]->{'Name'};
					my $fatom = $atoms[$_->{'FAtom'}]->{'Name'};
					foreach my $groundClause (@{$AgroundClauses[$ruleIndices[$_->{'Rule'}+1]]}){
						if (index($groundClause,$catom) > -1 and
							index($groundClause,$fatom) > -1)  {
							$clause	= $groundClause;
							last;
						}
					}	
				}
				if ($atoms[$_->{'Atom'}]->{'Name'} =~ 
						/([a-z0-9]+)_([a-z0-9]+)\((\w+),(\w+)\)/) {
					my $input = "$3:$4:$1:$2:$tvalue:0:".$_->{'Rule'};
					$input .= ":$clause" if(defined($clause));
					push @tableValues, $input;
				} elsif ($atoms[$_->{'Atom'}]->{'Name'} =~ 
						/([a-z0-9]+)_([a-z0-9]+)\((\w+)\)/) {
					my $input = "$3:-1:$1:$2:$tvalue:0:".$_->{'Rule'};
					$input .= ":$clause" if(defined($clause));
					push @tableValues, $input;
				}
				$visitedAtoms{$_->{'Atom'}} = 1;
			}
		}
		foreach (@{$node->{'Children'}}) {
			push @queue, $_;
		}
	}
	&showTable($ARGV[0]);
	MainLoop;
} elsif ($visual) { # Show the visual tree structure
	use Tk;
	use Tk::TreeGraph;
	showAssumptionTree();
	MainLoop;
} elsif ($assumptionTree) {
	$root->printAssumptionTree if($assumptionTree);
}
