#!/usr/bin/perl -w

=head1 NAME

tester: test out various logic engines for computing Constraint Lingo programs. 

=head1 SYNOPSIS

tester.pl [doall]

This program uses hard-wired information to try out various combinations of

1.  Input programs written in Constraint Lingo

2.  Converters from Constraint Lingo to logic back-ends

3.  Logic back ends

=head1 INPUT/OUTPUT

Input (from stdin) has lines like this:

100.in:
	sm2: 374 atoms, 2597 rules, 0 branches, 1 solutions, 0.180 seconds
	sm2c: 228 atoms, 1847 rules, 20 branches, 1 solutions, 0.200 seconds
	sm2t: 178 atoms, 483 rules, 2 branches, 1 solutions, 0.040 seconds
	dcs: 191 atoms, 2255 clauses, 0 branches, 1 solutions, 0.01 seconds
	ndcs: 203 atoms, 1238 clauses, 250 branches, 1 solutions, 0.09 seconds
	ecl: 0 branches, 1 solutions, 0.06 seconds
	dlv: 8 branches, 1 solutions, 0.09 seconds

The first line indicates which puzzle to run, and the subsequent lines show
what methods to try.  Any method that does not currently show 1 solution is
rerun.  If a doall parameter is given, then all methods are always rerun.

The output is a file in the same format.  Any line that doesn't follow
these conventions is a comment line and is preserved.

Author: Raphael Finkel 11/2001

=cut

use strict; # catch silly declaration errors
use vars qw(@CLProgs @Converters $user %solutions);
$user = $ENV{'USER'};

@CLProgs = glob "*.in";
# print "programs to test: ", join(", ", @CLProgs), "\n";

sub runIt {
	my ($CLProg, $nickname) = @_;
	my ($toRun, $result, $solution, $summary, @lines, $runOutput);
	# print "make $nickname GOAL=$CLProg | tail -1\n";
	$runOutput = `make $nickname GOAL=$CLProg`;
	$solution = (split(/\n\n/, $runOutput))[1];
	# print "solution is \n$solution\n";
	@lines = split(/\n/, $runOutput);
	$summary = $lines[$#lines];
	# print "summary is $summary\n";
	if ($summary =~ /$nickname:/) {
		print "\t$summary";
	} else {
		print "\t$nickname: failed";
	}
	if (defined($solutions{$CLProg})) {
		print "; solution differs"
			unless canonical($solution) eq $solutions{$CLProg};
		# print "solution to $CLProg matches\n";
	} else {
		$solutions{$CLProg} = canonical($solution);
		# print "solution to $CLProg stored\n";
	}
	print "\n";
} # runIt

sub reTest {
	my ($line, $CLProg, $converter, $nickname, $front, $back);
	while (defined($line = <STDIN>)) { # one line
		if ($line =~ /#/) { # comment
			print $line;
		} elsif ($line =~ /^([\w.]+):\s*$/) { # names a puzzle
			$CLProg = $1;
			print $line;
			$CLProg =~ s/\.in//;
		} elsif ($line =~ / 1 solutions/ and $line !~ /differs/ and
		!defined($ARGV[0])) {
			# ordinary success
			print $line;
		} elsif ($line =~ /^\s*$/) { # separator line
			print $line;
		} elsif ($line =~ /\s*([\w]+):/) { # failure: read nickname
			$nickname = $1;
			# print STDERR "got to run $nickname: $CLProg\n";
			runIt($CLProg, $nickname);
		} else { # some other sort of message
			print $line;
		}
	} # one line
} # reTest

sub canonical {
	# convert a solution into canonical form.
	my ($solution) = @_;
	# print "\nlooking at solution\n$solution\n";
	my @lines = split(/\n/, $solution);
	my @classes = split(/\s+/, $lines[0]);
	my @sortedClasses = sort @classes;
	my ($index1, $index2, $line, @result, @values, @outLine);
	my @order = ();
	# compute reordering
	for ($index1 = 0; $index1 <= $#classes; $index1 += 1) {
		for ($index2 = 0; $index2 <= $#classes; $index2 += 1) {
			if ($classes[$index2] eq $sortedClasses[$index1]) {
				push @order, $index2;
				next;
			}
		} # for $index2
	} # for $index1
	# print "classes are ", join(" ", @classes), "; order is ",
	#	join(" ", @order), "\n";
	for $line (@lines[2 .. $#lines]) {
		@values = split /\s+/, $line;
		@outLine = ();
		for ($index1 = 1; $index1 <= $#values; $index1 += 1) {
			push @outLine, @values[$order[$index1]];
		}
		push @result, join(" ", @outLine);
	}
	@result = sort @result;
	# print "\n\t\t", join(" ", @result), "\n";
	return join(" ", @result);
} # canonical

$| = 1; # flush immediately
reTest();
