/* sales.c 
 * Travelling salesman problem
 * Find the shortest path which allow every site to be visited 
 */ 

#include <stdio.h>
#include "conventions.h"  
#include "applic.h"  
#include "node.h"  

char Title[20] = "Salesman";  

#define T       	1 
#define F       	0 

#define MAXSIZE 	20  		/* Maximum number of sites to be visited */
#define MAXDIST 	100 		/* Maximum distance between sites */  

#define INFINITY (MAXDIST * MAXSIZE + 1)

/* structure definitions */

typedef int PathTaken[MAXSIZE + 1];  

/* Structure defining the problem */
typedef struct {  
	/* inherited constant */ 
	int Sites;          	/* Size - How many sites to visit */

	/* inherited level-dependent constant */ 
	int Depth;           	/* The generation we are at */
	int DistFromRoot;   	/* How far have we travelled */
	PathTaken Path;     	/* Route to this site */
 
	/* synthesized */ 
	int ShortestDist;   	/* Shortest Distance to all sites */
	PathTaken BestPath; 	/* Path taken which is ShortestDist long */
	int Leaves;			/* how many leaves were searched */
	int Nodes;			/* how many nodes were searched */

	/* other variables */ 
	int NextSite[MAXSIZE]; 	/* list of unvisited sites */
	int Children;       	/* How many children does this parent have? */
} 
ProblemType;  

/* structure for information we need to share with other workers */
typedef struct{
	int Min;				/* Minimum-length path known to us */
} InfoType;

typedef struct {
	int From, To;
	int Cost;
} EdgeType;

int  Distance[MAXSIZE + 1][MAXSIZE + 1];	/* our graph representation of
									paths and distances */
InfoType BestInfo;			  
int Initialized = F;

void ApplicInit(PSizePtr)  
int *PSizePtr;  
/*  Intialize any data structures and report the size of ProblemType
 *  to make sure that DIBs message-size assumptions are not  
 *  violated. We generate here our graph for distances 
 */ 
{  
	unsigned int Seed, From, To;
	extern int Nugget_MyNode;

	*PSizePtr = sizeof(ProblemType);  

	if (!Initialized) {
		Seed = 20; /* same each time to make it easier to debug */
		srand(Seed);
		for(From=1;From<=MAXSIZE;From++){
			for(To=1;To<=MAXSIZE;To++){
				if (From == To)
					Distance[From][To] = 0;
				else
					Distance[From][To] = rand()%(MAXDIST + 1);
			}
		}
		/* pseudo-data:  Distance from 0 is always 0 */  
		for(To=1;To<=MAXSIZE;To++)
			Distance[0][To] = 0;

		Initialized = T;
	}

	BestInfo.Min = INFINITY;
} /* ApplicInit */

void FirstProb(ProbPtr, Size)  
ProblemType *ProbPtr;  
int Size; 
/* Form the root node, so we can get started */ 
{ 
	int Pos; 
	int From,To;

	ProbPtr->Sites        = Size;  
	ProbPtr->Depth        = 0;  
	ProbPtr->Children     = 0; 
	ProbPtr->DistFromRoot = 0;  
	ProbPtr->ShortestDist = INFINITY; 
	ProbPtr->Leaves       = 0; 
	ProbPtr->Nodes       = 0; 
	for(Pos=1; Pos<=ProbPtr->Sites; Pos++)
		ProbPtr->NextSite[Pos] = Pos;

	/* print the graph we have generated */

	if (Nugget_MyNode==0){
		for(From=1;From<=Size;From++){
			for(To=1;To<=Size;To++){
				printf("%3d%c",Distance[From][To], 
					(To==Size) ? '\n' : ' ');
			}
		}
	}
} /* FirstProb */

void Generate(DonePtr, ParentPtr, ChildPtr)  
Boolean *DonePtr;  
ProblemType *ParentPtr, *ChildPtr;  
/*  Function to generate a child or call it a day for the parent,
 *  if the parent has produced all the children it wants.
 *  Note that this time's child is later a parent. Also that
 *  if a parent isn't done here, the child here will be a sibling
 *  to child the next time the function is called with the same
 *  parent and that the new child will be initialized with the
 *  first child's values. 
 */
{  
	int Pos, Child, NextDist[MAXSIZE + 1], Seq, Found;   

	if (ParentPtr->DistFromRoot +
	  MSTree(ParentPtr->NextSite, ParentPtr->Sites-ParentPtr->Depth)
		>= BestInfo.Min) {
		/* got a cutoff; already travelled further than a path some
		worker knows about through all sites */
		*DonePtr = DONE;
		return;
	} /* got a cutoff*/

	if (ParentPtr->Depth == ParentPtr->Sites) {  /* leaf */
		*DonePtr = DONE;  
		ParentPtr->ShortestDist = ParentPtr->DistFromRoot;
		ParentPtr->Leaves = 1;
		bcopy(ParentPtr->Path, ParentPtr->BestPath,
			sizeof(ParentPtr->BestPath));

		if (ParentPtr->ShortestDist < BestInfo.Min) { 
			/* better cutoff */
			BestInfo.Min = ParentPtr->ShortestDist;
			BroadcastInfo(&BestInfo);
		} /* better cutoff */

		ReportResult(ParentPtr);  
	}  /* leaf */
	else if (ParentPtr->Children == ParentPtr->Sites - ParentPtr->Depth) {
		/* Aged parent: finished generating children */
		*DonePtr = DONE;
	} /* aged parent */
	else { /* must generate a child */
		ParentPtr->Children++; 
		if (ParentPtr->Children == 1) {  /* first child */
			for (Pos=1; Pos<=ParentPtr->Depth; Pos++) 
				ChildPtr->Path[Pos] = ParentPtr->Path[Pos];
			ChildPtr->Sites = ParentPtr->Sites; 
			ChildPtr->Depth = ParentPtr->Depth + 1;
		} /* first child */		

		/* Assign attributes that are not inherited from first child */
		ChildPtr->ShortestDist = INFINITY;
		ChildPtr->Leaves       = 0;
		ChildPtr->Nodes        = 1;
		ChildPtr->Children     = 0;
		ChildPtr->Path[ChildPtr->Depth] = 
		  ParentPtr->NextSite[ParentPtr->Children]; /* Site this child is */
		ChildPtr->DistFromRoot = ParentPtr->DistFromRoot +
		  Distance[ParentPtr->Path[ParentPtr->Depth]]
				  [ChildPtr->Path[ChildPtr->Depth]];
		Found = 0;
		for (Seq = 1; Seq <= ParentPtr->Sites-ParentPtr->Depth; Seq++) {
			/* give Child a reasonable NextSite array */
			if (!Found && ParentPtr->NextSite[Seq] ==
			  ParentPtr->NextSite[ParentPtr->Children])
				Found = 1;
			else ChildPtr->NextSite[Seq-Found] = ParentPtr->NextSite[Seq];
		}
		*DonePtr = NOTDONE; 
	} /* generate a child */
} /* Generate */


void PrintProblem(ProbPtr)  
ProblemType *ProbPtr;  
/*  Invoked for each node after it has been generated
 *  if full reporting has been requested 
 */ 
{  
	int Pos;  

	printf("Node[%d]: Route [",Nugget_MyNode);
	for (Pos = 1; Pos <= ProbPtr->Depth; Pos++) 
		printf(" %d",ProbPtr->Path[Pos]); 
	printf(" ] has distance %d\n",
	ProbPtr->DistFromRoot);

} /* PrintProblem */

void Combine(ChildPtr,ParentPtr,UpdatePtr,UpdateUpPtr)  
ProblemType *ChildPtr, *ParentPtr;  
Boolean *UpdatePtr, *UpdateUpPtr;  
/*  The child has just finished; record any information to be passed
 *  up the tree to the parent.  If the parent's live children need
 *  to be notified asynchronously of parent's new situation, set 
 *  Update to true 
 */ 
{ 
	ParentPtr->Leaves += ChildPtr->Leaves;
	ParentPtr->Nodes += ChildPtr->Nodes;

	if (ChildPtr->ShortestDist < ParentPtr->ShortestDist) {
		/* This is the first child or a child showing a 
		             shorter root */
		ParentPtr->ShortestDist = ChildPtr->ShortestDist;
		bcopy(ChildPtr->BestPath, ParentPtr->BestPath,
		sizeof(ParentPtr->BestPath));
	} 

	/* Note that this child could contain a shorter path discovered
	   on another machine.  Therefore, compare it to BestInfo.Min and
	   if it is shorter, assign that value to BestInfo.Min.  Remember
	   that combine is called when a single child is done and not
	   when all leaves are reached, so this improved BestInfo.Min can
	   be used while the rest of the tree is still growing. */

	if (ParentPtr->ShortestDist < BestInfo.Min) {
		BestInfo.Min = ParentPtr->ShortestDist;
		BroadcastInfo(&BestInfo);
	} /* ParentPtr->ShortestDist < BestInfo.Min */

	*UpdatePtr = false;  
	*UpdateUpPtr = false;  
} /* Combine */

void PrintAnswer(ProbPtr)  
ProblemType *ProbPtr;  
/* intermediate answers (full reports) to the host */ 
{ 
	int Site, Leaves, Nodes, NextInt, Births, Generated;
	printf("The best path appears to be:\n");
	for(Site=1; Site<=ProbPtr->Sites; Site++){
		printf("%d ",ProbPtr->BestPath[Site]);
	}
	printf("\n");
	printf("The distance travelled in this case is %d.\n", 
		ProbPtr->ShortestDist);
	printf("We searched %d nodes and %d leaves.\n", 
		ProbPtr->Nodes, ProbPtr->Leaves);
	Leaves = 1;
	for (NextInt=1; NextInt <= ProbPtr->Sites; NextInt++)
		Leaves *= NextInt;
	Nodes = ProbPtr->Sites;
	if (Nodes != 1) {
		Generated = Nodes; 
		for (Births = ProbPtr->Sites - 1; Births != 1; Births--){
			Generated = Generated*Births;
			Nodes += Generated;
		}
		Nodes += Generated;
	}
	printf("%s %d %s %d %s\n",
		"A full search would have searched",
		Nodes,
		"nodes, and",
		Leaves,
		"leaves.");
	printf("Hence, we searched %.2f%% of the total nodes",
		(float)ProbPtr->Nodes/Nodes*100);
	printf(" and %.2f%% of the total leaves.\n",
		(float)ProbPtr->Leaves/Leaves*100);	
} /* PrintAnswer */

void AcceptAnswer(ProbPtr)  
ProblemType *ProbPtr; 
/* Info generated inside tree is available */  
{  
	PrintProblem(ProbPtr);  
}  

void AcceptRoot(ProbPtr)  
ProblemType *ProbPtr;  
/* When problem is finished. This is invoked at the host. */ 
{  
	PrintAnswer(ProbPtr);  
}  

void UseNewInfo(InfoPtr)
InfoType *InfoPtr;
/* How to use globally broadcasted information */
{
	if (InfoPtr->Min < BestInfo.Min)
		BestInfo.Min = InfoPtr->Min;
} /* UseNewInfo */

void PrintHelp()
/* Print any help information */
{
	printf("This is the travelling saleman problem\n");
} /* PrintHelp */

void ConvertProblemIn( External, Internal )
ProblemType *External, *Internal;
/* Convert incoming problem into local representation */
{
	int i;

	Internal->Sites = ntohl(External->Sites);
	Internal->Depth = ntohl(External->Depth);
	Internal->DistFromRoot = ntohl(External->DistFromRoot);
	Internal->ShortestDist = ntohl(External->ShortestDist);
	Internal->Leaves = ntohl(External->Leaves);
	Internal->Nodes = ntohl(External->Nodes);
	Internal->Children = ntohl(External->Children);
	Internal->Sites = ntohl(External->Sites);
	Internal->Sites = ntohl(External->Sites);

	for(i=0;i<=MAXSIZE;i++)
		Internal->Path[i] = ntohl(External->Path[i]);

	for(i=0;i<=MAXSIZE;i++)
		Internal->BestPath[i] = ntohl(External->BestPath[i]);

	for(i=0;i<MAXSIZE;i++)
		Internal->NextSite[i] = ntohl(External->NextSite[i]);
}

void ConvertProblemOut( Internal, External )
ProblemType *External, *Internal;
/* Convert outgoing problem into network representation */
{
	int i;

	External->Sites = htonl(Internal->Sites);
	External->Depth = htonl(Internal->Depth);
	External->DistFromRoot = htonl(Internal->DistFromRoot);
	External->ShortestDist = htonl(Internal->ShortestDist);
	External->Leaves = htonl(Internal->Leaves);
	External->Nodes = htonl(Internal->Nodes);
	External->Children = htonl(Internal->Children);
	External->Sites = htonl(Internal->Sites);
	External->Sites = htonl(Internal->Sites);

	for(i=0;i<=MAXSIZE;i++)
		External->Path[i] = htonl(Internal->Path[i]);

	for(i=0;i<=MAXSIZE;i++)
		External->BestPath[i] = htonl(Internal->BestPath[i]);

	for(i=0;i<MAXSIZE;i++)
		External->NextSite[i] = htonl(Internal->NextSite[i]);
}

void ConvertInfoIn( External, Internal)
InfoType *External, *Internal;
/* Convert incoming information into local representation */
{
	Internal->Min = htonl(External->Min);
}

void ConvertInfoOut( Internal, External )
InfoType *External, *Internal;
/* Convert outgoing information into network representation */
{
	External->Min = htonl(Internal->Min);
}

/* Auxillary Routines for finding Minimum spanning tree */

void MST_QHelp(First, Last, Arcs)
int First, Last;
EdgeType Arcs[];
{
	int SplitPoint;
	void MST_Split();
#define SMALL 10
	if (First < Last - SMALL) {
		MST_Split(First, Last, &SplitPoint, Arcs);
		MST_QHelp(First, SplitPoint - 1, Arcs);
		MST_QHelp(SplitPoint + 1, Last, Arcs);
	}
} /* MST_QHelp */

void MST_QSort(First, Last, Arcs)
int First, Last;
EdgeType Arcs[];
{
	int Fore, Aft;
	EdgeType TmpArc;
	MST_QHelp(First, Last, Arcs);
	Arcs[0].Cost = -1;
	for (Aft = 2; Aft <= Last; Aft++) {
		TmpArc = Arcs[Aft];
		for (Fore = Aft - 1; TmpArc.Cost < Arcs[Fore].Cost; Fore--) 
			Arcs[Fore+1] = Arcs[Fore];
		Arcs[Fore+1] = TmpArc;
	}
} /* MST_QSort */

void MST_Split(First, Last, SplitPoint, Arcs)
int First, Last, *SplitPoint;
EdgeType Arcs[];
{
	int SplitValue, Unknown;
	EdgeType TmpArc;
	SplitValue = Arcs[First].Cost;
	*SplitPoint = First;
	for (Unknown = First + 1; Unknown <= Last; Unknown++){
		if (Arcs[Unknown].Cost < SplitValue) {
			(*SplitPoint)++;
			TmpArc = Arcs[*SplitPoint];
			Arcs[*SplitPoint] = Arcs[Unknown];
			Arcs[Unknown] = TmpArc;
		}
	}
	TmpArc = Arcs[First];
	Arcs[First] = Arcs[*SplitPoint];
	Arcs[*SplitPoint] = TmpArc;
} /* MST_Split */

int Parent[MAXSIZE];
	/* index into array; negative means a weight of subtree */

Boolean SameSet(From, To)
int From, To;
/* returns whether From and To are in the same set with respect to the
Parent array.  If not, puts them in the same set.  Trival path compression. */
{
	int AncestorFrom, AncestorTo, OldWeight;
	for (AncestorFrom = From; Parent[AncestorFrom] > 0; AncestorFrom =
		Parent[AncestorFrom]);
	if (AncestorFrom != From) Parent[From] = AncestorFrom; /* compress */
	for (AncestorTo = To; Parent[AncestorTo] > 0; AncestorTo =
		Parent[AncestorTo]);
	if (AncestorTo != To) Parent[To] = AncestorTo; /* compress */
	if (AncestorTo == AncestorFrom) return T;
	if (Parent[AncestorFrom] < Parent[AncestorTo]) { /* From is weightier */
		OldWeight = Parent[AncestorTo];
		Parent[AncestorTo] = AncestorFrom;
		Parent[AncestorFrom] += OldWeight;
	} else { /* To is weightier */
		OldWeight = Parent[AncestorFrom];
		Parent[AncestorFrom] = AncestorTo;
		Parent[AncestorTo] += OldWeight;
	} /* To is weightier */
	return F;
} /* SameSet */

int MSTree(NextSite, Length)
int NextSite[], Length;
/*  Return the total path length of the minimal-cost spanning tree 
 *  among the sites listed in NextSite, which runs from 1 .. Length.  
 *  The sites are sorted by Union-find with path compression to check 
 *  for cycles. 
 */
{
	int Site;
	int EdgeCount = 0;
	EdgeType Edge[MAXSIZE * (MAXSIZE-1) / 2];
	int EdgeSeq;
	 
	{ /* build a list of edges */
		EdgeType *EdgePtr = &Edge[1];
		int From, To;
		for (From = 1; From <= Length; From++) {
			int f = NextSite[From];
			for (To = From+1; To <= Length; To++) {
				int t = NextSite[To];
				EdgePtr->From = f;
				EdgePtr->To = t;
				EdgePtr->Cost = Distance[f][t] < Distance[t][f] ?
					Distance[f][t] : Distance[t][f];
				EdgePtr++;
				EdgeCount++;
			} /* for To */
		} /* for From */
	} /* build list of edges */
	MST_QSort(1, EdgeCount, Edge);
	{ /* pick elements from list of edges */
		int Answer = 0, AnswerCount = 0;
		for (Site = 1; Site <= Length; Site++)
			/* initialize sets for union-find */
			Parent[Site] = -1;
		for (EdgeSeq = 1; EdgeSeq <= EdgeCount; EdgeSeq++) {
			if (!SameSet(Edge[EdgeSeq].From, Edge[EdgeSeq].To)) {
				/* belongs in spanning tree */
				Answer += Edge[EdgeSeq].Cost;
				AnswerCount++;
				if (AnswerCount == Length-1) return Answer;
			} /* belongs in spanning tree */
		} /* for Edge */
	} /* pick elements */
	return 0;
} /* MSTree */
