/*
 * Revision 1.10.1.1  92/05/01  11:16:10  jrezaie
 * When a node on the stack needs to be suspended, we put all its ancestors
 * which are also on the stack, in the list. But we make them available not to
 * be given away but to worked on by us(Suspended = SUSPWAS).
 * 
 * Revision 1.10  92/03/31  08:25:42  jrezaie
 * -MORELATER checked in BackTrack both going up and down
 * -ROOTID reintialized when leaving BackTrack
 * In TellParent, if parent has no unsolved children, it is checked fo
 * SUSPWAS inaddition to SUSPNOW
 * -When a node is passed to several machines, the middle machines keep
 * a copy of it so that they can pass update messages to the new owners
 * 
 * Revision 1.9  92/02/25  15:05:34  jrezaie
 * WorkFract is -1 for all machines
 * 
 * Revision 1.8  92/02/20  09:29:06  jrezaie
 * Made changes to BackTrack taking care of the case when Generate returns
 * MORELATER.  It puts all the nodes from GiveAwayDepth to Depth into the
 * work list and marks them as suspended.
 * Changes also were made to DoUpdate and case Updating of CheckIncoming
 * so that if an update message arrives, it unsuspends the node.
 * 
 * Revision 1.7  92/01/26  11:28:26  jrezaie
 * Changed WorkFract to 5 (give away half of the problems) for all machines
 * except machine 1.
 * 
 * Revision 1.6  92/01/21  11:40:01  jrezaie
 * The Update problem was corrected.  Problem was that the desc. of a node
 * on the stack were not updated.  This occurs only when the node is given
 * to us (its parent on the list is null).  Doupdate checks to see if
 * ChildPtr->Parent == ParentPtr (given to DoUpdate).  In the case mentioned
 * above, the nodes parent is null and therefore it does not go farther to
 * see if any desc. exist on the stack.  Correction: I added the check for
 * existance of desc. on stack and the update to CheckIncoming (case Updating).
 * 
 * Revision 1.5  91/11/01  14:05:45  jrezaie
 * All changes for UpdateUp has been made
 * 
 * Revision 1.4  91/10/08  14:03:06  jrezaie
 * After every call to UpdateUp, if the parent was suspended (SUSPNOW)
 * it was reactivated (SUSPWAS).
 * 
 * Revision 1.3  91/10/08  11:53:49  jrezaie
 * Added case UpdatingUp and a few other corrections
 * 
 * Revision 1.2  91/10/07  16:32:40  jrezaie
 * Added DoUpdateUp routine, made other changes to TellPeerUpdating
 * Need to make changes to the message type UpdatingUp in HandleRequest.
 * 
BackTrack is the main procedure.  It goes up and down the tree until it is
finished.  Whenever it moves up, it checks for incoming messages, calling
CheckIncoming if reasonable.  That routine dispatches such messages to various
routines, based on their Kind field.  Most difficult is a request for work;
HandleRequest is called for that.  It shares various global variables with
BackTrack.  On the side (mostly in other files) are routines for manipulating
the Work queue, which is the data structure used to keep track of all explicit
work (the current search is in a stack called CurrentProblem) and for keeping
track of the serial number of explicit nodes (Dewey numbers).  All globally
known declarations are in the .h files.

This is a descendent of 06.8.7.

Things to do:

Changes from previous version:
	1.  Generate is simpler:  only 3 arguments
	2.  Generate may return MORELATER in addition to DONE and NOTDONE
	3.  Major datastructure rework.  
	    WorkGiven and WorkGotten are now collapsed
	    into a single list.  
	    A given problem has only one entry on that list,
	    except if given away, both source and destination have a copy.  
	    This change allowed lots of simplifications.

Bugs:

	We are giving away SUSPWAS problems, which are not pristine.  
	We count them into AvailWorkCount, and they can be the only 
	entries there.  Perhaps best not to count them there?

	Can enter TellParent with a locally generated problem, null ParentPtr.
	Perhaps it was given away and eventually came back home. Just
	cancelling the problem can be ok and can lead to wrong # of solutions
	or traps elsewhere.
*/

/* Manager for distributed backtracking application */

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "conventions.h"
#include "nugget.h"
#include "dewey.h"
#include "host.h"
#include "applic.h"
#include "node.h"
#include "sap.h"
#include "nodework.h"
#ifdef newnode
	#include<unistd.h>	
#endif

extern int srand();

/* CONSTANTS */
#	define MaxDeweyReport 4
	/* the maximal size problem whose completion we report back to the
	host */

#	define MaxTimingDepth 15
	/* the maximal depth of different kinds of timing */
#ifdef notneeded
#	define LOOPLIMIT 50000 /* number of iterations before asking again */
#else
#	define LOOPLIMIT 50000 /* number of iterations before asking again */
#endif

/* MESSAGE TYPES */
#	define Request 0
#	define Work 2
#	define GlobalInfo 3
#	define Result 4
#	define Terminate 5
#	define Synch 6
#	define Repeating 7
#	define Updating 8
#	define UpdatingUp 9
#ifdef pvmred
#	define BadTask 10
#endif

/* DIRECTION TYPE */
#	define Up 10
#	define Down 11

/* TYPES */
	typedef int HelperListType[MAXWORKERS+1];

/* VARIABLES */
	Boolean Termination;
	/* I, 1, have discovered termination or 1 told me */
	int Debug;

	/* HELPER VARIABLES */
	HelperListType HelperList;
	/* initialized to 1, 2, ..., Nugget_MyNode-1, MyNode+1, ...,
	MAXWORKERS */
	int NextToAsk;

	/* COMMUNICATION VARIABLES */
	int
		PartitionSize,
		NumHelpers,
		/* how many neighbors we send requests to when out of work.  */
		ProblemNumber = 0,
		LoopLimit = LOOPLIMIT, 
		/* iterations before asking again for help */
		LocalSequence; 
		/* Local message sequence number, initialized to 1 */
		Boolean HaveBeenIdle;
		/* We send more requests while idle only when false */

#ifdef newnode
		int AddedLater;
		int StartedWorkers=0;
#endif

	/* SPECIFICATION OF THE PROBLEM WE ARE WORKING ON. */
		int ProblemSize;
		AnswerSpecification AnswerSpec;
		Boolean GlobalInUse; 	/* I may send to dest == -1 
							to go to 'anywhere' */
		int WorkFract; 		/* -1 to 10 ; currently set to -1 */
		int LastUniqueID;

	/* TIMING VARIABLES.  ALL TIME IS IN MICROSECONDS */
		TimeType Time;
		int TimeTemp; /* 0 .. 3+MaxRedundancy */
		long OldTime1; /* micros */
		int TimingStack[MaxTimingDepth];
		int TimingStackTop;

		/* TIMING MACROS */
#		define StartTiming(What) \
			TimingStackTop = 0; \
			TimingStack[TimingStackTop] = What; \
			OldTime1 = Nugget_ElapsedTime()
#		define PushTiming(What) \
			OldTime1 = Nugget_ElapsedTime() - OldTime1; \
			TimeTemp = TimingStack[TimingStackTop]; \
			if (TimeTemp <= 3+MaxRedundancy) { \
				Time[TimeTemp] += OldTime1; \
			} \
			(TimingStackTop)++; \
			TimingStack[TimingStackTop] = What; \
			OldTime1 = Nugget_ElapsedTime()
#		define PopTiming() \
			OldTime1 = Nugget_ElapsedTime() - OldTime1; \
			TimeTemp = TimingStack[TimingStackTop]; \
			if (TimeTemp <= 3+MaxRedundancy) { \
				Time[TimeTemp] += OldTime1; \
			} \
			(TimingStackTop)--; \
			ERROR(TimingStackTop < 0, 5, "Stack error"); \
			OldTime1 = Nugget_ElapsedTime()

	/* RECURSION STACK FOR BACKTRACK */
		int Depth; 
				/* how deep our dive into 
				the tree is at the moment; 
				starts at 0 */
		ProblemType CurrentProblem[MaxDepth]; 
				/* 0 is the parent of 1; 
				1 is the first one we are working on.  
				Don't trust
				CurrentProblem[i] for i <= GiveAwayDepth.  */
		int DeweyDelta[MaxDepth];
				/* CP[i] is the DD[i] child of CP[i-1].  
				Start counting at 0. */
		Dewey DeweyPrefix;
				/* Identifies CurrentProblem[GiveAwayDepth] */
		int RootID; 
				/* the unique id associated 
			    	with DeweyPrefix */
		WorkPtrType RootWorkPtr;
				/* could be local to BackTrack but 
				for debug */
		int GiveAwayDepth;
				/* if there are open problems, their depth.  
				If not, the depth of the parent of the next 
				set of open problems to generate.  Starts
				at 0. */
		int CurrentRedundancy;
				/* 0 is normal; 1 is redoing work given to 
				another, etc. */

/* PROVIDED TO APPLICATIONS */

void BroadcastInfo(InfoPtr)
InfoType *InfoPtr;
{
	OutBufPtr OutPtr;
	 
	PushTiming(CommTime);
	OutPtr = SAP_GetSndBuf();
	{	InterNodeMessage *IMsgPtr = (InterNodeMessage *) OutPtr;
		IMsgPtr->Kind = htonl(GlobalInfo);
		IMsgPtr->ProblemNumber = htonl(ProblemNumber);
		ConvertInfoOut(InfoPtr, &IMsgPtr->SomeInfo);
	}
	SAP_Broadcast(OutPtr, sizeof(InterNodeMessage), 1, PartitionSize);
	PopTiming();
} /* BroadcastInfo */

void ReportResult(ProbPtr)
ProblemPtr ProbPtr;
/* send a result value back to the host */
{
	OutBufPtr OutPtr;
	 
	if (AnswerSpec==Full) {
		PushTiming(TossTime);
		OutPtr = SAP_GetSndBuf();
		{	AnswerMessage *AMsgPtr = (AnswerMessage *) OutPtr;
			AMsgPtr->Status = htonl(AnAnswer);
			ConvertProblemOut(ProbPtr, &AMsgPtr->Prob, 
					"ReportResult");
			AMsgPtr->Sequence = htonl(LocalSequence);
			(LocalSequence)++;
		}
		SAP_Snd2(OutPtr, 0);
		PopTiming();
	} /* AnswerSpec == Full */
} /* ReportResult */

#ifdef interactive

void Dib_stdin(inputbuf,len)
char* inputbuf;
int len;
/* ask host to read in user input and provide in buffer */
{
	OutBufPtr OutPtr;
	InBufPtr InPtr;
	AnswerMessage *AMsgPtr;
	int i;
	 
	PushTiming(TossTime);
	OutPtr = SAP_GetSndBuf();
	{	AMsgPtr = (AnswerMessage *) OutPtr;
		AMsgPtr->Status = htonl(Do_stdin);
		AMsgPtr->Sequence = htonl(LocalSequence);
		(LocalSequence)++;
	}
	SAP_Snd2(OutPtr, 0);
	PopTiming();
	SAP_RcvSpecific(&InPtr,0,INPUTMSG);
	AMsgPtr = (AnswerMessage *) InPtr;
	if(len>IOCHARS)
		len=IOCHARS;
	for(i=0;i<len;i++){
		inputbuf[i]=AMsgPtr->buffer[i];
	}

} /* Dib_stdin */

void Dib_stdout(outputbuf,len)
char* outputbuf;
int len;
/* ask host to print the message provided in buffer */
{
	OutBufPtr OutPtr;
	 
	PushTiming(TossTime);
	OutPtr = SAP_GetSndBuf();
	{	int i;
		AnswerMessage *AMsgPtr = (AnswerMessage *) OutPtr;
		AMsgPtr->Status = htonl(Do_stdout);

		if(len>IOCHARS)
			len=IOCHARS;
		for(i=0;i<len;i++)
			AMsgPtr->buffer[i]=outputbuf[i];
		AMsgPtr->buffer[len]='\0';

		AMsgPtr->Sequence = htonl(LocalSequence);
		(LocalSequence)++;
	}
	SAP_Snd3(OutPtr, 0, sizeof(AnswerMessage));
} /* Dib_stdout */
#endif

/* COMMUNICATION WITH OTHER NODES */

void DropOut()
/* tell host we are quitting. */
{
	OutBufPtr OutPtr;
	 
	PushTiming(TossTime);
	OutPtr = SAP_GetSndBuf();
	PopTiming(); /* to make sure it counts in the message */
	PushTiming(TossTime);
	{
		AnswerMessage *AMsgPtr = (AnswerMessage *) OutPtr;
		AMsgPtr->Status = htonl(Quitting);
		AMsgPtr->Sequence = htonl(LocalSequence);
		TimeCopyOut(Time, AMsgPtr->Times);
		(LocalSequence)++;
	}
	SAP_Snd3(OutPtr, 0, sizeof(AnswerMessage));
	DEBUG(DEBUGNODE, 4, CheckWorkEmpty(); );
		/* just for information; there may be some redundant
		work left, or even non-redundant work that others 
		finished for me. */
	if (Nugget_MyNode == 1) {
		/* wait a while (to give other nodes a chance to send 
		final note to the host) and then tell the host to stop. */
		OutPtr = SAP_GetSndBuf();
		{
			AnswerMessage *AMsgPtr = (AnswerMessage *) OutPtr;
			AMsgPtr->Status = htonl(PleaseStop);
			AMsgPtr->Sequence = htonl(LocalSequence);
			(LocalSequence)++;
		}
		sleep(1); /* waste time */
		SAP_Snd3(OutPtr, 0, sizeof(AnswerMessage));
	}
	PopTiming();
} /* DropOut */

void TellTerminate()
/* Usually called only in node 1.  sends a 'terminate' message to every other
node and sets the 'termination' flag. */
{
	OutBufPtr OutPtr;
	 
	PushTiming(CommTime);

	DEBUG(DEBUGNODE, 3, fprintf(stdout, "Node[%d]: In TellTerminate\n",
			Nugget_MyNode); );

	OutPtr = SAP_GetSndBuf();
	{	InterNodeMessage *IMsgPtr = (InterNodeMessage *) OutPtr;
		IMsgPtr->Kind = htonl(Terminate);
		IMsgPtr->ProblemNumber = htonl(ProblemNumber);
	}
	SAP_Broadcast(OutPtr, sizeof(InterNodeMessage), 1, PartitionSize);
	/* 2..PartitionSize is not enough; we may be on another machine
	when termination is discovered, because the ultimate root may have
	migrated. */
	Termination = true ;
	PopTiming();
} /* TellTerminate */

void TellHostReporting(DewPtr)
Dewey *DewPtr;
/* We have finished a subproblem (or the whole problem); tell the Dewey
number of what we just finished to the host. */
{
	OutBufPtr OutPtr;
	 
	PushTiming(TossTime);
	OutPtr = SAP_GetSndBuf();
	{
		AnswerMessage *AMsgPtr = (AnswerMessage *) OutPtr;
		AMsgPtr->Status = htonl(Reporting);
		DeweyCopyOut(*DewPtr, AMsgPtr->AnswerDewey);
		AMsgPtr->Sequence = htonl(LocalSequence);
		LocalSequence++;
	}
	SAP_Snd3(OutPtr, 0, sizeof(AnswerMessage));
	PopTiming();
} /* TellHostReporting */

void TellHostBatchDone(ProbPtr)
ProblemPtr ProbPtr;
/* We have finished the whole problem; tell the updated
root (in ProbPtr) to the host */
{
	OutBufPtr OutPtr;
	 
	PushTiming(TossTime);
	OutPtr = SAP_GetSndBuf();
	PopTiming(); /* so this last wait for buffer is included in
		these statistics we now send to host. */
	PushTiming(TossTime);
	{
		AnswerMessage *AMsgPtr = (AnswerMessage *) OutPtr;
		AMsgPtr->Status = htonl(BatchDone);
		ConvertProblemOut(ProbPtr, &AMsgPtr->Prob, 
				"TellHostBatchDone");
		TimeCopyOut(Time, AMsgPtr->Times);
		AMsgPtr->Sequence = htonl(LocalSequence);
		(LocalSequence)++;
	}
	SAP_Snd2(OutPtr, 0);
	PopTiming();
} /* TellHostBatchDone */

#ifndef pvmred
void TellPeerRepeating (WorkPtr)
WorkPtrType WorkPtr;
/* I have decided to do WorkPtr again.  Tell its owner, (fault-tolerance
rule a) so that if the owner has other work for me to do, it will tell me. */
{
	OutBufPtr OutPtr;
	 
	PushTiming(CommTime);
	OutPtr = SAP_GetSndBuf();
	{	InterNodeMessage *IMsgPtr = (InterNodeMessage *) OutPtr;
		IMsgPtr->Kind = htonl(Repeating);
		IMsgPtr->Asker = htonl(Nugget_MyNode);
		IMsgPtr->Redundancy = htonl(WorkPtr->Redundancy);
		IMsgPtr->UniqueID = htonl(WorkPtr->UniqueID);
		IMsgPtr->ProblemNumber = htonl(ProblemNumber);
		DeweyCopyOut(WorkPtr->Dew, IMsgPtr->ChildDewey);
	}
	SAP_Snd3(OutPtr, WorkPtr->WhereTo, sizeof(InterNodeMessage));
	PopTiming();
} /* TellPeerRepeating */
#endif

void TellPeerUpdating(WorkPtr, Direction)
WorkPtrType WorkPtr;
int Direction; /* Tells what direction the updating will be done */
/* The problem in WorkPtr has changed, so we must inform the worker (definitely
not us) to which we have given it.
*/
{
	OutBufPtr OutPtr;
	 
	PushTiming(CommTime);
	ERROR(WorkPtr->WhereTo == OPEN, 29, "Updating unassigned work");
	OutPtr = SAP_GetSndBuf();
	{	InterNodeMessage *IMsgPtr = (InterNodeMessage *) OutPtr;
		if (Direction == Down)
			IMsgPtr->Kind = htonl(Updating);
		else
			IMsgPtr->Kind = htonl(UpdatingUp);
		IMsgPtr->UniqueID = htonl(WorkPtr->UniqueID);
		IMsgPtr->ProblemNumber = htonl(ProblemNumber);
		ConvertProblemOut(&WorkPtr->Value, &IMsgPtr->Child, 
			"TellPeerUpdating");

	}
	if (Direction == Down){
		DEBUG(DEBUGNODE, 3,
			fprintf(stdout, "Node[%d]: TellPeerUpdating to [%d]\n",
					Nugget_MyNode, WorkPtr->WhereTo);
		);
		SAP_Snd2(OutPtr, WorkPtr->WhereTo);
	}else{
		DEBUG(DEBUGNODE, 3,
			fprintf(stdout, "Node[%d]: TellPeerUpdating to [%d]\n",
					Nugget_MyNode, WorkPtr->WhereFrom);
		);
		SAP_Snd2(OutPtr, WorkPtr->WhereFrom);
	}
	PopTiming();
} /* TellPeerUpdating */

/* ROUTINES FOR MAINTAINING CONSISTENCY */

void DoUpdate(ParentPtr)
WorkPtrType ParentPtr;
/* When Combine or Update was called, it said that ParentPtr had changed,
so its Children need Update called.
*/
{
	int D; /* depth in CurrentProblem stack */
	WorkPtrType ChildPtr; /* steps through all work given */
	Boolean Again;
	ProblemType *ProbPtr;
	 
	DEBUG(DEBUGNODE, 3,
		fprintf(stdout, "Node[%d]: in DoUpdate, parent is (%o)\n", 
			Nugget_MyNode, ParentPtr->UniqueID);
	);
	for (ChildPtr = WorkHeaderPtr->Next; ChildPtr != null;
	  ChildPtr = ChildPtr->Next) {
		if (ChildPtr->ParentPtr == ParentPtr){
			/* ChildPtr is an affected child */
			/* forlater: we don't treat the case of something 
			I am redoing completely.  We do send to peer, 
			but fail to fix local copies. */
			if (ChildPtr->WhereTo != Nugget_MyNode && 
						!ChildPtr->Available) {
				if (ChildPtr->WhereTo == Open)
					fprintf(stdout, "Open But Not Available\n");
				/* I have given this child away, but may be redoing */
				DEBUG(DEBUGNODE, 3,
					fprintf(stdout, "Node[%d]: given away to [%d]\n",
						Nugget_MyNode,ChildPtr->WhereTo);
				);
				/* update my copy */
				Update(&ChildPtr->Value, &ParentPtr->Value, &Again);
				/* This node is given away, but it may have outstanding children */
				if (Again) DoUpdate(ChildPtr);
				TellPeerUpdating(ChildPtr, Down); /* could avoid if there is no
					change at all to the child.  It is not enough just to check
					Again; a change to the child could have affect without
					requiring deeper effects. */
				/* forlater: follow this chain on the peer. */
			} else if (ChildPtr->UniqueID == RootID) {
				/* this child is one I am currently working on */
				DEBUG(DEBUGNODE, 3,
					fprintf(stdout, "Node[%d]: Child is current work\n",
						Nugget_MyNode);
				);
				Update(&ChildPtr->Value, &ParentPtr->Value, &Again);
				if (Again) DoUpdate(ChildPtr);
				ProbPtr = (ProblemType *) &ChildPtr->Value;
				for (D = GiveAwayDepth; Again && D <= Depth; D++) {
					/* first iteration lets application put new inherited info
					in CP[GAD] without clobbering synthesized fields.  In this
					iteration, both problems given to Update are identical, but
					the second one has better inherited info.  Other iterations 
					are normal:  first arg to Update is child, second is
					parent. */
					DEBUG(DEBUGNODE, 3,
						fprintf(stdout, "Node[%d]: Diving\n",Nugget_MyNode);
					);
					Update(&CurrentProblem[D], ProbPtr, &Again);
					ProbPtr = &CurrentProblem[D];
				}
			} else if (ChildPtr->WhereTo == OPEN) {
				/* this problem has not been given away, not worked on */
				DEBUG(DEBUGNODE, 3, fprintf(stdout,"Node[%d]: Open\n",Nugget_MyNode););
				Update(&ChildPtr->Value, &ParentPtr->Value, &Again);
				if (ChildPtr->Suspended == SUSPNOW) { /* time to reactivate */
					DEBUG(DEBUGNODE, 0,
						fprintf(stdout, "Node[%d]: Reactivating AFTER UPDATE suspended process (%o)\n  ",
						Nugget_MyNode, ChildPtr->UniqueID);
						PrintProblem(&ChildPtr->Value);
						fprintf(stdout, "\n  Prev child:  ");
						PrintProblem(&ChildPtr->Suspension->PrevChild);
						fprintf(stdout, "\n");
					);
					ChildPtr->Suspended = SUSPWAS;
					ERROR(ChildPtr->Available, 85, "still available");
					ChildPtr->Available = true;
					WasSuspendedWorkCount++;
				}/* time to reactivate */
				if (Again) DoUpdate(ChildPtr);
			} else { /* catch all */
				/* the problem is about to be redone or is suspended */
				DEBUG(DEBUGNODE, 3,
					fprintf(stdout, "Node[%d]: Unexpected pattern in this problem\n");
					PrintWork(ChildPtr);
				);
				Update(&ChildPtr->Value, &ParentPtr->Value, &Again);
				if (ChildPtr->Suspended == SUSPNOW) { /* time to reactivate */
					DEBUG(DEBUGNODE, 0,
						fprintf(stdout, "Node[%d]: Reactivating AFTER UPDATE suspended process (%o)\n  ",
							Nugget_MyNode, ParentPtr->UniqueID);
						PrintProblem(&ChildPtr->Value);
						fprintf(stdout, "\n  Prev child:  ");
						PrintProblem(&ChildPtr->Suspension->PrevChild);
						fprintf(stdout, "\n");
					);
					ChildPtr->Suspended = SUSPWAS;
					ERROR(ChildPtr->Available, 85, "still available");
					ChildPtr->Available = true;
					WasSuspendedWorkCount++;
				}/* time to reactivate */
				if (Again) DoUpdate(ChildPtr);
			} /* catch all */
		} /* ChildPtr is an affected child */
	} /* for ChildPtr */
} /* DoUpdate */

void DoUpdateUp(ChildPtr)
WorkPtrType ChildPtr;
{/* When combine or UpdateUp were called, it said that child has new
    info which need to be passed up to the parent */

	WorkPtrType ParentPtr; /* will point to the parent */
	Boolean	AgainUp, AgainDown;

	ParentPtr = ChildPtr->ParentPtr;
	if (ParentPtr == null){ /* Two possibilities, either our parent is on 
			another machine or we are at the top level */ 
		if (ChildPtr->WhereFrom != Nugget_MyNode){
			/*My parent is on another machine */
			DEBUG(DEBUGNODE, 3,
				fprintf(stdout, "Node[%d]: Parent on different machine [%d]\n", Nugget_MyNode,ChildPtr->WhereFrom);
			);
				/* LOOK HERE */
			/*TellPeerUpdating(&ChildPtr->Value, Up);*/
			TellPeerUpdating(ChildPtr, Up);
		}else{ /* top level */
			DEBUG(DEBUGNODE, 3,
				fprintf(stdout, "Node[%d]: Top level\n",Nugget_MyNode););
			return;
		}
	} else { /* We have a parent, and it is on the same machine */
		DEBUG(DEBUGNODE, 3,
			fprintf(stdout, "Node[%d]: Updating the Parent\n",Nugget_MyNode););
		UpdateUp(&ChildPtr->Value, &ParentPtr->Value, 
			&AgainUp, &AgainDown);
		if (ParentPtr->Suspended == SUSPNOW) { /* time to reactivate */
			/* forlater: not clear that we should only try to reactivate parent
			if it has no more outstanding problems; maybe whenever a child
			finishes. */
			DEBUG(DEBUGNODE, 0,
				fprintf(stdout, "Node[%d]: Reactivating suspended process (%o)\n  ",
					Nugget_MyNode, ParentPtr->UniqueID);
				PrintProblem(&ParentPtr->Value);
				fprintf(stdout, "\n  Prev child:  ");
				PrintProblem(&ParentPtr->Suspension->PrevChild);
				fprintf(stdout, "\n");
			);
			ParentPtr->Suspended = SUSPWAS;
			ERROR(ParentPtr->Available, 85, "still available");
			ParentPtr->Available = true;
			WasSuspendedWorkCount++;
		}/* time to reactivate */
		if (AgainUp)
			DoUpdateUp(ParentPtr);
		if (AgainDown)
			DoUpdate(ParentPtr);
	}/* On the same machine as the parent */
}/* DoUpdateUp */
/* have re-read to here */

void TellParent(ChildPtr)
	WorkPtrType ChildPtr;
/* ChildPtr has been finished.  Update its parent and inform the parent's
origin if all its children have finished.  Reclaim Child work structure. */
{
	Boolean NeedUpdate, NeedUpdateUp;
	OutBufPtr OutPtr;
	WorkPtrType ParentPtr = ChildPtr->ParentPtr;

	if (ChildPtr->UnsolvedCount != 0){
		ChildPtr->Suspended = SUSPNOW;
		ChildPtr->Available = false;
		return;
	}
	ChildPtr->ParentPtr = null; /* remove link so DoUpdate won't see it. */
	DEBUG(DEBUGNODE, 3, fprintf(stdout, "Node[%d]: TellParent\n",Nugget_MyNode); );
	if (ParentPtr == null) { /* top-level problem */
		if (ChildPtr->WhereFrom == Nugget_MyNode) {
			/* This situation actually occurs.  Seems that always problem
			left over after someone else has already finished it. */
			DEBUG(DEBUGNODE, -10,
				fprintf(stdout, "Node[%d]: Early return from TellParent\n",
					Nugget_MyNode);
			);
			return;
		}
		ParentPtr = ChildPtr;
		DEBUG(DEBUGNODE, 3, fprintf(stdout, "Node[%d]: Top level\n",Nugget_MyNode); );
		ERROR(ParentPtr->UnsolvedCount != 0, 79, "more to come?");
	} else { /* has a parent */
		Combine(&ChildPtr->Value, &ParentPtr->Value, &NeedUpdate,
			&NeedUpdateUp);
		if (NeedUpdate) {
			DoUpdate(ParentPtr);
		}
		if (NeedUpdateUp) {
			DoUpdateUp(ParentPtr);
		}
		ParentPtr->UnsolvedCount--;
		DEBUG(DEBUGNODE, 3,
			fprintf(stdout, "Node[%d]: Parent is (%o) ", Nugget_MyNode,ParentPtr->UniqueID);
			PrintDewey(&ParentPtr->Dew);
			fprintf(stdout, "; %d still unsolved.\n", ParentPtr->UnsolvedCount);
		);
	} /* has a parent */
	if (ParentPtr->UnsolvedCount == 0) {
		if (ParentPtr->Suspended == SUSPNOW) { /* time to reactivate */
			/* forlater: not clear that we should only try to reactivate parent
			if it has no more outstanding problems; maybe whenever a child
			finishes. */
			DEBUG(DEBUGNODE, 0,
				fprintf(stdout, "Node[%d]: Reactivating suspended process (%o)\n  ",
					Nugget_MyNode, ParentPtr->UniqueID);
				PrintProblem(&ParentPtr->Value);
				fprintf(stdout, "\n  Prev child:  ");
				PrintProblem(&ParentPtr->Suspension->PrevChild);
				fprintf(stdout, "\n");
			);
			ParentPtr->Suspended = SUSPWAS;
			ERROR(ParentPtr->Available, 85, "still available");
			ParentPtr->Available = true;
			WasSuspendedWorkCount++;
		} /* reactivate */ 
		else if (ParentPtr->Suspended == SUSPNOT){ /* parent itself is finished */
			if (ParentPtr->WhereFrom == Nugget_MyNode) { /* handle locally */
				TellParent(ParentPtr);
			} else { /* send a message. */
				if (ParentPtr->WhereFrom == 0) { /* we must be node 1 */
					TellHostBatchDone((ProblemPtr) &ParentPtr->Value);
					TellTerminate(); /* tell all nodes, even me, to drop out. */
				} else { /* report to peer */
					PushTiming(CommTime);
					OutPtr = SAP_GetSndBuf();
					{	InterNodeMessage *IMsgPtr = (InterNodeMessage *) OutPtr;
						IMsgPtr->Kind = htonl(Result);
						IMsgPtr->UniqueID = htonl(ParentPtr->UniqueID);
						IMsgPtr->ProblemNumber = htonl(ProblemNumber);
						DeweyCopyOut(ParentPtr->Dew, IMsgPtr->ChildDewey);
						ConvertProblemOut(&ParentPtr->Value, &IMsgPtr->Child,
							"TellParent");
						DEBUG(DEBUGNODE, 2,
							fprintf(stdout, "Node[%d]: Telling (%o) Red=%d",
								Nugget_MyNode, ParentPtr->UniqueID,
								CurrentRedundancy);
							PrintDewey(&ParentPtr->Dew);
							fprintf(stdout, "to [%d]\n", ParentPtr->WhereFrom);
						);
					}
					SAP_Snd2(OutPtr, ParentPtr->WhereFrom);
					DEBUG(DEBUGNODE, 4,
						if (ParentPtr->Dew.Length < MaxDeweyReport) {
							/* bother the host only for high nodes */
							TellHostReporting(&ParentPtr->Dew);
						}
					);
					PopTiming();
				} /* report to peer */
				ReleaseWork(ParentPtr);
			} /* send a message */
		} /* report to parent */
	} /* Unsolved count went down to 0 */
#ifdef notneeded
	ReleaseSubsumedWork(&ChildPtr->Dew); /* including Child itself */
#endif
} /* TellParent */

void AskForHelp(FirstTime, Beggar, Hops)
Boolean FirstTime;
int Beggar, Hops;
/*
If FirstTime then we ask node 1; afterwards, send a request to NumHelpers
nodes next in line.  If Beggar != Nugget_MyNode, we are forwarding the
request, so we only send out one copy.  Tell the helper the number of hops.
*/
{
	int Seq, Count;
	HelperListType Helpers;
	OutBufPtr OutPtr;
	 
	PushTiming(CommTime);
	Count = (FirstTime || Beggar != Nugget_MyNode) ? 1 : NumHelpers;
	if (Beggar != Nugget_MyNode && PartitionSize == 2) {
		DEBUG(DEBUGNODE, 3, fprintf(stdout, "Node[%d]: Won't ask neighbor\n",
			Nugget_MyNode);
		);
		Count = 0;
	} else
		DEBUG(DEBUGNODE, 3, fprintf(stdout, "Node[%d]: Will ask neighbor\n",
			Nugget_MyNode); );
	/* Generate list of helpers */
		/* Our helpers are NextToAsk and its successors. */
		for (Seq = 1; Seq <= Count; Seq++) {
			Helpers[PartitionSize-Seq-1] = NextToAsk;
			NextToAsk = NextToAsk % PartitionSize + 1;
			if (NextToAsk == Nugget_MyNode) {
				NextToAsk = NextToAsk % PartitionSize + 1;
			}
		}
		if (FirstTime) {
			/* first helper should be 1 */
			Helpers[PartitionSize-2] = 1;
		}
	/* send out NumHelpers messages */
		for (Seq = 1; Seq <= Count; Seq++) {
			/* ask one helper */
			OutPtr = SAP_GetSndBuf();
			{	InterNodeMessage *IMsgPtr = (InterNodeMessage *) OutPtr;
				IMsgPtr->Kind = htonl(Request);
				IMsgPtr->Asker = htonl(Beggar);
				IMsgPtr->Count = htonl(Hops);
				IMsgPtr->ProblemNumber = htonl(ProblemNumber);
			}
			DEBUG(DEBUGNODE, 2, fprintf(stdout, "Node[%d]: Asking node[%d]\n", Nugget_MyNode, Helpers[PartitionSize-Seq-1]); );
			if (GlobalInUse) {
				SAP_Snd3(OutPtr, -1, sizeof(InterNodeMessage));
				break; /* only send one message, 
						whatever the helper list */
			} else {
				SAP_Snd3(OutPtr, Helpers[PartitionSize-Seq-1],
					sizeof(InterNodeMessage));
			}
		} /* for Seq */
	PopTiming();
} /* AskForHelp */

void HandleRequest(Who, Hops)
int Who, Hops;
/* Who has just asked us for work, or we have decided out of our good nature to
foist some work on Who.  If we have any to spare, send it; otherwise, forward
the request, if it is forwardable.  If we have no work to split the current
level of CurrentProblem if it is not trivial.  If it has already got
Hops==PartitionSize, don't forward it.  'Repeating' messages don't get
forwarded this way.
*/
{
	short Done;
	int GiveAwayCount; /* number of elements to give away */
	Dewey TmpDewey;
	WorkPtrType WorkPtr, SomeWorkPtr;
	int Seq;
	OutBufPtr OutPtr;
	ProblemType TmpProb;
 
	PushTiming(WorkTime);
	DEBUG(DEBUGNODE, 2,
		fprintf(stdout,"Node[%d]: Asked by node[%d]\n", Nugget_MyNode, Who);
	);
	ERROR(Who == Nugget_MyNode, 66, "Asking myself for work");
	while (AvailWorkCount == 0) { /* try to subdivide at GiveAwayDepth */
		if (GiveAwayDepth >= Depth) { /* BackTrack is about to terminate. */
			DEBUG(DEBUGNODE, 3,
				fprintf(stdout, "Node[%d]: Too deep to give away; depth %d.\n",
					Nugget_MyNode,GiveAwayDepth);
			);
			break;
		}
		if (! NonTrivial(&CurrentProblem[GiveAwayDepth+1])) { /* done soon */
			DEBUG(DEBUGNODE, 3,
				fprintf(stdout, "Node[%d]: Trivial when deeper than %d\n",
					Nugget_MyNode, GiveAwayDepth);
			);
			break;
		}
		WorkPtr = FindWork(RootID);
		if (WorkPtr == null) { /* the work that BackTrack is engaged in is
			redundant and has been finished. */
			DEBUG(DEBUGNODE, 3, fprintf(stdout, "Node[%d]: Nothing at the moment\n",Nugget_MyNode););
			break;
		}
		/*  split CurrentProblem[GiveAwayDepth] */
		{	/* adjust global variables shared with BackTrack */
			GiveAwayDepth++;
			DeweyPrefix.Length++;
			DeweyPrefix.DNums[DeweyPrefix.Length] = DeweyDelta[GiveAwayDepth];
			RootID = LastUniqueID++;
		}
		{	/* put CurrentProblem[new GiveAwayDepth] in Work list */
			WorkPtrType ChildPtr;
			ProblemCopy(&CurrentProblem[GiveAwayDepth], &TmpProb);
			ChildPtr = NewWork(&DeweyPrefix, RootID, Nugget_MyNode,
				WorkPtr, CurrentRedundancy, &TmpProb);
				/* increments WorkPtr->UnsolvedCount to 1 */
			RootWorkPtr = ChildPtr;
			ERROR(WorkPtr->UnsolvedCount < 1, 36, "Bad UnsolvedCount");
			/* adjust to show we are currently working on it */
			RootWorkPtr->WhereTo = Nugget_MyNode;
			RootWorkPtr->Available = false;
			AvailWorkCount--;
		}
		{	/* generate its siblings into TmpProb; put in Work list */
			DeweyCopy(&DeweyPrefix, &TmpDewey);
			while (true) {
				Generate(&Done, &WorkPtr->Value, &TmpProb);
				if (Done == MORELATER) {
					WorkPtr->Suspended = SUSPNOW;
					WorkPtr->Suspension = (SuspensionType *)
						malloc(sizeof(SuspensionType)+ProblemSize);
					WorkPtr->Suspension->GenSoFar =
						TmpDewey.DNums[TmpDewey.Length]+1;
					ProblemCopy(&TmpProb, &WorkPtr->Suspension->PrevChild);
					DEBUG(DEBUGNODE, 0,
						fprintf(stdout,
							"Node[%d]: Suspending (%o) having created %d kids\n",
							Nugget_MyNode, WorkPtr->UniqueID,
							WorkPtr->Suspension->GenSoFar);
						PrintProblem(&WorkPtr->Value);
						fprintf(stdout, "\n  Prev child:  ");
						PrintProblem(&WorkPtr->Suspension->PrevChild);
						fprintf(stdout, "\n");
					);
					break;
				} /* MORELATER */
				if (Done == DONE) break;
				TmpDewey.DNums[TmpDewey.Length]++;
				(void) NewWork(&TmpDewey, LastUniqueID++, Nugget_MyNode,
					WorkPtr, CurrentRedundancy, &TmpProb);
					/* incs AvailWorkCount, WorkPtr->UnsolvedCount */
				/* break; /* debug: just generate one for now */
			} /* while more siblings */
		} /* generate all siblings */
	} /* while need to subdivide at GiveAwayDepth */
	DEBUG(DEBUGNODE, 1,
		fprintf(stdout, "Node[%d]: Managed to find %d problems.\n", Nugget_MyNode,
			AvailWorkCount);
	);
	GiveAwayCount = AvailWorkCount;
	if (WorkFract < 0) {
		if (GiveAwayCount > 0) GiveAwayCount = 1;
	} else {
		if (GiveAwayCount > 0) {
			GiveAwayCount = ((GiveAwayCount * WorkFract)+5) / 10;
			if (GiveAwayCount == 0) GiveAwayCount = 1;
		}
	}
	PopTiming(); /* end of WorkTime; the rest is communication */
	PushTiming(CommTime);
	if (GiveAwayCount > 0) {
		/* send GiveAwayCount pieces of work to Who */
		for (Seq = 1; Seq <= GiveAwayCount; Seq++) {
			OutPtr = SAP_GetSndBuf();
			SelectWork(&SomeWorkPtr, true);

			DEBUG( DEBUGNODE,1,
				fprintf(stdout, "Node[%d]: Sending Work to %d", 
					Nugget_MyNode, Who);
			);

			{	InterNodeMessage *IMsgPtr = (InterNodeMessage *) OutPtr;
				IMsgPtr->Kind = htonl(Work);
				IMsgPtr->Count = htonl((long) 1); /* in increments of 1 unit */
				IMsgPtr->UniqueID = htonl(SomeWorkPtr->UniqueID);
				IMsgPtr->ProblemNumber = htonl(ProblemNumber);
				ConvertProblemOut(&SomeWorkPtr->Value, &IMsgPtr->Child,
					"HandleRequest");
				DeweyCopyOut(SomeWorkPtr->Dew, IMsgPtr->ChildDewey);
				IMsgPtr->Redundancy = htonl(SomeWorkPtr->Redundancy);
				DEBUG(DEBUGNODE, 2,
					fprintf(stdout, "Node[%d]: Sending red=%d", Nugget_MyNode,
						SomeWorkPtr->Redundancy);
					PrintDewey(&SomeWorkPtr->Dew);
					fprintf(stdout, "(%o) to node[%d]\n", SomeWorkPtr->UniqueID, Who);
				);
				ERROR(SomeWorkPtr->Redundancy < 0, 7, "Bad redundancy");
				if (SomeWorkPtr->WhereFrom != Nugget_MyNode) {
					/* This is work we have received from outside and have
					never started working on.  Forget about this work and tell
					the new holder not to report to us but to WhereFrom */
					IMsgPtr->Asker = htonl(SomeWorkPtr->WhereFrom);
					ERROR(SomeWorkPtr->ParentPtr != null, 39, "bad parent");
					SomeWorkPtr->Suspended = GIVENAWAY;
				} else {
					IMsgPtr->Asker = htonl(Nugget_MyNode);
				}
				SomeWorkPtr->WhereTo = Who;
				ERROR(Who == 0, 4, "Bad Who");	
			}
			SAP_Snd2(OutPtr, Who);
		} /* for Seq */
	} else if (Hops < PartitionSize) { /* no work for asker; must forward */
		AskForHelp(false, Who, Hops+1);
	}
	PopTiming();
} /* HandleRequest */

void CheckIncoming(BusyWait, AbortPtr, InPtr, Source)
Boolean BusyWait, *AbortPtr;
InBufPtr InPtr;
int Source;
/* called periodically to see if anyone has sent us a message we need
to deal with.  If BusyWait is set, we don't return if nothing has come in; we
keep checking.  We set Abort to true if work just came in that is less
redundant than the current level.  The InPtr argument, if non-null, is a
message that has been taken off the global queue; it came from Source.
*/
{
	WorkPtrType WorkPtr;
	InterNodeMessage Internal, *IMsgPtr;
	ProblemType Problem;
	int LoopCount;
#ifdef pvmred
	OutBufPtr OutPtr;
#endif
	 
	PushTiming(CommTime);
	*AbortPtr = false;
	LoopCount = 0;
	while (true) {
		if (InPtr != null || SAP_ProbeAny() ) {
			/* a message has come in; deal with it. */
			LoopLimit = LOOPLIMIT;
			if (InPtr == null){
				DEBUG(DEBUGNODE, 5,
					fprintf(stdout, "Node[%d]: In CheckIncoming Inptr==Null\n",Nugget_MyNode); 
				);
				SAP_Rcv2(&InPtr, &Source); 
			}
			else{
				DEBUG(DEBUGNODE, 5,
					fprintf(stdout, "Node[%d]: In CheckIncoming Inptr!=Null\n",Nugget_MyNode); 
				);
			}
			IMsgPtr = (InterNodeMessage *) InPtr;
			Internal.Kind = ntohl(IMsgPtr->Kind);
			Internal.Asker = ntohl(IMsgPtr->Asker);
			DeweyCopyIn(IMsgPtr->ChildDewey, Internal.ChildDewey);
			Internal.Redundancy = ntohl(IMsgPtr->Redundancy);
			Internal.UniqueID = ntohl(IMsgPtr->UniqueID);
			Internal.ProblemNumber = ntohl(IMsgPtr->ProblemNumber);
#ifdef newnode
			Internal.NewAddress = ntohl(IMsgPtr->NewAddress);
#endif

			if ( Internal.ProblemNumber != ProblemNumber 
					&& Internal.Kind!=NewNode ){
#ifdef newnode
#endif
				/* Do not discard a new node message just because 
				of a mismatch in problem number the guy has just 
				joined new. Problem numbers can differ. This can happen
				if a new node is added just at the time an earlier round 
				is finishing up with a problem. If we accept the messsage
				for new node the new node may not be able to work
				in current round due to mismatch in problem number
				but can work fine from next time on so let us
				accept new node messages with even mismatched problem
				numbers */

				/* fprintf(stdout, "Node[%d]: IntPN=%d, RecPN=%d Source=%d\n",
					Nugget_MyNode,ProblemNumber,Internal.ProblemNumber,
					Source); */

				DEBUG(DEBUGNODE, 3,
					fprintf(stdout,"Node[%d]: Got old message from %d\n", 
						Nugget_MyNode, Source);
				);

				SAP_ReuseMsg(InPtr);
				InPtr = null;
				continue;
			}
			if (! Termination) switch (Internal.Kind) {
			case Terminate:
				Termination = true;
				break;
			case Request:
				if (Internal.Asker != Nugget_MyNode) {
					HandleRequest(Internal.Asker, Internal.Count);
				}
				break; /* case Request */
			case Updating:
				DEBUG(DEBUGNODE, 3,
					fprintf(stdout, "Node[%d]: Got update\n",Nugget_MyNode););
				{	WorkPtrType ChildPtr = FindWork(Internal.UniqueID);
					Boolean Again;
					if (ChildPtr != null) { /* hasn't been finished */
						DEBUG(DEBUGNODE, 3,
							fprintf(stdout, "Node[%d]: Newer copy\n",
								Nugget_MyNode););
						ConvertProblemIn(&IMsgPtr->Child, &Problem, "Updating");
						if (ChildPtr->Suspended == GIVENAWAY){
							Update(&ChildPtr->Value, &Problem, &Again);
							if (ChildPtr->WhereTo == Open)
								fprintf(stdout, "GIVENAWAY but OPEN\n");
							TellPeerUpdating(ChildPtr,Down);
							/*ReleaseWork(ChildPtr);*/
						}else{/* Not given Away */
						Update(&ChildPtr->Value, &Problem, &Again);
						if (Again) DoUpdate(ChildPtr);
						/* Update must handle case where both args are the same
						problem, but second arg has new inherited info. */
						if (ChildPtr->WhereTo!=Nugget_MyNode && !ChildPtr->Available){
							if (ChildPtr->WhereTo== Open)
								fprintf(stdout,
									"Node[%d]: IN Case Updating Open but not available\n",Nugget_MyNode);
							TellPeerUpdating(ChildPtr, Down);
						}else if (ChildPtr->UniqueID == RootID) {
						/* I am currently working on this node */
						ProblemType *ProbPtr;
						int D;
						ProbPtr = (ProblemType *) &ChildPtr->Value;
						for (D = GiveAwayDepth; Again && D <= Depth; D++) {
							DEBUG(DEBUGNODE, 3,
							fprintf(stdout, "Node[%d]: Diving\n",Nugget_MyNode);
							);
							Update(&CurrentProblem[D], ProbPtr, &Again);
							ProbPtr = &CurrentProblem[D];
						}
						}/* I am  currently ... */
						else if (ChildPtr->Suspended == SUSPNOW) { /* time to reactivate */
							DEBUG(DEBUGNODE, 0,
								fprintf(stdout, "Node[%d]: Reactivating AFTER UPDATE suspended process (%o)\n  ",
								Nugget_MyNode, ChildPtr->UniqueID);
								PrintProblem(&ChildPtr->Value);
								fprintf(stdout, "\n  Prev child:  ");
								PrintProblem(&ChildPtr->Suspension->PrevChild);
								fprintf(stdout, "\n");
							);
							ChildPtr->Suspended = SUSPWAS;
							ERROR(ChildPtr->Available, 85, "still available");
							WasSuspendedWorkCount++;
							ChildPtr->Available = true;
						}/* time to reactivate */
					}/* Not given away */
					}
				}
				BusyWait = false;
				break; /* case Updating */
			case UpdatingUp:
				DEBUG(DEBUGNODE, 3,
					fprintf(stdout, "Node[%d]: Got updateUp\n",Nugget_MyNode););
				{	WorkPtrType ParentPtr = FindWork(Internal.UniqueID);
					Boolean AgainUp, AgainDown;
					if (ParentPtr != null) { /* hasn't been finished */
						DEBUG(DEBUGNODE, 3,
							fprintf(stdout, "Node[%d]: Newer copy\n",
								Nugget_MyNode););
						
						ConvertProblemIn(&IMsgPtr->Child, &Problem, "UpdatingUp");
						UpdateUp(&Problem, &ParentPtr->Value, &AgainUp, &AgainDown);
						/* Update must handle case where both args are the same
						problem, but first arg has new inherited info. */
						if (ParentPtr->Suspended == SUSPNOW) { /* time to reactivate */
						/* forlater: not clear that we should only try to reactivate parent
						if it has no more outstanding problems; maybe whenever a child
						finishes. */
						DEBUG(DEBUGNODE, 0,
							fprintf(stdout, "Node[%d]: Reactivating suspended process (%o)\n  ",
							Nugget_MyNode, ParentPtr->UniqueID);
							PrintProblem(&ParentPtr->Value);
							fprintf(stdout, "\n  Prev child:  ");
							PrintProblem(&ParentPtr->Suspension->PrevChild);
							fprintf(stdout, "\n");
						);
						ParentPtr->Suspended = SUSPWAS;
						ERROR(ParentPtr->Available, 85, "still available");
						ParentPtr->Available = true;
						WasSuspendedWorkCount++;
						}/* time to reactivate */
						if (AgainDown) DoUpdate(ParentPtr);
						if (AgainUp) DoUpdateUp(ParentPtr);
					}
				}
				break; /* case Updating */
			case Work:
#ifdef notneeded
				/* incoming message was initial work, a response to our
				request, or to our notice that we are repeating work. */
				DEBUG(DEBUGNODE, 3,
					fprintf(stdout, "Node[%d]: Received work (%o) from %d ",
						Nugget_MyNode, Internal.UniqueID, Internal.Asker);
					PrintDewey(&Internal.ChildDewey);
					fprintf(stdout, "\n");
				);
				if (Internal.Redundancy > 0)
					fprintf(stdout, "Node[%d] : Redundancy of incoming work is %d\n",
						Nugget_MyNode,Internal.Redundancy);
				if (Internal.Redundancy < CurrentRedundancy) {
					ERROR(Internal.Redundancy < 0, 6, "Got bad redundancy");
					*AbortPtr = true;
					CurrentRedundancy = Internal.Redundancy;
				}
				if (DeweySubsumes(&DeweyPrefix, &Internal.ChildDewey)) {
					/* fault tolerant rule c */
					DEBUG(DEBUGNODE, 3,
						fprintf(stdout, "Node[%d]: Got (%o)", Nugget_MyNode,
							Internal.UniqueID);
						PrintDewey(&Internal.ChildDewey);
						fprintf(stdout, " while working on ");
						PrintDewey(&DeweyPrefix);
						fprintf(stdout, "\n");
					);
					*AbortPtr = true;
					CurrentRedundancy = Internal.Redundancy;
				}
				if (Internal.Redundancy == CurrentRedundancy) {
					WorkPtr = FindWork(Internal.UniqueID);
					if (WorkPtr != null) {
						/* we gave it away, have gotten it back.  Just
						reinstall it as open; don't change its Redundancy. */
						ERROR(WorkPtr->WhereTo == OPEN, 11, "open?");
						ERROR(WorkPtr->WhereTo == Nugget_MyNode, 11, "why me?");
						ERROR(WorkPtr->Available, 12, "avail?");
						WorkPtr->WhereTo = OPEN;
						WorkPtr->Available = true;
						WorkPtr->Suspended = SUSPNOT;
						AvailWorkCount++;
					} else { /* new to us */
						ConvertProblemIn(&IMsgPtr->Child, &Problem, "NewWork");
						(void) NewWork(&Internal.ChildDewey, Internal.UniqueID,
							Internal.Asker, (WorkPtrType) null,
							CurrentRedundancy, &Problem);
							/* incs AvailWorkCount */
					}
					BusyWait = false; /* no need to wait any longer */
					/* fault tolerant rule d */
					for (WorkPtr = WorkHeaderPtr->Next;
					  WorkPtr != null;
					  WorkPtr = WorkPtr->Next) {
						if (!WorkPtr->Available &&
							WorkPtr->WhereTo != Nugget_MyNode &&
							DeweySubsumes(&Internal.ChildDewey, &WorkPtr->Dew))
						{ /* need to reprieve Worker */
							DEBUG(DEBUGNODE, 3,
								fprintf(stdout, "Node[%d]: Reprieving %d\n",
									Nugget_MyNode, WorkPtr->WhereTo);
							);
							/* not in */ HandleRequest(WorkPtr->WhereTo, PartitionSize);
							/* may not give worker a piece of Child,
							but at least will give better than what it
							is working on. */
						} /* need to reprieve Worker */
					} /* for WorkPtr */
				} else { /* new work at higher than CurrentRedundancy */
					/* work we were given may be discarded */
					DEBUG(DEBUGNODE, 3,
						fprintf(stdout, "Node[%d]: Ignoring (%o) [%d]", Nugget_MyNode,
							Internal.UniqueID, Internal.Redundancy);
						PrintDewey(&Internal.ChildDewey);
						fprintf(stdout, "from %d\n", Source);
					);
				}
				break; /* case Work */
#endif

				/* incoming message was work, a response to our request. */
				DEBUG(DEBUGNODE, 3,
					fprintf(stdout, "Node[%d]: Received work (%o) from %d ",
						Nugget_MyNode, Internal.UniqueID, Internal.Asker);
					PrintDewey(&Internal.ChildDewey);
					fprintf(stdout, "\n");
				);
				if (Internal.Redundancy > 0)
					DEBUG( DEBUGREDO, 1,
						fprintf(stdout, 
						"Node[%d] : Redundancy of incoming work is %d\n", 							Nugget_MyNode,Internal.Redundancy);
					);

				WorkPtr = FindWork(Internal.UniqueID);
				if (WorkPtr != null) {
					/* we gave it away, have gotten it back. */
					WorkPtr->WhereTo = OPEN;
					WorkPtr->Available = true;
					WorkPtr->Suspended = SUSPNOT;
					AvailWorkCount++;
					DEBUG( DEBUGNODE, 1,
						fprintf(stdout, 
							"Node[%d] : Got back given away work\n",
								Nugget_MyNode);
					);
				} else { /* new to us */
					ConvertProblemIn(&IMsgPtr->Child, &Problem, "NewWork");
#ifdef notneeded
					/* we got a new work from outside it should be marked
					with redundancy of incoming work using
					CurrentRedundancy means assigning it redundancy of
					whatever work we are currently working on 
					CHANGED BELOW */
					(void) NewWork(&Internal.ChildDewey, Internal.UniqueID,
						Internal.Asker, (WorkPtrType) null,
						CurrentRedundancy, &Problem);
#endif
					(void) NewWork(&Internal.ChildDewey, Internal.UniqueID,
						Internal.Asker, (WorkPtrType) null,
						Internal.Redundancy, &Problem);
						/* incs AvailWorkCount */
				}
				BusyWait = false; /* no need to wait any longer */
				break; /* case Work */

			case Result:
#ifdef notneeded
				DEBUG(DEBUGNODE, 3,
					fprintf(stdout, "Node[%d]: Report (%o) ", Nugget_MyNode,
						Internal.UniqueID);
					PrintDewey(&Internal.ChildDewey);
					fprintf(stdout, "from %d\n", Source);
				);
				PushTiming(WorkTime);
				WorkPtr = FindWork(Internal.UniqueID);
				if (WorkPtr != null) { /* we are still interested */
					ConvertProblemIn(&IMsgPtr->Child, &WorkPtr->Value,
						"NResult");
					TellParent(WorkPtr);
				}
				if (DeweySubsumes(&Internal.ChildDewey, &DeweyPrefix)) {
					/* we are doing work redundant to this answer */
					DEBUG(DEBUGNODE, 3,
						fprintf(stdout, "Node[%d]: Stopped redundant work",
							Nugget_MyNode);
						PrintDewey(&DeweyPrefix);
						fprintf(stdout, "\n");
					);
					*AbortPtr = true;
				}
				PopTiming();
				BusyWait = false; /* this result may unsuspend some problem
					and therefore allow me to make progress. */
				break; /* case Result */
#endif
				DEBUG(DEBUGNODE, 3,
					fprintf(stdout, "Node[%d]: Report (%o) ", Nugget_MyNode,
						Internal.UniqueID);
					PrintDewey(&Internal.ChildDewey);
					fprintf(stdout, "from %d\n", Source);
				);
				PushTiming(WorkTime);
				WorkPtr = FindWork(Internal.UniqueID);
				if (WorkPtr != null) { /* we are still interested */
					ConvertProblemIn(&IMsgPtr->Child, &WorkPtr->Value,
						"NResult");
					TellParent(WorkPtr);
				}
				PopTiming();
				BusyWait = false; /* this result may unsuspend some problem
					and therefore allow me to make progress. */
				break; /* case Result */

			case GlobalInfo:
				/* incoming message was new global info */
				PushTiming(WorkTime);
				{	InfoType Information;
					ConvertInfoIn(&IMsgPtr->SomeInfo, &Information);
					UseNewInfo(&Information);
				}
				PopTiming();
				break; /* case GlobalInfo */

			case Repeating:

#ifdef notneeded
				/* fault tolerance rule b */
				DEBUG(DEBUGNODE, 3,
					fprintf(stdout, "Node[%d]: %d is repeating (%o)", Nugget_MyNode,
						Source, Internal.UniqueID);
					PrintDewey(&Internal.ChildDewey);
					fprintf(stdout, "\n");
				);
				/* we could test to see if ChildDewey is an ancestor
					of the work we are now engaged in:
						if DeweySubsumes(&Internal.ChildDewey, &DeweyPrefix)
					but the following test seems more general.
				*/
				if (Internal.Redundancy >= CurrentRedundancy) {
					/* not in */ HandleRequest(Internal.Asker, PartitionSize);
					DEBUG(DEBUGNODE, 3,
					fprintf(stdout, 
					"Node[%d]: We treat as a request.\n",
					Nugget_MyNode);
					);
				}
				break; /* case Repeating */
#endif
				ERROR(1,1,"No one should send a repeat message");
				break;

#ifdef pvmred
			case BadTask:
				DEBUG( DEBUGREDO, 1,
					fprintf(stdout,"Node[%d] : Got bad work info\n",
						Nugget_MyNode);
				);

				/* release any subsumed work 
				of old generation (lower redundancy) */

				RemoveUselessWork(&Internal.ChildDewey,Internal.Redundancy);

				break;
#endif /* pvmred */

#ifdef newnode
			case NewNode:
				DEBUG( DEBUGNODE, 1,
					fprintf(stdout,"Node[%d] : Got NewNode Msg\n",
						Nugget_MyNode);
				);

				/* a new node has joined 
				   update parameters */

				Nugget_MySize++;
				NumHelpers++;
				PartitionSize++;
				HelperList[PartitionSize-2] = PartitionSize;
				Address[Nugget_MySize]=Internal.NewAddress;

				break;
#endif

			default:
				ERROR(true, 1, "Bad message type");	
			} /* if not termination switch */
			SAP_ReuseMsg(InPtr);
			/* prepare for next time through the loop */
			InPtr = null;
		} /* got a msg */

		if (Termination || *AbortPtr || (! BusyWait)) {
			/* whether or not a message came in, we get out of the loop
			if we know about termination, we want to abort the current work,
			or there is no need to wait any longer */
			break;
		} else {
			/* no incoming message, but we are expected to wait for one.
			While we are waiting, check if there is someone working
			on the expected work or there is a bad worker.
			If the worker is no more around, reclaim the work, and
			inform other workers about it */
#ifdef pvmred
			WorkPtr = null; 	/* no redundancy allowed */
			ReDoing(&WorkPtr); 	
					/* incs AvailWorkCount if successful */
					/* increments redundancy count for
					   work to be redone */

			if (WorkPtr != null){
				/* found a problem assigned to a worker
				who is no more around */

				/* clean your work list */
				RemoveUselessWork(&WorkPtr->Dew, (WorkPtr->Redundancy-1) );

				/* inform the workers working on subproblem to
				discard the subproblems of current problem */

				DEBUG( DEBUGNODE,1,
					fprintf(stdout, "Node[%d] : Sending bad work info\n", 
						Nugget_MyNode);
				);

				PushTiming(CommTime);
				OutPtr = SAP_GetSndBuf();
				{	
					InterNodeMessage *IMsgPtr = (InterNodeMessage *) OutPtr;
					IMsgPtr->Kind = htonl(BadTask);
					IMsgPtr->Redundancy = htonl(WorkPtr->Redundancy-1);
							/* send redundancy of old work */
					DeweyCopyOut(WorkPtr->Dew, IMsgPtr->ChildDewey);
				}
				SAP_Broadcast(OutPtr, sizeof(InterNodeMessage), 1, PartitionSize);
				PopTiming();

				DEBUG(DEBUGNODE, 1,
				    fprintf(stdout,"Node[%d]: Decided to reclaim (%o) [old redundancy %d] 
						given to %d ", Nugget_MyNode, WorkPtr->UniqueID,
						WorkPtr->Redundancy, WorkPtr->WhereTo);
					PrintDewey(&WorkPtr->Dew);
					if (WorkPtr->ParentPtr != null)
						fprintf(stdout, "parent has unsolved = %d\n",
							WorkPtr->ParentPtr->UnsolvedCount);
					fprintf(stdout, "\n");
				);

				/* no need to set Abort; it is ignored */

				break;
			} /* found a bad worker */
#else
			/* no incoming message, but we are expected to wait for one.
			While we are waiting, we might as well redo work. */

			if (Debug & DEBUGREDO) {
				ReDoing(&WorkPtr); /* incs AvailWorkCount if successful */
			} else {
				WorkPtr = null; /* no redundancy allowed */
			}
			if (WorkPtr != null){
				/* found something to redo */
				TellPeerRepeating(WorkPtr);
				DEBUG(DEBUGNODE, 1,
					fprintf(stdout, "Node[%d]: Decided to redo (%o) [%d] given to %d ",
						Nugget_MyNode, WorkPtr->UniqueID,
						WorkPtr->Redundancy, WorkPtr->WhereTo);
					PrintDewey(&WorkPtr->Dew);
					if (WorkPtr->ParentPtr != null)
						fprintf(stdout, "parent has unsolved = %d\n",
							WorkPtr->ParentPtr->UnsolvedCount);
					fprintf(stdout, "\n");
				);
				/* no need to set Abort; it is ignored */
				break;
			} /* found something to redo */
#endif
		} /* no incoming message */
		TimingStack[TimingStackTop] = IdleTime; /* just waiting */
		(LoopCount)++;
		if ( LoopCount > LoopLimit ) {
			/* changed 8/1/91: game players can take forever here */
			/* waited a long time.  Time to ask again. */
			LoopCount = 0;
			LoopLimit *= 2;
			DEBUG(DEBUGNODE, 3,
				fprintf(stdout, "Node[%d]: Asking again\n", Nugget_MyNode);
			);
			AskForHelp(false, Nugget_MyNode, 0);
		}
	} /* while true */
	if (Termination &&
	  (TimingStack[TimingStackTop] == CommTime ||
	  TimingStack[TimingStackTop] == IdleTime))
	{
		TimingStack[TimingStackTop] = TermTime; /* charge to termination,
			not to communication or idle */
	}
	PopTiming();
} /* CheckIncoming */

void BackTrack()
/* Recursive backtrack to solve the problem.
Uses an explicit stack of unfinished work:  CurrentProblem.
The problem to expand is the first in the Heap at the current GiveAwayDepth.
*/
{
	short Done;
	Boolean Abort;
	int PrevDepth;
	WorkPtrType TmpWorkPtr;
	ProblemPtr ParentPtr;
	Boolean NeedUpdate, NeedUpdateUp;
	InBufPtr InPtr;
	int Source;
	 
	HaveBeenIdle = false; /* not idle any more */
	if (AvailWorkCount + WasSuspendedWorkCount > 0) {
		SelectWork(&RootWorkPtr, false);
		ERROR(RootWorkPtr->Suspended == SUSPNOW, 77,"executing suspended node");
		/* initialize */
			CurrentRedundancy = RootWorkPtr->Redundancy;
			/* WorkClear(CurrentRedundancy); */
				/* anything more redundant may be scrapped, especially parts
				we have given away */
			PushTiming(WorkTime);
			PrevDepth = -1;
			Depth = 0;
			GiveAwayDepth = 0;
			ProblemCopy(&RootWorkPtr->Value, &CurrentProblem[GiveAwayDepth]);
			DeweyCopy(&RootWorkPtr->Dew, &DeweyPrefix);
			RootID = RootWorkPtr->UniqueID;
			if (RootWorkPtr->Suspended == SUSPWAS || RootWorkPtr->UnsolvedCount > 0) {
				RootWorkPtr->Suspended = SUSPNOT;
				DeweyDelta[1]=RootWorkPtr->Suspension->GenSoFar;
				ProblemCopy(&RootWorkPtr->Suspension->PrevChild,
					&CurrentProblem[GiveAwayDepth+1]);
				free(RootWorkPtr->Suspension);
				RootWorkPtr->Suspension = (SuspensionType *) 0;
				DEBUG(DEBUGNODE, 0,
					fprintf(stdout,
						"Node[%d]: Working on (%o) having created %d kids\n  ",
						Nugget_MyNode, RootWorkPtr->UniqueID, DeweyDelta[1]);
					PrintProblem(&RootWorkPtr->Value);
					fprintf(stdout, "\n  Prev child:  ");
					PrintProblem(&CurrentProblem[GiveAwayDepth+1]);
					fprintf(stdout, "\n");
				);
			} else
				DeweyDelta[1] = 0;
		DEBUG(DEBUGNODE, 2,
			fprintf(stdout, "Node[%d]: Bktrk (%o) ", Nugget_MyNode,
				RootID);
			PrintDewey(&RootWorkPtr->Dew);
			PrintProblem(&CurrentProblem[GiveAwayDepth]);
		);
		while (true) {
			/* each iteration advances the traversal up or down the tree */
			if (PrevDepth < Depth) { /* moving down */
				/* this is the first time CurrentProblem[Depth] has been
					generated (unless reactivated) */
				if (Depth == GiveAwayDepth){
					WorkPtrType WorkPtr = FindWork(RootID);
					ParentPtr = (ProblemType *) &(WorkPtr->Value);
				}else
					ParentPtr = &(CurrentProblem[Depth]);
				Generate(&Done, ParentPtr, &CurrentProblem[Depth+1]);
				if (Done == MORELATER){
					/* All problems in the stack must be put
					on the list and the one at Depth must be suspended*/
					WorkPtrType SuspWorkPtr, SuspChildPtr;
					ProblemType SuspTmpProb;
					int seq;
					SuspWorkPtr = FindWork(RootID);
					if (SuspWorkPtr == null) { /* the work
						that BackTrack is engaged in is
						redundant and has been finished. */
						DEBUG(DEBUGNODE, 3, fprintf(stdout, "Node[%d]: Nothing at the moment\n",Nugget_MyNode););
						break;
					}
					for (seq = GiveAwayDepth+1; seq<=Depth; seq++){
						ProblemCopy(&CurrentProblem[seq], &SuspTmpProb);
						SuspWorkPtr->Suspension = (SuspensionType *)
							malloc(sizeof(SuspensionType)+ProblemSize);
						SuspWorkPtr->Suspension->GenSoFar =
							DeweyDelta[seq];
						ProblemCopy(&SuspTmpProb, &SuspWorkPtr->Suspension->PrevChild);
						DeweyPrefix.Length++;
						DeweyPrefix.DNums[DeweyPrefix.Length]=DeweyDelta[seq];
						SuspWorkPtr->Available = true;
						SuspWorkPtr->Suspended = SUSPWAS;
						SuspWorkPtr->WhereTo = Nugget_MyNode;
						WasSuspendedWorkCount++;
						/*
						SuspChildPtr = NewWork(&DeweyPrefix, LastUniqueID++,
						Nugget_MyNode, SuspWorkPtr, CurrentRedundancy, &SuspTmpProb);
						*/
						SuspChildPtr = NewWork(&DeweyPrefix, LastUniqueID++,
						Nugget_MyNode, SuspWorkPtr, SuspWorkPtr->Redundancy, &SuspTmpProb);
						AvailWorkCount--;/* Was incremented in NewWork*/
						DEBUG(DEBUGNODE, 0,
							fprintf(stdout,
							"Node[%d]: Suspending IN BACKTRACK (%o) having created %d kids\n  ",
							Nugget_MyNode, SuspWorkPtr->UniqueID,
							SuspWorkPtr->Suspension->GenSoFar);
							PrintProblem(&SuspWorkPtr->Value);
							fprintf(stdout, "\n  Prev child:  ");
							PrintProblem(&SuspWorkPtr->Suspension->PrevChild);
							fprintf(stdout, "\n");
						);
						SuspWorkPtr = SuspChildPtr;
					}/* for */
					SuspWorkPtr->Suspension = (SuspensionType *)
						malloc(sizeof(SuspensionType)+ProblemSize);
					SuspWorkPtr->Suspension->GenSoFar =
						DeweyDelta[Depth+1];
					ProblemCopy(&CurrentProblem[Depth+1], &SuspWorkPtr->Suspension->PrevChild);
					/* Suspend the node at Depth */
					SuspWorkPtr->Suspended = SUSPNOW;
					SuspWorkPtr->WhereTo = Nugget_MyNode;
					SuspWorkPtr->Available = false;
					break;
				}
				PrevDepth = Depth;
				if (Done == DONE) {
					/* go back up; leave PrevDepth==Depth as a sign there
						were no children of this node. */
					DEBUG(DEBUGNODE, 5,
						fprintf(stdout, "Node[%d]: No more children down here.\n",
							Nugget_MyNode);
					);
				} else {
					Depth++; /* continue down */
					if (Depth > 1) DeweyDelta[Depth] = 0;
					ERROR(Depth >= MaxDepth, 83, "Exceeded max tree depth");
					DEBUG(DEBUGNODE, 5,
						fprintf(stdout, "Node[%d]: Going down to first child\n",
							Nugget_MyNode);
					);
				}
			} else { /* coming up to this level from below */
				/* have now finished CurrentProblem[Depth]. */
				ERROR(Depth < GiveAwayDepth, 29, "Bad depth");
				if (Depth == GiveAwayDepth) {
					/* report to problem's owner. */
					TmpWorkPtr = FindWork(RootID);
					if (TmpWorkPtr != RootWorkPtr)
						fprintf(stdout, "Node[%d]: Changed from under me!\n",
							Nugget_MyNode);
					if (TmpWorkPtr == null) {
						/* someone else finished this work for us! */
						DEBUG(DEBUGNODE, 2,
							fprintf(stdout, "Node[%d]: Someone else finished our work!\n",
								Nugget_MyNode);
						);
						break;
					}
					TellParent(TmpWorkPtr);
					break;
				} else { /* Depth > GiveAwayDepth */
					Boolean  AgainDown, AgainUp;
					WorkPtrType parent;
					if (Depth == GiveAwayDepth +1) {
					/*Special case where the parent is in
					  work list and it may have better
					  synthesized value */
						parent = FindWork(RootID);
						if (parent == null){
							break;
						}else{
							ParentPtr = (ProblemType *) &(parent->Value);
						}
					}/* Special case ... */
					else{
						ParentPtr = &(CurrentProblem[Depth-1]);
					}
					Combine(&CurrentProblem[Depth], ParentPtr, &NeedUpdate, &NeedUpdateUp);
					DEBUG(DEBUGNODE, 5,
						fprintf(stdout, "Node[%d]: Combining at depth=%d\n", 
								Nugget_MyNode,Depth);
					);
					if (NeedUpdateUp){
						
						int D = Depth - 2;
						ProblemPtr ChildPtr;
						ProblemPtr TemParentPtr;
						AgainUp = false;
						while( D >= GiveAwayDepth){
							ChildPtr = &(CurrentProblem[D+1]);
							TemParentPtr = &(CurrentProblem[D]);
							UpdateUp(ChildPtr, TemParentPtr, &AgainUp, &AgainDown);
							/* AgainDown must be false each time in this loop 
							   since other children of parent do not exist yet.
							   Also parent could not have been suspended */
							if (!AgainUp)
								break;
							D--;
						}/* while */
						if (AgainUp){/*The node at
						GiveAwayDepth is updated, but its
						copy in work list is not */
							WorkPtrType NewParentPtr = FindWork(RootID);
							if (NewParentPtr != null){
							UpdateUp(ChildPtr, &NewParentPtr->Value, &AgainUp, &AgainDown);
							if (AgainUp)
								DoUpdateUp(NewParentPtr);
							}/*NewParentPtr != null*/
						}/* The Node ...*/
					}/* NeedUpdateUp */
					if ((NeedUpdate || NeedUpdateUp) && Depth == GiveAwayDepth+1) {
						/* our parent is in Work list (pointed to by parent),
						may already have other children, which need to be
						given update */
							DEBUG(DEBUGNODE, 3,
								fprintf(stdout, "Node[%d]: Local newer copy\n",
								Nugget_MyNode););
							if (NeedUpdate)
								DoUpdate(parent);
							if (NeedUpdateUp)
								DoUpdateUp(parent);
					} /* update children of an explicit parent */
					/* generate the next sibling of CurrentProblem[Depth] */
					Generate(&Done, ParentPtr, &CurrentProblem[Depth]);
					if (Done == MORELATER){
						/* All problems in the stack must be suspended */
						WorkPtrType SuspWorkPtr, SuspChildPtr;
						ProblemType SuspTmpProb;
						int seq;
						SuspWorkPtr = FindWork(RootID);
						if (SuspWorkPtr == null) { /* the work
						that BackTrack is engaged in is
						redundant and has been finished. */
						DEBUG(DEBUGNODE, 3, fprintf(stdout, "Node[%d]: Nothing at the moment\n",
							Nugget_MyNode););
						break;
						}
						for (seq = GiveAwayDepth+1; seq<Depth; seq++){
							SuspWorkPtr->WhereTo = Nugget_MyNode;
							SuspWorkPtr->Available = true;
							SuspWorkPtr->Suspended= SUSPWAS;
							WasSuspendedWorkCount++;
							DeweyPrefix.Length++;
							DeweyPrefix.DNums[DeweyPrefix.Length]=DeweyDelta[seq];
							ProblemCopy(&CurrentProblem[seq], &SuspTmpProb);
							/*
							SuspChildPtr = NewWork(&DeweyPrefix, LastUniqueID++,
							Nugget_MyNode, SuspWorkPtr, CurrentRedundancy, &SuspTmpProb);
							*/
							SuspChildPtr = NewWork(&DeweyPrefix, LastUniqueID++,
							Nugget_MyNode, SuspWorkPtr, SuspWorkPtr->Redundancy, &SuspTmpProb);
							AvailWorkCount--;/*was incremented in NewWork */
							SuspWorkPtr->Suspension = (SuspensionType *)
								malloc(sizeof(SuspensionType)+ProblemSize);
							SuspWorkPtr->Suspension->GenSoFar =
								DeweyDelta[seq]+1;
							ProblemCopy(&SuspTmpProb, &SuspWorkPtr->Suspension->PrevChild);
							DEBUG(DEBUGNODE, 0,
								fprintf(stdout,
								"Node[%d]: Suspending IN BACKTRACK (%o) having created %d kids\n  ",
								Nugget_MyNode, SuspWorkPtr->UniqueID,
								SuspWorkPtr->Suspension->GenSoFar);
								PrintProblem(&SuspWorkPtr->Value);
								fprintf(stdout, "\n  Prev child:  ");
								PrintProblem(&SuspWorkPtr->Suspension->PrevChild);
								fprintf(stdout, "\n");
							);
							SuspWorkPtr = SuspChildPtr;
						}/* for */
						SuspWorkPtr->Suspension = (SuspensionType *)
							malloc(sizeof(SuspensionType)+ProblemSize);
						SuspWorkPtr->Suspension->GenSoFar =
							DeweyDelta[Depth]+1;
						ProblemCopy(&CurrentProblem[Depth], &SuspWorkPtr->Suspension->PrevChild);
						/*Suspend the node that was at level Depth-1 */
						SuspWorkPtr->Suspended = SUSPNOW;
						SuspWorkPtr->WhereTo = Nugget_MyNode;
						SuspWorkPtr->Available = false;
						break;
					}
					if (Done == DONE) {
						/* no more siblings; go back up */
						PrevDepth = Depth;
						Depth--;
					} else { /* not done */
						/* found a sibling */
						DeweyDelta[Depth]++;
						DEBUG(DEBUGNODE, 5,
							fprintf(stdout, "Node[%d]: Moving to sibling %d\n",
								Nugget_MyNode, DeweyDelta[Depth]);
						);
						PrevDepth = Depth - 1;
						if ( GlobalInUse && SAP_ProbeGlobal() ){
							SAP_RcvGlobal2(&InPtr, &Source); /* may be null */
							/* will be converted to internal format in
							CheckIncoming */
						} else {
							InPtr = null;
						}
						if (InPtr != null || SAP_ProbeAny() ) {
							/* check here */
							CheckIncoming(false, &Abort, InPtr, Source); 
							if (Abort) {
								DEBUG(DEBUGNODE, 2,
									fprintf(stdout, "Node[%d]: Aborting (%o) ",
										Nugget_MyNode, RootID);
									PrintProblem(&RootWorkPtr->Value);
									fprintf(stdout, " at redundancy %d\n",
										RootWorkPtr->Redundancy);
								);
								/* we have been working from a copy in
								CurrentProblem, so nothing to restore */
							}
							if (Termination || Abort) {
								break;
							} /* Termination || Abort */
						} /* SAP_ProbeAny() */
					} /* not done */
				} /* Depth > GiveAwayDepth */
			} /* coming up from below */
		} /* loop */
		/* we may have exited the loop because of Abort */
		DEBUG(DEBUGNODE, 2,
			fprintf(stdout, "Node[%d]: Stopped working on (%o) ", Nugget_MyNode,
				RootID);
			PrintDewey(&DeweyPrefix);
			fprintf(stdout, "\n");
		);
		PopTiming();
	} /* if available work */
	DeweyPrefix.Length = MaxDepth; /* absurd initial value */
	RootID = -1; /*absurd initial value */
} /* BackTrack */

void NodeDoIt() 
/* accept work from the host, either 'ask for help' or 'here is a problem'.
*/
{
	ProblemType Prob;
	int Seq;
	Dewey Dew; /* of initial problem */
	Boolean Dummy;
	InBufPtr InPtr;
	int Source;
	 
#ifdef newnode
	OutBufPtr OutPtr;
	InterNodeMessage Internal, *IMsgPtr;
#endif

	ProblemNumber++;
	ApplicInit(&ProblemSize);
	while (true) { 
		/* first message has to come from host to signal
		start of problem. Ignore spurious messages */
		DEBUG(DEBUGNODE, 2,
			fprintf(stdout, "Node[%d]: Waiting for message from host\n",
			Nugget_MyNode);
		);
		SAP_Rcv2(&InPtr, &Source);
		if (Source == 0) {
			break;
		} 
		else { /* not from host */
#ifdef newnode
			DEBUG(DEBUGNODE, 0,
				fprintf(stdout, 
					"Node[%d]: Expecting message from host got from %d\n",
						Nugget_MyNode, Source);
			);

			IMsgPtr = (InterNodeMessage *) InPtr;
			Internal.Kind = ntohl(IMsgPtr->Kind);
			Internal.ProblemNumber = ntohl(IMsgPtr->ProblemNumber);
			Internal.NewAddress = ntohl(IMsgPtr->NewAddress);

			if( Internal.Kind == NewNode ){
				DEBUG(DEBUGNODE, 0,
					fprintf(stdout, 
					"Node[%d]: Unexpected message is New Node message\n",
						Nugget_MyNode);
				);	

				/* update parameters */

				Nugget_MySize++;
				NumHelpers++;
				PartitionSize++;
				HelperList[PartitionSize-2] = PartitionSize;
				Address[Nugget_MySize]=Internal.NewAddress;

				/* you are a left behind person for the currently
				concluded round try join us for the round about to
				begin. If we fail to get you with us in time you
				will be in for the next round atleast. Worker 1
				takes the responsibility to inform you */
				if(Nugget_MyNode==1){
					OutPtr = SAP_GetSndBuf();
					{
						InterNodeMessage *IMsgPtr=
									(InterNodeMessage *)OutPtr;
						IMsgPtr->Kind = htonl(Terminate);
						IMsgPtr->ProblemNumber = 
									htonl(Internal.ProblemNumber);
					}
					SAP_Snd3(OutPtr,PartitionSize,
							sizeof(InterNodeMessage));
				}
			}
#endif
			SAP_ReuseMsg(InPtr);
		} /* message from host */
	}
	{	ProblemMessage *PMsgPtr = (ProblemMessage *) InPtr;
		ConvertProblemIn(&PMsgPtr->Prob, &Prob, "NDoIt");
		AnswerSpec = ntohl(PMsgPtr->Answer);
		GlobalInUse = ntohl(PMsgPtr->UseGlobal);
		PartitionSize = ntohl(PMsgPtr->PartSize);
		if (Nugget_MyNode > PartitionSize){
			/* I am a supernumerary on this calculation */
			SAP_ReuseMsg(InPtr);
			/* SAP_Synchronize(); /* to allow the others to continue */
			return;
		}
		WorkFract = ntohl(PMsgPtr->WorkFraction);
		if (Nugget_MyNode == 1){
			/* Machine 1 gives away only 1 of the problems the first time. */
			WorkFract = -1; 
		}
		else{
			/* we will always give away half of the problems. */
			WorkFract = -1; 
		}
		NumHelpers = ntohl(PMsgPtr->NumHelp);
		if (NumHelpers >= PartitionSize) {
			NumHelpers = PartitionSize - 1;
			/* I am never a helper for myself */
		};
		Debug = ntohl(PMsgPtr->DebugLevel);

#ifdef newnode
		AddedLater = ntohl(PMsgPtr->AddedLater);

		/* host only sends one message at start of each problem
		let us go by hosts count of problem number */

		ProblemNumber = ntohl(PMsgPtr->ProblemNumber);

		if(AddedLater){
			/* we will be old node for successive rounds */
			AddedLater=0;

			/* let others know about you */
			PushTiming(CommTime);
			OutPtr = SAP_GetSndBuf();
			{	InterNodeMessage *IMsgPtr = (InterNodeMessage *) OutPtr;
				IMsgPtr->Kind = htonl(NewNode);
				IMsgPtr->ProblemNumber = htonl(ProblemNumber);
				IMsgPtr->NewAddress = htonl(Address[Nugget_MySize]);
			}
			SAP_Broadcast(OutPtr,sizeof(InterNodeMessage), 1,PartitionSize);
			PopTiming();
		}
#endif
	} /* with ProblemMessage */

	setvbuf(stdout,NULL,_IOLBF,BUFSIZ);

	SAP_ReuseMsg(InPtr);
	Termination = false;
	NextToAsk = Nugget_MyNode % PartitionSize + 1; /* initialization */
	WorkInit(); /* clear any work from the last trial */
	DeweyPrefix.Length = MaxDepth; /* absurd initial value */
	StartTiming(CommTime);
	for (Seq = 0; Seq <= 3+MaxRedundancy; Seq++) {
		Time[Seq] = 0;
	}
	/*LocalSequence = 1;*/
	if (Nugget_MyNode != 1) {
		/* we will go begging */
		DEBUG(DEBUGNODE, 4, fprintf(stdout, "Node[%d]: Start by begging\n",
			Nugget_MyNode);
		);
#ifdef notneeded
		CurrentRedundancy = 10; /* will be reset to 0 by first work */
#else
		CurrentRedundancy = 0; 
#endif
		Depth = 0;
		GiveAwayDepth = 1;
		AskForHelp(true, Nugget_MyNode, 0);
		Dummy = false; /* to keep lint happy */
		CheckIncoming(true, &Dummy, (InBufPtr) null, Dummy);
		while( !Termination && AvailWorkCount+WasSuspendedWorkCount==0){
			AskForHelp(false, Nugget_MyNode, 0);
			CheckIncoming(true, &Dummy, (InBufPtr) null, Dummy);
		}
	} /* beggar */ 
	else { 
		/* Node 1 we are born rich */
		CurrentRedundancy = 0;
		Dew.Length = 0;
		(void) NewWork(&Dew, 0, 0, (WorkPtrType) null, 0, &Prob);
			/* incs AvailWorkCount */
	} /* born rich */
	while (! Termination && AvailWorkCount+WasSuspendedWorkCount != 0) {
		BackTrack();
		/* prevent attempting to give away work while we are empty */
		Depth = 0;
		GiveAwayDepth = 1;
		if (! Termination && AvailWorkCount+WasSuspendedWorkCount == 0) {
			AskForHelp(false, Nugget_MyNode, 0);
#ifdef notneeded
			CurrentRedundancy = 10; /* will be reset to 0 by first work */
#else
			CurrentRedundancy = 0; 
#endif
			while (! Termination && AvailWorkCount+WasSuspendedWorkCount == 0) {
				/* keep in touch with the world */
				CheckIncoming(true, &Dummy, (InBufPtr) null, Dummy);
			}
		} else {
			/* either (1) we aborted or (2) finished some work or (3) heard
			Termination.  In cases (1) and (2), there is already
			something waiting of greater importance in WorkGiven */
		}
	} /* while open work */
	if (Termination) {
		ApplicFinalize();
		DropOut();
	}
} /* NodeDoIt */

void NodeInit()
{
	int Seq, Tmp;

#ifdef newnode
	int numcpu;
	int numold;
	int num_tobeadded;
	int tasknumber;
	int newtid;
	OutBufPtr OutPtr;
#endif

	Nugget_UserRunCode = 0;

	LastUniqueID = Nugget_MyNode << 20;
	PartitionSize = Nugget_MySize;
	Tmp = Nugget_MyNode;
	srand((unsigned) Tmp);

	/* List of helpers */
	for (Seq = 0; Seq < Nugget_MyNode - 1; Seq++) {
		HelperList[Seq] = Seq + 1;
	}
	for ( ; Seq < PartitionSize-1; Seq++) {
		HelperList[Seq] = Seq + 2;
	}

	LocalSequence = 1;

#ifdef notneeded
	(void) ApplicInit(&Seq); /* discard result */
#endif

#ifdef newnode
	/* see if you started on a multiprocessor. Start task
	on each available CPU and inform controller */

	/* by default we expect numcpu to be 1 */
	numcpu=1;

	/* find correct value of num cpu based on your architecture */

#ifdef SUN4SOL2
	numcpu=sysconf(_SC_NPROCESSORS_ONLN);
#endif

	num_tobeadded=0;
	numold=SAP_WorkersOnMyHost();
	if(numold>0)
		num_tobeadded=numcpu-numold;

	if(num_tobeadded>0){
		for(tasknumber=1;tasknumber<=num_tobeadded;tasknumber++){
			newtid=SAP_SpawnNewNode(hostname);
			if(newtid>0){
				/* send message to host */
				PushTiming(TossTime);
				OutPtr = SAP_GetSndBuf();
				{	AnswerMessage *AMsgPtr = (AnswerMessage *) OutPtr;
					AMsgPtr->Status = htonl(StartedChild);
					AMsgPtr->ChildTid = htonl(newtid);
					AMsgPtr->Sequence = htonl(LocalSequence);
					(LocalSequence)++;
				}
				SAP_Snd2(OutPtr, 0);
				DEBUG(DEBUGNODE, 1, 
					fprintf(stdout, 
					"Node[%d]: Started on MPP. Starting more workers\n",
						Nugget_MyNode);
				);

				PopTiming();
			}
		}
	}
#endif

} /* NodeInit */
