
/*
 *
 *      Written for Advanced X Window Applications Programming
 *      modified by P. Cholewinski, University of Kentucky, 1995
 *
 */


#include   "xbook.h"



/*
 *      display.c
 *      X11 C routines for opening and closing display
 *      connections.
 *
 *      Routines:
 *      Display *SetUpDisplay( argc, argv, &screen )
 *      Display *OpenDisplay( display_name, &screen )
 *      CheckDisplayName( argc, argv, display_name )
 *      CloseDisplay( disp, window, gc )
 *      QuitX( display, error_message, error_file )
 *
 *	Written for Advanced X Window Applications Programming
 *
 */



Display *SetUpDisplay( argc, argv, screen )

int     argc;
char    *argv[];
int     *screen;

/*
 *      SetUpDisplay() checks the command-line
 *      parameters for a display name, and then
 *      calls OpenDisplay() to establish a connection
 *      to an X server.
 *
 *      If no connection can be established, the
 *      program will be terminated by a call to
 *      exit().
 *
 */

{       /* -- function SetUpDisplay */
        char    display_name[ 120 ];
        Display *display;


        CheckDisplayName( argc, argv,
                display_name );

        display = OpenDisplay( display_name, 
                                screen );

        if ( display == (Display *) NULL )
                {
                (void) fprintf( stderr, 
                                "Error could not open X display to %s.\n",
                                XDisplayName( display_name ) );
                exit( 1 );
                }

        return( display );

}       /* -- function SetUpDisplay */





Display *OpenDisplay( display_name, screen )

char    display_name[];
int     *screen;

/*
 *      OpenDisplay() tries to establish a connection to
 *      an X server.  If successful, it then gets the
 *      number of the default screen (used in many
 *      other X calls).  OpenDisplay() returns NULL
 *      on an error, so if you want to recover from
 *      a bad display name or errors in making
 *      the connection, use OpenDisplay(), rather
 *      than SetUpDisplay().  (For most X applications,
 *      failure to make a display connection is fatal.)
 *
 */


{       /* -- function OpenDisplay */
        Display *display;

        display = XOpenDisplay( display_name );

        if ( display != (Display *)NULL )
                {
                *screen = DefaultScreen( display );
                }

        return( display );

}       /* -- function OpenDisplay */





CheckDisplayName( argc, argv, display_name )

int     argc;
char    *argv[];
char    display_name[];

/*
 *      CheckDisplayName() checks the command-line arguments
 *      for a display name specification. It initializes display_name
 *      to NULL, if no -display command-line argument is found.
 *
 */

{       /* -- function CheckDisplayName */
        int     counter;

        display_name[ 0 ] = '\0';
        counter           = 1;

        while( counter < argc )
                {
                if ( strncmp( argv[ counter ], "-display", 8 ) == 0 )
                        {
                        counter++;
                        if ( counter < argc )
                                {
                                (void) strcpy( display_name, argv[ counter ] );
                                }
                        else
                                {
                                (void) fprintf( stderr, 
                                  "Error: usage is -display DisplayName\n" );
                                }
                        }

                counter++;
                }

}       /* -- function CheckDisplayName */




CloseDisplay( display, window, gc )

Display *display;
Window  window;
GC      gc;

/*
 *      CloseDisplay() collects a common set of routines
 *      executed together for programs that set up windows.
 *
 */

{       /* -- function CloseDisplay */

        XFreeGC( display, gc );

        XDestroySubwindows( display, window );

        XDestroyWindow( display, window );

        XFlush( display );

        XCloseDisplay( display );

}       /* -- function CloseDisplay */



QuitX( display, error_message, error_file )

Display *display;
char    error_message[], error_file[];

/*
 *      QuitX() is called to terminate a program, by calling exit().
 *      QuitX() also prints out an error message to stderr.
 */

{       /* -- function QuitX */

        (void) fprintf( stderr, "ERROR: %s%s\n", 
                error_message,
                error_file );

        XCloseDisplay( display );

        exit( 1 );

}       /* -- function QuitX */


/*
 *      end of file.
 */



unsigned long GetColor( display, name, colormap, default_color )

Display         *display;
char            name[];         
Colormap        colormap;       
unsigned long   default_color;  


{       
        unsigned long   color;
        XColor          rgbcolor, hardwarecolor;
        int             status;

        color = default_color;

        status = XLookupColor( display, colormap, name,
                        &rgbcolor, &hardwarecolor );

        if ( status != 0 )
                {
                status = XAllocColor( display, colormap,
                                &hardwarecolor );

                if ( status != 0 )
                        {
                        color = hardwarecolor.pixel;
                        }
                }

        return( color );

        
}       

Cursor MakeCursor( display, window, which_cursor )

Display         *display;
Window          window;
unsigned int    which_cursor;


{       
        Cursor  cursor;

        cursor = XCreateFontCursor( display, (unsigned) which_cursor );

        
        if ( cursor != (Cursor) None )
                {
                XDefineCursor( display, window, cursor );
                }

        return( cursor );

}       

Time    LastEventTime = CurrentTime;


CheckEvent( display, want_exposes, width, height, event, keysym )

Display *display;
int     want_exposes;   
int     width, height;  
XEvent  *event;
KeySym  *keysym;


{       
        int     status;


        status = False;         

        if ( XPending( display ) > 0 )
                {
                XNextEvent( display, event );

                status = ParseEvent( event, 
                                want_exposes, 
                                width, height, 
                                keysym );

                }

        
        return( status );

}       




NextEvent( display, want_exposes, width, height, event, keysym )

Display *display;
int     want_exposes;   
int     width, height;  
XEvent  *event;
KeySym  *keysym;


{       
        int     status;


        status = False;         

        while( status == False )
                {       
                XNextEvent( display, event );

                status = ParseEvent( event, 
                                want_exposes, 
                                width, height, 
                                keysym );
                }

}       



ParseEvent( event, want_exposes, width, height, keysym )

XEvent  *event;
int     want_exposes;
int     width, height;
KeySym  *keysym;

{       
        int     status;

        status = False;


        switch( event->type )
                {
                case ButtonPress:
                        LastEventTime = event->xbutton.time;

                        status = True;
                        break;
                case ClientMessage:
                        status = True;
                        break;
                case PropertyNotify:
                        LastEventTime = event->xproperty.time;

                        status = True;
                        break;
                case SelectionClear:
                        LastEventTime = event->xselectionclear.time;

                        status = True;
                        break;
                case SelectionNotify:
                        LastEventTime = event->xselection.time;

                        status = True;
                        break;
                case SelectionRequest:
                        LastEventTime = event->xselectionrequest.time;

                        status = True;
                        break;
                case ConfigureNotify:
                        if ( ( width != event->xconfigure.width ) || 
                             ( height != event->xconfigure.height ) )
                                {
                                status = True;
                                }
                        break;  
                case Expose:
                        status = True;

                        if ( want_exposes == False )
                                {
                                if ( event->xexpose.count != 0 )
                                        {
                                        status = False;
                                        }
                                }
                        break;

                case KeyPress:
                        LastEventTime = event->xkey.time;

                        status  = True;
                        *keysym = Key2Keysym( event );
                        break;
                case MappingNotify:
                        /*XRefreshKeyboardMapping( event );*/
			printf("don't know how to XRefreshKeyboardMapping\n");
                        break;
                }


        return( status );

}       



KeySym  Key2Keysym( keyevent )

XKeyEvent       *keyevent;

{       
        XComposeStatus  composestatus;
        KeySym          keysym;
        char            string[ 10 ];


        XLookupString( keyevent, string, 
                        9, &keysym, &composestatus );


        return( keysym );

}       


Time LastTimeStamp()


{       

        return( LastEventTime );

}       



/*
 *      font.c
 *      X11 routines to load in a font
 *
 *      Written for Advanced X Window Applications Programming
 *
 */


XFontStruct *LoadFont( display, gc, argc, argv, default_name )

Display *display;
GC      gc;
int     argc;
char    *argv[];
char    default_name[];


{       
        char            name[ BUFSIZE + 1 ];
	XFontStruct     *font;

	
        (void) strcpy( name, default_name );

        CheckFontName( argc, argv, name );


        font = XLoadQueryFont( display, name ); 


        if ( font == (XFontStruct *) NULL )
                {
                font = XLoadQueryFont( display, default_name ); 

                if ( font == (XFontStruct *) NULL )
                        {
                        font = XLoadQueryFont( display, "fixed" ); 
                        }       
                }       

        
        if ( font != (XFontStruct *) NULL )
                {
                XSetFont( display, gc, font->fid );
                }
        else
                {
                QuitX( display, "Error: could not load font ", name );
                }
        

        return( font );

}       



CheckFontName( argc, argv, fontname )

int     argc;
char    *argv[];
char    fontname[];


{       
        int     counter;

        counter  = 1;
        while( counter < argc )
                {
                if ( ( strncmp( argv[ counter ], "-font", 5 ) == 0 ) ||
                   ( strncmp( argv[ counter ], "-fn", 3 ) == 0 ) ) 
                        {
                        counter++;
                        if ( counter < argc )
                                {
                                (void) strcpy( fontname, argv[ counter ] );
                                }
                        else
                                {
                                (void) fprintf( stderr, 
                                  "Error: usage is -font FontName\n" );
                                }
                        }

                counter++;
                }


}       



GC MakeGC( display, drawable, fore, back )

Display         *display;
Drawable        drawable;
unsigned long   fore, back;


{       
        GC              gc;
        XGCValues       gcvalues;


        gcvalues.foreground = fore;
        gcvalues.background = back;

        gc = XCreateGC( display, drawable,
                        ( GCForeground | GCBackground ),
                        &gcvalues );

        if ( gc == 0 )
                {
                QuitX( display, 
                        "Error in creating a Graphics context", "" );
                }

        return( gc );

}       



SetGC( display, gc, fore, back )

Display         *display;
GC              gc;
unsigned long   fore, back;


{       

        XSetForeground( display, gc, fore );

        XSetBackground( display, gc, back );

}       



Window TopWindow( display, x, y, width, height, 
        bitmap_data, bitmap_width, bitmap_height, icon, gc )

Display         *display;
int             x, y, width, height;
char            *bitmap_data;
int             bitmap_width, bitmap_height;
Pixmap          *icon;
GC              *gc;


{       
        int             screen;
        Window          rootwindow, window;
        unsigned long   black, white;
        Cursor          cursor;


        screen     = DefaultScreen( display );

        rootwindow = RootWindow( display, screen );

        black      = BlackPixel( display, screen );
        white      = WhitePixel( display, screen );

#define BORDER_WIDTH    2

        window  = CreateWindow( display, 
                        rootwindow,
                        x, y, width, height,
                        BORDER_WIDTH,
                        black, white,
                        EVENT_MASK );

        *gc = MakeGC( display, window, black, white );


        *icon = XCreateBitmapFromData( display, window,
                        bitmap_data,
                        bitmap_width, bitmap_height );


        cursor  = MakeCursor( display, window, DEFAULT_CURSOR );

        XFreeCursor( display, cursor );

        XFlush( display );

        return( window );

}       

char    AppName[ BUFSIZE + 1 ];



Window CreateWindow( display,parent,x,y,width,height,border,fore,back, events )

Display         *display;
Window          parent;
int             x, y, width, height, border;
unsigned long   fore, back;
long            events;


{       
        Window                  window;
        XSetWindowAttributes    attributes;
        unsigned long           attribute_mask;
        Visual                  *visual = CopyFromParent;


        attributes.background_pixel = back;
        attributes.border_pixel     = fore;
        attributes.event_mask       = events;
        attribute_mask              = CWBackPixel | CWBorderPixel | CWEventMask;

        window = XCreateWindow( display,
                        parent,
                        x, y, width, height,    
                        border,
                        CopyFromParent,         
                        InputOutput,            
                        visual,
                        attribute_mask,
                        &attributes );


        if ( window == (Window) None )
                {
                QuitX( display, "",
                        "Error: Could not open window." );
                }

        return( window );
        
}       


CheckGeometry( argc, argv, screen_width, screen_height, x, y, width, height )

int     argc;
char    *argv[];
int     screen_width, screen_height;
int     *x, *y, *width, *height;

{       
        int     status;
        int     x1, y1, width1, height1;
        int     counter;

        counter  = 1;
        while( counter < argc )
                {
                if ( strncmp( argv[ counter ], "-geom", 5 ) == 0 )
                        {
                        counter++;
                        if ( counter < argc )
                                {
                                status = 
                                        XParseGeometry( argv[ counter ],
                                                &x1, &y1, &width1, &height1 );

                                if ( status & XValue )
                                        *x = x1;
                                if ( status & YValue )
                                        *y = y1;
                                if ( status & WidthValue )
                                        *width = width1;
                                if ( status & HeightValue )
                                        *height = height1;
                                if ( status & XNegative )
                                        *x = screen_width - *width + *x;
                                if ( status & YNegative )
                                        *y = screen_height - *height + *y;
                                }
                        }
        
                counter++;
                }

}       


MapWindow( display, window )

Display *display;
Window  window;

{       

        XMapRaised( display, window );

        XMapSubwindows( display, window );

        XFlush( display );

}       



NameWindow( display, window, name, class_name, class_type )

Display *display;
Window  window;
char    *name;
char    *class_name;
char    *class_type;


{       
        XClassHint              *classhints;



        classhints = XAllocClassHint();

        if ( classhints != (XClassHint *) NULL )
                {
                classhints->res_name  = class_name;
                classhints->res_class = class_type; 


                XSetClassHint( display, window, classhints );
        

                XFree( classhints );
                }


        /*
         * Window and icon name--we're using
         * The pre-R6 routines here, for simplicity.
         */
        XStoreName( display, window, name );

        XSetIconName( display, window, name );


        
        (void) strcpy( AppName, name );


}       /* -- function NameWindow */




SetWMHints( display, window, icon )

Display *display;
Window  window;
Pixmap  icon;

/*
 *      SetWMHints() sets hints for the window
 *      manager, contained in an XWMHints structure.
 */

{       
        XWMHints        *wmhints;


        wmhints    = XAllocWMHints();


        if ( wmhints != (XWMHints *) NULL )
                {
                wmhints->initial_state = NormalState;
                wmhints->input         = True;

                if ( icon != (Pixmap) None )
                        {
                        wmhints->icon_pixmap = icon;
                        wmhints->icon_mask   = icon;
                        wmhints->flags       =  StateHint      | 
                                                InputHint      | 
                                                IconPixmapHint |
                                                IconMaskHint;
                        }
                else
                        {
                        wmhints->flags       = StateHint | InputHint;
                        }


                XSetWMHints( display, window, wmhints );


                XFree( wmhints );
                }


}       




SetNormalHints( display, window, x, y, width, height )

Display         *display;
Window          window;
int             x, y, width, height;


{       
        XSizeHints              *sizehints;


        sizehints  = XAllocSizeHints();



        if ( sizehints != (XSizeHints *) NULL )
                {
                sizehints->x           = x;        
                sizehints->y           = y;        
                sizehints->width       = width;    
                sizehints->height      = height;   

                sizehints->min_width   = 100;
                sizehints->min_height  = 50;


                sizehints->flags       = USPosition | USSize | PMinSize;

                sizehints->base_width  = width;    
                sizehints->base_height = height;   

                sizehints->flags       |= PBaseSize; 

                XSetWMNormalHints( display, window, sizehints );

                XFree( sizehints );

                }


}       




GetAppName( name )

char    *name;


{       

        (void) strcpy( name, AppName );

}       
