/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <pwd.h>

#include "db.h"
#include "errs.h"

Usage ()
{
  fprintf (stderr, "Usage: btabread host relationname dataset wait [[attributename datatype] ...]\n");
  exit (BAD_ARGUMENTS);
}

char WriteCol[MAXATTR];		/* Write the column number stored here. */
int NumColsToWrite = 0;

static int WeAreRoot = FALSE;

char *Row[MAXATTR];

char *UserName ();

struct scheme sch;

main (argc, argv)
     int argc;
     char *argv[];
{
  int i;
  int time;
  int toolong;

  if (argc < 5)
    Usage ();

  if (getuid () == 0)
    WeAreRoot = TRUE;

  if (strcmp (argv[1], "-") != 0 && *argv[1] != '\0')
    {				/* argv[1] is a machine name; send request to that machine. */
      char *host;
      int fd, cnt;
      char buf[1024];

      host = argv[1];
      argv[1] = "-";
      if ((fd = DoRemote ("read", host, argc, argv)) == -1)
	exit (REMOTE_OPEN_FAILED);
      while ((cnt = read (fd, buf, sizeof buf)) > 0)
	if (write (1, buf, cnt) == -1)
	  exit (WRITE_ERROR);
      if (cnt == -1)
	exit (READ_ERROR);

      if (close (fd) == -1)
	exit (READ_ERROR);
      exit (GOOD);
    }				/* send request remote */

/* do it locally. */

  (void) UserName ();		/* To set username properly */

/* CHANGE raphael 6/92 used to be unconditional, just first branch */

  if (*argv[2] != '.' && *argv[2] != '/')
    {				/* ordinary; work as root */
      (void) setuid (SOMEONEOTHERTHANROOT);	/* Root cannot read over nfs */
      (void) setgid (getegid ());
    }
  else
    {				/* become the actual user; this is a personal relation */
      if (*argv[1] == '-')
	{
	  /* argv[0] is of the form user@host; we should be called from satd */
	  struct passwd *pw;
	  char shortusername[20], *atptr;
	  strcpy (shortusername, argv[0]);
	  atptr = strchr (shortusername, '@');
	  if (atptr)
	    *atptr = 0;		/* just want first part of the name */
	  else
	    exit (BAD_SCHEME_OPEN);
	  pw = getpwnam (shortusername);
	  if (!pw)
	    exit (BAD_SCHEME_OPEN);
	  (void) setgid (pw->pw_gid);
	  (void) setuid (pw->pw_uid);
	}
      else
	{			/* argv[0] is just btabread, and argv[1] is "".
		ReadScheme will make us the actual user */
	}
    }				/* become actual user */

  if (!ReadScheme (argv[2], &sch))
    {
      exit (BAD_SCHEME_OPEN);
    }

  time = 0;
  toolong = atoi (argv[4]);
/* Negative toolong implies ignore lock. */
  while (toolong >= 0 && CheckClobberLock (&sch))
    {
      sleep (2);
      time += 2;
      if (time >= toolong)
	exit (TIMED_OUT);
    }

  CheckPermissions (&sch, argv);

  for (i = 5; i < argc; i += 2)
    {
      int ai;
      for (ai = 0; ai < sch.numattr; ai++)
	if (strcmp (argv[i], sch.at[ai].name) == 0)
	  break;
      if (ai == sch.numattr)
	{
	  /* Unknown attribute */
	  exit (BAD_ATTRIBUTE);
	}
      switch (*argv[i + 1])
	{
	case 'R':
	  if (sch.at[ai].checktype != CT_REAL)
	    exit (BAD_DATATYPE);
	  break;
	case 'B':
	  if (sch.at[ai].checktype != CT_BOOLEAN)
	    exit (BAD_DATATYPE);
	  break;
	case 'I':
	  if (sch.at[ai].checktype != CT_INTEGER)
	    exit (BAD_DATATYPE);
	  break;
	case 'S':
	  if (sch.at[ai].checktype == CT_REAL ||
	      sch.at[ai].checktype == CT_BOOLEAN ||
	      sch.at[ai].checktype == CT_INTEGER)
	    exit (BAD_DATATYPE);
	  break;
	}
      WriteCol[NumColsToWrite++] = ai;
    }

  if (!OpenRow (&sch, WriteCol, NumColsToWrite, argv[3]))	/* argv[3] = dataset */
    {
      exit (BAD_OPEN);
    }

  while (ReadRow (Row))
    {
      int ai;
      for (ai = 0; ai < NumColsToWrite; ai++)
	{
	  if (!CheckItem (Row[WriteCol[ai]], &sch.at[WriteCol[ai]])) {
		fprintf(stderr, "bad data: [%s], attribute %d\n", Row[WriteCol[ai]]);
	    exit (BAD_DATA);
	  }
	  WriteString (Row[WriteCol[ai]]);
	}
    }

  (void) NetWriteEOF (stdout);

  if (ferror (stdout) || fclose (stdout) == -1)
    exit (WRITE_ERROR);

  exit (CloseRow ());
}


WriteString (s)
     char *s;
{
  (void) NetWriteString (stdout, s);
}


CheckPermissions (s, av)
     struct scheme *s;
     char *av[];
{
  char *uname;
  if (strcmp (av[1], "-") == 0)
    {
      /* We are being run by satd, satd will provide the username
	in argv[0], as we have no real way of knowing it was satd
	that started us, make sure that root is our real uid */
      if (!WeAreRoot)
	exit (NO_PERMISSION);
      uname = av[0];
    }
  else
    {
      uname = UserName ();
    }
  if (CheckUserPermission (uname, s->reader, s->useraliasfile))
    return;
  if (CheckUserPermission (uname, s->appender, s->useraliasfile))
    return;
  if (CheckUserPermission (uname, s->modifier, s->useraliasfile))
    return;
  if (CheckUserPermission (uname, s->owner, s->useraliasfile))
    return;
  exit (NO_PERMISSION);
}
