/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <syslog.h>
#include "ps.h"

#include "../sal/sal.h"
#include "errs.h"

char *NetReadString ();

static FILE *fp;
FILE *popen ();
static int Eof;

/* Must be that same as that produced by parse.y in sal */
union data
  {
    Integer In;
    Real Re;
    String Str;
    Boolean Bo;
  };

/* Must be that same as that produced by parse.y in sal */
struct Table
  {
    Integer numattr;
    String *names;
    char *datatypes;
    char *UsedInPut;
    Integer datacnt;
    Integer datasize;
    union data *data;
  };

void ClearTable (t)
     struct Table *t;
{
  union data *d;
  int i;

  d = t->data;

  for (i = 0; i < t->datacnt; i++)
    {
      int j;
      for (j = 0; j < t->numattr; j++)
	if (t->UsedInPut[j] && t->datatypes[j] == 'S')
	  free (d[j].Str);
      d += t->numattr;
    }
  t->datacnt = 0;
}


static int TableSyncError;

Boolean
OpenTable (tbl, host, file, dataset, wait)
     struct Table *tbl;
     char *host;
     char *file;
     char *dataset;
     int wait;
{
  char buf[10 * MAXSIMPSTR];
  Integer i;

  Eof = FALSE;
  TableSyncError = FALSE;

  (void) sprintf (buf, "%s \"%s\" \"%s\" \"%s\" \"%d\"",
		  BTABREAD, host, file, dataset, wait);

  for (i = 0; i < tbl->numattr; i++)
    {
      if (!tbl->UsedInPut[i])
	continue;
      (void) strcat (buf, " ");
      (void) strcat (buf, tbl->names[i]);
      (void) strcat (buf, " ");
      switch (tbl->datatypes[i])
	{
	case DT_BOOLEAN:
	  (void) strcat (buf, "B");
	  break;
	case DT_INTEGER:
	  (void) strcat (buf, "I");
	  break;
	case DT_REAL:
	  (void) strcat (buf, "R");
	  break;
	case DT_STRING:
	  (void) strcat (buf, "S");
	  break;
	default:
	  (void) strcat (buf, "Z");	/* will cause bad data type error */
	  break;
	}
    }
  fp = popen (buf, "r");

  syslog(LOG_DEBUG, "OpenTable: popen(%s,r): COE flags:", buf);
  coeflags();

  ClearNetReadError ();
  if (fp == NULL)
    return (FALSE);
  else
    return (TRUE);
}

Integer
CloseTable ()
{
  UNION_WAIT w;			/* CHANGE raphael 7/92 to remove 'ret' */

  if (!fp || ferror (fp) || (UW_STATUS (w) = pclose (fp)))
    {
      if (WIFSIGNALED (w))	/* CHANGE raphael 1/92 */
	return ((Integer) SIGNAL_CAUGHT);
      else
	return ((Integer) WEXITSTATUS (w));
    }
  if (NetReadError ())
    return (READ_ERROR);
  if (TableSyncError)
    return (DATA_OUT_OF_SYNC);
  return ((Integer) GOOD);
}


static String
GetString ()
{
  char *p;
  if (!(p = NetReadString (fp)))
    {
      Eof = TRUE;
      return ("");
    }
  else
    return ((String) strdup (p));
}

static Real
GetReal ()
{
  char *p;

  if (!(p = NetReadString (fp)))
    {
      Eof = TRUE;
      return (0.0);
    }
  else
    return ((Real) atof (p));
}

static Integer
GetInteger ()
{
  char *p;

  if (!(p = NetReadString (fp)))
    {
      Eof = TRUE;
      return (0);
    }
  else
    return ((Integer) atoi (p));
}

ReadData (t, tbl)
     union data *t;
     struct Table *tbl;
{
  Integer i;
  for (i = 0; i < tbl->numattr; i++)
    {
      if (tbl->UsedInPut[i])
	{
	  switch (tbl->datatypes[i])
	    {
	    case DT_BOOLEAN:
	      t[i].Bo = GetInteger ();
	      break;
	    case DT_INTEGER:
	      t[i].In = GetInteger ();
	      break;
	    case DT_REAL:
	      t[i].Re = GetReal ();
	      break;
	    default:
	      t[i].Str = GetString ();
	      break;
	    }
	}
      if (Eof)
	{
	  int j;
	  for (j = 0; j < tbl->numattr; j++)
	    if (tbl->UsedInPut[j])
	      break;
	  if (i != j)
	    {
	      int k;
	      TableSyncError = TRUE;
	      for (k = 0; k < i; k++)
		if (tbl->UsedInPut[k] && tbl->datatypes[k] == 'S')
		  free (t[k].Str);
	    }
	  break;
	}
    }
  return (!Eof);
}


void
CheckTableAlloc (t, p)
     struct Table *t;
     union data **p;
{

  if (t->datacnt >= t->datasize)
    {
      t->datasize *= 2;
      t->data = (union data *) realloc (t->data,
			  t->datasize * (sizeof (union data) * t->numattr));
      if (!t->data)
	OutOfDynamicStorage ();
      *p = t->data + (t->datacnt * t->numattr);
    }
}
