/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <syslog.h>

extern errno;

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include "global.h"
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#include <sys/file.h>
#endif
#include <sys/ioctl.h>

#include "db.h"
#include "permission.h"
#include "satd.h"
#include "nethead.h"

/* Define DEBUGLEVEL to zero to turn it off entirely */
#define DEBUGLEVEL 2

#define MODELORTRUSTED 1
#define NOTAMODEL 2
#define TRUSTED 3
#define MODEL 4

#define PERMISSIONDENIED "You do not have sufficient privileges to perform the requested operation.\n"

static int HostType;
#define H_MODEL 1
#define H_TRUSTED 2
#define H_COPY 3
#define H_ACCESS 4

static int Ownership;
#undef O_NONE
#define O_NONE 0
#define O_OWNER 1
#define O_MODIFIER 2
#define O_APPENDER 3

#if defined(NFS) && !(defined(__sparc__) && defined(__svr4__)) && !defined(__linux)
#	define GOOTHER (void) setreuid (0, SOMEONEOTHERTHANROOT)
#	define GOROOT (void) setuid (0); (void) setreuid (MyRealUid, 0);
#else
#	define GOOTHER
#	define GOROOT
#endif

char *UserName ();
char *GetLastModifiedTime ();
static char *Uname;
FILE *CloneFile(char *RealName, char *NewName);

long random ();

static int Locked;

static int NumTrustedToPull;

static int MyRealUid;

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 31
#endif
static char MyName[MAXHOSTNAMELEN];

Boolean WriteScheme ();

char *strdup ();
char *SelectRandomHost (); /* forward declaration */

Usage ()
{
  error ("Usage:\
	relmod relationname -add attrname initialvalue\n\
	relmod relationname -delete attrname\n\
	relmod relationname -lock [force]\n\
	relmod relationname -lockall\n\
	relmod relationname -pull [newmodifytime]\n\
	relmod relationname -receiveupdate lastmodtime\n\
	relmod relationname -rename oldattrname newattrname\n\
	relmod relationname -sendupdate lastmodtime hostname [check]\n\
	relmod relationname -test\n\
	relmod relationname -unlock [force]\n\
	relmod relationname -unlockall\n\
	relmod relationname -update\n\
");
  exit (1);
}

struct scheme sch;

extern char *fullhostname ();

main (argc, argv)
     int argc;
     char *argv[];
{

/*  if(!isatty(2)) */
#ifdef LOG_LOCAL5
  openlog ("relmod", LOG_PID, LOG_LOCAL5);
#else
  openlog ("relmod", LOG_PID);
#endif

#if DEBUGLEVEL > 2
  stdio_id(1,1,1);
#endif

  if (argc == 1) {
	int i;
	RelmodInit (argv[0], "Hosts");
	for (i = 0; i < 10; i++) {
		fprintf(stdout, "random host is %s\n",
			/* SelectRandomHost(MODELORTRUSTED)); */
			/* SelectRandomHost(NOTAMODEL)); */
			SelectRandomHost(TRUSTED));
	}
  }
  if (argc < 3 ||
      (strcmp (argv[2], "-add") == 0 && argc != 5) ||
      (strcmp (argv[2], "-delete") == 0 && argc != 4) ||
      (strcmp (argv[2], "-lock") == 0 && argc != 3 && argc != 4) ||
      (strcmp (argv[2], "-lockall") == 0 && argc != 3) ||
      (strcmp (argv[2], "-pull") == 0 && argc != 3 && argc != 4) ||
      (strcmp (argv[2], "-receiveupdate") == 0 && argc != 4) ||
      (strcmp (argv[2], "-rename") == 0 && argc != 5) ||
      (strcmp (argv[2], "-sendupdate") == 0 && argc != 5 && argc != 6) ||
      (strcmp (argv[2], "-test") == 0 && argc != 3) ||
      (strcmp (argv[2], "-unlock") == 0 && argc != 3 && argc != 4) ||
      (strcmp (argv[2], "-unlockall") == 0 && argc != 3) ||
      (strcmp (argv[2], "-update") == 0 && argc != 3)
    )
    Usage ();

  switch (argc)
    {
    case 3:
      syslog (LOG_DEBUG, "Relmod started with %s %s", argv[1], argv[2]);
      break;
    case 4:
      syslog (LOG_DEBUG, "Relmod started with %s %s %s",
	      argv[1], argv[2], argv[3]);
      break;
    case 5:
      syslog (LOG_DEBUG, "Relmod started with %s %s %s %s",
	      argv[1], argv[2], argv[3], argv[4]);
      break;
    case 6:
      syslog (LOG_DEBUG, "Relmod started with %s %s %s %s %s",
	      argv[1], argv[2], argv[3], argv[4], argv[5]);
      break;
    default:
      syslog (LOG_DEBUG, "Relmod started with %d args: %s %s",
	      argc, argv[1], argv[2]);
      break;
    }

  RelmodInit (argv[0], argv[1]);

#if DEBUGLEVEL > 3
  syslog(LOG_DEBUG, "relmod.c: After RelmodInit(): sleeping for 10");
  sleep(10);
  syslog(LOG_DEBUG, "relmod.c: After RelmodInit(): done sleeping for 10");
#endif

  if (strcmp (argv[2], "-add") == 0)
    {
      DoAdd (argv[3], argv[4]);
    }
  else if (strcmp (argv[2], "-delete") == 0)
    {
      DoDelete (argv[3]);
    }
  else if (strcmp (argv[2], "-lock") == 0)
    {
      DoLock (argc == 3 ? 0 : 1);
    }
  else if (strcmp (argv[2], "-lockall") == 0)
    {
      DoLockAll ();
    }
  else if (strcmp (argv[2], "-pull") == 0)
    {
      if (argc == 3)
	DoPull ((char *) 0);
      else
	DoPull (argv[3]);
    }
  else if (strcmp (argv[2], "-receiveupdate") == 0)
    {
      DoReceiveUpdate (argv[3]);
    }
  else if (strcmp (argv[2], "-rename") == 0)
    {
      DoRename (argv[3], argv[4]);
    }
  else if (strcmp (argv[2], "-sendupdate") == 0)
    {
      if (argc == 5)
	DoSendUpdate (argv[3], argv[4], (char *) 0);
      else
	DoSendUpdate (argv[3], argv[4], argv[5]);
    }
  else if (strcmp (argv[2], "-test") == 0)
    {
      DoTest ();
    }
  else if (strcmp (argv[2], "-unlock") == 0)
    {
      DoUnlock (argc == 3 ? 0 : 1);
    }
  else if (strcmp (argv[2], "-unlockall") == 0)
    {
      DoUnlockAll ();
    }
  else if (strcmp (argv[2], "-update") == 0)
    {
      DoUpdate ();
    }
  else
    {
      error ("Unknown command: %s.\n", argv[2]);
      Usage ();
    }
  exit (0);
}


SatError (h, e)
     char *h;
     char e;
{
  switch (e)
    {
    case BAD_SCHEME_READ:
      error ("Host %s could not read the scheme for %s.\n", h, sch.name);
      break;
    case NO_PERMISSION:
      error ("You do not have permission on host %s for %s.\n", h, sch.name);
      break;
    case THERE_IS_A_NEWER_VERSION:
      error ("Host %s has a newer version of the %s relation.\n", h, sch.name);
      break;
    case LOCK_ERROR:
      error ("Host %s could not make the lock for %s.\n", h, sch.name);
      break;
    case UNLOCK_ERROR:
      error ("Host %s could not remove the lock for %s.\n", h, sch.name);
      break;
    case LOCK_ALREADY_EXISTS:
      error ("Host %s already has a lock for %s.\n", h, sch.name);
      break;
    case UNARCHIVE_FAILED:
      error ("Host %s could not properly unarchive the new %s.\n", h, sch.name);
      break;
    case REPORT_FAILED:
      error ("Host %s could not properly make a report from the new %s.\n", h, sch.name);
      break;
    case CHDIR_FAILED:
      error ("Host %s could not change its working directory to the directory for %s.\n", h, sch.name);
      break;
    default:
      error ("Host %s responds with unknown error code for %s.\n",
	     h, sch.name);
      break;
    }
}

RelmodInit (arg0, relationname)
     char *arg0;
     char *relationname;
{
  char *p;

  setenv ("PATH", PATH);

  MyRealUid = getuid ();
  (void) setgid (getegid ());

  GOOTHER;

  if (!ReadScheme (relationname, &sch))
    {
      sleep (5);		/*
				** Sometimes an access host gets unlucky
				** in the midst of an update.
				*/
      if (!ReadScheme (relationname, &sch))
	{
	  error ("Error reading scheme for %s.\n", relationname);
	  exit (1);
	}
    }

  (void) strncpy (MyName, fullhostname (), sizeof MyName);

  if (sch.distributeflag)
    {
      HostType = GetHostType (MyName);
    }
  else
    {
      HostType = H_MODEL;
    }

  Locked = CheckClobberLock (&sch);

  GOROOT;
  p = strchr (arg0, '@');
  if (p)
    {
      if (MyRealUid != 0)
	{
	  error (PERMISSIONDENIED);
	  exit (1);
	}
      Uname = arg0;
      SetUserName (Uname);
    }
  else
    Uname = UserName ();
  GOOTHER;


  if (CheckUserPermission (Uname, sch.owner, sch.useraliasfile))
    Ownership = O_OWNER;
  else if (CheckUserPermission (Uname, sch.modifier, sch.useraliasfile))
    Ownership = O_MODIFIER;
  else if (CheckUserPermission (Uname, sch.appender, sch.useraliasfile))
    Ownership = O_APPENDER;
  else if (strncmp (Uname, "root@", 5) == 0)
    Ownership = O_NONE;
  else
    {
      error (PERMISSIONDENIED);
      exit (1);
    }

  (void) signal (SIGPIPE, SIG_IGN);
  (void) signal (SIGINT, SIG_IGN);
  (void) signal (SIGQUIT, SIG_IGN);
  (void) signal (SIGHUP, SIG_IGN);

  GOROOT;
}

GetHostType (h)
     char *h;
{
  PSL psl, GetDistList ();
  char *host;

  psl = GetDistList (sch.model_hosts, sch.distaliasfile);
  if (!psl)
    {
      error ("Aborting -- could not find %s alias in %s.\n",
	     sch.model_hosts, sch.distaliasfile);
      exit (1);
    }
  PSL_foreach (psl, host)
    if (strcmp (host, h) == 0)
    return (H_MODEL);
  PSL_endforeach

    psl = GetDistList (sch.trusted_hosts, sch.distaliasfile);
  if (!psl)
    {
      error ("Aborting -- could not find %s alias in %s.\n",
	     sch.trusted_hosts, sch.distaliasfile);
      exit (1);
    }
  PSL_foreach (psl, host)
    if (strcmp (host, h) == 0)
    return (H_TRUSTED);
  PSL_endforeach

    psl = GetDistList (sch.copy_hosts, sch.distaliasfile);
  if (!psl)
    {
      error ("Aborting -- could not find %s alias in %s.\n",
	     sch.copy_hosts, sch.distaliasfile);
      exit (1);
    }
  PSL_foreach (psl, host)
    if (strcmp (host, h) == 0)
    return (H_COPY);
  PSL_endforeach

    psl = GetDistList (sch.access_hosts, sch.distaliasfile);
  if (!psl)
    {
      error ("Aborting -- could not find %s alias in %s.\n",
	     sch.access_hosts, sch.distaliasfile);
      exit (1);
    }
  PSL_foreach (psl, host)
    if (strcmp (host, h) == 0)
    return (H_ACCESS);
  PSL_endforeach

  error ("Aborting -- host %s does not have access to %s.\n", h, sch.name);
  exit (1);
  /*NOTREACHED*/
}


DoAdd (attrname, initialvalue)
     char *attrname, *initialvalue;
{
  char GetCol[MAXATTR];
  char *Row[MAXATTR];
  char NewData[2048];
  char RealData[2048];
  FILE *fp;
  int i;

  (void) umask (0027);
  if (Ownership != O_OWNER)
    {
      error (PERMISSIONDENIED);
      exit (1);
    }
  if (!Locked)
    {
      error ("Error: You forgot to lock the relation.\n");
      exit (1);
    }
  else if (!MyClobberLock (&sch))
    {
      error ("Error: Someone else has the lock.\n");
      exit (1);
    }

  (void) sprintf (NewData, "%s/data.new", sch.path);
  (void) sprintf (RealData, "%s/data", sch.path);

  for (i = 0; i < sch.numattr; i++)
    if (strcmp (attrname, sch.at[i].name) == 0)
      break;

  if (i != sch.numattr)
    {
      error ("Error: Attribute %s already is in the relation.\n", attrname);
      exit (1);
    }

  if (sch.numattr >= MAXATTR)
    {
      error ("Error: Already have maximum number of allowed attributes.\n");
      exit (1);
    }

  sch.at[sch.numattr].name = strdup (attrname);
  sch.at[sch.numattr].desc = "";
  sch.at[sch.numattr].checktype = CT_STRING;

  for (i = 0; i < sch.numattr; i++)
    GetCol[i] = i;

  if ((fp = CloneFile (RealData, NewData)) == NULL)
    {
      fprintf (stderr, "Error opening new data file.\n");
      exit (1);
    }

  if (!OpenRow (&sch, GetCol, (int) sch.numattr, ""))
    {
      fprintf (stderr, "Error initializing data read.\n");
      exit (1);
    }

  while (ReadRow (Row))
    {
      int ai;

      for (ai = 0; ai < sch.numattr; ai++)
	{
	  if (!CheckItem (Row[ai], &sch.at[ai]))
	    {
	      fprintf (stderr, "Error: Bad data was read!\n");
	      exit (1);
	    }
	  (void) NetWriteString (fp, Row[ai]);
	}
      (void) NetWriteString (fp, initialvalue);
    }

  if (CloseRow () != 0)
    {
      fprintf (stderr, "Error reading data\n");
      exit (1);
    }

  (void) NetWriteEOF (fp);

  if (fclose (fp) == EOF)
    {
      fprintf (stderr, "Error writing data\n");
      exit (1);
    }

  if (rename (NewData, RealData) == -1)
    {
      perror ("Aborting -- Bad rename of new data to data");
      exit (1);
    }

  sch.numattr++;

  if (!WriteScheme (&sch))
    {
      error ("Error writing scheme -- WARNING DATA FILE ALREADY MODIFIED\n");
      exit (1);
    }

  if (!TouchLastModified (&sch))
    {
      error ("Could not set last modified time.\n");
      exit (1);
    }
}


DoDelete (attrname)
     char *attrname;
{
  char GetCol[MAXATTR];
  char *Row[MAXATTR];
  char NewData[2048];
  char RealData[2048];
  FILE *fp;
  int i;
  int delcol;

  (void) umask (0027);
  if (Ownership != O_OWNER)
    {
      error (PERMISSIONDENIED);
      exit (1);
    }
  if (!Locked)
    {
      error ("Error: You forgot to lock the relation.\n");
      exit (1);
    }
  else if (!MyClobberLock (&sch))
    {
      error ("Error: Someone else has the lock.\n");
      exit (1);
    }

  (void) sprintf (NewData, "%s/data.new", sch.path);
  (void) sprintf (RealData, "%s/data", sch.path);

  for (delcol = 0; delcol < sch.numattr; delcol++)
    if (strcmp (attrname, sch.at[delcol].name) == 0)
      break;

  if (delcol >= sch.numattr)
    {
      error ("Error: Attribute %s not found.\n", attrname);
      exit (1);
    }

  for (i = 0; i < sch.numattr; i++)
    GetCol[i] = i;

  if ((fp = fopen (NewData, "w")) == NULL)
    {
      fprintf (stderr, "Error opening new data file.\n");
      exit (1);
    }

  if (!OpenRow (&sch, GetCol, (int) sch.numattr, ""))
    {
      fprintf (stderr, "Error initializing data read.\n");
      exit (1);
    }

  while (ReadRow (Row))
    {
      int ai;

      for (ai = 0; ai < sch.numattr; ai++)
	{
	  if (ai == delcol)
	    continue;
	  if (!CheckItem (Row[ai], &sch.at[ai]))
	    {
	      fprintf (stderr, "Error: Bad data was read!\n");
	      exit (1);
	    }
	  (void) NetWriteString (fp, Row[ai]);
	}
    }

  if (CloseRow () != 0)
    {
      fprintf (stderr, "Error reading data\n");
      exit (1);
    }

  (void) NetWriteEOF (fp);

  if (fclose (fp) == EOF)
    {
      fprintf (stderr, "Error writing data\n");
      exit (1);
    }

  if (rename (NewData, RealData) == -1)
    {
      perror ("Aborting -- Bad rename of new data to data");
      exit (1);
    }

  {
    int ai;

    for (ai = delcol; ai < (sch.numattr - 1); ai++)
      sch.at[ai] = sch.at[ai + 1];
  }
  sch.numattr--;

  if (!WriteScheme (&sch))
    {
      error ("Error writing scheme -- WARNING DATA FILE ALREADY MODIFIED\n");
      exit (1);
    }

  if (!TouchLastModified (&sch))
    {
      error ("Could not set last modified time.\n");
      exit (1);
    }
}


DoLock (forceflag)
     int forceflag;
{
  if (Locked)
    {
      if (forceflag)
	{
	  if (Ownership != O_OWNER)
	    {
	      error (PERMISSIONDENIED);
	      exit (1);
	    }
	}
      else
	{
	  error ("Error: lock already exists.\n");
	  exit (1);
	}
    }

  if (Ownership != O_OWNER && Ownership != O_MODIFIER && Ownership != O_APPENDER)
    {
      error (PERMISSIONDENIED);
      exit (1);
    }

  if (forceflag && Ownership != O_OWNER)
    {
      error (PERMISSIONDENIED);
      exit (1);
    }

  if (!MakeClobberLock (&sch))
    {
      error ("Failed to make local lock.\n");
      exit (1);
    }

  if (sch.distributeflag)
    {
      PSL psl, GetDistList ();
      char *host;
      char retval;
      int lockcnt;

      fprintf (stderr, "Getting distributed locks.\n");
      psl = GetDistList (sch.model_hosts, sch.distaliasfile);
      if (!psl)
	{
	  error ("Aborting -- could not find %s alias in %s.\n",
		 sch.model_hosts, sch.distaliasfile);
	  exit (1);
	}

      lockcnt = 1;		/* One so we will be included */
      PSL_foreach (psl, host)
      {
	if (strcmp (MyName, host) == 0)
	  continue;
	if (!LockHost (host, &retval))
	  {
	    error ("Host %s not responding.\n", host);
	    continue;
	  }
	if (retval != GOOD_LOCK)
	  {
	    SatError (host, retval);
	    if (!forceflag)
	      {
		if (!RemoveClobberLock (&sch))
		  {
		    error ("Could not remove local lock after failing to get a remote lock.\n");
		  }
		exit (1);
	      }
	    if (retval == LOCK_ALREADY_EXISTS)
	      lockcnt++;
	  }
	else
	  lockcnt++;
      }
      PSL_endforeach
	if (lockcnt < (PSL_size (psl) + 1) / 2)
	{
	  if (!forceflag)
	    {
	      error ("Failed to get enough locks.\n");
	      DoUnlock (1);
	      exit (1);
	    }
	  error ("Warning: Failed to get enough locks.\n");
	}
    }
}


DoLockAll ()
{
  if (Ownership != O_OWNER)
    {
      error (PERMISSIONDENIED);
      exit (1);
    }

  if (Locked)
    warn ("Warning: Lock already exists.\n");
  else if (!MakeClobberLock (&sch))
    {
      error ("Failed to make local lock.\n");
    }

  if (sch.distributeflag)
    {
      PSL psl, GetDistList ();
      char *host;
      char retval;

      fprintf (stderr, "Getting distributed locks.\n");
      psl = GetDistList (sch.model_hosts, sch.distaliasfile);
      if (!psl)
	{
	  error ("Aborting -- could not find %s alias in %s.\n",
		 sch.model_hosts, sch.distaliasfile);
	  exit (1);
	}
      PSL_foreach (psl, host)
      {
	if (strcmp (MyName, host) == 0)
	  continue;
	if (!LockHost (host, &retval))
	  {
	    error ("Host %s not responding.\n", host);
	    continue;
	  }
	if (retval != GOOD_LOCK)
	  {
	    SatError (host, retval);
	  }
      }
      PSL_endforeach

	psl = GetDistList (sch.trusted_hosts, sch.distaliasfile);
      if (!psl)
	{
	  error ("Aborting -- could not find alias: %s in %s.\n",
		 sch.trusted_hosts, sch.distaliasfile);
	  exit (1);
	}
      PSL_foreach (psl, host)
      {
	if (strcmp (MyName, host) == 0)
	  continue;
	if (!LockHost (host, &retval))
	  {
	    error ("Host %s not responding.\n", host);
	    continue;
	  }
	if (retval != GOOD_LOCK)
	  {
	    SatError (host, retval);
	  }
      }
      PSL_endforeach

	psl = GetDistList (sch.copy_hosts, sch.distaliasfile);
      if (!psl)
	{
	  error ("Aborting -- could not find %s alias in %s.\n",
		 sch.copy_hosts, sch.distaliasfile);
	  exit (1);
	}
      PSL_foreach (psl, host)
      {
	if (strcmp (MyName, host) == 0)
	  continue;
	if (!LockHost (host, &retval))
	  {
	    error ("Host %s not responding.\n", host);
	    continue;
	  }
	if (retval != GOOD_LOCK)
	  {
	    SatError (host, retval);
	  }
      }
      PSL_endforeach
    }
}


/* modtime is (char *)0 if this is an anti-entropy pull,
   otherwise we are in the midst of a global update, and must
   do our part to tell the rest of the world about the update */
DoPull (modtime)
     char *modtime;
{
	int fd;
	int Checking;
	int i;
	int sizefactor;
	char *av[6];
	if (!sch.distributeflag) { /* not distributed; just do local stuff */
		LocalUpdate ();
		exit (0);
	}
#if DEBUGLEVEL > 2
	syslog(LOG_DEBUG, "relmod.c: DoPull: modtime check in 10 seconds");
	sleep(10);
#endif
	if (modtime)
		Checking = FALSE;
	else { /* no particular time given; this is a manual pull. */
		Checking = TRUE;
    GOOTHER;
		modtime = GetLastModifiedTime (&sch);
		GOROOT;;
	} /* manual pull */
#if DEBUGLEVEL > 2
	syslog(LOG_DEBUG, "relmod.c: DoPull: !Checking check in 10 seconds");
	sleep(10);
#endif
	if (!Checking) { /* auto pull */
#if DEBUGLEVEL > 2
		syslog(LOG_DEBUG, "relmod.c: DoPull: MakePullLock check in 10 seconds");
		sleep(10);
#endif
		if (!MakePullLock (sch.name, modtime)) { /* can't make pull lock */
			if (errno != EEXIST) /* but it doesn't exist */
				error ("open in DoPull: %m.\n");
			exit (0);
		}
#if DEBUGLEVEL > 2
		syslog(LOG_DEBUG, "relmod.c: DoPull: PullSome check in 10 seconds");
		sleep(10);
#endif
		PullSome (modtime); /* pass along auto pull */
#if DEBUGLEVEL > 2
		syslog(LOG_DEBUG, "relmod.c: DoPull: Randomize check in 10 seconds");
		sleep(10);
#endif
		Randomize ();
		if (HostType == H_ACCESS || HostType == H_COPY) {
			/* wait a reasonable amount of time */
			if (sch.datafromprog)
				sizefactor = 1;
			else { /* data from file */
				char buf[2048];
				struct stat sb;
				(void) sprintf (buf, "%s/data", sch.path);
				if (stat (buf, &sb) == -1)
					sizefactor = 2;
				else
					sizefactor = (sb.st_size / (25 * 1024)) + 2;
			} /* data from file */
			if (HostType == H_ACCESS) {
				sleep (COPYWAIT + ACCESSWAIT);
				sleep ((unsigned int) (((unsigned int) random ()) %
					(ACCESSWAIT * MAXACCESSERS * sizefactor)) + 2);
			} else if (HostType == H_COPY) {
				sleep (COPYWAIT);
				sleep ((unsigned int) (((unsigned int) random ())
					% (COPYWAIT * sizefactor)) + 2);
			}
		} /* wait a reasonable amount of time */
		/* model and trusted don't wait */
	} /* auto pull */

	if (HostType == H_ACCESS && !Checking) { /* auto pull */
		/* wait for my copy to get updated */
		int slept;
		GOOTHER;
		slept = 0;
		while (strcmp (GetLastModifiedTime (&sch), modtime) < 0) {
			if (slept++ > WAITFORUPDATE / 8) {
				error ("My copy to access of %s never got updated.\n",
				sch.name);
				exit (1);
			}
			sleep (8);
		}
		GOROOT;
	} /* auto pull; wait for my copy to get updated */
#if DEBUGLEVEL > 2
	syslog(LOG_DEBUG, "relmod.c: DoPull: calling LocalUpdate #2 in 10 seconds");
	sleep(10);
#endif
	LocalUpdate ();		/* Make sure we are in sync with the version we
		already have */
	if (HostType == H_ACCESS) exit(0); /* we are done; don't bug anyone */
	for (i = 0; i < 3; i++) { /* contact a host with -sendupdate */
		char *host;
		int tries = 0;
		do { /* find a host other than me */
			host = SelectRandomHost (
				HostType == H_TRUSTED || HostType == H_MODEL ?
				MODEL : MODELORTRUSTED);
			tries++;
		} while (strcmp (host, MyName) == 0 && tries < 10);
		if (tries == 10) { /* there are no such hosts! */
			error ("Seem to be no hosts to pull %s from", sch.name);
			exit(0);
		}
		av[0] = "";		/* filled in by DoRemote */
		av[1] = sch.name;
		av[2] = "";		/* filled as -sendupdate by satd */
		av[3] = modtime;
		av[4] = MyName;
		if (Checking) av[5] = "check";
#if DEBUGLEVEL > 1
		debug ("Requesting %s from %s.\n", sch.name, host);
#endif
		fd = DoRemote ("sendupdate", host, Checking ? 6 : 5, av);
		if (fd != -1) { /* success; exit here */
			(void) close (fd);
			exit (0);
		}
		/* failed; try again up to 3 times */
	} /* contact a host with -sendupdate */
	/* we fall through only if we fail */
	error ("DoPull failed to get %s from anybody.\n", sch.name);
	exit (1);
} /* DoPull */


DoReceiveUpdate (remotelastmodtime)
     char *remotelastmodtime;
{

#if DEBUGLEVEL > 0
  debug ("relmod.c:DoReceiveUpdate: Receiving %s from %s.\n", sch.name, Uname);
#endif

  if (strcmp (remotelastmodtime, GetLastModifiedTime (&sch)) < 0)
    NetReturn (THERE_IS_A_NEWER_VERSION);

  if (Ownership != O_OWNER && Ownership != O_MODIFIER && Ownership != O_APPENDER)
    {
      NetReturn (NO_PERMISSION);
    }
  if (!sch.distributeflag)
    {
      syslog (LOG_INFO, "The %s relation has distribution disabled.", sch.name);
      NetReturn (NO_PERMISSION);
    }


  if (HostType == H_MODEL || HostType == H_TRUSTED || HostType == H_COPY)
    {
      int retval;
      if (!Locked)
	{
	  NetReturn (LOCK_ERROR);
	}

      if (!MyClobberLock (&sch))
	{
	  error ("Error: Someone else has the lock.\n");
	  NetReturn (LOCK_ERROR);
	}
      if (chdir (sch.path) == -1)
	NetReturn (CHDIR_FAILED);

#if DEBUGLEVEL > 2
      syslog(LOG_DEBUG, "relmod.c: DoReceiveUpdate: calling system(UNARCHIVE) in 10 seconds");
      sleep(10);
#endif

      retval = system (UNARCHIVE);

#if DEBUGLEVEL > 2
      syslog(LOG_DEBUG, "relmod.c: DoReceiveUpdate: system(UNARCHIVE) returned %#x. 10 seconds", retval);
      sleep(10);
#endif

      if (retval != 0)
	{
	  syslog (LOG_ERR, "relmod.c:DoReceiveUpdate():system(%s) returned %#x",
		  UNARCHIVE, retval);
	  NetReturn (UNARCHIVE_FAILED);
	}
    }

#if DEBUGLEVEL > 2
  syslog(LOG_DEBUG, "relmod.c: DoReceiveUpdate: calling LocalUpdate in 10 seconds");
  sleep(10);
#endif

  LocalUpdate ();

  NetReturn (GOOD_RECEIVE);
}


DoRename (oldattrname, newattrname)
     char *oldattrname;
     char *newattrname;
{
  int i;

  (void) umask (0027);
  if (Ownership != O_OWNER)
    {
      error (PERMISSIONDENIED);
      exit (1);
    }
  if (!Locked)
    {
      error ("Error: You forgot to lock the relation.\n");
      exit (1);
    }
  else if (!MyClobberLock (&sch))
    {
      error ("Error: Someone else has the lock.\n");
      exit (1);
    }

  for (i = 0; i < sch.numattr; i++)
    if (strcmp (newattrname, sch.at[i].name) == 0)
      break;

  if (i != sch.numattr)
    {
      error ("Error: Attribute %s already is in the relation.\n", newattrname);
      exit (1);
    }

  for (i = 0; i < sch.numattr; i++)
    if (strcmp (oldattrname, sch.at[i].name) == 0)
      break;

  if (i == sch.numattr)
    {
      error ("Error: Cannot find attribute %s.\n", oldattrname);
      exit (1);
    }
  free (sch.at[i].name);
  sch.at[i].name = strdup (newattrname);
  if (!WriteScheme (&sch))
    {
      error ("Error writing scheme -- aborting.\n");
      exit (1);
    }
  if (!TouchLastModified (&sch))
    {
      error ("Could not set last modified time.\n");
      exit (1);
    }
}


DoSendUpdate (modtime, host, check)
     char *modtime;
     char *host;
     char *check;
{
  char retval;
  int slept;
  int remotehosttype;

  {
    char pidbuf[128];
    (void) sprintf (pidbuf, "%d", getpid ());
    (void) setenv ("SATID", pidbuf);
    SetUserName ("");
  }

#if DEBUGLEVEL > 1
  debug ("Sending %s to %s.\n", sch.name, host);
#endif

  if (!sch.distributeflag)
    {
      error ("Relation %s is not distributed.\n", sch.name);
      exit (1);
    }
  if (HostType != H_MODEL && HostType != H_TRUSTED)
    {
      error ("You can only send updates from a model or trusted host.\n");
      exit (1);
    }

  if (check)
    {
      if (strcmp (GetLastModifiedTime (&sch), modtime) == 0)
	exit (0);
    }

  slept = 0;
  while (strcmp (GetLastModifiedTime (&sch), modtime) < 0)
    {
      sleep (2);
      if (slept++ > WAITTOSEND / 2)
	{
	  error ("Never got updated, could not send update of %s.\n",
		 sch.name);
	  exit (1);
	}
    }

  slept = 0;
  while (!MakeClobberLock (&sch))
    {
      sleep (2);
      if (slept++ > WAITTOSEND / 2)
	{
	  error ("Could not get clobber lock sending update of %s.\n",
		 sch.name);
	  exit (1);
	}
    }

  if (strcmp (host, MyName) == 0)
    {
      error ("Attempted to send an update to myself.\n");
      exit (1);
    }

  remotehosttype = GetHostType (host);

  if (remotehosttype != H_ACCESS)
    {
      if (!LockHost (host, &retval) || retval != GOOD_LOCK)
	{
	  warn ("Warning: Could not lock remote relation during a send update; host %s, relation %s.\n", host, sch.name);
	  if (!RemoveClobberLock (&sch))
	    {
	      error ("Could not remove local lock after remote lock failed.\n");
	    }
	  exit (1);
	}
    }

  (void) UpdateHost (host);

  if (remotehosttype != H_ACCESS)
    {
      if (!UnlockHost (host, &retval) || retval != GOOD_LOCK)
	{
	  error ("Error: Could not unlock remote relation during a send update; host %s, relation %s.\n", host, sch.name);
	}
    }

  if (!RemoveClobberLock (&sch))
    {
      error ("Failed to remove local lock for %s.\n", sch.name);
      exit (1);
    }
#if DEBUGLEVEL > 0
  debug ("Finished sending %s to %s.\n", sch.name, host);
#endif
}



DoTest ()
{

#if DEBUGLEVEL > 2
  syslog(LOG_DEBUG, "relmod.c: DoTest: calling LocalUpdate in 10 seconds");
  sleep(10);
#endif

  LocalUpdate ();
}


DoUnlock (forceflag)
     int forceflag;
{
  if (!Locked)
    {
      if (forceflag)
	{
	  if (Ownership != O_OWNER)
	    {
	      error (PERMISSIONDENIED);
	      exit (1);
	    }
	}
      else
	{
	  error ("Error: lock does not exist.\n");
	  exit (1);
	}
    }
  else if (!MyClobberLock (&sch) && !forceflag)
    {
      error ("Error: Someone else has the lock.\n");
      exit (1);
    }

  if (Ownership != O_OWNER && Ownership != O_MODIFIER && Ownership != O_APPENDER)
    {
      error (PERMISSIONDENIED);
      exit (1);
    }

  if (forceflag && Ownership != O_OWNER)
    {
      error (PERMISSIONDENIED);
      exit (1);
    }

  if (sch.distributeflag)
    {
      PSL psl, GetDistList ();
      char *host;
      char retval;

      fprintf (stderr, "Releasing distributed locks.\n");
      psl = GetDistList (sch.model_hosts, sch.distaliasfile);
      if (!psl)
	{
	  error ("Aborting -- could not find %s alias in %s.\n",
		 sch.model_hosts, sch.distaliasfile);
	  exit (1);
	}
      PSL_foreach (psl, host)
      {
	if (strcmp (MyName, host) == 0)
	  continue;
	if (!UnlockHost (host, &retval))
	  {
	    error ("Host %s not responding.\n", host);
	    continue;
	  }
	if (retval != GOOD_LOCK)
	  {
	    SatError (host, retval);
	    if (!forceflag)
	      exit (1);
	    continue;
	  }
      }
      PSL_endforeach
    }

  if (!RemoveClobberLock (&sch))
    {
      error ("Failed to remove local lock.\n");
      exit (1);
    }
}


DoUnlockAll ()
{
  if (Ownership != O_OWNER)
    {
      error (PERMISSIONDENIED);
      exit (1);
    }

  if (sch.distributeflag)
    {
      PSL psl, GetDistList ();
      char *host;
      char retval;

      fprintf (stderr, "Removing distributed locks.\n");
      psl = GetDistList (sch.model_hosts, sch.distaliasfile);
      if (!psl)
	{
	  error ("Aborting -- could not find %s alias in %s.\n",
		 sch.model_hosts, sch.distaliasfile);
	  exit (1);
	}
      PSL_foreach (psl, host)
      {
	if (strcmp (MyName, host) == 0)
	  continue;
	if (!UnlockHost (host, &retval))
	  {
	    error ("Host %s not responding.\n", host);
	    continue;
	  }
	if (retval != GOOD_LOCK)
	  {
	    SatError (host, retval);
	  }
      }
      PSL_endforeach

	psl = GetDistList (sch.trusted_hosts, sch.distaliasfile);
      if (!psl)
	{
	  error ("Aborting -- could not find %s alias in %s.\n",
		 sch.trusted_hosts, sch.distaliasfile);
	  exit (1);
	}
      PSL_foreach (psl, host)
      {
	if (strcmp (MyName, host) == 0)
	  continue;
	if (!UnlockHost (host, &retval))
	  {
	    error ("Host %s not responding.\n", host);
	    continue;
	  }
	if (retval != GOOD_LOCK)
	  {
	    SatError (host, retval);
	  }
      }
      PSL_endforeach

	psl = GetDistList (sch.copy_hosts, sch.distaliasfile);
      if (!psl)
	{
	  error ("Aborting -- could not find alias: %s in %s.\n",
		 sch.copy_hosts, sch.distaliasfile);
	  exit (1);
	}
      PSL_foreach (psl, host)
      {
	if (strcmp (MyName, host) == 0)
	  continue;
	if (!UnlockHost (host, &retval))
	  {
	    error ("Host %s not responding.\n", host);
	    continue;
	  }
	if (retval != GOOD_LOCK)
	  {
	    SatError (host, retval);
	  }
      }
      PSL_endforeach
    }

  if (!Locked)
    warn ("Warning: Local Lock does not exist.\n");
  else if (!RemoveClobberLock (&sch))
    {
      error ("Failed to remove local lock.\n");
    }
}


DoUpdate ()
{
  if (Ownership != O_OWNER && Ownership != O_MODIFIER && Ownership != O_APPENDER)
    {
      error (PERMISSIONDENIED);
      exit (1);
    }

  if (!Locked)
    {
      error ("Error: relation lock has not been set.\n");
      exit (1);
    }
  else if (!MyClobberLock (&sch))
    {
      error ("Error: Someone else has the lock.\n");
      exit (1);
    }

  if (!TouchLastModified (&sch))
    {
      error ("Could not set last modified time.\n");
      exit (1);
    }

  if (sch.distributeflag && HostType != H_MODEL)
    {
      error ("You can do updates from a model host only.\n");
      exit (1);
    }
  else if (sch.distributeflag)
    {
      PSL psl, GetDistList ();
      char *host;

      fprintf (stderr, "Distributing updates.\n");
      psl = GetDistList (sch.model_hosts, sch.distaliasfile);
      if (!psl)
	{
	  error ("Aborting -- could not find %s alias in %s.\n",
		 sch.model_hosts, sch.distaliasfile);
	  exit (1);
	}
      PSL_foreach (psl, host)
      {
	if (strcmp (MyName, host) == 0)
	  continue;
	(void) UpdateHost (host);
      }
      PSL_endforeach
    }

#if DEBUGLEVEL > 2
  syslog(LOG_DEBUG, "relmod.c: DoUpdate: calling LocalUpdate in 10 seconds");
  sleep(10);
#endif

  LocalUpdate ();

  if (sch.distributeflag)
    PullSome (GetLastModifiedTime (&sch));
}


/*VARARGS*/
error (fmt, a1, a2)
     char *fmt;
     char *a1;
     char *a2;
{
  char buf[2048];

  syslog (LOG_ERR, fmt, a1, a2);
  (void) sprintf (buf, fmt, a1, a2);
  if (isatty (2))
    {
      fputs (buf, stderr);
    }
}


/*VARARGS*/
warn (fmt, a1, a2)
     char *fmt;
     char *a1;
     char *a2;
{
  char buf[2048];
  (void) sprintf (buf, fmt, a1, a2);
  if (isatty (2))
    {
      fputs (buf, stderr);
    }
  else
    {
      syslog (LOG_NOTICE, buf);
    }
}


/*VARARGS*/
debug (fmt, a1, a2)
     char *fmt;
     char *a1;
     char *a2;
{
  char buf[2048];

  (void) sprintf (buf, fmt, a1, a2);
  if (isatty (2))
    {
      fputs (buf, stderr);
    }
  else
    {
      syslog (LOG_DEBUG, buf);
    }
}


NetReturn (status)
     char status;
{

  (void) write (1, &status, 1);
#ifndef lint
  if (status != GOOD_RECEIVE)
    {
      error ("Returning bad receive update status: %o.\n", (int) status);
      exit (1);
    }
#endif
  (void) close (1);
}



UpdateHost (host)
     char *host;
{
  int fd;
  char retval;
  int pid;
  UNION_WAIT status;
  char *av[4];
  int remotehosttype;

  /* will cause an exit if the host has no status */
  remotehosttype = GetHostType (host);

  syslog (LOG_DEBUG, "relmod.c:UpdateHost():updating %s\n", host);
  av[0] = "";			/* filled in by DoRemote */
  av[1] = sch.name;
  av[2] = "";			/* filled as -receiveupdate by satd */
  av[3] = GetLastModifiedTime (&sch);

  if ((fd = DoRemote ("receiveupdate", host, 4, av)) == -1)
    {
      error ("Host %s not responding.\n", host);
      return (FALSE);
    }

  if (remotehosttype == H_MODEL || remotehosttype == H_TRUSTED
      || remotehosttype == H_COPY)
    {
      if (chdir (sch.path) == -1)
	{
	  error ("Could not chdir(%s) for %s relation.\n",
		 sch.path, sch.name);
	  return (FALSE);
	}

#if DEBUGLEVEL > 3
      syslog(LOG_DEBUG, "relmod.c: before fork() #1: sleeping for 10");
      sleep(10);
#endif

      if ((pid = fork ()) == 0)
	{
	  int retval;

	  closelog();
#ifdef LOG_LOCAL5
	  openlog ("relmod", LOG_PID, LOG_LOCAL5);
#else
	  openlog ("relmod", LOG_PID);
#endif

#ifdef HAVE_SETSID
	  (void) setsid ();
#else
	  (void) setpgrp (0, getpid ());
	  {
	    int null;

	    null = open ("/dev/tty");
	    (void) ioctl (null, TIOCNOTTY, (char *) 0);
	    (void) close (null);
	  }
#endif
	  (void) close (0);
	  (void) open ("/dev/null", O_RDONLY);
	  (void) close (1);
	  (void) dup (fd);

	  (void) close (2);
#if DEBUGLEVEL > 0
	  syslog (LOG_DEBUG, "UpdateHost(): attaching /tmp/arch to stderr");
	  (void) open ("/tmp/arch", O_WRONLY | O_CREAT | O_APPEND, 0644);
#else
	  syslog (LOG_DEBUG, "UpdateHost(): attaching /dev/null to stderr");
	  (void) open ("/dev/null", O_WRONLY);
#endif
#if DEBUGLEVEL > 2
	  syslog (LOG_DEBUG, "UpdateHost(): coeflags and stdio_id before ARCHIVE:");
	  coeflags();
	  stdio_id(1,1,1);
#endif
#ifdef NODUPSOCKET
	  syslog (LOG_DEBUG, "UpdateHost(): calling pipeopen(ARCHIVE)");
	  retval = pipeopen (WRITE_TO_PIPE, fd, ARCHIVE, 0);
#else
	  syslog (LOG_DEBUG, "UpdateHost(): calling system(ARCHIVE)");
	  retval = system (ARCHIVE);
#endif
	  if (retval != 0)
	    {
	      syslog (LOG_ERR, "UpdateHost():%s failed: %#x", ARCHIVE, retval);
	      exit (1);
	    }
	  syslog (LOG_DEBUG, "UpdateHost():%s returned: %#x", ARCHIVE, retval);

	  if (shutdown (fd, 1) == -1)
	    {
	      error ("shutdown failed: %m.\n");
	      exit (1);
	    }
	  exit (0);
	}

      if (pid == -1)
	{
	  error ("Could not fork in update of %s.\n", host);
	  return (FALSE);
	}

      while (wait (&status) != pid) /* NULL */ ;

      if (UW_STATUS (status) != 0)
	{
	  warn ("The archiver returned an error in update of %s.\n", host);
	  return (FALSE);
	}
    }

  if (read (fd, &retval, 1) != 1)
    {
      error ("Host %s would not return a return value for its update.\n", host);
      return (FALSE);
    }
  (void) close (fd);
  if (retval != GOOD_LOCK)
    {
      SatError (host, retval);
      return (FALSE);
    }
  return (TRUE);
}



LockHost (host, retval)
     char *host;
     char *retval;
{
  int fd;
  char *av[3];

  av[1] = sch.name;
  av[2] = GetLastModifiedTime (&sch);
  if ((fd = DoRemote ("lock", host, 3, av)) == -1 ||
      read (fd, retval, 1) != 1)
    {
      (void) close (fd);
      return (FALSE);
    }
  (void) close (fd);
  return (TRUE);
}



UnlockHost (host, retval)
     char *host, *retval;
{
  int fd;
  char *av[2];

  av[1] = sch.name;
  if ((fd = DoRemote ("unlock", host, 2, av)) == -1 ||
      read (fd, retval, 1) != 1)
    {
      (void) close (fd);
      return (FALSE);
    }
  (void) close (fd);
  return (TRUE);
}


LocalUpdate ()
{
  struct stat srcsb;

  /* The scheme may have changed, so reread it */

  GOOTHER;

#if DEBUGLEVEL > 2
  syslog(LOG_DEBUG, "relmod.c: LocalUpdate: ReadScheme in 10");
  sleep(10);
#endif

  if (!ReadScheme (sch.name, &sch))
    {
      error ("Error reading scheme for %s.\n", sch.name);
      goto out;
    }

  if (chdir (sch.path) == -1)
    {
      error ("Could not chdir(%s) for %s relation.\n", sch.path, sch.name);
      goto out;
    }


  if (HostType == H_MODEL || HostType == H_TRUSTED || HostType == H_COPY)
    {
#if DEBUGLEVEL > 2
      syslog(LOG_DEBUG, "relmod.c: LocalUpdate: ReadScheme in 10");
      sleep(10);
#endif

      GOROOT;
      if (stat ("defhead.mk", &srcsb) != -1)
	{
	  char buf[1024];

	  (void) sprintf (buf, "SALINCLUDE=%s", SALINCLUDE);
	  if (DoMake ("defhead.mk", buf, sch.name) != 0)
	    {
	      warn ("Make of 'defhead' for relation %s failed.\n",
		    sch.name);
	    }
	}

      if (sch.reportflag && stat ("report.mk", &srcsb) != -1)
	{
	  if (DoMake ("report.mk", "", sch.name) != 0)
	    {
	      warn ("Make of 'report' for relation %s failed.\n",
		    sch.name);
	      (void) unlink ("report");
	    }
	}

      if (sch.datafromprog && stat ("prog.mk", &srcsb) != -1)
	{
	  if (DoMake ("prog.mk", "", sch.name) != 0)
	    {
	      warn ("Make of 'prog' for relation %s failed.\n",
		    sch.name);
	      (void) unlink ("prog");
	    }
	}

      if (sch.checkflag && stat ("check.mk", &srcsb) != -1)
	{
	  if (DoMake ("check.mk", "", sch.name) != 0)
	    {
	      warn ("Make of 'check' for relation %s failed.\n",
		    sch.name);
	      (void) unlink ("check");
	    }
	}
    GOOTHER;
    }

  if (sch.reportflag)
    {
      int oldumask;
      char *p;
      char buf[2048];

      oldumask = umask (022);	/* Liberal umask when making reports. */

      p = strrchr (sch.name, '/');
      if (!p)
	p = sch.name;
      else			/* move beyond the '/' */
	p++;

      /* remove /tmp/report.RELATION.log in case a different owner */
      (void) sprintf (buf, "/tmp/report.%s.log", p); /* NEW raphael 11/95 */
      (void) unlink(buf); /* NEW raphael 11/95 */

      (void) sprintf (buf, "./report >/tmp/report.%s.log 2>&1", p);

#if DEBUGLEVEL > 2
      syslog(LOG_DEBUG, "relmod.c: LocalUpdate: calling system(%s) in 10", buf);
      sleep(10);
#endif

      if (system (buf) != 0)
	{
	  warn ("Report for relation %s failed. pid=%d\n", sch.name, getpid());
	}

#if DEBUGLEVEL > 2
      syslog(LOG_DEBUG, "relmod.c: LocalUpdate: back from system(%s). waiting 10", buf);
      sleep(10);
#endif

      (void) umask (oldumask);
    }

out:/* NULL */ ;
  GOROOT;

#if DEBUGLEVEL > 2
  syslog(LOG_DEBUG, "relmod.c: leaving LocalUpdate in 10");
  sleep(10);
#endif
}


char *
SelectRandomHost (flag)
     int flag;
{
  PSL psl, GetDistList ();
  char *host;
  static int inited = FALSE;
  static char *hostlist[MAXHOSTS];
  static int hlindex = 0;
  static int modelortrustedindex = 0;
  static int modelindex = 0;
  time_t time ();

  if (!inited)
    {
      GOOTHER;
      inited = TRUE;

      psl = GetDistList (sch.model_hosts, sch.distaliasfile);
      if (!psl)
	{
	  error ("Aborting -- could not find alias: %s in %s.\n",
		 sch.model_hosts, sch.distaliasfile);
	  exit (1);
	}
      if (PSL_size (psl) + hlindex >= MAXHOSTS)
	{
	  error ("Aborting -- out of room for hostnames.\n");
	  exit (1);
	}
      PSL_foreach (psl, host)
	hostlist[hlindex++] = host;
      PSL_endforeach

	modelindex = hlindex;

      psl = GetDistList (sch.trusted_hosts, sch.distaliasfile);
      if (!psl)
	{
	  error ("Aborting -- could not find alias: %s in %s.\n",
		 sch.trusted_hosts, sch.distaliasfile);
	  exit (1);
	}
      if (PSL_size (psl) + hlindex >= MAXHOSTS)
	{
	  error ("Aborting -- out of room for hostnames.\n");
	  exit (1);
	}
      PSL_foreach (psl, host)
	hostlist[hlindex++] = host;
      PSL_endforeach

	modelortrustedindex = hlindex;

      NumTrustedToPull = ((modelortrustedindex - modelindex) / modelindex) + 1;

      psl = GetDistList (sch.copy_hosts, sch.distaliasfile);
      if (!psl)
	{
	  error ("Aborting -- could not find alias: %s in %s.\n",
		 sch.copy_hosts, sch.distaliasfile);
	  exit (1);
	}
      if (PSL_size (psl) + hlindex >= MAXHOSTS)
	{
	  error ("Aborting -- out of room for hostnames.\n");
	  exit (1);
	}
      PSL_foreach (psl, host)
	hostlist[hlindex++] = host;
      PSL_endforeach

	psl = GetDistList (sch.access_hosts, sch.distaliasfile);
      if (!psl)
	{
	  error ("Aborting -- could not find alias: %s in %s.\n",
		 sch.access_hosts, sch.distaliasfile);
	  exit (1);
	}
      if (PSL_size (psl) + hlindex >= MAXHOSTS)
	{
	  error ("Aborting -- out of room for hostnames.\n");
	  exit (1);
	}
      PSL_foreach (psl, host)
	hostlist[hlindex++] = host;
      PSL_endforeach

	Randomize ();
    GOROOT;
    }

  if (flag == NOTAMODEL)
    {
      int tindex;

      tindex = hlindex - modelindex;
      host = tindex ? hostlist[(random () % tindex) + modelindex] : "";
    }
  else if (flag == TRUSTED)
    {
      int tindex;

      tindex = modelortrustedindex - modelindex;
      host = tindex ? hostlist[(random () % tindex) + modelindex] : "";
    }
  else if (flag == MODEL)
    {
      host = hostlist[(random () % modelindex)];
    }
  else				/* MODELORTRUSTED */
    host = hostlist[random () % modelortrustedindex];
  return (host);
}

Randomize ()
{
  unsigned int randvalue;
  char *p;

  randvalue = (unsigned int) time ((time_t *) 0);
  for (p = MyName; *p; p++)
    randvalue += *p;
  srandom (randvalue);
}

/* tell someone that an auto-pull session is in progress */
PullSome (modtime)
     char *modtime;
{
  int i;

  (void) SelectRandomHost (TRUSTED);	/* So NumTrustedToPull will be inited */

  if (HostType == H_MODEL)
    {
      /* We are just increasing the probablity that a trusted host
       will be contacted. */
      for (i = 0; i < NumTrustedToPull; i++)
	{
	  char *host;
	  int fd;
	  char *av[5];

	  host = SelectRandomHost (TRUSTED);

	  av[0] = "";		/* filled in by DoRemote */
	  av[1] = sch.name;
	  av[2] = "";		/* filled as -pull by satd */
	  av[3] = modtime;

	  fd = DoRemote ("pull", host, 4, av);
#if DEBUGLEVEL > 1
	{ char message[100];
	sprintf(message,"relmod: DoPull: sending auto pull to %s", host);
	syslog(LOG_DEBUG, message);
	}
#endif
	  (void) close (fd);
	}
    } /* I am Model */

  for (i = 0; i < HOSTSTOCONTACT; i++)
    {
      char *host;
      int fd;
      char *av[5];

      do
	{
	  host = SelectRandomHost (NOTAMODEL);
	}
      while (strcmp (host, MyName) == 0);

      av[0] = "";		/* filled in by DoRemote */
      av[1] = sch.name;
      av[2] = "";		/* filled as -pull by satd */
      av[3] = modtime;

      fd = DoRemote ("pull", host, 4, av);
      (void) close (fd);
      if (HostType != H_MODEL)
	sleep (WAITBETWEENPULLS);
    }
}


DoMake (makefile, defs, name)
     char *makefile, *defs, *name;
{
  char buf[2048];
  (void) sprintf (buf, "%s -v <%s | %s -s -f - %s >/tmp/%s.%s.log 2>&1",
		  SPP, makefile, MAKE, defs, name, makefile);
  return (system (buf));
}
