/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "sal.h"
#include "y.tab.h"

int yylex ()
{
  int c;
  static int EditStringFlag = FALSE;
  static int InExpr = 0;

again:
  getch (&c);

  if (EditStringFlag && !InExpr)
    {
      char string[MAXSIMPSTR];
      int spos;

      if (c == '\'')
	{
	  EditStringFlag = FALSE;
	  return (ENDEDITSTR);
	}
      if (c == '{')
	{
	  InExpr++;
	  return ('{');
	}
      spos = 0;
      ungetch (c);
      while (spos < sizeof string - 1)	/* -1 as below can produce 2 chars */
	{
	  getch (&c);
	  if (!isprint (c) && !isspace (c))
	    err ("Illegal character in edited string");
	  if (c == '"')
	    {
	      string[spos++] = '\\';
	    }
	  else if (c == '%')
	    {
	      string[spos++] = '%';	/* Double %'s for printf */
	    }
	  else if (c == '\n')
	    {
	      string[spos++] = '\\';
	      c = 'n';
	    }
	  else if (c == '\t')
	    {
	      string[spos++] = '\\';
	      c = 't';
	    }
	  else if (c == '\\')
	    {
	      getch (&c);
	      if (c == '\n')
		continue;
	      if (c == '%')
		{
		  string[spos++] = '%';
		  string[spos++] = '%';
		  continue;
		}
	      /* 'C' may need the backslash as well */
	      string[spos++] = '\\';
	    }
	  else if (c == '\'' || c == '{')
	    {
	      (void) ungetch (c);
	      string[spos] = '\0';
	      yylval.cp = strdup (string);
	      return (EDITSTRPART);
	    }
	  string[spos++] = c;
	}
      err ("Edited string much too long");
#ifdef lint
      return (0);
#endif
    }
  if (EditStringFlag && InExpr)
    {
      if (c == '}')
	{
	  InExpr--;
	  return ('}');
	}
      else if (c == '{')
	{
	  InExpr++;
	  return ('{');
	}
    }

  if (c == '@')
    {
      IncludeFile ();
      goto again;
    }
  else if (isalpha (c))
    {
      char name[1024];
      int npos;

      name[0] = c;
      npos = 1;
      while (npos < sizeof name)
	{
	  getch (&c);
	  if (!isalnum (c))
	    {
	      int kw;
	      if (c == '.')
		{
		  struct variable *v;
		  char *p;
		  name[npos] = '\0';
		  v = findvar (name);
		  if (!v)
		    err ("Bad table access");
		  if (!v->t_tpl.inuse)
		    err ("Attempt to use an inactive tuple variable");
		  name[npos++] = c;
		  p = &name[npos];
		  while (npos < sizeof name)
		    {
		      getch (&c);
		      if (!isalnum (c))
			{
			  int i;
			  (void) ungetch (c);
			  name[npos] = '\0';

			  for (i = 0; i < v->t_tbl->numattr; i++)
			    if (strcmp (p, v->t_tbl->names[i]) == 0)
			      break;
			  if (i == v->t_tbl->numattr)
			    err2 ("%s is not an attribute of %s", p, v->name);
			  if (!v->t_tbl->UsedInPut[i])
			    err2 ("%s has not been put into table %s", p, v->name);
			  goto foundtableaccess;
			}
		      name[npos++] = c;
		    }
		  err ("Table access name much too long");
		}
	      (void) ungetch (c);
	      name[npos] = '\0';
	      kw = findkw (name);
	      if (kw)
		return (kw);
	      if (!NamesOnly)
		{
		foundtableaccess:
		  if (InWhere)
		    {
		      yylval.vp = findTvar (name);
		    }
		  else
		    yylval.vp = (struct variable *) 0;
		  if (!yylval.vp)
		    yylval.vp = findvar (name);
		  if (yylval.vp)
		    {
		      switch (yylval.vp->datatype)
			{
			case DT_BOOLEAN:
			case DT_STRING:
			case DT_INTEGER:
			case DT_REAL:
			  return (VARNAME);
			case DT_LABEL:
			  return (LABEL);
			case DT_TABLE:
			  return (TABLE);
			case DT_TUPLE:
			  return (TUPLE);
			default:
			  CantHappen ("yylex, var datatype switch took default");
			}
		    }
		  yylval.fp = findfunc (name);
		  if (yylval.fp)
		    {
		      return (FUNCNAME);
		    }
		}
	      yylval.cp = strdup (name);
	      return (NAME);
	    }
	  name[npos++] = c;
	}
      err ("Name much too long");
#ifdef lint
      return (0);
#endif
    }
  else if (c == ':' && EditStringFlag)
    {
      char fdesc[50];
      int DecimalFound;
      int fdpos;

      DecimalFound = FALSE;
      getch (&c);
      fdesc[0] = c;
      fdpos = 1;
      if (c != '-' && !isdigit (c) && c != '.')
	err ("Bad Field Descriptor");
      while (fdpos < sizeof fdesc)
	{
	  getch (&c);
	  if (!(isdigit (c) || (!DecimalFound && c == '.')))
	    {
	      (void) ungetch (c);
	      fdesc[fdpos] = '\0';
	      yylval.cp = strdup (fdesc);
	      return (FIELDDESC);
	    }
	  fdesc[fdpos++] = c;
	  if (c == '.')
	    DecimalFound = TRUE;
	}
      err ("Field Descriptor much too long");
#ifdef lint
      return (0);
#endif
    }
  else if (isdigit (c))
    {
      char number[50];
      int DecimalFound;
      int eFound;
      int npos;

      DecimalFound = FALSE;
      eFound = FALSE;
      number[0] = c;
      npos = 1;
      while (npos < sizeof number)
	{
	  getch (&c);
	  /* Gather Hex Constant */
	  if (npos == 1 && number[0] == '0' && c == 'x')
	    {
	      number[npos++] = c;
	      while (npos < sizeof number)
		{
		  getch (&c);
		  if (!(isdigit (c) || (c <= 'f' && c >= 'a')
			|| (c <= 'F' && c >= 'A')))
		    {
		      (void) ungetch (c);
		      number[npos] = '\0';
		      yylval.cp = strdup (number);
		      return (INTEGER);
		    }
		  number[npos++] = c;
		}
	      err ("Number much too long");
	    }
	  if (!(isdigit (c) ||
		(!eFound && (c == 'e' || c == 'E')) ||
		(!DecimalFound && c == '.')))
	    {
	      (void) ungetch (c);
	      number[npos] = '\0';
	      yylval.cp = strdup (number);
	      if (DecimalFound)
		return (REAL);
	      else
		return (INTEGER);
	    }
	  number[npos++] = c;
	  if (c == '.')
	    DecimalFound = TRUE;
	  if (c == 'e' || c == 'E')
	    {
	      DecimalFound = TRUE;
	      eFound = TRUE;
	      getch (&c);
	      if (!isdigit (c) && c != '+' && c != '-')
		err ("Bad floating point number");
	      number[npos++] = c;
	    }
	}
      err ("Number much too long");
#ifdef lint
      return (0);
#endif
    }
  else if (c == ';')
    {
      goto again;
    }
  else if (isspace (c))
    {
      goto again;
    }
  else if (c == EOF)
    return (YY_EOF);
  else if (c == '\'')
    {
      if (EditStringFlag)
	err ("Edited strings not allowed inside other edited strings");
      EditStringFlag = TRUE;
      InExpr = 0;
      getch (&c);
      if (c != '\n')		/* Swallow first trailing Newline */
	ungetch (c);
      goto again;
    }
  else if (c == '/')		/* C style comments */
    {
      getch (&c);
      if (c != '*')
	{
	  ungetch (c);
	  return ('/');
	}
      else
	{
	  while (c != YY_EOF)
	    {
	      getch (&c);
	      if (c == '*')
		{
		  getch (&c);
		  if (c == '/')
		    goto again;
		  ungetch (c);
		}
	    }
	}
    }
  else if (c == '"')
    {
      char string[MAXSIMPSTR];
      int spos;

      spos = 0;
      while (spos < sizeof string - 1)	/* -1 because 2 can be produced below */
	{
	  getch (&c);
	  if (!isprint (c) && c != '\t')
	    err ("Illegal character in simple string");
	  if (c == '\\')
	    {
	      getch (&c);
	      if (c == '\n')
		continue;
	      /* 'C' will need the backslash as well */
	      string[spos++] = '\\';
	    }
	  else if (c == '"')
	    {
	      string[spos] = '\0';
	      yylval.cp = strdup (string);
	      return (STRING);
	    }
	  string[spos++] = c;
	}
      err ("String much too long");
#ifdef lint
      return (0);
#endif
    }
  else if (c == '$')		/* Rest of line after $ is a comment */
    {
      while (c != '\n' && c != YY_EOF)
	getch (&c);
      ungetch (c);
      goto again;
    }
  else
#define retop(s,t) {yylval.cp=s;return(t);}
    switch (c)
      {
      case '=':
	getch (&c);
	if (c == '=')
	  retop ("==", EQUOP);
	ungetch (c);
	return ('=');
      case '~':
	getch (&c);
	if (c == '=')
	  retop ("~=", MATCHES);
	ungetch (c);
	return ('~');
      case '>':
	getch (&c);
	if (c == '=')
	  retop (">=", RELOP);
	if (c == '>')
	  {
	    return (RIGHTSHIFT);
	  }
	ungetch (c);
	retop (">", RELOP);
      case '<':
	getch (&c);
	if (c == '=')
	  retop ("<=", RELOP);
	if (c == '<')
	  {
	    return (LEFTSHIFT);
	  }
	ungetch (c);
	retop ("<", RELOP);
      case '&':
	getch (&c);
	if (c == '&')
	  return (AND);
	ungetch (c);
	return ('&');
      case '|':
	getch (&c);
	if (c == '|')
	  return (OR);
	ungetch (c);
	return ('|');
      case '!':
	getch (&c);
	if (c == '=')
	  retop ("!=", EQUOP);
	ungetch (c);
	return ('!');
      default:
	return (c);
      }
}
