/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/file.h>
#include "global.h"
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <signal.h>
#include <syslog.h>
#include <errno.h>

time_t time ();

#include "nethead.h"
#include "db.h"
#include "satd.h"


#ifndef strdup
char *strdup();
#endif
char *GetLastModifiedTime ();
char *Gets ();

void DoLock();
void DoUnlock();

static int debug = 2;

RETSIGTYPE
  siginthandler ()
{
  (void) signal (SIGINT, siginthandler);
  debug = (debug + 1) % 3;
  syslog (LOG_INFO, "debugging %sabled", (debug) ? "en" : "dis");
}

#define MAXAP (3*MAXATTR)


main (argc, argv)
     int argc;
     char *argv[];
{
  char buf[1024];
  char user[2048];
  char command[128];
  char hostname[1024];
  int s;			/* File descriptor for socket */
  int ac;			/* argument count */
  int i;
  int retval;
  char *ap[MAXAP];
  int rc;
  
#ifndef undef /* debugging */
  switch (fork ())
    {
    case -1:
      {
	perror ("satd: ERROR: fork() failed");
	exit (-1);
      }
    case 0:
      {
	break;
      }
    default:			/* If pid > 0, process is parent, so exit parent */
      {
	exit (0);
      }
    }
#endif undef
  
#ifdef LOG_LOCAL5
  openlog ("satd", LOG_PID, LOG_LOCAL5);
#else
  openlog ("satd", LOG_PID);
#endif
  
  syslog (LOG_INFO, "SAT daemon started ok.");
  
  (void) setuid (0);
  (void) setgid (SECUREGROUP);
  
  (void) close (0);		/* Close stdin, stdout, and stderr */
  (void) close (1);
  (void) close (2);
  
  (void) open ("/dev/null", O_RDWR);
  (void) dup (0);
  (void) dup (0);
  
#ifdef SYSV
  (void) open ("/dev/null", O_WRONLY);	/* get rid of fd 3 for WIN3b TCP */
#endif
  
#ifdef HAVE_SETSID
  (void) setsid ();
#else
  (void) setpgrp (0, getpid ());
  {
    int fd;
    
    fd = open ("/dev/tty", O_RDONLY);
    (void) ioctl (fd, TIOCNOTTY, (char *) 0);
    (void) close (fd);
  }
#endif
  
  {
    char pidbuf[256];
    int fd;
    
    (void) sprintf (pidbuf, "%d\n", getpid ());
    fd = creat ("/etc/satd.pid", 0644);
    (void) write (fd, pidbuf, strlen (pidbuf));
    (void) close (fd);
  }
  
  (void) signal (SIGINT, siginthandler);
  
  setenv ("PATH", PATH, 1);	/* HMS: syslog this if debugging */
  
  /* Findclient returns the socket number */
  if ((s = findclient ("satd", hostname, sizeof hostname)) == -1)
    {
      syslog (LOG_ERR, "findclient(satd, <hostname>) failed: %m.");
      syslog (LOG_ERR, "satd exiting...");
      exit (1);
    }
  
  if (debug)
    {
      syslog (LOG_DEBUG, "Got a connection from %s", hostname);
    }
  
#if defined(NODUPSOCKET) && 0
#else
  (void) close (0);
  (void) close (1);
  (void) close (2);
  
  retval = dup (s);
  if (retval == -1)
    {
      syslog (LOG_ERR, "Bad dup of socket to stdin: %m");
      syslog (LOG_ERR, "satd exiting...");
      exit (1);
    }
  if(debug > 1)
    syslog (LOG_DEBUG, "satd.c:main():dup(findclient,0)=%d, COE flag=%d"
	    , retval, fcntl(retval, F_GETFD, 0));
  
  retval = dup (s);
  if (retval == -1)
    {
      syslog (LOG_ERR, "Bad dup of socket to stdout: %m");
    }
  if(debug > 1)
    syslog (LOG_DEBUG, "satd.c:main():dup(findclient,1)=%d, COE flag=%d"
	    , retval, fcntl(retval, F_GETFD, 0));
  
  retval = dup (s);
  if (retval == -1)
    {
      syslog (LOG_ERR, "Bad dup of socket to stderr: %m");
    }
  if(debug > 1)
    syslog (LOG_DEBUG, "satd.c:main():dup(findclient,2)=%d, COE flag=%d"
	    , retval, fcntl(retval, F_GETFD, 0));
#endif
  
  if (!Gets (command, sizeof command, stdin))
    {
      syslog (LOG_ERR, "Gets(command) failed");
    }
  if(debug > 1)
    {
      syslog (LOG_DEBUG, "Gets(command)=<%s>", command);
    }
  
  if (!Gets (buf, sizeof buf, stdin))
    {
	  char message[100]; /* bugtrap; raphael 1/95 */
	  sprintf (message, "Gets(SATID) failed; contacted by %s, command %s",
		  hostname, command);
	  syslog (LOG_ERR, message);
	  exit (1); /* raphael: added.  Otherwise, we core dump later! */
    }
  if(debug > 1)
    {
      syslog (LOG_DEBUG, "Gets(SATID)=<%s>", buf);
    }
  
  (void) setenv ("SATID", buf, 1);
  
  if (!Gets (buf, sizeof buf, stdin))
    {
      syslog (LOG_ERR, "Gets(ac) failed");
    }
  ac = atoi (buf);
  if(debug > 1)
    {
      syslog (LOG_DEBUG, "Gets(ac)=<%s>", buf);
    }
  
  
  if (ac >= (MAXAP - 1))
    {
      char ebuf[1024];
      
      (void) sprintf (ebuf, "Arg count too big from %s, command=%s, buf=%s",
		      hostname, command, buf);
      syslog (LOG_ERR, "%s.", ebuf);
      syslog (LOG_ERR, "exiting.");
      exit (1);
    }
  
  for (i = 0; i < ac; i++)
    {
      if (!Gets (buf, sizeof buf, stdin))
		{
		  syslog (LOG_ERR, "Gets(ap[%d]) failed", i);
		}
      if ((ap[i] = strdup (buf)) == NULL)
		{
		  syslog (LOG_ERR, "strdup(ap[%d]) failed", i);
		}
	  if (debug > 1)
		{
		  syslog (LOG_DEBUG, "arg[%d]=[%s]", i, ap[i]);
		}
    }
  ap[i] = (char *) 0;
  
  if (ferror (stdin))
    {
      syslog (LOG_ERR, "Bad read: %m.");
      syslog (LOG_ERR, "exiting.");
      exit (1);
    }
  
  if (ap[0] == 0) { /* bug trap; raphael 1/95 */
	char message[100];
	sprintf (message, "ap[0] is empty; contacted by %s", hostname);
	syslog (LOG_ERR, message);
	exit (1);
  }
  (void) sprintf (user, "%s@%s", ap[0], hostname);
  SetUserName (user);
#ifdef undef
  if (ap[0])			/* Return memory we got from strdup() */
    free((void *)ap[0]);
#endif
  ap[0] = user;
  
  if (strcmp (command, "read") == 0)
    {
#ifdef NODUPSOCKET
      /* Attach BTABREAD to the socket s, using the arglist in ap */
      rc = pipeopen (WRITE_TO_PIPE, s, BTABREAD, ap);
      exit(rc);
#else
      syslog (LOG_DEBUG, "execv(%s).", BTABREAD);
      (void) execv (BTABREAD, ap);
      syslog (LOG_ERR, "execv(%s) failed: %m", BTABREAD);
      _exit(127);
#endif
    }
  else if(strcmp(command, "version") == 0)
    {
      if(ap[1])
	free ((void *)ap[1]);
      ap[1] = "SATD v1.5";
#ifdef NODUPSOCKET
      rc = pipeopen(WRITE_TO_PIPE, s, "/bin/echo", ap);
      exit(rc);
#else
      (void) execv("/bin/echo", ap);
      syslog (LOG_ERR, "execv(%s) failed: %m", "/bin/echo");
      _exit(127);
#endif
    }
  else if (strcmp (command, "pull") == 0)
    {
      if (CheckPullLock (ap[1], ap[3]))
	{
	  syslog (LOG_DEBUG, "CheckPullLock(%s, %s) exists: exiting", ap[1], ap[3]);
	  exit (0);
	}
      if(ap[2])
	free((void *)ap[2]);
      ap[2] = "-pull";
#ifdef NODUPSOCKET
      rc = pipeopen (WRITE_TO_PIPE, s, RELMOD, ap);
      exit(rc);
#else
      syslog (LOG_DEBUG, "execv(%s).", RELMOD);
      (void) execv (RELMOD, ap);
      syslog (LOG_ERR, "execv(%s) failed: %m", RELMOD);
      _exit(127);
#endif
    }
  else if (strcmp (command, "sendupdate") == 0)
    {
      if(ap[2])
	free((void *)ap[2]);
      ap[2] = "-sendupdate";
#ifdef NODUPSOCKET
      rc = pipeopen (WRITE_TO_PIPE, s, RELMOD, ap);
      exit(rc);
#else
      syslog (LOG_DEBUG, "execv(%s).", RELMOD);
      (void) execv (RELMOD, ap);
      syslog (LOG_ERR, "execv(%s) failed: %m", RELMOD);
      _exit(127);
#endif
    }
  else if (strcmp (command, "receiveupdate") == 0)
    {
      if(ap[2])
	free((void *)ap[2]);
      ap[2] = "-receiveupdate";
#ifdef NODUPSOCKET
      rc = pipeopen (READ_FROM_PIPE, s, RELMOD, ap);
      exit(rc);
#else
      syslog (LOG_DEBUG, "execv(%s).", RELMOD);
      (void) execv (RELMOD, ap);
      syslog (LOG_ERR, "execv(%s) failed: %m", RELMOD);
      _exit(127);
#endif
    }
  else if (strcmp (command, "lock") == 0)
    {
      syslog (LOG_DEBUG, "Attempting DoLock(%s, %s, %s).", ap[0], ap[1], ap[2]);
      DoLock (ap[0], ap[1], ap[2]);
      syslog (LOG_DEBUG, "Done DoLock(%s, %s, %s).", ap[0], ap[1], ap[2]);
      exit (0);
    }
  else if (strcmp (command, "unlock") == 0)
    {
      DoUnlock (ap[0], ap[1]);
      syslog (LOG_DEBUG, "DoUnlock(%s, %s).", ap[0], ap[1]);
      exit (0);
    }
  else
    {
      char ebuf[1024];
      
      (void) sprintf (ebuf, "Bad command %s from %s, first args: %s, %s, %s",
		      command, ap[0], ap[1], ap[2], ap[3]);
      syslog (LOG_ERR, "%s.", ebuf);
      syslog (LOG_ERR, "exiting.");
      exit (1);
    }
  
  {
    char ebuf[1024];
    
    (void) sprintf (ebuf, "Command %s from %s, first 3 of %d args: %s, %s, %s",
		    command, ap[0], ac, ap[1], ap[2], ap[3]);
    syslog (LOG_ERR, "Execv failed: %m; %s.", ebuf);
    syslog (LOG_ERR, "exiting.");
  }
  exit (1);
}

char C;
#define Return(i) C = i, (void) write(1, &C, 1), exit(0)

void DoLock (user, name, lmodtime)
     char *user, *name, *lmodtime;
{
  time_t start;
  static struct scheme sch;
  
  if (!ReadScheme (name, &sch))
    {
      Return (BAD_SCHEME_READ);
    }
  
  syslog (LOG_DEBUG, "About to check permissons.");
  if (!CheckUserPermission (user, sch.owner, sch.useraliasfile) &&
      !CheckUserPermission (user, sch.modifier, sch.useraliasfile) &&
      !CheckUserPermission (user, sch.appender, sch.useraliasfile))
    {
      Return (NO_PERMISSION);
    }
  
  syslog (LOG_DEBUG, "About to check time.");
  if (strcmp (lmodtime, GetLastModifiedTime (&sch)) < 0)
    Return (THERE_IS_A_NEWER_VERSION);
  
  start = time ((time_t *) 0);
  syslog (LOG_DEBUG, "About to check lock.");
  while (CheckClobberLock (&sch))
    {
      if (MyClobberLock (&sch))
	Return (GOOD_LOCK);
      if ((time ((time_t *) 0) - start) > MAXLOCKWAIT)
	Return (LOCK_ALREADY_EXISTS);
      syslog (LOG_DEBUG, "waiting for lock");
      sleep (1);		/* most implementations do not guarantee a 1 second sleep */
    }
  
  if (!MakeClobberLock (&sch))
    Return (LOCK_ERROR);
  Return (GOOD_LOCK);
}


void DoUnlock (user, name)
     char *user, *name;
{
  static struct scheme sch;
  
  if (!ReadScheme (name, &sch))
    {
      Return (BAD_SCHEME_READ);
    }
  
  if (!CheckUserPermission (user, sch.owner, sch.useraliasfile) &&
      !CheckUserPermission (user, sch.modifier, sch.useraliasfile) &&
      !CheckUserPermission (user, sch.appender, sch.useraliasfile))
    {
      Return (NO_PERMISSION);
    }
  
  if (!RemoveClobberLock (&sch))
    Return (UNLOCK_ERROR);
  Return (GOOD_LOCK);
}
