
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "schedit.h"

char *getstr (), *getpat (), *gettext ();
long getlong ();
double getreal ();


void CreateScheme (name, sch)
     char *name;
     struct scheme *sch;
{
  int i;

  FreeScheme (sch);

  (void) printf ("Creating %s.\n\n", name);
  sch->name = strdup (name);

  if (*sch->name != '.' && *sch->name != '/')
    {
      char buf[2048];
      if (getuid () != 0)
	{
	  printf ("You must be root to create a new system relation\n");
	  leave (1);
	}
      (void) sprintf (buf, "%s/%s", SYSDBPATH, name);
      sch->path = strdup (buf);
    }
  else
    {
      (void) printf ("Using your real id to create this scheme\n");
      (void) setgid (getgid ());/* CHANGE raphael 6/92 */
      (void) setuid (getuid ());/* CHANGE raphael 6/92 */
      sch->path = strdup (sch->name);
    }

  {				/* Create the directory */
    char buf[2048];
    struct stat sb;
    if (stat (sch->path, &sb) == 0)
      {
	if (getuid () != 0 || sb.st_uid != 0 || sb.st_gid != SECUREGROUP)
	  {
	    fprintf (stderr, "Permission denied, directory already exists.\n");
	    leave (1);
	  }
	fprintf (stderr, "Warning: %s exists.\n", sch->path);
	if (!getyesno ("Continue?", "no"))
	  leave (1);
      }
    else
      {
	(void) sprintf (buf, "/bin/mkdir %s", sch->path);
	(void) printf ("Creating directory: %s\n", buf);
	if (system (buf) != 0)
	  {
	    fprintf (stderr, "Could not create directory with: %s\n", buf);
	    leave (1);
	  }
	/* Require that the user be able to create the directory,
		but for security reasons, only root should be able
		to make files there.  CHANGE 6/92 raphael: only set group
		if possible; readsch will complain if group is not right */
	(void) chown (sch->path, 0, SECUREGROUP);
	if (chmod (sch->path, 0750) == -1)
	  {
	    (void) rmdir (sch->path);
	    perror ("Should not happen, chown/chmod failed");
	    leave (1);
	  }
      }
  }

  sch->datafromprog = getbool ("Gets data from a program?", "false");
  sch->desc = gettext ("a description of this relation");

  sch->useraliasfile = getstr ("User alias file", USERALIASFILE);
  sch->distaliasfile = getstr ("Distribution alias file", DISTALIASFILE);

  sch->owner = getstr ("Owner alias", DEFOWNERALIAS);
  sch->modifier = getstr ("Modifier alias", DEFMODIFIERALIAS);
  sch->appender = getstr ("Appender alias", DEFAPPENDERALIAS);
  sch->reader = getstr ("Reader alias", DEFREADERALIAS);

  sch->distributeflag = getbool ("Automatically distribute this relation?", "false");
  if (sch->distributeflag)
    {
      sch->model_hosts = getstr ("Model hosts", DEFMODELALIAS);
      sch->trusted_hosts = getstr ("Trusted hosts", DEFTRUSTEDALIAS);
      sch->copy_hosts = getstr ("Copy hosts", DEFCOPYALIAS);
      sch->access_hosts = getstr ("Access hosts", DEFACCESSALIAS);
    }
  else
    {
      sch->model_hosts = strdup ("");
      sch->trusted_hosts = strdup ("");
      sch->copy_hosts = strdup ("");
      sch->access_hosts = strdup ("");
    }

  sch->logflag = getbool ("Log changes for this relation?", "true");
  if (sch->logflag)
    {
      sch->index_attribute = getstr ("Name of index attribute", "");
      sch->log_file = getstr ("File for change log", "./log");
    }
  else
    {
      sch->index_attribute = strdup ("");
      sch->log_file = strdup ("");
    }

  sch->checkflag = getbool ("Run a checking program?", "false");

  sch->reportflag = getbool ("Run a reporting program?", "false");

  sch->numattr = getboundint ("Number of attributes", "", 0, MAXATTR);

  for (i = 0; i < sch->numattr; i++)
    {
      printf ("\n\nGetting for attribute %d.\n", i + 1);
    again:
      sch->at[i].name = getstr ("Name", "");
      {
	int j;
	for (j = 0; j < i; j++)
	  if (strcmp (sch->at[j].name, sch->at[i].name) == 0)
	    {
	      fprintf (stderr, "You already used that name.\n");
	      free (sch->at[i].name);
	      goto again;
	    }
      }
      sch->at[i].desc = gettext ("a description of this attribute");
      sch->at[i].checktype = getmenu (CheckTypeDef, "Check type", "string");
      switch (sch->at[i].checktype)
	{
	case CT_STRING:
	case CT_TEXT:
	case CT_BOOLEAN:
	  break;
	case CT_PAT:
	  sch->at[i].at_patsrc = getpat ("Pattern", "");
	  break;
	case CT_INTEGER:
	  sch->at[i].at_Ibcheck = getbool ("Bound Check?", "false");
	  if (sch->at[i].at_Ibcheck)
	    {
	      sch->at[i].at_Ilow = getlong ("Low bound", "");
	      sch->at[i].at_Ihi = getlong ("High bound", "");
	    }
	  break;
	case CT_REAL:
	  sch->at[i].at_Rbcheck = getbool ("Bound Check?", "false");
	  if (sch->at[i].at_Rbcheck)
	    {
	      sch->at[i].at_Rlow = getreal ("Low bound", "");
	      sch->at[i].at_Rhi = getreal ("High bound", "");
	    }
	  break;
	default:
	  CantHappen ("Took default in creatsch.\n");
	}
    }
}

char *
getans (prompt, def)
     char *prompt, *def;
{
  static char answer[1024];

nodefloop:
  (void) printf ("%s [%s]: ", prompt, *def ? def : "no default");
  clearerr (stdin);		/* clear accidental EOF */
  (void) fgets (answer, sizeof answer, stdin);
  answer[strlen (answer) - 1] = '\0';	/* strip newline */
  if (*answer)
    return (answer);
  else
    {
      if (!*def)
	{
	  printf ("No default value.\n");
	  goto nodefloop;
	}
      return (def);
    }
}


char *
getstr (prompt, def)
     char *prompt;
     char *def;
{
  char *p;
  p = getans (prompt, def);
  return (strdup (p));
}


char *
getpat (prompt, def)
     char *prompt;
     char *def;
{
  char *p;
again:
  p = getans (prompt, def);
  if (!CheckPattern (p))
    goto again;
  return (strdup (p));
}


getint (prompt, def)
     char *prompt;
     char *def;
{
  int i;
  i = atoi (getans (prompt, def));
  return (i);
}


getboundint (prompt, def, low, high)
     char *prompt;
     char *def;
     int low;
     int high;
{
  int i;
again:
  i = getint (prompt, def);
  if (i > high || i < low)
    {
      fprintf (stderr, "Integer must be between %d and %d.\n", high, low);
      goto again;
    }
  return (i);
}


long
getlong (prompt, def)
     char *prompt;
     char *def;
{
  long l;
  l = (long) atof (getans (prompt, def));
  return (l);
}


double
getreal (prompt, def)
     char *prompt;
     char *def;
{
  double d;
  d = atof (getans (prompt, def));
  return (d);
}



getbool (prompt, def)
     char *prompt;
     char *def;
{
  char *p;
again:
  p = getans (prompt, def);
  if (strcmp (p, "true") != 0 && strcmp (p, "false") != 0)
    {
      fprintf (stderr, "Answer must be either \"true\" or \"false\".\n");
      goto again;
    }
  if (strcmp (p, "true") == 0)
    return (TRUE);
  return (FALSE);
}


char *
gettext (prompt)
     char *prompt;
{
  char text[MAXSTR];
  int i;
  printf ("Enter text for: %s,\n\tType ^D when finished.\n", prompt);
  for (i = 0; i < MAXSTR - 1; i++)
    {
      int c;
      c = fgetc (stdin);
      if (c == EOF)
	break;
      text[i] = c;
    }
  if (i > MAXSTR)
    {
      fprintf (stderr, "String too long.\n");
    }
  if (i > 0)
    text[i - 1] = '\n';
  text[i] = '\0';
  if (ferror (stdin))
    {
      perror ("Stdin");
      leave (1);
    }
  clearerr (stdin);		/* clear EOF */
  return (strdup (text));
}


getmenu (D, prompt, def)
     struct definition *D;
     char *prompt;
     char *def;
{
  char *p;
  struct definition *d;

again:
  printf ("\n\nChoose one of:\n");
  for (d = D; d->name; d++)
    {
      printf ("\t%s\n", d->name);
    }
  printf ("\n");
  p = getans (prompt, def);
  for (d = D; d->name; d++)
    if (strcmp (p, d->name) == 0)
      return (d->value);
  fprintf (stderr, "Choice not in menu.\n");
  goto again;

/*NOTREACHED*/
}



getyesno (prompt, def)
     char *prompt;
     char *def;
{
  char *p;
again:
  p = getans (prompt, def);
  if (strcmp (p, "yes") != 0 && strcmp (p, "no") != 0)
    {
      fprintf (stderr, "Answer must be either \"yes\" or \"no\".\n");
      goto again;
    }
  if (strcmp (p, "yes") == 0)
    return (TRUE);
  return (FALSE);
}
