/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
extern int errno;

#include "schedit.h"

#include <signal.h>
#ifndef BADSIG
#define BADSIG	((RETSIGTYPE (*)())(-1))
#endif

Boolean createflag;

void leave();

static
void Usage ()
{
  fprintf (stderr, "Usage: schedit [-c|-r] relationname\n");
  leave (1);
}


static struct scheme sch;

char *Gets ();

static char TempFileName[20];
static Boolean TempFileCreated = FALSE;
static int ReadOnly;
static int ParseError;

main (argc, argv)
     int argc;
     char *argv[];
{
  char *mktemp ();
  Boolean notdone;
  RETSIGTYPE sigleave ();

  (void) umask (0027);

  if (argc > 3 || argc < 2)
    Usage ();
  if (argc == 3)
    {
      if (strcmp (argv[1], "-c") != 0 && strcmp (argv[1], "-r") != 0)
	Usage ();
      else
	{
	  if (strcmp (argv[1], "-c") == 0)
	    createflag = TRUE;
	  else
	    ReadOnly = TRUE;
	}
    }


  strcpy (TempFileName, "/tmp/tabeditXXXXXX");	/* must be in modifiable space */
  if (signal (SIGINT, sigleave) == BADSIG
      || signal (SIGQUIT, sigleave) == BADSIG
      || signal (SIGHUP, sigleave) == BADSIG)
    {
      perror ("signal");
      leave (1);
    }


  if (createflag)
    {
      CreateScheme (argv[2], &sch);
    }
  else
    {
      if (!ReadScheme (ReadOnly ? argv[2] : argv[1], &sch))
	{
	  /* (void) fprintf(stderr,"Cannot open scheme for %s.\n",
			ReadOnly?argv[2]:argv[1]);
		*/
	  leave (1);
	}
    }

  if (!CheckUserPermission (UserName (), sch.owner, sch.useraliasfile)
      && getuid () != 0)
    {
      if (!CheckUserPermission (UserName (), sch.reader, sch.useraliasfile) &&
      !CheckUserPermission (UserName (), sch.modifier, sch.useraliasfile) &&
	!CheckUserPermission (UserName (), sch.appender, sch.useraliasfile))
	{
	  fprintf (stderr, "You don't have any access rights to this relation.\n");
	  leave (1);
	}
      if (!ReadOnly)
	{
	  fprintf (stderr, "You are not the owner of this relation so access is readonly.\n");
	  ReadOnly = TRUE;
	}
    }

  if (!ReadOnly && !createflag && !CheckClobberLock (&sch))
    {
      fprintf (stderr, "Warning: lock not set, this will be a readonly access.\n");
      sleep (2);
      ReadOnly = TRUE;
    }

  if (!ReadOnly && !MyClobberLock (&sch))
    {
      fprintf (stderr, "Someone else has the lock set, this access will be readonly.\n");
      sleep (2);
      ReadOnly = TRUE;
    }

  if (!ReadOnly && !CheckUserPermission (UserName (), sch.owner, sch.useraliasfile))
    {
      if (getuid () != 0)
	{
	  fprintf (stderr, "Only the owner of the relation can edit the scheme.\n");
	  leave (1);
	}
      else
	{
	  fprintf (stderr, "Warning: you do not own this relation.\n");
	  sleep (2);
	}
    }

  (void) mktemp (TempFileName);

  if (WriteAsciiFile (TempFileName, &sch))
    TempFileCreated = TRUE;
  else
    leave (1);

  (void) signal (SIGINT, SIG_IGN);
  (void) signal (SIGQUIT, SIG_IGN);
  (void) signal (SIGHUP, SIG_IGN);

  notdone = TRUE;

  ParseError = FALSE;

  while (notdone)
    {
      struct stat osb, nsb;

      if (stat (TempFileName, &osb) == -1)
	{
	  perror ("Bad stat of temporary file");
	  leave (1);
	}

      EditAsciiFile (TempFileName);

      if (stat (TempFileName, &nsb) == -1)
	{
	  perror ("Bad stat of temporary file");
	  leave (1);
	}

      if (ReadOnly ||
	  (osb.st_mtime == nsb.st_mtime && !ParseError && !createflag))
	leave (0);

      if (!ReadAsciiFile (TempFileName, &sch))
	{
	  ParseError = TRUE;
	  notdone = TRUE;
	}
      else if (sch.logflag &&
	  (strlen (sch.index_attribute) == 0 || strlen (sch.log_file) == 0))
	{
	  printf ("Logging requires both an index attribute and a log file.\n");
	  ParseError = TRUE;
	  notdone = TRUE;
	}
      else
	{
	  char ans[128];
	  printf ("Save changes? ");
	  (void) Gets (ans, sizeof ans, stdin);
	  if (*ans != 'y' && *ans != 'Y')
	    notdone = TRUE;
	  else
	    notdone = FALSE;
	}

      if (notdone)
	{
	  char buf[256];

	  (void) printf ("E(dit) or A(bort): ");
	  if (fgets (buf, sizeof buf, stdin) == NULL)
	    leave (1);
	  if (*buf == 'A' || *buf == 'a')
	    leave (1);
	}
    }


  if (!sch.datafromprog)
    {
      char fn[2048];
      struct stat sb;
      (void) sprintf (fn, "%s/data", sch.path);
      if (stat (fn, &sb) == -1 && errno == ENOENT)
	{
	  FILE *fp;
	  (void) unlink(fn); /* prevent attack whereby new file already exists
		as a symlink to something we don't want to overwrite */
	  if ((fp = fopen (fn, "w")) == NULL)
	    {
	      (void) fprintf (stderr, "Could not create `data' file.\n");
	      leave (1);
	    }
	  (void) NetWriteEOF (fp);
	  if (fclose (fp) == EOF)
	    {
	      (void) fprintf (stderr, "Could not write empty `data' file.\n");
	      leave (1);
	    }
	}
    }

  if (!WriteScheme (&sch))
    {
      (void) fprintf (stderr, "Error writing scheme -- aborting.\n");
      leave (1);
    }
  if (!TouchLastModified (&sch))
    {
      perror ("Could not set last modified time");
      leave (1);
    }
  leave (0);
}

void leave (status)
     int status;
{
  if (TempFileCreated)
    (void) unlink (TempFileName);
  exit (status);
}

RETSIGTYPE
sigleave ()
{
  leave (1);
}
