%{

/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <sys/types.h>
#include <setjmp.h>

#include "tabedit.h"

#define YY_EOF '\0'

static FILE *OutFile;

char *Gets();
FILE *CloneFile(char *RealName, char *NewName);

static int curattr;

static struct scheme *SCH;

static jmp_buf env;
#define Error(s) {error(s); longjmp(env,1);}

%}

%union {
	char *str;
	}

%token <str> ATTRNAME
%token <str> BOOLEAN
%token <str> NUMBER
%token <str> STRING
%token <str> TEXT

%type <str> StrExpr

%%

DataList:	Datum
	|	DataList Datum
	;

Datum:		ATTRNAME '=' StrExpr
			{
			if(strcmp($1,SCH->at[curattr].name) != 0)
				{
				Error("Unexpected attribute name.");
				}
			if(!CheckItem($3,&SCH->at[curattr]))
				{
				Error("Data does not match integrity constraint.");
				}
			(void) NetWriteString(OutFile,$3);
			if(++curattr >= SCH->numattr) curattr=0;
			free($1);
			free($3);
			}

StrExpr:	BOOLEAN
			{
			if(strcmp($1,"true") == 0)
				$$=strdup("1");
			else
				$$=strdup("0");
			free($1);
			}
	|	NUMBER
			{$$=$1;}
	|	STRING
			{$$=$1;}
	|	TEXT
			{$$=$1;}
	;

%%


yylex()
{
int c;

again:
getch(&c);

if(isalpha(c))
	{
	char name[1024];
	int npos;

	name[0] = c;
	npos=1;
	while(npos < sizeof name)
		{
		getch(&c);
		if(!isalnum(c) && c != '_')
			{
			int tk;
			if(c != EOF)
				ungetch(c);
			name[npos] = '\0';
			tk = findkw(name);
			if(!tk)
				{
				Error("Unknown attribute name.")
				}
			yylval.str = strdup(name);
			return(tk);
			}
		name[npos++] = c;
		}
	Error("Name much too long.")
#ifdef lint
	return(0);
#endif
	}
else if(isdigit(c))
	{
	char number[50];
	int DecimalFound;
	int eFound;
	int npos;

	DecimalFound = FALSE;
	eFound = FALSE;
	number[0] = c;
	npos=1;
	while(npos < sizeof number)
		{
		getch(&c);
		if(!(isdigit(c) ||
			(!eFound && (c == 'e' || c == 'E')) ||
			(!DecimalFound && c == '.')))
			{
			(void) ungetch(c);
			number[npos] = '\0';
			yylval.str = strdup(number);
			return(NUMBER);
			}
		number[npos++] = c;
		if(c == '.')
			DecimalFound = TRUE;
		if(c == 'e' || c == 'E')
			{
			DecimalFound = TRUE;
			eFound = TRUE;
			getch(&c);
			if(!isdigit(c) && c != '+' && c != '-')
				{
				fprintf(stderr,"Bad floating point number");
				leave(1);
				}
			number[npos++] = c;
			}
		}
	fprintf(stderr,"Number much too long");
	leave(1);
#ifdef lint
	return(0);
#endif
	}
else if(isspace(c))
	{
	goto again;
	}
else if(c == '"')
	{
	char string[MAXSIMPSTR];
	int spos;

	spos=0;
	while(spos < sizeof string)
		{
		getch(&c);
		if(!isprint(c))
			Error("Illegal character in simple string.")
		if(c == '\\')
			getch(&c);
		else if(c == '"')
			{
			string[spos] = '\0';
			yylval.str = strdup(string);
			return(STRING);
			}
		string[spos++] = c;
		}
	Error("String much too long.")
#ifdef lint
	return(0);
#endif
	}
else if(c == '[')
	{
	char text[MAXSTR];
	int txtpos;

	getch(&c);

	if(c != '\n')
		{
		text[0] = c;
		txtpos = 1;
		}
	else
		txtpos=0;

	while(txtpos < sizeof text)
		{
		getch(&c);
		if(!isprint(c) && !isspace(c))
			Error("Illegal character in text string.")
		if(c == '\\')
			getch(&c);
		else if(c == ']')
			{
			text[txtpos] = '\0';
			yylval.str = strdup(text);
			return(TEXT);
			}
		text[txtpos++] = c;
		}
	Error("Text string much too long.")
#ifdef lint
	return(0);
#endif
	}
else
	return(c);
}




Boolean
GetEditedData(file,s)
char *file;
struct scheme *s;
{
char NewName[2048];
char RealName[2048];
int parseisgood;
static int init=1;

SCH=s;

if(init)
	{
	InitKW();
	init=0;
	}

InFile=fopen(file,"r");
if(InFile == NULL)
	{
	(void) fprintf(stderr,"Cannot open temporary file.\n");
	return(FALSE);
	}

(void) sprintf(RealName,"%s/data",s->path);
(void) sprintf(NewName,"%s/data.new",s->path);
if ((OutFile = CloneFile(RealName, NewName)) == NULL) return(FALSE);

if(setjmp(env) == 0)
	{
	curattr=0;
	StartLN();
	if(yyparse() == 0)
		parseisgood=TRUE;
	else
		parseisgood=FALSE;
	}
else
	parseisgood=FALSE;

(void) NetWriteEOF(OutFile);

if(parseisgood && curattr != 0)
	{
	(void) fprintf(stderr,"Last set of attribute values incomplete.\n");
	return(FALSE);
	}

if(ferror(InFile) || fclose(InFile) == EOF)
	{
	(void) fprintf(stderr,"Read error.\n");
	return(FALSE);
	}

if(ferror(OutFile) || fclose(OutFile) == EOF)
	{
	(void) fprintf(stderr,"Write error.\n");
	return(FALSE);
	}

if(parseisgood)
	{
	char ans[128];
	char buf[2048];

	if(SCH->checkflag)
		{
		fprintf(stderr,"Checking data consistency.\n");
		(void) sprintf(buf,"%s/check",SCH->path);
		if(system(buf) != 0)
			return(FALSE);
		}

	printf("Save changes? ");
	(void) Gets(ans,sizeof ans,stdin);
	if(*ans != 'y' && *ans != 'Y')
		{
		(void) unlink(NewName);
		return(FALSE);
		}
	
	if(SCH->logflag)
		{
		fprintf(stderr,"Logging changes.\n");
		(void) sprintf(buf,"%s %s %s %s",TABDIFF,SCH->name,
			SCH->index_attribute,SCH->log_file);
		if(system(buf) != 0)
			return(FALSE);
		}

	if(rename(NewName,RealName) == -1)
		{
		perror("Bad rename");
		return(FALSE);
		}
	}
else
	{
	(void) unlink(NewName);
	return(FALSE);
	}

return(TRUE);
}

InitKW()
{
int i;

initkw();

for(i=0;i<SCH->numattr;i++)
	(void) insertkw(SCH->at[i].name,ATTRNAME);

(void) insertkw("true",BOOLEAN);
(void) insertkw("false",BOOLEAN);
}
