/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <syslog.h>
#include "tabedit.h"

#include <signal.h>
#ifndef BADSIG
#define BADSIG  ((RETSIGTYPE (*)())(-1))
#endif
#ifndef LOG_ODELAY
#       define LOG_ODELAY 0
#endif
#ifndef LOG_USER  
#       define LOG_USER LOG_INFO
#endif

FILE *InFile;
Boolean createflag;

RETSIGTYPE sigleave ();

void leave();

static
void Usage ()
{
  fprintf (stderr, "Usage: tabedit [-c|-r] relationname\n");
  leave (1);
}


static struct scheme sch;

static char TempFileName[20];
static Boolean TempFileCreated = FALSE;
static int ReadOnly;
static int ParseError;
static Boolean IMadeLock = FALSE;


char *UserName ();

main (argc, argv)
     int argc;
     char *argv[];
{
  FILE *fp;
  char *mktemp ();
  char cmdbuf[2048];
  Boolean notdone;

  (void) umask (0027);

  if (argc > 4 || argc < 2)
    Usage ();

  if (argc > 3 || argc < 2)
    Usage ();

  if (argc == 3)
    {
      if (strcmp (argv[1], "-c") != 0 && strcmp (argv[1], "-r") != 0)
	Usage ();

      if (strcmp (argv[1], "-r") == 0)
	ReadOnly = TRUE;
      else
	createflag = TRUE;
      argv++;
      argc--;
    }

  openlog ("tabedit", LOG_ODELAY, LOG_USER);
  syslog (LOG_DEBUG, "Sat:tabedit() opened syslog.");
  strcpy (TempFileName, "/tmp/tabeditXXXXXX");	/* must be in modifiable space */
  if (signal (SIGINT, sigleave) == BADSIG
      || signal (SIGQUIT, sigleave) == BADSIG
      || signal (SIGHUP, sigleave) == BADSIG)
    {
      perror ("signal");
      leave (1);
    }


  if (!ReadScheme (argv[1], &sch))
    {
      (void) fprintf (stderr, "Cannot open scheme for %s.\n", argv[1]);
      (void) fprintf (stderr,
	   "(Check root ownership, proper group ownership of directory)\n");
      leave (1);
    }

  if (!CheckUserPermission (UserName (), sch.owner, sch.useraliasfile) &&
      !CheckUserPermission (UserName (), sch.modifier, sch.useraliasfile))
    {
      if (getuid () == 0)
	{
	  fprintf (stderr, "Warning: you do not own this relation.\n");
	  sleep (2);
	}
      else if (CheckUserPermission (UserName (), sch.reader, sch.useraliasfile))
	{
	  if (ReadOnly == FALSE)
	    {
	      fprintf (stderr, "Warning: you are not allowed to modify, this will be a readonly access.\n");
	      sleep (2);
	      ReadOnly = TRUE;
	    }
	}
      else
	{
	  fprintf (stderr, "You do not have permission to access this data.\n");
	  leave (1);
	}
    }

  if (!ReadOnly && sch.datafromprog)
    {
      fprintf (stderr, "This relation gets its data from a program, so access will be readonly.\n");
      sleep (2);
      ReadOnly = TRUE;
    }

  if (!ReadOnly)
    {
      if (CheckClobberLock (&sch))
	{
	  if (!MyClobberLock (&sch))
	    {
	      fprintf (stderr, "Someone else has the lock.\n");
	      leave (1);
	    }
	}
      else
	{
	  (void) sprintf (cmdbuf, "%s %s -lock", RELMOD, sch.name);
	  if (system (cmdbuf) != 0)
	    leave (1);
	  IMadeLock = TRUE;
	}
    }

  (void) mktemp (TempFileName);

  (void) unlink(TempFileName); /* prevent attack whereby new file already
	  exists as a symlink to something we don't want to overwrite */
  fp = fopen (TempFileName, "w");
  if (fp == NULL)
    {
      (void) fprintf (stderr, "Cannot create temporary file.\n");
      leave (1);
    }

  TempFileCreated = TRUE;


  (void) fprintf (stderr, "Dumping data.\n");
  if (!DumpAsciiData (fp, &sch))
    {
      (void) fprintf (stderr, "Could not read data.\n");
      leave (1);
    }

  if (createflag)
    CreateData (fp, &sch);

  if (fclose (fp) == -1)
    {
      (void) fprintf (stderr, "Error writing to temporary file\n");
      leave (1);
    }


  (void) signal (SIGINT, SIG_IGN);
  (void) signal (SIGQUIT, SIG_IGN);
  (void) signal (SIGHUP, SIG_IGN);

  notdone = TRUE;

  ParseError = FALSE;
  while (notdone)
    {
      struct stat osb, nsb;

      if (stat (TempFileName, &osb) == -1)
	{
	  perror ("Bad stat of temporary file");
	  leave (1);
	}
      EditAsciiFile (TempFileName);

      if (stat (TempFileName, &nsb) == -1)
	{
	  perror ("Bad stat of temporary file");
	  leave (1);
	}

      if (osb.st_mtime == nsb.st_mtime && !ParseError && !createflag)
	leave (0);

      if (ReadOnly)
	{
	  (void) fprintf (stderr, "No changes saved: readonly access\n");
	  leave (0);
	}
      (void) fprintf (stderr, "Parsing your changes.\n");
      if (!GetEditedData (TempFileName, &sch))
	{
	  char buf[256];

	  ParseError = TRUE;
	  (void) printf ("E(dit) or A(bort): ");
	  if (fgets (buf, sizeof buf, stdin) == NULL)
	    leave (1);
	  if (*buf == 'A' || *buf == 'a')
	    leave (1);
	}
      else
	notdone = FALSE;
    }

  if (!TouchLastModified (&sch))
    {
      fprintf (stderr, "Could not set last modified time");
      leave (1);
    }

  if (IMadeLock)
    {
      (void) sprintf (cmdbuf, "%s %s -update", RELMOD, sch.name);
      if (system (cmdbuf) != 0)
	leave (1);
    }

  leave (0);
}

void leave (status)
     int status;
{
  char cmdbuf[2048];

  if (TempFileCreated)
    (void) unlink (TempFileName);

  if (IMadeLock)
    {
      (void) sprintf (cmdbuf, "%s %s -unlock", RELMOD, sch.name);
      if (system (cmdbuf) != 0)
	status = 1;
    }

  exit (status);
}

RETSIGTYPE
sigleave ()
{
  leave (1);
}
