
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>

#include "constants.h"
#include "db.h"
#include "errs.h"

char *UserName ();

Usage ()
{
  (void) fprintf (stderr, "Usage: tabread relationname [-c|-l] [-a attribute] ...\n");
  (void) fprintf (stderr, "	[-W lock_wait_time] [-ttabstring] [-d data_set_selector] [-w WhereExpr]\n");
  exit (1);
}

/* Formats of the printout */
#define PF_COL 1		/* each row on one line */
#define PF_LINE 2		/* each column list on it's own line */
int PrintFormat = PF_COL;

int Printed;

extern char NeedCol[MAXATTR];	/* Filled in by whereparse.y */
extern int NumColsNeeded;
static char PrintCol[MAXATTR];	/* Print the column number stored here. */
static int NumColsToPrint = 0;

char *TabString = "\t";

int HeaderFlag;

char *WhereExpr, *DataSetSelector = "";

char *Row[MAXATTR];

struct scheme sch;

int LockWait = (24 * 60 * 60);

main (argc, argv)
     int argc;
     char *argv[];
{
  int time;
  int i;

  if (argc < 2)
    Usage ();

  (void) UserName ();		/* To properly record our user name */

#ifdef NFS
  if (*argv[1] != '.' && *argv[1] != '/')	/* CHANGE raphael 6/92 */
    (void) setuid (SOMEONEOTHERTHANROOT);	/* Root cannot read over nfs */
#endif

  if (!ReadScheme (argv[1], &sch))
    {
      (void) fprintf (stderr, "Cannot read scheme for relation %s.\n", argv[1]);
      Usage ();
    }

  for (i = 2; i < argc; i++)
    {
      if (strcmp (argv[i], "-c") == 0)
	PrintFormat = PF_COL;
      else if (strcmp (argv[i], "-l") == 0)
	PrintFormat = PF_LINE;
      else if (strcmp (argv[i], "-h") == 0)
	HeaderFlag = TRUE;
      else if (strcmp (argv[i], "-a") == 0)
	{
	  int ai;
	  i++;
	  if (i == argc)
	    {
	      (void) fprintf (stderr, "Missing attribute name.\n");
	      Usage ();
	    }
	  for (ai = 0; ai < sch.numattr; ai++)
	    if (strcmp (argv[i], sch.at[ai].name) == 0)
	      break;
	  if (ai == sch.numattr)
	    {
	      (void) fprintf (stderr, "Unknown attribute: %s\n", argv[i]);
	      Usage ();
	    }
	  if (NumColsToPrint >= MAXATTR)
	    {
	      (void) fprintf (stderr,
			      "Too many attributes selected.\n");
	      exit (1);
	    }
	  PrintCol[NumColsToPrint++] = ai;
	  NeedCol[NumColsNeeded++] = ai;
	}
      else if (strncmp (argv[i], "-t", 2) == 0)
	TabString = &argv[i][2];
      else if (strcmp (argv[i], "-d") == 0)
	{
	  DataSetSelector = argv[++i];
	  if (i == argc)
	    {
	      (void) fprintf (stderr, "Missing data set argument.\n");
	      Usage ();
	    }
	}
      else if (strcmp (argv[i], "-W") == 0)
	{
	  LockWait = atoi (argv[++i]);
	  if (i == argc)
	    {
	      (void) fprintf (stderr, "Missing lock wait time.\n");
	      Usage ();
	    }
	}
      else if (strcmp (argv[i], "-w") == 0)
	{
	  WhereExpr = argv[++i];
	  if (i == argc)
	    {
	      (void) fprintf (stderr, "Missing where expression.\n");
	      Usage ();
	    }
	}
      else
	{
	  (void) fprintf (stderr,
			  *argv[i] == '-' ?
			  "Unknown option '%s'.\n" :
			  "Unexpected argument %s\n",
			  argv[i]);
	  Usage ();
	}
    }

  time = 0;
/* Negative LockWait implies ignore lock. */

/* CHANGE raphael 6/92: added !MyClobberLock in next line */
  while (LockWait >= 0 && CheckClobberLock (&sch) && !MyClobberLock (&sch))
    {
      sleep (2);
      time += 2;
      if (time >= LockWait)
	{
	  (void) fprintf (stderr,
			  "Timed out waiting for lock to be removed.\n");
	  exit (1);
	}
    }

  if (!(CheckUserPermission (UserName (), sch.reader, sch.useraliasfile) ||
	CheckUserPermission (UserName (), sch.appender, sch.useraliasfile) ||
	CheckUserPermission (UserName (), sch.modifier, sch.useraliasfile) ||
	CheckUserPermission (UserName (), sch.owner, sch.useraliasfile)))
    {
      (void) fprintf (stderr,
		      "You do not have permission to read this relation.\n");
      exit (1);
    }

  if (NumColsToPrint == 0)	/* No attributes specified implies all. */
    {
      for (i = 0; i < sch.numattr; i++)
	{
	  PrintCol[i] = i;
	  NeedCol[i] = i;
	}
      NumColsToPrint = sch.numattr;
      NumColsNeeded = sch.numattr;
    }

  if (WhereExpr)
    ParseWhere (&sch, WhereExpr);	/* Here possibly other additions will
					be made to NeedCol.  */

  if (HeaderFlag)
    {
      for (i = 0; i < NumColsToPrint; i++)
	{
	  printf ("%s", sch.at[PrintCol[i]].name);
	  if (PrintFormat == PF_COL)
	    {
	      if (i != NumColsToPrint - 1)
		(void) fputs (TabString, stdout);
	    }
	  else
	    (void) putchar ('\n');
	}
      if (PrintFormat == PF_COL)
	(void) putchar ('\n');
    }


  if (!OpenRow (&sch, NeedCol, NumColsNeeded, DataSetSelector))
    {
      (void) fprintf (stderr, "Error Initializing Data Read\n");
      exit (1);
    }

  Printed = FALSE;
  while (ReadRow (Row))
    {
      int ai;

      for (ai = 0; ai < NumColsToPrint; ai++)
	{
	  if (!CheckItem (Row[PrintCol[ai]], &sch.at[PrintCol[ai]]))
	    {
	      (void) fflush (stdout);
	      (void) fprintf (stderr,
			      "\n\n\nBad data was received!!!!\n");
	      exit (1);
	    }
	}

      if (WhereExpr && !CheckWhere (Row))
	continue;

      if (Printed && NumColsToPrint == 1 && PrintFormat == PF_COL)
	(void) fputs (TabString, stdout);

      for (ai = 0; ai < NumColsToPrint; ai++)
	{
	  switch (PrintFormat)
	    {
	    case PF_LINE:
	      if (sch.at[PrintCol[ai]].checktype == CT_BOOLEAN)
		printf ("%s\n", *Row[PrintCol[ai]] == '1' ?
			"true" : "false");
	      else
		printf ("%s\n", Row[PrintCol[ai]]);
	      break;
	    case PF_COL:
	      if (sch.at[PrintCol[ai]].checktype == CT_BOOLEAN)
		printf ("%s", *Row[PrintCol[ai]] == '1' ?
			"true" : "false");
	      else
		printf ("%s", Row[PrintCol[ai]]);
	      if (ai != NumColsToPrint - 1)
		(void) fputs (TabString, stdout);
	      break;
	    }
	}
      if (NumColsToPrint != 1 && PrintFormat == PF_COL)
	(void) putchar ('\n');
      Printed = TRUE;
    }
  if (Printed && NumColsToPrint == 1 && PrintFormat == PF_COL)
    (void) putchar ('\n');

  if (CloseRow () != GOOD)
    {
      (void) fprintf (stderr, "Error Reading Data\n");
      exit (1);
    }

  if (fclose (stdout) == EOF)
    {
      (void) fprintf (stderr, "Error Writing Data\n");
      exit (1);
    }

  exit (0);
}
