/*
edit: Interactive appointment editor

Raphael Finkel, raphael@ms.uky.edu, 1/89

*/

#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#ifdef __sparc__
#	include <strings.h>
#endif
#include "app.h"

/* menu procedures */
	MenuEntryReturnType
		/* main menu */
			MainDefault(),
			PrintGiven(), PrintKey(), NewAppt(), ModAppt(), ToggleVerbosity(),
			ToggleDiscard(), TakeAdvance(), GetOut(), Snoop(), Update(),
		/* change menu */
			ChangeDates(), ChangeTimes(), ChangeAdvanceNotice(),
			ChangeMessage(), ChangeDuringWeek(), ChangeDuringMonth(),
			ChangeDuringYear(), TakeWildCard(), TakeSecondTuesday(),
			AcceptChange(), DeleteChange(), AbortChange(),
		/* snoop menu */
			SnoopDefault();

/* constants */
	Menu MainMenu = {
		"Appointment calendar editor", 0, 0, Finalize,
		MainDefault, {
			{"Show me the calendar for a given date", 
				"[dd, mm/dd, mm/dd/yy, or one of UMTWRFS]", PrintGiven},
			{"Show me all appointments with message matching a string\n",
				"Search string (case is ignored)", PrintKey, 0},
			{"Make a new appointment", 0, NewAppt, 1},
			{"Modify an appointment\n",
				"Message string to search for (case is ignored)", ModAppt},
#		define TOGGLEITEM 4
			{0 /*Toggle verbosity*/, 0, ToggleVerbosity, 1},
#		define ADVANCEITEM 5
			{0 /*Set advance notice*/, 
				"New advance notice (negative to supress)", TakeAdvance, 1},
#		define DISCARDITEM 6
			{0 /*Toggle discard elapsed events*/, 0, ToggleDiscard, 1},
			{"Inspect someone else's calendar\n", "Whose", Snoop},
			{"Update files", 0, Update},
			{"Exit", 0, GetOut},
			{0, 0}
		}
	}; /* MainMenu */
	Menu SnoopMenu = {
		"Looking at                                                     ", 0,
		  0, 0, SnoopDefault, {
			{"Show me the calendar for a given date\n", 
				"[dd, mm/dd, mm/dd/yy, or one of UMTWRFS]", PrintGiven},
			{"Return to my own appointments", 0, GetOut},
			{0, 0}
		}
	}; /* SnoopMenu */

/* vars */
	int Disposition;
#		define DISP_OK 0
#		define DISP_ABORT -1
#		define DISP_DELETE -2
	int Changing; /* whether a change is in progress */

/*ARGSUSED*/
MenuEntryReturnType PrintGiven(Dummy, String)
char *String;
{
	DateType Date;
	int Wilds; /* number of wild specifications */

	InitializeNow(); /* in case we have crossed a date line */
	CheckDates(1);
	Wilds = InterpretDate(String, &Date);
	if (Wilds != 0) {
		fprintf(stdout, "%s\n",
			Wilds > 0 ? "Can't use wildcards here" :
			Wilds == -2 ? "Can't use default specification" :
			"Can't parse input");
		return MENU_BADARG;
	}
	PrintOneDayList(&Date);
	PauseMenu();
	return MENU_CONTINUE;
} /* PrintGiven */

/*ARGSUSED*/
MenuEntryReturnType PrintKey(Dummy, String)
char *String;
{
	AppointmentListType *HitPtr, *TmpPtr;

	CheckDates(1);
	HitPtr = ReportStringMatches(String);
	if (HitPtr->Next == HitPtr) {
		fprintf(stdout, "No matches\n");
		FreeAppointmentList(HitPtr);
		return MENU_BADARG;
	}
	for (TmpPtr = HitPtr->Next; TmpPtr != HitPtr; TmpPtr = TmpPtr->Next) {
		PrintAppointment(TmpPtr->ApptEntry, 0, 0, (DateType *) 0, stdout);
	}
	FreeAppointmentList(HitPtr);
	PauseMenu();
	return MENU_CONTINUE;
} /* PrintKey */

typedef struct {
	Menu MenuEntry;
	AppointmentPtrType ApptEntry;
	int Deletable; /* true if it is an option to delete */
	int More; /* true if there are more to come */
	} Changable;

#	define NewString() {\
		/* if (LinePtr->Description) free(LinePtr->Description); */ \
		LinePtr->Description = calloc(1, MAXLINE); \
	}

void MakeChangeMenu(ChangePtr)
Changable *ChangePtr;
/* construct a menu that displays the given appointment and allows changes */
{
	MenuLine *LinePtr;
	char Buffer[MAXLINE], Buffer1[MAXLINE];
	AppointmentPtrType AppointmentPtr = ChangePtr->ApptEntry;

	ChangePtr->MenuEntry.Name = "Would you like to change this appointment?";
	ChangePtr->MenuEntry.Initialize = ChangePtr->MenuEntry.Continue =
		ChangePtr->MenuEntry.Finalize = 0;
	LinePtr = ChangePtr->MenuEntry.Items;
	/* much of the following similar to PrintAppointment */
	{ 	/* Start date, end date */
		NewString();
		if (DateCompare(&AppointmentPtr->StartDate,
		  &AppointmentPtr->EndDate, 0) == 0) {
				PutDate(&AppointmentPtr->StartDate, Buffer, 1);
				sprintf(LinePtr->Description, "Limited to %s", Buffer);
		} else { /* start, end differ */
			if (AppointmentPtr->StartDate.Year == EARLIESTYEAR &&
			  AppointmentPtr->EndDate.Year == LATESTYEAR) {
				sprintf(LinePtr->Description, "No date limits");
			} else {
				PutDate(&AppointmentPtr->StartDate, Buffer, 1);
				PutDate(&AppointmentPtr->EndDate, Buffer1, 1);
				sprintf(LinePtr->Description, "Limited between %s and %s ",
					Buffer, Buffer1);
			}
		}
		LinePtr->Prompt = "New date limits (one or two generalized dates)";
		LinePtr->Routine = ChangeDates;
		LinePtr->Argument = (long) ChangePtr;
		LinePtr++;
	}
	{ 	/* Start time, end time */
		NewString();
		if (AppointmentPtr->StartTime == AppointmentPtr->EndTime) {
			if (AppointmentPtr->StartTime == -1) {
				sprintf(LinePtr->Description, "No time limits");
			} else {
				PutTime(AppointmentPtr->StartTime, Buffer, 1);
				sprintf(LinePtr->Description, "At %s", Buffer);
			}
		} else {
			PutTime(AppointmentPtr->StartTime, Buffer, 1);
			PutTime(AppointmentPtr->EndTime, Buffer1, 1);
			sprintf(LinePtr->Description, "Between %s and %s", Buffer, Buffer1);
		}
		LinePtr->Prompt = "New time limits (one or two of hh[[:]mm][p])";
		LinePtr->Routine = ChangeTimes;
		LinePtr->Argument = (long) ChangePtr;
		LinePtr++;
	}
	{ 	/* During week */
		NewString();
		if (AppointmentPtr->DuringWeek == AllWeekDays) {
			sprintf(LinePtr->Description, "No day-of-week restriction");
		} else {
			PutDuringWeek(AppointmentPtr->DuringWeek, Buffer);
			sprintf(LinePtr->Description, "Only on %s", Buffer);
		}
		LinePtr->Prompt = "New days of week [umtwrfs] ... ";
		LinePtr->Routine = ChangeDuringWeek;
		LinePtr->Argument = (long) ChangePtr;
		LinePtr++;
	}
	{ 	/* During year */
		NewString();
		if (AppointmentPtr->DuringYear == AllYearMonths) {
			sprintf(LinePtr->Description, "No month-in-year restriction");
		} else {
			PutDuringYear(AppointmentPtr->DuringYear, Buffer);
			sprintf(LinePtr->Description, "Only month %s", Buffer);
		}
		LinePtr->Prompt = "New months of year (like 1,4)";
		LinePtr->Routine = ChangeDuringYear;
		LinePtr->Argument = (long) ChangePtr;
		LinePtr++;
	}
	{ 	/* During month */
		NewString();
		if (AppointmentPtr->DuringMonth == AllMonthDays) {
			sprintf(LinePtr->Description, "No day-in-month restriction");
		} else {
			PutDuringMonth(AppointmentPtr->DuringMonth, Buffer);
			sprintf(LinePtr->Description, "Only on day %s", Buffer);
		}
		LinePtr->Prompt = "New days of month [eg 12,14]";
		LinePtr->Routine = ChangeDuringMonth;
		LinePtr->Argument = (long) ChangePtr;
		LinePtr++;
	}
	{ 	/* Advance notice */
		NewString();
		if (AppointmentPtr->AdvanceNotice == 0) {
			sprintf(LinePtr->Description, "No advance notice");
		} else {
			sprintf(LinePtr->Description, "With %d days advance notice",
				AppointmentPtr->AdvanceNotice);
		}
		LinePtr->Prompt = "New advance-notice interval (in days)";
		LinePtr->Routine = ChangeAdvanceNotice;
		LinePtr->Argument = (long) ChangePtr;
		LinePtr++;
	}
	{	/* message */
		NewString();
		sprintf(LinePtr->Description, "Message: %s\n", AppointmentPtr->Message);
		LinePtr->Prompt = "New message";
		LinePtr->Routine = ChangeMessage;
		LinePtr->Argument = (long) ChangePtr;
		LinePtr++;
	}
	{	/* wildcards */
		NewString();
		sprintf(LinePtr->Description, "Wildcard dates");
		LinePtr->Prompt = "Wild card input, such as 4/* or 1/*/89";
		LinePtr->Routine = TakeWildCard;
		LinePtr->Argument = (long) ChangePtr;
		LinePtr++;
	}
	{	/* Second tuesday */
		NewString();
		sprintf(LinePtr->Description, "n-th day of week every month\n");
		LinePtr->Prompt = "[12345] [UMTWRFS]";
		LinePtr->Routine = TakeSecondTuesday;
		LinePtr->Argument = (long) ChangePtr;
		LinePtr++;
	}
	{	/* disposition */
		NewString();
		sprintf(LinePtr->Description, "This appointment looks fine%s",
			ChangePtr->More ? "; go on to the next one" : "");
		LinePtr->Prompt = 0;
		LinePtr->Routine = AcceptChange;
		LinePtr++;
		if (ChangePtr->Deletable) {
			NewString();
			sprintf(LinePtr->Description, "Delete this appointment%s",
				ChangePtr->More ? "; go on to the next one" : "");
			LinePtr->Prompt = 0;
			LinePtr->Routine = DeleteChange;
			LinePtr++;
		} /* Deletable */
		NewString();
		sprintf(LinePtr->Description, "Don't %s this appointment%s",
			ChangePtr->Deletable ? "modify" : "add",
			ChangePtr->More ? ", and ignore all the other matches" : "");
		LinePtr->Prompt = 0;
		LinePtr->Routine = AbortChange;
		LinePtr++;
	}
	LinePtr->Description = 0;
} /* MakeChangeMenu */

MenuEntryReturnType AcceptChange(void)
{
	Disposition = DISP_OK;
	return MENU_BREAK;
} /* AcceptChange */

MenuEntryReturnType DeleteChange(void)
{
	Disposition = DISP_DELETE;
	return MENU_BREAK;
} /* AcceptChange */

MenuEntryReturnType AbortChange(void)
{
	Disposition = DISP_ABORT;
	return MENU_BREAK;
} /* AbortChange */

MenuEntryReturnType NewAppt(FromMenu, String)
char *String;
{
	Changable ChangeThis;
	DoMenuReturnType Result;
	char *CharPtr;
	char *BadMsg = "syntax:  =d date(s) =a advance =t time(s) =m message\n";

#	define SkipWhite() { \
		for (; isspace(*CharPtr); CharPtr++); \
	}
#	define SkipBlackWhite() { \
		for (; !isspace(*CharPtr); CharPtr++); \
		SkipWhite(); \
	}
	
	InitializeNow(); /* in case we have crossed a date line */
	CheckDates(1);
	bzero((char *) &ChangeThis, sizeof(ChangeThis));
	ChangeThis.ApptEntry = DefaultAppointment();
	ChangeThis.Deletable = ChangeThis.More = 0;
	if (*String) { /* parse the input string */
		CharPtr = String;
		while (*CharPtr) { /* treat arguments */
			if (strncmp(CharPtr, "=m", 2) == 0) {
				CharPtr += 2;
				SkipWhite();
				(void) strncpy(ChangeThis.ApptEntry->Message, CharPtr, MAXLINE);
				CharPtr = CharPtr + strlen(CharPtr); /* to end */
			} else if (strncmp(CharPtr, "=d", 2) == 0) {
				CharPtr += 2;
				SkipWhite();
				Result = InterpretDate(CharPtr,
					&ChangeThis.ApptEntry->StartDate);
				if (Result != 0) {
					fprintf(stdout, "Can't parse %s\n%s", CharPtr, BadMsg);
					if (FromMenu) PauseMenu();
					return MENU_BADARG;
				}
				SkipBlackWhite();
				if (*CharPtr && *CharPtr != '=') { /* a second date */
					Result = InterpretDate(CharPtr,
						&ChangeThis.ApptEntry->EndDate);
					if (Result != 0) {
						fprintf(stdout, "Can't parse second date\n%s", BadMsg);
						if (FromMenu) PauseMenu();
						return MENU_BADARG;
					}
					SkipBlackWhite();
				} else { /* no second date */
					ChangeThis.ApptEntry->EndDate =
						ChangeThis.ApptEntry->StartDate;
				}
			} else if (strncmp(CharPtr, "=t", 2) == 0) {
				CharPtr += 2;
				SkipWhite();
				ChangeThis.ApptEntry->StartTime = ParseTime(&CharPtr);
				SkipWhite();
				if (isdigit(*CharPtr)) {
					ChangeThis.ApptEntry->EndTime = ParseTime(&CharPtr);
					SkipWhite();
				} else {
					ChangeThis.ApptEntry->EndTime = 
						ChangeThis.ApptEntry->StartTime;
				}
				if (ChangeThis.ApptEntry->StartTime >
				  ChangeThis.ApptEntry->EndTime &&
				  ChangeThis.ApptEntry->EndTime != -1) {
					fprintf(stdout, "Times are out of order\n");
					if (FromMenu) PauseMenu();
					return MENU_BADARG;
				}
			} else if (strncmp(CharPtr, "=a", 2) == 0) {
				CharPtr += 2;
				SkipWhite();
				ChangeThis.ApptEntry->AdvanceNotice = atoi(CharPtr);
				SkipBlackWhite();
			} else {
				fprintf(stdout, "Can't parse %s\n%s", CharPtr, BadMsg);
				if (FromMenu) PauseMenu();
				return MENU_BADARG;
			}
		} /* treat arguments */
		MakeChangeMenu(&ChangeThis);
		Result = DoMenu(&(ChangeThis.MenuEntry));
		if (Result == MENU_BROKEN && Disposition == DISP_OK) {
			AddAppointment(ChangeThis.ApptEntry);
			Changed = 1;
		} else {
			free((char *) ChangeThis.ApptEntry);
		}
		FreeMenu(&ChangeThis.MenuEntry);
		return MENU_CONTINUE;
	} /* parse input string, if any */
	MakeChangeMenu(&ChangeThis);
	Result = DoMenu(&(ChangeThis.MenuEntry));
	if (Result == MENU_BROKEN && Disposition == DISP_OK) {
		AddAppointment(ChangeThis.ApptEntry);
		Changed = 1;
	} else {
		free((char *) ChangeThis.ApptEntry);
	}
	FreeMenu(&ChangeThis.MenuEntry);
	return MENU_CONTINUE;
} /* NewAppt */

/*ARGSUSED*/
MenuEntryReturnType ModAppt(Dummy, String)
char *String;
{
	Changable ChangeThis;
	DoMenuReturnType Result;
	AppointmentListType *HitPtr, *TmpPtr;
	
	InitializeNow(); /* in case we have crossed a date line */
	CheckDates(1);
	bzero((char *) &ChangeThis, sizeof(ChangeThis));
	HitPtr = ReportStringMatches(String);
	if (HitPtr->Next == HitPtr) { /* no matches */
		fprintf(stdout, "No matches\n");
		FreeAppointmentList(HitPtr);
		return MENU_BADARG;
	}
	for (TmpPtr = HitPtr->Next; TmpPtr != HitPtr; TmpPtr = TmpPtr->Next) {
		/* display and allow changes for TmpPtr->ApptEntry */
		ChangeThis.ApptEntry =
			(AppointmentPtrType) calloc(1, sizeof(AppointmentType));
		bcopy((char *) TmpPtr->ApptEntry, (char *) ChangeThis.ApptEntry,
			sizeof(AppointmentType));
		ChangeThis.Deletable = 1;
		ChangeThis.More = (TmpPtr->Next != HitPtr);
		Changing = 0;
		MakeChangeMenu(&ChangeThis);
		Result = DoMenu(&(ChangeThis.MenuEntry));
		if (Result == MENU_QUIT) {
			Disposition = DISP_ABORT;
			Result = MENU_BROKEN;
		}
		if (Result != MENU_BROKEN) {
			fprintf(stdout, "Can't figure out return value\n");
			PauseMenu();
		}
		if (Disposition == DISP_OK) {
			if (Changing) {
				AddAppointment(ChangeThis.ApptEntry);
				DeleteAppointment(TmpPtr->ApptEntry);
				Changed = 1;
				Changing = 0;
			}
		}
		else if (Disposition == DISP_DELETE) {
			free((char *) ChangeThis.ApptEntry);
			DeleteAppointment(TmpPtr->ApptEntry);
			Changed = 1;
		} else { /* Disposition == DISP_ABORT */
			free((char *) ChangeThis.ApptEntry);
			break;
		}
		FreeMenu(&ChangeThis.MenuEntry);
	} /* for TmpPtr */
	FreeAppointmentList(HitPtr);
	return MENU_CONTINUE;
} /* ModAppt */

MenuEntryReturnType ChangeDates(ChangePtr, String)
Changable *ChangePtr;
char *String;
{
	DateType Date;
	int Wilds;
	Wilds = InterpretDate(String, &Date);
	switch (Wilds) {
		case -2:
			Date.Year = EARLIESTYEAR;
			ChangePtr->ApptEntry->StartDate = Date;
			break;
		case -1: case 1: case 2: case 3:
			return MENU_BADARG;
		case 0:
			ChangePtr->ApptEntry->StartDate = Date;
			break;
	} /* switch Wilds */
	while (*String && !isspace(*String)) String++;
	while (isspace(*String)) String++;
	if (!*String) { /* no second part */
		ChangePtr->ApptEntry->EndDate = ChangePtr->ApptEntry->StartDate;
		if (ChangePtr->ApptEntry->EndDate.Year == EARLIESTYEAR)
				ChangePtr->ApptEntry->EndDate.Year = LATESTYEAR;
	} else { /* try for the second spec */
		Wilds = InterpretDate(String, &Date);
		switch (Wilds) {
			case -2:
				Date.Year = LATESTYEAR;
				ChangePtr->ApptEntry->EndDate = Date;
				break;
			case -1: case 1: case 2: case 3:
				return MENU_BADARG;
			case 0:
				ChangePtr->ApptEntry->EndDate = Date;
				break;
		}
		if (DateCompare(&ChangePtr->ApptEntry->EndDate,
		  &ChangePtr->ApptEntry->StartDate, 0) < 0)
			return MENU_BADARG;
	} /* second spec */
	Changing = 1;
	MakeChangeMenu(ChangePtr);
	return MENU_CONTINUE;
} /* ChangeDates */

MenuEntryReturnType TakeWildCard(ChangePtr, String)
Changable *ChangePtr;
char *String;
{
	DateType Date;
	int Wilds;
	Wilds = InterpretDate(String, &Date);
	ChangePtr->ApptEntry->DuringMonth = AllMonthDays;
	ChangePtr->ApptEntry->DuringYear = AllYearMonths;
	switch (Wilds) {
		case -2: case -1:
			return MENU_BADARG;
		case 0: case 1: case 2: case 3:
			if (Date.Month == -1) {
			} else {
				ChangePtr->ApptEntry->DuringYear = 1 << Date.Month;
			}
			if (Date.Day == -1) {
			} else {
				ChangePtr->ApptEntry->DuringMonth = 1 << Date.Day;
			}
			if (Date.Year == -1) { /* no date limits */
				Date.Day = 1;
				Date.Month = 1;
				Date.Year = EARLIESTYEAR;
				ChangePtr->ApptEntry->StartDate = Date;
				Date.Year = LATESTYEAR;
				ChangePtr->ApptEntry->EndDate = Date;
			} else { /* limited to one year only */
				Date.Day = 1;
				Date.Month = 1;
				ChangePtr->ApptEntry->StartDate = Date;
				Date.Day = 31;
				Date.Month = 12;
				ChangePtr->ApptEntry->EndDate = Date;
			}
			break;	
	} /* switch Wilds */
	Changing = 1;
	MakeChangeMenu(ChangePtr);
	return MENU_CONTINUE;
} /* TakeWildCard */

MenuEntryReturnType TakeAdvance(FromMenu, String)
int FromMenu;
char *String;
{
	MenuLine *LinePtr;
	DefaultAdvanceNotice = atoi(String);
	if (DefaultAdvanceNotice < 0) DefaultAdvanceNotice = -1;
	LinePtr = &MainMenu.Items[ADVANCEITEM];
	NewString();
	if (DefaultAdvanceNotice >= 0) {
		sprintf(LinePtr->Description,
			"Change default advance notice (currently %d)",
			DefaultAdvanceNotice);
		if (FromMenu)
			printf("%d day%s advance notice\n", DefaultAdvanceNotice,
				DefaultAdvanceNotice == 1 ? "" : "s");
	} else {
		sprintf(LinePtr->Description,
			"Change default advance notice (currently none)");
		if (FromMenu) printf("No advance notice\n");
	}
	if (FromMenu) PauseMenu();
	return MENU_CONTINUE;
} /* TakeAdvance */

/*ARGSUSED*/
MenuEntryReturnType GetOut(Dummy, String)
char *String;
{
	CheckDates(1);
	return MENU_BREAK;
} /* GetOut */

extern time_t ModifyTime;

/*ARGSUSED*/
MenuEntryReturnType Snoop(Dummy, String)
char *String;
/* look at someone else's appointments */
{
	struct passwd *PasswdPtr;
	time_t OldModifyTime;
	char OldUser[MAXLINE];
	char Name[MAXLINE]; /* New user's true name */
	if ((PasswdPtr = getpwnam(String)) == 0) {
		fprintf(stdout, "No such user: %s\n", String);
		return MENU_BADARG;
	}
	(void) strncpy(OldUser, User, sizeof(User));
	(void) strncpy(User, PasswdPtr->pw_dir, sizeof(User));
	SnoopAppointments.Next = SnoopAppointments.Prev = &SnoopAppointments;
	ActiveAppointments = &SnoopAppointments;
	ReadAppointments(&SnoopAppointments,0);
	OldModifyTime = ModifyTime;
	CheckDates(0);
	if (islower(*String)) { /*  capitalize initial letter of name */
		*String = _toupper(*String);
	}
	{ /* file Name */
		char *InPtr, *OutPtr, *TmpPtr;
		for (InPtr = PasswdPtr->pw_gecos, OutPtr = Name;
		  *InPtr && (*InPtr != ',');
		){
			if (*InPtr == '&') { /* substitute capitalized login name */
				InPtr++; /* past the '&' */
				for (TmpPtr = String; *TmpPtr; ) {
					*OutPtr++ = *TmpPtr++;
				}
			} else {
				*OutPtr++ = *InPtr++;
			}
		}
		*OutPtr = 0;
	} /* file name */
	if (SnoopAppointments.Next == &SnoopAppointments) { /* empty */
		fprintf(stdout, "%s has no appointments\n", Name);
		(void) strncpy(User, OldUser, sizeof(User));
		ModifyTime = OldModifyTime;
		return MENU_BADARG;
	} else {
		sprintf(SnoopMenu.Name+11, "%s%s appointments", Name,
			String[strlen(Name)-1] == 's' ? "'" : "'s");
		(void) DoMenu(&SnoopMenu);
		ActiveAppointments = &MyAppointments;
		ModifyTime = OldModifyTime;
		(void) strncpy(User, OldUser, sizeof(User));
	}
	return MENU_CONTINUE;
} /* Snoop */

/*ARGSUSED*/
MenuEntryReturnType Update(Dummy, String)
char *String;
/* write out changes */
{
	CheckDates(1);
	if (Changed) {
		WriteAppointments();
		ReadAppointments(&MyAppointments,1);
		Changed = 0;
		fprintf(stdout, "Updated\n");
	} else {
		fprintf(stdout, "No changes; not updated\n");
	}
	PauseMenu();
	return MENU_CONTINUE;
} /* Update */

MenuEntryReturnType TakeSecondTuesday(ChangePtr, String)
Changable *ChangePtr;
char *String;
{
	int Ordinal;
	char Day[10], *DayPtr;
	DateType Date;
	if (sscanf(String, " %d %[UMTWRFSumtwrfs]", &Ordinal, Day) != 2)
		return MENU_BADARG;
	String = 0;
	ChangePtr->ApptEntry->DuringWeek = 0;
	for (DayPtr = Day; *DayPtr; DayPtr++) { /* for each day specified */
		switch (*DayPtr) {
			case 'U': case 'u':
				ChangePtr->ApptEntry->DuringWeek |= Sunday; break;
			case 'M': case 'm':
				ChangePtr->ApptEntry->DuringWeek |= Monday; break;
			case 'T': case 't':
				ChangePtr->ApptEntry->DuringWeek |= Tuesday; break;
			case 'W': case 'w':
				ChangePtr->ApptEntry->DuringWeek |= Wednesday; break;
			case 'R': case 'r':
				ChangePtr->ApptEntry->DuringWeek |= Thursday; break;
			case 'F': case 'f':
				ChangePtr->ApptEntry->DuringWeek |= Friday; break;
			case 'S': case 's':
				ChangePtr->ApptEntry->DuringWeek |= Saturday; break;
			default: return MENU_BADARG;
		} /* switch *DayPtr */
	} /* for each day specified */
	ChangePtr->ApptEntry->DuringMonth = 0;
	for (; Ordinal; Ordinal /= 10) { /* for each digit of ordinal */
		switch (Ordinal % 10) {
			case 1: ChangePtr->ApptEntry->DuringMonth |= 0376; break;
			case 2: ChangePtr->ApptEntry->DuringMonth |= 0376 << 7; break;
			case 3: ChangePtr->ApptEntry->DuringMonth |= 0376 << 14; break;
			case 4: ChangePtr->ApptEntry->DuringMonth |= 0376 << 21; break;
			case 5: ChangePtr->ApptEntry->DuringMonth |= 0376 << 28; break;
			default: return MENU_BADARG;
		} /* switch digit */
	} /* for each digit */
	Date.Month = Date.Day = 1;
	Date.Year = EARLIESTYEAR;
	ChangePtr->ApptEntry->StartDate = Date;
	Date.Year = LATESTYEAR;
	ChangePtr->ApptEntry->EndDate = Date;
	Changing = 1;
	MakeChangeMenu(ChangePtr);
	return MENU_CONTINUE;
} /* TakeSecondTuesday */

MenuEntryReturnType ChangeTimes(ChangePtr, String)
Changable *ChangePtr;
char *String;
{
	TimeType Start, End;
	Start = ParseTime(&String);
	End = ParseTime(&String);
	if (Start > End && End != -1) {
		fprintf(stdout, "Times are out of order\n");
		return MENU_BADARG;
	}
	ChangePtr->ApptEntry->StartTime = Start;
	ChangePtr->ApptEntry->EndTime = End;
	Changing = 1;
	MakeChangeMenu(ChangePtr);
	return MENU_CONTINUE;
} /* ChangeTimes */

MenuEntryReturnType ChangeAdvanceNotice(ChangePtr, String)
Changable *ChangePtr;
char *String;
{
	int New;
	New = atoi(String);
	if (New < 0) return MENU_BADARG;
	ChangePtr->ApptEntry->AdvanceNotice = New;
	Changing = 1;
	MakeChangeMenu(ChangePtr);
	return MENU_CONTINUE;
} /* ChangeAdvanceNotice */

MenuEntryReturnType ChangeDuringWeek(ChangePtr, String)
Changable *ChangePtr;
char *String;
{
	ChangePtr->ApptEntry->DuringWeek = ParseDuringWeek(&String);
	Changing = 1;
	MakeChangeMenu(ChangePtr);
	return MENU_CONTINUE;
} /* ChangeDuringWeek */

MenuEntryReturnType ChangeDuringMonth(ChangePtr, String)
Changable *ChangePtr;
char *String;
{
	ChangePtr->ApptEntry->DuringMonth = ParseDuringMonth(&String);
	Changing = 1;
	MakeChangeMenu(ChangePtr);
	return MENU_CONTINUE;
} /* ChangeDuringMonth */

MenuEntryReturnType ChangeDuringYear(ChangePtr, String)
Changable *ChangePtr;
char *String;
{
	ChangePtr->ApptEntry->DuringYear = ParseDuringYear(&String);
	Changing = 1;
	MakeChangeMenu(ChangePtr);
	return MENU_CONTINUE;
} /* ChangeDuringYear */

MenuEntryReturnType ChangeMessage(ChangePtr, String)
Changable *ChangePtr;
char *String;
{
	char OldMessage[MAXLINE], *SourcePtr, *DestPtr, *TempPtr;
	int Quote; /* 1 if we just saw '\' */

	(void) strncpy(OldMessage, ChangePtr->ApptEntry->Message, MAXLINE);
	SourcePtr = String;
	DestPtr = ChangePtr->ApptEntry->Message;
	Quote = 0;
	while (*SourcePtr) {
		if (*SourcePtr == '\\') {
			if (Quote) {
				*DestPtr++ = '\\';
				Quote = 0;
			} else {
				Quote = 1;
			}
			SourcePtr++;
		} else if (*SourcePtr == '&' && !Quote) { /* insert original message */
			TempPtr = OldMessage;
			while (*TempPtr) *DestPtr++ = *TempPtr++;
			SourcePtr++;
		} else {
			*DestPtr++ = *SourcePtr++;
			Quote = 0;
		}
	} /* while *SourcePtr */
	*DestPtr++ = 0;
	Changing = 1;
	MakeChangeMenu(ChangePtr);
	return MENU_CONTINUE;
} /* ChangeMessage */

void FreeMenu(MenuPtr)
Menu *MenuPtr;
/* reclaim all the strings */
{
	int seq;
#	define FreeIt(a) {if (a) free(a);}
	/* better to leak memory here than to mis-free a static string! */ return;
	FreeIt(MenuPtr->Name);
	for (seq = 0; seq < ITEMSLIMIT; seq++){
		FreeIt(MenuPtr->Items[seq].Description);
	}
} /* FreeMenu */

MenuEntryReturnType ToggleVerbosity(FromMenu)
int FromMenu;
{
	MenuLine *LinePtr;
	Verbose = 1 - Verbose;
	LinePtr = &MainMenu.Items[TOGGLEITEM];
	NewString();
	sprintf(LinePtr->Description, "%s", Verbose ?
		"Short output (currently verbose)" :
		"Verbose output (currently short)");
	if (FromMenu) {
		fprintf(stdout, "%s\n", Verbose ? "Verbose output" : "Short output");
		PauseMenu();
	}
	return MENU_CONTINUE;
} /* ToggleVerbosity */

MenuEntryReturnType ToggleDiscard(FromMenu)
int FromMenu;
{
	MenuLine *LinePtr;
	DiscardRule = 1 - DiscardRule;
	LinePtr = &MainMenu.Items[DISCARDITEM];
	NewString();
	sprintf(LinePtr->Description, "%s\n", DiscardRule ?
		"Don't discard elapsed events (we currently do)" :
		"Discard elapsed events (currently not)" );
	if (FromMenu) {
		fprintf(stdout, "%s\n", DiscardRule ? "Will discard elapsed events" :
			"Won't discard elapsed events");
		PauseMenu();
	}
	return MENU_CONTINUE;
} /* ToggleDiscard */

MenuEntryReturnType MainDefault(String)
char *String;
{
	DateType DateDummy;
	AppointmentListType *HitPtr;

	if (strlen(String) == 1 && *String == 'q') return GetOut(0, String);
	if (getpwnam(String) != 0) return Snoop(0, String);
	if (InterpretDate(String, &DateDummy) == 0) return PrintGiven(0, String);
	HitPtr = ReportStringMatches(String);
	if (HitPtr->Next != HitPtr) { /* matches */
		FreeAppointmentList(HitPtr);
		return PrintKey(0, String);
	}
	FreeAppointmentList(HitPtr);
	if (*String == '=') {
		(void) NewAppt(1, String);
		return MENU_CONTINUE;
	}
	return MENU_BADARG;
} /* Default */

MenuEntryReturnType SnoopDefault(String)
char *String;
{
	if (strlen(String) == 1 && *String == 'q') return GetOut(0, String);
	else return PrintGiven(0, String);
}

void InitializeEdit(void)
{
	void srandom();
	long time();
	char Buffer[MAXLINE];
	(void) strncpy(User, getenv("HOME"), sizeof(User));
	InitializeNow();
	/* (void) sprintf((MainMenu.Name)+28, "%s", Version); */
	ActiveAppointments = &MyAppointments;
	MyAppointments.Next = MyAppointments.Prev = &MyAppointments;
	ReadAppointments(&MyAppointments,1);
	srandom((unsigned) time((long *) 0));
	(void) ToggleVerbosity(0); /* to put in main menu */
	(void) ToggleVerbosity(0); /* don't really want to toggle */
	sprintf(Buffer, "%d", DefaultAdvanceNotice);
	(void) TakeAdvance(0,Buffer); /* to put in main menu */
	(void) ToggleDiscard(0); /* to put in main menu */
	(void) ToggleDiscard(0); /* don't really want to toggle */
} /* Initialize */

int Finalize(void) {
	char Buffer[MAXLINE];
	if (Changed) {
		for (;;) { /* until good response */
			fprintf(stdout, "Write out changes [w] or abort [a]?  ");
			(void) fgets(Buffer, MAXLINE, stdin);
			if (strlen(Buffer) != 2 || Buffer[1] != '\n') continue;
			switch (*Buffer) {
				case 'w':
					(void) WriteAppointments();
					break;
				case 'a':
					break;
				default:
					continue;
			} /* switch */
			break;
		} /* until good response */
	} /* changed */
	return(0);
} /* Finalize */

void DoEdit(void){
	(void) DoMenu(&MainMenu);
} /* DoIt */

#ifdef undef

main(void){
	InitializeEdit();
	DoEdit();
} /* main */

#endif
