/*
parse.c

Raphael Finkel, raphael@ms.uky.edu, 1/89
*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "app.h"

/* shorthands */
#	define SkipWhite()\
		{ while (**CharPtrPtr && isspace(**CharPtrPtr)) (*CharPtrPtr)++; }

#	define SkipToWhite()\
		{ while (**CharPtrPtr && !isspace(**CharPtrPtr)) (*CharPtrPtr)++; }

/* procedures */

DateType ParseDate(char **CharPtrPtr)
                  
/* convert the ascii pointed to by CharPtrPtr into a DateType.  Advance over
blanks if needed.   Return an advanced CharPtrPtr. */
{
	DateType Answer;
	SkipWhite();
	Answer.Month = 1;
	Answer.Day = 1;
	if (strncmp(*CharPtrPtr, FromEver, 2) == 0) {
		(*CharPtrPtr) += strlen(FromEver);
		Answer.Year = EARLIESTYEAR;
		return(Answer);
	} else if (strncmp(*CharPtrPtr, ToEver, 2) == 0) {
		(*CharPtrPtr) += strlen(ToEver);
		Answer.Year = LATESTYEAR;
		return(Answer);
	}
	if (sscanf(*CharPtrPtr, "%d/%d/%d", &Answer.Month, &Answer.Day,
		&Answer.Year) != 3) {
		fprintf(stderr, "Could not read date %.8s on line %d.\n", *CharPtrPtr,
			LineNumber);
		exit(1);
	}
	Answer.Year = Standard(Answer.Year);
	if (Answer.Month < 1 || Answer.Month > 12 || Answer.Day < 1 ||
	  Answer.Day > DaysInMonth(Answer.Month, Answer.Year) ) {
		fprintf(stderr, "Bad date specification: %.8s on line %d.\n",
			*CharPtrPtr, LineNumber);
			exit(1);
	}
	SkipToWhite();
	return(Answer);
} /* ParseDate */

TimeType ParseTime(char **CharPtrPtr)
                  
/* convert the ascii pointed to by CharPtrPtr into a TimeType.  Advance over
blanks if needed.   Return an advanced CharPtrPtr.
Accepts hh[[:]mm][p]
*/
{
	int Hour, Minute;
	TimeType Answer;
	SkipWhite();
	if (**CharPtrPtr == '-') {
		(*CharPtrPtr)++;
		return(-1);
	}
	switch (sscanf(*CharPtrPtr, "%d:%d", &Hour, &Minute)) {
		case 0:
			return(-1);
		case 1:
			if (Hour > 24) {
				Minute = Hour % 100;
				Hour = Hour / 100;
			} else {
				Minute = 0;
			}
		default: ;
	}
	if (Hour < 0 || Hour > 23 || Minute < 0 || Minute > 59) {
		fprintf(stderr, "Bad time specification: %.5s, line %d.\n",
			*CharPtrPtr, LineNumber);
			return(-1);
	}
	Answer = 60*Hour + Minute;
	SkipToWhite();
	if (*(*CharPtrPtr-1) == 'p' && Hour < 12) Answer += 12 * 60;
	else if (*(*CharPtrPtr-1) == 'a' && Hour == 12) Answer -= 12 * 60;
	return(Answer);
} /* ParseTime */

DuringWeekType ParseDuringWeek(char **CharPtrPtr)
                  
/* convert the ascii pointed to by CharPtrPtr into a DuringWeekType.  Advance
over blanks if needed.   Return an advanced CharPtrPtr. */
{
	DuringWeekType Answer;
	char WeekString[MAXFIELD], *WeekStringPtr;
	SkipWhite();
	if (**CharPtrPtr == '-') {
		(*CharPtrPtr)++;
		return((DuringWeekType) AllWeekDays);
	}
	if (sscanf(*CharPtrPtr, "%s", WeekString) != 1) {
		fprintf(stderr, "Could not read days of week on line %d.\n",
			LineNumber);
		return((DuringWeekType) AllWeekDays);
	}
	Answer = 0;
	for (WeekStringPtr = WeekString; *WeekStringPtr; WeekStringPtr++) {
		switch (*WeekStringPtr) {
			case 'u': case 'U': Answer += Sunday; break;
			case 'm': case 'M': Answer += Monday; break;
			case 't': case 'T': Answer += Tuesday; break;
			case 'w': case 'W': Answer += Wednesday; break;
			case 'r': case 'R': Answer += Thursday; break;
			case 'f': case 'F': Answer += Friday; break;
			case 's': case 'S': Answer += Saturday; break;
			default:
				fprintf(stderr, "Bad day of week specifier: %c, line %d.\n",
					*WeekStringPtr, LineNumber);
		} /* switch */
	} /* for the entire string */
	SkipToWhite();
	return(Answer);
} /* ParseDuringWeek */

DuringMonthType ParseDuringMonth(char **CharPtrPtr)
                  
/* convert the ascii pointed to by CharPtrPtr into a DuringMonthType.  Advance
over blanks if needed.   Return an advanced CharPtrPtr. */
{
	DuringMonthType Answer;
	int Accumulator;
	char MonthString[MAXFIELD], *MonthStringPtr;
	SkipWhite();
	if (**CharPtrPtr == '-') {
		(*CharPtrPtr)++;
		return(AllMonthDays);
	}
	if (sscanf(*CharPtrPtr, "%[0123456789,]", MonthString) != 1) {
		fprintf(stderr, "Could not read days of month on line %d.\n",
			LineNumber);
		return(AllMonthDays);
	}
	Answer = 0;
	Accumulator = 0;
	for (MonthStringPtr = MonthString; *MonthStringPtr; MonthStringPtr++) {
		if ('0' <= *MonthStringPtr && *MonthStringPtr <= '9') {
			Accumulator = 10*Accumulator + *MonthStringPtr - '0';
		} else { /* comma */
			if (0 < Accumulator && Accumulator <= 31)
				Answer |= 1 << Accumulator;
			Accumulator = 0;
		}
	} /* for the entire string */
	if (0 < Accumulator && Accumulator <= 31)
		Answer |= 1 << Accumulator;
	SkipToWhite();
	return(Answer ? Answer : AllMonthDays);
} /* ParseDuringMonth */

DuringYearType ParseDuringYear(char **CharPtrPtr)
                  
/* convert the ascii pointed to by CharPtrPtr into a DuringYearType.  Advance
over blanks if needed.   Return an advanced CharPtrPtr. */
{
	DuringYearType Answer;
	int Accumulator;
	char YearString[MAXFIELD], *YearStringPtr;
	SkipWhite();
	if (**CharPtrPtr == '-') {
		(*CharPtrPtr)++;
		return(AllYearMonths);
	}
	if (sscanf(*CharPtrPtr, "%[0123456789,]", YearString) != 1) {
		fprintf(stderr, "Could not read months of year on line %d.\n",
			LineNumber);
		return(AllYearMonths);
	}
	Answer = 0;
	Accumulator = 0;
	for (YearStringPtr = YearString; *YearStringPtr; YearStringPtr++) {
		if ('0' <= *YearStringPtr && *YearStringPtr <= '9') {
			Accumulator = 10*Accumulator + *YearStringPtr - '0';
		} else { /* comma */
			if (0 < Accumulator && Accumulator <= 12)
				Answer |= 1 << Accumulator;
			Accumulator = 0;
		}
	} /* for the entire string */
	if (0 < Accumulator && Accumulator <= 12)
		Answer |= 1 << Accumulator;
	SkipToWhite();
	return(Answer ? Answer : AllYearMonths);
} /* ParseDuringYear */

int ParseAdvanceNotice(char **CharPtrPtr)
                  
/* convert the ascii pointed to by CharPtrPtr into an int.  Advance
over blanks if needed.   Return an advanced CharPtrPtr. */
{
	int Answer;
	SkipWhite();
	if (sscanf(*CharPtrPtr, "%d", &Answer) != 1) {
		fprintf(stderr, "Could not read advance notice count on line %d.\n",
			LineNumber);
		exit(1);
	}
	if (Answer < 0) {
		fprintf(stderr, "Invalid advance notice count %d on line %d.\n",
			Answer, LineNumber);
			exit(1);
	}
	SkipToWhite();
	return(Answer);
} /* ParseAdvanceNotice */

int ParseMessageNumber(char **CharPtrPtr)
                  
/* convert the ascii pointed to by CharPtrPtr into an int.  Advance
over blanks if needed.   Return an advanced CharPtrPtr. */
{
	int Answer;
	SkipWhite();
	if (sscanf(*CharPtrPtr, "%d", &Answer) != 1) {
		fprintf(stderr, "Could not read message number on line %d.\n",
			LineNumber);
		exit(1);
	}
	if (Answer < 0) {
		fprintf(stderr, "Invalid message number %d on line %d.\n",
			Answer, LineNumber);
			exit(1);
	}
	SkipToWhite();
	return(Answer);
} /* ParseMessageNumber */

void ParseLine(char *Buffer, AppointmentPtrType AppointmentPointer)
             
                                      
/* Parse the buffer into AppointmentPtr.
Print an error and terminate if the file is malformed. */
{
	AppointmentPointer->StartDate = ParseDate(&Buffer);
	AppointmentPointer->EndDate = ParseDate(&Buffer);
	if (DateCompare(&AppointmentPointer->EndDate,
	  &AppointmentPointer->StartDate, 0) < 0) {
		fprintf(stderr, "Starting and ending dates out of order, line %d.\n",
			LineNumber);
		exit(1);
	}
	AppointmentPointer->StartTime = ParseTime(&Buffer);
	AppointmentPointer->EndTime = ParseTime(&Buffer);
	if (AppointmentPointer->EndTime < AppointmentPointer->StartTime &&
		AppointmentPointer->EndTime != -1) {
		fprintf(stderr, "Starting and ending times out of order, line %d.\n",
			LineNumber);
		exit(1);
	}
	AppointmentPointer->DuringWeek = ParseDuringWeek(&Buffer);
	AppointmentPointer->DuringMonth = ParseDuringMonth(&Buffer);
	AppointmentPointer->DuringYear = ParseDuringYear(&Buffer);
	AppointmentPointer->AdvanceNotice = ParseAdvanceNotice(&Buffer);
	AppointmentPointer->MessageNumber = ParseMessageNumber(&Buffer);
} /* ParseLine */
