/*
periodic.c

Raphael Finkel 2/89

Periodic warnings of things about to happen
*/

#include <stdio.h>
#include <signal.h>
#include <time.h>
#include <sys/file.h>
#include <sys/types.h>
#include <fcntl.h>
#include <pwd.h>
#include <utmp.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#ifdef __sparc__
#	include <strings.h>
#endif
#include "app.h"

/* consts */
#	define SLEEPINTERVAL (60*60)

/* vars */
	int SecondsNow;
	int Resetting;
	int CheckFlag;
	char Mailer[40];

AppointmentListType *WhatNext()
/* returns a list of all MyAppointments at least AdvanceSeconds in the future
that start at the same time that start the soonest.  Returns NULL if no more
appointments today.  Does not consider appointments in future days with advance
notice that requires they be seen today.
*/
{
	AppointmentListType *Answer, *Ptr, *Tmp;
	int EarliestTime;
	
	Answer = ReportDateMatches(DateNow);
	if (Answer->Next == Answer) return NULL;
	SortAppointmentList(Answer);
	{ /* remove items that have already happened */
		for (Ptr = Answer->Next;
		  Ptr != Answer &&
		  60*Ptr->ApptEntry->StartTime <= SecondsNow+AdvanceSeconds; ) {
			Ptr = Ptr->Next;
			free((char *)Ptr->Prev);
		}
		Answer->Next = Ptr;
		Ptr->Prev = Answer;
		if (Answer->Next == Answer) return NULL;
	} /* remove expired items */
	{ /* remove items after the first ones */
		for (Ptr = Ptr->Next, EarliestTime = Ptr->Prev->ApptEntry->StartTime;
		  Ptr != Answer && Ptr->ApptEntry->StartTime == EarliestTime;
		  Ptr = Ptr->Next) ; /* move Ptr to after the ones we want */
		if (Ptr != Answer) { /* must delete from Ptr on */
			Ptr = Ptr->Prev; /* hold on to last good one */
			for (Tmp = Ptr->Next; Tmp != Answer; ) {
				Tmp = Tmp->Next;
				free((char *)Tmp->Prev);
			}
			Answer->Prev = Ptr;
			Ptr->Next = Answer;
		} /* must delete remainder */
	} /* remove items after first ones */
	return(Answer);
} /* WhatNext */

char *MyTty, *MyName;

void CheckTty()
/* exit if our tty is gone. */
{
	struct utmp ubuf;
	int ufd;
	ufd = open("/etc/utmp", O_RDONLY);
	if (ufd <= 0) {
		perror("/etc/utmp");
		exit(1);
	}
	for (;;) { /* read entries until we know if we are good */
		if (read(ufd, &ubuf, sizeof(ubuf)) == 0) { /* no such tty */
			printf("I can't find %s listed in /etc/utmp\n", MyTty);
			exit(1); 
		}
		/* printf("Checking %s\n", ubuf.ut_line); */
		if (strcmp(ubuf.ut_line, MyTty) != 0) continue; /* not our entry */
		if (strncmp(ubuf.ut_name, MyName, sizeof(ubuf.ut_name))) {
			/* our entry has gone west */
			printf("%s is now owned by %s\n", MyTty, ubuf.ut_name);
			exit(1); 
		}
		else { /* our entry looks just fine */
			close(ufd);
			return;
		}
	}
} /* CheckTty */

void WaitAndPrint(How)
int How;
/* wait until AdvanceSeconds before an appointment.  If How = BYTERMINAL,
print it out.  If How == BYMAIL, send mail. */
{
	AppointmentListType *ComingSoon, *Ptr;	
	struct tm *localtime(), *TimePtr;
	long time();
	long TotalSeconds;
	int WaitSeconds;
	MyAppointments.Next = MyAppointments.Prev = &MyAppointments;
	for (;;) { /* never exit */
		if (CheckFlag) CheckTty();
		Resetting = 0;
		InitializeNow();
		ReadAppointments(&MyAppointments, 0); /* frees old list first */
		TotalSeconds = time((long *)0);
		TimePtr = localtime(&TotalSeconds);
		SecondsNow = TimePtr->tm_sec + 60*(TimePtr->tm_min+60*TimePtr->tm_hour);
		ComingSoon = WhatNext();
		if (ComingSoon == 0) { /* no more appointments today */
			unsigned int SecondsLeftToday = 1 + (59 - TimePtr->tm_sec) +
				60*(59-TimePtr->tm_min + 60*(23-TimePtr->tm_hour));
			if (SecondsLeftToday > SLEEPINTERVAL) {
				if (Debug)
					printf("No more appointments today; waiting an interval\n");
				sleep(SLEEPINTERVAL);
			} else {
				if (Debug)
					printf("No more appointments today; waiting %d seconds\n",
						SecondsLeftToday);
				sleep(SecondsLeftToday);
			}
			continue;
		}
		WaitSeconds = 60*ComingSoon->Next->ApptEntry->StartTime - SecondsNow -
			AdvanceSeconds;
		if (WaitSeconds > SLEEPINTERVAL) { /* too long in future */
				if (Debug) printf("Waiting an interval\n");
			sleep(SLEEPINTERVAL - 20); /* don't miss anything! */
			continue;
		}
		if (Debug) printf("Sleeping %d seconds\n", WaitSeconds);
		sleep((unsigned) WaitSeconds);
		if (Resetting) { /* awakened by SIGINT */
			continue;
		}
		fprintf(stdout, "\n");
		for (Ptr = ComingSoon->Next; Ptr != ComingSoon; Ptr = Ptr->Next) {
			if (How == BYTERMINAL) {
				fprintf(stdout, "***  ");
				PrintAppointment(Ptr->ApptEntry, 0, 0, &DateNow, stdout);
			} else { /* BYMAIL */
				char buf[MAXLINE];

				freopen(TMPFILE, "w", stdout);
				PrintAppointment(Ptr->ApptEntry, 0, 0, &DateNow, stdout);
				fflush(stdout);
				sprintf(buf, "%s -s 'Appointment in %d minutes' $USER < %s",
					Mailer, AdvanceSeconds / 60, TMPFILE);
				CallProcess(0, SHELL, "-c", buf, 0);
				(void) unlink(TMPFILE);
			} /* BYMAIL */
		} /* for each appointment that matches now */
		FreeAppointmentList(ComingSoon);
	} /* never exit */
} /* WaitAndPrint */

void ResetTimer()
{
	Resetting = 1;
	sleep(1); /* cancels current longer sleep */
} /* ResetTimer */

#define NullToEmptyString(s) (s ? s : "")

void InitializePeriodic()
{
	long time();
	char *getenv(), *ttyname(), *rindex();
	struct passwd *getpwuid();
	(void) strncpy(User, NullToEmptyString(getenv("HOME")), sizeof(User));
	(void) strncpy(Mailer, NullToEmptyString(getenv("MAILER")), sizeof(Mailer));
	if (Mailer[0] == 0) strncpy(Mailer, MAILER, sizeof(Mailer));
	(void) signal(SIGINT, ResetTimer);
	(void) signal(SIGHUP, SIG_DFL); /* if started with '&', the shell
		has immunized us against SIGHUP; we don't want immunity. */
	MyAppointments.Next = MyAppointments.Prev = &MyAppointments;
	ActiveAppointments = &MyAppointments;
	InitializeNow();
	/* MyTty = rindex(ttyname(1), '/') + 1; */
	MyTty = ttyname(1) + 5;
	MyName = getpwuid(getuid())->pw_name;
	/* printf("My name is %s on %s\n", MyName, MyTty); */
} /* InitializePeriodic */

#ifdef undef

main(){
	InitializePeriodic();
	WaitAndPrint();
} /* main */

#endif
