.SH 1 "Mechanisms"
.lp
We will now turn to mechanisms that can be used to provide mutual exclusion and
synchronization.
We will start with simple but conservative techniques and move toward more
complex, liberal ones.
In each case we will show how the BeginRegion and EndRegion primitives are
built.
.pp
All mechanisms for mutual exclusion and synchronization depend ultimately on
the synchronous nature of hardware.
Some rely on the fact that processors can be made uninterruptible.
These \fBprocessor-synchronous\fP methods work only for individual processors.
.IN "Processor synchronous" "synchronization"
Others rely on the fact that main store can service only one access request at
a time, even in a multiprocessor.
.IN "Multiprocessor"
These \fBstore-synchronous\fP methods have a wider range of applicability.
.IN "Store synchronous" "synchronization"
.SH 2 "Disable interrupts"
.IN Interrupts Disabling
.lp
Most computer architectures allow interrupts to be disabled, perhaps on a
selective basis.
If a device signals an interrupt while interrupts are disabled,
the new interrupt will be postponed until interrupts are again allowed.
If activities are preempted only when an interrupt occurs,
exclusion can be enforced by disabling interrupts while any activity is in a
region.
Since no interrupts arrive, the current activity will be able to complete the
instructions in its region before any other activity starts executing
in a conflicting region.
.pp
Adjusting the interruptibility of the computer leads to a
processor-synchronous technique for mutual exclusion.
The BeginRegion and EndRegion procedures look like this:
.(b
.FS
 1	\fWprocedure\fP BeginRegion;
 2	\fWbegin\fP	
 3		DisableInterrupts;
 4	\fWend\fP BeginRegion;
 5	
 6	\fWprocedure\fP EndRegion;
 7	\fWbegin\fP
 8		EnableInterrupts;
.\" Nested regions discussed below
 9	\fWend\fP EndRegion;
.FE
.)b
.pp
Disabling interrupts can be quite fast.
A single machine instruction often suffices to change the
interrupt state of the machine.
However, this method has some unfortunate properties.
.sp 10p
.ip \(bu
Once an activity has entered a region, real-time events, like
devices that need service, cannot be treated until the activity has left the
region.
Therefore, lengthy processing in a region is usually unacceptable.
This property is due to the fact that disabling interrupts is extremely
conservative.
.ip \(bu
It excludes not only potentially conflicting activities that
might otherwise access shared data but also all activities whatsoever, no
matter what their purpose.
.ip \(bu
One reason a region might be lengthy is that the program might have to wait for
some event before continuing.
For example, the program might be waiting for a device to become available.
The programmer should not put waits inside regions.
.ip \(bu
The technique fails on multiprocessors because it is usually not possible for a
.IN Multiprocessors
program on one machine to block interrupts on other machines.
Even if interrupts can be disabled on all machines,
conflicting activities can still be executing simultaneously.
.ip \(bu
The programmer must be careful to ensure that all paths from the region go
through code to release exclusion by enabling interrupts once again.
It is easy to overlook some path and introduce a deadlock.
.ip \(bu
Another pitfall involves nested regions.
If procedure \fBA\fP calls procedure \fBB\fP from inside a region 
and \fBB\fP also enters a region, \fBB\fP will re-enable interrupts
during its EndRegion operation, relaxing exclusion before \fBA\fP was ready to
do so.
(See the exercises at the end of this chapter.)
.sp 10p
.lp
.SH 2 "Busy wait on MutEx"
.IN "Busy wait"
.lp
The method of disabling interrupts lets any activity into a region but prevents
any other activity from running while the first one is in the region.
A different approach is to wait at the entry to a region until it is safe to
continue.
We can introduce a Boolean variable MutEx (for Mutual Exclusion) that is set to 
true when an activity is in a region and false otherwise.
We then code BeginRegion and EndRegion as follows:
.(b
.FS
 1	\fWvar\fP MutEx : Boolean := false;
 2	
 3	\fWprocedure\fP BeginRegion;
 4	\fWbegin\fP
 5		\fWwhile\fP MutEx \fWdo\fP
 6			\fWnull\fP; { that is, do nothing }
 7		\fWend\fP;
 8		MutEx := true;
 9	\fWend\fP BeginRegion;
10	
11	\fWprocedure\fP EndRegion(Who : Activity);
12	\fWbegin\fP
13		MutEx := false;
14	\fWend\fP EndRegion;
.FE
.)b
Unfortunately, this store-synchronous solution is wrong.
If two activities enter BeginRegion at about the same time, they might
both get past the loop in line 5 and enter the region.
The problem is that an activity that decides to enter the region at line 5 does
not prevent others from entering as well until line 8.
By then it may be too late.
We will fix this problem when we look at locks.
.pp
We introduce this
.q solution
to demonstrate that even though it fails for mutual exclusion, busy
waiting \fIdoes\fP suffice to implement the synchronization graph
of Figure \n(ch.\n(F1.
We will use the following convention:
Every activity in the graph is associated with a shared Boolean variable
.q "Done." " "
For example, \fBA\fP is associated with Done[\fBA\fP].
We then build two utility procedures, Finishing and WaitFor:
.(b
.FS
 1	\fWvar\fP
 2		Done : \fWarray\fP [Activity] \fWof\fP Boolean := false;
 3	
 4	\fWprocedure\fP Finishing(Who : Activity);
 5	\fWbegin\fP
 6		Done[Who] := true;
 7	\fWend\fP Finishing;
 8	
 9	\fWprocedure\fP WaitFor(Whom : Activity);
10	\fWbegin\fP
11		\fWwhile\fP \fWnot\fP Done[Whom] \fWdo\fP
12			\fWnull\fP;
13		\fWend\fP;
14	\fWend\fP WaitFor;
.FE
.)b
The code for a few selected activities could look like this:
.(b
.FS
 1	activity \fWA\fP:
 2		perform \fWA\fP's work;
 3		Finishing(\fWA\fP);
 4	
 5	activity \fWB\fP:
 6		WaitFor(\fWA\fP);
 7		perform \fWB\fP's work;
 8		Finishing(\fWB\fP);
 9	
10	activity \fWE\fP:
11		WaitFor(\fWB\fP);
12		WaitFor(\fWD\fP);
13		perform \fWE\fP's work;
14		Finishing(\fWE\fP);
.FE
.)b
Even though both Finishing and WaitFor use the Done array, we don't have to
worry about simultaneous access, since only one activity will ever call
Finishing to modify any particular value in that array.
Other activities may call WaitFor, but they only inspect the array.
.pp
If we treat each row in the graph as a separate activity, some of the explicit
calls to WaitFor and Finishing are no longer needed:
.(l
.FS
.ne 8
 1	activity \fWABC\fP:
 2		perform \fWA\fP's work;
 3		Finishing(\fWA\fP);
 4		perform \fWB\fP's work;
 5		Finishing(\fWB\fP);
 6		perform \fWC\fP's work;
 7		Finishing(\fWC\fP);
 8	
.ne 12
 9	activity \fWDEF\fP:
10		WaitFor(\fWA\fP);
11		perform \fWD\fP's work;
12		Finishing(\fWD\fP);
13		WaitFor(\fWB\fP);
14		perform \fWE\fP's work;
15		Finishing(\fWE\fP);
16		WaitFor(\fWC\fP);
17		perform \fWF\fP's work;
18		Finishing(\fWF\fP);
19	
.ne 8
20	activity \fWGHI\fP:
21		WaitFor(\fWD\fP);
22		perform \fWG\fP's work;
23		WaitFor(\fWE\fP);
24		perform \fWH\fP's work;
25		WaitFor(\fWF\fP);
26		perform \fWI\fP's work;
.FE
.)l
.SH 2 "Switch variable" I
.lp
A more liberal store-synchronous mechanism for ensuring mutual exclusion between
two activities is for them to share a new variable called a \fBswitch\fP.
This variable lets either one or the other enter its region.
We depend on the synchronization atomicity of main store to arbitrate
simultaneous accesses to this variable.
.pp
In the following example, the switch variable is called Turn:
.(l
.FS
.ne 4
 1	\fWtype\fP Activity : (A, B);
 2	
 3	\fWvar\fP Turn : Activity := A;
 4	
.ne 7
 5	\fWprocedure\fP BeginRegion(Who : Activity);
 6	\fWbegin\fP
 7		\fWwhile\fP Turn <> Who \fWdo\fP
 8			\fWnull\fP;
 9		\fWend\fP;
10	\fWend\fP BeginRegion;
11	
.ne 8
12	\fWprocedure\fP EndRegion(Who : Activity);
13	\fWbegin\fP
14		\fWif\fP Who = A \fWthen\fP
15			Turn := B
16		\fWelse\fP
17			Turn := A;
18		\fWend\fP;
19	\fWend\fP EndRegion;
.FE
.)l
In line 3, we initialize the switch variable to point to activity \fBA\fP.
If \fBA\fP calls BeginRegion, the condition on the \fWwhile\fP in line 7 will be
false, so BeginRegion will return.
If \fBB\fP tries to enter its region, it will be blocked waiting for the
switch in line 7.
Turn will become \fBB\fP 
only after \fBA\fP has released the switch by calling EndRegion.
.pp
Switches are only slightly better than busy waiting.
.sp 10p
.ip \(bu
They correctly allow \fBA\fP and \fBB\fP to execute in
conflicting regions, even if
\fBA\fP and \fBB\fP are on different machines in a multiprocessor.
If \fBA\fP and \fBB\fP call BeginRegion simultaneously, they will not conflict
in their use of Turn, since BeginRegion only examines Turn without changing it.
\fBA\fP and \fBB\fP will never call EndRegion simultaneously because they
can't both be in their region at once.
.ip \(bu
They may be generalized to more than two processes.
.ip \(bu
Switches are more liberal than disabling interrupts.
Unrelated conflicts are handled by different switch variables.
Therefore, two activities can be executing in non-conflicting regions at the
same time.
For example, assume that there are two shared variables, x and y, and that
activity \fBA\fP uses
just x, activity \fBB\fP uses both x and y, and activity \fBC\fP uses just y.
We would use one switch variable, TurnX, with values \fBA\fP and \fBB\fP, to
protect variable x.
A different switch variable, TurnY, with values \fBB\fP and \fBC\fP,
could protect variable y.
When \fBB\fP wishes to enter a region, it would specify not only its own name
(\fBB\fP), but also which switch variable it wants to use.
.ip \(bu
If several switch variables exist and a single activity might want to be in a
combined region protected by several different switch variables, there is a
danger of deadlock because switch variables are serially reusable resources.
A hierarchy of switch variables can prevent this problem.
We used the same idea in Chapter 4 when we introduced hierarchical allocation
of sequentially reusable resources.
.ip \(bu
When an activity is waiting to enter its region (line 7), it performs
no useful work while it consumes computational resources.
If activities are competing with each other on a single processor, these busy
waits waste time because no other productive action takes place, even though
there is some activity that could make use of the time (namely, the activity
that is in the region).
If activities are on different processors, we might not mind this waste of time
so much because all activities that can make use of a processor resource might
be on machines of their own.
Still, a busy-waiting activity consumes main-store cycles, which are shared
among all the processors.
Even if other activities could run, the Hysteresis Principle tells us that
.IN "Hysteresis Principle"
busy waiting is still reasonable if the
expected waiting time is less than one process-switch time.
Very long busy-wait times tend to create a \fBconvoy phenomenon\fP,
.IN "Convoy phenomenon"
in which many activities end up waiting for the same region.
The entire convoy tends to collide at the next region, too, leading to worse
contention than might have been expected.
.ip \(bu
This solution forces \fBA\fP and \fBB\fP to alternate entry into their regions.
If \fBA\fP needs to enter its region only occasionally but \fBB\fP needs to do
so frequently, \fBB\fP will be blocked waiting for the region most of the
time, even though \fBA\fP is not currently in the region.
.sp 10p
.lp
.SH 2 "Non-alternating switch"
.IN Switch "non-alternating (start)"
.lp
The principal drawback of the switch method is the strict alternation it
imposes on conflicting activities. 
A non-alternating switch uses a switch variable and two auxiliary Boolean
variables that indicate whether the two activities are currently interested in
entering their regions.
We will store them in an array for convenience.
Here is the code that implements this store-synchronous method:
.(l
.FS
.ne 6
 1	\fWtype\fP Activity = (A,B);
 2	
 3	\fWvar\fP
 4		Turn : Activity := A;
 5		Interested : \fWarray\fP [Activity] \fWof\fP Boolean := false;
 6	
.ne 18
 7	\fWprocedure\fP BeginRegion(Who : Activity);
 8	\fWvar\fP Other : Activity; { the other activity besides Who }
 9	\fWbegin\fP
10		\fWif\fP Who = A \fWthen\fP Other := B \fWelse\fP Other := A \fWend\fP;
11		\fWloop\fP "outer"
12			{ we exit when we may enter our region }
13			Interested[Who] := true;
14			\fWrepeat\fP { busy wait if our turn, Other interested }
15				\fWwhen\fP \fWnot\fP Interested[Other] \fWexit\fP outer;
16			\fWuntil\fP Turn = Other;
17			{ we should stop objecting; it's not our turn. }
18			Interested[Who] := false;
19			\fWwhile\fP Turn = Other \fWdo\fP { busy wait for our turn }
20				\fWnull\fP;
21			\fWend\fP;
22		\fWend\fP \fWloop\fP;
23	\fWend\fP BeginRegion;
24	
.ne 5
25	\fWprocedure\fP EndRegion(Who : Activity);
26	\fWbegin\fP
27		\fWif\fP Who = A \fWthen\fP Turn := B \fWelse\fP Turn := A \fWend\fP;
28		Interested[Who] := false;
29	\fWend\fP EndRegion;
.FE
.)l
.pp
The Interested variables ensure that simultaneous execution does not occur:
Each activity indicates that it is interested (line 13) both before it enters
its region and during the entire time it is in the region.
An activity enters the region only if the other activity is not interested.
The Turn variable ensures that mutual blocking does not occur.
It does not change value during BeginRegion, and it is used only to determine
which activity backs off (lines 17\(mi21) if both wish to enter their region.
Only activity \fBA\fP ever modifies Interested[\fBA\fP], and, likewise, only
\fBB\fP modifies Interested[\fBB\fP], so there is no conflict for those
variables.
Only the activity currently in a region modifies Turn (line 27), so there is no
conflict for that variable either.
.pp
A significant amount of effort has been directed to
deriving the non-alternating switch.
There are many ways in which the switch can be built incorrectly.
For many years the program just shown, developed by Dekker,
was the only good solution.
A surprisingly simple solution was published by Peterson in 1981:
.(l
.FS
.ne 6
 1	\fWtype\fP Activity = (A,B);
 2	
 3	\fWvar\fP
 4		Turn : Activity := A;
 5		Interested : \fWarray\fP [Activity] \fWof\fP Boolean := false;
 6	
.ne 11
 7	\fWprocedure\fP BeginRegion(Who : Activity);
 8	\fWvar\fP Other : Activity; { the other activity besides Who }
 9	\fWbegin\fP
10		\fWif\fP Who = A \fWthen\fP Other := B \fWelse\fP Other := A \fWend\fP;
11		Interested[Who] := true;
12		Turn := Other;
13		\fWwhile\fP Interested[Other] \fWand\fP Turn = Other \fWdo\fP
14			\fWnull\fP;
15		\fWend\fP;
16	\fWend\fP BeginRegion;
17	
.ne 4
18	\fWprocedure\fP EndRegion(Who : Activity);
19	\fWbegin\fP
20		Interested[Who] := false;
21	\fWend\fP EndRegion;
.FE
.)l
.lp
We leave the proof of this implementation as an exercise.
.pp
The non-alternating switch has the following properties:
.sp 10p
.ip \(bu
It is correct in the same way that the ordinary switch is correct.
.ip \(bu
It is more liberal than the ordinary switch because it allows a faster
activity to enter its region many times before the other activity has entered.
However, it is still fair because priority is given to the activity that has
not entered its region for the longest time.
.ip \(bu
It is impossible for an activity to starve while waiting to enter its region,
if we assume that whenever the conflicting activity enters its region, it
eventually leaves that region.
.ip \(bu
The solution can be generalized to any number of activities that conflict over
the same shared variables.
The generalization is called a \fBspin switch\fP.
.IN "Spin switch"
.ip \(bu
The solution is complex and somewhat unclear, so it is easy to make a mistake
when programming it.
.ip \(bu
The non-alternating switch uses busy waiting if a conflicting activity is in
its region.
.sp 10p
.lp
.IN Switch "non-alternating (end)"
.SH 2 "Locks" I
.lp
The switch solutions we have seen are store-synchronous; they assume that
individual references to main store are atomic.
Many computers offer an explicit store-synchronous instruction
that makes mutual exclusion far easier and clearer.
This instruction atomically reads a variable and sets it to a new value.
The most common design is to set the new value to 1 no matter what the old
value.
In this case, the instruction is called Test and Set.
.IN "Test and set instruction"
It has the following behavior:
.(b
.FS
 1	\fWatomic\fP \fWfunction\fP TestAndSet(\fWvar\fP Lock : Boolean) : Boolean;
 2	\fWbegin\fP
 3		TestAndSet := Lock;
 4		Lock := true;
 5	\fWend\fP TestAndSet;
.FE
.)b
One of the exercises explores a different atomic action that increments the old
value.
.pp
Given Test and Set, we can build a simple mutual-exclusion mechanism called a
\fBlock\fP:
.(l
.FS
.ne 9
 1	\fWtype\fP Lock = Boolean;
 2	
 3	\fWprocedure\fP BeginRegion(\fWvar\fP L : Lock);
 4	\fWbegin\fP
 5		\fWwhile\fP TestAndSet(L) \fWdo\fP
 6			\fWnull\fP;
 7		\fWend\fP;
 8	\fWend\fP BeginRegion;
 9	
.ne 4
10	\fWprocedure\fP EndRegion(\fWvar\fP L : Lock);
11	\fWbegin\fP
12		L := false;
13	\fWend\fP;
.FE
.)l
Each region that an activity might enter uses a number of shared variables that
must be protected.
Each such group of shared variables is assigned a lock variable.
Before entering a region, an activity is expected to call BeginRegion on the
appropriate lock.
If several locks are needed, they should be acquired in a standard hierarchical
order to prevent deadlock.
It makes no difference in what order they are released at the end of the
region.
.pp
This solution works because the TestAndSet instruction (line 5) will return
true as long as some other activity is in a conflicting region.
When it returns false, it simultaneously sets the lock to true so that any
other activity trying to acquire the lock will fail.
If several activities are trying to get the same lock and the activity with
the lock releases it (line 12), only one of the contenders will get it because
TestAndSet is atomic.
(Which one of the contenders gets it is arbitrary and depends on hardware
arbitration.) 
.pp
The lock method has the following properties:
.sp 10p
.ip \(bu
It works for any number of activities on any number of processors sharing main
store.
.ip \(bu
It is simple and easy to verify.
.ip \(bu
It is liberal in that an activity that holds a lock does not prevent another
activity from acquiring a different lock.
.ip \(bu
It allows starvation because the choice of which activity to honor is
arbitrary, on the assumption that the hardware is unintelligent.
.ip \(bu
It uses busy waiting.
.sp 10p
.lp
.SH 2 "Semaphores" I
.lp
Each of the previous solutions employs busy waiting.
The following methods all share a new property that distinguishes them from the
busy-waiting methods:\ \&
They depend on the ability to schedule activities.
An activity that attempts to enter a region that is already occupied can be
blocked, just as a process that attempts to gain
a resource that is currently allocated might be blocked by the resource
manager.
While it waits, other activities may continue.
Instead of consuming computational resources in a fruitless loop,
the waiting activity only needs to suffer the cost of a context switch or two.
To achieve this effect, we need to embed the BeginRegion and EndRegion
operations in the scheduler instead of building them directly on the hardware.
The scheduler can also maintain a list of activities waiting to enter a region
so that a fair choice can be made when conditions allow.
.pp
Our first example of this approach is to
protect each group of shared variables with a
\fBsemaphore\fP, which has the following structure:
.(b
.FS
 1	\fWtype\fP
 2		Queue = \fWlist\fP \fWof\fP Activity
 3		Semaphore =
 4			\fWrecord\fP
 5				{ all fields are initialized as shown }
 6				MutEx : Lock := false;
 7				Value : integer := 1;
 8				Waiters : Queue := empty;
 9			\fWend\fP;
.FE
.)b
The scheduler provides two operations on semaphores.
We will call the first Up (sometimes people
call this operation P, Wait, or Acquire).
.IN Semaphore Up
The second operation is Down (also called V, Signal, or Release). 
.IN Semaphore Down
Informally, Down blocks the caller if Value (line 7) is 0.
Otherwise, it decrements Value.
Up increments Value and unblocks at most one waiting activity.
.pp
The correct use of semaphores to implement mutual exclusion is simple:\ \&
All regions that use the same shared variables are associated with the same
semaphore.
An activity that wishes to enter a region calls Down on the associated
semaphore (instead of BeginRegion).
When the activity exits the region, it calls Up on the same semaphore
(instead of EndRegion).
The first activity to try to enter its region succeeds, because Value is
initially 1.
Another activity that tries to enter while the first is still in its region
will be blocked.
When the first activity leaves the region, the second activity is unblocked.
The Value field is always either 0 or 1.
.pp
To give a more formal definition of Down and Up, we will implement them
using locks as a more primitive mutual-exclusion tool.
.(l
.FS
.ne 11
 1	\fWprocedure\fP Down(\fWvar\fP S : Semaphore);
 2	\fWbegin\fP
 3		BeginRegion(S.MutEx); { use TestAndSet }
 4		\fWif\fP S.Value = 0 \fWthen\fP
 5			Block(S.Waiters); { proceed when unblocked later }
 6		\fWelse\fP
 7			S.Value := S.Value \(mi 1;
 8		\fWend\fP;
 9		EndRegion(S.MutEx);
10	\fWend\fP Down;
11	
.ne 10
12	\fWprocedure\fP Up(\fWvar\fP S : Semaphore);
13	\fWbegin\fP
14		BeginRegion(S.MutEx);
15		\fWif\fP \fWnot\fP Empty(S.Waiters) \fWthen\fP
16			UnBlock(S.Waiters) { Unblock one waiter.  We continue. }
17		\fWelse\fP
18			S.Value := S.Value + 1;
19		\fWend\fP;
20		EndRegion(S.MutEx);
21	\fWend\fP Up;
.FE
.)l
.pp
It is important that Down and Up be mutually exclusive.
In particular, any access of the Value or the Waiters field must be atomic.
That is why we use the MutEx field of each semaphore to make sure that
Down and Up exclude each other.
We must make sure that both Down and Up retain exclusion for only a
short time; otherwise our attempt to avoid busy waiting has failed.
.pp
In the simple case, Down does not find S.Value equal to 0 in line 4 and need
not block in line 5.
If it does block, we require that the Block routine (in the scheduler) resume
some other runnable activity and that Block turn off exclusion before
resumption.
Block accomplishes this by calling EndRegion(S.MutEx).
Block therefore needs to know which semaphore is in use;
we omit this argument for clarity.
UnBlock, called by Up in line 16, marks one waiting activity as
runnable (for example, by placing it in a ready list).
.IN Ready list
The releasing activity then continues and soon releases exclusion at line 20.
When the newly runnable activity is scheduled to run, it is again given
exclusion, and it finds itself running at line 5 in the Down routine.
It will soon release exclusion itself in line 9.
.pp
There is some controversy over whether the scheduler should switch
immediately to the waiting activity that is activated in line 16.
Immediate switch guarantees that whatever condition is being awaited by that
activity still holds, since the Up operation has just been called and no
other activity has had a chance to run.
The disadvantage of an immediate switch is that it tends to increase the total
number of switches.
The activity that called Up is likely to call Down for a new region
soon, causing it to block in any case.
The Hysteresis Principle suggests that the current process should be allowed to
continue.
.IN "Hysteresis Principle"
.pp
Semaphores are so useful that some operating systems provide service calls so
that processes that share resources (particularly parts of virtual store) can
synchronize their accesses.
Four service calls are needed:
.sp 10p
.ip \(sq
\fHSemaphoreCreate(initial value).\fP
.IN "Service call" SemaphoreCreate
This call returns a new semaphore descriptor (a small integer) that the calling
process may use.
The semaphore structure itself is protected in the kernel and has its Value
field set to the given initial value.
This semaphore may be inherited by the children of the calling process so they
can all share the same semaphore.
.IN "Inheritance" "Semaphores"
.ip \(sq
\fHSemaphoreDestroy(semaphore descriptor).\fP
.IN "Service call" SemaphoreDestroy
This call informs the kernel that the given semaphore is no longer needed.
This call is implicitly performed when the last process using a semaphore
terminates.
Any process waiting on a SemaphoreDown call receives an error
return from that call.
.ip \(sq
\fHSemaphoreDown(semaphore descriptor).\fP
.IN "Service call" SemaphoreDown
This call performs the Down operation on the given semaphore.
An error is reported if the semaphore is not associated with this process
or if the semaphore is destroyed while the process is blocked waiting for it.
.ip \(sq
\fHSemaphoreUp(semaphore descriptor).\fP
.IN "Service call" SemaphoreUp
This call performs the Up operation on the given semaphore.
An error is reported if the semaphore is not associated with this process.
.sp 10p
.pp
Semaphores have the following properties:
.sp 10p
.ip \(bu
They correctly implement a liberal policy of mutual exclusion among any number
of activities on any number of processors.
Activities interfere with each other only if they refer to the same semaphore.
.ip \(bu
When an activity is blocked from entering its region, it does not busy wait.
.ip \(bu
Starvation is possible unless waiters are unblocked in first-come, first-served
order.
.ip \(bu
As with all the methods we have seen so far, there is no guarantee that
activities will call Down and Up at the correct times.
The wrong (or no) call may be made, or the wrong semaphore may be invoked.
.ip \(bu
A semaphore used for mutual exclusion is a serially reusable, non-preemptable
resource.
Its use is therefore subject to deadlock.
A hierarchical order for acquiring multiple semaphores can be used to prevent
deadlock.
.IN "Hierarchical allocation" semaphores
.sp 10p
.lp
It is easy to generalize the semaphore to allow any fixed number of activities
the right to enter their region at the same time.
For example, we might have seven tape drives and be
willing to allow up to seven activities to access tape-drive code.
To enforce this policy, we would build a semaphore for tape-drive code and
initialize its value to 7 instead of 1.
.pp
Semaphores allow us to implement synchronization without busy waiting.
For example, we could introduce a semaphore for each arrow in the
synchronization graph of Figure \n(ch.\n(F1: AB, BE, and so on.
Each semaphore would be initialized to 0.
A typical activity, like \fBE\fP, would have this sort of code:
.(b
.FS
 1	activity \fWE\fP:
 2		Down(BE);
 3		Down(DE);
 4		perform \fWE\fP's work;
 5		Up(EF);
 6		Up(EH);
.FE
.)b
We don't really need so many semaphores.
Instead, we could introduce just one semaphore per activity.
Again, it would be initialized to 0.
When the activity finishes, it invokes Up on the semaphore as many times
as there are activities waiting for this one to finish.
Before an activity starts, it invokes Down on the semaphores of all the
activities it needs to wait for.
For example, \fBE\fP looks like this:
.(b
.FS
 1	activity \fWE\fP:
 2		Down(\fWB\fP);
 3		Down(\fWD\fP);
 4		perform \fWE\fP's work;
 5		Up(\fWE\fP); { once for \fWF\fP, once for \fWH\fP }  
 6		Up(\fWE\fP);
.FE
.)b
One of the exercises at the end of the chapter explores another way to use
semaphores to solve the same problem.
.SH 2 "Critical regions" I
.lp
Semaphores provide the power we need, but they are easy to misuse.
The following sections deal with ways to embed semaphores in
programming-language constructs that try to minimize three problems:
.np
Failure to protect shared variables
.np
Failure to invoke semaphore operations in the right order and at the right
time
.np
Deadlock and starvation
.lp
We begin by a programming-language construct called the \fBcritical region\fP.
It is meant to be part of any language that explicitly allows several
activities to share variables.
Variables that are shared must be so declared;
they are the only global variables that activities are allowed to access.
Access to shared variables is restricted to code that lies within the
\fWregion\fP statement, as shown here:
.IN "\fWRegion\fP statement"
.(b
.FS
 1	\fWvar\fP
 2		v : \fWshared\fP integer;
 3	
 4	. . .
 5	
 6	\fWregion\fP v \fWdo\fP
 7		v := 15;
 8		write(v);
 9	\fWend\fP;
.FE
.)b
We no longer use explicit calls to procedures BeginRegion and EndRegion; the
compiler inserts those calls (most likely, calls to semaphores) for us.
.pp
There is a danger of deadlock because activity \fBA\fP might nest (that is,
embed) \fWregion\fP \fLv\fP inside \fWregion\fP \fLw\fP,
whereas activity \fBB\fP might nest them in the other order.
Sometimes the compiler can check the order of nesting and disallow conflicting
orders.
However, if a region includes a call to a procedure that then enters another
region, a nesting situation arises that the compiler cannot be expected to
detect.
Still, an impermissible order can be discovered when the program runs.
Another way to skirt the problem is to avoid nesting.
If a single region needs to use several shared variables, they may be listed
in the \fWregion\fP statement.
In this case, the compiler may choose any order it wishes to acquire the
semaphores.
A consistent order is equivalent to a hierarchical allocation rule.
.IN "Hierarchical allocation" "Critical regions"
.pp
If several activities are waiting for the same shared variable, we can protect
against starvation by making sure the underlying semaphores keep all waiting
activities in a first-in, first-out queue.
.pp
Regions protect the programmer against two mistakes that are common with
semaphores.
First, there is no way to access a shared variable from outside a region.
Variables are visible to more than one activity only if they are declared
\fWshared\fP.
.IN "\fWShared\fP declaration"
Shared variables are invisible except within critical regions.
Entry to critical regions automatically excludes conflicting activities.
Second, there is no way to forget to call Up at the end of a region.
Up is no longer explicitly invoked.
.pp
But several problems remain.
The first is that synchronization requires busy waiting if critical regions are
the only facility.
The second is that accesses to shared variables may be scattered throughout the
program, with the result that it is difficult to coordinate all the places
that modify them.
In addition, there is no way for several activities to read the same shared
variables simultaneously even though simultaneous reads need not be mutually
exclusive.
We will address these objections in the next few sections.
.SH 2 "Conditional critical regions" I
.lp
A slight modification to critical regions will allow us to solve the
synchronization problem without busy wait.
As we saw before, synchronization in general is the desire to block an action
until a particular condition becomes true.
For example, the
.q pop
operation makes sense only for a stack that is not empty.
The action of popping the stack should be delayed until that condition
holds.
Likewise, action \fBB\fP of Figure \n(ch.\n(F1 should continue only if \fBA\fP
has terminated.
.pp
To meet these needs, we add an \fWawait\fP
statement that may only be used inside a critical region.
.IN "\fWAwait\fP statement"
\fWAwait\fP names an arbitrary Boolean condition.
If the condition is not met, the activity executing the \fWawait\fP statement is
blocked until it is met.
For example, popping from a stack could be done as follows:
.(l
.FS
.ne 11
 1	\fWtype\fP
 2		stack =
 3			\fWrecord\fP
 4				Count : integer; { How many elements in stack }
 5				Values : \fWarray\fP 1:10 \fWof\fP element;
 6			\fWend\fP;
 7	\fWvar\fP
 8		Trays : \fWshared\fP stack;
 9	
10	...
11	
.ne 5
12	\fWregion\fP Trays \fWdo\fP
13		\fWawait\fP Trays.Count > 0;
14		Result := Trays.Values[Trays.Count];
15		Trays.Count := Trays.Count - 1;
16	\fWend\fP;
.FE
.)l
To solve the synchronization problem of Figure \n(ch.\n(F1, we can have a
Boolean variable indicate when each activity finishes and test that variable
in an \fWawait\fP statement.
.pp
We must decide exactly what the \fWawait\fP statement means.
Does it relax exclusion if it blocks?
If not, how can another activity ever modify the condition that is being
awaited?
If so, how can we be sure that no other activity has confused the data
structures while we were blocked?
One consistent approach is to demand that \fWawait\fP statements appear only at
the start of critical regions and that exclusion not start until the
condition is met.
(Of course, checking the condition must be atomic, a condition that can be
achieved through hidden semaphores.)\ \&
Another approach is to allow \fWawait\fP statements anywhere in the critical
region.
Any \fWawait\fP that blocks releases exclusion, so it is up to the programmer
to make sure that data structures are consistent before encountering
\fWawait\fP.
In this way other activities will not be confused when they enter their
conflicting region, and the blocked activity will not be confused when it
resumes.
.pp
We must also decide exactly when the conditions are to be tested.
If each activity is on a separate processor,
we may test conditions continually.
However, since such busy waiting is unacceptable for most implementations, we
need to limit the times when the condition may be tested.
Certainly there is no need to test the condition as long as some other
activity is in a conflicting region, because that other activity must maintain
exclusion until it leaves its region, even though it may have made the
condition true by its actions.
In practice, there may be no need to test conditions when all activities are
outside conflicting regions because useful conditions involve shared
variables, and shared variables are changed only by activities within critical
regions.
One time we need to check conditions is when an activity leaves a critical
region;
the only conditions that must be tested are those in conflicting regions.
The other time is when an activity in a conflicting region enters \fWawait\fP.
(This situation only arises if \fWawait\fP is permitted in the middle of
regions.)\ \&
Checking a condition can still be very expensive because it may require
switching to the environment of the blocked activity in order to access local
variables involved in the condition.
For efficiency, such complex conditions may be disallowed.
.pp
To show how conditional critical regions can be used to implement our
synchronization graph, we present the code for activity \fBE\fP:
.(b
.FS
 1	\fWvar\fP
 2		ADone, BDone, CDone, DDone, EDone, FDone, GDone :
 3			\fWshared\fP Boolean := false;
 4	
 5	activity E:
 6		\fWregion\fP BDone \fWdo\fP \fWawait\fP BDone \fWend\fP;
 7		\fWregion\fP DDone \fWdo\fP \fWawait\fP DDone \fWend\fP;
 8		perform E's work;
 9		\fWregion\fP EDone \fWdo\fP EDone := true; \fWend\fP;
.FE
.)b
Conditional critical regions are very easy to use for complex mutual-exclusion
and synchronization problems.
The exercises at the end of this chapter suggest some problems to try;
they are based on ideas we will introduce soon.
