.SH 2 "Monitors" I
.lp
One objection to using conditional critical regions is the cost involved in
checking conditions.
A second objection is that code that modifies shared data may be scattered
throughout a program.
The language construct called a \fBmonitor\fP was invented to address both
these issues.
It acts both as a data-abstraction device and a mutual-exclusion device.
Synchronization is provided by a new data type called a \fBcondition\fP.
.IN Monitors "\fWCondition\fP type"
.pp
A \fBdata-abstraction\fP device is a programming construct that separates 
.IN "Data abstraction"
operations on data from their implementation details, thus providing new
function while hiding unnecessary structure.
(We have seen abstraction as a form of the Beautification Principle.) \&
.IN "Abstraction"
.IN "Beautification Principle"
In languages that provide such devices, they usually have the form of a new
name scope (a block, in the Algol sense) that may export types and procedures
to the outside world but uses local variables to store state and perform
necessary computations.
Monitors have this structure, too, although they are principally used as a
mechanism for \fIshared\fP data abstractions.
They export procedures, which we will call \fBguard\fP procedures, that may be
called from outside the monitor.
.IN Monitors "\fWguard\fP procedures"
No matter how many activities are running, only one is allowed to execute a
guard procedure at a time.
.pp
The most straightforward use of monitors is to package all routines
that use a set of shared data into a single monitor.
All accesses to those data will be forced to use guard procedures because the
data themselves are hidden from the outside world.
For example, if we wish to implement a counter that records the number of times
some interesting event has happened, and different activities will wish to
modify and inspect the counter, we could build the following monitor.
The syntax chosen here is meant to convey the ideas without introducing
extraneous concepts; therefore, it differs from Modula, which also provides
monitors with a different syntax.
.(l
.FS
.ne 7
 1	\fWmonitor\fP Counter;
 2	
 3	\fWexport\fP RaiseCount, ReadCount;
 4	
 5	\fWvar\fP
 6		Count : integer;
 7	
.ne 5
 8	\fWguard\fP \fWprocedure\fP RaiseCount;
 9	\fWbegin\fP
10		Count := Count + 1;
11	\fWend\fP RaiseCount;
12	
.ne 8
13	\fWguard procedure\fP ReadCount : integer;
14	\fWbegin\fP
15		ReadCount := Count;
16	\fWend\fP ReadCount;
17	
18	\fWbegin\fP { initialize }
19		Count := 0;
20	\fWend\fP Counter.
.FE
.)l
One way to picture the monitor is shown in Figure \n(ch.\n+(FN, which shows
the monitor as a floor plan of a building.
.(z
.GS
height 1.5
file figs/fig.8.monitor
.GE
.FI "A simple monitor"
.)z
When activities wish to invoke guard procedures, they must enter through the
entry queue, where they are blocked until the guard procedures are free of any
activity.
Entrance 2 is unlocked only if there is no activity in the main room.
Door 1 is always unlocked; when it is opened to let an activity
out, door 2 is unlocked.
.pp
It is not hard to implement this kind of monitor using only semaphores.
A compiler could translate guard procedures into the following code:
.(b
.FS
	Down(MonitorEmpty);
		guard procedure body
	Up(MonitorEmpty);
.FE
.)b
The fact that the programmer does not need to remember the final Up
operation makes monitors easier and safer to use than bare semaphores.
The fact that all the code that can affect the shared variables is packaged in
one place also makes it easier to check that the variables are properly used.
.pp
It is not hard to find situations in which the simple monitors we have shown so
far are insufficient.
As it stands, they have no facility for synchronization.
The \fBproducers-consumers\fP problem is a good example of a situation where
both mutual exclusion and synchronization are needed.
.IN "Producers-consumers problem"
.(q
.ce
\fBProducers-consumers problem\fP
.sp
Any number of producer activities and consumer activities are running.
At any time, a producer activity may create some data.
At any time, a consumer activity may want to accept some data.
The data should be saved in a buffer until they are needed.
Since the buffer is finite, we want a producer to block if its new creation
would overflow the buffer.
We also want a consumer to block if there are no data available when it wants
them.
Data should be accepted in the order in which they are produced, although we
relax this rule somewhat for data produced or consumed simultaneously.
.)q
We can use a monitor to implement a \fBbounded buffer\fP shared among
all producers and consumers.
.IN "Bounded buffers"
The data-abstraction features of the monitor will collect the BufferPut and
BufferGet operations in one place, so the program is easy to debug and maintain.
The mutual-exclusion properties of the monitor will ensure that two producers
don't attempt to modify the buffer at the same time and, more generally, that
simultaneous conflicting uses are prevented.
Synchronization features will be used to block producers and consumers when the
buffer is full or empty.
.pp
Here is a monitor implementation of bounded buffers:
.IN "Producers-consumers problem" "Monitor solution"
.(l
.FS
.ne 15
 1	\fWmonitor\fP BoundedBuffer;
 2	
 3	\fWexport\fP GetBuffer, PutBuffer;
 4	
 5	\fWconst\fP
 6		Size = 10; { number of elements in the buffer at one time }
 7	\fWtype\fP
 8		Datum = ... { the data type of the contents of the buffer }
 9	\fWvar\fP
10		Buffer : \fWarray\fP 0:Size-1 \fWof\fP Datum;
11		Count : 0..Size; { how many elements currently in the buffer }
12		NextIn, NextOut : 0..Size-1; 
13			{ index of next datum to place in the buffer or remove }
14		NonEmpty, NonFull : \fWcondition\fP;
15	
.ne 11
16	\fWguard\fP \fWprocedure\fP PutBuffer(What : Datum);
17	\fWbegin\fP
18		\fWif\fP Count = Size \fWthen\fP
19			\fWwait\fP NonFull;
20		\fWend\fP;
21		Buffer[NextIn] := What;
22		NextIn := (NextIn + 1) \fWmod\fP Size;
23		inc(Count);
24		\fWsignal\fP NonEmpty;
25	\fWend\fP PutBuffer;
26	
.ne 11
27	\fWguard\fP \fWprocedure\fP GetBuffer(\fWvar\fP Result : Datum);
28	\fWbegin\fP
29		\fWif\fP Count = 0 \fWthen\fP
30			\fWwait\fP NonEmpty;
31		\fWend\fP;
32		Result := Buffer[NextOut];
33		NextOut := (NextOut + 1) \fWmod\fP Size;
34		dec(Count);
35		\fWsignal\fP NonFull;
36	\fWend\fP GetBuffer;
37	
.ne 5
38	\fWbegin\fP { initialization }
39		Count := 0;
40		NextIn := 0;
41		NextOut := 0;
42	\fWend\fP Buffer
.FE
.)l
.pp
This representation of bounded buffers satisfies our goals nicely.
We need not worry that some other piece of program modifies the shared
variables Buffer and Count because they are visible only within this monitor.
There is no danger that several activities will try to modify Count
simultaneously, will produce into the same cell of Buffer simultaneously, or
will take data from the same cell of Buffer simultaneously, because all such
behavior is restricted to guard procedures, which are mutually exclusive.
.pp
Finally, we have provided a way for activities to be blocked if necessary.
The mechanism is the new (predefined) type called \fWcondition\fP.
Both NonEmpty and NonFull are declared as conditions in line 14.
Let us concentrate on the condition NonEmpty.
At line 30, a consumer that finds the buffer empty executes a \fWwait\fP on
this condition.
.IN Monitors "\fWwait\fP statement"
The \fWwait\fP will block this activity until some other activity signals the
condition.
At that time, the blocked activity may continue.
When it gets to line 32, it assumes that the buffer is no longer empty.
Any producer that succeeds in placing new data in the buffer executes a
\fWsignal\fP on the NonEmpty condition at line 24.
.IN Monitors "\fWsignal\fP statement"
This operation unblocks a waiting consumer.
.pp
The foregoing discussion raises some troubling questions.
Exactly when does the blocked consumer continue?
If immediately, then we may have two activities in the monitor at once, and
our careful mutual exclusion is ruined.
If later, then by the time it continues, some other consumer may already
have taken the last datum, and the assumption on line 32 that the buffer is not
empty is wrong.
If several consumers are waiting at the same time, which one or ones are
unblocked by a signal?
If a producer executes a \fWsignal\fP when no consumers are waiting, is this
signal stored or discarded?
.pp
Not all definitions of monitors in the literature agree on the answers to these
questions.
We will first present a common approach and then discuss alternatives.
Figure \n(ch.\n+(FN expands the preceding figure to show the effect of
conditions.
.(z
.GS
height 4.5
width 4
file figs/fig.8.fullmonitor
.GE
.FI "Monitors with conditions"
.)z
For every condition we build a condition queue (shown on the bottom of the
monitor).
We also introduce one urgent queue (shown at the top of the monitor).
.IN Monitors queues
As before, we insist that at most one activity inhabit the guard procedures (the
center of the monitor) at any time.
Activities that are blocked will be placed in these queues.
Here are the rules that govern the doors in and out of the queues and the
central region:
.sp 10p
.np
New activities wait in the entry queue.
A new activity may enter through door 2 if no activity is currently in the
central region.
.np
If an activity leaves the central region through door 1 (the exit),
one activity is allowed in through door 4 (from the urgent queue) if
there is one waiting there.
If not, one activity is allowed through door 2 (from the entry queue) if there
is one waiting there.
.np
An activity that executes a \fWwait\fP enters the door to the appropriate
condition queue (for example, 5 or 7).
.np
When an activity executes \fWsignal\fP, the signaled condition queue is
inspected.
If some activity is waiting in that queue, the signaler enters the
urgent queue (door 3) and one waiter is allowed into the central region
(door 6 or 8).
If no activity is waiting in that queue, the signaler proceeds without leaving
the central region.
The signal is ignored.
.np
All queues are ordered first in, first out.
.sp 10p
.lp
.ne 1.5i \" bug override
These rules assure that a waiting consumer is unblocked immediately when a
producer signals NonEmpty and that the producer is blocked in the
urgent queue until the consumer has taken the datum.
We have maintained our rule that at most one activity inhabits the central
region.
.pp
People who use monitors have noticed that the signal operation is almost always
the last operation performed in a guard procedure.
We see this behavior in our producer-consumer solution (lines 24 and 35).
Our rules will often make the signaler wait in the urgent queue and
then return to the central region just to get out of the monitor altogether.
This scenario violates the Hysteresis Principle.
.IN "Hysteresis Principle"
Furthermore, if the signaler does have more work to do, it can't assume that
the situation of shared variables is the same when it returns from the
urgent queue because the activity that was unblocked is likely to change
them.
For these reasons, some people require that signal must be the last
operation of a guard procedure and must cause the signaling activity to leave
the monitor.
Then we don't need an urgent queue, and signalers never make invalid
assumptions about shared data.
However, this restriction makes monitors less flexible.
(Such monitors are strictly less powerful in a theoretical sense.)\ \&
.pp
A related suggestion is that \fWbroadcast\fP should be like \fWsignal\fP
.IN Monitors Broadcast
but should release \fIall\fP the members of the associated condition queue.
Since they can't all be allowed into the central region at once, most are
placed in the urgent queue.
In this way, when conditions are right, all the waiters may be released instead
of the awkward alternative of
having each released waiter releasing the next one in turn.
However, a released activity can no longer assume that the condition it has
awaited is truly met by the time it resumes.
Programs written with \fWbroadcast\fP usually replace the \fWif\fP statement
in lines 29 through 31 above with a \fWwhile\fP loop: 
.(b
.FS
29		\fWwhile\fP Count = 0 \fWdo\fP
30			\fWwait\fP NonEmpty;
31		\fWend\fP;
.FE
.)b
.pp
Proper use of monitors follows the guideline that no activity should take too
long in the central region.
It shouldn't take too long for an activity that is waiting in the entry queue
to get into the monitor and access the shared variables.
Any lengthy operation should relax exclusion by entering a condition queue or 
by doing its work outside the monitor.
A fascinating violation of this guideline arises if the activity in a guard
procedure makes a call on a guard procedure in another monitor.
.IN Monitors "Nested calls"
Under our rules, this activity is still considered to be in the first monitor,
preventing any other activity from entering.
However, it may take a long time before it returns because it may be forced to
wait in the second monitor in a condition queue.
(By our guideline, it shouldn't have to wait very long in either the entry
queue or the urgent queue.)\ \&
This delay in returning violates our guideline with respect to the first
monitor.
It can even lead to deadlock if the condition it awaits in the second monitor
can be signaled only by an activity that is currently waiting patiently to
enter the first monitor.
.pp
A number of solutions to this problem have been proposed.
.sp 10p
.np
Warn the programmer but allow the bad situation to develop.
That is, nested monitor calls maintain exclusion on the old monitor while in
the new one.
.np
Disallow nested monitor calls.
.np
Release exclusion on the old monitor and enforce it only on the new one.
When the activity is ready to return, it must wait in the urgent queue
of the first monitor until it can once again achieve exclusive use of the
central region.
.np
Let the programmer decide whether the nested call should
maintain exclusion in the old monitor or not.
By default, method 1 is used.
The programmer can say \fWduckout\fP to release exclusion while still in the
monitor and \fWduckin\fP to acquire exclusion again.
.IN "\fWduckout\fP" "Monitor operation"
.IN "\fWduckin\fP" "Monitor operation"
These calls can bracket a nested call to simulate method 3.
.sp 10p
.pp
Although monitors represent a significant advance over the earlier devices we
have studied, they do have some significant problems.
Monitors have been criticized for not providing any control over how the queues
are ordered.
The policy to treat them in first-in, first-out order is not always
appropriate.
For example, several activities simultaneously in the urgent queue are
like nested interrupts, which are usually released in first-in, last-out
(stack) order.
As another example, different waiters in a condition queue may have different
priorities, which could be taken into account in selecting an order.
Some people prefer a more general mechanism for
inspecting and reordering the various queues.
.pp
Monitors also display unexpected complexity with respect to nested calls.
It is not easy to describe the semantics of \fWwait\fP and \fWsignal\fP without
resorting to pictures like Figure \n(ch.\n(FN.
Complexity is also introduced by the artificial use of condition variables.
The programmer is more likely to understand the underlying condition (like
.q "Count > 0" )
than to represent that condition properly by judicious use of \fWwait\fP and
\fWsignal\fP.
.pp
Another objection to monitors comes from their data-abstraction ability.
If we have several bounded buffers to implement, we would be tempted to build
only one monitor and to have the PutBuffer and GetBuffer procedures take a
parameter that describes which buffer is to be manipulated.
This solution has two drawbacks.
One is that the buffer has an existence outside the monitor and so might be
inadvertently modified by a non-monitor procedure.
Luckily, modern programming languages have the idea of
.q "exported types"
that exist outside a data-abstraction module but can be manipulated only from
within that module.
The other drawback is that using only one monitor is too conservative:
Every manipulation of one buffer now excludes
operations on all other buffers because mutual exclusion is governed by which
monitor is to be entered, not by which data structure is being accessed.
Again, modern programming languages come to the rescue.
One can imagine a
.q "monitor type"
for bounded buffers.
A new monitor of that type is built for each buffer needed.
The program for monitors needs to be written only once, but each buffer gets
its own monitor with its own exclusion.
.pp
A serious objection to monitors is related to our guideline that exclusion
should not be in force for very long.
However, shared data might be needed for a very long time.
This is exactly the situation in the readers-writers problem, mentioned in our
chapter on file structures.
.IN "Readers-writers problem"
.(q
.ce
\fBReaders-writers problem\fP
.sp
Any number of reader activities and writer activities are running.
They share some data (for example, a file).
At any time, a reader activity may wish to read data from the file.
At any time, a writer activity may want to read or modify the data in the file.
Reading and writing are time-consuming operations, but they always finish
eventually.
During the time a writer is writing, no other reader or writer may access the
shared data.
Any number of readers may access the data simultaneously.
For fairness, we prefer that no reader or writer be \fBstarved\fP (that is,
forced to wait indefinitely).
.)q
If we make Read and Write guard procedures, two readers cannot execute at
the same time, a restriction that violates the spirit of the problem.
Therefore, Read must not be a guard procedure but rather a procedure external
to the monitor.
Proper use of Read would call the guard procedures StartRead and EndRead around
calls to Read, but there is no assurance that a reader will follow these rules.
Monitors therefore fail to protect the shared data adequately.
.SH 2 "Crowd monitors" I
.lp
One nice extension to monitors to avoid this last problem is called
\fBcrowd monitors\fP.
Crowd monitors distinguish guard procedures from ordinary procedures within the
monitor.
Only guard procedures are mutually exclusive.
Ordinary procedures may be invoked only by activities that have permission to
do so; this permission is granted and revoked by guard procedures.
A skeleton of the crowd-monitor solution to the readers-writers problem is as
follows:
.IN "Readers-writers problem" "Crowd-monitor solution"
.(l
.FS
.ne 8
 1	\fWcrowd monitor\fP ReadWrite;
 2	
 3	\fWexport\fP StartRead, EndRead, Read, StartWrite, EndWrite,
 4		Write;
 5	\fWvar\fP
 6		Readers : \fWcrowd\fP Read;
 7		Writers : \fWcrowd\fP Read, Write;
 8	
.ne 6
 9	\fWguard\fP \fWprocedure\fP StartRead;
10	\fWbegin\fP
11		... { block the caller until reading is safe }
12		\fWenter\fP Readers;
13		...
14	\fWend\fP StartRead;
15	
.ne 6
16	\fWguard\fP \fWprocedure\fP EndRead;
17	\fWbegin\fP
18		\fWleave\fP Readers;
19		... { bookkeeping, maybe signal a waiting writer }
20	\fWend\fP EndRead;
21	
.ne 6
22	\fWguard\fP \fWprocedure\fP StartWrite;
23	\fWbegin\fP
24		... { block the caller until writing is safe }
25		\fWenter\fP Writers;
26		...
27	\fWend\fP StartWrite;
28	
.ne 6
29	\fWguard\fP \fWprocedure\fP EndWrite;
30	\fWbegin\fP
31		\fWleave\fP Writers;
32		... { bookkeeping, maybe signal waiter }
33	\fWend\fP EndWrite;
34	
.ne 5
35	\fWprocedure\fP Read;
36	\fWbegin\fP
37		... { actually read from the shared data }
38	\fWend\fP Read;
39	
.ne 6
40	\fWprocedure\fP Write;
41	\fWbegin\fP
42		... { actually modify the shared data }
43	\fWend\fP Write;
44	
45	\fWend\fP ReadWrite
.FE
.)l
In lines 6 and 7, we declare two \fBcrowds\fP called Readers and Writers.
.IN "Crowd monitors" Crowds
Activities can be members of these crowds.
Any member of Readers may access the Read procedure (lines 35\(mi38),
and any member of Writers may access both the Read and the Write procedure
(lines 40\(mi43).
Activities initially belong to no crowds.
The guard procedures, which are mutually exclusive as before, decide when it is
appropriate for an activity to enter or leave a crowd.
They may use conditions to wait for the right situation.
When the guard procedure decides to let a reader proceed, it executes
\fWenter\fP for the Readers crowd (line 12).
.IN "Crowd monitors" "\fWenter\fP statement"
Similarly, a guard can let a writer enter the Writers crowd (line 25).
Although any activity may call Read and Write, since they are exported from the
monitor, a run-time check prevents activities from calling ordinary procedures
that are invisible to the crowds they are in.
A member only of Readers may not call Write.
However, a member of Writers may call either Read or Write, since both are
specified in the definition of Writers (line 7).
.pp
It is easy to implement crowd monitors once regular monitors have been
implemented.
A run-time check is applied every time an activity calls a procedure like Read
and Write in order to make sure that the activity is in the necessary crowd.
If not, a program error has been discovered.
No new blocking or unblocking rules are needed.
.SH 2 "Event counts and sequencers" I
.lp
So far we have investigated methods that serve for both mutual exclusion and
synchronization.
We now turn to a technique to achieve synchronization without requiring
mutual exclusion.
Mutual exclusion is not always desirable because it limits concurrency.
It is also difficult to enforce, as we have seen, and is not always needed on
physically distributed computers.
In fact, if one hasn't yet implemented mutual exclusion, the method
we are about to discuss can be used to build semaphores to provide mutual
exclusion, too.
The semaphores we can build will even allow simultaneous Down operations on
several semaphores.
.pp
The first tool we will introduce is the \fBevent count\fP.
An event count is represented as a non-decreasing integer variable, initially 0.
It keeps a count of the number of events of interest to the program,
such as the number of times a variable has been modified.
As an abstract data type, event counts have three operations:
Advance, Read, and Await.
.sp 10p
.ip \(bu
Advance(E)
.IN "Event count" "Advance operation"
is used to signal the occurrence of events associated with event count E.
It has the effect of incrementing E atomically.
.ip \(bu
Read(E)
.IN "Event count" "Read operation"
returns the value of the event count E.
If Read returns some number \fIn\fP, then at least \fIn\fP Advance operations
must have happened.
By the time this number is returned, the event count may have been advanced
again a number of times.
.ip \(bu
Await(E \fWreaches\fP \fIv\fP)
.IN "Event count" "Await operation"
waits for the event count E to have the value \fIv\fP.
It blocks the calling activity until at least \fIv\fP Advance operations have
occurred.
It is acceptable if more than \fIv\fP Advance operations
have occurred when it is finally unblocked.
This
.q overshoot
could result from very frequent Advance operations.
(We adopt the syntax shown here for clarity.
Await could be a procedure with two parameters, using a comma instead of
\fWreaches\fP.)
.sp 10p
.lp
These definitions allow both Await and Read to be concurrent with Advance,
since we don't care if Read gives us a somewhat stale value or if Await waits
a trifle too long.
.pp
We introduced the producers-consumers problem earlier to show how monitors
handle synchronization.
Here is a solution using event counts.
For the time being, we assume that there is only one producer and one consumer.
.IN "Producers-consumers problem" "Event-count solution"
.(l
.FS
.ne 8
 1	\fWconst\fP
 2		Size = 10; { number of elements in the buffer. }
 3	\fWtype\fP
 4		Datum = ... { data type for buffer contents. }
 5	\fWvar\fP
 6		In, Out : \fWeventcount\fP;
 7		Buffer  : \fWarray\fP 0:Size-1 \fWof\fP Datum;
 8	
.ne 13
 9	\fWactivity\fP Producer;
10	\fWvar\fP	
11		SequenceNumber : integer;
12		Value : Datum;
13	\fWbegin\fP
14		\fWfor\fP SequenceNumber := 1 \fWto\fP infinity \fWdo\fP
15			... { compute Value }
16			Await(Out \fWreaches\fP SequenceNumber \(mi Size);
17			Buffer[SequenceNumber \fWmod\fP Size] := Value;
18			Advance(In)
19		\fWend\fP
20	\fWend\fP Producer;
21	
.ne 12
22	\fWactivity\fP Consumer;
23	\fWvar\fP	
24		SequenceNumber : integer;
25		Value : Datum;
26	\fWbegin\fP
27		\fWfor\fP SequenceNumber := 1 \fWto\fP infinity \fWdo\fP
28			Await(In \fWreaches\fP SequenceNumber);
29			Value := Buffer[SequenceNumber \fWmod\fP Size];
30			Advance(Out)
31			... { use the Value }
32		\fWend\fP
33	\fWend\fP Consumer;
.FE
.)l
.lp
There is no need for us to worry that the consumer and producer will
simultaneously access the same cell in Buffer, since the producer will have to
wait until the consumer has used that cell before its Await in line 16 will
allow it to proceed.
Similarly, the consumer knows that when it accesses a cell of Buffer, the
producer must have placed data there, or the Await in line 28 would not have
unblocked.
In addition, we need not worry that both Advance operations (lines 18 and 30)
will happen at the same time, because they deal with different event counts.
This solution allows the bounded buffer to be used simultaneously by both
activities because it guarantees that the very same datum will never be touched
by both at once.
.pp
The other tool we will use is the \fBsequencer\fP,
which assigns an arbitrary order to unordered events.
We will use this order to decide which activity gets a resource first.
A sequencer is implemented as a non-decreasing integer variable, initialized
to 0.
It has only one operation: Ticket.
.sp 10p
.ip \(bu
Ticket(S)
.IN Sequencers "Ticket operation"
returns the current value of the sequencer S (initially 0) and then increments
S.
This operation is atomic.
We need some form of mutual exclusion to implement Ticket because no two
calls may return the same value.
.sp 10p
.pp
Now we can implement a producer-consumer situation in which there are many
producers.
For simplicity, we will still have only one consumer.
As before, consumption and production need not exclude each other.
However, two producers must take turns to make sure that they don't write into
the same cell in the Buffer.
Here is the new producer program.
.(l
.FS
.ne 8
 1	...
 2	\fWvar\fP
 3		Turn : \fWsequencer\fP;
 4	
 5	\fWactivity\fP Producer;
 6	\fWvar\fP	
 7		SequenceNumber : integer;
 8		Value : Datum;
.ne 11
 9	\fWbegin\fP
10		\fWloop\fP
11			... { compute Value }
12			SequenceNumber := Ticket(Turn);	{ discover turn }
13			Await(In \fWreaches\fP SequenceNumber);	{ wait for turn }
14			Await(Out \fWreaches\fP SequenceNumber \(mi Size + 1);
15				{ wait for Buffer }
16			Buffer[SequenceNumber \fWmod\fP Size] := Value;
17			Advance(In);
18		\fWend\fP;
19	\fWend\fP Producer;
.FE
.)l
Each producer must wait until it is its turn to produce.
The call to Ticket in line 12 orders active producers.
Usually there will be no wait in line 13, unless another producer has just
grabbed an earlier ticket and has not yet gotten to line 17.
The Await in line 14 is to make sure that the cell in Buffer that is about to
be overwritten has been consumed.
It differs from line 16 of the previous example by 1, since here SequenceNumber
starts at 0, not 1.
The Advance in line 17 tells waiting consumers that this cell in Buffer may be
consumed, and it tells waiting producers that we have finished our turn.
.pp
The call to Await in line 13 might seem unnecessary.
It's there to make sure that producers write cells of Buffer in order,
so that consumers may assume that when In is advanced in line 17, the next cell
of Buffer has new data.
Unfortunately, one effect of this imposed sequential behavior on producers is
that separate cells of Buffer cannot be written simultaneously.
If the data are large, producers may exclude each other for a long time.
.pp
We promised earlier that a semaphore could be built with event counts and
sequencers.
Here is an implementation:
.IN Semaphore "event counts and sequencers implementation"
.(l
.FS
.ne 8
 1	\fWtype\fP
 2		Semaphore =
 3			\fWrecord\fP
 4				T : \fWsequencer\fP;
 5				E : \fWeventcount\fP;
 6				I : integer; { initial value }
 7			\fWend\fP;
 8	
.ne 6
 9	\fWprocedure\fP Down(\fWvar\fP S : Semaphore);
10	\fWbegin\fP
11		\fWwith\fP S \fWdo\fP
12			Await(E \fWreaches\fP Ticket(T)\(miI+1);
13		\fWend\fP
14	\fWend\fP Down;
15	
.ne 4
16	\fWprocedure\fP Up(S : Semaphore);
17	\fWbegin\fP
18		Advance(S.E);
19	\fWend\fP Up;
.FE
.)l
The Down procedure gets a ticket (which orders all the waiters) and then
waits for the event count to be advanced as many times as the ticket indicates.
There is no need for Down and Up to exclude each other.
However, we require that every waiter get a different ticket.
.pp
We can extend this wait procedure to wait for two semaphores at once.
The dining-philosophers problem, described in Chapter 4, can be solved by such
a method.
Here is a procedure that waits for two semaphores.
It uses a third semaphore, MutualExclusion, for its own purposes.
.(l
.FS
.ne 6
 1	\fWvar\fP
 2		MutualExclusion : Semaphore; { initialized at I = 1 }
 3	
 4	\fWprocedure\fP DoubleDown(\fWvar\fP S1, S2 : Semaphore);
 5	\fWvar\fP
 6		Ticket1, Ticket2 : integer;
.ne 12
 7	\fWbegin\fP
 8		Down(MutualExclusion);
 9		Ticket1 := Ticket(S1.T);
10		Ticket2 := Ticket(S2.T);
11		Up(MutualExclusion);
12		\fWwith\fP S1 \fWdo\fP
13			Await(E \fWreaches\fP Ticket1);
14		\fWend\fP;
15		\fWwith\fP S2 \fWdo\fP
16			Await(E \fWreaches\fP Ticket2);
17		\fWend\fP;
18	\fWend\fP DoubleDown;	
.FE
.)l
.pp
We use the MutualExclusion semaphore in lines 8\(mi11 to make sure that
we acquire the two tickets as an atomic action.
It makes no difference in what order we wait for S1 and S2 (lines 13 and 16).
