.SH 2 "Path expressions" I
.lp
A completely different way to describe synchronization is to spell out what
routines may be called, and in what order, separately from the routines
themselves.
We will introduce a syntax known as \fBpath expressions\fP that specifies all
temporal relationships among invocations of procedures that control shared
data.
These expressions are built out of procedure names and connectors that indicate
restrictions.
For our examples, we will deal with the readers-writers problem and use
procedures such as Read, Write, and StartRead.
.pp
The simplest expression has just one procedure name:
.(b
.FS
	\fWpath\fP Read
.FE
.)b
This expression means that only one activity may call Read at a time
but that when Read is finished, another activity (or the same one again)
may call it again.
That is, when the path expression is finished, it may be started again.
If we want to specify that Read must be surrounded by StartRead and EndRead,
we can say
.(b
.FS
	\fWpath\fP StartRead ; Read ; EndRead
.FE
.)b
The semicolon operator indicates that after one procedure has finished,
.IN "Path expressions" "semicolon operator"
the next one may be called.
Equivalently, this operator indicates that before a particular procedure may be
called, the one before it must have finished.
However, it does not require that the same activity that calls StartRead
must call Read.
.pp
To indicate that any number of activities can simultaneously perform the same
procedure, we surround that procedure in brackets, as shown here:
.(b
.FS
	\fWpath\fP { Read }
.FE
.)b
.IN "Path expressions" "brackets"
This path expression says that any number of simultaneous uses of Read are
permissible.
If some activities are in Read, a new one may join them.
When they all finish, the path expression is finished and may be started again.
.pp
The last operator is the
.q "either-or"
operator, which we will write as a vertical bar, as in the following example:
.(b
.FS
	\fWpath\fP Read | Write
.FE
.)b
.IN "Path expressions" "Either-or operator"
This operator means
.q "the left side or the right side may be active, but not both." " "
This particular path expression specifies that one activity may read at a time
and one may write at a time, and reading and writing exclude each other.
After either finishes, any activity may start doing either again.
We assume that activities that wish to Read or Write are served in arrival
order; no activity is caused to wait forever for an either-or operator.
.pp
We can combine these operators to form more complex expressions.
For example, the following begins to address the actual readers-writers
problem:
.IN "Readers-writers problem" "Path-expression solution"
.(b
.FS
	\fWpath\fP { Read } | Write
.FE
.)b
This expression lets any number of readers alternate with a single writer.
Although this expression represents exactly the restriction we want, it is not
fair to writers.
If reader activities arrive fast enough so that there is always some activity
in the
.q "{\0Read\0}"
part of the path expression, writers will never be allowed to start.
.pp
This solution also has the danger that a reader will want to perform several
reads in a row.
A writer could be serviced between two reads, and inconsistent data would be
read.
To avoid this problem, we will assume that each reader wants to read only once.
Similarly, we will assume that each writer wants to write only once.
Otherwise, where we say Read, we could understand that to mean
.q "perform any number of read operations in the same activity." " "
Likewise, we can understand Write to mean
.q "perform any number of read or write operations in the same activity." " "
.pp
To represent fair synchronization rules, we need some more tools.
First, we will allow several path expressions to be in force at the same time,
as long as they don't mention the same procedures.
Next, we will want to introduce intermediate procedures.
For simplicity, we will use the following notation for these procedures:
.(b
.FS
	\fWlet\fP procedure = sequence of procedures
.FE
.)b
.pp
Let's illustrate this idea with a more complex readers-writers example.
.(b
.FS
 1	\fWpath\fP TryRead | TryWrite
 2	\fWpath\fP { StartRead; Read } | Write
 3	\fWlet\fP TryRead = StartRead
 4	\fWlet\fP TryWrite = Write
 5	\fWlet\fP DoRead = TryRead ; Read
 6	\fWlet\fP DoWrite = TryWrite
.FE
.)b
Readers call DoRead, which expands to \fLTryRead ; Read\fP
(line 5), which in turn becomes \fLStartRead ; Read\fP (line 3).
Similarly, writers call DoWrite, which becomes TryWrite
(line 6), which in turn becomes Write (line 4).
The StartRead procedure does not need to do any work; it is called only to
allow us to synchronize the rest of the operations.
Line 2 expresses the constraints we want on simultaneous readers and writers.
Line 1 tells us that activities must wait their turn to start TryRead and
TryWrite.
.pp
This set of path expressions works, in that it properly excludes readers and
writers from each other.
It also enforces a policy, in that it takes readers and writers in the order
in which they appear, except that readers that arrive next to each
other in line are allowed to overlap their executions.
It is not easy to be convinced, however, that this policy comes out of these
expressions.
.pp
Implementing path expressions is not difficult; they can be converted to
semaphore operations.
However, we will not examine the conversion because it is complex.
Luckily, a compiler can be expected to perform this work, so it can be done
before the program is run, and the programmer need not deal with its
intricacies.
.pp
Since path expressions have the advantage of separating the synchronization
specification from the routines themselves, the procedures that accomplish
work (like Read and Write) are uncluttered by extraneous considerations.
Path expressions also avoid specialized guard procedures whose only purpose
is to effect mutual exclusion and synchronization.
However, the procedures we introduced with the \fWlet\fP clause serve much the
same function.
.pp
Like monitors, path expressions are not descriptive enough to describe the
order in which blocked activities should be resumed.
The either-or operator unblocks activities in the order in which they arrive,
which may not be the order desired.
.pp
Let's step back from the complex synchronization problems we have been
examining and try Figure \n(ch.\n(F1.
We can certainly impose extra constraints and require that all the activities
follow a prescribed order.
If we wish to implement Figure \n(ch.\n(F1 without extra
constraints, we will have to let the same procedure appear in several path
expressions.
We then get the following solution:
.(b
.FS
 1	\fWpath\fP A ; B ; C ; F ; I
 2	\fWpath\fP D ; E ; F
 3	\fWpath\fP A ; D ; G ; H ; I
 4	\fWpath\fP B ; E ; H
.FE
.)b
.SH 2 "Invariant expressions" I
.lp
The method of \fBinvariant expressions\fP is reminiscent of path expressions.
We queue all activities that try to enter each procedure.
An invariant expression associated with each queue tells when it is
permissible to allow the first waiter in that queue into the crowd.
If there is no invariant expression for some procedure, activities may
enter it at any time.
.pp
Invariant expressions have a very limited syntax.
They may use built-in counters that record the number of interesting events for
any procedure P.
There are only five such counters:
.sp 10p
.ip \(bu
RequestCount(P) counts the number of requests (since the beginning of time)
for procedure P.
.ip \(bu
StartCount(P) counts the number of times procedure P has actually been
started.
.ip \(bu
FinishCount(P) counts the number of times procedure P has been
finished.
.ip \(bu
CurrentCount(P) counts the number of current executions of P; it is equivalent
to Start\%Count(P)\(miFinish\%Count(P).
.ip \(bu
WaitCount(P) counts the number of waiting executions of P; it is equivalent
to Request\%Count(P)\(miStart\%Count(P).
.sp 10p
.lp
An invariant expression has the form
.(b
.FS
	expression    comparison    constant
.FE
.)b
Expressions are just sums and differences of counters.
The comparison can be <, >, \(<=, \(>=, \(eq or \(!=.
For example, we might write
.(b
.FS
	WaitCount(P) \(mi RequestCount(Q) > 4
.FE
.)b
We associate such an expression with a procedure as a rule that must hold
before we allow an activity to execute that procedure.
(This interpretation is not the same as saying the rule must hold \fIwhile\fP
the activity is executing the procedure.)\ \&
The following is a solution to the multiple producer-consumer problem:
.IN "Producers-consumers problem" "Invariant-expression solution"
.(l
.FS
.ne 8
 1	\fWmodule\fP BoundedBuffer;
 2	
 3	\fWexport\fP GetBuffer, PutBuffer;
 4	
 5	\fWconst\fP
 6		Size = 10; { number of elements in the buffer at once }
 7	\fWtype\fP
 8		Datum = ... { data type of the contents of the buffer }
.ne 4
 9	\fWvar\fP
10		Buffer : \fWarray\fP 0:Size\(mi1 \fWof\fP Datum;
11		NextIn, NextOut : 0..Size\(mi1;
12			{ index of next datum to insert, delete }
.ne 4
13	\fWinvariant\fP PutBuffer
14		StartCount(PutBuffer) \(mi FinishCount(GetBuffer) < Size
15			{ no overproduction }
16		CurrentCount(PutBuffer) = 0 { prevent producer conflict }
.ne 5
17	\fWinvariant\fP GetBuffer
18		StartCount(GetBuffer) \(mi FinishCount(PutBuffer) < 0
19			{ no overconsumption }
20		CurrentCount(GetBuffer) = 0 { prevent consumer conflict }
21	
.ne 6
22	\fWprocedure\fP PutBuffer(What : Datum);
23	\fWbegin\fP
24		Buffer[NextIn] := What;
25		NextIn := (NextIn + 1) \fWmod\fP Size;
26	\fWend\fP PutBuffer;
27	
.ne 5
28	\fWprocedure\fP GetBuffer(\fWvar\fP Result : Datum);
29	\fWbegin\fP
30		Result := Buffer[NextOut];
31		NextOut := (NextOut + 1) \fWmod\fP Size;
32	\fWend\fP GetBuffer;
.FE
.)l
We have repeated the entire module to show how simple the PutBuffer and
GetBuffer procedures have become.
As with path expressions, we find that moving all the synchronization out of
the procedures clarifies the entire program.
Our invariant expressions are in lines 13 through 20.
Line 14 restricts calls to PutBuffer by requiring that enough calls to
GetBuffer have already happened so that the buffer won't be overfilled.
Similarly, line 18 restricts calls to GetBuffer by requiring that enough calls
to PutBuffer have already happened so that the buffer has some data to
be removed.
Lines 16 and 20 restrict both PutBuffer and GetBuffer to allow only one
activity in either.
.pp
We can also use invariant expressions to represent a solution to the
readers-writers problem.
We will show just the invariants and omit the rest of the module.
.IN "Readers-writers" "invariant-expression solution"
.(b
.FS
\fWinvariant\fP Read
	CurrentCount(Write) = 0
\fWinvariant\fP Write
	CurrentCount(Write) + CurrentCount(Read) = 0
.FE
.)b
We might make the same assumption we did for path expressions: that
activities forced to wait because of an invariant are unblocked, if possible,
in arrival order.
Therefore, if several readers and writers arrive while a writer is active,
all are blocked.
When the writer finishes, one reader or writer, whichever came first, is
allowed to start.
However, while one activity is reading, any number of readers may start;
only writers are blocked.
.pp
The solution just shown may starve writers.
To give writers precedence, we need to prevent new readers from
starting if there are any waiting writers.
The following invariant expressions serve this purpose:
.(b
.FS
\fWinvariant\fP Read
	WaitCount(Write) + CurrentCount(Write) = 0
\fWinvariant\fP Write
	CurrentCount(Write) + CurrentCount(Read) = 0
.FE
.)b
The only change is an extra clause in the invariant expression for Read.
.pp
Invariant expressions are easily used to represent the synchronization
graph of Figure \n(ch.\n(F1.
For example, the invariant expression for procedure \fBE\fP would be as follows:
.(b
.FS
\fWinvariant\fP E
	FinishCount(B) + FinishCount(D) = 2
.FE
.)b
.pp
Invariant expressions, like path expressions and monitors, are unable to
reorder waiting queues to serve some activities out of turn.
However, they are often far clearer than path expressions and can express
the same sort of synchronization constraints.
.ig xx
.SH 2 "Protected queues" I
.lp
Often a simple mechanism is adequate for controlling concurrency.
One elegant and simple device is the \fBprotected queue\fP, which, like all
queues, has these operations:
.sp 10p
.ip CreateQueue(Q) 1i
builds a new protected queue Q.
.ip PutQueue(Q,E) 1i
puts the element E on the rear of queue Q.
PutQueue is atomic with respect to all operations on this queue undertaken by
other activities.
.ip GetQueue(Q,E) 1i
returns the first element E from queue Q, removing it from the queue.
GetQueue blocks the calling activity if the queue is empty, and
is atomic with respect to all operations on this queue undertaken by
other activities.
.sp 10p
.lp
.xx
.SH 1 "Perspective"
.lp
It is necessary to distinguish carefully between simple mutual-exclusion
situations, where it is dangerous for two activities both to proceed,
and synchronization situations, where complex conditions might govern when an
activity may proceed.
To show pure forms of both situations, we considered modifying shared
data (pure mutual exclusion) and Figure \n(ch.\n(F1 (pure synchronization).
Later, we turned to the producer-consumer problem and the readers-writers
problem as more complex examples.
These latter examples display aspects both of mutual exclusion (we don't want
two producers simultaneously writing into the buffer, and we don't want two
writers simultaneously active) and of synchronization (consumers must wait
until data are produced; readers should wait if there are waiting writers).
The situations that occur within operating systems are typically of this hybrid
variety.
.pp
The methods we have seen range from simple hardware techniques to
programming-language constructs that try to guarantee that the restrictions
described are clear, efficiently implementable, and unlikely to be misused.
The more complex solutions are invariably implemented by behind-the-scenes
work using a simpler solution.
For example, monitors can be implemented by semaphores.
Semaphores themselves need to turn on mutual exclusion around critical parts of
their implementation, and might do that with test-and-set locks.
At the bottom of this ladder always stands the inherent atomicity of main
store, which serves only one access at a time.
.pp
One might evaluate different proposals on the basis of three characteristics:
modularity, expressive power, and ease of use.
Modularity is achieved by data abstraction and by separating concurrency
.IN "Data abstraction"
control from resource access.
Monitors, crowd monitors, path expressions, and invariant expressions all
display a high degree of modularity.
.pp
Expressive power involves ability to formulate exclusion,
synchronization, and priority requirements.
Each of the methods displayed is capable of formulating the first two kinds of
requirement.
None does well with respect to priority because queues are built
into most of the methods without any mechanism for inspecting and reordering
those queues.
.pp
Requirements may depend on a host of criteria.
A partial list follows.
For each, we show how it applies to the producers-consumers problem.
.sp 10p
.ip \(bu
What operation is being requested:
PutBuffer and GetBuffer have different requirements.
.ip \(bu
When the request is made:
If two activities call PutBuffer, the one that called first should take
precedence (in the absence of priority).
.ip \(bu
The synchronization state of the resource:
The fact that an activity is currently executing GetBuffer implies that another
activity should be delayed in executing the same procedure.
.ip \(bu
The data state of the resource:
GetBuffer is blocked if the buffer is empty.
.ip \(bu
The history of accesses:
We block PutBuffer if there haven't been enough calls to GetBuffer.
We could pose this same constraint as part of the data state of the resource.
.ip \(bu
The parameters to the operation:
The data that are being put in the buffer do not influence our solution, but we
could imagine that some data are more crucial than others.
.ip \(bu
The identity of the caller:
We would not like a producer to call GetBuffer.
.sp 10p
.lp
Path expressions ignore the data state of the resource,
the parameters to the operation, and the identity of the caller.
Invariant expressions concentrate heavily on the history of accesses,
expressing other items of interest (like the synchronization state and the data
state) in terms of that history.
.pp
Ease of use is a subjective criterion.
Conditional critical regions are much easier to use than monitors, but they
are not in as widespread use.
Invariant expressions seem more usable than path expressions in the examples we
have seen.
.pp
The choice of mechanisms for concurrency control depends on efficiency and
on intended use.
Inside the operating system kernel, it often suffices to use the simplest lock
based on raising priority for short access to shared data; longer accesses that
might involve synchronization considerations (like placing data in a buffer)
would have to use a more expensive solution.
If processes, which are outside the kernel, need to share data, the kernel
might give them a semaphore facility through service calls.
This facility might then be used to construct monitors or event counts.
.pp
The programming language chosen by the operating system writer has an obvious
influence on the solutions chosen for the kernel.
For example, an operating system written in Modula or Concurrent Pascal is
likely to use the monitor-like concepts in those languages.
Similarly, the facilities offered by the operating system have an influence on
concurrency control between processes that share data (either virtual store
or files).
.SH 1 "Further reading"
.lp
Most of our understanding of concurrency control is due to the early work
of a few giants:  Dijkstra, Hoare, and Brinch Hansen.
T. J. Dekker introduced the non-alternating switch in the early 1960s.
This concept was generalized by Dijkstra
.[
dijkstra problem concurrent
.]
to make the spin switch.
Peterson
.[
peterson myths mutual exclusion
.]
simplified the non-alternating switch.
Monitors were already in the works when conditional
critical regions were designed.
Hoare
.[
hoare parallel programming
.]
first mentioned conditional critical regions,
and Brinch Hansen
.[
edison
.]
later used them as the principal mechanism for concurrency
control in his language, Edison.
Dijkstra
.[
dijkstra hierarchical ordering
.]
introduced the notion of a
.q "secretary,"
.IN Secretary
the forerunner of the modern monitor, at the same time conditional critical
regions were introduced by Hoare.
Dijkstra's paper is worth reading for his insight into the mutual-exclusion
and synchronization problems and his foresight of problems such as the nested
monitor call.
The term
.q monitor
was introduced by Hoare.
.[
hoare concept
.]
Nested monitor calls have been discussed at length in the literature;
one place to find more
.IN Monitors "Nested calls"
discussion on this topic is in a paper by Haddon.
.[
haddon nested monitor
.]
Howard
.[
howard signaling
.]
has shown that monitors with signal only at the end are strictly less powerful
than ordinary monitors.
.IN Monitors "signal only at end"
The \fWbroadcast\fP version of \fWsignal\fP has been used successfully in Mesa.
.[
lampson redell
.]
Crowds are based on serializers,
.IN Serializers
which were introduced by Atkinson and Hewitt
.[
atkinson hewitt serializers
.]
and generalized by Horn and Honda.
.[
horn honda crowds
.]
Event counts and sequencers were introduced by Reed and Kanodia,
.[
reed kanodia
.]
path expressions by Campbell and Habermann,
.[
campbell habermann path expressions
.]
and invariant expressions by Robert and Verjus.
.[
robert verjus synchronization modules
.]
.pp
The readers-writers problem has been discussed by many authors;
Courtois and Parnas presented an early discussion.
.[
courtois parnas readers writers
.]
One good place to read more is in an article by Lamport.
.[
lamport concurrent reading writing
.]
The dining philosophers problem seems to have been introduced by Dijkstra.
.[
dijkstra hierarchical ordering
.]
.pp
A discussion of criteria for evaluating concurrency control mechanisms can be
found in a paper by Bloom.
.[
bloom evaluating synchronization
.]
Scott's survey of languages for distributed computing
.[
scott framework
.]
considers monitors and path expressions as a special case of the more general
topic of tools for distributed computation.
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
Consider the following program fragments for processes \fBA\fP and \fBB\fP.
.(b
.TS
center;
n l10b lb 
n lfL lfL .
	process A	process B
.sp .1i
1	\fWfor\fP SeqA := 1 \fWto\fP 10 \fWdo\fP	\fWfor\fP SeqB := 1 \fWto\fP 10 \fWdo\fP
2	      x := x + 1;	      x := x + 1;
.TE
.)b
Assume that the shared variable x starts at 0, that both \fBA\fP and \fBB\fP
execute once,
and that SeqA and SeqB are not shared variables.
The two processes may execute at any speed.
What are the possible resulting values of x?
Assume that x must be loaded into a register for incrementing.
.EP \n+(E+.
Consider Figure \n(ch.\n(F1 again.
.br
(a)
We have seen that the actions to be performed could be partitioned into three
activities, {A,B,C}, {D,E,F}, and {G,H,I}, to achieve
some constraints through implicit synchronization.
How many ways are there to partition the actions?
.br
(b)
Your answer to (a) most likely did not include any activities like {A,C} because
even though C must wait for A, which is implicit in {A,C}, it must also wait
for B, which would have to be explicit.
If we allow such activities, how many partitionings can you build?
.br
(c)
Your answer to (b) most likely did not include any activities like {C,D}.
Why not?
.EP \n+(E+.
Disabling interrupts does not work correctly for nested regions.
Rewrite the BeginRegion and EndRegion routines to fix this problem.
.EP \n+(E+.
Generalize the switch mechanism to three processes.
.EP \n+(E+.
Does Dekker's non-alternating switch use busy waiting if \fBB\fP is in its
region, Turn\ =\ A, and \fBA\fP is trying to enter its region?
.EP \n+(E+.
Prove that Dekker's non-alternating switch makes it
impossible for an activity to starve while it is waiting to enter its region,
on the assumption that whenever the conflicting activity enters its region, it
eventually leaves that region.
.EP \n+(E+.
Generalize Dekker's non-alternating switch to three conflicting activities.
(\fIsomewhat difficult\fP)
.EP \n+(E+.
Generalize Dekker's non-alternating switch to @n@ conflicting activities.
(\fIdifficult\fP)
.EP \n+(E+.
Prove that Peterson's non-alternating switch cannot starve either activity.
.EP \n+(E+.
Prove that Peterson's non-alternating switch guarantees mutual exclusion.
.EP \n+(E+.
Generalize Peterson's non-alternating switch to @n@ conflicting activities.
(\fIhard\fP)
.EP \n+(E+.
Instead of test-and-set, some computers provide an atomic instruction that
sets the new value to 1 greater than its old value, as shown here:
.(b
.FS
 1	\fWatomic\fP \fWfunction\fP TestAndInc(\fWvar\fP Lock : integer) : integer;
 2	\fWbegin\fP
 3		TestAndInc := Lock;
 4		Lock := Lock + 1;
 5	\fWend\fP TestAndInc;
.FE
.)b
Show how to use this instruction to implement locks.
(\fIHint\fP:  You must be wary of integer overflow.)
.EP \n+(E+.
The Finishing and WaitFor routines in the text work fine if the entire
synchronization graph is to be executed just once.
Show how to modify them so that if we wish to run the entire graph again, the
Done variables have the correct values.
.EP \n+(E+.
What does a cycle in a synchronization graph mean?
.EP \n+(E+.
Procedures Finishing and WaitFor associate with every activity a Done variable
that is set when the activity finishes.
An alternative is to associate a Start
variable with every activity and have the activity check it before it starts.
Show how the synchronization graph of Figure \n(ch.\n(F1 could be implemented
in this way.
.EP \n+(E+.
Show how to use semaphores to implement the synchronization graph so that each
semaphore is associated with an activity and the activity waits (perhaps a
number of times) for its own semaphore before starting.
.EP \n+(E+.
We claim that a compiler can check for conflicting orders of nesting in
\fWregion\fP statements.
Exactly how can the compiler accomplish this task?
.EP \n+(E+.
Show why critical regions cannot implement the synchronization graph of Figure
\n(ch.\n(F1 without busy waiting.
.EP \n+(E+.
Write a solution to the readers-writers problem using conditional critical
regions, making sure that readers and writers never starve.
All you need to provide is StartRead, EndRead, StartWrite, and EndWrite.
(\fIHint\fP:
Separate readers into platoons.
The next one starts filling as soon as a writer arrives.
Alternate a platoon of readers with a single writer.)
.nr E9 \n(E+
.EP \n+(E+.
Write a monitor solution to the readers-writers problem.
(See the discussion for exercise \n(E9.)
.EP \n+(E+.
Find a case where it would be desirable for an activity that executes
\fWsignal\fP in a monitor to perform more work before leaving the monitor.
.EP \n+(E+.
Implement semaphores using monitors.
.EP \n+(E+.
We need mutual exclusion to implement Ticket.
Do we also need mutual exclusion to implement Advance?
.EP \n+(E+.
Show the program for the consumers in an event count and sequencer solution to
multiple producers, multiple consumers.
.EP \n+(E+.
Show how to use the DoubleDown procedure to implement the dining philosophers
problem.
Describe what can go wrong if we don't use MutualExclusion in implementing
DoubleDown.
Would the same problem occur if the declaration of MutualExclusion were at
line 6 instead of line 2?
.EP \n+(E+.
Does the following path-expression solution to the readers-writers problem work?
.IN "Readers-writers" "Path expression solution"
What policy does it enforce?
.(b
.FS
 1	\fWpath\fP EnterRead
 2	\fWpath\fP TryRead | { TryWrite }
 3	\fWpath\fP { StartRead; Read } | Write
 4	\fWlet\fP EnterRead = TryRead
 5	\fWlet\fP TryRead = StartRead
 6	\fWlet\fP TryWrite = Write
 7	\fWlet\fP DoRead = EnterRead ; Read
 8	\fWlet\fP DoWrite = TryWrite
.FE
.)b
.EP \n+(E+.
Does the following path-expression solution to the readers-writers problem work?
.IN "Readers-writers" "Path expression solution"
What policy does it enforce?
.(b
.FS
 1	\fWpath\fP EnterWrite
 2	\fWpath\fP { TryRead } | TryWrite
 3	\fWpath\fP { Read } | ( StartWrite ; Write )
 4	\fWlet\fP TryWrite = StartWrite
 5	\fWlet\fP EnterWrite = TryWrite
 6	\fWlet\fP TryRead = Read
 7	\fWlet\fP DoRead = TryRead
 8	\fWlet\fP DoWrite = EnterWrite; Write
.FE
.)b
.EP \n+(E+.
Show how to implement path expressions with semaphores.
(\fIdifficult\fP)
.EP \n+(E+.
Write a path-expression solution to the multiple producer-consumer problem.
.EP \n+(E+.
How would you implement invariant expressions using conditional critical
regions?
.EP \n+(E+.
What advantage do invariant expressions have over conditional critical regions?
.EP \n+(E+.
Our invariant-expression solution to the producers-consumers problem serializes
concurrent producers.
Rewrite the solution so that producers are serialized only long enough to
reserve slots but may fill those slots in parallel.
Make sure that the GetBuffer procedure never takes data from an empty slot.
.EP \n+(E+.
The Sequent Balance 21000 multicomputer has a hardware test-and-set instruction.
.IN "Sequent Balance 21000" computer
How can this instruction be used for limited busy waiting, as suggested by the
Hysteresis Principle, but not cause long busy waits?
.IN "Hysteresis Principle"
.EE
