.CH "Concurrency"
.IN Concurrency (start)
.lp
As operating systems were first built, an annoying sort of problem was
discovered:
Routines invoked from interrupts sometimes modify data that are in the midst of
.IN Interrupts
being inspected by the interrupted program.
The modifications can be benign, like appending to the end of a list, or
disastrous, like modifying a linked list in such a way that the interrupted
program follows a dangling pointer.
.pp
Such problems crop up in various contexts within operating systems.
In Chapter 5, we saw that a circular buffer pool can be simultaneously filled
by a process and emptied by the kernel.
It would be erroneous for the process to put new data in the buffer that is
currently being emptied.
In Chapter 6, we encountered read/write and write/write conflicts in file use.
We invented transactions to provide synchronization atomicity to prevent such
conflicts.
.IN Transactions
If we allow several processes to share parts of their virtual store, similar
conflicts arise.
As we suggested in Figure 3.\n(F8,
the kernel of the operating system may be composed of separate
tasks, each of which fulfills one particular function.
These tasks may communicate through shared data.
.pp
All these cases involve several activities sharing a common data area.
What happens can depend on the order in which the activities access the data.
Some orders give fine results;
some give incorrect but not fatal results;
others are disastrous.
For example, in a video game, it is fine if we move the hero before the
monster.
It is incorrect to decide where each will go before moving the other,
because they could end up in the same location on the screen.
It is disastrous if such overlap leads to dangling pointers that cause the
game to fail.
When two cooks are working on the same soup, it is fine if one chops onions
while the other measures barley.
It may even be acceptable for them to add these ingredients at the same time.
But if both independently decide to add salt, the result can be inedible.
In this chapter we will present methods for ensuring that different activities
sharing access to common data give predictable results.
This general topic is known as \fBconcurrency control\fP.
.IN "Concurrency control"
We use the term \fBactivity\fP to avoid specifying whether our entities are
.IN Activity
processes, kernel tasks, interrupt-driven procedures, or DMA transfers.
.pp
From one point of view, shared data can be considered a resource that the
operating system must manage to ensure correct use.
The problem is therefore in the domain of operating systems.
From another point of view, most of the solutions we will look at have a
linguistic flavor; that is, we will build programming-language structures
to help us enforce our policies and hide the complexity of the algorithms.
To some extent, therefore, our problem is in the domain of programming
languages.
Concurrency control is not the only problem in which operating systems and
programming languages have mutual interests.
We will touch upon communication later as another example.
.SH 1 "The problems"
.lp
Concurrency control is a problem of resource management.
When we looked at allocation of sequentially reusable resources, we
characterized policies as liberal or conservative.
The same characterization applies in the realm of concurrency control.
A policy can be very liberal, allowing shared data to be accessed almost
.IN "Resource allocation" liberal
whenever an activity pleases.
Such a strategy achieves the goal of simultaneous access and a high degree of
parallelism.
The danger of liberal strategies is that some execution orders lead to
erroneous results.
At the other extreme, a policy can be very conservative, blocking accesses
.IN "Resource allocation" Conservative
until it is certain that erroneous results cannot arise.
The danger of conservative strategies is that activities are unable to respond
efficiently to real-time events.
.pp
Deadlock and starvation are dangers for any policy that occasionally blocks
activities.
As we saw in Chapter 4, \fBdeadlock\fP occurs when activities are waiting for
each other so that none can run, and \fBstarvation\fP is a situation in which
some blocked activity is consistently passed over and not allowed to run.
.IN Deadlock
.IN Starvation
A related danger is \fBlivelock\fP, a situation in which the algorithm 
.IN Livelock
that decides whether to block an activity continues to run without reaching a
decision.
A special case of livelock is \fBbusy waiting\fP, in which an activity uses
computational resources without making progress as an alternative to blocking.
.IN "Busy waiting"
Busy waiting is not objectionable if there is no other work worth doing during
the wait.
.pp
Before we proceed with solutions, let's review the typical situation in which
simultaneous access can lead to errors.
.(b
.TS
center;
n l10b lb 
n lfL lfL .
	process A	process B
.sp .1i
1	TmpA := x;	TmpB := x;
2	TmpA := TmpA + 1;	TmpB := TmpB + 1;
3	x := TmpA;	x := TmpB;
.TE
.)b
.lp
Assume that TmpA and TmpB are variables local to processes \fBA\fP and
\fBB\fP and that x is a variable shared between them.
Assume that x starts with value 0.
After both \fBA\fP and \fBB\fP have finished their fragments of code, x has
the value 2 if we are lucky and 1 if we are not:
If both processes execute statement 1 before either executes statement 3, then
each will read an old x value of 0 and update it in line 3 to a 1.
.pp
Here is a simpler situation that can also lead to trouble.
.(b
.TS
center;
n l10b lb 
n lfL lfL .
	process A	process B
.sp .1i
1	x := 3;	x := 5;
.TE
.)b
.lp
The resulting value of x can be either 3 or 5, depending on which process
finishes second.
If the constants were not small, like 3 and 5, but were so large that several
words were needed to store them, these assignments might take several machine
instructions.
The instructions could overlap in such a way that the answer is neither what
\fBA\fP intended nor what \fBB\fP intended.
.pp
We usually assume that certain operations, such as assignment into main store,
are atomic and that \fBarbitration\fP takes place at some level within the
computer hardware:
.IN Arbitration
If two requests arrive simultaneously, they are selected in some sequential
order.
All our methods will depend on hardware arbitration at some level.
We will assume that two requests made simultaneously to main store are
serviced consecutively.
Such requests can arise when several computers are connected as a
multiprocessor.
In uniprocessors, each machine instruction is usually atomic.
Traps and interrupts, which can cause a process switch, are delayed until the
current machine instruction is finished.
Even here, some long instructions (like block copy) may be interruptible.
In addition, missing-page traps can occur in the middle of an ordinary
instruction.
.SH 2 "Mutual exclusion" I
.lp
The problem of \fBmutual exclusion\fP is to ensure that two concurrent
activities do not access shared data at the same time.
As we have seen, arbitration in the hardware is a form of mutual exclusion.
We will see how it can be extended to provide exclusion for more complicated
operations.
.pp
Instructions that must be executed while another activity is excluded
are grouped into \fBregions\fP.
.IN Regions
A region often contains more than one access to the shared data.
For example, the assignment
.(b
.FS
	x := x + 1
.FE
.)b
must refer to x twice.
On some machines, two or more instructions are needed to accomplish this
assignment.
They are all part of the same region.
Two regions \fBconflict\fP if execution of one requires that the other not be
executed at the same time.
By definition, all regions conflict with themselves.
However, the region just shown would not conflict with a region that increments
y instead of x.
.pp
A typical way to express mutual exclusion is to surround regions by delimiters
like \fLBeginRegion\fP and \fLEndRegion\fP, as shown here.
.(b
.TS
center;
n l10b lb 
n lfL lfL .
	process A	process B
.sp .1i
1	BeginRegion;	BeginRegion;
2	x := 3;	x := 5;
3	EndRegion;	EndRegion;
.TE
.)b
The form of the delimiters, exactly what a region means, and how it is
implemented distinguish the methods we will survey.
In all cases we can look upon regions as a way of imposing a degree of
determinism on a non-deterministic situation, since we can't predict when
interrupts will happen, when activities will be scheduled, or the relative
speeds of truly simultaneous activities in a multiprocessor.
.SH 2 "Synchronization" I
.lp
Mutual exclusion prevents an activity from entering a region
until other activities are out of conflicting regions.
If we want to prevent an activity from continuing until some more general
condition is met, we say
that we wish to \fBsynchronize\fP the activity with the condition.
Synchronization is a generalization of mutual exclusion.
.pp
Occurrence of conditions marks the passage of time.
Synchronizing with those conditions achieves a common notion of time among
activities.
This idea of common time gives rise to the word
.q synchronization. " "
.pp
Since synchronization is a generalization of mutual exclusion, we can expect
that the techniques developed for solving the
mutual-exclusion problem are also useful for achieving synchronization.
We will test this expectation by presenting various synchronization problems.
The first one is best described by a \fBsynchronization graph\fP,
.IN "Synchronization graph"
which shows precedence constraints that govern when actions may begin.
Figure \n(ch.\n+(FN shows the graph we will discuss.
.(z
.GS
width 2
file figs/fig.8.synch.graph
.GE
.nr F1 \n(FN \" needed later
.FI "A synchronization graph"
.)z
Action \fBA\fP has no preconditions.
Once it has finished, actions \fBB\fP and \fBD\fP may start.
As soon as \fBB\fP has finished, \fBC\fP may start, but \fBE\fP must wait for
both \fBB\fP and \fBD\fP to finish.
\fBI\fP may not start until everything else has finished.
Each arrow in this graph is a synchronization step.
It represents the condition that the action at its tail has finished.
.pp
There are many ways to cast this graph into a program.
We might code each of the nine actions separately and provide for explicit
synchronization.
Alternatively, we could cast each row as a separate activity.
The first activity would execute \fBA\fP, then \fBB\fP, then \fBC\fP.
Synchronization between these pieces would be implicit.
The second activity, which executes \fBD\fP, \fBE\fP, and \fBF\fP, would
synchronize with conditions created by the first activity in three places.
The last activity is similar.
Another alternative is to represent each column as an activity.
.pp
Choosing an alternative affects which synchronizations are implicit and which
are explicit.
It also affects how we perceive the flow of control.
For example, this synchronization graph could represent the steps needed to
complete three Fortran compilations in a simple operating system.
Each compilation is represented in one row.
The row represents, for example, reading the source program from disk,
compiling, and writing the result to tape.
If there is only one disk drive and one tape drive, and we don't allow two
compilations to proceed simultaneously, the precedence constraints in the graph
properly limit the amount of concurrency.
.pp
If we treat each row as a separate activity, then interactions between reading
a program, compiling it, and outputting the result are implicit; interactions
between this program and another are explicit.
If we treat each column as a separate activity, then we have a card-reader
activity, a compilation activity, and an output activity.
Each signals the next when it has performed all it needs to with respect to one
Fortran compilation.
.pp
Later we will introduce the producers-consumers problem and the
readers-writers problem as more complex synchronization tests.
We will also return to the dining philosophers problem, which we introduced in
Chapter 4.
.so concur.1.nr
.so concur.2.nr
.so concur.3.nr
.IN Concurrency (end)
