#!/usr/local/bin/perl -T

# Report for mounts relation. Creates the following files:
#	 nfs files:
#						 /etc/exports (BSD) and /etc/dfs/dfstab (Solaris)
#	 amd files:
#						 /etc/amd.mounts.conf
#						 symlinks to /mounts

# imports packages:
#	Site.pl			site-specific customization
#	Definitions.pl	includes some constant variables
#	Relation.pl		includes some subroutines that are used most often.
#	PutRecords.pl	includes the PutRecords subroutine.
#	GetExpand.pl	a recursive-descent parser for HostClass expressions.
#
$SATHOME = "/usr/local/etc/sat";
require "$SATHOME/perlinclude/Site.pl";
require "$SATHOME/perlinclude/Definitions.pl";
require "$SATHOME/perlinclude/PutRecords.pl";
require "$SATHOME/perlinclude/Relation.pl";
require "$SATHOME/perlinclude/GetExpand.pl";

$Relation = "Mounts";

require "$SATHOME/perlinclude/StartUp.pl";

$BUFSIZE = 2000; # greatest length that an exports list may have

# where the scheme and data files for the Hosts relation are located.
$hostsschemefile = "$SatDBRoot/Hosts/scheme";
$hostsdatafile = "$SatDBRoot/Hosts/data";

# set relevant values
if ($Debug) {
	$Owner = $DebugOwner;
} else {
	$Owner = "root"; # for all files we produce here
		# /etc/exports must be owned by root on Linux
}

$AmdFile = "$Root/amd.mounts.conf";

# whatever we are currently working on.
$TempFile = "";

# host names from Hosts relation.
%FullNames = ();

sub MakeFull {
	#	 The parameter $s is a space-separated list of host names.
	#	 This subroutine replaces each entry by the full host name.
	local ($s) = @_;
	my ($entries_tuple, @Entries, $Answer);
	@Entries = split / /, $s;
	for $entries_tuple (@Entries) {
		if ($entries_tuple eq "") {
			next;
		}
		if ($FullNames{$entries_tuple} eq "") {
			print STDERR "Can't convert $entries_tuple to full name.\n";
			next;
		}
		if ($Answer eq "") {
			$Answer = $FullNames{$entries_tuple};
		} else {
			$Answer = "$Answer $FullNames{$entries_tuple}";
		}
	} # for each entry in s
	return $Answer;
} #MakeFull

# exit if the relation hasn't been modified since the last reporting,
if (! $Debug && SameAgeAsRelation ($lastmodified, $AmdFile)) {
	exit 0;
}

# we want to clean up the execution before the program terminates.
END { Clean (\$TempFile, $AmdFile, $lastmodified)};

# set the random number seed for the rand operator.
srand;

# read the records from the Host relation into @Hosts.
@Hosts = PutRecords ($hostsschemefile, $hostsdatafile);
$#Hosts > 10 || die "Looks like Hosts relation too short";
	 # terrible error; maybe we couldn't read data.

# read the records from the Mounts relation into @Mounts.
@Mounts = PutRecords ($schemefile, $datafile);
$#Mounts > 10 || die "Looks like Mounts relation too short";
	 # terrible error; maybe we couldn't read data.

# See if I am in a special group
$AmSolaris = IncludesMe ("SOLARIS", $MyHostName, $MyFullHostName);
$AmLinux = IncludesMe ("LINUX", $MyHostName, $MyFullHostName);
$AmIRIX = IncludesMe ("IRIX", $MyHostName, $MyFullHostName);
# $AmNEXT = IncludesMe ("NEXT", $MyHostName, $MyFullHostName);
$AmNEXT = 0; # we have no more NEXT machines; IncludesMe("NEXT") will fail.
$AmHPUX = IncludesMe ("HPUX", $MyHostName, $MyFullHostName);

# initialize %FullNames, $TrustList
$TrustList = "";
for $host_tuple (@Hosts) {
	$FullNames{$host_tuple->{'HostName'}} = $host_tuple->{'FullHostName'};
	if (IncludesMe ($host_tuple->{'Secure'}, $MyHostName, $MyFullHostName)) {
		# I trust this host; add it to the list of trusted hosts.
		$TrustList .= " $host_tuple->{'HostName'}";
	}
}
$TrustList =~ s/^ //; # remove initial space, if any

$ExportFile = "$Root/exports"; # usual case
if ($AmSolaris) {
	$ExportFile = "$Root/dfs/dfstab";
	$TrustList = MakeFull($TrustList);
	$TrustList =~ s/ /:/g;
} elsif ($AmLinux) {
	$TrustList =~ s/( |$)/(rw,no_root_squash)$1/g;
} elsif ($AmIRIX) {
	$TrustList =~ s/ /:/g;
} elsif ($AmHPUX) {
	$TrustList =~ s/ /:/g;
} else {
	$TrustList =~ s/ /:/g;
}
$ExportFile = Untaint($ExportFile);
$TrustList =~ Untaint($TrustList);

{ # ExportFile 
	#	 Each brand of Unix has its own formats and limitations.
	#	 Solaris:
	#		 the export file is called /etc/dfs/dfstab
	#		 it may use netgroups (but only from NIS or NIS+, not files)
	#		 it may not use IP numbers or short host names
	#		 it does not accept wildcards like *.cs.engr.uky.edu
	#		 to let trusted hosts access as root: root=host
	#		 the syntax is like
	#			 /usr/sbin/share -F nfs -o root=host:host,rw=host:host /u/file
	#		 The export file must be executed after it is built
	#	 Linux:
	#		 the export file is /etc/exports; no buffer overflow problem
	#		 it may not use netgroups
	#		 it may not use IP numbers
	#		 it accepts wildcards like *.cs.engr.uky.edu (9/96: seems to fail)
	#		 it accepts short names.
	#		 to let trusted hosts access as root: (no_root_squash)
	#		 the syntax is like
	#			 /u/ulf-d1 host(rw,no_root_squash) host(rw) 
	#		 rpc.nfsd and rpc.mountd must be HUPed after changes
	#	 IRIX:
	#		the export file is /etc/exports
	#		it may use netgroups, but not gotten via NIS+; only NIS
	#		it accepts short names.
	#		to let trusted hosts access as root: -root=host:host
	#		the syntax is like
	#			/u/bal-d1 -root=host:host -rw=norton:bedever:accolon foonly
	#		if you don't list any hosts without extra flags, export is ro to
	#			the world! (hence foonly) [I am not sure this is true.]
	#		/usr/etc/exportfs -a must be run after changes	
	#	 HPUX:
	#		the export file is /etc/exports
	#		I guess it may use netgroups via NIS+.  ?? Must check.
	#		it accepts short names.
	#		to let trusted hosts access as root: -root=host:host. ?? Must check.
	#		the syntax is like
	#			/u/bal-d1 -root=host:host -access=norton:bedever:accolon foonly
	#		if you don't list any hosts without extra flags, export is ro to
	#			the world! (hence foonly) ?? Must check
	#		/usr/etc/exportfs -a must be run after changes	 ?? Must check
	#	 OSF/1:
	#		 the export file is /etc/exports
	#		 it may use netgroups, only from NIS, if you can get them working!
	#		 it may use IP numbers
	#		 it does not accept wildcards like *.cs.engr.uky.edu
	#		 it allows continuation lines (start with whitespace)
	#		 to let trusted hosts access as root: -root=host:host
	#		 the syntax is like 
	#			 /u/ulf-d1 -root=host:host -access=a.domain:b.domain foonly
	#		 or, optionally,
	#			 /u/ulf-d1 a.domain b.domain
	#		if you don't list any hosts without extra flags, export is ro to
	#			the world! (hence foonly)
	#
	OpenFile ($ExportFile, \$TempFile, OUTFILE);
	print OUTFILE "# This $ExportFile file generated by SAT for $MyHostName"
		. " from Mounts.\n";
	print OUTFILE "# It indicates which remote hosts may mount my file"
		. " systems.\n";
	$FullToList = MakeFull (Expand ("$MySite-NA-PRINTER-DOS-WIN95"));
	$ExportCount = 0;
	for $mount_tuple (@Mounts) {
		if (($mount_tuple->{'Type'} !~ /^(nfs|dnfs|dswap)$/)
		  || ! ($mount_tuple->{'ListInExports'})
		  || ($mount_tuple->{'ToHost'} eq "")) {
			# not appropriate to list this in exports
			next;
		}
		$Found = 0;
		@FromHosts = split / /, $mount_tuple->{'FromHost'};
		@FromDirectories = split / /, $mount_tuple->{'FromDirectory'};
		for ($HostNum = 0; $HostNum <= $#FromHosts; $HostNum ++) {
			if (IncludesMe($FromHosts[$HostNum], $MyHostName, $MyFullHostName)){
				if (! $AmSolaris) {	
					print OUTFILE "$FromDirectories[$HostNum]";
				}
				$Found = $TRUE;
				last;
			}
		}
		if (! $Found) { # I am not involved in exporting this file system
			next;
		}
		# start of each line info
		if ($AmSolaris) {
			print OUTFILE "/usr/sbin/share -F nfs ";
			if (! ($mount_tuple->{'ExportOptions'} =~ /(rw|ro)/)) {
				die "$mount_tuple->{'Handle'} needs "
					. "ExportOptions = rw or ro.\n";
			}
			print OUTFILE "-o root=$TrustList," .
				"$mount_tuple->{'ExportOptions'}";
		} elsif ($AmLinux) {
			print OUTFILE " $TrustList ";
		} elsif ($AmIRIX || $AmHPUX) {
			if ($mount_tuple->{'ExportOptions'} ne "") {
				$mount_tuple->{'ExportOptions'} =~ s/^-/,/;
				print OUTFILE " -root=$TrustList" .
					"$mount_tuple->{'ExportOptions'}";
			}
		} else {
			print "I can't handle this OS type.  Sorry.\n";
			exit 1;
		} # start of line information
		if ($AmSolaris) { # groks netgroups
			$Separator = "=";
			$NewSeparator = ":";
			$ToSay = $MySite;
			$Extra = "";
		} elsif ($AmLinux) { # doesn't grok netgroups, but any size ok.
			$Separator = " ";
			$NewSeparator = " ";
			$ExportOptions = $mount_tuple->{'ExportOptions'};
			$ToSay = MakeFull (Expand ($mount_tuple->{'ToHost'}));
			$ToSay =~ s/\.[\w.]*/($ExportOptions)/g;
			$Extra = "";
		} elsif ($AmIRIX) { # groks netgroups, but not NIS+
			$Separator = "=";
			$NewSeparator = ":";
			$ToSay = "$FullToList";
			$ToSay =~ s/\.[\w.]*\s*/$NewSeparator/g;
			# $ToSay =~ s/$NewSeparator$/ $MyHostName/; # prevent world-read
			$ToSay =~ s/$NewSeparator$//;
		} elsif ($AmHPUX) { # groks netgroups, NIS+
			$Separator = " ";
			$NewSeparator = " ";
			$ToSay = $MySite;
		} else { # just build a vanilla list based on the ToHosts.
			print OUTFILE " ";
			$Separator = "";
			$ToSay = MakeFull (Expand ($mount_tuple->{'ToHost'}));
			$ToSay =~ tr/ /$NewSeparator/;
			$Extra = " $MyHostName"; # explicit to prevent world-read
		}
		if (length $ToSay < $BUFSIZE) {
			print OUTFILE "$Separator$ToSay$Extra";
		} else {
			die "$mount_tuple->{'Handle'} generates more than "
				. "$BUFSIZE chars in $ExportFile\n$ToSay";
		}
		# end of on each line info
		if ($AmSolaris) {
			print OUTFILE " $FromDirectories[$HostNum - 1]\n";
		} else {
			print OUTFILE "\n";
		}
		$ExportCount += 1;
	} # foreach tuple in Mounts
	CloseFile ($ExportFile, 0644, \$TempFile, OUTFILE, $Owner);
	# final actions when ExportFile has changed
	if (! $Debug && $AmSolaris) {
		chmod 0755, $ExportFile;
		system $ExportFile;
	} elsif (! $Debug && ($AmIRIX || $AmHPUX)) {
		$command = Untaint("/usr/etc/exportfs -a");
		system $command;
	} elsif (! $Debug && $AmNEXT) {
		$command = Untaint("rm -f /etc/xtab; /usr/etc/exportfs -a");
		system $command;
	} elsif (! $Debug && $AmLinux) { # raphael 12/7/1998: re-enabled
		system "/bin/killall -HUP rpc.mountd rpc.nfsd 2> /dev/null > /dev/null";
	}
} # $ExportFile

{ # $AmdFile 
	OpenFile ($AmdFile, \$TempFile, OUTFILE);
	$MountOpts = "intr,retrans=4,noquota,soft"; # had timeo=20: raphael
	print OUTFILE "# This $AmdFile file built by SAT for $MyHostName from"
		. " Mounts.\n";
	print OUTFILE "# It indicates what remote file systems to automount when"
		." needed.\n";
	print OUTFILE "/defaults type:=nfs;";
	if ($AmLinux) { # poor NFS implementation.
		$MountOpts .= ",rsize=8192,wsize=8192,timeo=3";
	}
	print OUTFILE "opts:=$MountOpts\n";
	for $mount_tuple (@Mounts) {
		$ToHosts = Expand ($mount_tuple->{'ToHost'});
		if ($mount_tuple->{'ToDirectory'} ne ""
		  && $mount_tuple->{'ToDirectory'} ne "\/"
		  && $ToHosts =~ /\b$MyHostName\b/
		  && $mount_tuple->{'FromHost'} !~ /\b$MyHostName\b/) { # import it
			@FromHosts = split / /, $mount_tuple->{'FromHost'};
			@FromDirectories = split / /, $mount_tuple->{'FromDirectory'};
			$MountPoint = $mount_tuple->{'ToDirectory'};
			$MountPoint =~ s/(^\/)//g;
			$MountPoint =~ s/\//\-/g;
			print OUTFILE "$MountPoint";
			
			# generate the entries in random order to avoid overloading any
			# remote host.
			$primary = (rand ( 2**15 - 1)) % ($#FromHosts + 1);
			print OUTFILE " rhost:=$FromHosts[$primary];";
			print OUTFILE "rfs:=$FromDirectories[$primary]";
			if ($mount_tuple->{'AmdOptions'} ne "") {
				print OUTFILE ";opts:=$MountOpts" . "," . 
					$mount_tuple->{'AmdOptions'};
			}
			for ($HostNum = 0; $HostNum <= $#FromHosts; $HostNum ++) {
				if ($HostNum != $primary) {
					print OUTFILE " rhost:=$FromHosts[$HostNum];";
					print OUTFILE "rfs:=$FromDirectories[$HostNum]";
					if ($mount_tuple->{'AmdOptions'} ne "") {
						print OUTFILE ";opts:=$mount_tuple->{'AmdOptions'}";
					}
				}
			} # for HostNum
			print OUTFILE "\n";
			# build symlinks
			if (! $Debug && ! FileLExists ($mount_tuple->{'ToDirectory'})) {
				# print "$mount_tuple->{'ToDirectory'} is missing\n";
				$MountPoint = Untaint($MountPoint);
				$Command = Untaint("ln -s /mounts/$MountPoint"
					. " $mount_tuple->{'ToDirectory'}");
				system  $Command;
			}
		} # import it
	} # for tuple in Mount
	CloseFile ($AmdFile, 0644, \$TempFile, OUTFILE, $Owner);
	if (! $Debug && -f "$Root/amq" ) { # flush the amd cache
		system "$Root/amq -f 2> /dev/null > /dev/null";
	}
} # $AmdFile

# exit stat.
exit 0;
