#!/usr/local/bin/perl -T

# Report for users relation. Creates the following files:
#   /etc/passwd               local user database
#              set shell /usr/local/bin/nologin for users not allowed on
#              this machine.
#   /etc/shadow               only for solaris (so far); others (linux, osf1)
#                             don't get.
#   /etc/group                local group database
#   /etc/amd.users.conf       automount daemon database
#   /usr/local/etc/sat/lib/useraliasfile
#                             SAT privilege alias file (if on this machine)
#   /etc/ftpusers             list of disallowed ftp accesses
#   /etc/shells               list of accepted shells

# imports packages:
#	Site.pl			site-specific customization
#	Definitions.pl	includes some constant variables
#	Relation.pl		includes some subroutines that are used most often.
#	PutRecords.pl	includes the PutRecords subroutine.
#   GetExpand.pl    a recursive-descent parser for HostClass expressions.
#
$SATHOME = "/usr/local/etc/sat";
require "$SATHOME/perlinclude/Site.pl";
require "$SATHOME/perlinclude/Definitions.pl";
require "$SATHOME/perlinclude/PutRecords.pl";
require "$SATHOME/perlinclude/Relation.pl";
require "$SATHOME/perlinclude/GetExpand.pl";

$Relation = "Users"; # which relation this program reports.

require "$SATHOME/perlinclude/StartUp.pl";

# files we create
$UserFile = "$Root/passwd"; # we set/check date only on this one.
$AliasFile = ($Debug ? $MyRoot : $SATHOME) . "/lib/useraliasfile";
$GroupFile = "$Root/group";
$AmdFile = "$Root/amd.users.conf";
$FTPFile = "$Root/ftpusers";
$ShellFile = "$Root/shells";
$Shells = "$SatDBRoot/$Relation/shells";
# in hpux, we also modify /tcb/files/auth/*

$TempFile = ""; # whatever we are currently working on.

if (! $Debug && SameAgeAsRelation ($lastmodified, $UserFile)) {
	exit 0; # not worth continuing; nothing new.
}

# we want to clean up the execution before the program terminates.
END { Clean (\$TempFile, $UserFile, $lastmodified)};

$HostClass = RunProgram ("$HOSTINFO HostClass") || die "$HOSTINFO: $!";
	# eg "STABLE SOLARIS"

$MyOS = RunProgram ("$HOSTINFO OSType"); # eg linux, or solaris, or aix

# for some OS types, we build a shadow file.
if ($MyOS eq "solaris") {
	$HaveShadow = $TRUE;
	$ShadowFile = "$Root/shadow";
} elsif ($MyOS eq "aix") {
	$HaveShadow = $TRUE;
	$ShadowFile = "$Root/security/passwd";
} elsif ($MyOS eq "linux" && -f "$Root/shadow") { # only for some Linux boxes.
	$HaveShadow = $TRUE;
	$ShadowFile = "$Root/shadow";
} elsif ($MyOS eq "hpux") {
	$HaveShadow = $TRUE; # not really, but close enough
} elsif ($MyOS eq "irix") {
	$HaveShadow = $FALSE;
		# actually, there is a shadow, but you should run pwconv to build it.
} else {
	$HaveShadow = $FALSE;
}

# read the records from the Users relation into @Users.
$> = getpwnam "operator" || die "Cannot become operator";
@Users = PutRecords ($schemefile, $datafile);
$> = 0;  # we are root again

{ # $UserFile
	OpenFile ($UserFile, \$TempFile, OUTFILE);
	%Uids = (); # stores the uids for each user. 
	%Groups = (); # stores the group information; each entry has all user names
		# in the group.
	# collect group and uid information.
	$NumUsers = 0;
	for $user_tuple (@Users) {
		$NumUsers++;
		$Uids{$user_tuple->{'Login'}} = $user_tuple->{'Uid'};
		@MyGroups = split / /, $user_tuple->{'Groups'};
		for $group (@MyGroups) {
			if (defined $Groups{$group}) {
				$Groups{$group} = "$Groups{$group},$user_tuple->{'Login'}";
			} else {
				$Groups{$group} = "$user_tuple->{'Login'}";
			}
		} # foreach group
		# check whether this user has an account here or not.
		$Class = $user_tuple->{'Class'};
		$Class =~ s/\s+$//; # remove trailing white space
		if (IncludesMe($Class, $MyHostName, $MyFullHostName)) {
			$MyShell = $user_tuple->{'Shell'};
		} else {
			$MyShell = "/usr/local/bin/nologin"; # default shell.
		}
		$Password = $HaveShadow ? ($MyOS eq "aix" ? "!" : ($MyOS eq "hpux" ?
			"*" : "x")) : $user_tuple->{'Password'};
		print OUTFILE "$user_tuple->{'Login'}:$Password:$user_tuple->{'Uid'}:"
			. "$Uids{$MyGroups[0]}:$user_tuple->{'Name'}:$user_tuple->{'Home'}:"
			. "$MyShell\n";
	}
	$NumUsers > 10 || die "Looks like User relation too short";
		 # terrible error; maybe we couldn't read data.
	CloseFile ($UserFile, 0444, \$TempFile, OUTFILE, "");
} # $UserFile

{ # $ShadowFile
	if ($HaveShadow && $MyOS ne "hpux") {
		OpenFile ($ShadowFile, \$TempFile, OUTFILE);
		if ($MyOS eq "aix") {
			for $user_tuple (@Users) {
			  print OUTFILE "$user_tuple->{'Login'}:\n"
				. "\tpassword = $user_tuple->{'Password'}\n\n";
			}
		} else { # not aix
			$ExtraStuff = ($MyOS eq "linux") ? ":9855:0:99999:7:::" : "::::::";
			for $user_tuple (@Users) {
				print OUTFILE "$user_tuple->{'Login'}:$user_tuple->{'Password'}"
					. "$ExtraStuff\n";
			}
		} # not aix		
		CloseFile ($ShadowFile, 0400, \$TempFile, OUTFILE, "");
	} elsif ($MyOS eq "irix") {
		if (!$Debug) {
			system "/sbin/pwconv" || die "/sbin/pwconv: $!"
		}
	} elsif ($MyOS eq "hpux") { # no shadow file; one file per user.
		$line = -1;
		$/ = ""; # read entire file at once.
		$filePrefix = ($Debug ? "$Root" : "") . "/tcb/files/auth";
		for $user_tuple (@Users) {
			$line++;
			$Recorded{$login} = 0;
			$login = $user_tuple->{'Login'};
			$Recorded{$login} = 0; # remember we have seen this user name
			$id = $user_tuple->{'Uid'};
			$passwd = $user_tuple->{'Password'};
			$login =~ /^(.)/; # first char
			$file = $filePrefix . "/$1/$login";
			$extra = $id == 0 ?  ":u_bootauth" : "";
			$file = Untaint($file);
			# print "Working on $login\n";
			if (-e $file) { # already exists.  Maybe ok.
				open INFILE, $file;
				$contents = <INFILE>;
				close INFILE;
				if ($contents =~ /pwd=$passwd:/) { # it is ok
					# print "No need to change\n";
					next;
				} else { # need to modify
					if ($Debug) {print "Just modifying password for $login\n";}
					$contents =~ s/pwd=[^:]*:/pwd=$passwd:/;
					OpenFile ($file, \$TempFile, OUTFILE);
					print OUTFILE $contents;
					CloseFile($file, 0664, \$TempFile, OUTFILE, "");
				}
			} else { # need to build
				if ($Debug) {print "Adding record $file for $login\n";}
				OpenFile ($file, \$TempFile, OUTFILE);
				print OUTFILE "$login:u_name=$login:u_id#$id:\\\n" .
					"\t:u_pwd=$passwd:\\\n" .
					"\t$extra:u_auditid#$line:\\\n" .
					"\t:u_auditflag#1:\\\n" .
					"\t:u_pswduser=$login:u_suclog#921005596:u_lock\@:\\\n" .
					"\t:chkent:\n";
				CloseFile($file, 0664, \$TempFile, OUTFILE, "");
			} # need to build
		} # for users
		# remove extraneous auth records (from previously valid users)
		open FILES, "/usr/bin/echo $filePrefix/?/*|"; # glob is unsafe; fake it
		$/ = " "; # separate at spaces
		while (defined($file = <FILES>)) { # foreach auth file
			chop $file;
			# print "Working on file: [$file]\n";
			$file =~ /$filePrefix\/.\/(.*)/;
			if (!defined($Recorded{$1}) && $1 ne "root") {
				print "Removing $file\n";
				unlink Untaint($file);
			}
		} # foreach auth file
		close FILES;
#		If you remove root's auth file, root can no longer run anything!
#			don't do this!!! system "/bin/rm /tcb/files/auth/[a-z]/[a-z]*";
#			system "/usr/sbin/pwconv" || die "/usr/sbin/pwconv: $!";
	} # hpux
} # Shadow file

{ # /etc/group
	OpenFile ($GroupFile, \$TempFile, OUTFILE);
	for $group (keys %Groups) {
		if ($HostClass =~ /LINUX/ && $HostClass =~ /ALPHA/) { # no long lines
			$LengthLimit = 400; # actual limit about 500, I think
			while ($Groups{$group} =~ /^(.{$LengthLimit,}?),(.*)$/) {
				$Groups{$group} = $2;
				print OUTFILE "$group\::$Uids{$group}:$1\n";
			}
			print OUTFILE "$group\::$Uids{$group}:$Groups{$group}\n";
		} else { # long lines allowed
			print OUTFILE "$group\::$Uids{$group}:$Groups{$group}\n";
		}
	}
	CloseFile ($GroupFile, 0744, \$TempFile, OUTFILE, "");
	system  ("sort -o $GroupFile -t: +2 -n $GroupFile");
	chmod 0444, $GroupFile;
} # /etc/group

{ # $AmdFile
	OpenFile ($AmdFile, \$TempFile, OUTFILE);
	print OUTFILE "# This file was automatically produced by SAT from the Users"
		. " relation.\n";
	print OUTFILE "/defaults type:=link\n";
	for $user_tuple (@Users) {
		print OUTFILE "$user_tuple->{'Login'} fs:=$user_tuple->{'Home'}\n";
	}
	CloseFile ($AmdFile, 0444, \$TempFile, OUTFILE, "");
} # $AmdFile

{ # $AliasFile
	$AliasLocation = RunProgram ("/usr/local/bin/whathost $SATHOME");
	$hostsschemefile = "$SatDBRoot/Hosts/scheme";
	$hostsdatafile = "$SatDBRoot/Hosts/data";
	if ($MyHostName eq $AliasLocation) { # I have a copy of $SATHOME
		@Hosts = PutRecords ($hostsschemefile, $hostsdatafile);
		OpenFile ($AliasFile, \$TempFile, OUTFILE);
		print OUTFILE "This file automatically produced by SAT from"
		  . " the Users and Hosts relations.\n\n";
		print OUTFILE "\nROOT\@MAIN\n";
		for $host_tuple (@Hosts) {
			if ($host_tuple->{'SatClass'} =~ /^model$/) {
				print OUTFILE "root\@$host_tuple->{'FullHostName'}\n";
			}
		}
		print OUTFILE "\nOPERATOR\@MAIN\noperator\@$SiteOfAuth\n\n";
		print OUTFILE "\nALLOPERATOR\n";
		for $host_tuple (@Hosts) {
			if ($host_tuple->{'HostClass'} =~ /$MyClass/) {
				print OUTFILE "root\@$host_tuple->{'FullHostName'}\n";
				print OUTFILE "operator\@$host_tuple->{'FullHostName'}\n";
			}
		}
		print OUTFILE "\nALLROOT\n";
		for $host_tuple (@Hosts) {
			if ($host_tuple->{'HostClass'} =~ /$MyClass/) {
				print OUTFILE "root\@$host_tuple->{'FullHostName'}\n";
			}
		}
		print OUTFILE "\nSTAFF\@MAIN\n";
		for $user_tuple (@Users) {
			if ($user_tuple->{'Groups'} =~ /\bstaff\b/) {
				print OUTFILE "$user_tuple->{'Login'}\@$SiteOfAuth\n";
			}
		}
		CloseFile ($AliasFile, 0444, \$TempFile, OUTFILE, "");
	} # I have a copy of $SATHOME
} # $AliasFile

{ # $FTPFile
	OpenFile ($FTPFile, \$TempFile, OUTFILE);
	print OUTFILE "# This file created by SAT from Users.\n";
	for $user_tuple (@Users) {
		if ($user_tuple->{'Shell'} eq "/bin/true"
		  && $user_tuple->{'Uid'} < 100) {
			print OUTFILE "$user_tuple->{'Login'}\n";
		}
	}
	print OUTFILE "student\n"; # special case
	print OUTFILE "root\n"; # security reasons - pjl
	CloseFile ($FTPFile, 0444, \$TempFile, OUTFILE, "");
} # $FTPFile

{ # $ShellFile
	# permissable shells master file is $Shells; copy to where expected
	if ($MyOS eq "aix") {
		# shells are just one line in login.cfg; copy the rest.
		$ShellFile = "$Root/security/login.cfg";
		OpenFile ($ShellFile, \$TempFile, OUTFILE);
		open (OLDSHELLFILE, "</etc/security/login.cfg")
			|| die "Can't open /etc/security/login.cfg: $!";
		while (defined($Line = <OLDSHELLFILE>)) {
			if ($Line =~ /^\tshells = /) { # this is the line to change
				open (IN, "<$Shells") || die "Can't open $Shells for read: $!";
				$separator = ""; # no comma before the first one
				print OUTFILE "\tshells = ";
				while (defined($Line = <IN>)) {
					if ($Line =~ /^\#/) {
						next; # skip comments
					}
					@ShellArray = split /[ \t]+/, $Line;
					# the first element is the name of the shell.
					if ($#ShellArray != $NULL) {
						print OUTFILE "$separator$ShellArray[0]";	
						$separator = ",";
					}
				}
				print OUTFILE "\n";
			} else { # ordinary line of OLDSHELLFILE
				print OUTFILE $Line
			}
		} # while Line
		close OLDSHELLFILE;
		CloseFile ($ShellFile, 0660, \$TempFile, OUTFILE, "");
	} else { # not aix
		OpenFile ($ShellFile, \$TempFile, OUTFILE);
			# OUTFILE is now a handle for a temp file for $ShellFile
		@ShellArray = ();
		open (IN, "<$Shells") || die "Can't open $Shells for read: $!";
		while (defined($Line = <IN>)) {
			if ($Line =~ /^\#/) {
				next; # skip comments.
			}
			@ShellArray = split /[ \t]+/, $Line;
			# the first element is the name of the shell.
			if ($#ShellArray != $NULL) {
				print OUTFILE "$ShellArray[0]\n";	
			}
		}
		CloseFile ($ShellFile, 0444, \$TempFile, OUTFILE, "");
	} # not aix	
} # $ShellFile

# exit status.
exit 0;
