.SH 1 "Perspective"
.lp
The file manager must provide an acceptable level of service for a wide range of
applications.
Some file managers accomplish this goal by providing a very simple but
efficient service and expect that complex applications can build on it.
For example, one can make a case for the idea that a flat file space of
sequential, direct-access files with minimal access restrictions is acceptable
for many applications and is an adequate base for extension to other
applications.
.pp
Other file managers accomplish the goal by providing many kinds of service.
For example, a process might have a choice of half a dozen file types and
allocation strategies at the time it creates a file.
The file manager supports all these options as efficiently as it can.
Such a file manager might actually implement a simple base structure and build
alternative structures on top of that base.
.pp
Database management systems are the most difficult to serve well.
.IN Databases
They structure data and access it in regular ways, but few file managers are
equipped to take advantage of those regularities.
For example, we examined storing huge files in depth-3 trees and were
pleased that a direct access required no more than four disk accesses.
However, if the huge file is storing a database, the database is likely to
contain internal pointers.
These pointers might be expressed in terms of bytes from the start of the file.
Following such a pointer requires about four read operations.
If the pointers were expressed in physical block number, each would need only
one read operation.
The structure used by the file manager effectively decreases
efficiency by a factor of 4.
For this reason, database programs like to avoid the file abstraction
and prefer to deal directly with physical addresses.
.pp
The internal structure of the file manager can be built in the following layers.
.sp 10p
.ip \(bu
\fBFile layer\fP:
This module creates a data structure for each open file.
The data structure includes the file descriptor and the read/write mark.
.IN "Read/write mark"
Procedures in this module open and close files on behalf of processes that
submit the Open and Close service calls.
.IN "Service call" Open
.IN "Service call" Close
The directory in which a file resides must be opened for a short time in order
to find out where the file descriptor is.
If the file is named by a path, each directory in the path must be opened for a
short time.
.ip \(bu
\fBByte layer\fP:
This module provides procedures for reading or writing any number of bytes from
or to any open file.
These procedures consult the open-file structure to determine which disk block
is involved in each transput.
They also modify the read/write mark.
These procedures are called by the Read, Write, and Position service calls.
.IN "Service call" Read
.IN "Service call" Write
.IN "Service call" Position
They are also called by procedures in the file layer that need to read
directories.
.ip \(bu
\fBBlock layer\fP:
This module provides procedures to read and write entire blocks.
A cache of recently used blocks may be maintained so that successive
operations to the same block do not require disk actions.
This layer is the upper layer of the disk driver.
.sp 10p
.lp
Each layer smooths out chunkiness created by the lower levels and produces
new structures, as suggested by the Beautification Principle.
.IN "Beautification Principle"
In the case of files, we can start with the atomic level, where
the disk surface is composed of discrete components (iron atoms, for example).
At the aggregate level, these discontinuities are evened out, and we consider
each disk surface to be essentially uniform and continuous.
The disk controller introduces structures by collecting together several
platters and subdividing each into tracks, sectors, and bytes.
The block layer flattens this hierarchy, leaving only an array of disk
blocks, each composed of a fixed number of bytes.
The byte layer hides disk blocks and the file-descriptor structures in order to
introduce byte streams.
The file layer organizes these streams of bytes into chunks called files.
Finally, directories convert a linear array of files (indexed by their
descriptor numbers) into a tree structure.
.pp
We can demonstrate these layers of the file manager by tracing their actions
when a process executes the following program:
.(b
.FS
 1	\fWprocedure\fP Example;
 2	\fWvar\fP
 3		MyFile : FileNumber;
 4	\fWbegin\fP
 5		MyFile := Open("/a/b",ToWrite);
 6		Position(MyFile,700); \(mi\(mi skip the first part of the file
 7		Write(MyFile,"testing",7); \(mi\(mi seven characters
 8		Close(MyFile);
 9	\fWend\fP Example;
.FE
.)b
.np
The Open call (line 5) causes a context switch to the kernel, which calls the
file layer of the file manager to open this file.
.np
The file layer needs to open \fL/\fP (the root directory) to find file
\fL/a\fP,
which is itself a directory.
This action involves asking the block layer to bring in the file descriptor for
\fL/\fP.
By convention, we use descriptor 0 for this file.
The file layer can calculate which disk block contains descriptor 0 by referring
to the disk descriptor, which should be kept in main store at all times.
Let us pretend that disk block 10 is needed.
.np
The block layer checks the cache of file blocks to see if block 10 is
currently in main store.
If so, this layer returns immediately, passing back a pointer to where in main
store it has cached block 10.
In our situation, a cache hit is very likely, because
block 10 is used extremely frequently.
If there is a cache miss, the block layer selects some cache entry to be
overwritten (based, for example, on an LRU strategy).
If the entry it has chosen is dirty, it must first write out that entry.
While the necessary transput is in progress, the block layer
places the current process on a device-wait list and queues the
appropriate read (and write) request to the lower part of the disk driver.
Other processes are allowed to run until the lower part of the disk driver
has finished writing, if necessary, and then reading in block 10.
Then the calling process, still executing in privileged state inside the block
layer, returns to the file layer.
It passes back a pointer to the main store where block 10 has been read into
the cache.
We will ignore how the block layer performs its job from now on.
.np
The file layer reads the descriptor for \fL/\fP from block 10.
It determines that the caller has the necessary access privileges by checking
the Permissions field of the descriptor.
From here on, we will ignore all privilege checks.
The file layer now creates a temporary open-file descriptor for \fL/\fP.
Here is a possible format for an open-file descriptor.
.(b L
.FS
10	\fWtype\fP
11		OpenFileDescriptor =
12			\fWrecord\fP
13				Descriptor : ^FileDescriptor; { points into disk-block cache }
14				Mark : integer; { the next byte to be read or written }
15				Mode : (ToRead, ToWrite, ToReadWrite);
16			\fWend\fP
17	\fWend\fP;
.FE
.)b
In our case, Descriptor (line 13) points into the disk-block cache where the
file descriptor has been read, Mark (line 14) starts at 0, and Mode (line 15)
is set ToRead.
.sp .1i
.ti +.5i
Let us assume that a file entry in a directory is always 20 bytes long.
The file layer now calls the byte layer to read the first 20 bytes from \fL/\fP
and store them in a 20-byte data structure (let us call it an entry descriptor)
allocated for this purpose:
.(b
.FS
18	\fWtype\fP
19		EntryDescriptor =
20			\fWrecord\fP
21				Name : \fWarray\fP [0..15] \fWof\fP char;
22				FileDescriptorNumber : integer; { 4 bytes }
23			\fWend\fP;
.FE
.)b
.np
The byte layer, referring to the open-file descriptor for \fL/\fP,
asks the block layer to fetch the first block (say 512 bytes) of \fL/\fP.
It then copies the first 20 bytes to the entry descriptor, updates the file's
read mark, and returns to the file layer.
If it had been asked for bytes that span several blocks, it would have asked
the block layer for each required block in turn and copied the relevant
information.
If the file is large enough so the file descriptor (line 13) is
insufficient to point to all the blocks, the byte layer can ask the block layer
to read in overflow file descriptors.
We will ignore how the byte layer works from now on.
.np
Upon reading the entry descriptor, the file layer may discover that it does not
describe \fL/a\fP but rather some other file, such as \fL%\fP or \fL^\fP .
In that case, the file layer repeatedly asks the byte layer for another 20
bytes from \fL/\fP until it finds an entry describing \fL/a\fP.
If it fails, the Open call was invalid, and an error return is taken to the
process.
.np
Having found a directory entry for \fL/a\fP, the file layer releases the
temporary open-file descriptor for \fL/\fP and opens a new one for \fL/a\fP in
a similar fashion.
.np
The file layer follows a similar course until it has created an open-file
descriptor for \fL/a/b\fP, set to allow writing (line 15).
The write mark (line 14) is initialized to 0.
All temporary open-file descriptors are released by now.
The index of this open-file descriptor is passed back to the calling process to
use as a file number.
.np
The process next submits a Position service call (line 6).
The kernel forwards this call to the byte layer, which advances the write mark
(line 14) to position 700.
.np
The process submits a Write service call (line 7).
The byte layer asks the block layer to read in the second block of file
\fL/a/b\fP, since the write mark (line 14) indicates that this is the
appropriate block.
This block now appears in the cache of file blocks.
The byte layer then copies 7 characters from the virtual space of the calling
process into the cache copy and marks the copy as
.q dirty. " " 
This block will eventually be cleaned by being copied back to the disk.
.np
The process submits a Close service call (line 8).
The file layer releases the open-file descriptor for the file.
.lp
We have ignored a few details here, such as how we ensure that the block layer
does not discard one of its cached blocks while it is still in use by a higher
layer, how room is made for overflow file descriptors when the byte layer
needs to bring them in, what happens if the file was only 400 bytes long at
the outset, and what is different if the file is not stored on disk at all,
but on a magnetic tape.
