#!/usr/local/bin/perl -w

# module for dynamic k-d trees

# run pod2text to read the documentation.

# note lines with string "bug" and "improve:"

=head1 NAME

KD: kd-tree storage of k-dimensional
numeric data and then nearest-neighbor queries on those data.

We allow the data within a single tree to disagree on k.  When we search,
we ignore extra dimensions in the target and descend to both children of
nodes that discriminate on a dimension missing in the target.  When we insert,
we place new nodes in the right child of nodes that discriminate on a dimension
missing in the new node.

=head1 DATA STRUCTURES

=head2 In-core trees

An in-core tree is composed of internal nodes and buckets.  A bucket contains
n-dimensional points in this data structure:

	'Type' => 'bucket'
	'Values' => ref to array of refs to n-dimensional arrays

It is not required that all points in buckets have the same dimension.

An internal node has this structure:

	'Type' => 'internal'
	'Direction' => scalar (0-based)
	'Discriminant' => scalar
	'leftChild' => $leftChild,
	'rightChild' => $rightChild

=head2 Trees in files

K-d trees stored in files are in ASCII, to make them easier to debug and
manipulate with tools written in a variety of languages.  The file starts with
a header:

	"kd" startPos

where startPos is the file offset of the root of the tree (which may not be at
the start of the file, due to reorganizations).

Buckets are stored as follows:

	"bucket" "parent" parentPos endPos (point "/")*

where parentPos is the file offset of the parent of this bucket, endPos is the
offset of the first position in the file after this bucket, and point is a
space-separated list of k numeric values.  For instance,

	bucket parent 456 200 149 348 / 312 819 /

represents a bucket with parent node at offset 456.  The bucket has two
points, (149,348) and (312,819); the last character used by the bucket is at
file offset 200.

Internal nodes are stored as follows:

	"internal" "parent" parentPos "direction" direction "discriminant"
		discriminant "left" leftPos "right" rightPos

where parentPos is the file offset of the parent of this internal node, leftPos
and rightPos are the file soffsets of its two children.  This internal node
discriminates in the dimension given by direction at the value given by
discriminant.

=head1 AUTHOR

Written by Raphael Finkel <raphael@cs.uky.edu>, 8/2000.

=cut

use IO::Seekable; # to define SEEK_SET
use strict 'vars'; # enforce declarations
use strict;

BEGIN {
	use Exporter ();
	use vars qw($VERSION @ISA @EXPORT %EXPORT_TAGS @EXPORT_OK);
	$VERSION = 1.00; # 8/2000 Raphael Finkel raphael@cs.uky.edu
	@ISA = qw(Exporter);
	@EXPORT = qw(&insertCore &insertFile &deleteCore &nearest &TestKD
		&readFile &writeFile &printPoint &printTree);
	@EXPORT_OK   = qw(&randomPoint &probe &TimeSoFar);
	%EXPORT_TAGS = ();
}


## declarations

	# constants
	use vars qw($bucketSize $infinity $tabs $useMedian $tokenLength
		$gapLength $numberLength);
	$bucketSize = 4; # 50 is a good number; constant
	$infinity = 1e10; # to initialize minimization computations
	$tabs = 0; # for printing purposes; varies
	$useMedian = 0; # experiments show sorting is faster than my median routine
	$tokenLength = 30; # long enough for any word or token in a kd file
	$gapLength = 100; # length of gap at end of a new file-format bucket
	$numberLength = 10; # digits needed to represent a file offset

	# global variables
	use vars qw(@Trace $bestDimension $dummy $numDims);
	@Trace = (); # current call stack, for debugging

=head1 APPLICATION PROGRAMMING INTERFACE

=cut

## insertion subs

=head2 insertCore($tree, $newValue)

$tree is a reference to an in-core k-d tree (possibly empty, in which
case the reference is undefined).
$newValue is a reference to an array of numeric values.
insertCore() returns a reference to an in-core k-d tree containing
$newValue as an additional point.

=cut

sub insertCore {
	# $tree is a ref to a k-d tree; $newValue is a ref to an array of floats.
	push @Trace, "in insertCore";
	my ($tree, $newValue) = @_;
	if (!defined($tree)) { # new tree
		# print "a new tree\n";
		$tree = {'Type' => 'bucket', 'Values' => [$newValue]};
	} elsif ($tree->{'Type'} eq 'bucket') {
		# print "Adding to bucket\n";
		if (scalar(@{$tree->{'Values'}}) < $bucketSize) {
			# just add another value
			# print "bucket not too big: $#{$tree->{'Values'}}\n";
			push @{$tree->{'Values'}}, $newValue;
		} else {
			# print "must split bucket\n";
			my ($Direction, $Discriminant, $left, $right) =
				splitBucket(@{$tree->{'Values'}});
			my $newTree = {'Type' => 'internal',
				'Direction' => $Direction, 'Discriminant' => $Discriminant,
				'leftChild' => {'Type' => 'bucket', 'Values' => $left},
				'rightChild' => {'Type' => 'bucket', 'Values' => $right},
			};
			$tree = insertCore($newTree, $newValue);
		}
	} else {
		# print "Adding to internal node\n";
		my $direction = $tree->{'Direction'};
		if (defined($$newValue[$tree->{'Direction'}]) and
		  $$newValue[$tree->{'Direction'}] < $tree->{'Discriminant'}) {
			$tree->{'leftChild'} = insertCore($tree->{'leftChild'}, $newValue);
		} else {
			$tree->{'rightChild'} = insertCore($tree->{'rightChild'}, $newValue);
		}
	} # internal node
	pop @Trace;
	return($tree);
} # insertCore;

sub fixupParent {
	my ($tree, $parentPos, $oldStartPos, $newStartPos) = @_;
	# The internal node at $parentPos points to a node at $oldStartPos.
	# Fix it to point instead to the node at $newStartPos.
	# File position when we return is arbitrary.
	# print "fixupParent: $parentPos, $oldStartPos, $newStartPos\n";
	my $childPos; # where in file we find the parent't pointer to $oldStartPos
	seek $tree, $parentPos, SEEK_SET;
	if ($parentPos == 0) { # old parent was the root
		expect("kd", $tree);
		if ($oldStartPos == expect("^\\d+\$", $tree)) {
			seek $tree, $parentPos, SEEK_SET;
			print $tree sprintf("kd %${numberLength}d", $newStartPos);
		} else {
			die "Can't update parent at $parentPos to point to child " ,
				"that used to be at $oldStartPos";
		}
	} else { # old parent was internal
		expect("internal", $tree);
		expect("parent", $tree);
		expect("^\\d+\$", $tree);
		expect("direction", $tree);
		expect("^\\d+\$", $tree);
		expect("discriminant", $tree);
		expect("^\\d+\$", $tree);
		$childPos = tell($tree);
		expect("left", $tree);
		if ($oldStartPos == expect("^\\d+\$", $tree)) {
			seek $tree, $childPos, SEEK_SET;
			print $tree sprintf("left %${numberLength}d", $newStartPos);
		} else {
			$childPos = tell($tree);
			expect("right", $tree);
			$childPos = tell($tree);
			if ($oldStartPos == expect("^\\d+\$", $tree)) {
				seek $tree, $childPos, SEEK_SET;
				print $tree sprintf("right %${numberLength}d", $newStartPos);
			} else {
				die "Can't update parent at $parentPos to point to child " ,
					"that used to be at $oldStartPos";
			}
		}
	} # old parent was internal
} # fixupParent

=head2 insertFile($tree, $newValue)

$tree is
(1) an open file pointing at the start of a node, or
(2) a file name, in which
case insertion is to start at the top.
$newValue is a reference to an array of numeric values.
insertFile() modifies the k-d tree to include the given point.
There is no return value; the file read/write marker is left at an arbitrary
place.
In case (2), the file is closed upon return.

=cut

sub insertFile {
	my ($tree, $newValue) = @_;
	if (!ref($tree)) { # a file name
		open INSERTFILE, "+>$tree";
		insertFile(\*INSERTFILE, $newValue);
		close INSERTFILE;
		return;
	}
	push @Trace, "insertFile at mark " . tell($tree) . ", point " .
		join(" ", @$newValue);
	my $startPos = tell($tree); # start of my node, before header word
	my $token = ref($tree) eq "HASH" ? $tree->{'Type'} : readToken($tree);
	my $fixupPoint; # offset in file we need to return to and insert value
	my $filePosEnd; # offset in file of end of the node we insert
	if ($token eq 'null') { # new tree
		# print "a new tree\n";
		seek $tree, $startPos, SEEK_SET; # will overwrite "null" with a bucket
		print $tree "bucket parent 0 ";
		$fixupPoint = tell $tree;
		print $tree " " x ($numberLength+1); # will point to after bucket
		print $tree &printPoint($newValue) . " / ";
		print $tree " " x $gapLength;
		$filePosEnd = tell $tree;
		seek $tree, $fixupPoint, SEEK_SET;
		print $tree sprintf("%${numberLength}d ", $filePosEnd);
	} elsif ($token eq 'bucket') {
		# print "Adding to bucket\n";
		my ($parentPos, $endPos, @values) = readBucket($tree);
		# print "values so far: ", join("/", @values), "\n";
		if (scalar(@values) < $bucketSize) {
			# just add another value
			# print "bucket not too big: $#values\n";
			push @values, $newValue;
			my $toPrint;
			$toPrint = "bucket parent $parentPos ";
			$toPrint .= sprintf("%${numberLength}d ", $endPos);
			my $point;
			foreach $point (@values) {
				$toPrint .= printPoint($point) . " / ";
			} # foreach point
			if ($endPos - $startPos >= length($toPrint)) { # it fits
				seek $tree, $startPos, SEEK_SET;
				print $tree $toPrint;
			} else { # abandon this region; go to end of file.
				# print "abandoning at $startPos\n";
				seek $tree, 0, SEEK_END;
				$endPos = tell($tree);
				# print "moving to at $endPos\n";
				$toPrint .= " " x $gapLength;
				my $newEndPos = sprintf("%${numberLength}d",
					$endPos + length($toPrint));
				$toPrint =~ s/(bucket parent \d+) \d+/$1 $newEndPos/;
				print $tree $toPrint;
				fixupParent($tree, $parentPos, $startPos, $endPos);
			}
		} else {
			# print "must split bucket starting at $startPos\n";
			my ($Direction, $Discriminant, $leftValues, $rightValues) =
				splitBucket(@values);
			# improve: we just abandon the old bucket
			my $newStartPos;
			seek $tree, 0, SEEK_END;
			$newStartPos = tell();
			# improve: merge the following code with part of writeFileRecursive
			print $tree "internal parent $parentPos " ,
				"direction $Direction " .
				"discriminant $Discriminant ";
			$fixupPoint = tell $tree;
			print $tree
				sprintf("left %${numberLength}d right %${numberLength}d ", 0,
				0);
			my $filePosLeft = tell $tree;
			writeFileBucket($tree, $newStartPos, $gapLength, @$leftValues);
			my $filePosRight = tell $tree;
			writeFileBucket($tree, $newStartPos, $gapLength, @$rightValues);
			seek $tree, $fixupPoint, SEEK_SET;
			print $tree
				sprintf("left %${numberLength}d right %${numberLength}d ",
					$filePosLeft, $filePosRight);
			seek $tree, $newStartPos, SEEK_SET;
			insertFile($tree, $newValue);
			fixupParent($tree, $parentPos, $startPos, $newStartPos);
		}
	} elsif ($token eq 'internal') {
		# print "Adding to internal\n";
		my ($Direction, $Discriminant, $leftPos, $rightPos) =
			readInternal($tree);
		if (defined($$newValue[$Direction]) and
		  $$newValue[$Direction] < $Discriminant) {
			seek $tree, $leftPos, SEEK_SET;
			insertFile($tree, $newValue);
		} else {
			seek $tree, $rightPos, SEEK_SET;
			insertFile($tree, $newValue);
		}
	} elsif ($token eq 'kd') { # start of file
		my $rootPos = expect("^\\d+\$", $tree);
		seek $tree, $rootPos, SEEK_SET;
		insertFile($tree, $newValue);
	} else {
		die "bad representation at mark " . tell($tree);
	}
	pop @Trace;
} # insertFile;

sub splitBucket {
	my (@values) = @_;
	push @Trace, "in splitBucket";
	# @values contains all the values in a bucket.
	# Split by median of the dimension of the largest span.
	# Return dimension, value in that dimension, [left values], [right values]
	my (@lowest, @highest, $dimension);
	my $point;
	foreach $point (@values) { # determine dimension of largest span
		$dimension = 0;
		my $coordinate;
		foreach $coordinate (@$point) {
			if (defined($lowest[$dimension])) {
				# print "adjusting boundaries for $dimension\n";
				$lowest[$dimension] = $coordinate
					if $coordinate < $lowest[$dimension];
				$highest[$dimension] = $coordinate
					if $coordinate > $highest[$dimension];
			} else {
				# print "setting boundaries for $dimension\n";
				$lowest[$dimension] = $coordinate;
				$highest[$dimension] = $coordinate;
			}
			$dimension += 1;
		} # foreach coordinate
	} # foreach point
	$bestDimension = -1; # global; imported into median() and sort().
	my ($biggestSpan) = -1;
	for ($dimension = 0; $dimension < scalar(@highest); $dimension += 1) {
		my $span = $highest[$dimension] - $lowest[$dimension];
		if ($span > $biggestSpan) {
			$bestDimension = $dimension;
			$biggestSpan = $span;
		}
	} # foreach $dimension
	# print "Best dimension is $bestDimension with span $biggestSpan\n";
	my ($leftValues, $rightValues, $medianValue);
	if ($useMedian) {
		# use median; # we don't use, so commented out.
		sub comparer {
			return ($_[0][$bestDimension] <=> $_[1][$bestDimension]);
		} # comparer
		($medianValue, $leftValues, $rightValues) = median(\&comparer, @values);
		push @{$leftValues}, $medianValue;
	} else {
		my @sortedValues =
			sort {$$a[$bestDimension] <=> $$b[$bestDimension]} @values;
		my $serial;
		for ($serial = 0; $serial < $bucketSize / 2; $serial += 1) {
			push @{$leftValues}, $sortedValues[$serial];
		}
		for (; $serial < $bucketSize; $serial += 1) {
			push @{$rightValues}, $sortedValues[$serial];
		}
		$medianValue = $sortedValues[$bucketSize / 2];
	}
	pop @Trace;
	return($bestDimension, ${$medianValue}[$bestDimension], $leftValues,
		$rightValues);
} # splitBucket

## searching subs

=head2 nearest($tree, $target, $identical)

$tree is (1) a reference to an in-core k-d tree,
(2) an open file descriptor with the read mark at the start of a node, or
(3) a file name, in which case search starts from the root of the tree.
$target is a reference to an array of numeric values representing the query
point.
$identical is Boolean: true means return $target itself if it is in the tree;
false means find the nearest except for $target if it is the tree.

Nearest() returns a reference to a hash of this form:

	{'Distance' => scalar, 'Point' = ref to array of refs}

It uses Euclidean distance as its metric.
In case (2), the read mark is left at an arbitrary place.  In case (3), the
file is closed.

=cut

sub nearest {
	my ($tree, $target, $identical, $bestDistance, $bestPoint) = @_;
	# The best distance found so far and the point at that distance from the
	# target are the last two parameters, which are optional.
	# print "nearest: best so far is $bestDistance\n";
	$bestDistance = $infinity if !defined($bestDistance);
	my ($maxdim, $dimension, $distance, $spread, $answer);
	if (!ref($tree)) { # a file name
		open INSERTFILE, $tree;
		my $answer = nearest(\*INSERTFILE, $target);
		close INSERTFILE;
		return($answer);
	}
	my $token = ref($tree) eq "HASH" ?  $tree->{'Type'} : readToken($tree);
	if ($token eq 'bucket') {
		# print "Searching in bucket\n";
		push @Trace, "bucket";
		my @values;
		if (ref($tree) eq "HASH") {
			@values = @{$tree->{'Values'}};
		} else {
			($dummy, $dummy, @values) = readBucket($tree);
		}
		my $point;
		foreach $point (@values) {
			$distance = 0;
			$maxdim = scalar(@$target);
			$maxdim = scalar(@$point) if scalar(@$point) < $maxdim;
			for ($dimension = 0; $dimension < $maxdim; $dimension++) {
				$spread = $$point[$dimension] - $$target[$dimension];
				$distance += $spread * $spread;
			}
			# print "distance to ", printPoint($point), ":", $distance, "\n";
			if ($distance < $bestDistance and ($distance != 0 or $identical)) {
				$bestDistance = $distance;
				$bestPoint = $point;
			}
		} # foreach point
		# print "returning distance $bestDistance, point ",
		#	printPoint($bestPoint), "\n";
		$answer = {'Distance' => $bestDistance, 'Point' => $bestPoint};
	} elsif ($token eq 'internal') {
		# print "Searching in internal node\n";
		push @Trace, "internal";
		my ($Direction, $Discriminant, $leftChild, $rightChild) =
			ref($tree) eq "HASH" ? 
					($tree->{'Direction'}, 
					$tree->{'Discriminant'}, 
					$tree->{'leftChild'}, 
					$tree->{'rightChild'}) 
				: readInternal($tree);
		# print "Dir $Direction dis $Discriminant\n";
		my $span;
		if (!defined($$target[$Direction]) or
		  $$target[$Direction] < $Discriminant) {
			# print "left ";
			seek $tree, $leftChild, SEEK_SET if ref($tree) eq "GLOB";
			$answer = nearest(ref($tree) eq "HASH" ? $leftChild : $tree,
				$target, $identical, $bestDistance, $bestPoint);
			$span = $answer->{'Point'}[$Direction] - $Discriminant;
			$span *= $span;
			if ($answer->{'Distance'} > $span) { 
				seek $tree, $rightChild, SEEK_SET if ref($tree) eq "GLOB";
				$answer = nearest(ref($tree) eq "HASH" ? $rightChild : $tree,
					$target, $identical, $answer->{'Distance'},
					$answer->{'Point'});
			}
		}
		if (!defined($$target[$Direction]) or
		  $$target[$Direction] >= $Discriminant) {
			# print "right ";
			seek $tree, $rightChild, SEEK_SET if ref($tree) eq "GLOB";
			$answer = nearest(ref($tree) eq "HASH" ? $rightChild : $tree,
				$target, $identical, $bestDistance, $bestPoint);
			$span = $answer->{'Point'}[$Direction] - $Discriminant;
			$span *= $span;
			if ($answer->{'Distance'} > $span) { 
				seek $tree, $leftChild, SEEK_SET if ref($tree) eq "GLOB";
				$answer = nearest(ref($tree) eq "HASH" ? $leftChild : $tree,
					$target, $identical, $answer->{'Distance'},
					$answer->{'Point'});
			}
		}
	} elsif ($token eq 'kd') { # start of file
		my $parentPos = expect("^\\d+\$", $tree);
		seek $tree, $parentPos, SEEK_SET;
		$answer = nearest($tree, $target, $identical, $bestDistance,
			$bestPoint);
	} else {
		die "empty tree or bad representation.";
	}
	pop @Trace;
	return($answer);
} # nearest

## tree to string subs

=head2 printPoint($point)

$point is a reference to an array of numeric values.
printPoint() returns a string representing that array, separated by spaces.

=cut

sub printPoint { # returns a string
	my ($point) = @_;
	return(join(" ", @$point));
} # printPoint

=head2 printTree($tree)

$tree is (1) a reference to an in-core k-d tree,
(2) a file name, in which case printing starts from the root of the tree.
$tree is a reference to an in-core k-d tree.
printTree() returns a string representing that tree for printing.
In case (2), the file is closed.

=cut

sub printTree { # returns a string
	# $tree is a ref to a k-d tree
	my ($tree) = @_;
	if (!ref($tree)) { # a file name
		return printTree(readFile($tree));
	}
	my $answer = " " x $tabs;
	if (!defined($tree)) { # empty tree
		$answer = "undefined tree";
	} elsif ($tree->{'Type'} eq 'bucket') {
		my $point;
		foreach $point (@{$tree->{'Values'}}) {
			$answer .= printPoint($point) . " / ";
		} # foreach point
	} else {
		$answer .= "direction $tree->{'Direction'}, " .
			"discriminant $tree->{'Discriminant'},\n";
		$tabs += 2;
		$answer .= printTree($tree->{'leftChild'}) . "\n";
		$answer .= printTree($tree->{'rightChild'});
		$tabs -= 2;
	}
	return($answer);
} # print;

## File write subs

=head2 writeFile($fileName, $tree)

Writes the in-core file represented by the reference in $tree to the file whose
name is given by $fileName.
Returns with a warning and without writing if the file already exists.
Uses filehandle FILE, conflicting with any other use.

=cut

sub writeFile {
	my ($fileName, $tree) = @_;
	if (-e $fileName) {
		print STDERR "writeFile: $fileName already exists";
		return;
	}
	open FILE, "+>$fileName";
	print FILE sprintf("kd %${numberLength}d ", 0);
	my $endPos = tell(FILE);
	seek FILE, 0, SEEK_SET;
	print FILE sprintf("kd %${numberLength}d ", $endPos);
	writeFileRecursive(\*FILE, $tree, 0);
	close FILE;
} # writeFile

sub writeFileBucket {
	my ($FILE, $parentPos, $extraSpace, @values) = @_;
	# assumes file marker is ready for start of bucket.
	# leaves file marker at end.
	print $FILE "bucket parent $parentPos ";
	my $filePosStart = tell $FILE;
	print $FILE sprintf("%${numberLength}d ", 0);
	my $point;
	foreach $point (@values) {
		print $FILE printPoint($point) . " / ";
	} # foreach point
	print $FILE " " x $extraSpace;
	my $filePosEnd = tell $FILE;
	seek $FILE, $filePosStart, SEEK_SET;
	print $FILE sprintf("%${numberLength}d ", $filePosEnd);
	seek $FILE, $filePosEnd, SEEK_SET;
} # writeFileBucket

sub writeFileRecursive {
	my ($FILE, $tree, $parentPos) = @_;
	my ($string, $filePosStart, $filePosEnd, $filePosLeft, $filePosRight);
	if (!defined($tree)) { # null child
		print $FILE "null ";
	} elsif ($tree->{'Type'} eq 'bucket') {
		writeFileBucket($FILE, $parentPos, 0, @{$tree->{'Values'}});
	} else {
		my $myPos = tell $FILE;
		print $FILE "internal parent $parentPos " ,
			"direction $tree->{'Direction'} " .
			"discriminant $tree->{'Discriminant'} ";
		$filePosStart = tell $FILE;
		print $FILE
			sprintf("left %${numberLength}d right %${numberLength}d ", 0, 0);
		$filePosLeft = tell $FILE;
		writeFileRecursive($FILE, $tree->{'leftChild'}, $myPos);
		$filePosRight = tell $FILE;
		writeFileRecursive($FILE, $tree->{'rightChild'}, $myPos);
		$filePosEnd = tell $FILE;
		seek $FILE, $filePosStart, SEEK_SET;
		print $FILE sprintf("left %${numberLength}d right %${numberLength}d ",
			$filePosLeft, $filePosRight);
		seek $FILE, $filePosEnd, SEEK_SET;
	}
} # writeFileRecursive

## File read subs

sub expect { # reads a token and fails if it is not what is expected
	# returns the token read
	my ($pattern, $openFile) = @_;
	my $where = tell($openFile);
	my $token = readToken($openFile);
	die "At offset $where in file expected to read \"$pattern\"; " .
		"got \"$token\" instead.  Stopping"
		if ($token !~ /$pattern/);
	return($token);
} # expect

sub readToken { # returns a token and advances input to next character.
	my ($openFile) = @_;
	my $startPos = tell $openFile;
	my $stringIn;
	read $openFile, $stringIn, $tokenLength;
	$stringIn =~ /^(\S*)(\s*)/; # first token, white space
	my $token = $1;
	my $whiteStuff = $2;
	if (length($whiteStuff) + length($token) == $tokenLength) {
		# must read more to exhaust the white space
		do { read $openFile, $stringIn, $tokenLength;}
			until $stringIn !~ /^\s+$/;
		if ($stringIn eq "") { # end of file
			# print "end of file\n";
		} else { # move to start of coming token
			# print "got something now: [$stringIn]\n";
			$stringIn =~ /(^\s*)/;
			seek $openFile, length($1) - $tokenLength, SEEK_CUR;
		}
	} else { # move to start of coming token
		seek $openFile, $startPos + length($token) + length($whiteStuff) ,
			SEEK_SET;
	}
	# print "got string [$token]\n";
	return($token);
} # readToken

sub readBucket {
	my ($tree) = @_;
	# input is a file at the start of a bucket, past word 'bucket'
	# returns ($parent, $endPos, @values)
	# print "readBucket\n";
	expect("parent", $tree);
	my $parent = expect("^\\d+\$", $tree);
	my $endPos = readToken($tree);
	my (@point, @values);
	while (tell($tree) < $endPos) { # one point
		@point = ();
		while (1) { # one coordinate
			my $value = readToken($tree);
			last if ($value eq '/');
			push @point, $value;
		} # while one coordinate
		push @values, [@point];
	} # while one point
	return($parent, $endPos, @values);
} # readBucket

sub readInternal {
	my ($tree) = @_;
	expect("parent", $tree);
	expect("^\\d+\$", $tree); # where is the parent; ignore
	expect("direction", $tree);
	my $Direction = readToken($tree);
	expect("discriminant", $tree);
	my $Discriminant = readToken($tree);
	expect("left", $tree);
	my $leftPos = readToken($tree);
	expect("right", $tree);
	my $rightPos = readToken($tree);
	return(($Direction, $Discriminant, $leftPos, $rightPos));
}

=head2 readFile($tree)

$tree is
(1) an open file pointing at the start of a node, or
(2) a file name, in which case reading is to start at the top.
Reads the external file into an in-core structure, which it then returns.

=cut

sub readFile {
	my ($tree) = @_;
	my ($answer, $token);
	if (!ref($tree)) { # a file name
		open READFILE, "+<$tree" or die "readFile: $!";
		$answer = readFile(\*READFILE);
		close READFILE;
		return $answer;
	}
	$token = readToken($tree);
	# print "reading file at offset ", tell($tree), ": ";
	if ($token eq "null") {
		# print "null\n";
		$answer = undef;
	} elsif ($token eq "bucket") {
		# print "bucket\n";
		my ($dummy, @values);
		($dummy, $dummy, @values) = readBucket($tree);
		$answer = {'Type' => 'bucket', 'Values' => [@values]};
	} elsif ($token eq "internal") { # internal node
		# print "internal\n";
		my ($Direction, $Discriminant, $leftPos, $rightPos) =
			readInternal($tree);
		# print "going to left; seeking to $leftPos\n";
		seek $tree, $leftPos, SEEK_SET;
		my $leftChild = readFile($tree);
		# print "going to right; seeking to $rightPos\n";
		seek $tree, $rightPos, SEEK_SET;
		my $rightChild = readFile($tree);
		$answer = {'Type' => 'internal', 'Direction' => $Direction,
			'Discriminant' => $Discriminant,
			'leftChild' => $leftChild,
			'rightChild' => $rightChild };
	} elsif ($token eq 'kd') { # start of file
		my $rootPos = expect("^\\d+\$", $tree);
		seek $tree, $rootPos, SEEK_SET;
		$answer = readFile($tree);
	} else {
		die "Not the start of a node: $token";
	}
	return($answer);
} # readFile

## deletion subs

=head2 deleteCore($tree, $oldValue)

$tree is a reference to an in-core k-d tree (possibly empty, in which
case the reference is undefined).
$oldValue is a reference to an array of numeric values.
deleteCore() returns the same reference to the in-core k-d tree, but
$oldValue has been removed.  No error if $oldValue wasn't there to begin with.

=cut

sub deleteCore {
	my ($tree, $oldValue) = @_;
	return $tree if (!defined($tree)); # empty tree; can't delete
	if ($tree->{'Type'} eq 'bucket') {
		my @values = @{$tree->{'Values'}};
		my $stringified = join(' ', @$oldValue);
		# print "Deleting $stringified from bucket\n";
		my $index;
		for ($index = 0; $index <= $#values; $index++) {
			if ($stringified eq join(' ', @{$values[$index]})) { # found it
				# print "removing ", printPoint($oldValue), "\n";
				splice @values, $index, 1;
				$tree->{'Values'} = [@values];
				last;
			}
		}
	} else {
		# print "Deleting from internal node\n";
		my $direction = $tree->{'Direction'};
		if (defined($$oldValue[$tree->{'Direction'}]) and
		  $$oldValue[$tree->{'Direction'}] < $tree->{'Discriminant'}) {
			$tree->{'leftChild'} = deleteCore($tree->{'leftChild'}, $oldValue);
		} else {
			$tree->{'rightChild'} =
				deleteCore($tree->{'rightChild'}, $oldValue);
		}
	} # internal node
	return($tree);
} # deleteCore;

=head2 deleteFile($tree, $oldValue)

$tree is
(1) an open file pointing at the start of a node, or
(2) a file name, in which case deletion is to start at the top.
$oldValue is a reference to an array of numeric values.
deleteFile() removes $oldValue from the tree.
No error if $oldValue wasn't there to begin with.
There is no return value; the file read/write marker is left at an arbitrary
place.
In case (2), the file is closed upon return.

=cut

sub deleteFile {
	my ($tree, $oldValue) = @_;
	if (!ref($tree)) { # a file name
		open DELETEFILE, "+<$tree";
		deleteFile(\*DELETEFILE, $oldValue);
		close DELETEFILE;
		return;
	}
	my $token = readToken($tree);
	return if ($token eq 'null'); # empty tree; can't delete
	if ($token eq 'bucket') {
		my $startPos = tell($tree);
		my ($parentPos, $endPos, @values) = readBucket($tree);
		my $stringified = join(' ', @$oldValue);
		# print "Deleting $stringified from bucket\n";
		my $index;
		for ($index = 0; $index <= $#values; $index++) {
			if ($stringified eq join(' ', @{$values[$index]})) { # found it
				# print "removing ", printPoint($oldValue), "\n";
				splice @values, $index, 1;
				seek $tree, $startPos, SEEK_SET;
				my $toPrint = "parent $parentPos ";
				$toPrint .= sprintf("%${numberLength}d ", $endPos);
				my $point;
				foreach $point (@values) {
					$toPrint .= printPoint($point) . " / ";
				}
				print $tree $toPrint;
				$startPos = tell $tree;
				print $tree " " x ($endPos - $startPos); # fill rest with blank
				last;
			} # if found
		} # foreach point
	} elsif ($token eq 'internal') {
		# print "Deleting from internal node\n";
		my ($Direction, $Discriminant, $leftPos, $rightPos) =
			readInternal($tree);
		if (defined($$oldValue[$Direction]) and
		  $$oldValue[$Direction] < $Discriminant) {
			seek $tree, $leftPos, SEEK_SET;
			deleteFile($tree, $oldValue);
		} else {
			seek $tree, $rightPos, SEEK_SET;
			deleteFile($tree, $oldValue);
		}
	} elsif ($token eq 'kd') { # start of file
		my $rootPos = expect("^\\d+\$", $tree);
		seek $tree, $rootPos, SEEK_SET;
		deleteFile($tree, $oldValue);
	} else {
		die "bad representation at mark " . tell($tree);
	}
} # deleteFile;

## test code subs

sub TestKD {
	my $method = "file";
	my ($myTree);
	srand 2;
	$| = 1;
	$numDims = 2;
	my $numPoints = 10;
	# initialize tree
	TimeSoFar("");
	if ($method eq "file") {
		my $rootPos;
		unlink "tmp.out";
		writeFile("tmp.out", undef); # just an empty tree
		my ($serial);
		open DOITFILE, "+<tmp.out" or die "Can't open tmp.out; stopped ";
		$myTree = \*DOITFILE;
		bless $myTree, 'IO::Handle'; # so we can flush it
		for ($serial = 0; $serial < $numPoints; $serial += 1) {
			seek $myTree, 0, SEEK_SET;
			# expect("kd", $myTree);
			# $rootPos = expect("^\\d+\$", $myTree);
			# seek $myTree, $rootPos, SEEK_SET;
			insertFile($myTree, randomPoint());
			$myTree->flush(); # so we can read it without closing
			print "after inserting in file, I reread and see: \n",
				printTree(readFile("tmp.out")), "\n";
		}
		close($myTree);
		exit(0);
	} else { # internal
		my $serial;
		for ($serial = 0; $serial < $numPoints; $serial += 1) {
			$myTree = insertCore($myTree, randomPoint());
		}
		print "initial tree:\n", printTree($myTree), "\n";
		exit(0);
	}
	TimeSoFar("Setup of $numPoints $numDims-dimensional points," .
		" bucket size $bucketSize");
	probe($myTree);
	TimeSoFar("Single probe");
	unlink("tmp.out");
	writeFile("tmp.out", $myTree);
	TimeSoFar("Writing out tree");
	my $newTree = readFile("tmp.out");
	TimeSoFar("Reading tree");
	print "written, then read back:\n", printTree($newTree), "\n";
} # TestKD

sub randomPoint {
	my $answer = [];
	my $dim;
	for ($dim = 0; $dim < $numDims; $dim += 1) {
		push @$answer, int(rand 1000);
	}
	# print "random point: ", printPoint($answer), "\n";
	return($answer);
} # randomPoint

sub probe {
	my ($tree) = @_;
	my $point = randomPoint();
	my $result;
	push @Trace, "probing " . printPoint($point);
	$result = nearest($tree, $point, 1);
	print "nearest to ", printPoint($point), " is ",
		printPoint($result->{'Point'}), " at distance ",
		$result->{'Distance'}, "\n";
	pop @Trace;
} # probe

sub TimeSoFar {
	my ($msg) = @_;
	my $time = (times())[0];
	my $lastTime;
	$lastTime = 0 if !defined($lastTime);
	print $msg, ": ", sprintf("%.2f", $time - $lastTime),
		" seconds\n" if $msg ne "";
	$lastTime = $time;
}

END {
	print "Trace: ", join(";", @Trace), "\n" if $#Trace != -1;
}

1
