.CH "Concurrency"
.SH 1 "Presenting the material"
.lp
This is a technically difficult chapter, which is one reason I delay it until
the students have a good intuition about what an operating system is all about
and they have begun to realize the need for concurrency control.
By now, they have seen that simultaneous access to files can cause havoc,
and they have a feeling that the different managers inside the kernel are
likely to share some data structures, possibly with chaotic results.
.pp
Many of the mechanisms shown in this chapter can be discussed with enough
precision to warrant use of program fragments.
You will find it valuable to go over these in class directly from the book.
To save you effort in copying from book to blackboard, I have put line numbers
in the programs so you can
.q point
to particular places.
.pp
It is important to point out that all mechanisms for
synchronization rest ultimately on some atomicity in the hardware.
On uniprocessors, most instructions cannot be interrupted.
On multiprocessors, accesses to the same location in central store are
serialized.
The existence of local caches (such as the one on the Sequent Balance
computers) makes the situation more complex, but does not change the general
picture.
.pp
There is a discontinuity in this chapter when semaphores are introduced.
For the first time, the scheduler is brought into the picture to avoid busy
wait.
In addition, programming language concerns begin to loom larger:
How can we protect the programmer from silly errors, assist in modularizing
code, and make it easy to write and understand synchronization code?
Most of the methods shown after semaphores can be treated as syntactic sugar
added to make semaphores palatable.
The semaphores are still there, hidden in the implementation,
.pp
I have found that distinguishing between \fImutual exclusion\fP and the more
general \fIsynchronization\fP aids in interpreting the strengths of the various
proposals covered in this chapter.
I use the synchronization graph (you may prefer to call it a \fBprecedence
graph\fP) of Figure 8.1 as a running example of the latter.
It is complex enough so that cobegin/coend are not sufficient to describe it.
.pp
The non-alternating switch solutions were not easy to invent, and they are not
easy to explain.
I find that a timeline picture is useful, showing two processes that have
interest in entering the critical region.
First, I let \fBA\fP go ahead unhindered.
Then I let \fBB\fP do the same.
Then I let \fBB\fP try to enter while \fBA\fP is in the region, showing how it
is delayed.
Finally, I let both \fBA\fP and \fBB\fP try to enter at about the same time.
You could use different color chalk to indicate passage of time during which a
process is uninterested in the critical region, interested and busy waiting,
blocked, and actually in the critical region.
.pp
Don't feel you have to show Dekker's solution if you prefer to skip directly to
Peterson's.
I left Dekker's in because it was the first published correct
solution.
There are other nice solutions, too.
Here is one due to Kessels.
.[
kessels arbitration
.]
.(l
.FS
.ne 6
 1	\fWtype\fP
 2		Activity = 0..1;
 3	\fWvar\fP
 4		Code : \fWarray\fP [Activity] \fWof\fP 0..1;
 5		WantIn : \fWarray\fP [Activity] \fWof\fP Boolean := false;
 6	
.ne 12
 7	\fWprocedure\fP BeginRegion(Who : Activity);
 8	\fWvar\fP	Other : Activity := 1 - Who;
 9	\fWbegin\fP
10		WantIn[Who] := true;
11		Code[Who] := Code[Other] \fWxor\fP Who;
12		\fWwhile\fP WantIn[Other] \fWand\fP
13			(Code[Who] = Code[Other] \fWxor\fP Who)
14		\fWdo\fP
15			\fWnull\fP;
16		\fWend\fP;
17	\fWend\fP BeginRegion;
18	
.ne 4
19	\fWprocedure\fP EndRegion(Who : Activity);
20	\fWbegin\fP
21		WantIn[Who] := false;
22	\fWend\fP EndRegion;
.FE
.)l
This solution has the marvelous property that there are no variables (like Turn
in Peterson's solution) that are writable by both activities.
.pp
The text doesn't mention it, but a lock is sufficient to implement the
synchronization constraints of Figure 8.1.
Each arrow is associated with a different lock.
Initially, all locks are set to
.q held.
When a process at the tail of an arrow finishes, it \fIreleases\fP the
associated lock by calling EndRegion.
A process at the head of an arrow executes BeginRegion to wait for its
predecessor.
(There are other solutions as well, cognate to the three solutions given with
semaphores.)
.pp
I have tried various names for the P and V operations for semaphores.
In the previous edition, I called them Acquire and Release.
This nomenclature was usually fine, but it caused confusion between
allocating/deallocating a semaphore data structure and acquiring/releasing.
Also, when semaphores are used for more general synchronization, the language
of acquisition and release is not appropriate.
I like the Up/Down names a lot better.
The only problem is that for acquisition, one tends to say,
.q "The process picks up this semaphore." "  "
Unfortunately, one picks up a semaphore by executing Down.
.pp
You may want to introduce an atomic UpDown operation that takes two semaphores.
It is the only clear way to implement monitors using semaphores.
.pp
Conditional critical regions are my favorite structure.
They tend to lead to the clearest programs.
The Edison programming language uses only conditional critical regions.
.[
edison brinch
.]
.pp
Monitors have been fairly successful.
Modula (not Modula-2) implements them, as does Mesa.
Unfortunately, everyone uses a different variant.
.pp
The producers-consumers (bounded buffer) problem and the readers-writers
problem are standard ways to show off how well different synchronization
mechanisms work.
There are more complex problems as well, such as various extensions to the
dining philosophers in which there are several resource types.
It is not at all trivial to fully
.q solve
the readers-writers problem; a full solution must avoid starvation of either
class.
.pp
I find path expressions hard to invent and harder to understand.
They do have the beauty that they prevent meaningless sequences of accesses,
which distinguishes them from the other methods in this chapter.
On the other hand, I find invariant expressions very natural.
Still, students have a tough time writing reasonable programs with them.
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
After both \fBA\fP and \fBB\fP have finished, x could be any integer from 10 to
20.
.EP \n+(E+.
(a)
There are a lot of ways to partition the actions.
For example, there is one way that has 9 separate activities (make no
connections), 12 ways to get 8 separate activities (make one connection),
and C(12,2)\(mi4 ways to get 7 separate activities (make two of the possible 12
direct connections; disallow the four that look like {D,E,B}.
It gets unreasonably hard to enumerate the rest.
.br
(b)
This is too hard to solve reasonably.
.br
(c)
Activities like {C,D} include elements with no precedence constraints between
them at all.
.EP \n+(E+.
We will assume that instead of EnableInterrupts and DisableInterrupts, we have
an atomic routine called SetInterrupts which not only establishes a new
interrupt scheme (either \fIon\fP or \fIoff\fP), but returns the previous value
of that scheme.
We will stack old values.
.(b
.FS
 1	\fWprocedure\fP BeginRegion;
 2	\fWbegin\fP	
 3		Push(SetInterrupts(off));
 4	\fWend\fP BeginRegion;
 5	
 6	\fWprocedure\fP EndRegion;
 7	\fWbegin\fP
 8		SetInterrupts(Pop);
 9	\fWend\fP EndRegion;
.FE
.)b
.EP \n+(E+.
Here is a three-process switch.
.(l
.FS
.ne 4
 1	\fWtype\fP Activity : (A, B, C);
 2	
 3	\fWvar\fP Turn : Activity := A;
 4	
.ne 7
 5	\fWprocedure\fP BeginRegion(Who : Activity);
 6	\fWbegin\fP
 7		\fWwhile\fP Turn <> Who \fWdo\fP
 8			\fWnull\fP;
 9		\fWend\fP;
10	\fWend\fP BeginRegion;
11	
.ne 10
12	\fWprocedure\fP EndRegion(Who : Activity);
13	\fWbegin\fP
14		\fWif\fP Who = A \fWthen\fP
15			Turn := B
16		\fWelsif\fP Who = B \fWthen\fP
17			Turn := C
18		\fWelse\fP
19			Turn := A;
20		\fWend\fP;
21	\fWend\fP EndRegion;
.FE
.)l
.EP \n+(E+.
Yes, Dekker's non-alternating switch uses busy waiting if \fBB\fP is in its
region, Turn\ =\ A, and \fBA\fP is trying to enter its region.
It loops through the following:
.(b
.FS
14          \fWrepeat\fP { busy wait if our turn, Other interested }
15              \fWwhen\fP \fWnot\fP Interested[Other] \fWexit\fP outer;
16          \fWuntil\fP Turn = Other;
.FE
.)b
.EP \n+(E+.
Say \fBA\fP is trying to enter, and \fBB\fP is currently in its region.
Say that Turn = A.
Then \fBA\fP is in lines 14\(mi16, and Interested[A] = true.
When \fBB\fP exits its region, it sets Interested[B] to false.
In order to starve \fBA\fP, we must keep executing \fBB\fP now, because
\fBA\fP would immediately encounter the \fWexit\fP in line 15 and enter its
region.
When \fBB\fP tries again to enter the region, it will languish in lines
19\(mi21, during which Interested[B] is false.
It can't enter before \fBA\fP, so \fBA\fP is not starved.
.PP
In the other case, Turn = B.
In this case, \fBA\fP is in lines 19\(mi21, and Interested[A] = false.
When \fBB\fP leaves its region, it sets Turn = A.
Again, let us try to get \fBB\fP back into the region before scheduling \fBA\fP
again.
Since Interested[A] = false, \fBB\fP can enter the region again.
\fBA\fP seems to be starving.
However, Turn = A, so when we next run \fBA\fP, it will leave lines 19\(mi21
and enter lines 14\(mi16.
We have reduced the situation to the previous case, in which \fBA\fP is not
starved.
.EP \n+(E+.
Not solved here.
.EP \n+(E+.
Not solved here.
.EP \n+(E+.
Let's try to starve \fBA\fP.
It must be looping in lines 13\(mi15, which means Interested[B] = true
and Turn = B.
When \fBB\fP leaves the region, it sets Interested[B]=false.
We will keep executing \fBB\fP to try to get back into the region.
When \fBB\fP gets to line 12, it sets Turn = A and starts looping in lines
13\(mi15.
Now \fBA\fP can continue, because Turn = A.
So \fBB\fP cannot enter the critical region again without A going first.
A does not starve.
.EP \n+(E+.
If mutual exclusion fails, both \fBA\fP and \fBB\fP are in the critical region.
Interested is only modified in line 11 and 20; we know that Interested[A] =
true and Interested[B] = true.
If both \fBA\fP and \fBB\fP have passed line 12 but not the loop of lines
13\(mi15, only one will pass the loop.
If one has not gotten to line 11, the other can get past the loop.
The only possible conflict is if one has gotten past 11 but not 12,
and the other is at the loop.
The first will soon pass 12, and we are back to a previous case.
.EP \n+(E+.
Not solved here.
The method involves recursion.
.EP \n+(E+.
The simplest solution is this:
.(l
.FS
.ne 9
 1  \fWtype\fP Lock = integer := 0;
 2
 3  \fWprocedure\fP BeginRegion(\fWvar\fP L : Lock);
 4  \fWbegin\fP
 5      \fWwhile\fP TestAndInc(L) > 1 \fWdo\fP
 6          Dec(L); (* must be atomic *)
 7      \fWend\fP;
 8  \fWend\fP BeginRegion;
 9
.ne 4
10  \fWprocedure\fP EndRegion(\fWvar\fP L : Lock);
11  \fWbegin\fP
12      Dec(L);
13  \fWend\fP;
.FE
.)l
This solution suffers from two potential problems.
First, the lock may overflow.
However, it can only grow as high as the number of competing activities,
which should be fairly small.
Second, when two activities both try to acquire the lock simultaneously,
they can repeatedly fail, causing livelock.
The likelihood of livelock increases with the number of competing activities.
A solution is to put a random wait after line 6, possibly coupled with an
exponential backoff.
That means that on the @i@th failure iteration of the loop, one waits a random
amount of time uniformly distributed between 0 and @2 sup i@.
One could also use the current value of L to determine the current amount of
contention, and set the backoff by that.
.EP \n+(E+.
It is hard to know when to reset Done to false; one must wait until all the
successor activities that need to wait for Done have seen the value.
We can't have different successors able to reset Done, because Done then
becomes an unprotected write-shared variable.
.PP
The best solution seems to be to let Done be an integer that counts how many
times its associated activity has finished.
Finishing increments Done[Who], and WaitFor takes a new argument: the value to
expect.
.EP \n+(E+.
A cycle in a synchronization graph means that some activity depends on itself,
which means there is no way that activity can ever start.
To get the effect of repetition of a graph, one needs new constructs, such as a
starter activity that is needed only once, and afterwards is not required as a
prerequisite for any other activity.
.EP \n+(E+.
Without mutual exclusion, one must ensure that each shared variable can only
be written with one activity, although many may read it.
The Done variable is only written by the associated activity, so that method
works.
A Start variable would have to be written in an activity by all the
activities that precede it, which is unsafe.
So this method does \fInot\fP work.
However, the third option is still open:
Associate a variable with each arrow in the synchronization graph.
.EP \n+(E+.
This problem, although on the surface identical to the previous one,
can be solved, since semaphores are more powerful than unprotected shared
variables.
Whenever an activity finishes, let it perform Up on the semaphore of each of
its successors.
When an activity wants to start, let it perform Down on its semaphore once for
each of its predecessors.
.EP \n+(E+.
Each time a \fWregion\fP statement is seen within another, the compiler may add
a directed edge in a graph of shared variables that shows the order of nesting.
If there is ever a cycle in that directed graph, there is a conflicting order.
However, \fWregion\fP statements could include procedure calls, and the called
procedure could itself have conflicting \fWregion\fP statements.
To be on the safe side, the compiler needs to include procedures as well as
shared variables in the graph; any call on a procedure within a \fWregion\fP
statement adds an edge, and any \fWregion\fP statement within a procedure adds
an edge.
This method will be conservative (there will be cycles where there is in fact
no conflict, because the seemingly illegal \fWregion\fP statement is actually
within a conditional that does not get executed), but it will be safe.
An exact solution is not only input-dependent (conditionals may depend on input
whose values are unavailable to the compiler), but also undecidable even if
input is known.
.EP \n+(E+.
Critical regions (not conditional critical regions) protect shared variables by
providing mutual exclusion.
If we have a variable to indicate that an activity is finished,
it must be shared between that activity and its successors.
Therefore, the variable must be accessed through a \fWregion\fP statement.
However, if the value of the variable indicates that the successor may not yet
proceed, it is necessary to leave the \fWregion\fP statement to let the
predecessor modify the variable.
Therefore, a solution will have this appearance:
.(b
.FS
1	\fWloop\fP "outer"
2		\fWregion\fP v \fWdo\fP
3			\fWif\fP MayProceed(v) \fWthen\fP
4				\fWexit\fP "outer";
5			\fWend\fP;
6		\fWend\fP;
7	\fWend\fP;
.FE
.)b
The outer loop constitutes busy waiting.
.EP \n+(E+.
We use some auxiliary variables to indicate which platoon is currently filling
and executing for readers and writers.
.(l
.FS
.ne 5
\fWvar\fP
	FillingReaders, FillingWriters : integer := 0; (* platoons *)
	ExecutingReaders, ExecutingWriters : integer := 0; (* platoons *)
	ActiveReaders : integer := 0; (* counts *)

.ne 9
\fWprocedure\fP StartRead;
\fWvar\fP
	MyPlatoon : integer;
\fWbegin\fP
	MyPlatoon := FillingReaders;
	\fWregion\fP MutEx \fWdo\fP
		\fBif\fP ActiveReaders = 0 \fBthen\fP
			inc(FillingWriters);
		\fBend\fP;
		\fWawait\fP MyPlatoon <= ExecutingReaders;
	\fBend\fP; (* region *)
	inc(ActiveReaders);
\fWend\fP StartRead;

.ne 5
\fWprocedure\fP EndRead;
\fWbegin\fP
	\fWregion\fP MutEx \fWdo\fP
		dec(ActiveReaders);
		\fWif\fP ActiveReaders = 0 \fBthen\fP
			inc(ExecutingWriters);
		\fBend\fP;
	\fBend\fP; (* region *)
\fWend\fP EndRead;

.ne 8
\fWprocedure\fP StartWrite;
\fBvar\fP
	MyPlatoon : integer;
\fWbegin\fP
	\fWregion\fP Mutex \fWdo\fP
		MyPlatoon := FillingWriters;
		inc(FillingWriters); (* Every writer is in a new platoon. *)
		inc(FillingReaders); (* start a new reader platoon *)
		\fWawait\fP MyPlatoon <= ExecutingWriters;
	\fWend\fP; (* region *)
\fWend\fP StartWrite;

.ne 4
\fWprocedure\fP EndWrite;
\fWbegin\fP
	\fWregion\fP Mutex \fWdo\fP
		inc(ExecutingWriters);
		inc(ExecutingReaders);
	\fWend\fP; (* region *)
\fWend\fP EndWrite;
.FE
.)l
.EP \n+(E+.
Not solved here.
.EP \n+(E+.
It is hard to find a case where it would be desirable for an activity that
executes \fWsignal\fP in a monitor to perform more work before leaving the
monitor.
Generally, the \fWsignal\fP can be delayed.
One reason to put it earlier is to avoid remembering that it may need to be
done later.
Another is that there must be several \fWsignal\fPs, and they must be placed in
some order.
One far-fetched example is for a shared stack of integers where we require
that there be no more than 5 odd-valued entries.
There are two conditions:  NonEmpty and FewerThanFiveOdds.
When an odd entry is popped, both conditions need to be signalled.
.EP \n+(E+.
We will take the liberty of placing condition variables inside records.
We also export a type.
.(l
.FS
.ne 4
 1	\fWmonitor\fP Semaphores;
 2	
 3	\fWexport\fP Semaphore, Up, Down;
 4	
.ne 7
 5	\fWtype\fP
 6		Semaphore =
 7			\fWrecord\fP
 8				Value : integer;
 9				Unlocked : \fWcondition\fP;
10			\fWend\fP;
11	
.ne 8
12	\fWguard\fP \fWprocedure\fP Down(S : Semaphore);
13	\fWbegin\fP
14		\fWif\fP S.Value < 1 \fWthen\fP
15			\fWwait\fP S.Unlocked;
16		dec(S.Value);
17	\fWend\fP Down;
18	
.ne 7
19	\fWguard\fP \fWprocedure\fP Up(S : Semaphore);
20	\fWbegin\fP
21		inc(S.Value);
22		\fWsignal\fP S.Unlocked;
23	\fWend\fP Up;
24	
25	\fWend\fP Semaphores;
.FE
.)l
.EP \n+(E+.
It is necessary to ensure that Advance be atomic, that is, that two
simultaneous calls to Advance actually advance the counter twice.
Mutual exclusion is likely necessary here, unless there is an atomic
increment instruction.
.EP \n+(E+.
Here is the program for the consumers in an event count and sequencer solution
to multiple producers, multiple consumers.
.(l
.FS
.ne 8
 1	...
 2	\fWvar\fP
 3		ConsumerTurn : \fWsequencer\fP;
 4	
 5	\fWactivity\fP Consumer;
 6	\fWvar\fP	
 7		SequenceNumber : integer;
 8		Value : Datum;
.ne 11
 9	\fWbegin\fP
10		\fWloop\fP
11			... { compute Value }
12			SequenceNumber := Ticket(Consumer);	{ discover turn }
13			Await(Out \fWreaches\fP SequenceNumber);	{ wait for turn }
14			Await(In \fWreaches\fP SequenceNumber + 1);
15				{ wait for production }
16			Value := Buffer[SequenceNumber \fWmod\fP Size];
17			Advance(Out);
18		\fWend\fP;
19	\fWend\fP Consumer;
.FE
.)l
.EP \n+(E+.
Here is how to use the DoubleDown procedure to implement the dining philosophers
problem.
.(b
.FS
 1	\fWprocedure\fP Philosopher(Which : 0..4);
 2	\fWbegin\fP
 3		\fWloop\fP { think \(em eat loop }
 4			Think; { occupy some finite amount of time in thought. }
 5			DoubleDown(Chopstick[Which], Chopstick[(Which + 1) \fWmod\fP 5);
 6			Eat;   { for a finite amount of time }
 7			Up(Chopstick[Which]);
 8			Up(Chopstick[(Which+1) \fWmod\fP 5]);
 9		\fWend\fP; { think \(em eat loop }
10	\fWend\fP Philosopher;
.FE
.)b
The easiest way to see the problem if we don't use MutualExclusion in
implementing DoubleDown is to imagine a reduced, 2-philosopher problem.
Say \fBA\fP gets past line 9, having gotten Ticket1 = 1.
Then \fBB\fP gets two tickets, Ticket1 = 2 and Ticket2 = 1.
Then \fBA\fP gets its last ticket, Ticket2 = 2.
\fBA\fP will block at line 16, and \fBB\fP will block at line 13;
we have a deadlock.
.PP
It would not help to place the declaration of MutualExclusion at
line 6 instead of line 2.
In that case, MutualExclusion would be a local variable, and would not exclude
anything.
.EP \n+(E+.
The given path expression shows how hard it is to interpret these things.
The following figure shows some of the constraints.
.(b
.GS
height 3
width 3
file ../figs/fig.8.ex.path
.GE
.)b
The squares indicate one-at-a-time procedures;
the ovals indicate any number at a time.
Some of the constraints (rule 3, for instance) are not shown at all.
.PP
Certainly the expression prevents reader-writer conflict, by rule 3.
Say there are lots of eager writers and readers waiting, and one writer is
active.
Only one reader is allowed to wait in TryRead (rule 2), and only one in
EnterRead (rule 1); the others wait in DoRead.
If that is the situation, then all writers are waiting in DoWrite, not TryWrite
(rule 2).
When the current writer finishes, the reader in TryRead will move to StartRead
(rule 5).
At this point, all the readers could move through EnterRead and TryRead to
StartRead, and then the entire platoon of writers will move to TryWrite.
.PP
New readers will now be prevented from entering TryRead (rule 2), and will be
backed up at EnterRead and DoRead.
All the writers in the current platoon will therefore be allowed to precede any
new writer.
In fact, newly arriving writers will immediately join the TryWrite platoon,
so Writers have preference of Readers, who therefore are subject to starvation.
.EP \n+(E+.
This path expression avoids readers-writers conflict (rule 3).
If there are both waiting readers and writers, one writer will alternate with a
platoon of readers.
Reader platoons form in TryRead, unless there is a waiting writer in TryWrite,
in which case the reader platoon forms in DoRead.
.EP \n+(E+.
Not solved here.
.EP \n+(E+.
Not solved here.
It is not hard to make all producers mutually exclusive, and likewise all
consumers:
.(b
.FS
	\fWpath\fP Produce
	\fWpath\fP Consume
.FE
.)b
The hard part is to say that the number of Consumes may not exceed the number
of Produces.
Path expressions are not good at counting.
.EP \n+(E+.
To implement invariant expressions using conditional critical regions, one can
pepper each procedure with code that increments the necessary counters.
RequestCount is incremented at the top, StartCount after an initial \fWawait\fP
statement (more on that soon), and FinishCount right before exit.
Right after the StartCount is incremented, the procedure should \fWawait\fP the
necessary condition.
In the bounded buffer example, PutBuffer could be rewritten as follows:
.(b
.FS
 1	\fWprocedure\fP PutBuffer(What : Datum);
 2	\fWbegin\fP
 3		\fWregion\fP Mutex \fWdo\fP
 4			inc(RequestCountPutBuffer);
 5			\fWawait\fP StartCountPutBuffer - FinishCountGetBuffer < Size;
 6			\fWawait\fP FinishCountPutBuffer = StartCountPutBuffer;
 7			inc(StartCountPutBuffer);
 8		\fWend\fP;
 9		Buffer[NextIn] := What;
10		NextIn := (NextIn + 1) \fWmod\fP Size;
11		\fWregion\fP Mutex \fWdo\fP
12			inc(FinishCountPutBuffer);
13		\fWend\fP;
14	\fWend\fP PutBuffer;
.FE
.)b
.EP \n+(E+.
Invariant expressions automatically increment the counters, and they restrict
the programmer to a particularly simple form of expression.
However, I still find conditional critical regions easier to use.
.EP \n+(E+.
The following invariant-expression solution to the producers-consumers problem
serializes concurrent producers only long enough to
reserve slots but may fill those slots in parallel.
Unfortunately, a producer might reserve a slot and not fill it for a very long
time, during which time the other producer may busily fill other slots in the
array.
The second producer could even overwrite the reserved slot.
In addition, GetBuffer could take data from the reserved but still empty
slot.
The only obvious solution is ugly:  Let the consumer zero out each datum as it
is read, and let it wait before reading a slot that is still zero.
Unfortunately, I don't know how to express this solution with invariant
expressions.
.(l
.FS
.ne 4
 1	\fWmodule\fP BoundedBuffer;
 2	
 3	\fWexport\fP GetBuffer, PutBuffer;
 4	
.ne 4
 5	\fWconst\fP
 6		Size = 10; { number of elements in the buffer at once }
 7	\fWtype\fP
 8		Datum = ... { data type of the contents of the buffer }
.ne 4
 9	\fWvar\fP
10		Buffer : \fWarray\fP 0:Size\(mi1 \fWof\fP Datum;
11		NextIn, NextOut : 0..Size\(mi1;
12			{ index of next datum to insert, delete }
.ne 5
13	\fWinvariant\fP StartPut
14		StartCount(StartPut) \(mi FinishCount(GetBuffer) < Size
15			{ no overproduction }
16		CurrentCount(StartPut) = 0 { prevent producer conflict }
.ne 2
17	\fWinvariant\fP PutBuffer
18			(* empty *)
.ne 5
19	\fWinvariant\fP GetBuffer
20		StartCount(GetBuffer) \(mi FinishCount(PutBuffer) < 0
21			{ no overconsumption }
22		CurrentCount(GetBuffer) = 0 { prevent consumer conflict }
23	
.ne 6
24	\fWprocedure\fP StartPut() : integer; (* returns where *)
25		\fWbegin\fP
26			StartPut := NextIn;
27		NextIn := (NextIn + 1) \fWmod\fP Size;
28		\fWend\fP StartPut;
29	
.ne 5
30	\fWprocedure\fP PutBuffer(What : Datum; Where : integer);
31	\fWbegin\fP
32		Buffer[Where] := What;
33	\fWend\fP PutBuffer;
34	
.ne 5
35	\fWprocedure\fP GetBuffer(\fWvar\fP Result : Datum);
36	\fWbegin\fP
37		Result := Buffer[NextOut];
38		NextOut := (NextOut + 1) \fWmod\fP Size;
39	\fWend\fP GetBuffer;
.FE
.)l
.EP \n+(E+.
The trick to using a busy-waiting construct without always busy waiting is to
try for a while, and then ask the scheduler to continue trying later.
In particular, a process needing a lock can try with test-and-set for, say,
100 attempts.
If the lock is still held by another process, the waiter can submit a service
call:
.q "Try to get this lock for me".
The kernel will periodically try again, and when it can get the lock, it
releases the blocked process.
In this way, most acquisitions of the lock will not require a context switch,
but those that would waste too much time busy-waiting do.
The number of attempts in the first phase can be tuned to take about as much
time as a context switch.
The Sequent Balance 21000 multicomputer has a hardware test-and-set instruction.
.EE
.SH 1 "Examination Questions"
.ES
.nr E+ 0 1
.EP \n+(E+.
What purpose is there in initializing a semaphore value to 0?
What purpose is there in initializing a semaphore value to 1?
What purpose is there in initializing a semaphore value to 2?
.PP
\fBAnswer\fP:
Initializing to 0 is useful in synchronization when one activity must wait
until another finishes.
The waiter will perform Down and be blocked until the other finishes and
performs Up.
Initializing to 1 is useful for mutual exclusion.
Initializing to 2 is useful for allowing at most 2 activities to use a sharable
resource simultaneously.
.EP \n+(E+.
Show how to implement a semaphore by using a critical region.
In particular, fill in the following code:
.(l
.FS
.ne 7
\fWtype\fP
	Semaphore =
		\fWrecord\fP
			MutEx : integer;
			Value : integer;
		\fWend\fP;

.ne 4
\fWprocedure\fP Up(\fWshared var\fP Sem : Semaphore);
\fWbegin\fP
	<fill this in>
\fWend\fP Up;

.ne 4
\fWprocedure\fP Down(\fWshared var\fP Sem : Semaphore);
\fWbegin\fP
	<fill this in>
\fWend\fP Down;
.FE
.)l
.PP
\fBAnswer\fP:
.(l
.ne 7
\fWprocedure\fP Up(\fWshared var\fP Sem : Semaphore);
\fWbegin\fP
	\fWregion\fP Sem \fWdo\fP
		inc(Sem.Count);
	\fWend\fP;
\fWend\fP Up;

.ne 7
\fWprocedure\fP Down(\fWshared var\fP Sem : Semaphore);
\fWbegin\fP
	\fWregion\fP Sem \fWdo\fP
		\fWawait\fP Sem.Count >= 0;
		dec(Sem.Count);
	\fWend\fP;
\fWend\fP Down;
.)l
.EP \n+(E+.
In what sense do \fBmonitors\fP provide synchronization atomicity?
.PP
\fBAnswer\fP:
They provide mutual exclusion of guard procedures, which means that invocations
of those procedures are serialized.
.EP \n+(E+.
The following program fragment shows how the Dining Philosophers problem can be
written with invariant expressions.
Fill in the invariants in lines 7 and 9.
It is not necessary to prevent deadlock, only to make sure chopsticks are not
over-allocated.
.(l
.FS
.ne 4
 1	\fWmodule\fP Sticks;
 2	
 3		\fWexport\fP	 GetStick1, GetStick2, GetStick3, GetStick4, GetStick5,
 4			PutStick1, PutStick2, PutStick3, PutStick4, PutStick5;
 5	
.ne 5
 6		\fWinvariant\fP GetStick1
 7			<fill this one in>
 8		\fWinvariant\fP PutStick1
 9			<fill this one in>
10		(* All the other Chopticks have similar invariants, so we omit them. *)
11	
.ne 4
12		\fWprocedure\fP GetStick1
13		\fWbegin\fP
14			(* empty procedure *)
15		\fWend\fP GetStick1;
16	
.ne 4
17		\fWprocedure\fP PutStick1
18		\fWbegin\fP
19			(* empty procedure *)
20		\fWend\fP PutStick1;
21	
.ne 4
22		(* All the other procedures are similar. *)
23	
24	\fWend\fP Sticks;
25	
.ne 13
26	(* sample Philospher *)
27	
28	\fWprocedure\fP Philosopher1;
29	\fWbegin\fP
30		\fWloop\fP
31			Think;
32			GetStick1;
33			GetStick2;
34			Eat;
35			PutStick1;
36			PutStick2;
37		\fWend\fP;
38	\fWend\fP Philosopher1;
.FE
.)l
.PP
\fBAnswer\fP:
.(b
.FS
 7			StartCount(GetStick1) \(mi FinishCount(PutStick1) = 0;
				{ prevent over\(miallocation }

 9			FinishCount(GetStick1) \(mi StartCount(PutStick1) = 1;
				{ optional: prevent giving back an unallocated stick }
.FE
.)b
To prevent deadlock, make sure no more than 4 philosophers get a stick at any
one time.
We do this by adding an additional invariant to line 7:
.(b
.FS
			StartCount(GetStick2) \(mi FinishCount(PutStick2) +
			StartCount(GetStick3) \(mi FinishCount(PutStick3) +
			StartCount(GetStick4) \(mi FinishCount(PutStick4) +
			StartCount(GetStick5) \(mi FinishCount(PutStick5) < 4
.FE
.)b
.EP \n+(E+.
Invariant expressions are usually implemented with semaphores.
It is also possible to implement semaphores by invariant expressions.
Fill in lines 8 and 10 in the following program fragment.
.(l
.FS
.ne 4
 1	\fWmodule\fP Semaphore;
 3		{ simulates a semaphore with an initial value of 1 }
 4	
 5		\fWexport\fP Down, Up;
 6	
.ne 4
 7		\fWinvariant\fP Down
 8			<fill this one in>
 9		\fWinvariant\fP Up
10			<fill this one in>
12	
.ne 4
13		\fWprocedure\fP Down;
14		\fWbegin\fP
15			{ empty procedure }
16		\fWend\fP Down;
17	
.ne 6
18		\fWprocedure\fP Up;
19		\fWbegin\fP
20			{ empty procedure }
21		\fWend\fP Up;
22	
23	\fWend\fP Semaphore;
24	
.ne 11
25	{ sample application }
26	
27	\fWprocedure\fP Application;
28	\fWbegin\fP
29		\fWloop\fP
30			NonCritical;
31			Down;
32			Critical;
33			Up;
34		\fWend\fP;
35	\fWend\fP Application;
.FE
.)l
.PP
\fBAnswer\fP:
.(b
.FS
 8	StartCount(Down) \(mi FinishCount(Up) <= 1

10	(* empty invariant *)
.FE
.)b
.EP \n+(E+.
Process \fBA\fP should finish before \fBB\fP starts, and \fBB\fP should finish
before \fBC\fP or \fBD\fP starts.
.br
(a)  Draw the synchronization graph for this set of precedence constraints.
.br
(b)  Show how to use two semaphores to provide the necessary synchronization.
.PP
\fBAnswer\fP:
.br
(a)
.(b
.GS
width 1
height 1
file ../figs/fig.8.exam1
.GE
.)b
.br
(b)  Call one semaphore DoneA, the other DoneB, and initialize them both to 0.
Let \fBA\fP perform Up(DoneA) when it is finished.
Let \fBB\fP perform Down(DoneA) as it starts and Up(DoneB) \fItwice\fP when it
is finished.
Let both \fBC\fP and \fBD\fP perform Down(DoneB) as they start.  
.EP \n+(E+.
What properties must a solution to the readers-writers problem have?
.PP
\fBAnswer\fP:
If a writer is active, it must exclude all other readers and writers.
No reader should starve.
No writer should starve.
It is not required, but is highly desirable, that multiple readers be allowed
to act simultaneously.
.EP \n+(E+.
What is the effect of a programming error that calls EndRegion instead of
BeginRegion under the following implementations?
.br
(a)  Disabling interrupts
.br
(b)  Semaphores
.br
(c)  Test-and-set locks
.PP
\fBAnswer\fP:
.br
(a)  The region will not be exclusive, but no other damage will accrue.
.br
(b)  Every time the region is executed, the MutEx semaphore will have two
extra Up operations performed.
The result is that the semaphore will become useless; no region depending on it
will remain mutually exclusive.
.br
(c)  The region will not be exclusive.
On a multiprocessor, the mistake will break exclusion currently in force for
any other competing region.
.EE
.lp
.rs
