#!/usr/local/bin/tclsh

# check logs for suspicious activity
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set myHost [info hostname]
while {[file rootname $myHost]!=$myHost} {
	set myHost [file rootname $myHost]
}

switch -regexp $tcl_platform(os) {
	{Linux} {
		set tailCommand tac
	} default {
		set tailCommand "tail -r"
	}
} ;# switch

# Form 1: check for a bad thing within the log.  7 fields.
# OS:  pattern that specifies what OS this entry applies to
# logFile:  name of log file to search
# bad-pattern:  specifies lines of log file that look suspicious
# except-pattern:  mitigating circumstance.  Use $^ for 'none'.
# major: major name of alarm.  Use "%h" to put in name of current host
# minor: minor name of alarm.  Use "%h" to put in name of current host
# bad: badness level of alarm

proc Form1 {line} {
	global tcl_platform myHost tailCommand
	if {![regexp [lindex $line 0] $tcl_platform(os)]} return ;# not ours
	puts "Working on $line"
	set logFileName [lindex $line 1]
	set badPattern [lindex $line 2]
	set exceptPattern [lindex $line 3]
	regsub {(.*)(%h)(.*)} [lindex $line 4] "\\1$myHost\\3" major
	regsub {(.*)(%h)(.*)} [lindex $line 5] "\\1$myHost\\3" minor
	set badness [lindex $line 6]
	set logFile [open "|$tailCommand $logFileName" r]
	if {[catch {set logFile [open "|$tailCommand $logFileName" r]} errorMsg]} {
		report $major $minor 10 "$tailCommand $logFileName: $errorMsg"
		puts "Couldn't run $tailCommand $logFileName"
		return
	}
	while {[gets $logFile logLine]>=0} { # one line of log file
		# puts "line $logLine"
		if {[regexp $badPattern $logLine] &&
		  ![regexp $exceptPattern $logLine]} {
			puts "found: $major $minor $badness \"$logLine\""
			report $major $minor $badness "$logLine"
		break ; # no need to continue in this log file.
		}
	}
}; # Form1

# Form 2: check for a bad log length.  5 fields.
# OS:  pattern that specifies what OS this entry applies to
# logFile:  name of log file to search
# bad-length:  number of bytes at which a yellow warning is due
# minor: minor name of alarm.  Use "%h" to put in name of current host

proc Form2 {line} {
	global tcl_platform myHost
	if {![regexp [lindex $line 0] $tcl_platform(os)]} return ;# not ours
	puts "Working on $line"
	set logFileName [lindex $line 1]
	set badLength [lindex $line 2]
	regsub {(.*)(%h)(.*)} [lindex $line 3] "\\1$myHost\\3" minor
	if {[catch {set fileLength [file size $logFileName]} errorMsg]} {
		report $myHost $minor 10 "$logFileName: $errorMsg"
		puts "Couldn't find $logFileName"
	} elseif {$fileLength > $badLength} {
		report $myHost $minor 15 "$logFileName is $fileLength bytes"
		puts "($myHost,$minor): $logFileName is $fileLength bytes"
	} else {
		report $myHost $minor 0 "$logFileName is $fileLength bytes"
	}
}; # Form2

proc Form3 {line} {
	global tcl_platform myHost tailCommand
	if {![regexp [lindex $line 0] $myHost]} return ;# not ours
	puts "Working on $line"
	set logFileName [lindex $line 1]
	set badPattern [lindex $line 2]
#	set exceptPattern [lindex $line 3]
	regsub {(.*)(%h)(.*)} [lindex $line 3] "\\1$myHost\\3" major
	regsub {(.*)(%h)(.*)} [lindex $line 4] "\\1$myHost\\3" minor
	set badness [lindex $line 5]
	set logFile [open "|$tailCommand $logFileName" r]
	if {[catch {set logFile [open "|$tailCommand $logFileName" r]} errorMsg]} {
		report $major $minor 10 "$tailCommand $logFileName: $errorMsg"
		puts "Couldn't run $tailCommand $logFileName"
		return
	}
	while {[gets $logFile logLine]>=0} { # one line of log file
		# puts "line $logLine"
		if {[regexp $badPattern $logLine] } {
#		  ![regexp $exceptPattern $logLine]} {
			puts "found: $major $minor $badness \"$logLine\""
			report $major $minor $badness "$logLine"
		break ; # no need to continue in this log file.
		}
	}
}; # Form3

set confFile [open ./log.conf r]
while {[gets $confFile line]>=0} { # one line of configuration
    if [regexp ^# $line] continue ;# comment line
    if [regexp ^$ $line] continue ;# empty line
	if {[llength $line] == 7} {
		Form1 $line
	} elseif {[llength $line] == 6} {
		Form3 $line
	} elseif {[llength $line] == 4} {
		Form2 $line
	} else {
		puts "Bad line in ./log.conf: $line"
		report internal log.test 20 "Bad line in log.conf: $line"
	}
}
