#!/usr/local/bin/perl

# Use SNMP to find the status of network printers
#  Copyright (c) 1999 Raphael Finkel.  See the COPYRIGHT file for details.

# Not all net printers speak this protocol accurately.  Our experience:
# Lexmark Optra LaserPrinter
# 	Version 79.123.1 only responds to uptime
# 	Version 79.120.1 Can give "warning" with no error bits set
# 	Version 15.84.1 Can give "warning" with no error bits set
# Lexmark Optra C LaserPrinter
# 	Version 79.130.1 responds accurately
# Lexmark Optra S 2455
# 	Version 1.8.19 responds accurately

use Net::SNMP; # not part of the standard distribution; you need to install it.

# MIB paths
$internet = '1.3.6.1'; # RFC 1155
	$mgmt = $internet . '.2'; # RFC 1155
		$mib2 = $mgmt . '.1'; # RFC 1213, 1155
			$host = $mib2 . '.25'; # RFC 1514 (and for all the rest, too)
				$hrDevice = $host . '.3'; 
					$hrDeviceTable = $hrDevice . '.2'; 
						$hrDeviceEntry = $hrDeviceTable . '.1'; # guess
							$hrDeviceType = $hrDeviceEntry . '.2'; 
							$hrDeviceStatus = $hrDeviceEntry . '.5'; 
					$hrPrinterTable = $hrDevice . '.5';
						$hrPrinterEntry = $hrPrinterTable . '.1'; # guess
							$hrPrinterStatus = $hrPrinterEntry.'.1';
							$hrPrinterDetectedErrorState = $hrPrinterEntry.'.2';

my $identifierString = $mib2 . '.1.1.0';
my $sysUpTime = $mib2 . '.1.3.0';

# MIB constants
$hrDevicePrinter = '1.3.6.1.2.1.25.3.1.5';

$debug = 0;
$debug = 1 if defined($ARGV[0]);

# Printer messages
@deviceStatusMessage = ('unused entry', 'unknown', 'running', 'warning',
	'testing', 'down');
@printerStatusMessage = ('unused entry', 'other', 'unknown', 'idle',
	'printing', 'warmup');

@errorBits = ('low paper.', 'no paper.', 'low toner.', 'no toner.',
	'door open.', 'jammed.', 'offline.', 'needs service.');

sub ShowErrors { # convert bits in error message
	my ($code) = @_;
	$code =~ s/\[(.*)\]/$1/;
	$code = hex $code;
	if ($code == 0) {
		$messages .= "ok.";
		# print "ok. ";
		return;
	}
	$mask = 1;
	for ($bit = 0; $bit < 8; $bit++) {
		if ($mask & $code) {
			$messages .= "$errorBits[$bit] ";
			# print "$errorBits[$bit] ";
		}
		$mask *= 2;
	} # for each bit in code
} # ShowErrors

sub makeCall {
	my ($request, $name, $messageArray) = @_;
	my ($response, $answer, $message);
	$response = $sess->get_request($request);
	if (!defined($response)) {
		# print "no response for $name\n";
		return(-1);
	} else {
		$answer = $response->{$request};
		if (defined($messageArray)) {
			$message = $$messageArray[$answer];
			# print "$message. ";
			$messages .= "$message. ";
			return($answer);
		} else {
			# print "returning [$answer]\n";
			return($answer);	
		}
	}
} # makeCall

foreach $dest (split /\s+/, `enumerate.tcl netPrinters`) { # one printer
# foreach $dest (('indigo')) { # debug 

	# print "$dest: ";

	($sess,$error) = Net::SNMP->session(-hostname=>$dest);
	if (!defined($sess)) {
		# print "Could not set up session: $error\n";
		next;
	}

	my $uptime = makeCall($sysUpTime, "up time");
	if ($uptime == -1) {
		report("netprinters", $dest, 15, "no response");
		next;
	}

	$messages = makeCall($identifierString, "identifier string") . ' Up ' .
		$uptime . '. ';

	my $base = $hrDeviceType;
	$patience = 10;
	for ($element = 1; $element < $patience; $element++) { # search for printer
		$answer = makeCall($base . '.' . $element, "device");
		if ($answer eq $hrDevicePrinter) {
			last;
		}
	} # search for printer element
	if ($element == $patience) {
		# print "Cannot find printer device.\n";
		report("netprinters", $dest, 13,
			"$messages No other information.");
		next;
	}

	makeCall($hrDeviceStatus . '.' . $element, "device",
		\@deviceStatusMessage);

	makeCall($hrPrinterStatus . '.' . $element, "printer",
		\@printerStatusMessage);

	$answer = makeCall($hrPrinterDetectedErrorState . '.' . $element,
		"error state"); 
	# print "reported error state [$answer]\n";
	ShowErrors($answer);

	# print "\n";

	$badness = ($messages =~ /ok/) ? 1 : 14;
	report("netprinters", $dest, $badness, "$messages");
	$sess->close();
} # one printer

sub report {
	my ($major, $minor, $badness, $msg) = @_;
	if (!$debug) {
		system('report', $major, $minor, $badness, $msg);
	}
	print "$minor $badness \"$msg\" \n";
}
