# Package to expand HostClass-based set-theoretic expressions
#
# Expand(string), where string is a Boolean combination of host
#	names and host macros.	The host macros are defined by the HostClass
#	attribute of the SAT Hosts relation.	Operators are
#		+ set union (a missing operator is taken as +)
#		- set subtraction
#		* set intersection
#	Usual precedence and associativities apply.	Parentheses may be used for
#	grouping.

%STDHOSTNAMES = (); # list of host names
%STDHOSTMACROS = (); # expanded host macros
%KNOWNEXPANSIONS = (); # previous expansion results, to avoid redoing
$EXPANDFAULT = $FALSE; # result code

# where the scheme file and data file are.
$HOSTSDATAFILE = "/usr/local/etc/sat/db/Hosts/data";
$HOSTSSCHEMEFILE = "/usr/local/etc/sat/db/Hosts/scheme";

sub AppendHostMacros {
	my ($macro, $value) = @_;
	if (exists $STDHOSTMACROS{$macro}) {
		$STDHOSTMACROS{$macro} = "$STDHOSTMACROS{$macro} $value";
	} else {
		$STDHOSTMACROS{$macro} = $value;
	}
} # AppendHostMacros

sub InitStdHostMacros {
#	for each host, record its name in %STDHOSTNAMES and add its name to all
#	keys in %STDHOSTNAMES mentioned in its HostClass.
	my ($host_tuple, $class, $key, @Classes);
	my @Hosts = PutRecords ($HOSTSSCHEMEFILE, $HOSTSDATAFILE);
	for $host_tuple (@Hosts) {
		$STDHOSTNAMES{$host_tuple->{'HostName'}} = $host_tuple->{'HostName'};
		AppendHostMacros ("ALL", $host_tuple->{'HostName'});
		@Classes = split / +/, $host_tuple->{'HostClass'};
		for $class (@Classes) {
			AppendHostMacros($class, $host_tuple->{'HostName'});
		}
	}
} # InitStdHostMacros

%ANSWERS = (); # temporary hash storing previous expansion results
$SCANSTRING = ""; # Globle variable for parsing.

#    Recursive-descent parser for expressions that use this BNF:
#				S --> T | T + T | T - T
#				T --> P | P * P
#				P --> atom | null | ( S )
#	Answers are associative arrays, whose indices are machine names and whose
#	values are those same names.  The result is the union of all the elements
#	in the array.
#	The routines put their answers in %ANSWERS{}, clearing it before using it.
#	The string being scanned is the global $SCANSTRING so that it can be
#	modified (front stripped off) by all rountines. Stripping always leaves
#	the string with no white spaces at the start.
#
sub ExpandP {
#	P --> atom | null | ( S )
	# print "ExpandP: scanstring is now [$SCANSTRING]\n";
	my ($host_tuple, $key, $atom, @Expansion);
	if ($SCANSTRING =~ /^\s*\(/) { # ( S )
		$SCANSTRING =~ s/^\s*\(//; # remove matched part
		ExpandS();
		if ($SCANSTRING =~ /^ *\)/) {
			$SCANSTRING =~ s/^ *\)//;
		} else {
			warn "UnMatched parenthesis; remainder is $SCANSTRING";
			$EXPANDFAULT = $TRUE;
		}
	} elsif ($SCANSTRING eq "") { # null
		%ANSWERS = ();
	} else { # atom
		%ANSWERS = ();
		$SCANSTRING =~ s/^\s*(\w*)//; # strip off the first atom. 
		$atom = $1; # atom found in previous line
		if (defined($STDHOSTNAMES{$atom}) && $STDHOSTNAMES{$atom} eq $atom) {
			$ANSWERS{$atom} = $atom;
		} elsif (defined $STDHOSTMACROS{$atom}) { # fill in the expansion
			@Expansion = split / /, $STDHOSTMACROS{$atom};
			for $host_tuple (@Expansion) {
				$ANSWERS{$host_tuple} = $host_tuple;
			}
		} else { # syntax error
			warn "I cannot recognize $atom.\n";
			$EXPANDFAULT = $TRUE;
		}
	} # atom
} # ExpandP

sub ExpandT {
#	T --> P | P * P
	# print "ExpandT: scanstring is now [$SCANSTRING]\n";
	my ($key, %SubAnswer);
	ExpandP();
	while ($TRUE) { # Iterate to get multiple P * P * P
		if ($SCANSTRING =~ /^ *\*/) { # found *
			$SCANSTRING =~ s/^ *\*//;
			%SubAnswer = %ANSWERS;
			ExpandP(); # next operand
			for $key (keys %ANSWERS) { # intersect
				if (! exists $SubAnswer{$key}) {
					delete $ANSWERS{$key};
				}
			}
		} else { # no more P instances
			last;
		}
	} # interate to get multiple P * P * P
} # Expand T

sub ExpandS {
#	S --> T | T + T | T - T
	# print "ExpandS: scanstring is now [$SCANSTRING]\n";
	my ($key, %SubAnswer);
	ExpandT();
	while ($TRUE) { # make "+" and "-" left associative by iteration.
		# Check for the leading "+" or whitespaces, strip them off
		# and perform the second parsing. Perform the operation
		# afterwards.
		if ($SCANSTRING =~ /^\s*(\+|\w)/) { # either a + or another T
			$SCANSTRING =~ s/\s*\+//; # remove +, if any
			%SubAnswer = %ANSWERS; # temporary storage
			ExpandT(); # next operand.
			for $key (keys %SubAnswer) { # union
				$ANSWERS{$key} = $SubAnswer{$key};
			}
		} elsif ($SCANSTRING =~ /^\s*-/) { # -
			$SCANSTRING =~ s/^\s*-//; # remove it
			%SubAnswer = %ANSWERS; # temporary storage
			ExpandT(); # next operand.
			for $key (keys %ANSWERS) { # subtraction
				delete $SubAnswer{$key}
			}
			%ANSWERS = %SubAnswer;
		} else { # no more T instances
			last;
		}
	} # iterate to get multiple T + T - T
} # ExpandS

sub Expand {
	($SCANSTRING) = @_;
	# print "Expand: scanstring is now [$SCANSTRING]\n";
	my ($key, $Result, $Hold);
	$EXPANDFAULT = $FALSE;
	$SCANSTRING =~ s/^\s*//; # Strip off leading whitespaces.
	$Hold = $SCANSTRING; # for comparison later
	if (exists $KNOWNEXPANSIONS{$SCANSTRING}) {
		# print "known\n";
		return $KNOWNEXPANSIONS{$SCANSTRING};
	} else { # first time expanding this string
		ExpandS();
		if ($SCANSTRING ne "") {
			warn "Garbage at end of string: $SCANSTRING";
			$EXPANDFAULT = $TRUE;
		}
		$Result = "";
		for $key (keys %ANSWERS) {
			$Result = "$Result $key";
		}
		$Result =~ s/^\s*//; # remove leading whitespace
		if (! $EXPANDFAULT) {
			$KNOWNEXPANSIONS{$Hold} = $Result; # for future reference
		}
		return $Result;
	} # first time expanding this string
} # Expand

sub TestIt{
	my $Result;
	while (<STDIN>) {
		chop;
		$Result = Expand($_);
		print "$Result ";
		if ($EXPANDFAULT) {
			print "[there was a fault]"
		}
		print "\n";
	}
} # TestIt

InitStdHostMacros();

1; # package return value.
