#	Description:
#		This subroutine retrieves formatted data from $datafile,
#		and returns the data in forms of an array.
#
#		The formats of the data are described in $schemefile, which
#		itself is in a special format. In SAT, scheme files are all
#		of the same format: each file has 18 general descriptions
#		that describe a particular relation, and followed by various
#		number of attributes of that relation.
#
#		The format of each data item, scheme description item, or
#		attribute item is of the following form: it starts with two
#		bytes, followed by an item of the length "byte1 * 256 + byte2".
#		Here, byte1 and byte2 are the decimal values of those two
#		bytes. They are obtained by using "ord".
#
#		PutRecords first looks into $schemefile,
#		retrieves the names of the attributes into an array, reads
#		in the data from $datafile according to the number of
#		attributes this relation has, stores the data into an array
#		of associative arrays, where each associative array has the
#		attributes as keys and data as values accordingly, and then
#		returns the array.
#
#	Parameters:
#		$schemefile: the absolute path of the scheme file that
#			describes that format of $datafile.
#		$datafile:	where the data is stored.
#
#	Return:
#		expected data in an array of hash tables.

sub PutRecords {
	my ($schemefile, $datafile) = @_;
	my @Scheme = (); # The description part of the scheme
	my @Attributes = (); # The names of attributes
	my ($FirstChar, $NumberOfChars, $NumberOfFields, $Type, $dummy,
		$NumberOfRecords);
	local (*SCHEME); # file handle
	open (SCHEME, "<$schemefile") || die "Can't open $schemefile: $!";

	# GetString reads the desired number of bytes from *FD and stores
	# the bytes in $Ptr, which is a reference to a scalar, and returns
	# the number of bytes to be read next time.
	#	*FD:		the File-Handler to be read.
	#	$Ptr:		where the string is to be stored after read.
	#	$Length: the offset of the string.
	sub GetString {
		local (*FD, $Ptr, $Length) = @_;
		my ($Number1, $Number2);
		read FD, $$Ptr, $Length;
		$Number1 = ord chop $$Ptr;	
		$Number2 = ord getc FD;		
		$Number2 += $Number1 * $ASCII;
		return $Number2;	# return length.
	} # GetString

	$FirstChar = ord getc SCHEME; # length of the first item
	$NumberOfChars = ord getc SCHEME;
	$NumberOfChars += $FirstChar * $ASCII;
	$NumberOfFields = 0; 
	while ($TRUE) { # read the descriptions
		$NumberOfChars = GetString (SCHEME, \$Scheme[$NumberOfFields],
			$NumberOfChars);
		if ($NumberOfFields == $NUMBEROFCLAUSES) {
			last;
		}
		$NumberOfFields++;
	} # read desscriptions
	
	$NumberOfFields = 0;
	while ($NumberOfChars) { # read the attributes
		$NumberOfChars = GetString (SCHEME, \$Attributes[$NumberOfFields],
			$NumberOfChars);
		$NumberOfChars = GetString (SCHEME, \$dummy, $NumberOfChars);
		$NumberOfChars = GetString (SCHEME, \$Type, $NumberOfChars);
		if ($Type == $PATTERN) {
			$NumberOfChars = GetString (SCHEME, \$dummy, $NumberOfChars);
		} elsif ($Type == $RANGE) {
			$NumberOfChars = GetString (SCHEME, \$dummy, $NumberOfChars);
			$NumberOfChars = GetString (SCHEME, \$dummy, $NumberOfChars);
			$NumberOfChars = GetString (SCHEME, \$dummy, $NumberOfChars);
		}
		$NumberOfFields ++;
	}
	close SCHEME;
	# Read data
	my @Data = ();
	local (*DATA); # file handle
	open (DATA, "<$datafile") || die "Can't open $datafile: $!";
	$FirstChar = ord getc DATA; # length of the first data item
	$NumberOfChars = ord getc DATA;
	$NumberOfChars += $FirstChar * $ASCII;
	$NumberOfRecords = 0;
	my $field;
	while($NumberOfChars) { # read the data into @Data
		$field = 0;
		while ($field < $NumberOfFields) {
			$NumberOfChars =
			GetString (DATA, \$Data[$NumberOfRecords]{$Attributes[$field]},
				$NumberOfChars);
			$field ++;
		}
		$NumberOfRecords ++;
	} # read data
	close DATA;
	return @Data; # return @Data
} # PutRecords

1; # package return value
