#!/usr/local/bin/perl -T

# Report for hosts relation. Creates the following files:
#	 /etc/hosts					local host database
#	 /.rhosts					remote login permissions (based on Secure)
#	 /etc/hosts.equiv			equivalent main hosts (based on Secure)
#	 /etc/hosts.lpd (BSD)		hosts known to the printer daemon(CS|EXTERNAL)
#	 /etc/lp/Systems (Solaris)	same purpose
#	 /etc/mail/sendmail.cw		hosts that have mail sent here
#	 ~operator/.rhosts			remote login for operator
#	 /etc/hostinfo				data for the 'hostinfo' program
#	 /var/named/				files for name daemon (only at SOA)
#		CurrentLevel			Current level of named files
#			named.hosts_cs
#			named.rev_cs146
#			named.rev_cs154
#			named.local
#	 /usr/local/etc/sat/lib/distaliasfile
#								Definition of host macros for SAT schemas
#	 /etc/netgroup				List of all machines at our site
#	 /etc/fingerdir/			GNU finger configuration
#		serverhost				on every machine
#		clients					just on the server host

# imports packages:
#	Site.pl			site-specific customization
#	Definitions.pl	includes some constant variables
#	Relation.pl		includes some subroutines that are used most often.
#	PutRecords.pl	includes the PutRecords subroutine.
#	GetExpand.pl	a recursive-descent parser for HostClass expressions.
#
$SATHOME = "/usr/local/etc/sat";
require "$SATHOME/perlinclude/Site.pl";
require "$SATHOME/perlinclude/Definitions.pl";
require "$SATHOME/perlinclude/PutRecords.pl";
require "$SATHOME/perlinclude/Relation.pl";
require "$SATHOME/perlinclude/GetExpand.pl";

$Relation = "Hosts";

require "../../perlinclude/StartUp.pl";

# debug versions output files not usually in /etc
if ($Debug) {
	$DistAlias = "$MyRoot/lib/distaliasfile";
	$Rhosts = "$MyRoot/.rhosts";
	$NamedDir = "$MyRoot/named";
	$Operator = "$MyRoot/operator/.rhosts";
} else {
	$DistAlias = "$SATHOME/lib/distaliasfile";
	$Rhosts = "/.rhosts";
	$NamedDir = "/var/named";
	($dummy,$dummy,$dummy,$dummy,$dummy,$dummy,$dummy,$dir,$dummy) =
		getpwnam("operator");
	$Operator = "$dir/.rhosts";
}

# files usually located in /etc
$Equiv = "$Root/hosts.equiv";
$Lpd = "$Root/hosts.lpd";
$Systems = "$Root/lp/Systems";
$SendMail = "$Root/mail/sendmail.cw";
$HostInfo = "$Root/hostinfo";
$NetGroup = "$Root/netgroup";
$FingerServer = "$Root/fingerdir/serverhost";
$FingerClients = "$Root/fingerdir/clients";

# get OS type for localhost
$OSType = RunProgram ("$HOSTINFO OSType");

# if the host is of type "mach", special HostFile.
if ($OSType eq "mach") {
	$HostFile = "$Root/hosts.network";
} else {
	$HostFile = "$Root/hosts";
}
$HostFile = Untaint($HostFile);

# a temporary scalar used in OpenFile and CloseFile subroutines.
$TempFile = "";

# exit if the relation hasn't been modified since the last reporting,
if (! $Debug && SameAgeAsRelation ($lastmodified, $HostFile)) {
	exit (0);
}

sub AppendSecureHosts {
#	append $what to $StdSecureHosts{$which}.
	local ($which, $what) = @_;
	$StdSecureHosts{$which} = "$StdSecureHosts{which} $what";
}

# we want to clean up the execution before the program terminates.
END { Clean (\$TempFile, $HostFile, $lastmodified)};

# read the records from the Hosts relation into @Hosts.
@Hosts = PutRecords ($schemefile, $datafile);

{ # /etc/hosts
	OpenFile ($HostFile, \$TempFile, OUTFILE);
	print OUTFILE "# This file is automatically generated as a SAT report from"
	." Hosts.\n";
	print OUTFILE "# Internet host table.\n";
	print OUTFILE "127.0.0.1\tlocalhost\tloghost\n";
	
	for $host_tuple (@Hosts) { # include this host in the file
		print OUTFILE "$host_tuple->{'Address'}";
		print OUTFILE "\t$host_tuple->{'FullHostName'}";
		print OUTFILE "\t$host_tuple->{'HostName'}";
		if ($host_tuple->{'ShortHostName'} ne $host_tuple->{'HostName'}
			&& $host_tuple->{'ShortHostName'} ne "") {
			print OUTFILE "\t$host_tuple->{'ShortHostName'}";
		}
		if ($host_tuple->{'FullHostName'} eq $MyFullHostName) {
			$MyMailDest = $host_tuple->{'MailDest'};
		}
		if ($host_tuple->{'FullHostName'} =~ /$MyDomain[0]/
			&& $host_tuple->{'OtherAliases'} ne "") {
			# no aliases for hosts not in my domain.
			print OUTFILE "\t$host_tuple->{'OtherAliases'}";
		}
		print OUTFILE "\n";
	} # for all hosts
	CloseFile ($HostFile, 0644, \$TempFile, OUTFILE, "");
	# If the localhost is a NeXT machine,
	# load the information about hosts into NetInfo.
	if (IncludesMe ("NEXT", $MyHostName, $MyFullHostName) ) {
		system "/usr/bin/niload -d hosts / < $HostFile";
	}
} # /etc/hosts

{ # /.rhosts
	# side effect:  constructs StdSecureHosts
	%StdSecureHosts = ();
	OpenFile ($Rhosts, \$TempFile, OUTFILE);
	for $host_tuple (@Hosts) {
		$SecureHostList = Expand ($host_tuple->{'Secure'});
		@SecureHosts = split / /, $SecureHostList;
		for $securehost_tuple (@SecureHosts) {
			if ($securehost_tuple eq $MyHostName
			  && $host_tuple->{'HostName'} ne $MyHostName) {
				AppendSecureHosts $securehost_tuple,
					$host_tuple->{'FullHostName'};
				print OUTFILE "$host_tuple->{'FullHostName'} root\n"; 
				last;
			}
		}
	}
	CloseFile ($Rhosts, 0600, \$TempFile, OUTFILE, "");
} # /.rhosts

{ # /etc/hosts.equiv
	# uses StdSecureHosts
	OpenFile ($Equiv, \$TempFile, OUTFILE);
	@SecureHosts = split / /, $StdSecureHosts{$MyHostName};
	for $securehost_tuple (@SecureHosts) {
		# We could put this host in $Equiv, but we choose not to.
	}
	# close the temporary file and rename it to $Equiv.
	# Set the permission to 0644.
	CloseFile ($Equiv, 0644, \$TempFile, OUTFILE, "");
} # /etc.hosts.equiv

{ # /etc/hosts.lpd or /etc/lp/Systems
	if ($OSType ne "solaris") {
		# For non-solaris hosts, create /etc/hosts.lpd.
		OpenFile ($Lpd, \$TempFile, OUTFILE);
		print OUTFILE "# This file is automatically generated as a SAT report"
			. " from Hosts.\n";
		print OUTFILE "# List of hosts that may send to my lpd.\n";
		print OUTFILE "# (Those listed in /etc/hosts.equiv may also).\n";
		for $host_tuple (@Hosts) {
			if ($host_tuple->{'HostClass'} =~ /(EXTERNAL|$MySite)\b/) {
			print OUTFILE "$host_tuple->{'FullHostName'}\n";
			}
		}
		CloseFile ($Lpd, 0644, \$TempFile, OUTFILE, "");
	} elsif ($OSType eq "solaris") {
		# For solaris hosts, create /etc/lp/Systems.
		OpenFile ($Systems, \$TempFile, OUTFILE);
		print OUTFILE "# This file is automatically generated as a SAT report"
			. " from Hosts.\n";
		print OUTFILE "# List of hosts that may communicate (both directions)"
			. " to my lpd.\n";
		print OUTFILE "# You can read this list with /usr/sbin/lpsystem -l.\n";
		print OUTFILE "#VERSION=1\n";
		print OUTFILE "# Format, all on one line:
#	Host:Pass:-:HostType:-:TimeOut:Retry:-:-:Comment
#
#	Host may be + to allow all.	May be a short name.	May not be an IP addr.
#	Pass must be x
#	HostType is s5 or bsd.	**NOTE**	As of 2.4 Solaris, use bsd
#	TimeOut and Retry are in minutes or n for never.
#		TimeOut
#			 0: quit lpNet as soon as requests are done
#			 never: leave lpNet running (often a bad idea)
#		Retry gives the time limit during which 5 connections are tried.

";
		print OUTFILE "+:x:-:bsd:-:0:10:-:-:\n";
		for $host_tuple (@Hosts) {
			if ($host_tuple->{'HostClass'} =~ /(EXTERNAL|$MySite)/) {
			print OUTFILE "$host_tuple->{'FullHostName'}:x:-:bsd"
				.":-:0:10:-:-:\n";
			}
		}
		CloseFile ($Systems, 0644, \$TempFile, OUTFILE, "");
	} # Solaris
} # /etc/hosts.lpd or /etc/lp/Systems

{ # /etc/mail/sendmail.cw
	# only if mail server is running on localhost
	if ($MyMailDest eq $MyFullHostName) {
		OpenFile ($SendMail, \$TempFile, OUTFILE);
		foreach $Domain (@MyDomain) {
			print OUTFILE "$Domain\n";
		}
		for $host_tuple (@Hosts) {
			if ($host_tuple->{'MailDest'} eq $MyFullHostName) {
				print OUTFILE "$host_tuple->{'FullHostName'}\n";
			}
		}
		CloseFile ($SendMail, 0644, \$TempFile, OUTFILE, "");
	} # I am running mail server
} # /etc/mail/sendmail.cw

{ # /etc/hostinfo
	# Used by the hostinfo program
	OpenFile ($HostInfo, \$TempFile, OUTFILE);
	print OUTFILE "# This file is automatically generated as a SAT report from"
		." Hosts.\n";
	print OUTFILE "#	It is data for the hostinfo and spp programs.\n";
	for $host_tuple (@Hosts) {
		if ($MyFullHostName ne $host_tuple->{'FullHostName'}) {
			next;
		} else { # $host_tuple refers to me
			print OUTFILE "FullHostName\n$host_tuple->{'FullHostName'}\n";
			print OUTFILE "HostName\n$host_tuple->{'HostName'}\n";
			print OUTFILE "ShortHostName\n$host_tuple->{'ShortHostName'}\n";
			print OUTFILE "OtherAliases\n$host_tuple->{'OtherAliases'}\n";
			print OUTFILE "Address\n$host_tuple->{'Address'}\n";
			print OUTFILE "OSType\n$host_tuple->{'OSType'}\n";
			print OUTFILE "OSVer\n$host_tuple->{'OSVer'}\n";
			print OUTFILE "MachType\n$host_tuple->{'MachType'}\n";
			$shortdesc = $host_tuple->{'Description'};
			$shortdesc =~ tr/\n/;/;
			print OUTFILE "Description\n$shortdesc\n";
			print OUTFILE "Location\n$host_tuple->{'Location'}\n";
			print OUTFILE "SatClass\n$host_tuple->{'SatClass'}\n";
			$MySatClass = $host_tuple->{'SatClass'};
			print OUTFILE "Secure\n$host_tuple->{'Secure'}\n";
			print OUTFILE "MailDest\n$host_tuple->{'MailDest'}\n";
			print OUTFILE "HostClass\n$host_tuple->{'HostClass'}\n";
			last;
		}
	} # for each Host
	CloseFile ($HostInfo, 0644, \$TempFile, OUTFILE, "");
} # /etc/hostinfo

{ # ~operator/.rhosts
	# only if localhost belongs to "model" SatClass
	if ($MySatClass eq "model") {
		OpenFile ($Operator, \$TempFile, OUTFILE);
		print OUTFILE "- This file is automatically generated as a SAT report"
			. " from Hosts.\n";
		print OUTFILE "$SiteOfAuth"." root\n";
		for $host_tuple (@Hosts) {
			if ($host_tuple->{'HostClass'} =~ /$MyClass/) {
			print OUTFILE "$host_tuple->{'FullHostName'}"." operator\n";
			}
		}
		# when debugging, can't chown and chgrp for "operator".
		CloseFile ( $Operator, 0644, \$TempFile, OUTFILE,
			$Debug ? $DebugOwner : "operator");
	} # MySatClass is "model"
} # ~operator/.rhosts

{ # DNS stuff: /var/named
	# only on $SiteOfAuth
	$LevelFile = "$NamedDir"."/CurrentLevel";
	$Refresh = "3600";		# Time to refresh ( in seconds: 1 hours)
	$Retry = "600";			# Time to retry	 ( in seconds: 10 minutes)
	$Expire = "3600000";	# Time to expire	( in seconds: 1000 hours)
	$Minimum = "86400";		#  ( in seconds: 24 hours)
	if ($SiteOfAuth eq $MyFullHostName) { # worth building DNS files
		# find out the mail host's IP address
		for $host_tuple (@Hosts) {
			if ($host_tuple->{'FullHostName'} eq $MyMailDest) {
				$MailIP = $host_tuple->{'Address'};
				last;
			}
		}
		# read, adjust and write level file
		open (LEVEL, "<$LevelFile") || die "Can't open $LevelFile for read $!";
		$Level = <LEVEL>;
		close LEVEL;
		chop $Level;
		@LevelArray = split /\./, $Level;
		if ($#LevelArray != 1) {
			die "Bad level: $Level\n";
		}
		$firstpart = $LevelArray[0];
		$secondpart = $LevelArray[1] + 1;
		if ($secondpart > 9999) {
			$firstpart = $firstpart + 1;
			$secondpart = 0;
		}
		$Level = "$firstpart".".$secondpart";
		OpenFile ($LevelFile, \$TempFile, OUTFILE);
		print OUTFILE "$Level\n";
		print OUTFILE "# used by Sat:Hosts to determine the serial # for named"
			." files.\n";
		CloseFile ($LevelFile, 0644, \$TempFile, OUTFILE, "");
		
		# generate the files in /var/named/*
		foreach $IP (keys %IPs) { # eg: $IPs{"146"} = "cs.engr.uky.edu"
			($Header,$Rest) = ($IPs{$IP} =~ /([^.]*).(.*)/);
				# eg: $Header = "cs", $Rest = "engr.uky.edu"
			MakeRev ($IP, $Header, $Rest); # /var/named/named.rev_cs$IP.
			# sub MakeRev is at end of this file
		}
		$Pattern = $MyDomain[0]; # eg: cs.engr.uky.edu
		$Pattern =~ s/^[^.]*\.//; # eg: engr.uky.edu 
		$Pattern =~ s/\./\\./g; # eg: engr\.uky\.edu
		foreach $Domain (@MyDomain) {
			($Header,$Rest) = ($Domain =~ /([^.]*).(.*)/);
			MakeHosts ($Header, $Rest, $Pattern);
			# /var/named/named.hosts_$Header.$Rest.
			# sub MakeHosts is at end of this file
		}
		MakeLocal (); # /var/named/named.local.
			# sub MakeLocal is at end of this file
		if (!$Debug) { # tell named that the files have changed
			system "kill -HUP `cat /etc/named.pid`";
		}
	} # I am $SiteOfAuth
} # DNS stuff

{ # /usr/local/etc/sat/lib/distaliasfile
	sub GenerateList {
		# Write list of hostnames of the same SATCLASS to OUTFILE
		local ($Header, $Pattern) = @_;
		print OUTFILE "\n$Header\n";
		my ($host_tuple);
		for $host_tuple (@Hosts) {
			if ($host_tuple->{'FullHostName'} =~ /$MyDomain[0]/
			  && $host_tuple->{'SatClass'} eq $Pattern) {
				print OUTFILE "$host_tuple->{'FullHostName'}\n";
			}
		}
	}

	$AliasLocation = RunProgram ("/usr/local/bin/whathost $SATHOME");
	# if I have my own copy, refresh /usr/local/etc/sat/lib/distaliasfile.
	if ($MyHostName eq $AliasLocation) {
		OpenFile ($DistAlias, \$TempFile, OUTFILE);
		print OUTFILE "# This file is automatically generated as a SAT report"
			. " from Hosts.\n";
		print OUTFILE "#	It is data for SAT schemata.\n";
		GenerateList ($MyClass . "MODEL", "model");
		GenerateList ($MyClass . "TRUSTED", "trusted");
		GenerateList ($MyClass . "COPY", "copy");
		GenerateList ($MyClass . "ACCESS", "access");
		CloseFile ($DistAlias, 0644, \$TempFile, OUTFILE, "");
	}
} # /usr/local/etc/sat/lib/distaliasfile

{ # /etc/netgroup
	# Only if I export netgroups via NFS.
	# Netgroups are used for exporting file systems and for rwall.
	if ($MySatClass eq "model") {
		$AllMine = $MySite . " ";
		OpenFile ($NetGroup, \$TempFile, OUTFILE);
		print OUTFILE "# This file is automatically generated as a SAT "
			. "report from Hosts.\n";
		print OUTFILE "#	It is data for exported file systems.\n";
		print OUTFILE "#	There seems to be a line-length limit, " .
			"so we define $MySite in stages.\n";
		for $host_tuple (@Hosts) { # put host into the /etc/netgroup file.
			if ($host_tuple->{'HostClass'} =~ /$MySite/ &&
			  $host_tuple->{'HostClass'} !~ /DOWN|PRINTER|DOS|WIN95/) {
				# reasonable to export to this host
				if ($host_tuple->{'ShortHostName'} ne "") {
					$AllMine = $AllMine . " $host_tuple->{'ShortHostName'}";
					print OUTFILE "$host_tuple->{'ShortHostName'}"
						." ($host_tuple->{'FullHostName'},,$MyDomain[0].)\n";
				} else { # short name not available; use long name
					$AllMine = $AllMine . " $host_tuple->{'FullHostName'}";
				}
			} # reasonable to export
		} # foreach $host
		print OUTFILE "$AllMine\n";
		CloseFile ($NetGroup, 0644, \$TempFile, OUTFILE, "");
		if (!$Debug) {
			system "/usr/lib/nis/nispopulate -F -f -u -p /etc netgroup";
		}
	} # I am a model host
} # /etc/netgroup

{ # /etc/fingerdir files
	# check out the finger host
	for $host_tuple (@Hosts) {
		if ($host_tuple->{'FullHostName'} =~ /$MyDomain[0]/
		  && $host_tuple->{'OtherAliases'} =~ /finger/) {
			$FingerHost = $host_tuple->{'FullHostName'};
			last;
		}
	}
	if (! defined $FingerHost) {
		print STDERR "No finger host\n";
	} else { # found the finger host
		OpenFile ($FingerServer, \$TempFile, OUTFILE);
		print OUTFILE "$FingerHost\n";
		CloseFile ($FingerServer, 0644, \$TempFile, OUTFILE, "");
		if ($MyFullHostName eq $FingerHost) { # I am the finger host
			OpenFile ($FingerClients, \$TempFile, OUTFILE);
			print OUTFILE "# This file contains a list of hosts that GNU"
				. " finger polls every so often\n";
			print OUTFILE "# to get user login information.\n";
			print OUTFILE "# Created by SAT for $FingerHost from Hosts.\n";
			for $host_tuple (@Hosts) {
				if ($host_tuple->{'FullHostName'} =~ /$MyDomain[0]/
				  && $host_tuple->{'HostClass'} =~ /SOLARIS|LINUX|IRIX|AIX/) {
					print OUTFILE "$host_tuple->{'FullHostName'}\n";
				}
			}
			CloseFile ($FingerClients, 0644, \$TempFile, OUTFILE, "");
		} # I am finger host
	} # found finger host
} # /etc/fingerdir files

# exit status.
exit (0);

sub MakeRev {
#	Create /var/named/named.rev* files.
	local ($id, $header, $rest) = @_;
	local (*OUTFILE);
	my $FileName = "$NamedDir"."/named.rev_"."$header"."$id";
	OpenFile ($FileName, \$TempFile, OUTFILE);
	print OUTFILE "; This file is automatically generated as a SAT report from"
		." Hosts.
;	Please don't modify directly.
;
; $FileName -- $header.$rest
;
$id.$RevIP.in-addr.arpa.\tIN\tSOA\t$SiteOfAuth"
.". $DudeToContact. (
\t\t\t\t\t\t$Level ; Serial
\t\t\t\t\t\t$Refresh ; Refresh
\t\t\t\t\t\t$Retry ; Retry
\t\t\t\t\t\t$Expire ; Expire
\t\t\t\t\t\t$Minimum ) ; Minimum
\t\t\t\tIN\tNS\t$SiteOfAuth.
\t\t\t\tIN\tNS\tncc.uky.edu.
\t\t\t\tIN\tNS\tnic.net.uky.edu.\n";
	my ($host_tuple);
	for $host_tuple (@Hosts) {
		if ($host_tuple->{'Address'} =~ /$IPPrefix.$id/) {
			# under my authority
			@SpAd = split /\./, $host_tuple->{'Address'};
			print OUTFILE "$SpAd[3]\tIN\tPTR\t$host_tuple->{'FullHostName'}.\n";
		}
	} # for hosts
	CloseFile ($FileName, 0644, \$TempFile, OUTFILE, "");
} # MakeRev

sub MakeHosts {
#	Generate /var/named/named.hosts_*
	local ($site, $domain, $pattern) = @_;
	my $FileName = "$NamedDir"."/named.hosts_"."$site".".$domain";
	local (*OUTFILE);
	OpenFile ($FileName, \$TempFile, OUTFILE);
	print OUTFILE "; This file is automatically generated as a SAT report from"
		. " Hosts.
;	Please don't modify directly.
;
; $FileName -- $site".".$domain
;
$site".".$domain.\t\tIN\tSOA\t$SiteOfAuth".". $DudeToContact. (
\t\t\t\t\t$Level ; Serial
\t\t\t\t\t$Refresh ; Refresh
\t\t\t\t\t$Retry ; Retry
\t\t\t\t\t$Expire ; Expire
\t\t\t\t\t$Minimum ) ; Minimum
\t\t\tIN\tNS\t$SiteOfAuth.
\t\t\tIN\tNS\tncc.uky.edu.
\t\t\tIN\tNS\tnic.net.uky.edu.
$site".".$domain.\t\tIN	A	 $MailIP
$site".".$domain.\t\tMX	0	 $MyMailDest.\n\n";
	my ($host_tuple, $als_tuple);
	for $host_tuple (@Hosts) {
		if ($host_tuple->{'FullHostName'} =~ /.*\.$site\.$pattern$/) {
			# under my authority
			printf (OUTFILE "%-16s\tIN\tA\t%s\n",
				$host_tuple->{'HostName'}, $host_tuple->{'Address'});
			printf (OUTFILE "%-16s\tIN\tMX\t0 %s.\n",
				$host_tuple->{'HostName'}, $host_tuple->{'MailDest'});
			printf (OUTFILE "%-16s\tIN\tMX\t8 pop.uky.edu.\n",
				$host_tuple->{'HostName'});
			printf (OUTFILE "%-16s\tIN\tMX\t10 ncc.uky.edu.\n",
				$host_tuple->{'HostName'});
			# don't output HINFO; it is nobody's business.
			if ($host_tuple->{'HostName'} ne $host_tuple->{'ShortHostName'} &&
			  $host_tuple->{'ShortHostName'} ne "") {
				printf (OUTFILE "%-16s\tIN\tCNAME\t%s\n",
					$host_tuple->{'ShortHostName'}, $host_tuple->{'HostName'});
			}
			if ($host_tuple->{'OtherAliases'} ne "") {
				my @als = split / /, $host_tuple->{'OtherAliases'};
				for $als_tuple (@als) {
				printf (OUTFILE "%-16s\tIN\tCNAME\t%s\n",
					 $als_tuple, $host_tuple->{'HostName'});
				}
			}
		} # under my authority
	} # for all hosts
	CloseFile ($FileName, 0644, \$TempFile, OUTFILE, "");
} # MakeHosts

sub MakeLocal {
#	Generates /var/named/named.local.
	my $FileName = "$NamedDir"."/named.local";
	local (*OUTFILE);
	OpenFile ($FileName, \$TempFile, OUTFILE);
	print OUTFILE "; This file is automatically generated as a SAT report from"
		. " Hosts.
;	Please don't modify directly.
;
; $FileName -- $SiteOfAuth
;
@\t\tIN\tSOA\t$SiteOfAuth".". $DudeToContact. (
\t\t\t\t\t$Level ; Serial
\t\t\t\t\t$Refresh ; Refresh
\t\t\t\t\t$Retry ; Retry
\t\t\t\t\t$Expire ; Expire
\t\t\t\t\t$Minimum ) ; Minimum
\t\t\tIN\tNS\t$SiteOfAuth.
1\t\t\tIN\tPTR\tlocalhost.\n";
	CloseFile ($FileName, 0644, \$TempFile, OUTFILE, "");
} # MakeLocal
