.CH "Resource Deadlock"
.lp
We have dealt so far with two fundamental resources, time and space.
Allocating these resources is a matter of scheduling \(em that is, deciding
which process should receive which resources at a given time.
The time resource is granted by placing the process in the ready list and then
letting it execute.
The space resource is granted by allowing part or all of the virtual space of
the process to reside in main store.
.pp
Both time and space are special.
Time \(em the ability to execute \(em is held by only one process at
a time (unless the computer is a multiprocessor).
.IN "multiprocessor"
After the resource has been held, it can be preempted and given for a while
to another process.
Space can be held by many processes at once and may be preempted after
a while.
All regions of physical space are alike to the process.
Every process requires both time and space.
.pp
We will devote this chapter to allocation of resources that represent
non-sharable, non-preemptable resources.
Such resources represent extra facilities beyond space and time, such as
devices: files, printers, and tape drives.
For example, a file might be restricted so that only one process may use it at
a time.
Similarly, a printer or a tape drive may be acquired by a process for a while
and then given back.
Finally, a process may need five tape drives during the course of its
computation, but not all at once.
.pp
Data structures within the kernel are often non-sharable, non-preemptable
resources.
For example, the kernel might have an array of context blocks.
When a new process is started, a free context block must be allocated.
As long as the process exists, that context block is neither shared nor
preempted.
Similarly, the file manager (which we discuss in Chapter 5) might have an array
of descriptors for open files.
When a file is opened, a free descriptor is allocated, and must not be shared
or preempted as long as the file is in use.
In a related vein, page frames that contain tied-down pages
.IN "Pages" "Tying down"
cannot be shared (unless processes are sharing pages) and are not preempted.
.pp
Once a process has been granted such a resource, it is not possible
to preempt the resource without injuring the process or the resource.
For example, if the use of a file is preempted, the process cannot
assume that when it gets to use the file again, it will be in the same state as
before.
The process has been injured and most likely should just quit.
Even worse, the process may have left the file in an inconsistent state.
For example, the file may represent a load image that the preempted process
.IN "Load image"
was in the middle of linking.
.IN "Linking"
Some links are complete and others are not.
In this case, preemption injures the resource, and it might never again be
usable.
The incompletely linked image cannot be loaded, nor can the linking be
finished.
.pp
Even though they are not preemptable, the resources we will discuss are all
\fBserially reusable\fP.
.IN Resources "Serially reusable"
After one process has finished using a tape drive, some other process may
certainly use it without harm.
.pp
Another property of the resources we will discuss here is that they are both
\fBdiscrete\fP and \fBbounded\fP.
.IN Resources "Discrete"
.IN Resources "Bounded"
This means that they come in individual chunks and that the amount of the
resource is limited, usually physically.
For example, if an installation has only seven tape drives,
it makes no sense to acquire 3.4 of them or 8 of them.
Time, in contrast, is neither discrete nor bounded.
The scheduler can divide it into pieces of any size, and there is always enough
to satisfy any finite process.
.pp
We will speak of \fBresource classes\fP, which are sets of identical resources.
.IN "Resources" "classes"
If a process does not care which tape drive it gets, then all tape drives are
identical as far as it is concerned, and they are therefore in the same class.
Likewise, in an installation with several line printers, it is often immaterial
which line printer is acquired; all line printers are in the same class.
.SH 1 "Nuts and bolts"
.lp
We will assume that when a process needs a resource, it will ask the kernel
for it by making a service call.
The part of the kernel responsible for these requests is the
\fBresource manager\fP.
.IN "Resource manager"
The same service call might ask for several resources at a time.
If the resources desired are available and the resource manager decides to
grant them, the process is allowed to continue.
The fact that the process now possesses the resources is marked in its context
block, and the resources are marked as belonging to that process.
.pp
The resource manager might decide not to grant the resources.
In this case, the process is placed in a resource-wait
.IN "Resource-wait list"
list, where it remains until the resource manager decides to honor it with the
resources.
(Instead, the resource manager could report failure immediately or after a
delay.)
.pp
When the process has finished using one or more resources, it returns them by
issuing another service call.
This call does not block the process, although the scheduler may take this
opportunity to switch to a different process for a while.
.IN "Process switch"
If a process tries to use a resource it has not acquired, the resource manager
treats this action as an error.
.pp
Here is a sample of the declarations needed for these service calls:
.(b
.FS
 1	\fWtype\fP
 2		ResourceClass = (Tape, Printer, ... ); { generic name of resource }
 3		ResourceIdentifier = integer; { specific name of resource }
 4		ResourceDescriptor = 
 5			\fWrecord\fP
 6				Kind : ResourceClass; { set by process }
 7				Which : ResourceIdentifier; { set by resource manager }
 8			\fWend\fP;
 9		Resources = \fWlist of\fP ResourceDescriptor;
.FE
.)b
The service calls we provide might have the following form.
.sp 10p
.ip \(sq
\fHGetResource(request).\fP
The request is of type Resources, with only the Kind field set.
If a process wants, for example, three tapes and a printer, it makes a
request with a Resources list containing four resource descriptors.
When this call returns to the calling process (perhaps after a delay),
the Request list has been updated by the resource manager to indicate which
resources have been granted.
.ip \(sq
\fHReleaseResource(request).\fP
Again, request is of type Resources.
The resource manager reclaims the resources listed.
.sp 10p
.pp
The standard scenario for a process that wishes to use some resources is as
follows:
.(b L
.FS
	GetResource(Request);  { may block }
	Use(resource);    { must terminate eventually }
	ReleaseResource(Request);
.FE
.)b
.pp
If process \fBA\fP requests three tapes and a printer
and the installation has five tapes and five printers,
the resource manager will choose from these to give to \fBA\fP.
When \fBA\fP later decides to write information to the tape, it might
submit a service call that looks like this:
.(b L
.FS
	Write(tape,info);
.FE
.)b
where
.q "info"
is the data to be written, and
.q "tape"
is of type
ResourceDescriptor
and describes which tape drive to use.
The
Write
call will fail if the tape has not been allocated to process \fBA\fP.
(We will discuss trans\%put service calls in Chapter 5.)\ \&
Likewise, ReleaseResource will fail if \fBA\fP tries to release a resource that
has not been allocated.
.pp
If a process terminates before it has released all its resources, the resource
manager pretends that it has released them properly.
This pretense is not as bad as preemption, because the terminating process
cannot be hurt by taking away its resources.
However, the resources may not be in a consistent state.
The resource manager could just trust processes to leave resources in the right
condition.
For example, a tape drive should be rewound.
It is better for the resource manager to handle this \fBfinalization\fP part
.IN Device "Finalization"
of the release operation itself, whether release was performed explicitly by the
process or implicitly by the termination of the process.
But the internal consistency of some resources, like files, cannot
be enforced by the resource manager because it does not know how the file is
interpreted by the processes that use it.
The concept of \fBatomic transactions\fP,
.IN "Atomic transactions"
discussed in Chapter 6, can be applied to
achieve proper finalization even in the case of files.
.SH 1 "Deadlock and starvation" I
.lp
At first glance, resource allocation seems trivial.
It seems that we don't need any policy at all.
If the resource is available, grant it;
if not, block the process until the resource is released by whoever has
it.
This approach is the most \fBliberal\fP method imaginable.
.IN "Resource allocation" "Liberal"
.pp
The problem with this approach is that it can get stuck.
The simplest scenario that fails has two processes, \fBA\fP and \fBB\fP.
Both \fBA\fP and \fBB\fP will eventually want two tape drives, and the
operating system has exactly two drives.
However, since neither \fBA\fP nor \fBB\fP needs both tape drives to start out
with, they ask for the drives one at a time, even though they could have asked
for both at once.
Here is the scenario that runs into trouble:
.(b
.FS
\fWA\fP requests one tape drive.
The resource manager grants tape drive 1 to \fWA\fP.
\fWB\fP requests one tape drive.
The resource manager grants tape drive 2 to \fWB\fP.
\fWA\fP requests another tape drive.
The resource manager blocks \fWA\fP, since no drives are available.
\fWB\fP requests another tape drive.
The resource manager blocks \fWB\fP, since no drives are available.
.FE
.)b
Both \fBA\fP and \fBB\fP are stuck:
Each is waiting for a resource that is held by the other.
This sort of circular waiting is called \fBdeadlock\fP.
In Europe, it is known by the more striking name
.q "deadly embrace." " "
.pp
Figure \n(ch.\n+(FN is a \fBresource graph\fP that shows processes and
resources.
.IN "Resource graph"
.(z
.GS
height 1.5
file figs/fig.4.deadlock
.GE
.nr F1 \n(FN \" needed later
.FI "A deadlocked resource graph"
.)z
We draw an arrow from a process (like \fBA\fP) to a resource it is waiting for
(like tape 2).
Similarly, we draw an arrow from a resource (like tape 2) to a process that
currently is using it, if any (like \fBB\fP).
We see that this resource graph has a cycle.
In general, deadlock is a situation in which the resource graph has a cycle.
(We will return to resource graphs later and generalize this principle.)
.pp
The only way to break a deadlock is by preempting a resource, which, as we
have seen, can injure both the process and the resource.
It is far better to avoid deadlock in the first place,
as a less liberal policy might:
.(b
.FS
 1	\fWA\fP requests one tape drive.
 2	The resource manager grants tape drive 1 to \fWA\fP.
 3	\fWB\fP requests one tape drive.
 4	The resource manager blocks \fWB\fP, applying some clever policy.
 5	\fWA\fP requests another tape drive.
 6	The resource manager grants tape drive 2 to \fWA\fP.
 7	\fWA\fP returns tape drive 1.
 8	The resource manager grants tape drive 1 to \fWB\fP.
 9	\fWB\fP requests one tape drive.
10	The resource manager blocks \fWB\fP, since no drives are available.
11	\fWA\fP returns tape drive 2.
12	The resource manager grants tape drive 2 to \fWB\fP.
13	\fWB\fP returns tape drive 2.
14	\fWB\fP returns tape drive 1.
.FE
.)b
Figure \n(ch.\n+(FN shows the resource graph at three points in this scenario.
.(z
.GS
height 1.5
file figs/fig.4.avoid
.GE
.FI "Avoiding deadlock"
.)z
After line 6, \fBB\fP is waiting for either tape drive, so we have shown two
arrows from \fBB\fP. 
However, there is no cycle in the graph.
At the end, both \fBA\fP and \fBB\fP can successfully terminate.  
.pp
Deadlock can exist among more than two processes.
For example, we could accidentally get into the following state:
.(b L
.TS
box, center;
cb | cb cb
cb | c c .
.sp 3p
Process	Has	Wants
.sp 3p
_
A	tape	printer
B	printer	plotter
C	plotter	tape
.sp 3p
.TE
.)b
This situation is shown in the resource graph of Figure \n(ch.\n+(FN.
.(z
.GS
height 2.25
file figs/fig.4.deadlock.3
.GE
.FI "Three-way deadlock"
.)z
The essence of a deadlock is the existence of a cycle of processes, each of
which is waiting for something held by the next process in the cycle.
Our first example had a cycle of two processes.
This second example has three.
.pp
Deadlock is an obvious problem when it happens; all the processes in the
deadlock stop making any progress, even though other processes might continue
unaffected.
The way to prevent deadlock is for the resource manager sometimes to deny a
resource when a process requests it and to block the process instead, even
though the resource is available.
Policies that deny resources even when they are available are
\fBconservative\fP, as opposed to the liberal policy shown earlier.
.IN "Resource allocation" Conservative
.pp
Another way to picture deadlock is to draw a \fBprogress diagram\fP such as the
.IN "Progress diagram"
one shown in Figure \n(ch.\n+(FN.
.(z
.GS
height 3
file figs/fig.4.progress
.GE
.FI "Progress diagram"
.)z
In this diagram, the two dimensions refer to virtual time
.IN "Virtual" time
in processes \fBA\fP and \fBB\fP.
Process \fBA\fP requires a tape drive during part of its execution and a
printer during another part.
Likewise, process \fBB\fP requires both tape drive and printer.
Let us assume that there is only one tape drive and one printer.
The dashed path through the diagram shows how the processes might be scheduled.
When process \fBA\fP is running, the path heads to the right. 
When \fBB\fP is running, the path heads upwards.
If we have only one processor, there are no other possible directions for the
path.
The shaded regions are forbidden because of resource conflicts.
The region on the left must be skirted, because there is only one tape,
and both processes need it in that region.
The region on the right may not be entered because of conflict for the only
printer.
.pp
The dots in the diagram are placed at important times in the execution of these
processes.
At point 1, process \fBB\fP acquires the printer.
At point 2, it also gets the tape.
It releases the printer at point 3 and the tape at point 4.
Process \fBA\fP gets the tape at point 5 and the printer at point 6. 
It releases them at points 7 and 8.
.pp
This picture shows a safe path through the obstacle course.
There is a dangerous turn, however, between points 1 and 2.
If the path had continued to the right instead of heading upward, it would not
immediately enter a forbidden region.
However, since the path may never head down or left, it would have inevitably
reached an obstacle that could not be surmounted.
Virtual time would have to stop for both processes.
They would be deadlocked.
To avoid deadlock, the path must avoid the unsafe (but otherwise permissible)
region in which \fBA\fP has a tape and \fBB\fP has a printer.  
We will see how various methods achieve this goal.
.pp
The most conservative approach of all is called \fBserialization\fP,
.IN "Serialization"
in which processes are not allowed to have resources concurrently at all.
Serialization in our two-tape example would result in the following
scenario:
.(b
.FS
 1	\fWA\fP requests one tape drive.
 2	The resource manager grants tape drive 1 to \fWA\fP.
 3	\fWB\fP requests one tape drive.
 4	The resource manager blocks \fWB\fP, since \fWA\fP has a resource.
 5	\fWA\fP requests another tape drive.
 6	The resource manager grants tape drive 2 to \fWA\fP.
 7	\fWA\fP returns tape drive 1.
 8	\fWA\fP returns tape drive 2.
 9	The resource manager grants tape drive 1 to \fWB\fP.
10	\fWB\fP requests one tape drive.
11	The resource manager grants tape drive 2 to \fWB\fP.
12	\fWB\fP returns tape drive 2.
13	\fWB\fP returns tape drive 1. 
.FE
.)b
Serialization works because all waiting processes are waiting for the same
running process to release resources.
The running process never waits for resources, so there can be no cycle.
We would like to avoid serialization when we can because reduced concurrency
means reduced throughput.
The process that has resources and is running might be blocked for a while in a
trans\%put-wait list.
.IN "Transput-wait list"
During that time no other processes would run, because they are
all in a resource-wait list.
.IN "Resource-wait list"
.pp
Even if deadlock is avoided, a more subtle problem remains.
If a process is blocked even when resources are available, it might never be
allowed to run again.
We call this danger \fBstarvation\fP.
.IN "Starvation"
Starvation arises from consistently awarding resources to the competitors
of a blocked process.
As long as one competitor has the resources, the blocked process cannot
continue.
Increased conservatism is needed to prevent this problem.
.pp
The spectrum from liberal to conservative is shown in Figure \n(ch.\n+(FN.
.(z
.GS
height 1.5
file figs/fig.4.spectrum
.GE
.FI "The liberal-conservative spectrum"
.IN liberal
.IN conservative
.)z
Policies that are too liberal are subject to deadlock.
Those that are too conservative reduce concurrency.
Our goal is to achieve the most liberal resource-allocation policy we can
without encountering deadlock.
Serialization avoids deadlock but is very conservative.
Figure \n(ch.\n(FN also shows the resource-allocation policies we will examine.
The one-shot, hierarchical, and advance-claim algorithms put increasingly less
onerous restrictions on the processes that wish to request resources.
The advance-claim algorithm also requires a certain
amount of prior knowledge about the resources required by each process.
The more information of this nature we know, the closer we can approach our
goal.
However, the algorithms that make use of such knowledge are increasingly
expensive to execute.
.pp
Before we turn to these other algorithms, we will introduce a
particularly instructive example of resource allocation.
.SH 1 "Dining philosophers" I
.lp
The classical
.q "Dining philosophers"
.IN "Dining philosophers"
problem demonstrates both deadlock and starvation issues.
.(q
.ce
\fBDining philosophers problem\fP

Five philosophers are seated around a circular table.
Between each pair of adjacent philosophers is a chopstick,
so there are five chopsticks all together.
In front of each philosopher is a bowl of rice.
It takes two chopsticks to eat rice, so a philosopher needs to grab both the
chopstick to the left and the one to the right in order to start eating.
While philosopher @n@ is eating, neither neighbor, that is, neither
philosopher @n + 1@ nor @n - 1@, can be eating,
because both need a chopstick that is in use by philosopher @n@.
Each philosopher thinks for a while, then decides to eat, gets the chopsticks
needed, eats, and puts the chopsticks down again, in an endless cycle.
The challenge is to grant requests for chopsticks while avoiding
deadlock and starvation.
.)q
.pp
Figure \n(ch.\n+(FN shows each philosopher as a process and each chopstick as
a resource.
.(z
.GS
height 2.5
file figs/fig.4.dining
.GE
.FI "Dining philosophers"
.)z
Each philosopher executes the following program.
.(b
.FS
 1	\fWprocedure\fP Philosopher(Which : 0..4);
 2	\fWvar\fP
 3		LeftChopstick, RightChopstick : ResourceDescriptor;
 4	\fWbegin\fP
 5		{ initialize Chopstick resource variables }
 6		LeftChopstick.Kind := Chopstick(Which);
 7		RightChopstick.Kind := Chopstick((Which+1) \fWmod\fP 5);
 8		\fWloop\fP { think \(em eat loop }
 9			Think; { occupy some finite amount of time in thought. }
10			GetResource(LeftChopstick);
11			GetResource(RightChopstick);
12			Eat;   { for a finite amount of time }
13			ReleaseResource(LeftChopstick);
14			ReleaseResource(RightChopstick);
15		\fWend\fP; { think \(em eat loop }
16	\fWend\fP Philosopher;
.FE
.)b
Deadlock can occur if every philosopher tries to get chopsticks at once and each
gets as far as line 11.
Each has asked for a left chopstick, and the allocator might grant all these
requests.
Now every philosopher is stuck at line 11, because each right chopstick is someone
else's left chopstick; thus each is in use.
Figure \n(ch.\n+(FN shows the resource graph in this situation.
.(z
.GS
height 2.5
file figs/fig.4.phil.dead
.GE
.FI "Deadlocked philosophers"
.)z
A clever allocation policy might have prevented this disaster by keeping one
chopstick free, which would have been more conservative.
.pp
The dining philosophers are also subject to starvation.
Let's say that philosophers 2 and 4 are currently eating.
Philosophers 1 and 5 are engaged in thinking great thoughts and will not be
eating for a long time.
Now philosopher 3 requests both chopsticks (one at a time or both together).
Philosopher 3 is blocked, since the chopsticks are held by philosopher 2 and
philosopher 4.
Figure \n(ch.\n+(FN shows the resource graph.
.(z
.GS
height 2.5
file figs/fig.4.phil.starve
.GE
.FI "Philosopher 3 starves"
.)z
Later, philosopher 2 finishes eating.
Instead of granting 2's right chopstick as 3's left chopstick, which we have seen can
lead to deadlock, let's let the resource manager be more conservative and
leave that chopstick free.
Later, philosopher 2 wants to eat again.
Since philosopher 1 is engrossed in thought and
oblivious to the real world, the resource manager gives
philosopher 2 two chopsticks, leading again to the situation in Figure
\n(ch.\n(FN.
Then philosopher 4 finishes eating.
Again the resource manager decides not to grant philosopher 4's left chopstick
to philosopher 3, because deadlock could result.
Philosophers 2 and 4 keep eating whenever they want, but poor philosopher 3,
although no doubt quite clever, starves to death.
.pp
The problem, then, is to follow a strategy that is not so liberal that it
leads to deadlock or starvation, but not so conservative that it serializes,
because serialization is a waste of resources.
.SH 1 "One-shot allocation" I
.lp
To prevent deadlock, we only need to ensure that a cycle of waiting can never
build up.
One way to avoid a cycle is to partition all processes into two kinds: those
that have all the resources they will ever need and those that have none at
all.
Processes in the first group are not blocked.
Processes in the second group are blocked waiting for resources held by the
first group.
Therefore, no cycle of waiters can exist.
This line of reasoning leads to \fBone-shot allocation\fP, which requires each
.IN "Resource allocation" "One-shot"
process to acquire all its resources at once.
The process is blocked until the request can be satisfied.
Any attempt to request resources after the first time is an error.
.pp
In its pure form, the one-shot method requires each process to acquire all its
needed resources at one time and never request resources again.
Under this rule, each philosopher would request both chopsticks at the outset and
never return them.
Only two philosophers would ever eat; the other three would starve.
.pp
A more liberal form of one-shot allocation follows this rule:
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
One-shot allocation
.sp 0p
_
.sp 3p
A process may request resources only when it has none at all.
.sp 3p
.TE
.)b
A process may repeatedly acquire a number of resources,
use them, return them all, and then later acquire a new set.
To follow this policy, the philosophers should ask for both chopsticks at once.
A resource manager that employs the one-shot policy should be careful to block
any philosopher that cannot get both chopsticks at once.
It must never grant requests partially (like allocating one of the two
chopsticks requested by a philosopher), because such an allocation is
equivalent to letting the philosopher make two requests.
.pp
Now if all the philosophers ask for chopsticks at the same time, the resource
manager will grant one philosopher, say number 1, two chopsticks.
Philosophers 2 and 5 cannot be satisfied at the moment, so they are given no
chopsticks at all.
Philosophers 3 and 4 are still candidates for resources; say the resource
manager gives 3 two chopsticks.
Now 4 must also be blocked.
There is no deadlock, since there is no cycle in the graph, as shown in
Figure \n(ch.\n+(FN.
.(z
.GS
height 2.5
file figs/fig.4.oneshot
.GE
.FI "One-shot allocation"
.)z
This allocation strategy has prevented deadlock, but it does not address
starvation.
.pp
Our earlier situation in which processes \fBA\fP and \fBB\fP each need two
tapes is also solved, since the one-shot method will enforce serialization.
.pp
However, one-shot allocation is too conservative to be practical.
A process that will eventually need many resources will have to acquire them
all at once and thus prevent other processes from using the resources, even
though the process that holds them is not yet using them.
Furthermore, a process may not know all the resources it will eventually need.
.SH 1 "Hierarchical allocation" I
.lp
Another way to prevent a cycle is to associate some number with each process and
somehow make sure that numbers only increase as we follow any chain of
waiting processes.
There clearly can't be a cycle of processes, each of which has a higher number
than the previous one.
.pp
That is the theory, but how can it be made practical?
First, give each resource class a
.q "level"
number.
For example, tape drives can have level 1, plotters 2, and printers 3.
The numbers need not all be different.
We will associate each process with the number of the highest level resource it
currently holds.
To make sure the numbers increase on any chain of waiting processes,
we impose the following rule.
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Hierarchical allocation
.sp 3p
_
.sp 3p
A process may request resources only at a higher level
than any resource it currently holds.
.sp 3p
.TE
.)b
For example, if process \fBA\fP has a tape drive and a plotter, it can ask for
two printers, but it may not request another tape drive or another three
plotters.
A process that only has a tape drive may ask for both plotters and printers
but not for another tape drive.
This rule is not so severe as the one-shot rule.
.pp
The hierarchical allocation rule solves the deadlock problem.
To see why, let's say that there is a deadlock.
We can show that some process must have broken the rules.
Pick any member of the deadlock and call it process \fBA\fP.
It must be blocked while waiting for some resource.
That resource is at some level, say \fIa\fP.
Some process \fBB\fP has that resource and is waiting for another resource.
Since \fBB\fP already has a resource at level \fIa\fP, the one it is waiting for
must be at a higher level, say \fIb\fP.
This argument continues around the cycle
until we find that process \fBA\fP must be holding a
resource at level \fIz\fP , where @z ~>~ a@.
But then \fBA\fP has broken the rules, because it is asking for a resource at
level \fIa\fP.
If \fBA\fP hasn't broken the rules, then somewhere else along the cycle a
mistake was made.
To enforce a hierarchical allocation, the resource manager should return an
error indication instead of granting a resource when a process breaks the rule.
.pp
In our scenario of Figure 4.\n(F1, neither process has followed the hierarchical
rule, since both have requested a tape drive after they already have one.
Each should request both drives simultaneously.
If they had followed the rule, they would not have encountered a deadlock.
If we wish, we can distinguish the two tape drives and put them at different
levels.
Then both \fBA\fP and \fBB\fP must ask for the lower-level tape drive before
the higher-level one.
The second process to ask for its lower-level tape drive will be blocked.
Again, deadlock is avoided.
.pp
The hierarchical strategy also protects the dining philosophers from deadlock.
If all chopsticks are at the same level, the philosophers must ask for
both chopsticks at once, which we have already seen prevents deadlocks.
If each chopstick is at a different level, each philosopher must always ask for
the chopsticks in the proper order.
It cannot happen that the proper order for every philosopher is to ask first
for the left chopstick and then for the right chopstick.
.pp
Unfortunately, the hierarchical strategy is still fairly conservative.
Processes must ask for low-level resources long before they are really needed
because after higher-level resources have been acquired, it is no longer
possible to get low-level resources without giving the high-level ones back.
Since resources are acquired in this artificial order, some processes are
blocked for resources they don't need yet but must acquire now.
Other processes are blocked for resources they need but are being held by
another process that does not need them yet.
In the extreme case in which all resources are at the same level,
hierarchical allocation reduces to the one-shot technique.
.pp
When hierarchical allocation is used,
the scarcest resources are usually given the highest levels.
Requests for them will therefore be made only when they are actually needed by
a process.
.SH 1 "Advance-claim algorithm" I
.lp
The advance-claim algorithm moves us closer to our goal of liberalism without
deadlock by requiring extra information in advance.
Consider a banker with an initial amount of cash with which to grant loans
and customers who wish to take loans and later repay them.
Customers often request additional loans before they can pay back anything.
The banker must be conservative enough to avoid deadlocks among the customers
but wishes to be liberal enough to make as many loans as possible.
.pp
We will require that each process make a
\fBclaim\fP before it acquires any resources.
.IN "Advance-claim algorithm" "Claim"
This claim indicates the greatest number of each resource class it will need
at any one time.
This claim allows the resource manager to foresee and prevent dangerous
situations.
The claim can be presented as a service call:
.sp 10p
.ip \(sq
\fHClaim(resources).\fP
This call returns an error if the process already holds resources.
It sets an upper limit on the number of resources the process may later
use at one time.
The resources argument may list several classes of resources.
.sp 10p
.pp
A process is constrained by its claim; if it ever makes a request that, if
satisfied, would cause the process to hold more of some resource class than
its initial claim, the request is rejected.
The sum of all the claims made by all the processes can be much larger than the
total amount of resources available.
However, no one process may claim more than the total number of
physical resources because there would be a danger that this process could
then legally request more than can be granted, and the process would not be
able to continue.
.pp
The resource manager records the current \fBallocation state\fP for each
.IN "Allocation state"
resource class.
For simplicity, we will assume that there is only one resource class, like tape
drives.
The current allocation state can be declared as follows:
.(b
.FS
 1	\fWtype\fP
 2		ProcessResourceState =
 3			\fWrecord\fP
 4				Claim : integer;
 5				Held : integer;
 6				Request : integer; { if not zero, the process
 7					is blocked waiting for this many more }
 8			\fWend\fP;
 9		AllocationState =
10			\fWrecord\fP
11				\fWarray\fP 1:NumberOfProcesses \fWof\fP ProcessResourceState;
12				Free : integer; { unallocated resources }
13			\fWend\fP;
.FE
.)b
An allocation state is called \fBrealizable\fP if it makes sense \(em that is,
if the following conditions prevail:
.IN "Allocation state" "realizable"
.np
No one claim is for more than the total resources available.
.np
No process is holding more than its claim.
.np
The sum of all the Held fields is not more than the total resources available.
.lp
Otherwise, the state is called \fBunrealizable\fP.
.IN "Allocation state" "Unrealizable"
The resource manager ensures that unrealizable states are never entered.
Each claim is checked against the total resources, and each request is checked
against the associated claim.
.pp
A realizable state is safe if it cannot possibly lead to a deadlock.
Formally, a state is \fBsafe\fP if
.IN "Allocation state" "Safe"
there is a sequence of processes, called a \fBsafe sequence\fP,
.IN "Safe sequence"
such that these three conditions are met:
.np
The first process in the sequence can certainly finish, because
even if it should request all the resources its claim allows, there are enough
free resources to satisfy this request.
.np
The second process in the sequence can finish if the first finishes and
releases everything it now has, because adding what the first has to
the free resources can satisfy the entire claim of the second process.
.np
The \fIi\fPth process can finish if all the previous processes do,
because the sum of all their resources and the currently free resources can
satisfy the entire claim of the \fIi\fPth process. 
.lp
Informally, the state is safe because at the worst,
we can avoid deadlock by blocking all new requests except
those made by the first process.
When it finishes we will allow only the second to continue, and so forth.
One hopes, of course, that such serialization will not be necessary.
.pp
Here is an example.
.(b L
.TS
box, center;
cb | cb cb
cb | n n .
.sp 3p
Process	 Holding	Claims
.sp 3p
_
A	4	6
B	2	7
C	4	11
.sp 3p
_
.T&
cb n .
.sp 3p
Unallocated	2
.sp 3p
.TE
.)b
\fBA\fP can certainly finish, because the 2 unallocated resources can be
granted to it if necessary to bring its holdings to the 6 that it claims.
When it finishes, it will release all 6 resources.
We could then certainly finish \fBB\fP by applying 5 of those resources to
bring its holdings up to its claim of 7.
When it terminates, it will release all 7, giving us 8 unallocated resources.
Those resources suffice to finish \fBC\fP. 
We have shown that {\fBA\fP, \fBB\fP, \fBC\fP} is a safe sequence, so the
current state is safe.
However, a very similar allocation state is unsafe:
.(b L
.TS
box, center;
cb | cb cb
cb | n n .
.sp 3p
Process	Holding	Claims
.sp 3p
_
.sp 3p
A	4	6
B	2	9
C	4	11
.sp 3p
_
.sp 3p
.T&
cb n .
Unallocated	2
.sp 3p
.TE
.)b
\fBA\fP could finish as before, but we cannot guarantee that either \fBB\fP
or \fBC\fP would finish because after \fBA\fP's termination, there are only
6 unallocated resources.
We might be lucky, though, and \fBB\fP might finish without requesting its
full claim.
So an unsafe state does not necessarily have a deadlock.
.pp
Having established these concepts, we can define the advance-claim algorithm:
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Advance-claim algorithm
.sp 0p
_
.sp 3p
Never allocate a request if it causes the current
allocation state to become unsafe.
.sp 3p
.TE
.)b
This algorithm is often called the \fBbanker's algorithm\fP.
.IN "Banker\'s algorithm"
.pp
At first, this criterion appears expensive to evaluate.
Whenever a process makes a new request, the resource manager must find a safe
sequence under the assumption that the request has been granted.
There are @n !@
(@n@ factorial, or
@ n ~\(mu~ ( n ~ - ~ 1 ) ~\(mu~ ( n ~ - ~ 2 ) ~\(mu~ ... ~\(mu~ 2 ~\(mu~ 1@)
sequences of @n@ processes, all of which would be tried before the state is
pronounced unsafe.
However, this cost can be reduced to work proportional to @n sup 2@.
If the state is safe and if some process \fBA\fP can finish given the
currently available resources, there is a safe sequence that starts with
\fBA\fP.
We do not miss the only solutions by starting with a reasonable-looking first
element.
The @{roman O} (n sup 2 )@ algorithm (the
.q "O"
means
.q "order of" )
looks roughly like this:
.(b
.FS
 1	S := [all the processes];
 2	\fWwhile\fP S <> [] \fWdo\fP 
 3		find \fWA\fP, an element of S that can finish
 4		if impossible, fail:  the state is unsafe
 5		remove \fWA\fP from S, add \fWA\fP's resources to the unallocated pool
 6	\fWend\fP;
 7	succeed: the state is safe.
.FE
.)b
The resource manager need not allocate resources to requesting
processes according to the order of the safe sequence.
It may allocate resources in the order of requests as long as there exists some
safe sequence after each request is satisfied.
.pp
The advance-claim algorithm can be applied with a still cheaper algorithm
developed by Habermann.
The resource manager maintains an array S[0..@r^-^1@] of integers,
where @r@ is the number of copies of the resource.
For simplicity, we assume that there is only one resource class.
Initially,  S@[ i ] ~=~ r~-~i@ for all @0 ~<=~ i ~<~ r@.
If a process with a claim of @c@ holds @h@ units and requests 1 more,
decrement S[@i@] for all @0 ~<=~ i~ <~ c-h@.
If any S[@i@] becomes negative, the new state is unsafe.
.pp
As an example, we return to the safe state shown above.
To find the current values of S, we will assume that \fBA\fP, \fBB\fP, and
\fBC\fP each started with no resources, then asked for their current holdings
one by one.
We will mark each decrement by \(mu.
.(b L
.TS
box, center;
cb | n n n n n n n n n n n n .
.sp 3p
Index	0	1	2	3	4	5	6	7	8	9	10	11
.sp 3p
_
.sp 3p
Original S	12	11	10	9	8	7	6	5	4	3	2	1
.sp 3p
_
A	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu
A	\(mu	\(mu	\(mu	\(mu	\(mu
A	\(mu	\(mu	\(mu	\(mu
A	\(mu	\(mu	\(mu
.sp 3p
_
.sp 3p
B	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu
B	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu
.sp 3p
_
.sp 3p
C	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu
C	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu
C	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu
C	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu	\(mu
.sp 3p
=
.sp 3p
Final S	2	1	0	0	0	0	1	1	1	1	1	1
.sp 3p
.TE
.)b
The current state is safe.
If we grant \fBA\fP one more resource, we would decrement S[0] and S[1], which
does not make those entries negative.
Therefore it would be safe to grant \fBA\fP another resource.
But granting \fBB\fP one more resource would decrement S[0]
through S[4], three of which would become negative.
That state would be unsafe.
.pp
To see why this strange method works, let us see what the first
column, S[0], represents.
Before any resources are granted, S[0] is equal to @r@, the number of
resources.
Every time a resource is granted, the process that requested it must have had a
claim higher than the number of resources granted so far, so \(mu will be
placed in at least the S[0] column, if not in others.
Therefore, the current value of S[0] is the number of available resources,
since we subtract one for each \(mu in the column.
If S[0] < 0, the state is not only unsafe,
it is also unrealizable, since more resources have been granted than exist.
.pp
Next, let us see what happens if a process, say \fBD\fP, claims the entire
set of resources and then acquires them all.
We can see how S is calculated for r = 4:
.(b L
.TS
box, center;
cb | n n n n n .
.sp 3p
Index	0	1	2	3
.sp 3p
_
.sp 3p
Original S	4	3	2	1
.sp 3p
_
D	\(mu	\(mu	\(mu	\(mu
D	\(mu	\(mu	\(mu
D	\(mu	\(mu
D	\(mu
.sp 3p
=
.sp 3p
Final S	0	0	0	0
.sp 3p
.TE
.)b
A process that acquires its entire claim makes a triangular
.q "footprint"
of \(mu marks in our picture.
In the case in which the initial claim is the entire set of resources, the
footprint is large enough to reduce every entry in S to 0.
When a process has received its entire claim, the footprint is complete and
makes a triangle.
.pp
Now let us say that one or more entries of S have become \(mi1.
For concreteness, let us say that S[3] is the leftmost column with an entry of
\(mi1.
What is the largest possible value for S[0], S[1], and S[2]?
S[2] cannot be greater than 0 \(em that is, one larger than S[3] \(em
since every \(mu in column 3 is in a line that also puts \(mu in column 2,
and column 2 starts only 1 larger than column 3.
S[2] cannot be negative, since column 3 was the first with a negative value.
So S[2] must be 0.
Similar reasoning shows that S[1] is either 0 or 1, and S[0] is 0, 1, or 2.
.pp
Since S[2] = 0 and S[3] = \(mi1, there cannot be any line of exactly three
\(mu marks.
Such a line would have \(mu in column 2 but not in column 3 and therefore
would have decreased S[2] more than S[3], so S[2] would have to be less than 0.
Therefore, no footprint has a line with exactly three \(mu marks.
All footprints are either large and incomplete, with their last row having at
least four \(mu marks, or they are small, with their first row having at most
two \(mu marks.
All the small footprints can finish.
They belong to processes that form the start of a safe sequence.
(There are at most two such processes, one with a claim of 2 and only one line
of footprint and one with a claim of 1 and one line of footprint.)\ 
After these processes have terminated, S[0] will be 2 and S[1] will be 1.
Only processes with large incomplete footprints are left.
Not a single one of them can finish, because each has at least three lines of
footprint to complete, so S[0] must be at least 3.
But S[0] is only 2.
Therefore, there is no safe sequence.
This argument does not depend on the fact that we chose column 3 as the first
with a negative value.
.pp
We have been assuming throughout that there is only one resource class.
The situation becomes more complicated if there are several resource classes.
It is not enough to find a safe sequence for each resource class independently;
a sequence must be found that is safe with respect to each resource class.
.pp
The advance-claim algorithm is conservative insofar as it cautiously avoids
entering an unsafe state even if this unsafe state has no deadlock.
However, it achieves the most liberal policy that is guaranteed not to
deadlock.
One cost of this achievement is the requirement that processes enter claims,
which may be unnatural.
For example, a process that needs to modify a number of files needs to exclude
other processes, but it might not know which files must be modified until it has
read a significant amount of data from many files.
A second cost of the advance-claim algorithm is the time and space needed for
finding a safe sequence.
.pp
More complex methods exist that lead still closer to our goal.
They require still more prior information about what the processes will need.
For example, if we know the exact order in which each process will request and
release resources, we can organize a more liberal schedule.
The necessary calculation, however, is quite time-consuming.
.SH 1 "Deadlock detection" I
.lp
The methods described so far can be characterized as policies for
\fBdeadlock prevention\fP.
.IN "Deadlock prevention"
If, however, we are willing to allow deadlocks, we must be able to discover
them when they occur.
We call this strategy \fBdeadlock detection\fP.
A method similar to the advance-claim algorithm can be used to check for a
deadlock-free sequence when the current allocation state is unsafe.
We say that a state is \fBdeadlock-free\fP if
.IN "Allocation state" "Deadlock free"
there is a sequence of processes, called a \fBdeadlock-free sequence\fP,
.IN "Deadlock-free sequence"
such that the following three conditions are met.
.np
The first process in the sequence might finish because
there are enough free resources to fulfill its current outstanding
request (what it has actually asked for) even if there are not enough to
satisfy its entire claim (what it may potentially ask for).
.np
The second process in the sequence might finish if the first finishes and
releases everything it now has, because adding what the first has to
the free resources can satisfy the entire outstanding request
of the second process.
.np
The \fIi\fPth process might finish if all the previous processes do
because the sum of all their resources and the currently free resources can
satisfy the entire outstanding request of the \fIi\fPth process. 
.lp
A deadlock-free state has no current deadlock
because the chain of waiting processes is not cyclic.
However, if it is unsafe, an unlucky set of requests can turn it into a
deadlocked state.
Let us continue the previous unsafe example, adding a column for the
outstanding request, that is, the request that has been made that we have not
yet decided to honor.
(We are either at the point of deciding whether to honor it or we have placed
the requesting process in a resource-wait list.)
We no longer need the Claims
column, since we are not using the advance-claim algorithm.
.(b L
.TS
box, center;
cb | cb cb cb
cb | n n n .
.sp 3p
Process	Holding	Outstanding request
.sp 3p
_
.sp 3p
A	4	2
B	2	6
C	4	7
.sp 3p
.T&
cb n .
_
Unallocated	2
.sp 3p
.TE
.)b
As we saw before, this state could be unsafe (given the claims).
However, {\fBA\fP, \fBB\fP, \fBC\fP} is a deadlock-free sequence, so there is
no deadlock yet.
Again, a slight change can make a big difference.
If \fBB\fP should ask for one more resource, this is what we have:
.(b L
.TS
box, center;
cb | cb cb cb
cb | n n n .
.sp 3p
Process	Holding	Outstanding request
.sp 3p
_
A	4	2
B	2	7
C	4	7
.sp 3p
.T&
cb n .
_
.sp 3p
Unallocated	2
.sp 3p
.TE
.)b
Now \fBA\fP might finish, yielding 6 free resources,
but there is no way to satisfy either \fBB\fP or \fBC\fP, both of which need 7.
Processes \fBB\fP and \fBC\fP are deadlocked, but \fBA\fP is not.
.pp
Deadlock can be detected in such cases by building the resource graph and
looking for cycles.
If one resource class has several equivalent members (like tape
drives, which might be functionally identical), the simple resource graphs
must be improved slightly.
We cluster all resources from one class into the same node of the graph.
An arrow goes from any process waiting for a resource to the class node for
that resource.
If a process is waiting for several resources, we draw an arrow from the
process to each of the appropriate class nodes.
An arrow goes from a class node to any process using one of its resources.
.pp
A deadlock is no longer signaled by a simple cycle.
For example, the generalized resource graph of Figure \n(ch.\n+(FN is an
.IN "Resource graph" generalized
extension of Figure \n(ch.\n+(F1.
.(z
.GS
height 1.5
file figs/fig.4.graph.free
.GE
.FI "A deadlock-free generalized resource graph"
.nr F2 \n(FN \" needed later
.)z
There are three tape drives and three processes.
Each of \fBA\fP, \fBB\fP, and \fBC\fP has one tape drive.
\fBA\fP and \fBB\fP are waiting for a second tape drive.
The same cycle we saw in Figure \n(ch.\n(F1 is present (passing through
\fBA\fP, the tapes, \fBB\fP, and the tapes), but there is no deadlock.
\fBC\fP may well finish its work and release its tape drive.
Then either \fBA\fP or \fBB\fP can get it, finish, and let the other finish as
well.
.pp
In generalized resource graphs, we look for knots, which generalize cycles
in ordinary resource graphs.
A \fBknot\fP
.IN "Knot" "Resource graph"
is a set of vertices (including processes and resources) such
that starting at any vertex of the knot, paths lead to \fIall\fP the
vertices in the knot and to no vertices outside the knot.
In our example, paths leading from \fBA\fP, \fBB\fP, or the tapes lead to
\fBC\fP, but no path at all leads from \fBC\fP.
Therefore, this graph has no knots and no deadlock.
.pp
If we enlarge the example we can build a deadlock, as shown in
Figure \n(ch.\n+(FN.
.(z
.GS
height 2.0
file figs/fig.4.graph.dead
.GE
.FI "A deadlocked resource graph"
.)z
Now \fBC\fP is waiting for a printer, but both printers are allocated.
This graph contains a knot, namely, all the vertices except for \fBD\fP.
All the processes in the knot are deadlocked.
Terminating any one of them will break the deadlock.
For example, if we terminate \fBC\fP, one tape is released.
Either \fBA\fP or \fBB\fP can then proceed.
(Until we allocate that tape to either \fBA\fP or \fBB\fP, we ignore the knot,
because it involves a resource that is not fully allocated.)\ \&
.pp
One way the resource manager can detect a knot, and therefore a deadlock,
is to try to \fBreduce\fP
.IN "Resource graph" "Reducing"
the generalized resource graph.
If it can't reduce the graph completely, there is a knot.
To reduce the graph, we remove any process that is not waiting for any
resources and any process that is waiting only for resource classes that are not
fully allocated.
These processes are not contributing to any deadlock; there is some chance they
will finish and return their resources.
As we remove these processes from the graph, we also remove any arrows pointing
to them from resources in order to indicate that those resources are given back.
This action may turn resources that were fully allocated before into partially
allocated resources, thereby allowing us to remove more processes.
If we can remove all the processes, there is no knot and no deadlock.
.pp
In Figure \n(ch.\n(F2, we would first remove \fBC\fP, since it is not waiting.
Then the tapes resource is no longer fully allocated, so both \fBA\fP and
\fBB\fP can be removed.  
In Figure \n(ch.\n(FN, however, no processes can be removed, so there is a
deadlock.
.pp
The onset of deadlock is always at the time a process makes a request and is
blocked waiting for a resource.
A resource manager that depends on deadlock detection needs to check for
deadlock only at this time.
A reasonable alternative, taken by VMS,
.IN VMS "operating system"
is to start a timer for 10 seconds (or however long is specified by the
deadlock-wait parameter) whenever a request blocks.
If the request is still blocked when the timer expires, an attempt is made to
detect deadlock.
The goal of the attempt is to remove the
blocked process that triggered the detection episode from the resource graph.
If that process can be removed, it is not necessary to continue and remove the
others; there is no deadlock involving that process.
.SH 1 "Deadlock recovery"
.lp
When a deadlock occurs, one of the processes participating in the deadlock
must be terminated for the good of the community.
A policy that chooses a victim process is called a method of \fBdeadlock
recovery\fP.
.IN "Deadlock recovery"
Terminating a process might injure data, as we have seen.
A process currently modifying shared data should be able to warn the
resource manager not to terminate it to recover from deadlock.
If data consistency presents no problems, the terminated victim might be
restarted later.
The designers of database management packages have developed very
sophisticated mechanisms to allow database transactions to be aborted in such
a way that restart is always possible.
We will discuss some of those mechanisms in Chapter 6.
.pp
To select a victim process, the resource manager could take into account the
external priority of the process and how expensive it will be to restart it.
One good measure is the amount of computation the process has already
performed.
The larger this amount, the more expensive it is to restart the process and
get it to its present state.
Also, one would hope that processes that have consumed large amounts of
computation are almost finished.
.SH 1 "Starvation"
.lp
Our discussion has concentrated on deadlock, which is due to overly liberal
policies for allocation of resources.
Starvation, in contrast, is the result of overly liberal policies for
reassigning resources once they are returned.
When resources are freed, they can be applied to any process waiting for them
in the resource-wait list.
.IN "Resource-wait list"
(The short-term scheduler should most likely not switch to the waiting process
but should continue with the releasing process, in accord with the Hysteresis
Principle.)\ \&
.IN "Hysteresis Principle"
Not all policies for applying those resources prevent starvation.
.pp
For example, consider a situation with one resource class.
A straightforward policy is to apply all released resources to the
first-blocked process.
However, the blocked process may be the last in all possible safe sequences
for the current allocation state.
We can't unblock it, or the state becomes unsafe.
We can't apply the resources to other jobs because they were not blocked first.
Hence this policy does not work at all.
.pp
We might allow other small requests to be satisfied meanwhile, in the hope that
the processes we favor in this way will eventually release enough resources to
satisfy the first-blocked process.
More precisely, we might sort the resource-wait list by the order in which
.IN "Resource-wait list"
processes block.
When resources are freed, we scan this list and grant resources only to
processes whose current request can be fully satisfied.
Unfortunately, this \fBfirst-fit\fP policy will starve the first-blocked process
.IN "First-fit" "resource allocation"
if there is a continuous stream of new jobs with smaller requirements.
.pp
One way to modify this strategy is to allow partial allocation.
As many units of the resource are granted to the first process in the list as
may be done safely; the rest may be applied (if safe) to later processes in the
list.
Even this policy fails, because there are situations in which no resources may
safely be granted to the first blocked process, even though others may
continue.
.pp
Another modification is to order the resource-wait list according to some
safe sequence.
(If there are several safe sequences, any one will do.)\ \&
Resources are granted, either partially or fully, starting at the head of this
list.
However, there is no way to guarantee that any particular process will ever
reach the start of this list.
A process may remain near the end of the safe sequence and never be allocated
resources.
.pp
The \fBstarvation detection\fP approach is modeled after deadlock detection.
.IN "Starvation detection"
By the time deadlock is detected, it is too late,
but it is never too late to fix starvation.
Starvation might be signaled by a process remaining on the resource-wait list
for too long, measured in either time units or units of process completions.
.pp
Once starvation has been detected,
new processes may be denied resources until starving processes have completed.
Of course, processes that already have resources must be allowed to get more,
since they may appear earlier on the safe sequence.
This approach is certain to work, but it must be tuned carefully.
If starvation detection is too sensitive, new processes are banned too often,
with the result that there is an overall decrease in throughput.
In other words, the policy is too conservative.
If detection is not sensitive enough, however, processes can starve
for a long time before they are finally allowed to run.
.pp
Luckily, starvation is seldom a problem in actual operating systems, because
non-preemptable, serially reusable resources such as printers and tape drives
tend to be underused, and the existence of idle periods tends to prevent
starvation.
.SH 1 "Perspective"
.lp
Deadlock arises from dynamic resource sharing.
If resources are not shared, there can never be a deadlock.
Deadlocks are especially dangerous in time-critical work.
If the operating system is used to run programs that control factory processes
or pilot aircraft, deadlock is unacceptable.
.pp
The danger of deadlocks is mitigated by several considerations.
First, deadlocks can often be avoided by proper design of the algorithms that
share resources.
For example, different parts of the kernel may share information, like process
context blocks.
Those parts will want exclusive access during the time they modify that
information; such access is like holding a resource.
A hierarchical design inside the kernel is often used to make sure that no
deadlocks arise from using these resources.
We will return to this point in Chapter 8, where we discuss mechanisms for
achieving exclusive access.
.pp
Second, deadlock is an unacceptable situation only if everyone is infinitely
patient.
In interactive environments, the resource manager need not recover from or
.IN "Interactive multiprogramming"
even detect deadlocks because the user will eventually get bored waiting for an
answer and will manually terminate the process.
.pp
The service call we described for requesting resources could be enhanced by
including a \fBtimeout\fP parameter.
.IN "Timeout"
If the resource manager cannot or will not honor the request within the time
specified by this parameter, it unblocks the calling process and gives it
a failure indication.
The process could try again, try a different request, back out of whatever
action it is trying to perform, or terminate.
.pp
Some authors distinguish deadlock prevention from avoidance.
Prevention includes any method that negates one of the following three
conditions that lead to deadlock.
.sp 10p
.ip \(bu
Resources may not be shared.
One can sometimes get the effect of sharing an otherwise non-sharable device
by buffering all output to that device (in parallel) and then serializing the
output of those buffers.
Buffering can therefore be used as a prevention method, but deadlock can still
arise over the use of the buffers themselves.
.ip \(bu
Resources may be requested while others are already being held.
As we have seen, the one-shot method prohibits such action, and is therefore a
prevention method.
The hierarchical method prevents such requests within a level;
by induction, it prevents circular wait across levels as well.
.ip \(bu
Resources may not be preempted.
Under batch multiprogramming,
.IN Batch multiprogramming
some jobs may allow cancellation and restarting.
However, jobs that have made changes to files are unlikely to be restartable.
.sp 10p
.lp
Avoidance, on the other hand, includes methods like the banker's
algorithm that take advantage of advance knowledge (such as maximum claims).
That advance knowledge allows them more freedom to navigate the progress diagram
.IN "Progress diagram"
and still avoid dangerous regions.
.pp
The conservative-liberal metaphor helps in rating various policies that we have
seen.
Liberal policies have a greater potential for parallelism and throughput
because they allow more flexible navigation of the progress diagram, but
they have an equal potential for deadlock and starvation.
The advance-claim algorithm provides a reasonably liberal method that is still
conservative enough to prevent deadlock.
To prevent starvation as well, other conservative steps must be taken, such as
banning new arrivals or concentrating resources on particular processes.
Overly conservative methods like one-shot allocation remove all worries about
deadlock and starvation, but the cost is severe reduction of concurrency.
.pp
Although we have devoted a significant amount of attention to the advance-claim
algorithm, it is mostly interesting in the theoretical, not the practical,
sense.
As we pointed out, many applications are unable to compute reasonable claims
before they start working; they only discover what they will need once they
have made some progress.
Furthermore, resources can disappear suddenly if the hardware malfunctions.
Therefore, most operating systems don't implement the advance-claim algorithm.
It is much more common to grant resources to processes following the most
liberal policy imaginable.
If this policy leads to deadlock, the deadlock is detected (either by the
resource manager or by disgruntled users) and broken by injuring some process.
Hierarchical allocation has also been used successfully, especially within the
kernel to make sure that its modules never become deadlocked.
.SH 1 "Further reading"
.lp
A good mathematical treatment of deadlocks is given in Shaw's textbook on
operating systems.
.[
shaw logical operating systems
.]
The problem of deadlock avoidance was formalized and analyzed several years ago.
.[
Holt deadlock properties
.]
.[
Shoshani Elphick Coffman
.]
.[
Hebalkar resource sharing
.]
.[
Shoshani Coffman avoid deadlocks
.]
.[
Minoura Deadlock revisited
.]
The advance-claim algorithm was first proposed by E. Dijkstra.
.[
dijkstra cooperating sequential genuys
.]
A generalization of the binary safe-unsafe criterion to a measure of the
minimum number of resources needed for safety has been suggested by Madduri and
Finkel.
.[
Madduri Finkel Extension of Banker's Algorithm
.]
Starvation has also been called
.q "effective deadlock"
by Holt,
.[
Holt prevention system deadlocks
.]
who, along with others, has pointed out that partial allocation does not
guarantee freedom from starvation.
.[
Parnas Habermann reply Holt
.]
.[
Rossi Fontao deadlock-free resource allocation
.]
.pp
Starvation control has not been dealt with extensively.
Holt
suggested maintaining periodically incremented counters for each blocked
process.
.[
system deadlocks holt surveys
.]
When a counter exceeds a critical value the scheduler has to find a safe
sequence and finish jobs in that order.
He also suggested partial allocations but required that resources not be granted
to jobs with zero holdings.
The approach of banning new jobs was introduced by Madduri and Finkel.
.[
madduri finkel note starvation
.]
Another approach to starvation,
which allocates resources according to a safe-sequence order,
is found in the Boss 2
.IN "Boss 2 operating system"
operating system for the RC4000 computer.
.[
Lauesen Job scheduling guaranteeing reasonable turn-around
.]
.IN "RC4000 computer"
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
The service calls for allocating and releasing resources look suspiciously like
the calls for allocating and releasing segments.
Is there any fundamental difference between the two situations?
.EP \n+(E+.
The service call for allocation of resources cannot specify
.q "either one resource of class 1 or one of class 2." " "
However, our dining philosophers might want to ask for whichever chopstick is
available first, then ask for the other one specifically.
.br
(a)  Why aren't all the chopsticks in the same resource class?
.br
(b)  What could happen if we allow this sort of
.q "selective resource allocation"
and use a liberal allocation policy?
.br
(c)  How well would selective allocation mesh with the hierarchical allocation
policy?
.br
(d)  How well would selective allocation mesh with the advance-claim algorithm?
.EP \n+(E+.
What is a one-process deadlock?
Why does this situation not cause any problems to the resource manager?
.EP \n+(E+.
Under the hierarchical resource allocation policy,
suppose that a process is holding a resource of level 3 and finds it needs to
acquire a resource of level 2.
Is there any way it can do this?
.EP \n+(E+.
Let's number all the chopsticks in the dining philosophers problem and require
that each philosopher request an even-numbered chop\%stick before an
odd-numbered chop\%stick.
Will this allocation strategy work?
Is it a form of a well-known strategy?
.EP \n+(E+.
Assume that there are two classes of resources.
For the following resource-allocation state, tell which, if any, of
the processes are deadlocked.
.(b
.TS
box, center;
cb | cb s | cb s
cb | cb cb | cb cb
cb | n n | n n.
.sp 3p
Process	1	2	3	4
\^	Holding	Request	Holding	Request	Holding	Request	Holding	Request
.sp 3p
_
.sp 3p
A	3	1	4	5
B	2	3	3	6
C	4	2	2	0
.sp 3p
.T&
cb | c s | c s.
_
.sp 3p
Unallocated	2	3
.sp 3p
.TE
.)b
.nr E9 \n(E+
.EP \n+(E+.
Assume that there are two classes of resources.
For the following resource-allocation state, tell which, if any, of
the processes are deadlocked.
.(b
.TS
box, center;
cb | cb s | cb s
cb | cb cb | cb cb
cb | n n | n n.
.sp 3p
process	1	2	3	4
\^	Holding	Request	Holding	Request	Holding	Request	Holding	Request
.sp 3p
_
.sp 3p
A	3	7	4	5
B	2	3	3	6
C	4	2	2	0
.sp 3p
.T&
cb | c s | c s.
_
.sp 3p
Unallocated	2	3
.sp 3p
.TE
.)b
.nr E8 \n(E+
.EP \n+(E+.
Draw the resource graphs for the situations in exercises \n(E9 and \n(E8.
.EP \n+(E+.
Construct an unsafe resource-allocation state in which there are three
processes, there is one class of resource with four resources, and each
process has a claim of three resources.
.EP \n+(E+.
Construct an unsafe resource-allocation state in which there are three
processes, there is one class of resource with four resources, and each
process has a claim of two resources.
.EP \n+(E+.
Which of the following events can lead from a safe state to an unsafe state?
Which can lead from a deadlock-free state to a deadlocked state?
.br
(a)  A process makes a request within its claim.
.br
(b)  A process makes a request beyond its claim.
.br
(c)  A request within the claim is satisfied.
.br
(d)  A process is blocked on a resource-wait list.
.br
(e)  A process releases a resource it had previously acquired.
.br
(f)  A process increases its claim.
.br
(g)  A process decreases its claim.
.br
(h)  A process terminates.
.br
(i)  A process is created and it makes its claim.
.EP \n+(E+.
We have shown for Habermann's method that if S@[ i ] ~=~ -1@ for some
@i@, there is no safe sequence.
Show that if all S@[ i ] ~>=~ 0@ for all @i@, there is a safe sequence.
.EP \n+(E+.
Show a situation in which there are several resource classes and for each
class there is a safe sequence but the situation is still unsafe.
.EP \n+(E+.
Find a situation in which the first-fit starvation control policy fails.
.EP \n+(E+.
Find a situation in which the first-fit starvation-control policy fails even
when we make partial allocations.
.EP \n+(E+.
Why is testing the length of the resource-wait list an inadequate measure of
starvation?
.EP \n+(E+.
Why don't we measure the length of time on the resource-wait list in virtual
time?
.IN "Virtual time"
.EE
