/* lp_smrnd.c
 * Neil Moore <neil@cs.uky.edu>
 * 30 Mar 1999
 *
 * Part of liblp.  See COPYRIGHT for copyright information.
 */
#include <stdlib.h>
#include <lp.h>
#include <lp_lm.h>
#include <lp_wfs.h>
#include <lp_util.h>
#include <lp_smrnd.h>

static int *try_model(struct lp *p, int *model, int *flips, int flipmode);
static int size_diff(struct lp *p, int *model);


/* Randomly find a stable model of p.  A random model (containing the
 * well-founded model) is selected; then try_model() is called.  This is
 * repeated until a stable model is found, or until the specified number
 * of restarts have been used.  This function returns the model (an
 * array of 0s and 1s; the nth element is true if p->atoms[n] is in the
 * stable model).  *restarts and *flips are changed to indicate the
 * number of restarts and average number of flips per restart made.
 *
 */
int *lp_sm_rand(struct lp *p, int *restarts, int *flips, int flipmode)
{
	int *wfm;
	int *sm=NULL;
	int i;
	int total=0;
	
	wfm = lp_wfs(p);

	for(i=0; i<*restarts; i++) {
		int nflips = *flips;
		sm = try_model(p, wfm, &nflips, flipmode);
		total += nflips;
		if(sm != NULL) {
			i++;
			break;
		}
		free(sm);
	}
	
	*restarts = i;
	*flips = total/i;
	
	return sm;
}


/* Randomly find a stable model of p based on wfm.  A random model
 * containing the well-founded model wfm is formed.  Then the truth value
 * of each element not in the wfm is inverted; the resulting models are
 * compared, and one of those flips resulting in the best model is
 * performed, and the process is repeated (up to flips times). If flipmode
 * is LP_FM_ONCE, no atom will be flipped more than once.  If flipmode
 * is LP_FM_RANDOM, ties are broken by random selection.  If flipmode is
 * LP_FM_LR, ties are broken by choosing the least-recently made of the
 * tied flips.  This function returns an array of 0s and 1s, one for each
 * atom, indicating the stable model found (NULL if no stable model was
 * found).  *flips is set to the number of flips actually made.
 *
 */
int *try_model(struct lp *p, int *wfm, int *flips, int flipmode) 
{
	int i, j;
	int best, elt, diff, lr;
	int k=1;
	
	int *recent=NULL;
	int *try; /* array of bools */
	int *model;

	if(flipmode == LP_FM_ONCE)
		model = lp_copy_array(wfm, p->natoms);
	else
		model = wfm;

	if(flipmode == LP_FM_LR)
		recent = lp_calloc(p->natoms, sizeof(int));
	
	try = lp_malloc(p->natoms * sizeof(int));
	
	for(i=0; i<p->natoms; i++) {
		if(model[i] == LP_L_UNKNOWN)
			try[i] = lp_randtf();
		else if(model[i] == LP_L_TRUE)
			try[i] = 1;
		else
			try[i] = 0;
	}

	for(i=0; i<*flips; i++) {
		best = size_diff(p,try);
		
		if(best == 0) {
			*flips = i;
			if(flipmode == LP_FM_ONCE)
				free(model);
			return try;
		}
		
		k=1;
		lr = p->natoms+1;
		elt = -1;

		/* best is the smallest difference between M and LM(red(P,M))
		 *  for all flips.
		 * elt is the element to flip to give that difference.
		 * lr is the time of the element which gives that flip.
		 *
		 * We wish to choose the element (the least recent such if
		 * flipmode is FM_LR) which gives the best model when flipped.
		 * if there is a tie, we choose randomly.
		 *
		 * k is used for random choice.  We make the random choice
		 * using the algorithm:
		 *
		 *   k := 1;
		 *   for i := first to last
		 *     if(match(i))
		 *       r := random number between 0 and 1.
		 *       if(r < 1/k)
		 *           choose this one.
		 *       k := k+1.
		 *
		 * The source of this algorithm is unknown, but it appears in
		 * the Perl FAQ (in the context of choosing a random line from
		 * a file).
		 *
		 */
		for(j=0; j<p->natoms; j++) {
			if(model[j] != LP_L_UNKNOWN)
				continue;
			
			try[j] = !try[j];
			diff = size_diff(p,try);
			if(diff < best) {
				best = diff;
				elt = j;
				if(flipmode == LP_FM_LR)
					lr = recent[elt];
			}
			
			if(flipmode == LP_FM_LR) {
				if(diff == best && recent[j] < lr) {
					elt = j;
					lr = recent[j];
				}
				if(diff == best && recent[j] == lr && lp_randd() < 1.0/(k++))
					elt = j;
			} else if(diff == best && lp_randd() < 1.0/(k++))
				elt = j;
			
			
			try[j] = !try[j];
		}

		if(elt == -1) {
			i++;
			break;
		}
		
		try[elt] = !try[elt];
		
		if(flipmode == LP_FM_LR)
			recent[elt] = i+1;
		else if(flipmode == LP_FM_ONCE)
			model[elt] = try[elt] ? LP_L_TRUE : LP_L_FALSE;
	}

	if(flipmode == LP_FM_ONCE)
		free(model);
	if(flipmode == LP_FM_LR)
		free(recent);
	free(try);
	*flips = i;
	return NULL;
}


/* Test a model.  This function returns the size of the symmetric difference
 * of model and lm(P^{model}) (the least model of the reduct of p with respect
 * to model).  Smaller values are closer to stable models; zero indicates a
 * stable model.
 *
 */
int size_diff(struct lp *p, int *model) 
{
	struct lp *r;
	int *lmodel;
	int i, diff=0;

	r = lp_reduct(p,model);
	lmodel = lp_lm(r);
	lp_free_lp(r);

	for(i=0; i<p->natoms; i++)
		if(lmodel[i] != model[i])
			diff++;
	
	free(lmodel);
	return diff;
}
