/* simpmain.c
 * Neil Moore <neil@cs.uky.edu>
 * 24 Apr 1999
 *
 * Part of sm2dt.  See COPYRIGHT for copyright information.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <lp.h>
#include <lp_smrd.h>
#include <lp_smpr.h>
#include <lp_util.h>
#include <lp_graph.h>

#ifndef VERSION
#define VERSION "unknown"
#endif

static void uniquify(struct lp *p);
static void handle_reqs(struct lp *p);
static void handle_pos_req(struct lp *p, struct lp_atom *a);
static void handle_neg_req(struct lp *p, struct lp_atom *a);
static void classify_rules(struct lp *p);
static void resize_rules(struct lp *p, int nrules);
static void resize_atoms(struct lp *p, int natoms);
static void usage(void);
static void version(void);

int main(int argc, char *argv[]) 
{
	char *infilename = NULL;
	char *outfilename = NULL;
	FILE *infile = NULL;
	clock_t start, end;
	int quiet=0;
	int len;

	struct lp *prog = NULL;
	
	while (*++argv != NULL) {
		if(!strcmp(*argv, "-o") || !strcmp(*argv, "--output")) {
			++argv;
			if(*argv == NULL) {
				usage();
				return 1;
			}
			outfilename = *argv;
		} else if(!strncmp(*argv, "-o", 2)) {
			outfilename = *argv + 2;
		} else if(!strncmp(*argv, "--output=", 9)) {
			outfilename = *argv + 9;
		} else if(!strcmp(*argv, "-q") || !strcmp(*argv, "--quiet")) {
			quiet = 1;
		} else if(!strcmp(*argv, "-i") || !strcmp(*argv, "--print-info")) {
			quiet = 0;
		} else if(!strcmp(*argv, "-v") || !strcmp(*argv, "--version")) {
			version();
			return 0;
		} else if(!strcmp(*argv, "-h") || !strcmp(*argv, "--help")) {
			usage();
			return 0;
		} else if(!strcmp(*argv, "-")) {
			infilename = NULL;
		} else if ((*argv)[0] == '-') {
			usage();
			return 1;
		} else {
			infilename = *argv;
		}		
	}
	
	if(infilename == NULL) {
		infile = stdin;
		if(outfilename == NULL)
			outfilename = "sm2dt.sm.simp";
	} else {
		if((infile = fopen(infilename, "r")) == NULL) {
			fprintf(stderr, "cannot open %s: %s\n", infilename, strerror(errno));
			goto failure;
		}
	}

	if(outfilename == NULL) {
		len = strlen(infilename);
		outfilename = lp_malloc(len + 6);
		strncpy(outfilename, infilename, len+4);
		strcat(outfilename, ".simp");
	}	

	start=clock();
	prog = lp_malloc(sizeof(struct lp));
	if(lp_sm_read_file(infile, prog) < 0) {
		fprintf(stderr, "error in input: output not generated\n");
		goto failure;
	}

	handle_reqs(prog);
	classify_rules(prog);
	uniquify(prog);
	lp_sm_print_program(prog, outfilename);

	end = clock();
	if(!quiet)
		printf("simplification: %f seconds\n", ((double)end - start)/ CLOCKS_PER_SEC);
	
	if (infile != stdin)
		fclose(infile);
	lp_free_lp(prog);
	return 0;
	
 failure:
	if(infile && infile != stdin)
		fclose(infile);
	if(prog)
		lp_free_lp(prog);
	return 1;
}


void classify_rules(struct lp *p) 
{
	int i, apos;

	/* appear[a]    Meaning for p->atoms[a]
	 * -------------------------------------------------
	 *   -1        Appears other than as a killing atom.
	 *    0        Has not yet appeared.
	 *    1        Has appeared only as a killing atom.
	 */
	int *appear = lp_calloc(p->natoms, sizeof(int));
	
	for(i=0; i<p->nrules; i++) {
		apos = lp_apos(p, p->rules[i].head);
		
		if(appear[apos] == -1)
			continue;
		
		if(lp_is_local_killing(p, &p->rules[i]))
			appear[apos] = 1;
		else
			appear[apos] = -1;
	}

	for(i=0; i<p->nrules; i++) {
		apos = lp_apos(p, p->rules[i].head);

		if(appear[apos] == 1)
			p->rules[i].type = LP_R_KILLING;
		else
			p->rules[i].type = LP_R_NORMAL;
	}
	free(appear);
}


void uniquify(struct lp *p) 
{
	int i, j, curr=0, orignatoms;
	int *nappears;

	nappears = lp_calloc(p->natoms, sizeof(int));

	for(i=0;i<p->nrules; i++)
		if(p->rules[i].type == LP_R_KILLING)
			nappears[lp_apos(p, p->rules[i].head)]++;
	

	orignatoms = p->natoms;
	
	for(i=0; i<p->nrules; i++) {
		if(p->rules[i].type == LP_R_KILLING) {
			/* 1 Apr 1999 */
			if(--nappears[lp_apos(p, p->rules[i].head)] == 0)
				continue;
			
			for(j=0; j<p->rules[i].nneg; j++) {
				if(p->rules[i].neg[j] == p->rules[i].head)
					p->rules[i].neg[j] = &p->atoms[p->natoms + curr];
			}
				p->rules[i].head = &p->atoms[p->natoms + curr++];
		}
	}

	resize_atoms(p, p->natoms + curr);

	for(i=orignatoms; i<p->natoms; i++)
		sprintf(p->atoms[i].name, "Gka%05d", i-orignatoms);

	free(nappears);
	return;
}


void handle_reqs(struct lp *p) 
{
	int i;
	if(p->nposreq > 0) {
		resize_atoms(p, p->natoms+1);
		
		strcpy(p->atoms[p->natoms-1].name, "Gpr00000");
	}
	
	
	for(i=0; i<p->nposreq; i++)
			handle_pos_req(p, p->posreq[i]);
	for(i=0; i<p->nnegreq; i++)
			handle_neg_req(p, p->negreq[i]);

	free(p->posreq);
	p->posreq = 0;
	p->nposreq = 0;
	free(p->negreq);
	p->negreq = 0;
	p->nnegreq = 0;
}


void handle_pos_req(struct lp *p, struct lp_atom *a)
{
	int i, j, k;
	int nrules;
	struct lp_rule *newrule;
	struct lp_rule *last;
	struct lp_rule *r;
	
	nrules = p->nrules;

	/* remove all rules with not(a) in the body.  Rule order is not
	 * preserved.
	 */
	for(i=0; i<nrules; i++) {
		r = &p->rules[i];
		for(j=0; j<r->nneg; j++) {
			if(r->neg[j] == a) {
				p->size -= (r->npos + r->nneg + 1);
				
				/* copy last rule into this one */
				last = &p->rules[p->nrules-1];
				r->head = last->head;
				r->npos = last->npos;
				r->nneg = last->nneg;
				r->type = last->type;
				
				for(k=0; k<last->nneg; k++)
					r->neg[k] = last->neg[k];
				for(k=0; k<last->npos; k++)
					r->pos[k] = last->pos[k];

				/* remove last rule */
				nrules--;
				i--;   /* reproecess changed rule */
				break; /* go to next rule */
			}
		}	
	}

	if(nrules != p->nrules-1)
		resize_rules(p, nrules+1);
	newrule = &p->rules[p->nrules - 1];
	
	newrule->head = &p->atoms[p->natoms - 1];
	newrule->npos = 0;
	newrule->nneg = 2;
	newrule->neg[0] = a;
	/* p->natoms - 1 is the killing atom added by handle_reqs() */
	newrule->neg[1] = &p->atoms[p->natoms - 1];

	p->size += 3;
}


void handle_neg_req(struct lp *p, struct lp_atom *a)
{
	int i, j, k;
	int nrules;
	struct lp_rule *r;
	struct lp_rule *last;
	
	nrules = p->nrules;

	/* remove all rules with a in the body.  Rule order is not
	 * preserved.
	 */
	for(i=0; i<nrules; i++) {
		r = &p->rules[i];
		for(j=0; j<r->npos; j++) {
			if(r->pos[j] == a) {
				p->size -= (r->npos + r->nneg + 1);
				
				/* copy last rule into this one */
				last = &p->rules[p->nrules-1];
				r->head = last->head;
				r->npos = last->npos;
				r->nneg = last->nneg;
				r->type = last->type;
				
				for(k=0; k<last->nneg; k++)
					r->neg[k] = last->neg[k];
				for(k=0; k<last->npos; k++)
					r->pos[k] = last->pos[k];

				/* remove last rule */
				nrules--;
				i--;   /* reproecess changed rule */
				break; /* go to next rule */
			}
		}	
	}

	if(nrules != p->nrules)
		resize_rules(p, nrules);
	
	for(i=0; i<p->nrules; i++) {
		r = &p->rules[i];
		/* The next two steps must be performed in this order */ 
		
		/* Eliminate not(a) from rule bodies */
		for(j=0; j<r->nneg; j++)
			if(r->neg[j] == a)
				/* j-- to reprocess this literal */
				r->neg[j--] = r->neg[--r->nneg];

		/* Add not(a) to the body of rules with head a */
		if(r->head == a)
			r->neg[r->nneg++] = a;
	}
}

void resize_rules(struct lp *p, int nrules)
{
	p->nrules = nrules;
	p->rules = lp_realloc(p->rules, sizeof(struct lp_rule) * nrules);
}


void resize_atoms(struct lp *p, int natoms) 
{
	int i,j;
	int diff;
	struct lp_atom *origbase = p->atoms;
		
	p->natoms = natoms;
	p->atoms = lp_realloc(p->atoms, sizeof(struct lp_atom) * natoms);

	diff = p->atoms - origbase;
	if(diff == 0)
		return;
		
	for(i=0; i<p->nrules; i++) {
		p->rules[i].head += diff;
		for(j=0; j<p->rules[i].nneg; j++)
			p->rules[i].neg[j] += diff;
		
		for(j=0; j<p->rules[i].npos; j++)
			p->rules[i].pos[j] += diff;
	}
	for(i=0; i<p->nposreq; i++)
		p->posreq[i] += diff;
	for(i=0; i<p->nnegreq; i++)
		p->negreq[i] += diff;
}

void usage(void) 
{
	printf("usage: uniquify [options] [filename]\n");
	printf("Make killing atoms in a logic program unique.\n\n");
	printf(" -o, --output=file    output goes to file (default of filename.simp,\n");
	printf("                      or \"sm2dt.sm.simp\" if input is stdin)\n");
	printf(" -i, --print-info     print timing information to stdout (default)\n");
	printf(" -q, --quiet          do not print timing information\n");
	printf(" -v, --version        print version information to stdout\n");
	printf(" -h, --help           display this message\n\n");
	printf("Input (in smodels format) is read from filename (defaults to stdin).\n");
}

void version(void) 
{
	printf("uniquify, from sm2dt " VERSION "\n");
}
