/* strmain.c
 * Neil Moore <neil@cs.uky.edu>
 * 24 Apr 1999
 *
 * Part of sm2dt.  See COPYRIGHT for copyright information.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <lp.h>
#include <lp_lists.h>
#include <lp_smrd.h>
#include <lp_smpr.h>
#include <lp_util.h>
#include <lp_graph.h>

#ifndef VERSION
#define VERSION "unknown"
#endif

static int stratify(struct lp *prog, struct lp_rlist ***strata);
static void usage(void);
static void version(void);

int main(int argc, char *argv[]) 
{
	char *infilename = NULL;
	struct lp_rlist **strata=NULL;
	FILE *infile=NULL;
	int nstrata, len;
	clock_t start, end;
	int quiet=0;

	char *outfilebase = NULL;
	char *outfilename = NULL;
	char *strfilename = NULL;

	struct lp *prog = NULL;

	while (*++argv != NULL) {
		if(!strcmp(*argv, "-o") || !strcmp(*argv, "--output")) {
			++argv;
			if(*argv == NULL) {
				usage();
				return 1;
			}
			outfilebase = *argv;
		} else if(!strncmp(*argv, "-o", 2)) {
			outfilebase = *argv + 2;
		} else if(!strncmp(*argv, "--output=", 9)) {
			outfilebase = *argv + 9;
		} else if(!strcmp(*argv, "-v") || !strcmp(*argv, "--version")) {
			version();
			return 0;
		} else if(!strcmp(*argv, "-h") || !strcmp(*argv, "--help")) {
			usage();
			return 0;
		} else if(!strcmp(*argv, "-q") || !strcmp(*argv, "--quiet")) {
			quiet = 1;
		} else if(!strcmp(*argv, "-i") || !strcmp(*argv, "--print-info")) {
			quiet = 0;
		} else if(!strcmp(*argv, "-")) {
			infilename = NULL;
		} else if ((*argv)[0] == '-') {
			usage();
			goto failure;
		} else {
			infilename = *argv;
		}		
	}
	
	if(infilename == NULL) {
		infile = stdin;
		if(outfilebase == NULL)
			outfilebase = "sm2dt.sm";
	} else {
		if((infile = fopen(infilename, "r")) == NULL) {
			fprintf(stderr, "cannot open %s: %s\n", infilename, strerror(errno));
			goto failure;
		}
	}

	if(outfilebase == NULL)
		outfilebase = infilename;
	
	len = strlen(outfilebase);
	
	outfilename = lp_malloc(sizeof(char) * (len + 5));
	strncpy(outfilename,outfilebase,len+1);
	strcat(outfilename, ".ord");

	strfilename = lp_malloc(sizeof(char) * (len + 5));
	strncpy(strfilename,outfilebase,len+1);
	strcat(strfilename, ".str");

	start=clock();
	prog = lp_malloc(sizeof(struct lp));
	if(lp_sm_read_file(infile, prog) < 0) {
		fprintf(stderr, "error in input: output not generated\n");
		goto failure;
	}
	
	nstrata = stratify(prog, &strata);
	lp_sm_print_program_str(prog, strata, nstrata, outfilename, strfilename);
	
	end = clock();
	if(!quiet)
		printf("stratification: %f seconds\n", ((double)end - start)/ CLOCKS_PER_SEC);
	
	if (infile != stdin)
		fclose(infile);

	lp_free_lp(prog);
	return 0;
	
 failure:
	if(infile && infile != stdin)
		fclose(infile);
	if(prog)
		lp_free_lp(prog);
	return 1;
}

int stratify(struct lp *prog, struct lp_rlist ***strata)
{
	struct lp_vlist **components, *j;
	struct lp_rlist *curr;
	struct lp_graph *graph;
	
	int ncomps, nstrata=0, i;
	
	graph = lp_construct_strat_graph(prog);
	
	ncomps = lp_find_scc(graph, &components);
	*strata = lp_malloc(sizeof(struct rule *) * ncomps);
	
	for(i=0; i<ncomps; i++) {
		(*strata)[i] = NULL;
		
		for(j=components[i]; j!=NULL; j=j->next) {
			/* XXX lp_vl_add */
			if(j->vertex->type != LP_DT_RULE)
				continue;
			curr = lp_malloc(sizeof(struct lp_vtx));
			curr->next = (*strata)[nstrata];
			curr->rule = (struct lp_rule *)j->vertex->data;
			(*strata)[nstrata] = curr;
		}
		if((*strata)[nstrata] != NULL)
			nstrata++;
	}

	for(i=0; i<ncomps; i++)
		lp_vl_free(components[i]);

	*strata = lp_realloc(*strata, sizeof(struct lp_rule *) * nstrata);
	
	free(components);
	lp_free_graph(graph);
	
	return nstrata;
}

void usage(void) 
{
	printf("usage: stratify [options] [filename]\n");
	printf("Compute relaxed stratification of a logic program.\n\n");
	printf(" -o, --output=file    output goes to file.{ord,str} (defaults to input\n");
	printf("                      filename, or \"sm2dt.sm\" if input is stdin)\n");
	printf(" -i, --print-info     print timing information to stdout (default)\n");
	printf(" -q, --quiet          do not print timing information\n");
	printf(" -v, --version        print version information to stdout\n");
	printf(" -h, --help           display this message\n\n");
	printf("Input (in smodels format) is read from filename (defaults to stdin).\n");
}

void version(void) 
{
	printf("stratify, from sm2dt " VERSION "\n");
}
