.SH 1 "Paging" I
.lp
We now turn to the most common storage management method in use.
Paging has the same hardware flavor as segmentation except that the problem of
allocating physical store among different-sized segments is removed by
insisting that each segment-like chunk be exactly the same length.
We will call each such chunk a \fBpage\fP.
The virtual space of a process is subdivided into as many pages as are needed,
depending on its length.
.pp
We saw that segmentation supported software modularity by allowing a program's
name space to be partitioned.
.IN "Name space"
Paging, in constrast, partitions virtual space without regard to logical
entities residing in that space.
Paging is therefore \fBtransparent\fP to the process.
.IN "Transparency"
The fact that virtual space is subdivided into pages does not affect
the way the process is programmed or the way it treats virtual space.
Later, we will examine combinations that
incorporate the semantics of segmentation as well as the regularity of paging.
.pp
Paging is a pleasant contrast to segmentation from the point of view of the
storage-manager implementer.
Shuffling is eliminated because all free areas in physical store
.IN Shuffling
are one page-length long.
A page may be placed in any such free area; there is no need to treat adjacent
free areas together as part of a larger free area.
We will treat physical store as an array of bins into which pages may be
placed; these bins are called
\fBpage frames\fP:
.IN "Page frames"
.(b
.FS
 1	\fWconst\fP
 2		BitsPerVirtualAddress = 18; { for example }
 3		BitsPerPhysicalAddress = 24; { for example }
 4		BitsPerVirtualPageAddress = 8; { for example }
 5		BitsPerOffsetAddress =
 6			BitsPerVirtualAddress \(mi BitsPerVirtualPageAddress;
 7		PageSize = 2\u\s-2BitsPerOffsetAddress\s+2\d\x'-3p';	
 8		NumberOfPageFrames = 2\u\s-2BitsPerPhysicalAddress\s+2\d\x'-3p' / PageSize;
 9			{ maximum allowed by the hardware; we may have fewer }
10		PageLimit = 2\u\s-2BitsPerVirtualAddress\s+2\d\x'-3p' / PageSize;
11			{ number of pages per virtual space }
12	\fWtype\fP
13		PageFrame = \fWarray\fP 0 : PageSize \(mi 1 \fWof\fP byte;
14		PageFrameNumber = 0 .. NumberOfPageFrames \(mi 1;
15		PhysicalStore = \fWarray\fP PageFrameNumber \fWof\fP PageFrame;
.FE
.)b
If we use the constants suggested in lines 2\(mi4, virtual
addresses are 18 bits long.
Virtual space is therefore limited to 256K bytes.
A process may have a virtual space ranging anywhere between 0 bytes and this
limit in increments of one page size.
A virtual page address is 8 bits.
The 256K bytes of virtual store are therefore divided into 256 pages,
each having 1K bytes.
The number of bits in an offset is @18~-~8~=~10@, which fits this page size.
The largest virtual space has 256K\|/\|1K = 256 pages.
A physical address is described by 24 bits.
As a result, the largest amount of
physical store that may be placed on the machine is 16M bytes.
A particular machine may have less store.
Physical store can have as many as 16M\|/\|1K = 16K page frames.
On this particular machine, physical space is much larger than any virtual
space.
This situation is found, for example, on some 16-bit computers like the 
DEC PDP-11.
.IN "DEC PDP-11" "computer"
More typically, virtual spaces are allowed to be larger
than the amount of physical store present.
.pp
Pages fit exactly into page frames in physical store, so there is no
external waste.
.IN "external waste"
However, internal waste is likely because a program might not need
.IN "Internal waste"
a virtual store whose length is exactly divisible by the page size.
On the average, each process will waste one half of the last page.
The choice of page size, which is a policy bound at the time the operating
system is designed, involves a tradeoff among various considerations:
.sp 10p
.np
overhead space \(em least with large pages, since page tables have
fewer entries.
.np
internal waste \(em smallest with small pages, since one-half of a small
page is small.
Figure \n(ch.\n+(FN shows the contrast between small and large pages for the
same virtual space.
.(z
.GS
width 4
file figs/fig.3.wasted
.GE
.FI "Wasted space and page size"
.)z
When pages are small, there is less internal waste, but the page table is
longer.
.np
cold-start faults \(em lower with large pages.
.IN "Cold start" "page faults"
These are the faults needed to bring the program into main store the first
time.
With larger pages, there are fewer total pages, hence fewer cold-start faults.
.np
page-fault rate \(em generally lower with small pages, assuming that a fixed
amount of main store is available.
Regions of heavy use within a virtual space tend to be relatively small, and
the computer can fit more such regions into a fixed-size main store if page
size is small.
.np
efficiency of transfer between main and backing store \(em best with
large pages, since they take about the same amount of time as small pages to
transfer.
The total transfer time is dominated by disk rotational and seek latency, not
transfer latency.
.IN Disk Latencies
(We discuss disk transfers in Chapter 5.)\ \&
.sp 10p
.lp
Larger pages are generally better, unless they become so large that too few
processes in the ready list can have pages in main store at the same time.
As main store has become less expensive, page sizes have increased, until now
they are sometimes as large as 8K bytes,
although sizes from 512 bytes to 4K bytes are more common.
.SH 2 "Implementation"
.lp
The address-translation table used by the hardware is called a \fBpage table\fP.
.IN "Page table"
Its structure is determined by the architecture of the computer.
Entries in the page table do not contain a Size field, since they are all the
same length.
As with segmentation,
address translation first splits the virtual address into two parts:
the \fBpage number\fP and the \fBoffset\fP.
.IN Paging "Offset"
Here is how a typical translation table might look:
.(l
.FS
.ne 8
 1	\fWtype\fP
 2		AccessType = (Read, Write, Execute);
 3		PageTableEntry =
 4			\fWrecord\fP
 5				Location : PageFrameNumber;
 6				Permissions : \fWset\fP \fWof\fP AccessType;
 7				Present : Boolean;
 8			\fWend\fP;
.ne 3
 9	\fWvar\fP
10		PageTable: \fWarray\fP 0 : PageLimit-1 \fWof\fP PageTableEntry;
11		NumberOfPages : 0 .. PageLimit;
.FE
.)l
Because pages are aligned to page frames in physical store,
the table need not keep physical addresses for the Location field;
the page frame number suffices.
Here is the algorithm followed by the hardware to translate each access:
.(b
.FS
12	\fWprocedure\fP Translate (VA : VirtualAddress; AT : AccessType) : PhysicalAddress;
13	\fWvar\fP
14		VirtualPage : 0 .. PageLimit \(mi 1;
15		Offset : 0 .. PageSize \(mi 1;
16	\fWbegin\fP
17		VirtualPage := VA \fWdiv\fP PageSize; { performed by extracting bits }
18		Offset := VA \fWmod\fP PageSize; { performed by extracting bits }
19		\fWif\fP VirtualPage \(>= NumberOfPages \fWthen\fP
20			trap("invalid page number");
21		\fWelse\fP
22			\fWwith\fP PageTable[VirtualPage] \fWdo\fP
23				\fWif\fP \fWnot\fP Present \fWthen\fP
24					trap("missing page"); { page fault; not an error }
25				\fWelsif\fP \fWnot\fP AT \fWin\fP Permissions \fWthen\fP
26					trap("security violation");
27				\fWelse\fP
28					Translate := Location*PageSize + Offset;
29				\fWend\fP; { if }
30			\fBend\fP; { with }
31		\fBend\fP; { if }
32	\fBend\fP Translate;
.FE
.)b
.pp
Figure \n(ch.\n+(FN shows the algorithm followed by the hardware on every
access to translate and check addresses.
.(z
.GS
width 3
height 3
file figs/fig.3.pg.translate
.GE
.sp .2i
.FI "Hardware algorithm for translating page-oriented access"
.)z
This algorithm involves checking the page table, which is typically
in main store, for every translation.
(Some early machines, such as the Nova 3/D,
.IN "Nova 3/D" computer
stored page tables in high-speed registers, but this
design has become less popular with potentially large page tables.)\ \&
To reduce the time required for translation, many machines provide a
\fBtranslation look-aside buffer\fP (TLB)
.IN "Translation look-aside buffer"
to hold the most frequently used parts of the current page table.
The hardware simultaneously tries to find the desired entry in the TLB
and in main store.
If the entry is in the TLB (there is a cache hit),
.IN "Cache hit"
the hardware finds it very quickly (on the order
of 100 nanoseconds) and aborts its attempt to get the entry from main store,
an operation that would have taken about one microsecond to complete.
If the entry is not in the TLB (there is a cache miss),
.IN "Cache miss"
the main-store reference is allowed to complete,
and its value is not only used for address translation but also placed in the
TLB for future use.
We will postpone discussing which entry in the TLB is discarded in this case;
the policies are very similar to policies we will introduce for page
replacement.
Typical translation look-aside buffers can achieve a 95 percent hit ratio.
.IN "Hit ratio"
.pp
Here is an example of a paging situation with two processes, \fBA\fP and
\fBB\fP.
We will assume that both are simultaneously using parts of physical store.
For simplicity, we will make virtual addresses two decimal digits, of which the
first is the page number and the second is the offset.
We will also show the contents of a translation look-aside buffer
that holds two entries of the current page table.
For this example, we will not let the TLB entries change, even if a
cache miss occurs.
\fBA\fP is currently running, so entries in the translation look-aside buffer
.IN "Translation look-aside buffer"
refer to it.
.(b L
.sz -2
.ce
\fXPage table for process A\fP
.ns
.sz +2
.TS
box, center;
cb | cb
n | c.
.sp 3p
Virtual page number	Page frame
.sp 3p
_
.sp 3p
0	7
1	8
2	absent
3	4
.sp 3p
.T&
l s.
_
.sp 3p
Number of pages = 4
.sp 3p
.TE
.)b
.(b L
.sz -2
.ce
\fXPage table for process B\fP
.ns
.sz +2
.TS
box, center;
cb | cb
n | c.
.sp 3p
Virtual page number	Page frame
.sp 3p
_
.sp 3p
0	2
1	absent
2	8
.sp 3p
.T&
l s.
_
.sp 3p
Number of pages = 3
.sp 3p
.TE
.)b
.(b L
.sz -2
.ce
\fXTranslation look-aside buffer\fP (\fBA\fP is active)
.ns
.sz +2
.TS
box, center;
cb | cb
n | c.
.sp 3p
virtual page number	page frame
.sp 3p
_
.sp 3p
1	8
3	4
.sp 3p
.TE
.)b
.(b L
.sz -2
.ce
\fXMain Store\fP
.ns
.sz +2
.TS
box, center;
ce ce ce ce ce ce ce ce ce ce
l l l l l l l l l l .
.sp 3p
0	1	2	3	4	5	6	7	8	9
.sp 3p
_
.sp 3p
		\fBB\fP(0)		\fBA\fP(3)			\fBA\fP(0)	\fBA\fP(1)		
								\fBB\fP(2)		
.sp 3p
.TE
.)b
.(b L
.sz -2
.ce
\fXTranslations\fP
.sz +2
.TS
box, center;
cb | cb
n | l.
.sp 3p
Virtual address	Physical address
.sp 3p
_
.sp 3p
71	cache miss, invalid page trap
14	84
23	cache miss, page fault
06	76 (cache miss)
.sp 3p
.TE
.)b
.SH 2 "Security and sharing"
.lp
The same security scheme that we saw for segmentation can be applied to
.IN Security
.IN Sharing
paging.
Page table entries may specify permissible access independently for each
virtual page.
Because each page has separate permissions,
different items that require different access restrictions, such as
instructions and data, should not reside on the same page.
Sharing instructions between processes also requires that data be separated,
or else some data may be inadvertently shared, a situation that could confuse
both processes.
However, because the process does not involve itself in the concept of pages,
separating instructions from data on separate pages is not natural to
the process.
Still, compilers and assemblers can distinguish instructions from
data and can pass this information on to the linker.
The linker, in turn, can segregate regions that require separate access
restrictions
.IN "Linker"
and can make sure that each starts on a new page.
(This practice increases the amount of internal waste.)\ \&
.IN "Internal waste"
The load image is then marked so that the loader can identify and properly
.IN "Load image"
.IN "Loader"
restrict access to each region.
.pp
One clever idea, found in VMS and Dynix,
.IN VMS "operating system"
.IN Dynix "operating system"
is for the linker to put nothing in the first page of virtual store.
This page is then marked invalid by the storage manager.
If the process accidentally tries to follow a null or uninitialized pointer,
there is a good chance that it will generate an access into that first page.
which will cause a trap.
