.SH 1 "Segmentation"
.lp
We will now allow each process to have a number of segments, numbered 0 through
@n-1@, where the process has @n@ segments.
There will be some limit on the number of segments permitted, determined, of
course, by the hardware.
On the DEC PDP-11, there can be eight segments, each 8K
bytes long.
.IN "DEC PDP-11" "computer"
This limitation, both in the number of segments and in their maximum length, has
led operating systems for the PDP-11 either to ignore segmentation entirely
and give each process just one contiguous space or to restrict the use of
segments.
Unix, for example, gives each process three segments.
.IN Unix
The first is for instructions and may be shared.
The second is for data, and the third is for the stack.
Each of these segments of virtual space may be composed of several adjacent
hardware segments.
For example, one segment of virtual space could be 23K bytes long by occupying
three hardware segments on the PDP-11.
.pp
Having lots of variable-sized segments can be an advantage.
Segmentation allows the programmer or the compiler to partition the \fBname
space\fP
.IN "Name space"
of a program\|\(em\|that is, the space of identifiers such as variables and
procedures\|\(em\|into discrete segments.
Hardware segmentation supports program modularity.
.IN "Modularity"
For example, a program written in Pascal usually has to check the subscripts of
each array reference to make sure they are within bounds.
If a separate segment is devoted to that array, the hardware can automatically
check the subscripts and relieve the software of that burden.
It is also natural to put library routines in segments by themselves and then
share those routines among all the processes that might need them.
.SH 2 "Implementation"
.lp
It is no longer sufficient to use the most significant bit of a virtual address
to distinguish which segment is being used.
Instead, the top @m@ bits of each virtual address are used to select one of
@2 sup m@ segments.
Of course, the maximum number of segments is therefore @2 sup m@.
The remaining bits are used to select the \fBoffset\fP within the segment.
.IN Segments offset
The translation table now has an entry for each segment, so we call it a
\fBsegment table\fP.
.IN "Segment table"
.IN "Translation table"
That entry contains the start, size, and permissions of that segment:
.(l
.FS
.ne 7
 1	\fWconst\fP
 2		BitsPerPhysicalAddress = 24; { for example }
 3		BitsPerVirtualAddress = 16; { for example }
 4		BitsPerSegmentAddress = 6; { for example }
 5		BitsPerOffsetAddress  = BitsPerVirtualAddress \(mi BitsPerSegmentAddress;
 6		SegmentLimit = 2\u\s-2BitsPerSegmentAddress\s+2\d\x'-3p';	
 7		OffsetLimit  = 2\u\s-2BitsPerOffsetAddress\s+2\d\x'-3p';
.ne 10
 8	\fWtype\fP
 9		PhysicalAddress = integer; { containing BitsPerPhysicalAddress bits }
10		AccessType = (Read, Write, Execute);
11		SegmentTableEntry =
12			\fWrecord\fP
13				Start : PhysicalAddress;
14				Size : integer; { length of virtual space }
15				Permissions : \fWset\fP \fWof\fP AccessType;
16				Present : Boolean;
17			\fWend\fP;
.ne 3
18	\fWvar\fP
19		SegmentTable : \fWarray\fP 0 : SegmentLimit-1 \fWof\fP SegmentTableEntry;
20		NumberOfSegments : 0 .. SegmentLimit;
.FE
.)l
If we use the numbers suggested in the declarations (lines 2\(mi4), virtual
addresses are 16 bits long.
Virtual space is therefore limited to 64K bytes.
A process may have a virtual space ranging anywhere between 0 bytes and this
limit, divided into segments.
Of the 16 bits, 6 are used to name the segment.
A process may therefore have at most 64 segments,
each with a maximum size of 1K bytes.
The only way for a process to have a full 64K-byte virtual space is to have all
64 segments and for each segment to be the full 1K bytes in length.
In our example, a physical address is described by 24 bits.
The largest amount of physical store that may be placed on the machine is
therefore 16M bytes.
A particular machine may have less store, of course.
Main store for computers of this size
is usually purchased in increments of 256K, 512K, or 1M bytes.
.pp
The algorithm used by the hardware on every access can be represented as
follows:
.(b
.FS
21	\fWprocedure\fP Translate (VA : VirtualAddress; AT : AccessType) : PhysicalAddress;
22	\fWvar\fP
23		Segment : 0 .. SegmentLimit - 1;
24		Offset : 0 .. OffsetLimit - 1;
25	\fWbegin\fP
26		Segment := VA \fWdiv\fP OffsetLimit;
27		Offset := VA \fWmod\fP OffsetLimit;
28		\fWif\fP Segment \(>= NumberOfSegments \fWthen\fP
29			trap("invalid segment number");
30		\fWelse\fP
31			\fWwith\fP SegmentTable[Segment] \fWdo\fP
32				\fWif\fP \fWnot\fP Present \fWthen\fP
33					trap("missing segment");
34					{ segment fault; not an error }
35				\fWelsif\fP Offset \(>= Size \fWthen\fP
36					trap("offset out of range";
37				\fWelsif\fP \fWnot\fP AT \fWin\fP Permissions \fWthen\fP
38					trap("security violation");
39				\fWelse\fP
40					Translate := Start + Offset;
41				\fWend\fP; { if }
42			\fWend\fP; { with }
43		\fWend\fP; { if }
44	\fWend\fP Translate;
.FE
.)b
Figure \n(ch.\n+(FN shows the algorithm followed by the hardware on every
access to translate and check addresses.
.(z
.GS
height 4
file figs/fig.3.seg.translate
.GE
.FI "Hardware translation for segmentation"
.)z
As a concrete example, consider the following situation.
Again, let us use decimal notation.
Assume that a virtual address is three digits long and physical store has 20,000
bytes, as before.
Assume that the first digit determines the segment number and the other two
digits determine the offset.
.(b L
.sz -2
.ce
\fXSegment Table\fP
.ns
.sz +2
.TS
box, center;
cb | cb cb cb cb
n | n n a a.
.sp 3p
Segment number	Start	Size	Permissions	Present
.sp 3p
_
.sp 3p
0	13426	26	RWX	yes
1	2309	00	R	yes
2	1000	23	 WX	yes
3	\(em	95	 W	no
4	10000	100	RWX	yes
.sp 3p
.T&
l s s s s.
_
.sp 3p
Number of segments = 5
R = Read, W = Write, X = Execute
.sp 3p
.TE
.)b
.(b L
.TS
box, center;
cb cb | cb
n c | l.
.sp 3p
Virtual address	Access type	Physical address
.sp 3p
_
.sp 3p
0	R	13426
217	W	1017
217	R	trap: security violation
232	R	trap: offset out of range
394	R	trap: missing segment, security violation
499	X	10099
725	X	trap: invalid segment number
.sp 3p
.TE
.)b
Several traps can apply to the same reference.
For example, the Read reference to location 394 traps both because the segment
is missing and because of security violation.
If the machine can only report one trap for a reference,
it is better to report traps that indicate program error (like security
violation) than those that indicate that help from the storage manager is
needed (like missing segment).
.SH 2 "Storing the necessary information"
.lp
Missing segments are, of course, swapped out.
The segment table must indicate that accesses to such segments should
generate traps.
We have shown a Boolean (one-bit) field called
.q "Present"
.IN "Present field"
in the segment table (line 16 of the declarations) to indicate this fact.
Instead, we could have set the segment length to zero or removed all access
privileges.
Then the hardware would still trap to the kernel, but now the kernel would have
to refer to the context block for the process to find out whether the trap
signals that the process is attempting an invalid access or a valid access
to a missing segment.
.pp
Such treatment of the segment-length or permissions field
is an example of the \fBmirror tables\fP technique.
.IN "Mirror tables"
Under this technique, the hardware table is intentionally misleading in order
to have the hardware take some special action that would be too expensive in
software.
The actual state of affairs is honestly represented in a second set of tables
stored by the kernel.
Mirror tables are used either when the hardware lacks some needed field
(for example, the hardware segment table lacks a Present
field) or for reasons of efficiency (for example, composite page tables for
processes that are themselves operating systems).
.pp
Before a missing segment may be accessed,
the storage manager must swap that segment back into main store.
.IN "Storage manager"
To find it on backing store, the storage manager must remember its
location.
Some hardware provides another field in the segment table for this purpose.
This field is ignored by the hardware.
If the hardware provides only a Present
field, the storage manager can use the
Start and Permissions fields to store the backing-store address, if there
is room.
The use of the same field to mean different things at different times is called
\fBoverloading\fP.
.IN "Overloading"
If overloading is not possible, the storage manager can save a mirror copy of
.IN "Mirror tables"
the translation table in the context block, and this copy can have a
.q "backing-store location"
field.
.SH 2 "Swapping"
.lp
Once the missing segment has been swapped back in, the kernel switches back to
the context in which the process was running.
Its program counter is reset to the
.IN "Context switch"
instruction that failed previously.
The hardware then tries to execute that instruction again.
If all is well, no new trap will occur.
However, the instruction might still fail for several reasons.
It might require another operand from another segment that is also missing.
In fact, a single instruction on a three-address machine can cause five
segment faults: one for each of the two operands, one for the result,
and two for the instruction itself if it lies on a boundary between two
segments.
(We are not even considering the case where operands can cross segment
boundaries!)\ \&
The storage manager keeps bringing in segments until the process can continue.
With luck, some other process is ready and can make use of the processor
while the segments are being swapped in.
.pp
The decision to swap out a segment does not completely inactivate the process
that was using it.
As long as the process avoids accessing virtual addresses within that segment,
it can continue to run.
When it does require that segment, the hardware will trap to the kernel.
We call this event a \fBsegment fault\fP.
.IN "Segment fault"
The storage manager will swap that segment back in.
While the segment is being swapped in, which takes on the order of 20 to 50
milliseconds (depending on what sort of backing store is used and how long the
segment is), the kernel switches to some other process.
Once the segment is in, the first process will be allowed to continue.
It has no inkling that a segment fault occurred.
(Otherwise, the operating system would violate the Beautification Principle.)
.IN "Beautification principle"
.pp
When the storage manager needs to swap some segment out in order to make room
for some other segment, it is best to choose a segment
that is not currently in use by any process.
Otherwise, a segment fault in one process may impede progress in other
processes as well.
We will devote a great deal of attention to mechanisms to accomplish this feat
when we discuss paging.
.pp
One danger should be mentioned.
Since a single instruction might require five or more different segments in
order to be executed,
segments that have just been brought in should not be subject
to swapping out before they have even been used the first time, or else the
operating system might thrash by victimizing segments that are needed for the
.IN Thrashing
current instruction.
.pp
Swapping out a segment is not always expensive.
A segment that does not have write permission (by any process that uses it)
need not be written to backing store because it has not changed since it was
last on backing store.
It is only necessary to reset its Present field to false in the segment table
.IN "Present field"
and to reclaim its space.
Similarly, writable segments that have not been changed
since they were last swapped in are cheap to swap out.
Some machines set a Boolean field called
.q "Dirty"
.IN "Dirty field"
in the segment table entry to true
whenever a write access occurs to the corresponding segment.
The storage manager initializes Dirty
to false whenever it swaps the segment in.
When a segment is about to be swapped out, it only needs to be written to
backing store if Dirty is true.
Of course, this mechanism requires that a copy be kept on the backing store of
all segments that are brought into main store.
.SH 2 "Large segment spaces"
.lp
On some machines, the number of bits devoted to the segment number is very
large.
For example, the Honeywell 645, which supports the Multics
.IN "Honeywell 645" "computer"
.IN "Multics"
operating system, has 18 bits of segment number.
Instead of using segments in a range from zero to some reasonably small number,
programs in such machines may use a few segment numbers scattered within the
enormous space of possibilities.
The reason Multics uses so many segments is that every subprogram that exists
has its own segment number that is reserved for as long as the subprogram is
saved on secondary store; this time could be years.
.pp
The implementation just described does not work very well for large, sparse
segment spaces because the segment table has one entry for each potential
segment.
The hardware is certainly not going to refer directly to a table with
@2 sup 18@ = 256K entries;
such a table would require more physical store than most computers have!
Instead, a hashing technique can be used, a technique that first converts the
.IN Hashing
segment number into an index in a hash table where the desired entry can be
found.
Another alternative is to apply paging to the segment table.
To understand that solution, we must first understand paging,
which will be presented shortly.
.pp
The hardware either stores the entire segment table in registers
if it is short or uses a cache to store the most frequently used entries,
in conformance with the Cache Principle.
.IN "Cache principle"
If the hardware had to refer to main store each time it needed to translate an
address, the effective speed of the machine would be halved.
The storage manager does not need to know anything at all about the cache,
except that whenever the kernel modifies the segment table itself,
for example when switching processes,
the hardware must find out which entries in the cache are invalid.
Machines that have such \fBtranslation look-aside buffers\fP,
.IN "Translation look-aside buffer"
as we will call these caches, have an instruction called
.q "flush the cache,"
which informs the hardware that all the entries in its cache, if any, are no
longer valid.
The kernel should execute this instruction as part of process switching.
.SH 2 "Shared segments"
.lp
Two processes that share a segment don't have to use the same segment number,
nor need they have identical privileges over that segment.
One process could call it segment 5 and have RX privileges, and the other
could call it segment 1 and have WR privileges.
This segment has two separate entries,
one in the segment table for each process.
The same physical segment may even appear twice
in the virtual space of one process.
However, sharing an instruction segment with different segment numbers can be
difficult.
Such segments tend to refer to themselves (in jumps, for example).
These references cannot have the segment number as part of the instruction.
Instead, they must refer to a register that holds the segment number.
.pp
Shared segments cause problems with Dirty
.IN "Dirty field"
fields because a shared segment may be written by process \fBA\fP,
and then swapped out when process \fBB\fP is running.
The hardware only knows about process \fBB\fP's segment table at this point, and
the table will not indicate that the segment is dirty unless \fBB\fP has
written into it.
The storage manager must consult a shared-segment table in order to track
.IN "Shared-segment table"
down all the Dirty
fields that correspond to this shared segment in all the
context blocks for the processes that share it.
.SH 2 "Unusable space"
.lp
Segmentation creates three kinds of unusable space that cannot be applied to
any process.
.sp 10p
.ip \(bu
Free pieces of physical store that are not in use at the moment are wasted.
Some of these pieces may be too small to be of any use until the storage
manager shuffles the current segments.
We will call such unusable space \fBexternal waste\fP,
.IN "External waste"
since it is outside the space of any process.
We also saw external waste under the partition method.
.IN "Partitions"
.ip \(bu
Context blocks must hold copies of the segment table
during the time that a process is not running.
Space devoted to the segment table cannot be put to other uses
when the process is running; it is a fixed cost of segmentation.
If the segment tables are large, they may be swapped out to reduce this
cost.
Space that is unusable because it is needed for these tables we will call
\fBoverhead space\fP.
.IN "Overhead space"
.ip \(bu
A process might not be using all the space in a segment; it may have
asked for a segment that is bigger than is really needed.
This type of unused area is called \fBinternal waste\fP.
.IN "Internal waste"
.sp 10p
.lp
Segmentation tends to have very little internal waste because processes
generally request exactly the amount of space they need.
Any internal waste that exists is not the storage manager's fault; it is due to
restrictions imposed by the hardware that require segment lengths to be
divisible by some power of 2.
.pp
External waste can be reduced by shuffling, but that solution is expensive.
.IN "Shuffling"
If segments are aligned to chunk boundaries, at the end of each segment there
will be a small amount of unavoidable external waste ranging from 0
bytes to 1 byte less than the chunk size.
On the average, this external waste will be one-half a chunk size for each
segment in main store.
The amount of overhead space depends on how many segments are allowed, a
number that itself depends on the number of bits in a virtual address
allocated to the segment number.
The more segments that are allowed, the more space is devoted to overhead, and
the smaller the segments may be.
The amount of overhead space is not under the control of the operating system
designer except for the decision to swap unused segment tables.
.SH 2 "Initial loading and execution"
.lp
A program needs separate segments to store different data
items or subroutines.
Cooperation between the storage manager and the loader establishes the initial
assignment of segments in virtual store, and cooperation between the storage
manager and the running process can change that assignment dynamically.
.IN "Storage manager"
The load image must specify in which segment each instruction or data section
belongs.
Instruction segments typically get Execute permission only;
data sections can be specified to be either Read/Write or only Read.
When the loader
.IN Loader
initializes virtual space, it can bring all the segments into main store or can
leave them on backing store.
In the latter case, the program will slowly acquire the segments it needs by
faulting.
.pp
The Split service call
.IN "Service call" Split
might indicate the disposition of each segment in the parent's virtual space:
.sp 10p
.ip \(bu
\fBUnique\fP.
The segment becomes the sole property of either the parent or the child.
The segment table for the other process indicates that the segment is
absent.
.ip \(bu
\fBShared\fP.
The segment is listed in the segment tables of both parent and child.
.ip \(bu
\fBCopied\fP.
The child gets a private version, initialized as a copy of the segment still
owned by the parent.
The storage manager copies the data and builds a new segment,
allocating backing store at the same time.
Alternatively, the storage manager can employ a form of \fBlazy evaluation\fP:
.IN "Lazy evaluation"
Let child and parent share the segment, but deny both of them Write
permission.
Perhaps neither will ever try to modify the data.
As soon as either one does, the hardware will trap to the storage manager,
which can make a new copy for the child and give both processes Write
permission over their (now private) copies.
This technique is called \fBcopy on write\fP.
.IN "Copy on write"
.sp 10p
.pp
A process that wants to change the structure of its virtual store dynamically
can use a variety of service calls:
.sp 10p
.ip \(sq
\fHGetSegment(descriptor) / ReleaseSegment(segment).\fP
.IN "Service call" GetSegment
.IN "Service call" ReleaseSegment
The first call returns the segment number of a newly allocated segment.
The descriptor indicates the desired size and permissions.
The second call removes the given segment from the caller's segment table.
.ip \(sq
\fHGetStore(segment, size) / ReleaseStore(segment, size).\fP
.IN "Service call" GetStore
.IN "Service call" ReleaseStore
These calls lengthen or shorten a segment.
.ip \(sq
\fHChangeSegment(segment, permissions).\fP
This call lets the process add or remove Read, Write, and Execute permissions
to or from its segment.
The operating system may disallow changing permissions on shared segments.
.sp 10p
.lp
