.SH 1 "Device drivers" I
.lp
The bewildering variety and number of devices for computers is one of the
severest challenges of the operating system writer.
Each brand of disk, channel, tape, or communication device has its own
control protocol, and new devices come on the market with great frequency.
It is essential to reduce this variety to some sort of order; otherwise, every
time a new device is attached to the computer, the operating system will have
to be completely reworked.
Forcing a regular structure, at least at some level inside the kernel, works
across a wide range of devices.
The device-specific parts of device control can then be sequestered into
well-defined modules.
.pp
The Unix operating system sets a good example of forcing order on the chaos of
.IN Unix "operating system"
device types.
Each device can be described by two numbers: the type (each
different brand of tape drive is usually a different type) and the instance
(each separate drive of the same type has its own instance number).
Each type is associated with a \fBdevice driver\fP that accommodates the
.IN "Device driver"
peculiarities of that device type.
The device driver for a given type maintains separate data structures for each
instance.
.pp
Device drivers provide a standard interface to the rest of the kernel.
They provide routines for \fBopening\fP and \fBclosing\fP the device
.IN Device open
.IN Device close
(used by the kernel when a process wants to start or stop using the device)
as well as routines for transmitting data to or from the device.
Only a few data formats are used, so that the device driver knows what to
expect.
For example, data may be transmitted in fixed-size blocks, with a standard
header.
This header might include the following information.
.sp 10p
.ip \(bu
device type
.ip \(bu
device number
.ip \(bu
address of the data on the device
.ip \(bu
address of the data in main store
.ip \(bu
amount of data to transfer
.ip \(bu
whether the data are to be read or written
.ip \(bu
what kernel module is waiting for the completion of this operation
.sp 10p
.lp
.SH 2 "Two levels of device driver"
.lp
Device drivers must serve two masters: the rest of the kernel and the device
itself.
The kernel indicates its desires by calling the routines just
mentioned, asking for data to be read or written.
When these transfers are complete, the device driver is expected to inform the
appropriate kernel module.
(We will discuss how the individual kernel modules can signal each other in
Chapters 8 and 9.
For now, imagine that there are little processes, which we will call
\fBtasks\fP, in the kernel, and that one is waiting for the completion of this
.IN "Kernel tasks"
trans\%put operation.
It is the job of the device driver to unblock that task once its request has
been satisfied.)\ \&
The device itself is given commands by the device driver.
When these commands have been accomplished, the device causes an interrupt.
This interrupt is set to transfer control to the appropriate device driver.
.pp
One useful organization for device drivers to help them deal with these very
different sorts of requests is to divide them into two pieces.
Each driver has an
.q "upper"
and a
.q "lower"
part.
The two parts communicate by sharing data structures.
This organization is shown in Figure \n(ch.\n+(FN.
.(z
.GS
width 4
height 3
file figs/fig.5.twolevel
.GE
.FI "Two-level structure of a device driver"
.)z
.pp
The upper part accepts requests from the rest of the kernel.
For example, the upper part of the disk driver might accept a request from the
storage manager to write out a page onto backing store.
Similarly, a process that wants to write to a printer might pass data by a
system call, which some kernel task might then pass
on to the printer driver.
The upper part of the device driver transforms these requests into
entries in a list of pending work for the lower part.
It may even sort the entries in the pending-work list so that some trans\%put
operations precede others.
.pp
The lower part only wakes up when there is an interrupt or when new work is
placed on its pending-work list.
It usually sets the processor status
.IN "Processor status"
in such a way that other interrupts from the same device are prevented until
it is finished with its current work.
If the interrupt that awakens the lower part indicates completion of some
trans\%put, the appropriate data structure is so marked, and the kernel task
that was waiting is unblocked.
The upper driver need not be informed.
The lower driver then checks to see if there is any more work queued up for it,
and if so, it initiates the next operation.
.pp
To give some feeling for the sources of complexities in device drivers, we will
consider three devices: the clock, terminals and disks.
.SH 2 "Clock device driver" I
.lp
Clocks come in several styles.
The simplest, a \fBline clock\fP,
.IN Line Clocks
generates an interrupt every
.q tick, 
that is, every 60th (or 50th) of a second.
It might also have a register that indicates how many ticks have occurred since
the register was reset.
The clock may be protected against power failure by a battery, which allows it
to continue ticking even when the computer is down or without power.
Slightly more complex clocks have a register that indicates how many interrupts
were missed due to the cpu running at a priority too high to let the clock
interrupt.
\fBProgrammable clocks\fP
.IN Programmable Clocks
also have a count register that may be set by the software.
The clock decrements that register at a standard rate (often every
microsecond) until it reaches 0, at which time it generates an interrupt.
Some computers have both line clocks, which are used for recording the time of
day, and programmable clocks, which are used for marking scheduling events.
If there is no programmable clock, the line clock alone can be used for
scheduling.
.pp
When a line clock interrupts, the clock driver typically
performs the following functions:
.sp 10p
.ip \(bu
Increment the software time data structure.
.ip \(bu 
If this clock is used for scheduling, decrement the remaining-time field of the
currently running process, if any.
If that value has become zero, invoke the scheduler.
.ip \(bu
Perform simple accounting.
For example, the currently running process might be charged for one tick worth
of time and for its current main-store use.
.ip \(bu
If there is no programmable clock, decrement the counter for the next alarm,
.IN Alarms
and if it is now zero, invoke whatever kernel task was waiting for an alarm.
Alarms are used for timeouts, which are used in communication protocols and
.IN Timeouts
elsewhere.
If there is a programmable clock, it will be used to schedule alarms.
.ip \(bu
If the current process wishes to be profiled,
.IN Profiling
use its program counter to determine in which region of the program it is
executing and increment the associated counter.
(For example, each page of the virtual space of the process could have a
separate counter.)\ \&
These counters can be used by the programmer to determine where the program is
spending its time.
.sp 10p
.lp
The necessary actions can often be accomplished very quickly, and context
can then switch back to the interrupted process.
.SH 2 "Terminal device drivers" I
.lp
The Beautification Principle applies to both the process interface and the
user interface.
.IN "process interface"
.IN "user interface"
The user interface is heavily influenced by the device driver for the terminal.
When the user types a key on the keyboard, the lower part of the driver might
\fBecho\fP the character on the associated display device.
.IN Terminal "Echo"
(A terminal on a half-duplex line does not need this service.)\ \&
.pp
Often a user will type ahead before any program has asked to read the
information.
.IN "Terminal" "type-ahead"
The device driver might store a certain amount of type-ahead in a buffer so
that it is not lost.
Most operating systems echo typed-ahead characters as soon as they are typed,
but some, like Tenex,
.IN Tenex
wait until the characters are read by some process before echoing.
This latter approach has the advantage that the display is never cluttered with
typed-ahead characters.
In addition, the process that wants to read the characters might not want them
echoed at all; deferring the echo lets the driver accommodate such wishes.
But it is hard to type accurately if you can't see what you are typing.
.pp
Many operating systems provide \fBediting\fP facilities that
.IN Terminal "Editing"
the user can invoke before a program reads data that have been typed in.
Mistakes can be removed by deleting or overwriting the offending characters.
Most operating systems define at least two editing characters, one to erase the
entire line that has been typed in so far and another to erase the last
character on the line.
These facilities are available until the user \fBactivates\fP the line by
.IN Terminal "Input activation"
pressing the carriage-return key.
Special characters are reserved for editing purposes.
The backspace key, which is the same as <control>H in the ASCII character set,
.IN "ASCII"
is quite common for erasing a single character.
The <control>U key is often used for erasing an entire line.
These keys echo in such a way that the user can tell how much has been erased.
One convention is to repeat the character that is being erased.
Another, which works only on video displays, is to erase the character from the
screen.
Erasure is usually performed by moving the cursor back, printing a space,
and then moving the cursor back again.
Erasing a tab character from the screen requires special effort, since the
number of repetitions of this erasure sequence depends on the exact position of
the character before the tab.
.pp
More sophisticated operating systems provide additional features, including the
following.
.sp 10p
.ip \(bu
Erasing one word
.ip \(bu
Moving backward or forward in the input line
.ip \(bu
Deleting or adding text at the current location in the input line
.ip \(bu
Searching for characters in the line
.ip \(bu
Recalling previous lines that have already been activated.
This facility is particularly useful if the user types in a command to a
program, but the command has an error.
After the program has rejected it, the user can recall the erroneous line,
correct it by using the editing features, and then resubmit it.
.sp 10p
.lp
There are many different terminals on the market, and each seems to have its
own conventions for such actions as clearing the screen and moving to a
particular place on the screen.
In addition, terminals are capable of transferring data at different rates.
Some operating systems provide some amount of \fBcustomization\fP in how they
.IN "Customization" terminals
deal with terminals.
For example, once the terminal driver knows what kind of terminal is being used,
it can decide how to echo editing characters.
Users have personal preferences, as well.
Some prefer to use <control>X to erase lines.
Customization can be provided by service calls that establish these special
characters.
The user who wants a particular set of conventions can run a program that
submits these calls.
Even the activation character, which is usually just the carriage return,
can be reset in many operating systems.
.SH 2 "Disk device drivers" I
.lp
One trivial function of a disk device driver is to smooth out the chunky
structure of the physical disk, with its sectors, tracks, and cylinders,
and provide the abstraction of a virtual disk that is a linear array of
sectors.
Sector @s@ on track @t@ in cylinder @c@ has this identifying number:
.EQ
SectorId ~~=~~ ( c cdot TracksPerCylinder ~+~ t ) cdot SectorsPerTrack ~+~ s
.EN
Higher levels of software in the kernel can use sector identifiers instead of
triples to describe locations on a disk.
.pp
A less trivial function of the disk driver is to
reduce the effect of latencies in accessing the disk.
Several techniques are available.
.sp 10p
.ip \(bu
Processes that require disk access may be blocked in a disk-wait list
until the request can be satisfied.
Meanwhile, other processes can run.
We call this technique \fBsynchronous trans\%put\fP.
.IN "Blocking" "Transput"
Alternatively, a process that can perform useful computation before the trans\%put
completes may be allowed to continue.
We call this technique \fBasynchronous trans\%put\fP.
We discuss this difference later.
In either case, it is important to overlap trans\%put and computation.
.ip \(bu
Large data objects can be stored in contiguous regions on the disk
so that once the first seek has finished, the rest of the data can be accessed
without additional delays.
For this reason, allocation of disk space is often made in increments of
entire tracks or even entire cylinders.
.ip \(bu
Similarly, the blocks that constitute a file might be constrained to reside
close to each other.
(We will discuss physical allocation of files in Chapter 6.)
.ip \(bu
When the same disk pack is used both for secondary store (files) and backing
store (swap space), the swap space should occupy the tracks halfway between the
center and the edge, because items in this location can be accessed with the
lowest average seek latency.
.IN "Seek" latency
.ip \(bu
Disk requests are continually generated by main-store management (swapping of
pages or segments) and file usage; often several requests are pending.
We can impose an order on the outstanding disk requests so that they can be
served efficiently.
.sp 10p
.lp
We will direct our attention to this last strategy, which is known as
\fBdisk-head scheduling\fP.
.IN "Disk-head scheduling"
.pp
In general, disk-head scheduling makes the following assumptions:
.sp 10p
.ip \(bu
There is only one disk drive.
If there are several, each is scheduled independently.
.ip \(bu
All requests are for single, equal-sized blocks.
.ip \(bu
Requested blocks are randomly distributed on the disk pack.
.ip \(bu
The disk drive has only one movable arm, and all read/write heads are attached
to that one arm.
.ip \(bu
Seek latency is linear in the number of tracks crossed.
.IN Latency seek
(This assumption fails if the disk controller has mapped tracks at the end of
the disk to replace ones that have bad sectors.)
.IN Bad Sectors
.ip \(bu
The disk controller does not introduce any appreciable delays.
.ip \(bu
Read and write requests take identical time to service.
.sp 10p
.lp
We will evaluate various policies by describing how long requests have to wait
as a function of the load.
Both the average wait time and the variance of the wait time are important.
A low average with a high variance means that some requests are taking a very
long time even though most are served quickly.
The load is measured in requests per second.
We could also use the length of the outstanding-request list.
.pp
The simplest policy is to order disk requests on a first-come, first-served
(FCFS) basis.
.IN "Disk-head scheduling" "First-come, first served (FCFS)"
FCFS makes sure that every request is eventually served; there is no starvation.
However, every request is likely to suffer a seek operation.
This method works fine for very low loads, but it saturates quickly \(em that
is, the load doesn't have to grow very large before the disk can no longer
keep up with the work.
The variance of the mean wait time is fairly low.
.pp
FCFS disk-head scheduling and its refinements once again illustrate the
Law of Diminishing Returns, which we first saw in Chapter 1.
.IN "Law of Diminishing Returns"
There are many alternatives to FCFS, and they are all much better.
For example, the Pickup
.IN "Disk-head scheduling" "Pickup method"
method keeps the requests in FCFS order to prevent starvation,
but on its way to the track where the next request lies, it
.q "picks up"
any requests that can be serviced along the way.
.pp
Another alternative policy is Shortest Seek-latency First (SSF),
.IN "Disk-head scheduling" "Shortest Seek-latency First (SSF)"
which next serves the request whose track is closest to the current one.
This policy does allow starvation, but if a request is consistently bypassed,
the disk must be incapable of keeping up with the disk requests in any case.
This kind of starvation is not as bad as the underlying problem, which is very
likely that the main-store allocation policy is causing thrashing.
.IN "Thrashing"
SSF has the advantage of minimizing the total seek latency and
.IN Seek latency
saturating at the highest load of any policy.
On the negative side, it tends to have a larger wait-time variance.
.pp
One very popular policy is called Look.
.IN "Disk-head scheduling" "Look"
The disk head is either in an inward-seeking phase or an outward-seeking phase
at any time.
As it seeks inward, it picks up any requests it passes.
It changes direction only when there are no more requests ahead;
it then switches to seeking outwards and fills all requests in that direction as
it gets to them.
When requests come so frequently that the disk can barely keep up,
a circular variant, C-Look, can lead to better variance among the request
latencies.
.IN "Disk-head scheduling" "C-Look"
This variant always seeks inward.
When the last request has been serviced in that direction,
a single seek is performed to the outermost track with a pending request
without picking up any requests on the way out.
.pp
To illustrate these disk-scheduling algorithms, let us consider a disk with 200
cylinders (a typical number).
Let's assume that at a given instant, the read/write head is sitting at
cylinder 100.
Transfers to or from the following cylinders have been requested.
The requests are listed in the order in which they arrived.
It is irrelevant whether they are read or write requests.
.(b
	23 89 132 42 187 165 21 34 101 102 34 2 167 3 34 199 20
.)b
We will assume that no new requests arrive while we service this set.
(By the way, it is quite unusual for more than three or four requests to be
outstanding in a well-tuned operating system.)\ \&
The algorithms we have seen have the following behavior.
.(b M
.sz -2
.TS
center, box;
cb n n n n n n n n n n n n n n n n n | cb
lb n n n n n n n n n n n n n n n n n | nb.
.sp 3p
Requests	23	89	132	42	187	165	21	34	101	102	34	2	167	3	34	199	20	motion
.sp 3p
_
.sp 3p
FCFS	23	89	132	42	187	165	21	34	101	102	34	2	167	3	34	199	20	1472
Pickup	89	42	34	34	34	23	101	102	132	165	167	187	21	20	3	2	199	623
SSF	101	102	89	132	165	167	187	199	42	34	34	34	23	21	20	3	2	322
Look	101	102	132	165	167	187	199	89	42	34	34	34	23	21	20	3	2	296
C-Look	101	102	132	165	167	187	199	2	3	20	21	23	34	34	34	42	89	383
.sp 3p
.TE
.sz +2
.)b
The SSF and Look methods have similar results, but if requests had continued to
arrive during the time others were serviced, it is likely that SSF would have
turned back to pick up some new requests close to its current position.
Look turns back only when there is no further work in the current direction.
As expected, FCFS has the worst behavior.
.pp
To compare these methods further, we turn to simulation results.
The simulated disk has 200 tracks.
Seek latency is
.IN "Seek latency"
@ ( 0.5 ~+~ 0.4 ^ cdot ^ Distance ) @
milliseconds if the distance is at least one track and zero otherwise.
(The 0.5 term represents time needed just to start and stop the head.)\ \&
There are 20 sectors per track.
The disk spins at 3600 revolutions per minute.
Rotational latency depends on when the head arrives at the track
and the sector that is needed.
The simulation compares the behavior of the five policies on a workload of 1000
requests, with interarrival time selected from an exponential distribution.
The average arrival rate is allowed to vary between 5 and 100 arrivals per
second.
Both the track and the sector needed by each request are selected from a
uniform distribution.
(Very similar results were obtained from using a normal distribution for
tracks, with average value 100 and standard deviation 100.)\ \&
.pp
Figure \n(ch.\n+(FN shows the average time for a disk request to be serviced
for the different methods across the given workloads.
.(z
.GS
height 3
file figs/fig.5.meantime
.GE
.FI "Average response time"
.)z
The FCFS method is clearly terrible.
The other methods are not particularly easy to distinguish.
Surprisingly, C-Look did not perform as well as Look, even with high arrival
rate.
(It did perform slightly better than Look with requested tracks selected from a
normal distribution.)\ \&
Figure \n(ch.\n+(FN shows the throughput \(em that is, the number of requests
served per second \(em for the five methods.
.(z
.GS
height 3
file figs/fig.5.throughput
.GE
.FI "Throughput"
.)z
This figure shows even more clearly that FCFS is by far the worst but
that all the other methods are essentially indistinguishable.
If the arrival rate is less than 22 requests per second, FCFS manages just as
well as the others with respect to throughput, although its response time
starts falling behind the others at about 10 requests per second.
.SH 1 "Data modification"
.lp
Data sent to a device need not be an identical copy of the data found in main
store.
Three distinct kinds of modification may be performed, each for
a different reason.
.SH 2 "Data compaction" I
.lp
When a great deal of information is to be sent across an expensive
transmission line, modifying the data to occupy fewer bits is worthwhile.
Similarly, if data must be saved for a long time on an expensive medium, it
will be cheaper if the data can be crammed into fewer bits.
.pp
There are often obvious ways to save space.
For example, a file that represents a computer program may have many space
characters.
Instead of sending all the spaces, each blank region can be encoded by a
count of how many spaces there are.
This method is an example of \fBrun-length encoding\fP.
.IN "Run-length encoding"
Television pictures are large arrays of intensity values.
Instead of sending a 6-bit intensity value for each element of the array,
a 3-bit difference between that element and the previous one often suffices.
Some escape mechanism can be used when the difference cannot be represented in
3 bits.
This method is an example of \fBdifference encoding\fP.
.IN "Difference encoding"
More sophisticated coding methods exist, such as Huffman codes, which
.IN "Huffman codes"
assign often-used characters a short code, and less used ones a longer code.
.SH 2 "Data encryption" I
.lp
Another sort of transformation can be applied to data to hide their contents.
Actually, \fBsecrecy\fP (can anyone else read the data?) is only one part of a
.IN "Secrecy"
larger set of security considerations
that includes \fBauthentication\fP (who generated the data?) and
\fBprotection\fP
.IN "Authentication"
.IN "Protection"
(can anybody modify the data?).
The transformation used for these purposes is called \fBencryption\fP.
.IN "Encryption"
Data before the transformation is applied are in \fBcleartext\fP form.
.IN "Cleartext"
The encrypted data are in \fBciphertext\fP form.
A particular encryption algorithm may have an associated \fBdecryption\fP
.IN "Decryption"
algorithm that is used to transform ciphertext back into the original cleartext.
.pp
You might think that an encryption algorithm that has no decryption would be
useless.
Actually, such \fBtrap-door\fP encryption algorithms have an important use in
.IN "Trap-door"
authentication.
Many operating systems identify each user by a user name and a password,
.IN Passwords
which the user is expected to keep secret.
Any person who knows that name and password has all the privileges of that user.
Most operating systems store the passwords in a file;
anyone with access to that file can therefore learn everyone's password.
This situation is very vulnerable because any loss of security endangers
everyone's security, and security depends on powers entrusted to humans.
.pp
A far better approach (taken by Unix) is to encrypt each
.IN Unix authentication
password with a trap-door function and to store the encrypted passwords in a
publicly accessible file.
The encryption method is publicly available.
Now any program that wishes to authenticate a user may ask for a password,
encrypt it, and compare the result to the value in the file.
If the values match, the password is correct.
In particular, the logon program, which greets the user at the beginning of the
interactive computer session, can tell if the user knows the correct password.
However, the huge number of possible passwords and the fact that no decryption
method exists make it infeasible for an intruder to guess the password, because
every conceivable password would have to be encrypted and compared with the
file.
This scheme assumes, of course, that people choose fairly long and fairly
irregular passwords.
Some installations of Unix refuse to allow a password that is too short or too
.IN Unix "operating system"
closely related to an English word.
VMS can be set to insist that specific users change their password every so
often.
.IN Passwords
.IN VMS "operating system"
.pp
Most applications of encryption, however, require that the ciphertext be
convertible back into cleartext.
Unfortunately, all current encryption methods have problems.
The
.q "one-time pad"
.IN "One-time pad"
is the only provably secure method known at this time.
It derives the ciphertext by a byte-wise exclusive-or operation between the
cleartext and a
.q "key"
whose length must be at least as long as the message.
The decryption method uses the same key and the same algorithm.
The key must never be used again, or it becomes easier and easier to guess.
Since both the originator and the recipient of the ciphertext must agree on the
key, it must be transmitted as well, a procedure that is just as difficult as
the original problem of secure transmission.
Computer-generated pseudo-random numbers create keys that are too easy to
guess, by the way.
.pp
A currently popular encryption method is the Data Encryption Standard (DES).
.IN Cryptography "Data Encryption Standard (DES)"
The DES method was developed by the National Bureau of Standards with several
goals in mind.
Many commercial applications are currently using DES.
The algorithm (either encryption or decryption) can be performed
efficiently with a special-purpose chip, but
far less efficiently with a program.
Since the key is 56 bits long, there is plenty of room for picking unique keys.
Unfortunately, some keys turn out to be weak; if a weak key is used to encrypt,
the ciphertext can be analyzed quite easily to discover the cleartext.
It is not certain how many keys are weak or
.q "semi-weak." " "
There is some suspicion that 56 bits is insufficient to prevent a successful
attack that uses massive amounts of computer time.
It is possible that a longer key (perhaps 100 bits) would be sufficient.
.pp
The one-time pad and DES are considered
.q "conventional;"
.IN Cryptography "conventional"
they share the property that the same key is used for encryption and decryption.
A non-conventional approach called
.q "public-key cryptography"
.IN Cryptography "public-key"
uses different keys for the two transformations.
One public-key method, called RSA after its inventors (Rivest, Shamir, and
.IN Cryptography "RSA"
Adelman), has resisted attacks so far, although it has not been proved
to be computationally secure.
.pp
The RSA method itself is too complex to discuss here.
The general properties of public-key methods are elegant, however, so we will
outline them.
Every user U has a pair of keys, one for encryption, @E sub U@, and
one for decryption, @D sub U@.
It is impossible to guess @E sub U@ from @D sub U@ or vice versa.
Ciphertext encrypted with @E sub U@ is decrypted by @D sub U@.
Everyone's E key is considered public; it is saved in a public place where
anyone may access it.
Let's say \fBA\fP and \fBB\fP are two parties (programs, computers, or people)
that want to send secure messages to each other.
\fBA\fP encrypts messages for \fBB\fP using \fBB\fP's public key @E sub B@.
\fBB\fP can decrypt them using its private key @D sub B@, but no one
else can decrypt them because no one else knows @D sub B@.
Similarly, \fBB\fP encrypts messages for \fBA\fP using \fBA\fP's public key
@E sub A@.
These messages are secret.
However, \fBB\fP cannot be sure that \fBA\fP has sent the messages because
anyone could have used @E sub B@, \fBB\fP's public key.
.pp
If we want \fBB\fP to be able to authenticate \fBA\fP's messages, to be sure
that \fBA\fP sent them, we also need to assume that
if we take cleartext and apply @D sub U@, we get a sort of reverse ciphertext,
which is also secure and can be decrypted by applying @E sub U@.
Now let \fBA\fP encrypt its messages to \fBB\fP by using \fBA\fP's secret key
@D sub A@.
\fBB\fP can decrypt the messages by using \fBA\fP's public key @E sub A@.
\fBB\fP is sure that \fBA\fP sent the message because only \fBA\fP
knows @D sub A@.
However, anyone else who gets this message can also apply \fBA\fP's public key
@E sub A@ and decrypt it, so the message is not secure.
.pp
These two ideas can be combined to create messages that are both secure and
authenticated:
Let \fBA\fP first encrypt the message with its private key @D sub A@,
then encrypt the result with \fBB\fP's public key @E sub B@.
Only \fBB\fP can read this message because only \fBB\fP knows @D sub B@.
Once \fBB\fP has applied @D sub B@, it then applies \fBA\fP's public
key @E sub A@ to derive the cleartext and be sure that \fBA\fP
composed the message.
Figure \n(ch.\n+(FN shows the three cases we have just discussed.
.(z
.GS
height 3
file figs/fig.5.publickey
.GE
.FI "Secure and authenticated messages with public-key encryption"
.)z
.pp
We see that the RSA method is able to send secure, authenticated messages.
The transformations, however, are very time-consuming.
It is currently used principally as a way of securely transmitting keys that
can then be used by other, cheaper methods such as DES.
.IN Cryptography "Data Encryption Standard (DES)"
.SH 2 "Reliability"
.lp
We have seen that data transformations are useful to compress data and to
introduce security.
A third type of transformation is used for reliability.
Communication lines suffer from damaged messages for many reasons, depending
on the physical nature of the line.
Electrical disturbances, such as lightning and nearby power sources, can
introduce spurious signals.
Contention for a shared line can cause competing messages to destroy each
other.
The signal may be attenuated over distance to the point that it cannot be
reliably detected.
Similarly, bits stored on a disk or a tape might be recorded too faintly, or the
medium might have a small flaw.
When they are read back, they may appear different from how they were written.
Even bits stored in main store are subject to a low level of error.
.pp
There is only one mechanism for providing reliability \(em
\fBredundancy\fP, which means using extra bits.
.IN "Redundancy"
This single mechanism takes many forms, depending on the types of
errors that are expected and their expected frequency.
For example, an extra bit, known as the \fBparity\fP bit, can be added
.IN "Parity bit"
to each word in main store.
It is set to ensure that the total number of one bits in the word is even.
(Sometimes odd parity is used instead.
They are equivalent in purpose.)\ \&
Any word with a one-bit error will have the wrong parity, which can be detected.
Most computers have at least one bit of redundancy in their main store for this
purpose.
If two errors strike the same word, the error will not be detected.
The probability of this event is fairly low.
If the probability of double errors is not low enough,
more than one redundant bit can be placed on each word.
Now not only can one-bit errors be detected, but they can also be corrected on
the fly.
The correct word is the
.q "closest"
good word to the damaged word,
where
.q "closest"
is measured in \fBHamming distance\fP: the number of bits with different values.
.IN "Hamming distance"
Two-bit errors can be detected but not properly corrected.
If @n@-bit errors are to be detected, legal bit patterns must not have a Hamming
distance of less than @n~+~1@.
In general, by adding enough redundant bits, an \fBerror-correcting code\fP can be
designed to correct any desired number of errors.
The correct amount of redundancy is the amount that reduces undetected or
uncorrected errors to an acceptable level.
Main store is often designed to correct single-bit errors.
.pp
Parity and its generalizations are also used on devices like tape drives and
disk units.
Here the typical error is not a random bit but rather a consecutive string of
damaged bits.
A method known as the
.q "cyclic redundancy check,"
.IN "cyclic redundancy check (CRC)"
or CRC, calculates the redundant bits
to be stored at the end of each chunk of data.
This set of bits is often known as a \fBchecksum\fP.
.IN "Checksum"
.pp
Checksums and parity bits are usually below the level where the operating
system has any control.
The operating system takes over when a device (including main store) discovers
an uncorrectable error.
This discovery is usually converted into an interrupt or a trap to the kernel.
For devices like tape or disk, the device driver will retry the operation
that failed until either it succeeds or some retry limit is reached.
In the latter case, the device is broken (perhaps the tape or disk surface is
damaged), and whatever application was trying to use it must be cancelled.
The driver might make an entry in a log detailing the error.
.IN Error log
.pp
For communication devices talking to other computers, retry should eventually
succeed.
The recipient of the damaged message could send a note to the sender requesting
retransmission.
However, the request for retry might itself be lost, and the recipient might
not even know who the sender is, because that part of the message may have been
damaged.
Instead, it is customary for the recipient to acknowledge each undamaged
message it gets.
The sender retransmits any message it has sent that is not acknowledged within
a reasonable time, on the assumption that the original message was damaged in
transit.
Of course, the acknowledgement may be the one that was lost or damaged, so the
recipient must be prepared to deal with duplicate messages that the sender
accidentally sent.
Sequence numbers are used to distinguish messages so that the recipient can
discard any message with a duplicate sequence number.
