#!/usr/local/bin/perl -w

=head1 SYNOPSIS

expand.pl
	[--defsFile defs-file | --storedDefs]
	[--saveDefs]
	[--wordList | --immediate]
	[--utf8List]
	[--respellList]
	[--defList]
	[--uDefList]
	[--text text-file]
	[--wordFilter]
	[--utf8Filter]
	[--defFilter]
	[--uDefFilter]
	[--uRhymeList]

=head1 DESCRIPTION

Morphological definitions are given in the defs-file or in a database saved by
a previous run of this program with the --saveDefs option.

These definitions are used to generate reports and to filter input (given by
--text).  If --text is missing, it is taken as stdin; a "--text -" parameter
also indicates stdin.

One may generate reports and filter input in the same run, but that is not
likely useful.

Reports:
	--wordList: show Romanization of all forms;
	--immediate: like --wordList, but output online (for interactive use)
	--utf8List: show UTF-8 encoding of all forms
	--uRhymeList: same as utf8List, but sorted by ending.
	--respellList: show all words that have non-phonetic spelling (in
		Romanization) in a format appropriate for the respell program
	--defList: show all definitions
	--uDefList: show all definitions using Unicode

Filters:
	--wordFilter: show all words in the text file not generated by the defs-file.
	--utf8Filter: show all words in a utf8 text file not generated by the defs-file.
	--defFilter: echo the text, but add after each word a definition if known.
	--uDefFilter: same as defFilter, but for utf8 input.

=head1 METHOD

References to CY are to the Synopsis of Grammar at the end of College
Yiddish.

The following codes are used in the defs-file, prefixed by "/".

	D: diminutive
	N: plural in -n
	S: plural or possessive in -s
	A: adjective (with optional suffix)
	V: verb, with optional irregular infinitive
	I: noun, can take -ik or -dik or -edik
	H: Hebrew; needs separators around it
	P: preposition
	T: past tense in ge .. t
	G: separable prefix on verb, already conjugated to the past form (without
		ge)
	B: form of the past participle (including ge, if appropriate)
	C: name (needs dative, possessive)
	E: suffix
	J: -ns, as in bobens, beydns.
	K: positive, comparative, superlative adjective (can indicate new base)
	L: prefix
	-: nonexistent form, to be suppressed

Respellings are indicated by {internal codes} after a word or bar-delimited
word part.  The internal codes must not contain a space; use underscore (_)
instead.  Usually, hyphen is better.

=head1 BUGS

	/F seems to duplicate /G
	/I seems to duplicate /A

=cut

# initialized endings
	@DiminAfterVowel = ("le", "lekh");
	@DiminAfterCons = ("l", "ele", "lekh", "elekh");
	@Adjective = ("s", "er", "e", "n");
	@Verb = ("st", "stu", "t", "n");

# patterns
	$lengthenName = "(m|n|[bdfgkmnpstvz]l|[ae]y|[iu])\\|?\$"; # CY A.3
	$lengthenInf = "(m|n[gk]?|(kh|[bdfgkmnpstvz])l|[oae]y|[iou])\\|?\$";
		# CY G.35.b

if ($] >= 5.008) {
	require encoding; import encoding  "utf8";
	require DB_File;
}

keys %fullList = 50000; # hash containing all expansions
my ($utf8) = 0;  # set to 1 in initialization if output is in utf8

sub emit { # a morphological form has been generated
	local ($theWord) = @_;
	if ($theWord =~ / /) {
		print STDERR "BAD: [$theWord]\n"
	}
	if (defined($suppress{$theWord})) {
		# print "suppressing $theWord\n";
		return;
	}
	if ($theWord eq "\n") {
		print "\n" unless $#ARGV >= 1;
		return;
	}
	my $toPrint = $theWord;
	$toPrint =~ s/^\||\|$//g; # remove | fore and aft
	$toPrint = $spells{$theWord} if ($utf8 && defined($spells{$theWord}));
	if ($allArgs =~ /--immediate\b/) {
		print $toPrint;
		print " [$definition]" unless $definition eq "";
		print " ";
	}
	if (defined($fullList{$toPrint})) {
		$fullList{$toPrint} .= ", $definition"
			unless ($fullList{$toPrint} =~ /\b\Q$definition\E\b/);
	} else {
		$fullList{$toPrint} = $definition;
	}
	if ($allArgs =~ /--(utf8Filter|uDefFilter)\b/i) {
		$internal = Internal_to_UTF(Yivo_2_Internal($toPrint));
		if (defined($knownUtf8{$internal})) {
			$knownUtf8{$internal} .= ", $definition"
				unless $knownUtf8{$internal} =~ /\b\Q$definition\E\b/;
		} else {
			$knownUtf8{$internal} = $definition;
			# print "I know [$internal]\n";
		}
	}
	return($theWord);
} # emit

sub doAdjective {
	local ($theBase) = @_;
	local($base);
	$theBase = $spells{$theBase} if ($utf8 && defined($spells{$theBase}));
	foreach $suffix (@Adjective) {
		if ($theBase =~ /(or|oyr|[^aeiouyr|])n$/ && $suffix =~ /^(e|n)/) {
			# eg: gevezn + er => gevezener; gevezn + n => gevezenem
			# CY C.13: supporting (epenthetic) "e".
			# I added case r for modern + er => moderner, not moderener
			$theBase =~ /(.*)n$/;
			$base = $1 . "en";
		} else {
			$base = $theBase;
		}
		if ($suffix eq "n") {
			if ($base =~ /(.*)e$/) {
				# eg: rozeve + n => rozevn
				doAddN($1, "(oy|i|m|m\\w)\$", "(ne|ay|n)\$");
			} else {
				# CY C.11
				doAddN($base, "([oae]y|[ioum])\$", "(^nay|ne|n)\$");
			}
			# eg: bloyen, frien; yenem, nayem
		} elsif ($suffix eq "s" && $base =~ /(s|st)$/) {
			# CY C.12 (covers "s", not "st")
			# eg: groys + s, alt|st + s
		} elsif ($suffix eq "s" && $base =~ /t$/) {
			# eg: geshvitst|s, gezukhts
			emit $base . "|" . $suffix;
		} elsif ($suffix =~ "^e" && $base =~ /e$/) {
			# eg: azoyne + er
			# ignore an extra -e ending
			$shortbase = $base; # make a copy
			$shortbase =~ s/e$//;
			emit $shortbase . $suffix;
		} elsif ($suffix eq "s" && $base =~ /(sh|i)$/) {
			emit $base . "es";
			# eg: englishes, fries
		} else {
			emit $base . $suffix;
		}
	}
} # doAdjective

sub doAddN {
	local ($base, $forEn, $forEm) = @_;
	$base = $spells{$base} if ($utf8 && defined($spells{$base}));
	$shortbase = $base; # make a copy
	$shortbase =~ s/e$//; # remove any final e, but not from "ze"
	# eg: yene + em => yenem
	if ($forEm ne "" && $base =~ /$forEm/) {
		return(emit $shortbase . "em");
	} elsif ($base =~ /$forEn/) {
		return(emit $shortbase . "en");
	} else {
		return(emit $base . "n");
	}
} # do AddN

sub joinUp { # join two strings, separating with | if necessary
	my ($first, $second, $shtum) = @_;
	$first = $spells{$first} if ($utf8 && defined($spells{$first}));
	$second = $spells{$second} if ($utf8 && defined($spells{$second}));
	# print "joining $first to $second\n";
	if  ("$first\01$second" =~ /s\01h|k\01h|t\01s|z\01h|e\01y/) {
		return($first . '|' . $second);
	} elsif ($shtum and $second =~ /^(oy|ey|ay|i)/) {
		return($first . '#' . $second); # shtumer alef
	} else {
		return($first . $second);
	}
} # joinUp

sub ToUTF {
	local($left, $right) = @_;
	$mychar = ord($left) * 256 + ord($right);
	if ($mychar <= 0x7F) { # 7 sig bits; plain 7-bit ascii
		return chr $mychar;
	} elsif ($mychar <= 0x7FF) { # 11 sig bits; Hebrew is in this range */
		$first = chr(0300 | ($mychar >> 6)&037);
		$second = chr(0200 | $mychar & 077);
		return "$first$second";
	} elsif ($mychar <= 0xFFFF) { # 16 sig bits */
		$first = chr(0340 | (($mychar >> 12)&017));
		$second = chr(0200 | (($mychar >> 6)&077));
		$third = chr(0200 | ($mychar & 077));
		return "$first$second$third";
	}
} # ToUTF

sub Internal_to_UTF {
	my ($text) = @_;
	use utf8;
	$text =~ tr{#bgdhwzHtyXxlMmNnseFPCckrSTOv-ʼ}
	           {אבגדהוזחטיךכלםמןנסעףפץצקרשתױװ־'};
	$text =~ s/a/אַ/g;
	$text =~ s/B/בֿ/g;
	$text =~ s/K/כּ/g;
	$text =~ s/f/פֿ/g;
	$text =~ s/p/פּ/g;
	$text =~ s/Q/שׂ/g;
	$text =~ s/W/תּ/g;
	$text =~ s/I/ײַ/g;
	$text =~ s/A/ײ/g;
	$text =~ s/i/יִ/g;
	$text =~ s/o/אָ/g;
	$text =~ s/u/וּ/g;
	return $text;
	no utf8;
} # Internal_to_UTF

sub Yivo_2_Internal {
	my ($text) = @_;
	@respelled = ();
	foreach $part (split /\|/, $text) {
		# print "part $part, spelled $spells{$part}\n";
		push @respelled, defined($spells{$part}) ? $spells{$part} : $part;
	}
	$text = join("|", @respelled);
	$text =~ s/\bfar(ey|in|ib|um)/far#$1/g; # "fareynik"
	$text =~ s/\bfarur/far#ur/g; # ur has to have a shtumer alef here
	$text =~ s/\bur(ey)/ur#$1/g; # "ureynikl"
		# maybe need a rule for geiblt, which also needs a shtumer alef?
    $text =~ s/ay/I/g; # pasakh tsvey-yud
    $text =~ s/ey/A/g; # tsvey-yud
	$text =~ s/oy/O/g; # vov-yud
	$text =~ s/u/w/g; # vov
	$text =~ s/iy/Jy/g; # khirik-yud yud
	$text =~ s/vw/vu/g; # tsvey-vov melupm-vov
	$text =~ s/wv/uv/g; # melupm-vov tsvey-vov
	$text =~ s/ww/uw/g; # melupm-vov vov
	$text =~ s/wi/wJ/g; # vov khirik-yud
	$text =~ s/(^|[^\w#~|]|_)([OAIiuw])/$1#$2/g; # shtumer alef
	$text =~ s/ie/Je/g; # khirik-yud ayin
	$text =~ s/ii/JJ/g; # khirik-yud khirik-yud
	$text =~ s/iw/Jw/g; # khirik-yud vov
	$text =~ s/([aAeIoOu])i/$1J/g; # vowel khirik-yud
	$text =~ s/i([aAeIoOu])/J$1/g; # khirik-yud vowel
	$text =~ s/i/y/g; # yod
	$text =~ s/J/i/g; # khirik-yud
	$text =~ s/yy/|yi/g; # yud khirik-yud
	$text =~ s/kh/x/g; # khof
	$text =~ s/tsh/tS/g; # tes shin
	$text =~ s/ts/c/g; # tsadik
	$text =~ s/(\w|[#'"\x80-\x91]|\|)([nxfmc])($|[^\w^#~|\x80-\x91]|_)/$1\u$2$3/g; # final letters
		# must be at least a 2-letter word 
	$text =~ s/(\w[kh])It(N?)\b/$1At$2/g; # xxxkayt => xxxkeyt
	$text =~ s/sh/S/g; # shin
	$text =~ s/zh/zS/g; # zayin shin
	$text =~ s/dj/dzS/g; # daled zayin shin
	$text =~ s/\|//g; # remove prophylactic  |
	$text =~ s/J/i/g; # explicit pintl-yud
	$text =~ s/U/u/g; # explicit melupm-vov
	$text =~ s/V/w/g; # explicit single vov
	return($text);
} # Yivo_2_Internal

sub getData {
	if ($allArgs =~ /--storedDefs\b/) {
		# print "Using stored defs.\n";
		# tie %i2y, 'DB_File', 'i2y.data', O_RDWR|O_CREAT, 0664, $DB_HASH;
		tie %spells, 'DB_File', 'spells.data', O_RDWR|O_CREAT, 0664, $DB_HASH;
		tie %fullList, 'DB_File', 'defs.data', O_RDWR|O_CREAT, 0664, $DB_HASH;
		return;
	}
	my ($fileName);
	$allArgs =~ /--defsFile ([-\w.\/]+)/ or usage();
	$fileName = $1;
	if ($fileName eq "-") {
		$DATAFILE = *STDIN;
	} else {
		open INPUT, $fileName || die "Can't open $fileName; stopping";
		$DATAFILE = *INPUT;
	}
	while (defined($line = <$DATAFILE>)) { # one line of dictionary input
		$definition = "" unless $line =~ /^\s/;
		# $line =~ s/\[[^]]*\]//g; # remove definitions, spelling annotations
		$line =~ s/%.*//; # remove comments
		next if ($line =~ /^\s*$/); # blank line; ignore
		# print "line: $line";
		@parts = ();
		while ($line =~ s/^\s*([-\w#{}|\/'ʼ"]+|\[[^]]*\])//) {
			# one unit or signif. cmnt
			push @parts, $1;
		}
		# print "\tparts: ", join(', ', @parts), "\n";
		@newparts = ();
		$localdef = undef;
		foreach $unit (reverse @parts) { # backwards to get defs first
			$unit =~ s/'//g; # remove accent; we ignore it
			if ($unit =~ /\[/) { # significant comment
				# print "significant: $unit\n";
				next unless $unit =~ /def: ([^]]*)\]/;
				push @newparts, $localdef if defined($localdef);
				$localdef = $unit;
			} else {
				push @newparts, $unit;
			}
		} # backwards sweep through @parts
		push @newparts, $localdef if defined($localdef);
		@parts = reverse @newparts;
		# print "after shifting: ", join(", ", @parts), "\n";
		foreach $unit (@parts) { # one unit
			# print "\tunit: $unit\n";
			if ($unit =~ /\[/) { # significant comment
				next unless $unit =~ /def: ([^]]*)\]/;
				$definition = $1;
				# print "\tnew def: $localdef\n";
				next;
			}
			$unit =~ /([-\w|{}#]*)(.*)/; 
			$rest = $2;
			# print "base is $1 rest is $2\n";
			if ($1 ne "") { # new base
				$base = removeBraces($1);
				$participle = "";
				emit $base;
			} # new base
			$base = "" unless defined($base); # for very start of file with /
			while ( $rest =~ s/^\/([-\w|{}#]+)//) { # one code
				$code = $1;
				# print "code $code\n";
				decode: {
					if ($code =~ "^E") { # ending
						$code =~ /E(.+)/;
						$suffix = $1;
						emit (joinUp($base, $suffix, 0));
						last decode;
					} elsif ($code =~ /^X(.*)/) { # plural form
						$pluralForm = removeBraces($1);
						if ($pluralForm ne "") {
							# print STDERR "Plural: $base -> $pluralForm\n";
							emit ($pluralForm);
						} else {
							# print STDERR "No plural form for $base\n";
						}
						last decode;
					} elsif ($code =~ "^B") { # participle form
						$code =~ /B(.+)/;
						$participle = removeBraces($1);
						emit $participle;
						doAdjective($participle);
						last decode;
					} elsif ($code eq "H") { # Hebrew
						$base = '|' . $base . '|';
						last decode;
					} elsif ($code =~ /^D/) { # diminutives
						if ($code =~ /^D(.+)/) {
							$base = removeBraces($1);
						}
						if ($base =~ /[aeiou](\|)?$/) {
							foreach $suffix (@DiminAfterVowel) {
								emit (joinUp($base, $suffix, 0));
							}
						} else {
							foreach $suffix (@DiminAfterCons) {
								emit (joinUp($base, $suffix, 0));
							}
						}
						last decode;
					} elsif ($code eq "C") { # name
						if ($base =~ /^(tat|zeyd)e$/) { # special dative
							# CY A.4
							$root = $1;
							&doAddN($root, "", ""); # dative
						} else { # ordinary dative
							&doAddN($base, $lengthenName, "");
						}
						if ($base =~ /(s|ts|sh|z)$/) {
							# CY A.5
							emit (joinUp($base, 'es', 0));
						} else { # unless -s, last syllable accent: use '
							emit (joinUp($base, 's', 0));
						}
						last decode;
					} elsif ($code eq "J") { # -ns
						# possessive form
						if ($base =~ "(.*)e\$") {
							# eg: tatn
							$root = $1;
							if ($root =~ /(ndl|n|ng|oy|ay|m|m\w|nk)$/) {
								# eg: mamens
								emit (joinUp($root, "ens", 0)); 
							} else {
								# eg: tatns
								emit (joinUp($root, "ns", 0)); 
							}
						} else {
							if ($base =~ /(ndl|n|ng|oy|ay|m|m\w|nk)\$/) {
								emit (joinUp($base, "ens", 0)); 
							} else {
								# eg: yidns
								emit (joinUp($base, "ns", 0)); 
							}
							# eg: Sheyndl + n => Sheyndlen
						}
						last decode;
					} elsif ($code =~ "^A(.*)") { # adjective
						if ($1 ne "") {
							&doAdjective($1);
						} else {
							&doAdjective($base);
						}
						last decode;
					} elsif ($code =~ /^K(.*)/) { # pos, comparative, superlative
						if ($1 eq "") { # standard, eg: flink, flinker, flinkster
							$comparative = $base;
						} else { # exceptional, eg: gut, beser, bester
							$comparative = $1;
							emit (joinUp($comparative, "er", 0));
						}		
						doAdjective($base);
						# comparative: CY C.14.a
						doAdjective($comparative . "er");
						# superlative: CY C.14.b; requires case, gender ending
						if ($comparative =~ /s$/) {
							# eg: zis + ster => zister; CY C.14.b
							doAdjective($comparative . "t");
						} else {
							doAdjective(joinUp($comparative, "st", 0));
						}
						last decode;
					} elsif ($code =~ /I(.*)/) { # word taking -ik; explicit option
						if ($1 ne "") {
							$ending = $1;
						} elsif ($base =~ /(l|kht|d|ns|er|ts|tst)$/) {
							# eg: amoliker
							$ending = "ik";
						} elsif ($base =~ /(sh|es|em|l|kh|n)$/) {
							# eg: khoydesh, shabes, sholem, shul, peysekh,
							$ending = "dik";
						} else {
							$ending = "edik";
						}
						$most = emit (joinUp($base, $ending, 0));
						doAdjective($most);
						last decode;
					} elsif ($code =~ /^V(.*)/) { # verb, optional infinitive
						$infinitive = removeBraces($1);
						if ($1 ne "") { # explicit infinitive
							emit($infinitive);
						}
						# print "setting infinitive to $infinitive\n";
						foreach $suffix (@Verb) {
							# CY G.35
							if ($suffix eq "n") {
								local($newBase) = $base;
								if ($base =~ /(.*[^aeiouy])n$/) {
									# eg: efn + n => efenen
									# epenthetic e
									$newBase = $1 . "en";
								}
								if ($infinitive eq "") {
									$infinitive =
										doAddN($newBase, $lengthenInf, "");
									# print "setting infinitive to $infinitive\n";
								} else {
									doAddN($newBase, $lengthenInf, "");
								}
								# College Yiddish p. 55
							} elsif ($base =~ /s$/ && $suffix =~ /^s/) {
								# CY G.35.e: fold s with s
								$new = $suffix; # must copy; $suffix is in @Verb
								$new =~ s/^s//;
								emit (joinUp($base, $new, 0));
							} elsif ($base =~ /t[|]?$/ && $suffix =~ /^t/) {
								# CY G.35.d: fold t with t
								$new = $suffix; # must copy; $suffix is in @Verb
								$new =~ s/^t//;
								emit (joinUp($base, $new, 0));
							} else {
								emit (joinUp($base, $suffix, 0));
							}
						}
						$most = joinUp($infinitive, "dik", "", 0);
						emit($most);
						doAdjective($most);
						last decode;
					} elsif ($code eq "T") { # verb in ge .. t
						local($newBase);
						if ($base =~ /^ge(.*)/) {
							# eg: gefel
							$newBase = $1;
						} else {
							$newBase = $base;
						}
						if ($newBase =~ /t[|]?$/) {
							$participle = joinUp('ge', $newBase, 1);
							emit $participle;
							doAdjective($participle);
						} else {
							$participle = joinUp('ge',
								 joinUp($newBase, 't', 0), 1);
							emit $participle;
							doAdjective($participle);
						}
						last decode;
					} elsif ($code =~ /^L(.+)/ ) { # prefix
						$prefix = $1;
						emit (joinUp($prefix, $base, 1));
						last decode;
					} elsif ($code =~ "^G") { # separable prefix on past participle
						$code =~ /G(.*)/;
						$prefix = $1;
						# print "infinitive $infinitive \n";
						emit(joinUp(joinUp($prefix, "tsu", 0), $infinitive, 1));
						$most = emit(joinUp($prefix, $infinitive, 1));
						emit($most . "dik");
						emit $prefix . $participle;
						doAdjective($most . "dik");
						doAdjective($prefix . $participle);
						last decode;
					} elsif ($code eq "N") { # plural in -n
						doAddN($base, $lengthenInf, "");
						# eg: eplen, zunen, tsaytungen, froyen, bekerayen
						last decode;
					} elsif ($code eq "P") { # preposition
						# CY B.8
						if ($base =~ /n$/) {
							if ($base =~ /(.*[^aeiouy])([nl])$/) {
								# eg: tsvishn + m => tsvishenem
								emit $1 . "e" . $2 . "em"
							} else {
								emit $base . "em"
							}
						} elsif ($base =~ /(u|y)$/) {
							# eg: tsum, baym
							emit $base . "m"
						} else {
							emit $base . "n"
						}
						last decode;
					} elsif ($code eq "S") { # plural in -s
						if ($base =~ /(s|sh)$/) {
							emit $base . "es"
						} else {
							emit joinUp($base, 's', 0);
						}
						last decode;
					} elsif ($code =~ /^-(.*)/) { # form to suppress
						$base = $1;
						$suppress{$base} = 1;
						# print "Will suppress $base\n";
					} else {
						print STDERR "Cannot decode $code\n";
					}
				} # decode loop
			} # deal with one code
		} # one unit
		print "\n" if ($allArgs =~ /--immediate\b/);
	} # foreach line of dictionary input
	# $stat = "" . %fullList;
	# print "done; hash has $stat\n";
	if ($allArgs =~ /--saveDefs\b/) {
		print "about to save defs\n";
		unlink 'spells.data', 'defs.data';
		# tie %i2y, 'DB_File', 'i2y.data', O_RDWR|O_CREAT, 0664, $DB_HASH;
		tie %saveSpells, 'DB_File', 'spells.data', O_RDWR|O_CREAT, 0664, $DB_HASH;
		tie %saveFullList, 'DB_File', 'defs.data', O_RDWR|O_CREAT, 0664, $DB_HASH;
		%saveSpells = %spells;
		%saveFullList = %fullList;
		print "finished saving defs\n";
	}
} # getData

sub removeBraces {
	my ($unit) = @_;
	my ($romanized, $respelling, $header);
	@results = ();
	for $piece (split /}/, $unit) { # one piece
		if ($piece =~ /(.*){([^}]*)/) { # spelling rule
			# print STDERR "working on $piece}:";
			$romanized = $1;
			$respelling = $2;
			$header = "";
			if ($romanized =~ /(.*)\|([^|]*)/) {
				$header = $1;
				$romanized = $2;
				# print STDERR "($header , $romanized) ";
			}
			$romanized =~ s/^-//g;
			if (defined($spells{$romanized}) &&
					$spells{$romanized} ne $respelling) {
				print STDERR "$romanized is spelled both as ",
					$spells{$romanized}, " and as $respelling\n";
			}
			$spells{$romanized} = $respelling;
			# print STDERR "I see that you spell $romanized $respelling\n";
			push @results, $header . $romanized;
			# print STDERR " got ", $header . $respelling, "\n";
		} else {
			push @results, $piece;
		}
	} # one piece
	return (join("", @results));
} # removeBraces

sub doWork {
	return if ($allArgs =~ /--saveDefs/);
	if ($allArgs =~ /--text (-|\S+)/) {
		$fileName = $1;
		if ($fileName eq "-") {
			$TEXTFILE = *STDIN;
		} else {
			open INPUT, $fileName || die "Can't open $fileName; stopping";
			$TEXTFILE = *INPUT;
		}
	} else {
		$TEXTFILE = *STDIN;
	};
	if ($allArgs =~ /--wordList\b/) {
		# print "All words\n";
		while (($form,undef) = each %fullList) {
			print "$form\n" unless defined($suppress{$form});
		}
	};
	if ($allArgs =~ /--defList\b/) {
		while (($form,$def) = each %fullList) {
			print "$form [$def]\n"
				unless defined($suppress{$form}) or $def eq "";
		}
	}
	if ($allArgs =~ /--uDefList\b/) {
		while (($form,$def) = each %fullList) {
			next if $suppress{$form};
			next if $def eq "";
			$internal = Yivo_2_Internal($form);
			print Internal_to_UTF($internal), "\t", $def, "\n";
		}
	}
	if ($allArgs =~ /--wordFilter\b/) {
		while (defined($line = <$TEXTFILE>)) { # one line of text input
			if ($line =~ /\\/) {next;} # don't spellcheck TeX lines
			while ($line =~ s/([\w|#]+)//) { # one word of input
				$word = $1;
				if (!defined $fullList{$word} or defined($suppress{$word})) {
					print $word , "\n"; # wasn't found
				}
			} # one word of input
		} # one line of text input
	}
	if ($allArgs =~ /--utf8Filter\b/i) {
=head1
		while (($form,undef) = each %fullList) {
			next if $suppress{$form};
			$knownUtf8{Internal_to_UTF(Yivo_2_Internal($form))} = $definition;
			print "I know [", Internal_to_UTF(Yivo_2_Internal($form)), "]\n";
		}
=cut
		# print "known: ", join(", ", keys %knownUtf8), "\n";
		while (defined($line = <$TEXTFILE>)) { # one line of text input
			use utf8;
			$line =~ s/[][.,?:;*'"!()״„“־׳-]/ /g;
			# print "line is $line\n";
			foreach $word (split /\s+/, $line) {
				# one word of input
				next if ($word =~ /^\d+$/);
				next if ($word =~ /^$/);
				if (!defined($knownUtf8{$word})) {
					print "$word", "\n";
					$knownUtf8{$word} = 2;
					# print "known: ", join(", ", keys %knownUtf8), "\n";
				} else {
					# print "I knew $word\n";
				}
			} # one word of input
			no utf8;
		} # one line of text input
	}
	if ($allArgs =~ /--defFilter\b/) {
		while (defined($line = <$TEXTFILE>)) { # one line
			$line =~ s/([\w|#]+)/defined($fullList{$1}) ? "$1 ($fullList{$1})" : "$1"/eg;
			print $line;
		}
	}
	if ($allArgs =~ /--uDefFilter\b/) {
		# print "doing uDefFilter\n";
		$alphabet = "אבגדהוזחטיךכלםמןנסעףפץצקרשתױװאַבֿכּפֿפּשׂתּײַײיִאָוּ";
		while (defined($line = <$TEXTFILE>)) { # one line
			$line =~ s/([$alphabet]+)/defined($knownUtf8{$1}) ? "$1 \/$knownUtf8{$1}\/" : "$1"/eg;
			print $line;
		}
	}
	if ($allArgs =~ /--respellList\b/) {
		print "Alphabetical.  ",
			"Use internal form for RHS (without final forms).\n";
		print "All respelled Hebrew words\n";
		foreach $spell (sort (keys %spells)) {
			print "$spell $spells{$spell}\n";
		}
	}
	if ($allArgs =~ /--utf8List\b/) {
		while (($form,undef) = each %fullList) {
			# print Yivo_2_Internal($form), ": ";
			next if $suppress{$form};
			$internal = Yivo_2_Internal($form);
			# $i2y{$internal} = $form;
			print Internal_to_UTF($internal), "\n";
		}
	}
	if ($allArgs =~ /--uRhymeList\b/) {
		@words = ();
		while (($form,undef) = each %fullList) {
			# print Yivo_2_Internal($form), ": ";
			next if $suppress{$form};
			push @words, scalar reverse($form);
			# $internal = Yivo_2_Internal($form);
			# $i2y{$internal} = $form;
			# push @words, myReverse(Internal_to_UTF($internal));
		}
		foreach $word (sort @words) {
			# print myReverse($word), "\n";
			print scalar reverse($word), "\t",
				Internal_to_UTF(Yivo_2_Internal(scalar reverse($word))), "\n";
		}
	}
} # doWork

sub usage {
	print STDERR
		"Usage: $0\n",
		"\t[--defsFile defs-file | --storedDefs]\n" ,
		"\t[--saveDefs]\n",
		"\t[--wordList | --immediate]\n",
		"\t[--utf8List]\n",
		"\t[--uRhymeList]\n",
		"\t[--respellList]\n",
		"\t[--defList]\n",
		"\t[--uDefList]\n",
		"\t[--text text-file]\n" ,
		"\t[--wordFilter]\n",
		"\t[--utf8Filter]\n",
		"\t[--defFilter]\n",
		"\t[--uDefFilter]\n",
		;
	exit(1);
} # usage

sub myReverse {
	# reverse utf-8 with composing characters together with base characters
	my ($word) = @_;
	# print "[$word]\n";
	if ($word =~ /^((\p{Ll}|\p{Lo}|\p{Lu}|\p{Lt})\p{Mn}*)(.*)((\p{Ll}|\p{Lo}|\p{Lu}|\p{Lt})\p{Mn}*)$/) {
		my ($first, $middle, $last) = ($1, $3, $4);
		$word = $last . myReverse($middle) . $first;
	} else {
		# print "not found\n";
	}
	# print "<$word>\n";
	return($word);
} # myReverse

# main program here
usage() if ($#ARGV == -1);
$allArgs = join(' ', @ARGV);
$utf8 = ($allArgs =~ /utf8/);
getData();
doWork();
# untie %i2y;
untie %spells;
untie %fullList;
